/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Terry Fong, Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://imtsg14.epfl.ch/tlib>
//
/////////////////////////////////////////////////////////////////////////////////


#include <iostream.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Intrinsic.h>
#include <X11/extensions/XShm.h>
#include <sys/shm.h>
#ifdef __cplusplus
}
#endif /* __cplusplus */

#include "tlDisplay_x11.h"

// uncomment the following to use shared-memory X11 display. This runs
// faster than regular X11 display, but will not work across the network 
// (which is not recommended anyway for real-time work...)
#define USE_SHM

////////////////////////////////////////////////////////////////////////
// module globals
////////////////////////////////////////////////////////////////////////

// globals
Display*   _tlDisplay_x11_theDisplay = NULL;
char       _tlDisplay_x11_list[TL_MAX_DISPLAY];
Window     _tlDisplay_x11_window[TL_MAX_DISPLAY];
tlDisplay* _tlDisplay_x11_ptlDisplay[TL_MAX_DISPLAY];

// private globals
static Colormap         colormap = (Colormap) NULL;
static Visual*          visual = NULL;
static int              fieldmode = 0x0;
static int              depth;
static GC               gc = (GC) NULL;
static int              _tlDisplay_init = 0;
static XImage*          xImage[TL_MAX_DISPLAY];
static XShmSegmentInfo* shminfo[TL_MAX_DISPLAY];


////////////////////////////////////////////////////////////////////////
// non-class methods
////////////////////////////////////////////////////////////////////////

static int
_tlDisplayImageCreate (int width, int height, int winID)
{
  int size;

#ifdef USE_SHM
  /*
   * setup MIT-SHM X server extension
   */
  if (!shminfo[winID]) {
    if (NULL == (shminfo[winID] = (XShmSegmentInfo *) 
		 malloc (sizeof(XShmSegmentInfo)))) {
      TL_ERROR_MSG("tlDisplay::create: malloc error");
      return (TL_ERROR);		/* failure */
    }
  }

  if (!xImage[winID]) {
    /* create a shared memory image */
    if (NULL == (xImage[winID] = XShmCreateImage (_tlDisplay_x11_theDisplay, 
						  visual, depth, ZPixmap, 
						  NULL, shminfo[winID], 
						  width, height))) {
      TL_ERROR_MSG("tlDisplay::create: XShmCreateImage error");
      return (TL_ERROR);		/* failure */
    }
  }

  /* get the shared memory image id */
  size = (xImage[winID])->bytes_per_line * (xImage[winID])->height;

  if (((shminfo[winID])->shmid = 
       shmget (IPC_PRIVATE, size, IPC_CREAT|0777)) == -1) {
    TL_ERROR_MSG("tlDisplay::create: shmget error");
    return (TL_ERROR);		/* failure */
  }
    
  /* attach to the shared memory image */
  if (((shminfo[winID])->shmaddr = (char *) 
       shmat((shminfo[winID])->shmid, (void *)0, 0)) 
      == (char *)(-1)) {
    TL_ERROR_MSG("tlDisplay::create: shmat error");
    return (TL_ERROR);		/* failure */
  }
    
  /* point to the shared memory buffer */
  (xImage[winID])->data = (shminfo[winID])->shmaddr;
    
  /* attach the server to the shared memory segment */
  if (FALSE == XShmAttach (_tlDisplay_x11_theDisplay, shminfo[winID])) {
    TL_ERROR_MSG("tlDisplay::create: unable to attach shared memory\n");
    return (TL_ERROR);		/* failure */
  }

#else  

  /*
   * setup regular X image
   */
  if (!xImage[winID]) {
    char *data;
    
    /* create the X image */
    if (NULL == (xImage[winID] = XCreateImage (_tlDisplay_x11_theDisplay, 
					       visual, depth, ZPixmap, 0,
					       NULL, width, height, 32, 0))) {
      TL_ERROR_MSG ("tlDisplay::create: XCreateImage error");
      return (TL_ERROR);		/* failure */
    }

    size = (xImage[winID])->bytes_per_line * (xImage[winID])->height;

    /* allocate storage */
    if (NULL == (data = (char*)malloc (size))) {
      TL_ERROR_MSG ("tlDisplay::create: unable to allocate X image storage\n");
      return (TL_ERROR);		/* failure */
    }

    (xImage[winID])->data = data;
  }
#endif
    
  XSync (_tlDisplay_x11_theDisplay, False);	/* force server update */

  return(TL_NO_ERROR);

}


static int
_tlDisplayImageDestroy ( int winID )
{
  int status=0;
  
#ifdef USE_SHM
  struct shmid_ds dummy;

  if (shminfo[winID]) {
    /* detach the server from the shared memory segment */
    if (FALSE == XShmDetach (_tlDisplay_x11_theDisplay, shminfo[winID])) {
      TL_ERROR_MSG("tlDisplay::Close: unable to detach shared memory");
      status = -1;
    }
    
    /* detach from the shared memory image */
    if (shmdt ((shminfo[winID])->shmaddr) < 0) {
      TL_ERROR_MSG("tlDisplay::Close: shmdt");
      status = -1;		/* failure */
    }
    
    /* delete shared memory segment */
    if (shmctl ((shminfo[winID])->shmid, IPC_RMID, &dummy) < 0) {
      TL_ERROR_MSG("tlDisplay::Close: shmctl");
      status = -1;		/* failure */
    }
    
    if(shminfo[winID]) {
      free(shminfo[winID]);
      shminfo[winID]=NULL;
    }
  }
#endif
  
  if (xImage[winID]) {
    if (FALSE == XDestroyImage (xImage[winID])) {
      TL_ERROR_MSG("tlDisplay::Close: unable to destroy image\n");
      status = -1;
    }
  }
  
  /* note: do not free xImage since XDestroyImage has already
   * freed xImage. Freeing xImage twice will cause heap corruption under
   * Linux (see the malloc man page)
   */
  xImage[winID] = NULL;
  
  return (status);
}


/*
 * RGB bits (from xdpyinfo)
 *
 * TrueColor-24 masks
 *   blue  0x0000ff = 00000000 00000000 11111111
 *   green 0x00ff00 = 00000000 11111111 00000000
 *   red   0xff0000 = 11111111 00000000 00000000
 * pixel data (MSB LSB): rrrrrrrr gggggggg bbbbbbbb
 *
 * TrueColor-16 masks
 *   blue  0x001f = 00000000 00011111
 *   green 0x07e0 = 00000111 11100000
 *   red   0xf800 = 11111000 00000000
 * pixel data (MSB LSB): rrrrrggg gggbbbbb
 * LSB: bits 2-4 green and upper 5 bits of blue 
 * MSB: upper 5 bits of red and upper 3 bits of green
 *
 * TrueColor-8 masks
 *   blue  0xc0 = 11000000
 *   green 0x38 = 00111000
 *   red   0x7  = 00000111
 * pixel data (msb lsb): bbgggrrr 
 */
#define _TL_DISPLAY_PUT_PIXEL(r,g,b,x,y,winID) \
{ \
  unsigned long pixel; \
  if (depth==32) pixel = ((r<<16) | (g<<8) | (b<<0) | 0xff000000); \
  else if (depth==24) pixel = ((r<<16) | (g<<8) | b); \
  else if (depth==16) { \
    pixel = (((g & 0x1c) << 3) | ((b & 0xf8) >> 3)); \
    pixel |= (((r & 0xf8) | ((g & 0xe0) >> 5)) << 8); \
  } \
  else if (depth==8) pixel=((b&0xc0) | ((g&0xe0)>>2) |((r&0xe0)>>5)); \
  else return (TL_ERROR); \
  XPutPixel (xImage[winID], x, y, (unsigned long) pixel); \
}


int
_tlDisplayImagePaint (int winID)
{
  // shortcut
  int width  = (xImage[winID])->width;
  int height = (xImage[winID])->height;

#ifdef USE_SHM
  XShmPutImage (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[winID], 
		gc, xImage[winID], 0, 0, 0, 0, width, height, False);
  XSync (_tlDisplay_x11_theDisplay, False);	/* force server update */
#else
  XPutImage (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[winID], 
	     gc, xImage[winID], 0, 0, 0, 0, width, height);
#endif

  return (TL_NO_ERROR);
}


static void
cleanup (void)
{
  if (colormap != (Colormap) NULL) {
    XFreeColormap (_tlDisplay_x11_theDisplay, colormap);
  }

  if (gc != (GC) NULL) {
    XFreeGC (_tlDisplay_x11_theDisplay, gc);
  }
}


/**
 ***********************************************************************
 *  
 *  Constructor;
 *  create default display.
 *
 ***********************************************************************/
tlDisplay::tlDisplay  ()
{
  init (TL_DISPLAY_NAME, 0, 0);
}


/**
 ***********************************************************************
 *  
 *  Constructor
 *
 *  @param name        window name (displayed in title bar)
 *
 ***********************************************************************/
tlDisplay::tlDisplay  (char *name)
{
  init (name, 0, 0);
}


/**
 ***********************************************************************
 *  
 *  Constructor;
 *  specifies window position on the screen.
 *
 *  @param x       initial window X position  [pix]
 *  @param y       initial window Y position  [pix]
 *
 ***********************************************************************/
tlDisplay::tlDisplay  (int x, 
		       int y)
{
  init (TL_DISPLAY_NAME, x, y);
} 


/**
 ***********************************************************************
 *  
 *  Constructor;
 *  specifies both position and window name.
 *
 *  @param name    window name (displayed in title bar)
 *  @param x       initial window X position  [pix]
 *  @param y       initial window Y position  [pix]
 *
 ***********************************************************************/
tlDisplay::tlDisplay (char *name, 
		      int  x, 
		      int  y)
{
  init (name, x, y);
}


/**
 ***********************************************************************
 *  
 *  ~tlDisplay --
 *    TL library display & event management 
 *    destructor
 *
 ***********************************************************************/
tlDisplay::~tlDisplay ()
{
  close ();
  if (this->name) {
    free (this->name);
  }
}


/**
 ***********************************************************************
 *
 *  Window creation and registration
 *
 *  @param name    window name (displayed in title bar)
 *  @param x       initial window X position  [pix]
 *  @param y       initial window Y position  [pix]
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::init (char *name, int x, int y)
{
  int         k;
  XGCValues   gcvalues;
  XVisualInfo visualTemplate;
  XVisualInfo *visualList; 
  int         numVisualMatched;
  int         screen;
  unsigned long        valuemask;
  XSetWindowAttributes attributes;


  // store parameters
  this->x      = x;
  this->y      = y;
  this->width  = TL_DEFAULT_COLS;
  this->height = TL_DEFAULT_ROWS;
  this->name   = strdup (name);

  // verbose
  if (tlVerbose > 1) {
    printf ("opening XWindow display...\n");
  }

  // if this is the first time
  if (_tlDisplay_init == 0) {
    if (tlVerbose > 1) {
      printf ("first time display - initializing...\n");
    }

    if (!XInitThreads() ||
	(True != XtToolkitThreadInitialize())) {
      TL_ERROR_MSG ("tlDisplay::create: no thread support\n");
      return (TL_ERROR);		/* failure */
    }
    
    // initialize device list
    for (k=0; k<TL_MAX_DISPLAY; k++) { 
      _tlDisplay_x11_list[k] = 0;
      xImage [k]    = NULL;
      shminfo[k]    = NULL;
      _tlDisplay_x11_ptlDisplay[k] = NULL;
    }
  }

  // find first available device index
  this->ID = -1;
  for (k=0; k<TL_MAX_DISPLAY; k++) {
    if (_tlDisplay_x11_list[k] == 0) {
      this->ID = k;
      _tlDisplay_x11_list[k] = 1;
      break;
    }
  }

 
  // check for availability
  if ((this->ID) == -1) {
    TL_ERROR_MSG ("tlDisplay::create: maximum number of windows exceeded");
    return (TL_ERROR);		/* failure */
  }
 

  /*
   * create a simple X window
   */


  if (!_tlDisplay_x11_theDisplay) {
    /* connect to server specified by DISPLAY environment variable */
    if (NULL == ( _tlDisplay_x11_theDisplay = XOpenDisplay (NULL)  )) {
      TL_ERROR_MSG ("tlDisplay::create: unable to open display\n");
      return (TL_ERROR);	/* failure */
    }
    
  }

  screen = DefaultScreen(_tlDisplay_x11_theDisplay);

  /*
   * find a TrueColor visual 
   */
  if (visual == NULL) {
    visualTemplate.screen = screen;
    visualTemplate.c_class = TrueColor;
    visualList = XGetVisualInfo (_tlDisplay_x11_theDisplay, 
				 VisualClassMask | VisualScreenMask,
				 &visualTemplate,
				 &numVisualMatched);
  
    if (numVisualMatched > 0) {
      /* found a TrueColor visual, use the highest depth */
      depth = 0;
      do {
	numVisualMatched--;
	if (visualList[numVisualMatched].depth > depth) {
	  depth = visualList[numVisualMatched].depth;
	  visual = visualList[numVisualMatched].visual;
	}
      } while (numVisualMatched > 0);
      printf("detected TrueColor visual (depth %d)\n",depth);
      XFree (visualList);	/* q: will this screw up "visual" ??? */
    }
    else {
      TL_ERROR_MSG ("tlDisplay::create: could not find a TrueColor visual\n");
      return (TL_ERROR);	/* failure */
    }
  }
  

  /* create a private color map for the selected visual */ 
  if (_tlDisplay_init == 0) {
    colormap = 
      XCreateColormap (_tlDisplay_x11_theDisplay, 
		       RootWindow (_tlDisplay_x11_theDisplay, screen),
		       visual, AllocNone);
  }

  /* setup window attributes */
  attributes.background_pixel = BlackPixel(_tlDisplay_x11_theDisplay, screen);
  attributes.border_pixel     = BlackPixel(_tlDisplay_x11_theDisplay, screen);
  attributes.colormap = colormap;
  valuemask = CWBackPixel | CWBorderPixel | CWColormap;

  /* create a new window */
  printf ("creating %dx%d window at (%d, %d)\n", width, height, x, y);

  int borderWidth = 2;
  _tlDisplay_x11_window[ID] = 
    XCreateWindow (_tlDisplay_x11_theDisplay, 
		   RootWindow (_tlDisplay_x11_theDisplay, 
			       screen), /* parent */
		   x, y, width, height, borderWidth, depth,
		   InputOutput, visual, valuemask, &attributes);
  
  /* set the window size & hints */
  resize (width, height);

  /* set the window title */
  XStoreName (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], name);
  XSetIconName (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], name);

  /* make the window viewable */
  //XMapWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID]);
  //visible = True; 

  /* ask for input events */
  XSelectInput (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID],
		ButtonPressMask|
		ButtonReleaseMask | 
		KeyPressMask |
		KeyReleaseMask |
		//StructureNotifyMask|
		ExposureMask  |
		DestroyNotify
		);
 
  
  gcvalues.graphics_exposures = False;
  if (_tlDisplay_init == 0) {
    gc = XCreateGC (_tlDisplay_x11_theDisplay, 
		    _tlDisplay_x11_window[ID], 
		    GCGraphicsExposures, 
		    &gcvalues);  
  } 

  /* set Image */      

  if (0 > _tlDisplayImageCreate(this->width, this->height, this->ID) ) {
    TL_ERROR_MSG ("tlDisplay::create: _tlDisplayImageCreate\n");
    return (TL_ERROR);
  }

  _tlDisplay_x11_ptlDisplay[this->ID] = this;

  if (_tlDisplay_init == 0) {
    atexit (cleanup);		// make sure we cleanup on exit...
    _tlDisplay_init = 1;
  }

  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::create: done\n");
  }

  return TL_NO_ERROR;
}



/**
 ***********************************************************************
 *
 *  Update the window position and size.
 *
 *  \return
 *  TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
inline int 
tlDisplay::update ()
{
  XWindowAttributes wattr;
  Window junkwin;
  int rx,ry;

  
  XGetWindowAttributes(_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], &wattr);
  XTranslateCoordinates (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], wattr.root, -wattr.border_width, -wattr.border_width, &rx, &ry, &junkwin);

  x = rx - wattr.x;
  y = ry - wattr.y;
  width = wattr.width;
  height = wattr.height;
    
 
  return 0;
}



/**
 ***********************************************************************
 *
 *  Reposition the window and display area.
 *
 *  \param x  new window X position
 *  \param y  new window Y position
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::setPos (int x, int y)
{
  XSizeHints sizehints;

  
  // issue warning
  if (tlVerbose > 1) {
    printf ("moving window to %dx%d\n", x, y);
  }

  // window size
  this->x = x;
  this->y = y;

  // move
  sizehints.flags = PPosition;
  sizehints.x = x;
  sizehints.y = y;
  XSetNormalHints (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], &sizehints);  
  XMoveWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], x, y);
  
  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::resize: done\n");
  }

  
  return update ();  
}



/**
 ***********************************************************************
 *
 *  Resize the window and the display area.
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::resize (int width, int height)
{
  XSizeHints sizehints;

  // issue warning
  if (tlVerbose > 1) {
    printf ("resizing window to %dx%d\n", width, height);
  }

  // window size
  this->width = width;
  this->height = height;

  // resize
  sizehints.flags = PMaxSize|PAspect|PResizeInc;
  sizehints.max_width = width;
  sizehints.max_height = height;
  sizehints.min_aspect.x = 1;
  sizehints.min_aspect.y = 2;
  sizehints.max_aspect.x = (fieldmode) ? 4 : 2;
  sizehints.max_aspect.y = 1;
  sizehints.width_inc = 8;
  sizehints.height_inc = 2;
  XSetNormalHints (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], &sizehints);
  XResizeWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], width, height);

  if (xImage[ID] != NULL) {

    if (0 > _tlDisplayImageDestroy(this->ID)) {
      TL_ERROR_MSG ("tlDisplay::resize: _tlDisplayImageDestroy\n");
      return (TL_ERROR);		/* failure */
    }

    if (0 > _tlDisplayImageCreate (width, height, this->ID)) {
      TL_ERROR_MSG ("tlDisplay::resize: _tlDisplayImageCreate\n");
      return (TL_ERROR);		/* failure */
    }
  }

  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::resize: done\n");
  }

  return update ();  
}


/**
 ***********************************************************************
 *
 *  Close and destroys the window.
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::close ()
{
    
  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::close: closing window...\n");
  }

  if (0 > _tlDisplayImageDestroy(this->ID)) {
    TL_ERROR_MSG ("tlDisplay::close: _tlDisplayImageDestroy\n");
    //return (TL_ERROR);		/* failure */
  }

  if (_tlDisplay_x11_theDisplay) {
    XUnmapWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID]);
    XDestroyWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID]);
  }

  _tlDisplay_x11_ptlDisplay[this->ID]=NULL;   
  _tlDisplay_x11_list[this->ID] = 0;

  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::close: done\n");
  }
 
  return 0;
}


/**
 ***********************************************************************
 *
 *  Make window visible.
 *
 *  \return
 *    TL_EVENT status (see tlVision.h for details)
 *
 ***********************************************************************/
int
tlDisplay::show ()
{
  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::show: activating window (%d)\n", ID);
  }

  // show window
  XMapWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID]); 
  XMoveWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID], x, y);
  XSync (_tlDisplay_x11_theDisplay, False);	/* force server update */

  visible = True;

  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::show: done\n");
  }

  return update ();
}


/**
 ***********************************************************************
 *
 *  Make window invisible.
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int
tlDisplay::hide ()
{
  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::hide: deactivating window (%d)\n", ID);
  }

  // show window
  XUnmapWindow (_tlDisplay_x11_theDisplay, _tlDisplay_x11_window[ID]);
  XSync (_tlDisplay_x11_theDisplay, False);	/* force server update */

  visible = False;

  // verbose
  if (tlVerbose > 1) {
    printf ("tlDisplay::hide: done\n");
  }

  return update ();
}


/**
 ***********************************************************************
 *
 *  Display an image in the display area.
 *
 *  @param image     the image to display
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 *  \todo
 *    make display multithreaded
 *
 ***********************************************************************/
int 
tlDisplay::display (tlPixelArray *image)
{
  if (!visible) {		// do nothing
    return TL_NO_ERROR;
  }

  // sanity check
  if (!image) {
    TL_ERROR_MSG ("tlDisplay::display: NULL image\n");
    return TL_ERROR_NULL_POINTER;
  }

  tlPixel *src = image->pixel;

  if (!src) {
    TL_ERROR_MSG ("tlDisplay::display: NULL image data\n");
    return TL_ERROR_NULL_POINTER;
  }

  // verify X window size
  if ((height != image->height) || (width != image->width)) {    
    if (0 > resize (image->width, image->height)) {
      return (TL_ERROR);		/* failure */
    }
  }
  
  // convert to a displayable RGB format
  register int x, y;
  unsigned char r, g, b;

  switch (image->format) {
  case RGB:
    if (tlVerbose > 1) {
      printf ("tlDisplay::display: RGB image\n");
    }
    assert (TL_RGB_WIDTH == 3);
    for (y=0; y<height; y++) {
      for (x=0; x<width; x++) {
	r = (unsigned char) *src++;
	g = (unsigned char) *src++;
	b = (unsigned char) *src++;
	_TL_DISPLAY_PUT_PIXEL (r, g, b, x, y, ID);
      }
    }
    break;
      
  case GRAY:
    if (tlVerbose > 1) {
      printf ("tlDisplay::display: GRAY image\n");
    }
    assert (TL_GRAY_WIDTH == 1);
    for (y=0; y<height; y++) {
      for (x=0; x<width; x++) {
	unsigned char val = (unsigned char) *src++;
	r = g = b = val;
	_TL_DISPLAY_PUT_PIXEL (r, g, b, x, y, ID);
      }
    }
    break;
      
  case HSI:
    if (tlVerbose > 1) {
      printf ("tlDisplay::display: HSI image\n");
    }
    assert (TL_HSI_WIDTH == 4);
    for (y=0; y<height; y++) {
      for (x=0; x<width; x++) {
	unsigned char val = (unsigned char) *src; // H
	r = g = b = val;
	src += TL_HSI_WIDTH;
	_TL_DISPLAY_PUT_PIXEL (r, g, b, x, y, ID);
      }
    }
    break;
      
  case BIN:
    if (tlVerbose > 1) {
      printf ("tlDisplay::display: BIN image\n");
    }
    assert (TL_BIN_WIDTH == 1);
    for (y=0; y<height; y++) {
      for (x=0; x<width; x++) {
	if (*src++ != 0) {
	  r = g = b = 255;
	} 
	else {
	  r = g = b = 0;
	}
	_TL_DISPLAY_PUT_PIXEL (r, g, b, x, y, ID);
      }
    }
    break;
      
  case NRG:
    if (tlVerbose > 1) {
      printf ("tlDisplay::display: NRG image\n");
    }
    assert (TL_NRG_WIDTH == 2);
    b = (unsigned char) 0;

    for (y=0; y<height; y++) {
      for (x=0; x<width; x++) {
	r = (unsigned char) *src++;
	g = (unsigned char) *src++;
	_TL_DISPLAY_PUT_PIXEL (r, g, b, x, y, ID);
      }
    }
    break; 
      
  case STEREO:
    if (tlVerbose > 1) {
      printf ("tlDisplay::display: STEREO image\n");
    }
    // needs scaling
    tlPixel temp;
    for (y=0; y<height; y++) {
      for (x=0; x<width; x++) {
	if ((temp = *src++) < 254 ) {
	  unsigned char val = (unsigned char) (50 + (205 * temp)/(64));
	  r = g = b = val;
	}
	else {
	  r = g = b = 0;
	}
	_TL_DISPLAY_PUT_PIXEL (r, g, b, x, y, ID);
      }
    }
    break;
      
  default:
    if (tlVerbose > 1) {
      TL_ERROR_MSG ("tlDisplay::display: unknown image format");
    }
    return TL_ERROR_INVALID_FORMAT;
    break;
  }

  update ();
  
  return (_tlDisplayImagePaint (ID));
}


/**
 ***********************************************************************
 *
 *  Display a histogram in the display area.
 *
 *  @param hist     the histogram to display
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::display (tlHist *hist)
{
  if (!visible) {		// do nothing
    return TL_NO_ERROR;
  }

  if (!hist) {
    TL_ERROR_MSG ("tlWindow::display: NULL histogram\n");
    return TL_ERROR_NULL_POINTER;
  }

  int histWidth = hist->getLength();
  int histHeight = hist->getMaxVal() - hist->getMinVal();

  // verify X window size
  if ((histWidth != this->width) || (histHeight != this->height)) {
    if (0 > resize (histWidth, histHeight)) {
      return (TL_ERROR);		/* failure */
    }
  }

  // convert to a displayable RGB format
  int *src = hist->data;
  register int x, y;
  unsigned char r, g, b;

  for (y=0; y<height; y++) {
    for (x=0; x<width; x++) {
      if (*(src) < (height-y)) {
	r = g = b = 0;
      } 
      else {
	r = g = b = 255;
      }
      _TL_DISPLAY_PUT_PIXEL (r, g, b, x, y, ID);

      src++;
    }
    src = hist->data;
  }

  update ();
  
  return (_tlDisplayImagePaint (ID));
}




/**
 ***********************************************************************
 *
 *  Unimplemented window management.
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::maximize ()
{
  return 0;
}



/**
 ***********************************************************************
 *
 *  Unimplemented window management.
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::minimize ()
{
  return 0;
}



/**
 ***********************************************************************
 *
 *  Unimplemented window management.
 *
 *  \return
 *    TL_ERROR status (see tlVision.h for details)
 *
 ***********************************************************************/
int 
tlDisplay::restore ()
{
  return 0;
}
