/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Terry Fong, Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://imtsg14.epfl.ch/tlib>
//
/////////////////////////////////////////////////////////////////////////////////


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>		/* ioctl */
#include <linux/videodev.h>
#include <malloc.h>
#include <string.h>
#include <ccvt.h>
#include <pwc-ioctl.h>

#ifdef __cplusplus
}
#endif /* __cplusplus */

#include "tlDigitizerV4L.h"


// documentation

///  \file tlDigitizerV4L.h
///  \brief Video4Linux digitizer interface.

///  \class tlDigitizerV4L
///  Video4Linux digitizer interface for TLIB.



#ifndef NO_DOC

static int
tryPalette (int ifd, int palette, int depth)
{
  struct video_picture vpic;

  if (ioctl (ifd, VIDIOCGPICT, &vpic) < 0) {
    return (-1);		/* failure */
  }

  vpic.palette = palette;
  vpic.depth = depth;
  if (ioctl (ifd, VIDIOCSPICT, &vpic) < 0) {
    return (-1);		/* failure */
  }

  if (ioctl (ifd, VIDIOCGPICT, &vpic) < 0) {
    return (-1);		/* failure */
  }

  if (vpic.palette != palette) {
    return (-1);		/* failure */
  }

  return (0);			/* success */
}

#endif // NO_DOC





////////////////////////////////////////////////////////////////////////
tlDigitizerV4L::tlDigitizerV4L() : tlDigitizer ()
{
  if (tl_errno != TL_NO_ERROR) return;
  ifd             = -1;
  mmbuf           = NULL;
  rgbImage        = NULL;
  grabQueuedInit  = 0;
}


////////////////////////////////////////////////////////////////////////
///
///  Prepare a digitizer for image acquisition using a given acquisition size.
///  A subsequent call to open() will open the digitizer with the given settings.
///
///  \param width           acquisition width
///  \param height          acquisition height
///
///  \note
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
tlDigitizerV4L::tlDigitizerV4L (int width, 
                                int height) : tlDigitizer (width, height)
{
  if (tl_errno != TL_NO_ERROR) return;
  ifd             = -1;
  mmbuf           = NULL;
  rgbImage        = NULL;
  grabQueuedInit  = 0;
}



////////////////////////////////////////////////////////////////////////
///
///  Prepare a digitizer for image acquisition using a given acquisition size
///  and specific acquisition channel. A subsequent call to open() will open
///  the digitizer with the given settings.
///
///  \param width           new acquisition width
///  \param height          new acquisition height
///  \param channel         channel index on the digitizer
///  \param input           acquisition channel video format
///
///  \note
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
tlDigitizerV4L::tlDigitizerV4L (int            width, 
                                int            height,
                                int            channel,
                                unsigned short input) : tlDigitizer (width, height)
{
  if (tl_errno != TL_NO_ERROR) return;
  ifd             = -1;
  mmbuf           = NULL;
  rgbImage        = NULL;
  grabQueuedInit  = 0;
  this->channel   = channel;
  this->input     = format;
}


////////////////////////////////////////////////////////////////////////
tlDigitizerV4L::~tlDigitizerV4L ()
{
  close ();
}


////////////////////////////////////////////////////////////////////////
int
tlDigitizerV4L::open (int index)
{
  return (open (width, height, index, channel, input));
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::open (int width, 
                      int height,
		      int index)
{
  return (open (width, height, index, channel, input));
}


////////////////////////////////////////////////////////////////////////
///
///  Open a digitizer for image acquisition using a given acquisition size
///  and a specific acquisition channel.
///
///  \param width           new acquisition width
///  \param height          new acquisition height
///  \param channel         channel index on the digitizer
///  \param input           input channel video format
///  \param index           \b [default=0] open device at a specific index
///                         (for use of multiple devices)
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::open (int            width, 
                      int            height,
		      int            index,
                      int            channel,
		      unsigned short input)
{
  char device_name[256];

  // sanity check
  if (width < 1)  width  = TL_DEFAULT_COLS;
  if (height < 1) height = TL_DEFAULT_ROWS;

  // save configuration
  this->width   = width;
  this->height  = height;
  this->channel = channel;
  this->format  = format;

  // open the frame grabber 
  sprintf (device_name, "%s%d", TLDIGITIZER_V4L_BASE_NAME, index);  

  if (tlVerbose) {
    printf ("opening %s\n", device_name);
  }

  //  if ((ifd = _grab_open (device_name, O_RDWR)) < 0){
  if ((ifd = ::open (device_name, O_RDWR)) < 0){
    perror ("tlDigitizerV4L::open");
    TL_ERRNO_MSG ("tlDigitizerV4L::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }    

  if (ioctl (ifd, VIDIOCGCAP, &vcap) < 0) {
    perror ("tlDigitizerV4L::open: VIDIOCGCAP");
    TL_ERRNO_MSG ("tlDigitizerV4L::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }
  if (tlVerbose) {
    printf ("tlDigitizerV4L::open: v4l capture device name = %s\n", vcap.name);
  }

  if (strstr (vcap.name, "Philips") &&
      strstr (vcap.name, "webcam")) {
    int compression;
    int agc;
    int val;
    struct video_window vw;
    struct pwc_whitebalance wb;

    if (tlVerbose) {
      printf ("tlDigitizerV4L::open: Philips webcam detected. Configuring\n");
    }

    // set image capture size
    memset (&vw, 0, sizeof(vw));
    vw.width = width;
    vw.height = height;
    vw.clipcount = 0;
    if (ioctl (ifd, VIDIOCSWIN, &vw) < 0) {
      perror ("Philips: VIDIOCSWIN");
    }

    if (ioctl (ifd, VIDIOCGWIN, &vw) < 0) {
      perror ("Philips: VIDIOCGWIN");
    }
    if (vw.width != (unsigned int) width) {
      fprintf (stderr, "tlDigitizerV4L::open: warning: Philips image width = %d (%d not supported)\n",
	      vw.width, width);
    }
    if (vw.height != (unsigned int) height) {
      fprintf (stderr, "tlDigitizerV4L::open: warning: Philips image width = %d (%d not supported)\n",
	      vw.height, height);
    }
    if (vw.flags & PWC_FPS_FRMASK) {
      fprintf (stderr, "tlDigitizerV4L::open: Philips framerate is currently set to: %d fps\n", 
	      (vw.flags & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT);
    }

    /* Set preferred compression quality
     * (0 = uncompressed, 3 = highest compression)
     *
     * This is only a preference: if a combination of size & framerate is
     * not supported uncompressed, the driver will silently switch
     * to a compressed mode. 
     */
    compression = 0;
    if (ioctl (ifd, VIDIOCPWCSCQUAL, &compression) < 0) {
      fprintf (stderr, "tlDigitizerV4L::open: could not set compression=%d\n", compression);
    }

    /* Get preferred compression quality */
    if (ioctl (ifd, VIDIOCPWCGCQUAL, &compression) < 0) {
      fprintf (stderr, "tlDigitizerV4L::open: could not get compression setting\n");
    }
    if (tlVerbose) {
      printf ("tlDigitizerV4L::open: Philips compression = %d\n", compression);
    }

    /* 
     * Set manual white balance
     */
    wb.mode = PWC_WB_MANUAL;
    wb.manual_red = 21845;	// 1/3 full scale (0-65535)
    wb.manual_blue = 32768;	// 1/2 full scale (0-65535)
    if (ioctl (ifd, VIDIOCPWCSAWB, &wb) < 0) {
      fprintf (stderr, "tlDigitizerV4L::open: could not set manual white balance\n");
    }

    /* 
     * Set manual gain
     */
    //    agc = 16384;		// 1/4 full scale (0-65535)
    agc = -1;			// enable auto-gain
    if (ioctl (ifd, VIDIOCPWCSAGC, &agc) < 0) {
      fprintf (stderr, "tlDigitizerV4L::open: could not set manual gain\n");
    }

    /* 
     * Disable backlight compensation
     */
    val = 0;			// 0 = disable, 1 = enable
    if (ioctl (ifd, VIDIOCPWCSBACKLIGHT, &val) < 0) {
      fprintf (stderr, "tlDigitizerV4L::open: could not disable backlight compensation\n");
    }

    /* 
     * Disable noise reduction
     */
    val = 0;			// 0 (none) to 3 (high)
    if (ioctl (ifd, VIDIOCPWCSDYNNOISE, &val) < 0) {
      fprintf (stderr, "tlDigitizerV4L::open: could not disable noise reduction\n");
    }

    /* 
     * Disable contour sharpening
     */
    val = 0;			// 0 (none) to 3 (high)
    if (ioctl (ifd, VIDIOCPWCSCONTOUR, &val) < 0) {
      fprintf (stderr, "tlDigitizerV4L::open: could not disable contour sharpening\n");
    }

    /* probably should include pwc-ioctl.h from your kernel source
       (somewhere around /usr/src/linux/drivers/usb);
       must match version of pwc.o and pwcx.o modules!!!
    */

    /* probably should also do:
       VIDIOCPWCSSHUTTER   Set shutter speed
    */
  }

  setChannel (channel, input);


  /* set up the capture type and size */
  if (NULL == (rgbImage = (unsigned char *) malloc (width * height * 3))) {
		TL_ERRNO_MSG ("tlDigitizerV4L::open()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  if (tlVerbose) {
    printf ("tlDigitizerV4L::open: rgbImage malloc'd at 0x%x\n", 
            (unsigned int) rgbImage);
  }
  
  /* set up the capture type and size */
  vbuf.frame  = 0;
  vbuf.height = height; /* # of lines in output image */
  vbuf.width  = width; /* # of pixels in a row in output image */
  
  if (!tryPalette (ifd, VIDEO_PALETTE_RGB24, 24)) {
    if (tlVerbose) {
      printf ("tlDigitizerV4L::open: RGB24 palette detected\n");
    }
    vbuf.format = VIDEO_PALETTE_RGB24; /* RGB in 3 bytes: B,G,R */
  }
  else if (!tryPalette (ifd, VIDEO_PALETTE_YUV420P, 16)) {
    if (tlVerbose) {
      printf ("tlDigitizerV4L::open: YUV420P (4:2:2 Planar) detected\n");
    }
    vbuf.format = VIDEO_PALETTE_YUV420P; /* YUV 4:2:2 Planar */
  }
  else {
    TL_ERRNO_MSG ("tlDigitizerV4L::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  
  // if everything succeeded, register slot
  isOpen = true;
  
  
  return 0; 
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::setSize (int width, 
                         int height)
{
  close();
  open (width, height);

  return (TL_NO_ERROR);
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::close ()
{
  int status = 0;
  
  if (!isOpen) {
    TL_ERROR_MSG ("tlDigitizerV4L->close() : digitizer not open");
  }

  if (ifd != -1) {
    if ((0 > munmap (mmbuf, mbuf.size)) || (0 > ::close(ifd))) {
      TL_ERRNO_MSG ("tlDigitizerV4L::close()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
      status = -1;
    }
  }

  ifd = -1;  
  mmbuf = NULL;
  if (tlVerbose) {
    printf ("tlDigitizerV4L::close: freeing rgbImage at 0x%x\n", (unsigned int) rgbImage);
  }
  free (rgbImage);
  rgbImage = NULL;

  // notify
  isOpen = false;

  return (status); 
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::grab (tlImage *image)
{
  int                     i=0, row, col, pixels;
  register tlPixel*       data=NULL;
  register unsigned char* p;

  if (ifd == -1 || !isOpen) {
    TL_ERROR_MSG ("tlDigitizerV4L::grab: error: frame grabber not initialized\n");
    return (TL_ERROR);		/* failure */
  } 

  //
  // capture one frame
  //

  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grab: capturing %dx%d frame\n", 
	    vbuf.width, vbuf.height);
  }

  if (ioctl (ifd, VIDIOCMCAPTURE, &vbuf) <0) {
    perror ("VIDIOCMCAPTURE");
		TL_ERRNO_MSG ("tlDigitizerV4L::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  /* wait for a frame to complete */
  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grab: waiting for frame\n");
  }

  if (ioctl (ifd, VIDIOCSYNC, &vbuf.frame) < 0) {
    TL_ERRNO_MSG ("tlDigitizerV4L::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grab: transferring image\n");
  }


  //
  // transfer data from the framegrabber to the image
  //

  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grab: init image to RGB\n");
  }

  // initialize image to grabber format
  image->init (RGB, width, height);

  // start of image data
  p = (unsigned char *) mmbuf + 
    mbuf.offsets[vbuf.frame];

  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grab: copying image data\n");
  }

  // copy image data
  switch (vbuf.format) {
  case VIDEO_PALETTE_YUV420P:
    ccvt_420p_rgb24 (width, height, p, rgbImage);
    p = rgbImage;
    pixels = width * height * 3;

    data = image->pixel;
    for (i=0; i<pixels; i++) {
      *data = *p;
      data++;
      p++;
    }
    break;

  case VIDEO_PALETTE_RGB24:
    data = image->pixel;
    for (row=0; row<height; row++) {
      for (col=0; col<width; col++) {
        data[i+2] = *p++;	/* blue */
        data[i+1] = *p++;	/* green */
        data[i] = *p++;		/* red */
        i+=3;
      }
    }
    break;
  }

  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grab: got frame\n");
  }

  return TL_NO_ERROR;
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::grab (tlImage *image, 
                      tlRect  *rect)
{
  if (ifd == -1 || !isOpen) {
    TL_ERROR_MSG ("tlDigitizerV4L::grab: error: frame grabber not initialized\n");
    return (TL_ERROR);		/* failure */
  } 

#if 0
  int                     x1, y1, x2, y2;
  struct video_capture vc;

  if (vcap.type && VID_TYPE_SUBCAPTURE) {
    printf ("digitizer can subcapture\n");
  }
  else {
    printf ("no subcapture\n");
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (x1+(rect->width), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (y1+(rect->height), height);

  vc.x = x1; vc.y = y1; vc.width = x2 - x1; vc.height = y2 - y1;
  vc.decimation = 1; vc.flags = VIDEO_CAPTURE_ODD;

  if (ioctl (ifd, VIDIOCSCAPTURE, &vc) < 0) {
    TL_ERRNO_MSG ("tlDigitizerV4L::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

#else

  //
  // capture a full frame 
  //

	return grab (image);

#endif

  return (TL_NO_ERROR);
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::grabQueued (tlImage *image)
{
  int                     i=0, row, col, pixels;
  register tlPixel*       data=NULL;
  register unsigned char* p;

    
  if (ifd == -1 || !isOpen) {
    TL_ERROR_MSG ("tlDigitizerV4L::grab: error: frame grabber not initialized\n");
    return (TL_ERROR);		/* failure */
  } 

  if (!grabQueuedInit) {
    //
    // capture one frame
    //
    if (tlVerbose > 1) {
      printf ("tlDigitizerV4L::grabQueued: capturing %dx%d frame\n", 
	      vbuf.width, vbuf.height);
    }

    if (ioctl (ifd, VIDIOCMCAPTURE, &vbuf) <0) {
      TL_ERRNO_MSG ("tlDigitizerV4L::grabQueued()", TL_ERROR_GRAB_FAILED);
			return -1;
    }

    grabQueuedInit = 1;
  }

  /* wait for a frame to complete */
  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grabQueued: waiting for frame\n");
  }

  if (ioctl (ifd, VIDIOCSYNC, &vbuf.frame) < 0) {
    TL_ERRNO_MSG ("tlDigitizerV4L::grabQueued()", TL_ERROR_GRAB_FAILED);
		return -1;
  }

  //
  // transfer data from the framegrabber to the image
  //

  // initialize image to grabber format
  image->init (RGB, width, height);

  // start of image data
  p = (unsigned char *) mmbuf + 
    mbuf.offsets[vbuf.frame];

  // copy image data
  switch (vbuf.format) {
  case VIDEO_PALETTE_YUV420P:
    ccvt_420p_bgr24 (width, height, p, rgbImage);
    p = rgbImage;
    pixels = width * height * 3;

    data = image->pixel;
    for (i=0; i<pixels; i++) {
      *data = *p;
      data++;
      p++;
    }
    break;

  case VIDEO_PALETTE_RGB24:
    data = image->pixel;
    for (row=0; row<height; row++) {
      for (col=0; col<width; col++) {
        data[i+2] = *p++;	/* blue */
        data[i+1] = *p++;	/* green */
        data[i] = *p++;		/* red */
        i+=3;
      }
    }
    break;
  }

  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grabQueued: got frame\n");
  }

  //
  // capture one frame
  //
  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grabQueued: capturing %dx%d frame\n", 
	    vbuf.width, vbuf.height);
  }

  if (ioctl (ifd, VIDIOCMCAPTURE, &vbuf) <0) {
    TL_ERRNO_MSG ("tlDigitizerV4L::grabQueued()", TL_ERROR_GRAB_FAILED);
		return -1;
  }

  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L::grabQueued: queueing next frame\n");
  }

  return 0;
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L::grabQueued (tlImage *image, 
                            tlRect  *rect)
{
  //
  // capture a full frame
  //

  return grabQueued (image);
}



/////////////////////////////////////////////////////////////////////////
///
///  Set the acquisition channel.
///    
///  \param channel     channel index on the digitizer
///  \param input       input video format
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
int
tlDigitizerV4L::setChannel (int            channel,
                            unsigned short input)
{
  struct video_channel vc;

  
  vc.channel = channel;
  if (ioctl (ifd, VIDIOCGCHAN, &vc) < 0 ) {
    TL_ERRNO_MSG ("tlDigitizerV4L::setChannel()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  
  if (tlVerbose) {
    printf ("tlDigitizerV4L::setChannel(): v4l video source (channel %d) = %s\n", channel, vc.name);
  }
  
  if (ioctl (ifd, VIDIOCGMBUF, &mbuf) < 0) {
    TL_ERRNO_MSG ("tlDigitizerV4L::setChannel()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  
  if (tlVerbose) {
    int i;
    printf ("tlDigitizerV4L::setChannel(): mapped memory: %d buffers, %d frames, 0x%x bytes/buf\n",
	    mbuf.frames, mbuf.frames, 
	    mbuf.size);
    printf ("tlDigitizerV4L::setChannel(): offsets: ");
    for (i=0; i<mbuf.frames; i++) {
      printf ("%d=0x%x ", i, mbuf.offsets[i]);
    }
    printf ("\n");
  }
  
  if ((mmbuf = (unsigned char *) mmap ((caddr_t) 0, mbuf.size,
				       PROT_READ | PROT_WRITE, 
				       MAP_SHARED, ifd, (off_t) 0)) == (unsigned char*) (-1)) {
    TL_ERRNO_MSG ("tlDigitizerV4L::setChannel()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  
  if (tlVerbose) {
    printf ("tlDigitizerV4L::setChannel(): mmap at 0x%x\n", (int) mmbuf);
  }
  
  vc.channel = channel;
  if (tlVerbose) {
    printf ("tlDigitizerV4L::setChannel(): setting channel %d\n", vc.channel);
  }
  
  // memorize and configure
  this->channel = channel;
  return setInput (input);
}



/////////////////////////////////////////////////////////////////////////
///
///  Set the acquisition channel input format.
///    
///  \param input      input channel video format
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
int
tlDigitizerV4L::setInput (unsigned short input)
{
  struct video_channel vc;


  // apply input config to current channel
  vc.channel = channel;

  // set input video format
  if (input & TL_SOURCE_AUTO) {
    vc.norm = VIDEO_MODE_AUTO;
  }
  else if (input & TL_SOURCE_SECAM) {
    vc.norm = VIDEO_MODE_SECAM;
  }
  else if (input & (TL_SOURCE_NTSCJ | TL_SOURCE_NTSC)) {
    vc.norm = VIDEO_MODE_NTSC;
  }
  else if (input & (TL_SOURCE_PALM | TL_SOURCE_PALN | TL_SOURCE_PAL)) {
    vc.norm = VIDEO_MODE_PAL;
  }
  vc.type = VIDEO_TYPE_CAMERA;  
  if (ioctl (ifd, VIDIOCSCHAN, &vc) < 0) {
    TL_ERRNO_MSG ("tlDigitizerV4L::setChannel()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // store parameters
  this->input = input;



  return 0;
}


////////////////////////////////////////////////////////////////////////
///
///  Set the frame rate of the DirectX compatible digitizer.
///
///  \param fps     frame rate [frame per second]
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
int
tlDigitizerV4L::setFrameRate (int fps)
{
  if (strstr (vcap.name, "Philips") &&
      strstr (vcap.name, "webcam")) {
    struct video_window vw;
    
    printf ("tlDigitizerV4L::setFrameRate : configuring Philips webcam for %d fps\n", fps);
    
    if (ioctl (ifd, VIDIOCGWIN, &vw) < 0) {
      TL_ERRNO_MSG ("tlDigitizerV4L::setFrameRate()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
    
    // set new framerate
    vw.flags &= ~PWC_FPS_FRMASK;
    vw.flags |= (fps << PWC_FPS_SHIFT);
    
    if (ioctl (ifd, VIDIOCSWIN, &vw) < 0) {
      TL_ERRNO_MSG ("tlDigitizerV4L::setFrameRate()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
  }
  else {
    TL_ERRNO_MSG ("tlDigitizerV4L::setFrameRate()", TL_ERROR_NOT_AVAILABLE);
    return -1;
  }
  
  return TL_NO_ERROR;
}
