/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Terry Fong, Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://imtsg14.epfl.ch/tlib>
//
/////////////////////////////////////////////////////////////////////////////////


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>		      /* ioctl */
#include <asm/types.h>          /* for videodev2.h */
#include <linux/videodev2.h>
#include <malloc.h>
#include <string.h>
#include <ccvt.h>

#ifdef __cplusplus
}
#endif /* __cplusplus */

#include "tlDigitizerV4L2.h"


// documentation

///  \file tlDigitizerV4L2.h
///  \brief Video4Linux2 digitizer interface.

///  \class tlDigitizerV4L2
///  Video4Linux2 digitizer interface for TLIB.


#ifndef NO_DOC

#define CLEAR(x) memset (&(x), 0, sizeof (x))

static int
xioctl (int fd, int request, void *arg)
{
  int r;

  do r = ioctl (fd, request, arg);
  while (-1 == r && EINTR == errno);

  return r;
}

// Map the driver buffers to user memory.
void
tlDigitizerV4L2::init_mmap (void)
{
  struct v4l2_requestbuffers req;
  CLEAR (req);
  // Use 2 buffers. More buffers are probably better but as we enqueue
  // and dequeue them every time in grab(), less is probably not worst.
  req.count  = 2;
  req.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;

  if (-1 == xioctl (ifd, VIDIOC_REQBUFS, &req)) {
    if (EINVAL == errno) {
      fprintf (stderr, "%s does not support memory mapping\n", vcap.card);
      exit (EXIT_FAILURE);
    } else {
      exit (EXIT_FAILURE);
    }
  }

  if (req.count < 2) {
    fprintf (stderr, "Insufficient buffer memory on %s\n", vcap.card);
    exit (EXIT_FAILURE);
  }

  buffers = (struct buffer*) calloc (req.count, sizeof (*buffers));

  if (!buffers) {
    fprintf (stderr, "Out of memory\n");
    exit (EXIT_FAILURE);
  }

  for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
    struct v4l2_buffer buf;
    CLEAR (buf);

    buf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index  = n_buffers;

    if (-1 == xioctl (ifd, VIDIOC_QUERYBUF, &buf))
      exit (EXIT_FAILURE);

    buffers[n_buffers].length = buf.length;
    buffers[n_buffers].start = mmap (NULL /* start anywhere */,
                                     buf.length,
                                     PROT_READ | PROT_WRITE /* required */,
                                     MAP_SHARED /* recommended */,
                                     ifd, buf.m.offset);

    if (MAP_FAILED == buffers[n_buffers].start)
      exit (EXIT_FAILURE);
  }
}

// Unmap buffers.
void
tlDigitizerV4L2::uninit_mmap (void)
{
  unsigned int i;

  for (i = 0; i < n_buffers; ++i)
    if (-1 == munmap (buffers[i].start, buffers[i].length))
      exit (EXIT_FAILURE);

  free (buffers);
}

// Start the capture.
void
tlDigitizerV4L2::start_capturing (void)
{
  enum v4l2_buf_type type;

  // Enqueue buffers with VIDIOC_QBUF
  for (unsigned i = 0; i < n_buffers; ++i) {
    struct v4l2_buffer buf;
    CLEAR (buf);

    buf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index  = i;

    if (-1 == xioctl (ifd, VIDIOC_QBUF, &buf))
      exit (EXIT_FAILURE);
  }

  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  // Start capture
  if (-1 == xioctl (ifd, VIDIOC_STREAMON, &type))
    exit (EXIT_FAILURE);
}

// Stop the capture.
void
tlDigitizerV4L2::stop_capturing (void)
{
  enum v4l2_buf_type type;

  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  // Stop the capture and remove buffers from queues
  if (-1 == xioctl (ifd, VIDIOC_STREAMOFF, &type))
    exit (EXIT_FAILURE);
}

#endif // NO_DOC


////////////////////////////////////////////////////////////////////////
tlDigitizerV4L2::tlDigitizerV4L2 () : tlDigitizer ()
{
  if (tl_errno != TL_NO_ERROR) return;
  ifd             = -1;
  rgbImage        = NULL;
  grabQueuedInit  = 0;
}


////////////////////////////////////////////////////////////////////////
///
///  Prepare a digitizer for image acquisition using a given acquisition size.
///  A subsequent call to open() will open the digitizer with the given settings.
///
///  \param width           acquisition width
///  \param height          acquisition height
///
///  \note
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
tlDigitizerV4L2::tlDigitizerV4L2 (int width, 
                                  int height) : tlDigitizer (width, height)
{
  if (tl_errno != TL_NO_ERROR) return;
  ifd             = -1;
  rgbImage        = NULL;
  grabQueuedInit  = 0;
}



////////////////////////////////////////////////////////////////////////
///
///  Prepare a digitizer for image acquisition using a given acquisition size
///  and specific acquisition channel. A subsequent call to open() will open
///  the digitizer with the given settings.
///
///  \param width           new acquisition width
///  \param height          new acquisition height
///  \param channel         channel index on the digitizer
///  \param input           input channel video format
///
///  \note
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
tlDigitizerV4L2::tlDigitizerV4L2 (int            width, 
                                  int            height,
                                  int            channel,
                                  unsigned short input) : tlDigitizer (width, height)
{
  if (tl_errno != TL_NO_ERROR) return;
  ifd             = -1;
  rgbImage        = NULL;
  grabQueuedInit  = 0;
}


////////////////////////////////////////////////////////////////////////
tlDigitizerV4L2::~tlDigitizerV4L2 ()
{
  close ();
}


////////////////////////////////////////////////////////////////////////
int
tlDigitizerV4L2::open (int index)
{
  return (open (width, height, index, channel, format));
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::open (int width, 
                       int height,
		                   int index)
{
  return (open (width, height, index, channel, format));
}


////////////////////////////////////////////////////////////////////////
///
///  Open a digitizer for image acquisition using a given acquisition size
///  and a specific acquisition channel.
///
///  \param width           new acquisition width
///  \param height          new acquisition height
///  \param channel         channel index on the digitizer
///  \param input           input channel video format
///  \param index           \b [default=0] open device at a specific index
///                         (for use of multiple devices)
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::open (int            width, 
                       int            height,
		                   int            index,
                       int            channel,
		                   unsigned short input)
{
  char device_name[256];


  // sanity check
  if (width < 1) width = TL_DEFAULT_COLS;
  if (height < 1) height = TL_DEFAULT_ROWS;

   // save configuration
  this->width   = width;
  this->height  = height;

  // open the frame grabber 
  sprintf (device_name, "%s%d", TLDIGITIZER_V4L2_BASE_NAME, index);  
  if (tlVerbose) {
    printf ("opening %s\n", device_name);
  }

  // Do not use O_NONBLOCK thus VIDIOC_DQBUF in grab() will block until
  // a buffer is full. With O_NONBLOCK, VIDIOC_DQBUF will return EAGAIN
  // if no buffer is available. Could be better in combination with a
  // timeout (see video capture example in V4L2 documentation).
  if ((ifd = ::open (device_name, O_RDWR /*| O_NONBLOCK*/, 0)) < 0){
    perror ("tlDigitizerV4L2::open");
    TL_ERRNO_MSG ("tlDigitizerV4L2::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }    

  // Query video device capabilities
  if (xioctl (ifd, VIDIOC_QUERYCAP, &vcap) < 0) {
    perror ("tlDigitizerV4L2::open: VIDIOC_QUERYCAP");
    TL_ERRNO_MSG ("tlDigitizerV4L2::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }
  if (tlVerbose) {
    printf ("tlDigitizerV4L2::open: v4l capture device name = %s\n", vcap.card);
  }

  // We want to capture video...
  if (!(vcap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
    fprintf (stderr, "%s is no video capture device\n", vcap.card);
    return -1;
  }
  // ...using streaming I/O
  if (!(vcap.capabilities & V4L2_CAP_STREAMING)) {
    fprintf (stderr, "%s does not support streaming i/o\n", vcap.card);
    return -1;
  }

  // Set the channel and input format 
  setChannel (channel, input);

  struct v4l2_cropcap cropcap;
  struct v4l2_crop crop;

  /* Select video input, video standard and tune here. */
  cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

  if (-1 == xioctl (ifd, VIDIOC_CROPCAP, &cropcap)) {
    /* Errors ignored. */
  }

  crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  crop.c = cropcap.defrect; /* reset to default */

  // Set cropping to default. A good V4L2 application must take care
  // of restoring default values.
  if (-1 == xioctl (ifd, VIDIOC_S_CROP, &crop)) {
    switch (errno) {
      case EINVAL:
        /* Cropping not supported. */
        break;
      default:
        /* Errors ignored. */
        break;
    }
  }

  /* set up the capture type and size */
  if (NULL == (rgbImage = (unsigned char *) malloc (width * height * 3))) {
    TL_ERRNO_MSG ("tlDigitizerV4L2::open()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  if (tlVerbose) {
    printf ("tlDigitizerV4L2::open: rgbImage malloc'd at 0x%x\n", 
            (unsigned int) rgbImage);
  }
  
  struct v4l2_format	 fmt;
  CLEAR (fmt);
  
  fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width       = width; 
  fmt.fmt.pix.height      = height;
  // The em28xx driver for my Pinnacle Dazzle DVC 90 seems to
  // only support YUYV. May be not the case for other drivers.
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;

  if (-1 == xioctl (ifd, VIDIOC_S_FMT, &fmt)) {
    printf("VIDIOC_S_FMT failed\n");
    return -1;
  }
  
  if (tlVerbose > 0) {
  	if (-1 == xioctl (ifd, VIDIOC_G_FMT, &fmt)) {
      printf("VIDIOC_G_FMT failed\n");
      return -1;
    }
  	printf("Size        : %dx%d\n", fmt.fmt.pix.width, fmt.fmt.pix.height);
  	printf("Pixel format: %x\n", fmt.fmt.pix.pixelformat);
  	printf("Field       : %d\n", fmt.fmt.pix.field);
  }

  /* Buggy driver paranoia. */
  unsigned min = fmt.fmt.pix.width * 2;
  if (fmt.fmt.pix.bytesperline < min)
    fmt.fmt.pix.bytesperline = min;
  min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
  if (fmt.fmt.pix.sizeimage < min)
    fmt.fmt.pix.sizeimage = min;
  
  // Map driver buffers to user memory
  init_mmap();
  
  // if everything succeeded, register slot
  isOpen = true;

  
  return 0; 
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::setSize (int width, 
                          int height)
{
  close();
  open (width, height);

  return (TL_NO_ERROR);
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::close ()
{
  int status = 0;
  
  if (!isOpen) {
    TL_ERROR_MSG ("tlDigitizerV4L2->close() : digitizer not open");
  }

  stop_capturing();
  uninit_mmap();

  if (-1 == ::close (ifd))
    return -1;

  ifd = -1;
        
  if (tlVerbose) {
    printf ("tlDigitizerV4L2::close: freeing rgbImage at 0x%x\n", (unsigned int) rgbImage);
  }
  free (rgbImage);
  rgbImage = NULL;

  // notify
  isOpen = false;

  return (status); 
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::grab (tlImage *image)
{
  register tlPixel        *data = NULL;
  register unsigned char  *p;
  struct v4l2_buffer      buf;


  if (ifd == -1 || !isOpen) {
    TL_ERRNO_MSG ("tlDigitizerV4L2::grab()", TL_ERROR);
    return (TL_ERROR);
  } 

  // Start capturing
  start_capturing();
  
  CLEAR (buf);
  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  // Dequeue one buffer (one image)
  if (-1 == xioctl (ifd, VIDIOC_DQBUF, &buf)) {
    switch (errno) {
      case EAGAIN:
        // Should never happen is open() did not set the
        // O_NONBLOCK flag.
        printf("grab(): EAGAIN\n");
        return 0;
      case EIO:
        /* Could ignore EIO, see spec. */
        /* fall through */
      default:
        return -1;
    }
  }

  // assert (buf.index < n_buffers);

  // initialize image to grabber format
  image->init (RGB, width, height);

  // Convert image from YUYV to RGB24
  p = (unsigned char *) buffers[buf.index].start;
  ccvt_yuyv_rgb24 (width, height, p, rgbImage);
  p = rgbImage;
  data = image->pixel;
  memcpy (p, data, width * height * 3);
  /*
  unsigned pixels = width * height * 3;
  for (unsigned i=0; i<pixels; i++) {
    *data = *p;
    data++;
    p++;
  }
  */
  
  // Stop capturing and dequeue all buffers from queues.
  stop_capturing();
  
  if (tlVerbose > 1) {
    printf ("tlDigitizerV4L2::grab: got frame\n");
  }

  return TL_NO_ERROR;
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::grab (tlImage *image, 
                       tlRect  *rect)
{
  // To be implemented.
  return grab (image);
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::grabQueued (tlImage *image)
{
  // To be implemented. Maybe this could be done by copying/pasting
  // the content of grab(), removing the call to start_capturing()
  // and stop_capturing() and requeuing "buf".
  return grab(image);
}


////////////////////////////////////////////////////////////////////////
int 
tlDigitizerV4L2::grabQueued (tlImage *image, 
                             tlRect  *rect)
{
  return grabQueued (image);
}



/////////////////////////////////////////////////////////////////////////

int
tlDigitizerV4L2::setChannel (int            channel,
			                       unsigned short format)
{
  // set the channel.
  if (-1 == xioctl (ifd, VIDIOC_S_INPUT, &channel)) {
     TL_ERRNO_MSG ("tlDigitizerV4L2::setChannel()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
    
  // store channel
  this->channel = channel;
  return setInput (input);
}



/////////////////////////////////////////////////////////////////////////

int
tlDigitizerV4L2::setInput (unsigned short format)
{
  // set input video format (auto by default)
  v4l2_std_id std = V4L2_STD_ALL;
  if (input & TL_SOURCE_AUTO) {
    std = V4L2_STD_ALL;
  }
  else if (input & TL_SOURCE_SECAM) {
    std = V4L2_STD_SECAM;
  }
  else if (input & (TL_SOURCE_NTSCJ | TL_SOURCE_NTSC)) {
    std = V4L2_STD_NTSC;
  }    
  else if (input & (TL_SOURCE_PALM | TL_SOURCE_PALN | TL_SOURCE_PAL)) {
    std = V4L2_STD_PAL;
  }  
  if (-1 == xioctl (ifd, VIDIOC_S_STD, &std)) {
    TL_ERRNO_MSG ("tlDigitizerV4L2::setChannel()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }  
  if (tlVerbose > 0) {
    if (-1 == xioctl (ifd, VIDIOC_G_STD, &std)) {
      TL_ERRNO_MSG ("tlDigitizerV4L2::setChannel()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
    printf("Norm: %x\n", (unsigned int)std);
  }
  
  // store input format
  this->input = input;

  
  return 0;
}
