/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlDisplay.h"
#include "tlImage.h"
#include "tlHist.h"
#include "tlHist2D.h"
#include "tlDisplay_dx3.h"
#include "tlCameraStereo.h"

#include <windows.h>
#include <string.h>

#define INITGUID
#include <ddraw.h>



///  \file tlDisplay.h

///  \class tlDisplay
///  \brief Simple display class.
///
///  Provides a window to render images, as well as simple event management 
///  (mouse and keyboard input).
///  The window ID number will always appear in the window.

///  \def TL_MAX_DISPLAY
///  Maximum number of displays opened simultaneously.

///  \def TL_DISPLAY_NAME
///  String that will be displayed on the window by default.  



///  \fn int tlDisplay::getID()
///  Return the unique identifier for the given window.



/// \fn int tlDisplay::getWidth()
///  Return window height (displayable area without borders).



///  \fn int tlDisplay::getHeight()
///  Return window width (displayable area without borders).



///  \fn bool tlDisplay::getVisible()
///  Return window state (visible or hidden).



// system specifics constants and globals

HWND            _tl_display_dx3_window[TL_MAX_DISPLAY];
RECT            _tl_display_dx3_rect[TL_MAX_DISPLAY];
int             _tl_display_dx3_open_display = 0;
int             _tl_display_dx3_init = 0;
char            _tl_display_dx3_list[TL_MAX_DISPLAY];
tlDisplay*      _tl_display_dx3_ptr[TL_MAX_DISPLAY];
int             _tl_display_dx3_actual_width[TL_MAX_DISPLAY];
int             _tl_display_dx3_actual_height[TL_MAX_DISPLAY];

static LPDIRECTDRAW        _tl_lpDD[TL_MAX_DISPLAY];
static LPDIRECTDRAWSURFACE _tl_lpDDSPrimary[TL_MAX_DISPLAY];
static LPDIRECTDRAWCLIPPER _tl_lpClip[TL_MAX_DISPLAY];
static LPDIRECTDRAWSURFACE _tl_lpImage[TL_MAX_DISPLAY];

static int ScreenWidth, ScreenHeight;


///  Constructor, create default display.
///
///  \return
///  See \ref errors "error management" for details.

tlDisplay::tlDisplay  ()
{
  init (TL_DISPLAY_NAME, CW_USEDEFAULT, CW_USEDEFAULT);
}



///  Constructor.
///
///  \param name        window name (displayed in title bar)
///
///  \return
///  See \ref errors "error management" for details.

tlDisplay::tlDisplay  (char *name)
{
  init (name, CW_USEDEFAULT, CW_USEDEFAULT);
}



///  Constructor, specifies window position on the screen.
///
///  \param x       initial window X position  [pix]
///  \param y       initial window Y position  [pix]
///
///  \return
///  See \ref errors "error management" for details.

tlDisplay::tlDisplay  (int x, 
                       int y)
{
  init (TL_DISPLAY_NAME, x, y);
}



///  Constructor, specifies both position and window name.
///
///  \param name    window name (displayed in title bar)
///  \param x       initial window X position  [pix]
///  \param y       initial window Y position  [pix]
///
///  \return
///  See \ref errors "error management" for details.

tlDisplay::tlDisplay  (char *name, 
                       int  x, 
                       int  y)
{
  init (name, x, y);
}



///  TL library display & event management  destructor.
///
///  \return
///  See \ref errors "error management" for details.

tlDisplay::~tlDisplay ()
{
  close ();
  if (this->name) {
    free (this->name);
  }

  tl_errno = TL_NO_ERROR;
}



///  Constructor helper.
///
///  \param name    window name (displayed in title bar)
///  \param x       initial window X position  [pix]
///  \param y       initial window Y position  [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::init (char *name,
                 int  x,
                 int  y)
{
  int               k;
  static WNDCLASS   w;
  HINSTANCE         hInst;
  DDSURFACEDESC     ddsd;
  HWND              screen;
  RECT              screenRect;


  // if this is the first display
  if (_tl_display_dx3_init == 0) {

    // initialize display size
    screen = GetDesktopWindow ();
    GetWindowRect (screen, &screenRect);
    ScreenWidth  = screenRect.right - screenRect.left;
    ScreenHeight = screenRect.bottom - screenRect.top;

    // initialize device and pointer list
    for (k=0; k<TL_MAX_DISPLAY; k++) {
      _tl_display_dx3_list[k] = 0;
      _tl_display_dx3_ptr[k]  = NULL;
    }

    // initialize all DirectX vars
    for (k=0; k<TL_MAX_DISPLAY; k++) {
      _tl_lpDD[k]         = NULL;
      _tl_lpDDSPrimary[k] = NULL;
      _tl_lpClip[k]       = NULL;
      _tl_lpImage[k]      = NULL;
    }

    // done with first time initialization
    _tl_display_dx3_init = 1;
  }

  // find first available device index
  this->ID = -1;
  for (k=0; k<TL_MAX_DISPLAY; k++) {
    if (_tl_display_dx3_list[k] == 0) {
      this->ID = k;
      break;
    }
  }

  // check for availability
  if ((this->ID) == -1) {
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_MAX_REACHED);
    return -1;
  }

  // store params
  this->width  = TL_DEFAULT_COLS;
  this->height = TL_DEFAULT_ROWS;
  this->x      = x;
  this->y      = y;
  this->name   = (char *) malloc ((strlen(name)+5)*sizeof(char));
  sprintf (this->name, "%s [%d]", name, ID);

  // get application instance handle
  hInst = GetModuleHandle (NULL);

  // register the control window class
  w.style = 0;
  w.lpfnWndProc = _tlCtlProc;
  w.cbClsExtra = 0;
  w.cbWndExtra = sizeof(DWORD);
  w.hInstance = hInst;
  w.hIcon = NULL;
  w.hCursor = LoadCursor(NULL,IDC_ARROW);
  w.lpszClassName = this->name;
  if (!RegisterClass(&w)) {
    TL_ERRNO_MSG("tlDisplay::create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // create the display window
  _tl_display_dx3_window[this->ID] = CreateWindow(
    this->name,
    this->name,
    WS_OVERLAPPEDWINDOW,
    x,
    y,
    this->width  + _TL_BORDER_LEFT + _TL_BORDER_RIGHT,
    this->height + _TL_BORDER_TOP  + _TL_BORDER_BOTTOM,
    NULL,
    NULL,
    hInst,
    NULL);

  if (_tl_display_dx3_window[this->ID] == NULL) {
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // store class pointer
  _tl_display_dx3_ptr[this->ID] = this;

  // create DirectX instance
  if (DD_OK != DirectDrawCreate (NULL, &(_tl_lpDD[this->ID]), NULL)) {
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set DirectX cooperative level
  if (DD_OK != _tl_lpDD[this->ID]->SetCooperativeLevel (_tl_display_dx3_window[this->ID], DDSCL_NORMAL)) {
    _tl_lpDD[this->ID]->Release();
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // create clipper
  if (DD_OK != _tl_lpDD[this->ID]->CreateClipper (NULL, &(_tl_lpClip[this->ID]), NULL)) {
    TL_ERROR_MSG("tlDisplay::create() : DirectX clipper creation failed");
    _tl_lpDDSPrimary[this->ID]->Release();
    return TL_ERROR;
  }
  _tl_lpClip[this->ID]->SetHWnd (0, _tl_display_dx3_window[ID]);

  // create primary surface
  ZeroMemory (&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof(ddsd);
  ddsd.dwFlags = DDSD_CAPS;
  ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
  if (DD_OK != _tl_lpDD[this->ID]->CreateSurface (&(ddsd), &(_tl_lpDDSPrimary[this->ID]), NULL)) {
    _tl_lpClip[this->ID]->Release();
    _tl_lpDD[this->ID]->Release();
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // check surface properties
  _tl_lpDDSPrimary[this->ID]->GetSurfaceDesc (&ddsd);
  if (ddsd.ddpfPixelFormat.dwRGBBitCount < 16) {
    _tl_lpDDSPrimary[this->ID]->Release();
    _tl_lpClip[this->ID]->Release();
    _tl_lpDD[this->ID]->Release();
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // define clipper for primary surface
  if (DD_OK != _tl_lpDDSPrimary[this->ID]->SetClipper (_tl_lpClip[this->ID])) {
    _tl_lpDDSPrimary[this->ID]->Release();
    _tl_lpClip[this->ID]->Release();
    _tl_lpDD[this->ID]->Release();
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // create image surface associated with this display
  ZeroMemory (&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof (DDSURFACEDESC);
  ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
  ddsd.dwWidth = this->width;
  ddsd.dwHeight = this->height;
  if (DD_OK != _tl_lpDD[this->ID]->CreateSurface (&ddsd, &(_tl_lpImage[this->ID]), NULL)) {
    _tl_lpDDSPrimary[this->ID]->Release();
    _tl_lpClip[this->ID]->Release();
    _tl_lpDD[this->ID]->Release();
    TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // if everything was successfull, register slot
  _tl_display_dx3_list[this->ID] = 1;

  // update actual window position and size
  update ();


  return update ();
}


///  Helper function, retrieve window position in size on the fly.
///
///  \return
///  See \ref errors "error management" for details.

inline int
tlDisplay::update ()
{
  RECT winRect;


  GetWindowRect (_tl_display_dx3_window[this->ID], &winRect);
  this->x = winRect.left;
  this->y = winRect.top;


  return 0;
}



///  Reposition the window and the display area.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::setPos (int x, 
                   int y)
{
  DDSURFACEDESC ddsd;


  // reposition the display window
  if (0 == SetWindowPos (
    _tl_display_dx3_window[ID], 
    HWND_TOP, 
    x, 
    y, 
    width + _TL_BORDER_LEFT + _TL_BORDER_RIGHT, 
    height + _TL_BORDER_TOP + _TL_BORDER_BOTTOM,
    SWP_NOZORDER))
  {
    TL_ERRNO_MSG ("tlDisplay::setPos()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // store new location
  this->x = x;
  this->y = y;

  // resize image surface
  _tl_lpImage[this->ID]->Release();
  ZeroMemory (&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof (DDSURFACEDESC);
  ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
  ddsd.dwWidth = this->width;
  ddsd.dwHeight = this->height;
  if (DD_OK != _tl_lpDD[this->ID]->CreateSurface (&ddsd, &(_tl_lpImage[this->ID]), NULL)) {
    TL_ERRNO_MSG ("tlDisplay::setPos()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }


  return update ();  
}



///  Resize the window and the display area.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::resize (int width, 
                   int height)
{
  DDSURFACEDESC ddsd;


  // check geometry
  if ((width > ScreenWidth) || (height > ScreenHeight)) {
    TL_ERRNO_MSG ("tlDisplay::resize()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // store new size
  this->width = width;
  this->height = height;
  
  // resize the display window
  if (0 == SetWindowPos (
    _tl_display_dx3_window[ID], 
    HWND_TOP, 
    x, 
    y, 
    width + _TL_BORDER_LEFT + _TL_BORDER_RIGHT, 
    height + _TL_BORDER_TOP + _TL_BORDER_BOTTOM,
    SWP_NOZORDER | SWP_NOMOVE))
  {
    TL_ERRNO_MSG ("tlDisplay::resize()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // resize image surface
  _tl_lpImage[this->ID]->Release();
  ZeroMemory (&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof (DDSURFACEDESC);
  ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
  ddsd.dwWidth = this->width;
  ddsd.dwHeight = this->height;
  if (DD_OK != _tl_lpDD[this->ID]->CreateSurface (&ddsd, &(_tl_lpImage[this->ID]), NULL)) {
    TL_ERRNO_MSG ("tlDisplay::resize()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }


  return update ();  
}



///  Maximize the window.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::maximize ()
{
  // set property structure
  WINDOWPLACEMENT wp;
  GetWindowPlacement (_tl_display_dx3_window[ID], &wp);
  wp.showCmd = SW_MAXIMIZE;

  // resize the display window
  if (0 == SetWindowPlacement (_tl_display_dx3_window[ID], &wp)) {
    TL_ERRNO_MSG ("tlDisplay::maximize()", TL_ERROR_GEOMETRY);
    return -1;
  }

  
  return 0;
}



///  Minimize the window.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::minimize ()
{
  // set property structure
  WINDOWPLACEMENT wp;
  GetWindowPlacement (_tl_display_dx3_window[ID], &wp);
  wp.showCmd = SW_MINIMIZE;

  // resize the display window
  if (0 == SetWindowPlacement (_tl_display_dx3_window[ID], &wp)) {
    TL_ERRNO_MSG ("tlDisplay::minimize()", TL_ERROR_GEOMETRY);
    return -1;
  }

  
  return 0;
}



///  Restore the window to its nominal size.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::restore ()
{
  // set property structure
  WINDOWPLACEMENT wp;
  GetWindowPlacement (_tl_display_dx3_window[ID], &wp);
  wp.showCmd = SW_RESTORE;

  // resize the display window
  if (0 == SetWindowPlacement (_tl_display_dx3_window[ID], &wp)) {
    TL_ERRNO_MSG ("tlDisplay::restore()", TL_ERROR_GEOMETRY);
    return -1;
  }

  
  return 0;
}



///  Close and destroys the window.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::close ()
{
  // hide
  ShowWindow (_tl_display_dx3_window[this->ID], SW_HIDE);
  DestroyWindow (_tl_display_dx3_window[this->ID]);

  // release DirectX pointers
  _tl_lpDDSPrimary[this->ID]->Release();
  _tl_lpClip[this->ID]->Release();
  _tl_lpImage[this->ID]->Release();
  _tl_lpDD[this->ID]->Release();

  // deallocate globals
  _tl_display_dx3_list[this->ID] = 0;
  _tl_display_dx3_ptr[this->ID] = NULL;

  return update ();
}



///  Make window visible.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlDisplay::show ()
{
  // show window
  ShowWindow(_tl_display_dx3_window[ID],SW_SHOW);
  visible = true;


  return update ();
}



///  Make window invisible.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlDisplay::hide ()
{
  // show window
  ShowWindow(_tl_display_dx3_window[ID],SW_HIDE);
  visible = false;


  return update ();
}



///  Display a pixel array in the display area.
///
///  \param pixelArray  the pixel array to display
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::display (tlPixelArray *pixelArray)
{
  int                       i, j;
  unsigned char *           offset;
  int                       incr, pitch;
  static DDSURFACEDESC      ddsd[TL_MAX_DISPLAY];
  register tlPixel*         src;
  register unsigned char*   pixmap;


  // check if the window is visible (!)
  if (!visible) 
    return 0;

  // check image integrity
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlWindow->display", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check image size
  if ((pixelArray->width  != width) ||
      (pixelArray->height != height)) {
      if (0 > resize(pixelArray->width,pixelArray->height)) return -1;
    }

    // assign some useful vars
    src = pixelArray->pixel;

    // lock surface
    ZeroMemory (&(ddsd[this->ID]), sizeof(ddsd[this->ID]));
    ddsd[this->ID].dwSize = sizeof(ddsd[this->ID]);
    HRESULT res = _tl_lpImage[this->ID]->Lock (
      NULL, 
      &(ddsd[this->ID]), 
      DDLOCK_WRITEONLY | DDLOCK_SURFACEMEMORYPTR, 
      NULL);

    // make sure it worked
    if (DD_OK != res) {
      TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
      return -1;
    } else {
      // get surface address and line increment
      offset = (unsigned char *) ddsd[this->ID].lpSurface;
      pitch  = ddsd[this->ID].lPitch;
      incr   = (ddsd[this->ID].ddpfPixelFormat.dwRGBBitCount) / 8;
    }

    // convert to a displayable RGB format
    switch (ddsd[this->ID].ddpfPixelFormat.dwRGBBitCount) {
  case 16:
    {
      switch(pixelArray->format) {
  case(RGB):
  case (YUV):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++){
          *(pixmap  ) = (unsigned char)((*(src+2)>>3) | (*(src+1)<<3)&(0xE0));
          *(pixmap+1) = (unsigned char)((*(src+1)>>5) | (*(src  )   )&(0xF8));
          src    += TL_RGB_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case (GRAY):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++){
          *(pixmap  ) = (unsigned char)((*(src)>>3) | (*(src)<<3)&(0xE0));
          *(pixmap+1) = (unsigned char)((*(src)>>5) | (*(src)   )&(0xF8));
          src    += TL_GRAY_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(HSI):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) = (unsigned char)((*(src)>>3) | (*(src)<<3)&(0xE0));
          *(pixmap+1) = (unsigned char)((*(src)>>5) | (*(src  ) )&(0xF8));
          src    += TL_HSI_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(BIN):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          if ((*src) != 0) {
            *(pixmap  ) = 0xFF;
            *(pixmap+1) = 0xFF;
          } else {
            *(pixmap  ) = 0x0;
            *(pixmap+1) = 0x0;
          }
          src    += TL_BIN_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(NRG):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) = (unsigned char)((*(src+1)<<3)&(0xE0));
          *(pixmap+1) = (unsigned char)((*(src+1)>>5) | (*(src))&(0xF8));
          src    += TL_NRG_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(STEREO):
    {
      tlHRPixel *hrsrc = pixelArray->hrpixel;
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          if (*(hrsrc) >= 0) {
            *(pixmap  ) = (unsigned char)((*(hrsrc) >> 10) | (*(hrsrc) << 10)&(0xE0));
            *(pixmap+1) = (unsigned char)((*(hrsrc) >> 12) | (*(hrsrc) <<  7)&(0xF8));
          }
          else {
            *(pixmap  ) = 0;
            *(pixmap+1) = 0;
          }
          hrsrc  += TL_STEREO_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  default:
    {
      TL_ERROR_MSG ("tlDisplay::display() : unknown image format");
      return TL_ERROR_INVALID_FORMAT;
    }
      }
    }
    break;

  default:
    {
      switch(pixelArray->format) {
  case(RGB):
  case(YUV):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) = *(src+2);
          *(pixmap+1) = *(src+1);
          *(pixmap+2) = *(src  );
          src    += TL_RGB_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(GRAY):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) =
            *(pixmap+1) =
            *(pixmap+2) = *(src);
          src    += TL_GRAY_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(HSI):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) =
            *(pixmap+1) =
            *(pixmap+2) = *(src);
          src    += TL_HSI_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(BIN):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          if ((*src) != 0) {
            *(pixmap  ) =
              *(pixmap+1) =
              *(pixmap+2) = TL_MAX_PIXEL_VAL;
          } else {
            *(pixmap  ) =
              *(pixmap+1) =
              *(pixmap+2) = TL_MIN_PIXEL_VAL;
          }
          src    += TL_BIN_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(NRG):
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) = 0;
          *(pixmap+1) = *(src+1);
          *(pixmap+2) = *(src);
          src    += TL_NRG_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  case(STEREO):
    {
      tlHRPixel *hrsrc = pixelArray->hrpixel;
      unsigned char shift = pixelArray->hrShift;
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          if (*(hrsrc) < 0) {
            *(pixmap  ) =
            *(pixmap+1) =
            *(pixmap+2) = 0;
          }
          else {
            *(pixmap  ) =
            *(pixmap+1) =
            *(pixmap+2) = (unsigned char) ((*(hrsrc)) >> shift);
          }
          hrsrc  += TL_STEREO_WIDTH;
          pixmap += incr;
        }
      }
    }
    break;

  default:
    {
      TL_ERRNO_MSG ("tlDisplay::display()", TL_ERROR_INVALID_FORMAT);
      return -1;
    }
      }
    }
    }

    // unlock intermediary surface
    _tl_lpImage[this->ID]->Unlock(NULL);

    // blit image onto primary surface (i.e. directly to screen video memory)
    _tl_lpDDSPrimary[this->ID]->Blt (
      &(_tl_display_dx3_rect[this->ID]),
      _tl_lpImage[this->ID], 
      NULL, 
      NULL, 
      NULL);


  return update ();
}



///  Display a histogram in the display area.
///
///  \param hist     the histogram to display
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::display (tlHist *hist)
{
  int                   i,j;
  unsigned char *       offset;
  tlHistData            *src;
  unsigned char*        pixmap;
  int                   incr, pitch;
  static DDSURFACEDESC  ddsd[TL_MAX_DISPLAY];


  // check if the window is visible (!)
  if (!visible) 
    return 0;

  // check image integrity
  if (!hist) {
    TL_ERRNO_MSG ("tlDisplay::display()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check image size
  if ((hist->getLength()  != width) ||
    (((hist->getMaxVal())-(hist->getMinVal())) != height)) {
      if (0 > resize(hist->getLength(),(hist->getMaxVal())-(hist->getMinVal()))) {
        return -1;
      }
    }

    // assign some useful vars
    src = hist->data;

    // lock surface
    ZeroMemory (&(ddsd[this->ID]), sizeof(ddsd[this->ID]));
    ddsd[this->ID].dwSize = sizeof(ddsd[this->ID]);
    HRESULT res = _tl_lpImage[this->ID]->Lock (
      NULL, 
      &(ddsd[this->ID]), 
      DDLOCK_WRITEONLY | DDLOCK_SURFACEMEMORYPTR, 
      NULL);

    // make sure it worked
    if (DD_OK != res) {
      TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
      return -1;
    } else {
      // get surface address and line increment
      offset = (unsigned char *) ddsd[this->ID].lpSurface;
      pitch  = ddsd[this->ID].lPitch;
      incr   = (ddsd[this->ID].ddpfPixelFormat.dwRGBBitCount) / 8;
    }

    // convert to a displayable RGB format
    switch (ddsd[this->ID].ddpfPixelFormat.dwRGBBitCount) {
  case 16:
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          if (*(src) < (height-j)) {
            *(pixmap  ) = 0;
            *(pixmap+1) = 0;
          } 
          else {
            *(pixmap  ) = 255;
            *(pixmap+1) = 255;
          }
          src    -= 1;
          pixmap += incr;
        }
        src = hist->data;
      }
    }
  default:
    {
      for (j=0; j<height; j++) {
        src    = hist->data;
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          if (*(src++) < (height-j)) {
            *(pixmap  ) = 0;
            *(pixmap+1) = 0;
            *(pixmap+2) = 0;
          } 
          else {
            *(pixmap  ) = 255;
            *(pixmap+1) = 255;
            *(pixmap+2) = 255;
          }
          pixmap += incr;
        }
      }
    }
    }

    // unlock intermediary surface
    _tl_lpImage[this->ID]->Unlock(NULL);

    // blit image onto primary surface (i.e. directly to screen video memory)
    _tl_lpDDSPrimary[this->ID]->Blt (
      &(_tl_display_dx3_rect[this->ID]),
      _tl_lpImage[this->ID], 
      NULL, 
      NULL, 
      NULL);


  return update ();
}



///  Display a 2D histogram in the display area.
///
///  \param hist     the histogram to display
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDisplay::display (tlHist2D *hist)
{
  int                   i,j;
  unsigned char *       offset;
  tlHistData            **src;
  unsigned char*        pixmap;
  int                   incr, pitch;
  static DDSURFACEDESC  ddsd[TL_MAX_DISPLAY];


  // check if the window is visible (!)
  if (!visible) 
    return 0;

  // check image integrity
  if (!hist) {
    TL_ERRNO_MSG ("tlWindow->display", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check image size
  if ((hist->getXLength () != width) ||
    (hist->getYLength () != height)) {
      if (0 > resize (hist->getXLength (),hist->getYLength ())) return -1;
    }

    // assign some useful vars
    src = hist->data;

    // lock surface
    ZeroMemory (&(ddsd[this->ID]), sizeof(ddsd[this->ID]));
    ddsd[this->ID].dwSize = sizeof(ddsd[this->ID]);
    HRESULT res = _tl_lpImage[this->ID]->Lock (
      NULL, 
      &(ddsd[this->ID]), 
      DDLOCK_WRITEONLY | DDLOCK_SURFACEMEMORYPTR, 
      NULL);

    // make sure it worked
    if (DD_OK != res) {
      TL_ERRNO_MSG ("tlDisplay::create()", TL_ERROR_CONFIG_FAILED);
      return -1;
    } else {
      // get surface address and line increment
      offset = (unsigned char *) ddsd[this->ID].lpSurface;
      pitch  = ddsd[this->ID].lPitch;
      incr   = (ddsd[this->ID].ddpfPixelFormat.dwRGBBitCount) / 8;
    }

    // convert to a displayable RGB format
    switch (ddsd[this->ID].ddpfPixelFormat.dwRGBBitCount) {
  case 16:
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) = 
            *(pixmap+1) = (unsigned char)src[i][j];
          pixmap += incr;
        }
      }
    }
  default:
    {
      for (j=0; j<height; j++) {
        pixmap = offset + j*pitch;
        for (i=0; i<width; i++) {
          *(pixmap  ) = 
            *(pixmap+1) = 
            *(pixmap+2) = (unsigned char)src[i][j];
          pixmap += incr;
        }
      }
    }
    }

    // unlock intermediary surface
    _tl_lpImage[this->ID]->Unlock(NULL);

    // blit image onto primary surface (i.e. directly to screen video memory)
    _tl_lpDDSPrimary[this->ID]->Blt (
      &(_tl_display_dx3_rect[this->ID]),
      _tl_lpImage[this->ID], 
      NULL, 
      NULL, 
      NULL);


  return update ();
}

