/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlDisplay.h"
#include "tlPoint.h"
#include "tlRect.h"
#include "tlDisplay_dx3.h"

#include <windows.h>



///  \def TL_EVENT_DELAY
///  Minimum delay between two events.

///  \def TL_EVENT_JIGGLE_MAX
///  Mouse click jiggle tolerance.

///  \def TL_NO_EVENT
///  No event occured.

///  \def TL_EVENT
///  Some event occured.

///  \def TL_EVENT_EXIT
///  Program exit event.

///  \def TL_EVENT_MOUSE
///  Mouse event

///  \def TL_EVENT_MOUSEDOWN
///  Mouse button pressed event.

///  \def TL_EVENT_MOUSEUP
///  Mouse button released event.

///  \def TL_EVENT_LMOUSE
///  Mouse left button event

///  \def TL_EVENT_LMOUSEDOWN
///  Mouse left button pressed event.

///  \def TL_EVENT_LMOUSEUP
///  Mouse left button released event.

///  \def TL_EVENT_RMOUSE
///  Mouse right button event

///  \def TL_EVENT_RMOUSEDOWN
///  Mouse right button pressed event.

///  \def TL_EVENT_RMOUSEUP
///  Mouse right button released event.

///  \def TL_EVENT_MMOUSE
///  Mouse middle button event

///  \def TL_EVENT_MMOUSEDOWN
///  Mouse middle button pressed event.

///  \def TL_EVENT_MMOUSEUP
///  Mouse middle button released event.

///  \def TL_EVENT_MOUSEDBLCLK
///  Mouse button double-clicked event.

///  \def TL_EVENT_KEY
///  Keyboard event.

///  \def TL_EVENT_KEYDOWN
///  Keyboard key pressed event.


// events management
bool tlDisplay::init_global = false;
static bool       _tl_event_is_mouserect    ;
static bool       _tl_event_is_doubleclick  ;
static int        _tl_event_last_event      ;
static tlPoint*   _tl_event_mouse           ;
static tlPoint*   _tl_event_mouse_down      ;
static tlPoint*   _tl_event_mouse_up        ;
static tlPoint*   _tl_event_mouse_double    ;
static tlRect*    _tl_event_mouse_rect      ;
static char       _tl_event_key_down        ;
static char       _tl_event_key_up          ;
static tlDisplay* _tl_event_last_display    ;

// safety mutex for multiple display handling
static HANDLE _tl_display_mutex = CreateMutex (NULL, false, "TLIB display mutex");



///  Window message processing callback
///
///  \return
///  Windows defined

long _stdcall
_tlCtlProc (HWND   hWnd,
            UINT   msg,
            WPARAM wparam,
            LPARAM lparam)
{
  int k;


  tlDisplay::initialize ();

  // gain exclusive access to the callback and globals
  WaitForSingleObject (_tl_display_mutex, INFINITE);

  // identify window
  for (k=0; k<TL_MAX_DISPLAY; k++) {
    if (_tl_display_dx3_list[k] != 0) {
      if (_tl_display_dx3_window[k] == hWnd) {
        _tl_event_last_display = _tl_display_dx3_ptr[k];
        break;
      }
    }
  }

  // process messages
  switch (msg) {
  case WM_MOUSEMOVE:
    {
      _tl_event_mouse->x = (int) (LOWORD(lparam) * _tl_event_last_display->getWidth()  / _tl_display_dx3_actual_width [k]);
      _tl_event_mouse->y = (int) (HIWORD(lparam) * _tl_event_last_display->getHeight() / _tl_display_dx3_actual_height[k]);
    }
    break;

  case WM_QUIT:
    _tl_event_last_event = TL_EVENT_EXIT;
    break;

  case WM_DESTROY:
    // return 0;
    _tl_event_last_event = TL_EVENT_EXIT;
    PostQuitMessage(0);
    break;

  case WM_SIZE:
    _tl_display_dx3_actual_width [k] = LOWORD(lparam);
    _tl_display_dx3_actual_height[k] = HIWORD(lparam);
    break;

  case WM_KEYDOWN:
    // the actual key (ASCII code) is set by WM_CHAR
    _tl_event_last_event = TL_EVENT_KEYDOWN;
    break;

  case WM_KEYUP:
    // the actual key (ASCII code) is set by WM_CHAR
    _tl_event_last_event = TL_EVENT_KEYUP;
    break;

  case WM_CHAR:
    _tl_event_key_down = (char) wparam;
    _tl_event_key_up   = (char) wparam;
    break;

  case WM_LBUTTONDOWN:
  case WM_MBUTTONDOWN:
  case WM_RBUTTONDOWN:
    switch (msg) {
    case WM_LBUTTONDOWN:
      _tl_event_last_event = TL_EVENT_LMOUSEDOWN;
      break;
    case WM_MBUTTONDOWN:
      _tl_event_last_event = TL_EVENT_MMOUSEDOWN;
      break;
    case WM_RBUTTONDOWN:
      _tl_event_last_event = TL_EVENT_RMOUSEDOWN;
      break;
    }
    _tl_event_mouse_down->x = (int) (LOWORD(lparam) * _tl_event_last_display->getWidth()  / _tl_display_dx3_actual_width [k]);
    _tl_event_mouse_down->y = (int) (HIWORD(lparam) * _tl_event_last_display->getHeight() / _tl_display_dx3_actual_height[k]);
    break;

  case WM_LBUTTONUP:
  case WM_MBUTTONUP:
  case WM_RBUTTONUP:
    switch (msg) {
    case WM_LBUTTONUP:
      _tl_event_last_event = TL_EVENT_LMOUSEUP;
      break;
    case WM_MBUTTONUP:
      _tl_event_last_event = TL_EVENT_MMOUSEUP;
      break;
    case WM_RBUTTONUP:
      _tl_event_last_event = TL_EVENT_RMOUSEUP;
      break;
    }
    if (_tl_event_is_doubleclick) {
      _tl_event_is_doubleclick = false;
      break;
    }
    _tl_event_mouse_up->x = (int) (LOWORD(lparam) * _tl_event_last_display->getWidth()  / _tl_display_dx3_actual_width [k]);
    _tl_event_mouse_up->y = (int) (HIWORD(lparam) * _tl_event_last_display->getHeight() / _tl_display_dx3_actual_height[k]);
    _tl_event_mouse_rect->set (_tl_event_mouse_down, _tl_event_mouse_up);

    // check if we have a rect or a point
    if ((_tl_event_mouse_rect->width <= TL_EVENT_JIGGLE_MAX) && (_tl_event_mouse_rect->height <= TL_EVENT_JIGGLE_MAX)) {
      _tl_event_is_mouserect = false;
    }
    else {
      _tl_event_last_event = TL_EVENT_MOUSERECT;
      _tl_event_is_mouserect = true;
    }
    break;

  case WM_LBUTTONDBLCLK:
  case WM_MBUTTONDBLCLK:
  case WM_RBUTTONDBLCLK:
    _tl_event_last_event = TL_EVENT_MOUSEDBLCLK;
    _tl_event_is_doubleclick = true;
    _tl_event_mouse_double->x = (int) (LOWORD(lparam) * _tl_event_last_display->getWidth()  / _tl_display_dx3_actual_width [k]);
    _tl_event_mouse_double->y = (int) (HIWORD(lparam) * _tl_event_last_display->getHeight() / _tl_display_dx3_actual_height[k]);
    break;

  case WM_WINDOWPOSCHANGED:
  case WM_SHOWWINDOW:
    {
      RECT *rect = &(_tl_display_dx3_rect[k]);
      GetWindowRect (_tl_display_dx3_window[k], rect);
      rect->top    += _TL_BORDER_TOP;
      rect->bottom -= _TL_BORDER_BOTTOM;
      rect->left   += _TL_BORDER_LEFT;
      rect->right  -= _TL_BORDER_RIGHT;
    }
    break;

  default:
    _tl_event_last_event = TL_NO_EVENT;
    break;
  }

  // release exclusive access to the callback and globals
  ReleaseMutex (_tl_display_mutex);


  // call default for unprocessed messages
  return DefWindowProc(hWnd,msg,wparam,lparam);
}



///  Check if an event has occured and is in the event queue.
///
///  \return
///  TL_EVENT if an event occured,\n
///  TL_NO_EVENT otherwise.

int
tlDisplay::queryEvent (void)
{
  static MSG msg;


  if(PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE) != 0) return TL_EVENT;


  return TL_NO_EVENT;
}



///  Process TL library windows events;
///  result is stored in TL specific data structure.
///  This is a blocking function.
///
///  \return
///  TL_EVENT status (see \ref tlDisplay.h for details)

int
tlDisplay::event (void)
{ 
  static MSG msg;


  if (0 == PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE)) {
    WaitMessage();
  }

  if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
    if ((msg.message == WM_QUIT) || (msg.message == WM_DESTROY) || (msg.message == WM_CLOSE))
      return TL_EVENT_EXIT;
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }


  return _tl_event_last_event;
}



///  Return last received event. Is set to TL_NO_EVENT once called.
///
///  \return
///  TL_EVENT status (see tlDisplay.h for details)
///  TL_NO_EVENT if there is no new event

int 
tlDisplay::getLastEvent ()
{
  int res;


  res = _tl_event_last_event;
  _tl_event_last_event = TL_NO_EVENT;


  return res;
}



///  Return a pointer to the display that generated the 
///  last processed event.
///
///  \return
///  A pointer to the display that received the last event 
///  (as given by tlDisplay::getLastEvent()),\n
///  NULL if no display is associated with the event.

tlDisplay* 
tlDisplay::getLastDisplay ()
{
  return (_tl_event_last_display);
}



///  Get mouse coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
///
///  \return
///  A pointer to a \ref tlPoint object.

tlPoint*
tlDisplay::getMouse ()
{
  if (!init_global) initialize ();

  return _tl_event_mouse->copy ();
}



///  Get mouse coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
///
///  \param point    \b [out] mouse coordonates
///
///  \return
///  A pointer to a \ref tlPoint object.

int
tlDisplay::getMouse (tlPoint *point)
{
  if (!init_global) initialize ();

  return _tl_event_mouse->copyTo (point);
}



///  Get mouse down click coordonates. User can have clicked in *any* 
///  TL window. Use tlDisplay::getLastDisplay() to retrieve the window 
///  that generated the event.
///
///  \return
///  A pointer to a \ref tlPoint object.

tlPoint*
tlDisplay::getMouseDown ()
{
  if (!init_global) initialize ();

  return _tl_event_mouse_down->copy ();
}



///  Get mouse down click coordonates. User can have clicked in *any* 
///  TL window. Use tlDisplay::getLastDisplay() to retrieve the window 
///  that generated the event.
///
///  \param point    \b [out] mouse coordonates
///
///  \return
///  A pointer to a \ref tlPoint object.

int
tlDisplay::getMouseDown (tlPoint *point)
{
  if (!init_global) initialize ();

  return _tl_event_mouse_down->copyTo (point);
}



///  Get mouse up click coordonates. User can have clicked in *any* 
///  TL window. Use tlDisplay::getLastDisplay() to retrieve the window 
///  that generated the event if necessary.
///
///  \return
///  A pointer to a \ref tlPoint object.

tlPoint*
tlDisplay::getMouseUp ()
{
  if (!init_global) initialize ();

  return _tl_event_mouse_up->copy();
}



///  Get mouse up click coordonates. User can have clicked in *any* 
///  TL window. Use tlDisplay::getLastDisplay() to retrieve the window 
///  that generated the event if necessary.
///
///  \param point    \b [out] mouse coordonates
///
///  \return
///  A pointer to a \ref tlPoint object.

int
tlDisplay::getMouseUp (tlPoint *point)
{
  if (!init_global) initialize ();

  return _tl_event_mouse_up->copyTo (point);
}



///  Get mouse double click coordonates. User can have clicked in *any* 
///  TL window. Use tlDisplay::getLastDisplay() to retrieve the window 
///  that generated the event if necessary.
///
///  \return
///  A pointer to a \ref tlPoint object.

tlPoint*
tlDisplay::getMouseDouble ()
{
  if (!init_global) initialize ();

  return _tl_event_mouse_double->copy ();
}



///  Get mouse double click coordonates. User can have clicked in *any* 
///  TL window. Use tlDisplay::getLastDisplay() to retrieve the window 
///  that generated the event if necessary.
///
///  \param point    \b [out] mouse coordonates
///
///  \return
///  A pointer to a \ref tlPoint object.

int
tlDisplay::getMouseDouble (tlPoint *point)
{
  if (!init_global) initialize ();

  return _tl_event_mouse_double->copyTo (point);
}



///  Get mouse defined rectangle coordonates. User can have clicked 
///  in \c any TL window. Use tlDisplay::getLastDisplay() to retrieve 
///  the window that generated the event if necessary.
///
///  \return
///  A pointer to a \ref tlPoint object.

tlRect*
tlDisplay::getMouseRect ()
{
  if (!init_global) initialize ();

  return  _tl_event_mouse_rect->copy ();
}



///  Get mouse defined rectangle coordonates. User can have clicked 
///  in \c any TL window. Use tlDisplay::getLastDisplay() to retrieve 
///  the window that generated the event if necessary.
///
///  \param point    \b [out] mouse defined rectangle coordonates
///
///  \return
///  A pointer to a \ref tlPoint object.

int
tlDisplay::getMouseRect (tlRect *rect)
{
  if (!init_global) initialize ();

  return _tl_event_mouse_rect->copyTo (rect);
}



///  Get keycode for keyboard downstroke.
///
///  \return
///  The character ASCII code.

char
tlDisplay::getKeyDown ()
{
  if (!init_global) initialize ();

  return _tl_event_key_down;
}



///  Get keycode for keyboard uptroke.
///
///  \return
///  The character ASCII code.

char
tlDisplay::getKeyUp ()
{
  if (!init_global) initialize ();

  return _tl_event_key_up;
}



///  Warp the mouse pointer to an absolute screen position. 
///  (x, y) = (0, 0) = top-left corner
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlDisplay::setMousePos (int x, 
                        int y)
{
  if (SetCursorPos (x, y)) {
    return (0);
  }
  else {
    TL_ERRNO_MSG ("tlDisplay::setMousePos()", TL_ERROR_NOT_IN_IMAGE);
    return (-1);
  }
}



///  Wait for a keypress (keyboard upstroke)
///
///  \return
///  The character ASCII code.

char
tlDisplay::waitKey (void)
{

  if (!init_global) initialize ();

  while (event() != TL_EVENT_KEYUP) {};
  return (getKeyUp());
}



///  Wait for a mouse click (button down)
///
///  \return
///  The character ASCII code.

void
tlDisplay::waitMouse (void)
{
  int done = 0;


  if (!init_global) initialize ();

  while (!done) {
    switch (event()) {
    case TL_EVENT_MOUSEUP:
    case TL_EVENT_LMOUSEUP:
    case TL_EVENT_RMOUSEUP:
    case TL_EVENT_MMOUSEUP:
      done = 1;
      break;
    default:
      break;
    }
  }
}



inline void
tlDisplay::initialize ()
{
  if (!init_global) {
    _tl_event_is_mouserect    = false;
    _tl_event_is_doubleclick  = false;
    _tl_event_last_event      = TL_NO_EVENT;
    _tl_event_mouse           = new tlPoint ();
    _tl_event_mouse_down      = new tlPoint ();
    _tl_event_mouse_up        = new tlPoint ();
    _tl_event_mouse_double    = new tlPoint ();
    _tl_event_mouse_rect      = new tlRect  ();
    _tl_event_key_down        = -1;
    _tl_event_key_up          = -1;
    _tl_event_last_display    = NULL;
    init_global = true;
  }
}
