/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlDigitizerDT3130.h"
#include "tlImage.h"



///  \file tlDigitizerDT3130.h
///  \brief DatX DT3130 digitizer interface.

///  \class tlDigitizerDT3130
///  DatX DT3130 digitizer interface for TLIB.



///  Default constructor. Uses default parameters.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerDT3130::tlDigitizerDT3130 () : tlDigitizer ()
{
  if (tl_errno != TL_NO_ERROR) return;
}



///  Prepare a digitizer for image acquisition using a given acquisition size.
///  A subsequent call to open() will open the digitizer with the given settings.
///
///  \param width           acquisition width
///  \param height          acquisition height
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerDT3130::tlDigitizerDT3130 (int width,
                                      int height) : tlDigitizer ()
{
  if (tl_errno != TL_NO_ERROR) return;
  this->width   = width;
  this->height  = height;
}



///  Prepare a digitizer for image acquisition using a given acquisition size
///  and specific format parameters. A subsequent call to open() will open
///  the digitizer with the given settings.
///
///  \param width     new acquisition width
///  \param height    new acquisition height
///  \param channel   channel index on the digitizer
///  \param input     input format
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerDT3130::tlDigitizerDT3130 (int            width,
                                      int            height,
                                      int            channel,
                                      unsigned short input) : tlDigitizer ()
{
  if (tl_errno != TL_NO_ERROR) return;
  this->width   = width;
  this->height  = height;
  this->channel = channel;
  this->input   = input;
}



///  Default destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerDT3130::~tlDigitizerDT3130 ()
{
  close ();
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerDT3130::open (int index)
{
  return open (width, height, index, channel, input);
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerDT3130::open (int width, 
                         int height,
                         int index)
{
  return open (width, height, index, channel, input);
}



///  Open a digitizer for image acquisition using a given acquisition size
///  and specific format parameters.
///
///  \param width           new acquisition width
///  \param height          new acquisition height
///  \param index           the index of the digitizer in the system 
///                         (for use of multiple digitizers at the same time)
///  \param channel         channel index on the digitizer
///  \param input           input format
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDigitizerDT3130::open (int            width, 
                         int            height,
                         int            index,
                         int            channel,
                         unsigned short input)
{
  int               deviceCount;
  LPOLT_IMGDEVINFO  lpDevInfoList;
  ULNG	            oldValue;
  OLT_LNG_RANGE	    maxHeight, maxWidth;
  OLT_SCALE_PARAM   scale;


  if (isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // save configuration
  this->width       = width;
  this->height      = height;

  // how MANY digitizers are available ?
  if (OlImgGetDeviceCount(&deviceCount) != OLC_STS_NORMAL) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // currently, only support a single DT3130
  if (index > 0) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // allocate space for info
  lpDevInfoList = (LPOLT_IMGDEVINFO) VirtualAlloc(NULL, deviceCount * sizeof(OLT_IMGDEVINFO), MEM_COMMIT, PAGE_READWRITE);
  if (lpDevInfoList == NULL) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // fill in the device info
  lpDevInfoList->StructSize = sizeof(OLT_IMGDEVINFO);
  if (OlImgGetDeviceInfo(lpDevInfoList, deviceCount * sizeof(OLT_IMGDEVINFO))) {
    VirtualFree(lpDevInfoList, deviceCount * sizeof(OLT_IMGDEVINFO), MEM_DECOMMIT);
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // try to to open the device	
  if (!OlImgIsOkay(OlImgOpenDevice(lpDevInfoList->Alias, &devId))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // store the name of the device
  strncpy(name, lpDevInfoList->Alias, OLC_MAX_ALIAS_STR_SIZE);

  // free memory
  VirtualFree(lpDevInfoList, deviceCount * sizeof(OLT_IMGDEVINFO), MEM_DECOMMIT);

  // set a 5 second timeout for driver calls (useful ?? beuh...)
  if (!OlImgIsOkay(OlImgSetTimeoutPeriod(devId, 5, &oldValue))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set channel 
  if (0 != setChannel (channel, input)) {
    return TL_ERROR;
  }

  // check consitency of height value	
  if (!OlImgIsOkay(OlFgQueryInputCaps(devId, OLC_FG_IC_FRAME_HEIGHT_LIMITS, &maxHeight, sizeof(OLT_LNG_RANGE)))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  if (this->height < maxHeight.Min || this->height > maxHeight.Max) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // check consitency of width value	
  if (!OlImgIsOkay(OlFgQueryInputCaps(devId, OLC_FG_IC_FRAME_WIDTH_LIMITS, &maxWidth, sizeof(OLT_LNG_RANGE)))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  if (this->width < maxWidth.Min || this->width > maxWidth.Max) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set the scaling ratio
  scale.hscale = (unsigned short) floor(100.0 * this->width / maxWidth.Max);
  scale.vscale = (unsigned short) floor(100.0 * this->height / maxHeight.Max);
  if (!OlImgIsOkay(DtColorHardwareScaling(devId, channel, OLC_WRITE_CONTROL, &scale))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::passthruStart()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set the desired height and width frame
  if (!OlImgIsOkay(OlFgSetInputControlValue(devId, (USHRT)channel, OLC_FG_CTL_FRAME_HEIGHT, this->height, (LPINT)&oldValue))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  if (!OlImgIsOkay(OlFgSetInputControlValue(devId, (USHRT)channel, OLC_FG_CTL_FRAME_WIDTH, this->width, (LPINT)&oldValue))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // allocate device frame memory
  if (!OlImgIsOkay(OlFgAllocateBuiltInFrame(devId, OLC_FG_DEV_MEM_VOLATILE, OLC_FG_NEXT_FRAME, &frameId))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // maps the frame buffer into the applications virtual address space.
  frameInfo.StructSize = sizeof(OLT_FG_FRAME_INFO);
  if (!OlImgIsOkay(OlFgMapFrame(devId, frameId, &(frameInfo)))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // notify
  isOpen = true;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDT3130::setSize (int width, 
                            int height)
{
  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::set()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  close ();

  return open (width, height);
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerDT3130::close ()
{
  BOOL          done;
  OLT_APISTATUS jobStatus;
  ULNG          bytesWritten;


  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::close()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // stop passthru
  if (jobPassId != 0) {
    if (!OlImgIsOkay(OlFgStopAsyncPassthru(devId, jobPassId, &jobStatus))) {
      TL_ERRNO_MSG ("tlDigitizerDT3130::close()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
    jobPassId = 0;
  }

  // wait for asyncronous job completed
  if (jobId != 0) {	
    do {
      if (!OlImgIsOkay(OlFgIsAsyncAcquireJobDone(devId, jobId, &done, &jobStatus, &bytesWritten))) {
        TL_ERRNO_MSG ("tlDigitizerDT3130::close()", TL_ERROR_CONFIG_FAILED);
        return -1;
      }
    } while (!done);
    jobId = 0;
  }

  // unmaps the frame buffer into the applications virtual address space.
  if (!OlImgIsOkay(OlFgUnmapFrame(devId, frameId, &(frameInfo)))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::close()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // deallocate frame
  OlFgDestroyFrame(devId, frameId);

  // close current device
  OlImgCloseDevice(devId);

  // notify
  isOpen = false;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDT3130::grab (tlImage *image)
{
  int                     i, j;
  register tlPixel*       data;
  register unsigned char* p;


  // do an Acquire
  if (!OlImgIsOkay(OlFgAcquireFrameToDevice(devId, frameId))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // initialize image to grabber format
  image->init(RGB,width,height);

  // get a pointer to the frame buffer
  p = (unsigned char *)(frameInfo.BaseAddress);

#ifdef TL_USE_x86_ASSEMBLY

  // assembly version
  int nbPix = height * width;
  data      = image->pixel + (nbPix-1)*TL_RGB_WIDTH-1;
  __asm {
    mov   esi, p      // move p to esi register
      mov   edi, data   // move data to edi register
      mov   ecx, nbPix  // move loop counter to ecx register
      std               // set direction flag
iploop:
    lodsd             // load dword esi into eax
      bswap eax         // swap byte orders in eax
      stosd             // store eax in dword edi
      inc   edi         // update pixel counter
      dec   ecx         // update loop counter
      jne   iploop
      cld
  }

#else

  // copy buffer onto globals data structure
  data = image->pixel;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }

#endif


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDT3130::grab (tlImage *image, 
                         tlRect  *rect)
{
  int                     i, j;
  int                     x1,x2,y1,y2;
  register tlPixel*       data;
  register unsigned char* p;


  // do an Acquire
  if (!OlImgIsOkay(OlFgAcquireFrameToDevice(devId, frameId))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // initialize image to grabber format
  image->init(RGB,width,height);

  // get a pointer to the frame buffer
  p = (unsigned char *)(frameInfo.BaseAddress);

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((rect->x)+(rect->width), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((rect->y)+(rect->height), height);


  // copy buffer onto globals data structure
  data = image->pixel;
  for (j=y1; j<y2; j++) {
    for (i=x1; i<x2; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDT3130::grabQueued (tlImage *image)
{
  int                       i, j;
  register tlPixel*         data;
  register unsigned char*   p;
  static OLT_FG_ACQJOB_ID   jobId;
  BOOL                      done;
  OLT_APISTATUS             jobStatus;
  ULNG                      written;


  // wait for acquisition to complete
  while (TRUE != OlFgIsAsyncAcquireJobDone (devId, jobId, &done, &jobStatus, &written)) Sleep (0);

  // initialize image to grabber format
  image->init(RGB,width,height);

  // get a pointer to the frame buffer
  p = (unsigned char *)(frameInfo.BaseAddress);

#ifdef TL_USE_x86_ASSEMBLY

  // assembly version
  int nbPix = height * width;
  data      = image->pixel + (nbPix-1)*TL_RGB_WIDTH-1;
  __asm {
    mov   esi, p      // move p to esi register
      mov   edi, data   // move data to edi register
      mov   ecx, nbPix  // move loop counter to ecx register
      std               // set direction flag
iploop:
    lodsd             // load dword esi into eax
      bswap eax         // swap byte orders in eax
      stosd             // store eax in dword edi
      inc   edi         // update pixel counter
      dec   ecx         // update loop counter
      jne   iploop
      cld
  }

#else

  // copy buffer onto globals data structure
  data = image->pixel;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }

#endif

  // start the next acquisition
  if (!OlImgIsOkay(OlFgAsyncAcquireFrameToDevice(devId, frameId, &jobId))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDT3130::grabQueued (tlImage *image, 
                               tlRect  *rect)
{
  int                       i, j;
  int                       x1,x2,y1,y2;
  register tlPixel*         data;
  register unsigned char*   p;
  static OLT_FG_ACQJOB_ID   jobId;
  BOOL                      done;
  OLT_APISTATUS             jobStatus;
  ULNG                      written;


  // wait for acquisition to complete
  while (TRUE != OlFgIsAsyncAcquireJobDone (devId, jobId, &done, &jobStatus, &written)) Sleep (0);

  // initialize image to grabber format
  image->init(RGB,width,height);

  // get a pointer to the frame buffer
  p = (unsigned char *)(frameInfo.BaseAddress);

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((rect->x)+(rect->width), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((rect->y)+(rect->height), height);

  // copy buffer onto globals data structure
  data = image->pixel;
  for (j=y1; j<y2; j++) {
    for (i=x1; i<x2; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }

  // start the next acquisition
  if (!OlImgIsOkay(OlFgAsyncAcquireFrameToDevice(devId, frameId, &jobId))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerDT3130::setChannel (int            channel,
                               unsigned short input)
{
  ULNG oldValue;
  ULNG sourceCount;


  // how many input sources does the device support ?
  OlFgQueryInputCaps(devId, OLC_FG_IC_INPUT_SOURCE_COUNT, &sourceCount, sizeof(ULNG));

  // check if the given channel is valid
  if ((unsigned)channel >= sourceCount) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::setInput()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set the channel
  if (!OlImgIsOkay(OlFgSetInputVideoSource(devId, (USHRT)channel, (LPUSHRT)&oldValue))) {
    TL_ERRNO_MSG ("tlDigitizerDT3130::setChannel()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // save parameters
  this->channel = channel;

  // set format
  setInput (input);


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerDT3130::setInput (unsigned short input)
{
  ULNG            oldValue;
  OLT_SIGNAL_TYPE	signal;


  // set the desired input format
  if (input & TL_SOURCE_COMPOSITE) {
    signal = OLC_COMPOSITE_SIGNAL;
    if (!OlImgIsOkay(DtColorSignalType (devId, channel, OLC_WRITE_CONTROL, &signal))) {
      TL_ERRNO_MSG ("tlDigitizerDT3130::setChannel()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
  }
  else if (input & TL_SOURCE_SVIDEO) {
    signal = OLC_YC_SIGNAL;
    if (!OlImgIsOkay (DtColorSignalType (devId, channel, OLC_WRITE_CONTROL, &signal))) {
      TL_ERRNO_MSG ("tlDigitizerDT3130::setChannel()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
  }

  // set the desired video format
  if (input & TL_SOURCE_PAL) {
    if (!OlImgIsOkay(OlFgSetInputControlValue(devId, (USHRT)channel, OLC_FG_CTL_INPUT_FILTER, OLC_FG_FILT_AC_50, (LPLNG) &oldValue))) {
      TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
  }
  else if (input & TL_SOURCE_NTSC) { 
    if (!OlImgIsOkay(OlFgSetInputControlValue(devId, (USHRT)channel, OLC_FG_CTL_INPUT_FILTER, OLC_FG_FILT_AC_60, (LPLNG) &oldValue))) {
      TL_ERRNO_MSG ("tlDigitizerDT3130::open()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
  }

  // save configuration
  this->input = input;


  return 0;
}
