/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlDigitizerDX.h"
#include "tlImage.h"



///  \file tlDigitizerDX.h
///  \brief DirectX digitizer interface.

///  \class tlDigitizerDX
///  DirectX digitizer interface for TLIB.



// DirectX headers
#include <atlbase.h>
#include <comip.h>
#include <qedit.h>
#include <dshow.h>
#include <streams.h>



class SampleGrabberCallback : public ISampleGrabberCB
{
private:
  bool *ready;

public:
  SampleGrabberCallback (bool *imageReady) { ready = imageReady; };
  virtual ~SampleGrabberCallback () {return; };
  STDMETHODIMP_(ULONG) AddRef  () { return 1; }
  STDMETHODIMP_(ULONG) Release () { return 2; }

  STDMETHODIMP QueryInterface(REFIID riid, void **ppvObject) {
    if (NULL == ppvObject) return E_POINTER;
    if (riid == __uuidof(IUnknown)) {
      *ppvObject = static_cast<IUnknown*>(this);
      return S_OK;
    }
    if (riid == __uuidof(ISampleGrabberCB)) {
      *ppvObject = static_cast<ISampleGrabberCB*>(this);
      return S_OK;
    }
    return E_NOTIMPL;
  }

  STDMETHODIMP SampleCB(double Time, IMediaSample *pSample) {
    return E_NOTIMPL;
  }

  STDMETHODIMP BufferCB(double Time, BYTE *pBuffer, long BufferLen) {
    *ready = true;
    return S_OK;
  }
};



class DXGrabObjects {
public:
  ISampleGrabber          *pGrabber;
  IGraphBuilder           *pGraph;
  ICaptureGraphBuilder2   *pCapture;
  IAMStreamConfig         *pVConfig;
  IMediaEvent             *pEvent;  
};



///  Prepare a digitizer for image acquisition using a given acquisition size.
///  A subsequent call to open() will open the digitizer with the given settings.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerDX::tlDigitizerDX () : tlDigitizer ()
{
  if (tl_errno == TL_NO_ERROR) return;

  width  = TL_DEFAULT_COLS;
  height = TL_DEFAULT_ROWS;
  format = RGB;
  
  fps = 30;
  cb  = NULL;
}



///  Prepare a digitizer for image acquisition using a given acquisition size.
///  A subsequent call to open() will open the digitizer with the given settings.
///
///  \param width           acquisition width
///  \param height          acquisition height
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerDX::tlDigitizerDX (int width, 
                              int height) : tlDigitizer (width, height)
{
  if (tl_errno != TL_ERROR) return;

  this->width  = width;
  this->height = height;
  this->format = RGB;
  
  fps = 30;
  cb  = NULL;
}



///  \note
///  See \ref errors "error management" for details.

tlDigitizerDX::~tlDigitizerDX ()
{
  if (cb) delete cb;
  tl_errno = TL_NO_ERROR;
  close ();
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerDX::open (int index)
{
  return open (width, height, index);
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDX::open (int width, 
                     int height,
                     int index)
{
  HRESULT hr;
  int     k;
  ULONG   cFetched;
  bool    found = false;

  // DirectX controls
  IMoniker                *pMoniker;
  CComPtr<IBaseFilter>    pFilterIn;
  IBaseFilter             *pF;
  IBaseFilter             *pNR;
  CMediaType              VideoType;
  IPin                    *pPinOut;
  IPin                    *pPinIn;

  grabber = new DXGrabObjects;

  // local callback
  imageReady = false;
  cb = new SampleGrabberCallback (&imageReady);

  // verbose
  if (tlVerbose) printf ("opening DirectX digitizer\n");

  // check
  if (isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_NOT_REQUIRED);
    return -1;
  }

  // store geometric vars
  this->width  = width;
  this->height = height;
  this->format = RGB;


  // 
  // initialize COM
  //

  CoInitialize(NULL);


  // 
  // create the filter graph builder
  //

  hr = CoCreateInstance (CLSID_FilterGraph, NULL, CLSCTX_INPROC, IID_IGraphBuilder, (void **) &(grabber->pGraph));
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // create the capture graph builder
  hr = CoCreateInstance(CLSID_CaptureGraphBuilder2 , NULL, CLSCTX_INPROC, IID_ICaptureGraphBuilder2, (void **) &(grabber->pCapture));
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // attach the filter graph to the capture graph
  hr = grabber->pCapture->SetFiltergraph(grabber->pGraph);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }


  //
  // find a camera device
  //

  // create the system device enumerator
  CComPtr<ICreateDevEnum> pSysDevEnum;
  hr = pSysDevEnum.CoCreateInstance(CLSID_SystemDeviceEnum);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // create an enumerator for the video capture devices
  CComPtr<IEnumMoniker> pEnumCat = NULL;
  hr = pSysDevEnum->CreateClassEnumerator(CLSID_VideoInputDeviceCategory, &pEnumCat, 0);
  if (FAILED(hr) || !pEnumCat) {
    grabber->pGraph->Release();
    grabber->pCapture->Release();
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // enumerate the monikers.
  pEnumCat->Skip (index);
  for (k=0; (pEnumCat->Next (1,&pMoniker,&cFetched) == S_OK); k++) {
    if (pMoniker->IsRunning (0, 0, 0) == S_OK) {
      printf ("monitor is found to be running\n");
      continue;
    }
    found = true;
    this->index = index;
    break;
  }

  // release enumerator
  pSysDevEnum.Release();
  pEnumCat.Release();

  // check if a free device was found
  if (!found) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // display device name if necessary
  if (tlVerbose) {
    printf ("found ");
    std::flush(cout);
    CComPtr<IPropertyBag> pProp;
    pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void **) &pProp);
    VARIANT varName;
    VariantInit(&varName); 
    hr = pProp->Read(L"FriendlyName", &varName, 0); 
    if (SUCCEEDED(hr)) wprintf(L"%s\n", varName.bstrVal);
  }


  //
  // configure the device
  //

  // bind the moniker to the input filter
  hr = pMoniker->BindToObject(0,0,IID_IBaseFilter, (void**)&pFilterIn);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  if(!pFilterIn) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // add the input filter to the graph
  hr = grabber->pGraph->AddFilter(pFilterIn, L"Input Filter");
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // get the device configuration interface
  hr = grabber->pCapture->FindInterface(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video, pFilterIn, IID_IAMStreamConfig, (void **)&(grabber->pVConfig));
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set frame size
  AM_MEDIA_TYPE *pMediaType;
  grabber->pVConfig->GetFormat(&pMediaType);
  if (pMediaType->formattype != FORMAT_VideoInfo) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  (((VIDEOINFOHEADER *)(pMediaType->pbFormat))->bmiHeader).biWidth      = width;
  (((VIDEOINFOHEADER *)(pMediaType->pbFormat))->bmiHeader).biHeight     = height;
  (((VIDEOINFOHEADER *)(pMediaType->pbFormat))->bmiHeader).biSizeImage  = width*height*3;
  hr = grabber->pVConfig->SetFormat(pMediaType);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }


  //
  // create sample grabber
  //

  // sample grabber
  hr = CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (LPVOID *)&pF);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // get that thing
  hr = pF->QueryInterface(IID_ISampleGrabber, (void **)&(grabber->pGrabber));
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // add sample grabber to the graph
  hr = grabber->pGraph->AddFilter(pF, L"Grabber");
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // force it to connect to video, 24 bit
  VideoType.SetType    (&MEDIATYPE_Video);
  VideoType.SetSubtype (&MEDIASUBTYPE_RGB24);
  hr = grabber->pGrabber->SetMediaType (&VideoType);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // connect output pin from capture device to input pin on samplegrabber
  hr = grabber->pCapture->FindPin(pFilterIn, PINDIR_OUTPUT, &PIN_CATEGORY_CAPTURE, NULL, TRUE, 0, &pPinOut);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  hr = grabber->pCapture->FindPin(grabber->pGrabber, PINDIR_INPUT, NULL, NULL, TRUE, 0, &pPinIn);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  hr = grabber->pGraph->Connect(pPinOut, pPinIn);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // create NULL renderer and add to filtergraph
  hr = CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (LPVOID *)&pNR);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  hr = grabber->pGraph->AddFilter(pNR, L"NullRenderer");
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // connect output pin from samplegrabber to input pin on NULL renderer
  hr = grabber->pCapture->FindPin(grabber->pGrabber, PINDIR_OUTPUT, NULL, NULL, TRUE, 0, &pPinOut);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  hr = grabber->pCapture->FindPin(pNR, PINDIR_INPUT, NULL, NULL, TRUE, 0, &pPinIn);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  hr = grabber->pGraph->Connect(pPinOut, pPinIn);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }


  //
  // launch the beast
  //

  // allocate image buffer
  buffer = (BYTE *) malloc (width*height*TL_RGB_WIDTH);

  // attach event manager
  hr = grabber->pGraph->QueryInterface(IID_IMediaEvent,(void**)&(grabber->pEvent));
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // create the control media interface
  CComQIPtr<IMediaControl, &IID_IMediaControl> pControl = grabber->pGraph;

  // set buffering mode
  hr = grabber->pGrabber->SetBufferSamples (TRUE);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set callback
  hr = grabber->pGrabber->SetCallback (cb, 1);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // run the graph
  hr = pControl->Run();
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }


  //
  // cleanup
  //

  pMoniker->Release ();
  pFilterIn.Release ();
  isOpen = true;


  return 0;
}




////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDX::setSize (int width, 
                        int height)
{
   // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDX::set()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }


  close ();
  return open (width, height, index);
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerDX::close ()
{
  // silently ignore unnecessary close calls
  if (!isOpen) {
    return 0;
  }

  // stop the graph  

  // create the control media interface
  CComQIPtr<IMediaControl, &IID_IMediaControl> pControl = grabber->pGraph;
  pControl->Stop();
  if (buffer) free(buffer);

  // release the grabber and the graph
  grabber->pEvent->Release();
  grabber->pGrabber->Release();
  grabber->pCapture->Release();
  grabber->pGraph->Release();
  grabber->pVConfig->Release();

  // notify
  isOpen = false;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDX::grab (tlImage *image)
{
  int              i,j;
  long             size;
  HRESULT          hr;
  register BYTE    *buf;
  register tlPixel *pixel;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDX::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // reset image to RGB
  image->init (RGB, width, height);

  // shortcuts
  size   = width*height*3;
  buf    = buffer;
  pixel  = image->pixel;

  // wait for image completion
  while (!imageReady) Sleep (0);

  // copy the captured image to the buffer
  hr = grabber->pGrabber->GetCurrentBuffer (&size, (long *)buffer);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::grab()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // convert
  for (j=height-1; j>=0; j--) {
    register tlPixel *dest = pixel + j*width*TL_RGB_WIDTH;
    for (i=0; i<width; i++) {
      *(dest++) = *(buf+2);
      *(dest++) = *(buf+1);
      *(dest++) = *(buf  );
      buf += 3;
    }
  }

  imageReady = false;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDX::grab (tlImage *image, 
                     tlRect  *rect)
{
  int              i,j;
  long             size;
  HRESULT          hr;
  register BYTE    *buf;
  register tlPixel *pixel;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDX::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // reset image to RGB
  image->init (RGB, width, height);

  // shortcuts
  const int x1 = TL_MAX ((rect->x), 0);
  const int x2 = TL_MIN (((rect->x)+(rect->width)), width);
  const int y1 = TL_MAX ((rect->y), 0);
  const int y2 = TL_MIN (((rect->y)+(rect->height)), height);
  size         = width*height*3;
  buf          = buffer;
  pixel        = image->pixel;

  // wait for image completion
  while (!imageReady) Sleep (0);

  // copy the captured image to the buffer
  hr = grabber->pGrabber->GetCurrentBuffer(&size, (long *)buffer);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlDigitizerDX::grab()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // convert
  for (j=y2-1; j>=y1; j--) {
    register tlPixel *dest = pixel  + (j*width+x1)*TL_RGB_WIDTH;
    // register BYTE    *buf  = buffer + ((y2-1-j)*width+x1)*TL_RGB_WIDTH;
    register BYTE    *buf  = buffer + ((height - j-1)*width+x1)*TL_RGB_WIDTH;
    for (i=x1; i<x2; i++) {
      *(dest++) = *(buf+2);
      *(dest++) = *(buf+1);
      *(dest++) = *(buf  );
      buf += 3;
    }
  }

  imageReady = false;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDX::grabQueued (tlImage *image)
{
  return grab (image);
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDX::grabQueued (tlImage *image, 
                           tlRect  *rect)
{
  return grab (image, rect);
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerDX::setFrameRate (int framerate)
{
  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerDX::setFrameRate()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // stop the graph temporarily
  CComQIPtr<IMediaControl, &IID_IMediaControl> pControl = grabber->pGraph;
  if (FAILED(pControl->Stop())) {
    TL_ERRNO_MSG ("tlDigitizerDX::setFramerate()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // set frame rate
  AM_MEDIA_TYPE *pMediaType;
  grabber->pVConfig->GetFormat(&pMediaType);
  if (pMediaType->formattype != FORMAT_VideoInfo) {
    TL_ERRNO_MSG ("tlDigitizerDX::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  fps = framerate;
  ((VIDEOINFOHEADER *)(pMediaType->pbFormat))->AvgTimePerFrame = (LONGLONG)(10000000 / fps);
  if (FAILED(grabber->pVConfig->SetFormat(pMediaType))) {
    TL_ERRNO_MSG ("tlDigitizerDX::setFramerate()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // restart the graph
  if (FAILED(pControl->Run())) {
    TL_ERRNO_MSG ("tlDigitizerDX::setFramerate()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  

  return 0;
}