/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlDigitizerPXC200.h"
#include "tlImage.h"
#include "tlRect.h"

#include <windows.h>



#define TL_PXC_DLL_NT         "pxc2_NT.dll"
#define TL_PXC_DLL_9X         "pxc2_95.dll"
#define TL_FRAME_DLL          "frame_32.dll"
#define TL_GRAB_PIXEL_TYPE    PBITS_RGB24



///  \file tlDigitizerPXC200.h
///  \brief ImageNation PXC200 digitizer interface.

///  \class tlDigitizerPXC200
///  ImageNation PXC200 digitizer interface for TLIB.



///  Default constructor. Uses default parameters.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerPXC200::tlDigitizerPXC200 () : tlDigitizer ()
{
  if (tl_errno != TL_NO_ERROR) return;
  this->channel = 0;
}



///  Prepare a digitizer for image acquisition using a given acquisition size.
///  A subsequent call to open() will open the digitizer with the given settings.
///
///  \param width           acquisition width
///  \param height          acquisition height
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerPXC200::tlDigitizerPXC200 (int width, 
                                      int height) : tlDigitizer (width, height)
{
  if (tl_errno != TL_NO_ERROR) return;
  this->channel = 0;
}



///  Prepare a digitizer for image acquisition using a given acquisition size
///  and specific format parameters. A subsequent call to open() will open
///  the digitizer with the given settings.
///
///  \param width           new acquisition width
///  \param height          new acquisition height
///  \param channel         channel index on the digitizer
///  \param input           input format
///
///  \note
///  See \ref errors "error management" for details.
///
///  \sa
///  \ref tlDigitizer.h "TLIB input formats"

tlDigitizerPXC200::tlDigitizerPXC200 (int            width, 
                                      int            height,
                                      int            channel,
                                      unsigned short input) : tlDigitizer (width, height)
{
  if (tl_errno != TL_NO_ERROR) return;
  this->input   = input;
  this->channel = channel;
}



///  Default destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerPXC200::~tlDigitizerPXC200 ()
{
  tl_errno = TL_NO_ERROR;
  close ();
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerPXC200::open (int index)
{
  return open (width, height, index, channel, input);
}


////////////////////////////////////////////////////////////////////////

int
tlDigitizerPXC200::open (int width,
                         int height,
                         int index)
{
  return open (width, height, index, channel, input);
}



///  Open a digitizer for image acquisition using a given acquisition size
///  and specific format parameters.
///
///  \param width           new acquisition width
///  \param height          new acquisition height
///  \param index           the index of the digitizer in the system 
///                         (for use of multiple digitizers at the same time)
///  \param channel         channel index on the digitizer
///  \param input           input format
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDigitizerPXC200::open (int            width, 
                         int            height, 
                         int            index,
                         int            channel,
                         unsigned short input)
{
  short         input_format;
  short         px_cols = (short) width;
  short         px_rows = (short) height;
  int           runningNT;
  OSVERSIONINFO os;


  if (isOpen) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::open()", TL_ERROR_NOT_REQUIRED);
    return -1;
  }

  // init framegrabber handles
  fgHandle = 0;

  // init framebuffers
  buffer = 0L;

  // check input consistency
  if(px_cols < 1) px_cols = (short) TL_DEFAULT_COLS;
  if(px_rows < 1) px_rows = (short) TL_DEFAULT_ROWS;

  // initialize the hardware library
  memset(&os, NULL, sizeof(OSVERSIONINFO));
  os.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx(&os);
  runningNT = (os.dwPlatformId == VER_PLATFORM_WIN32_NT);
  if (runningNT) {
    if (!imagenation_OpenLibrary(TL_PXC_DLL_NT,&(pxc),sizeof(pxc))) {
      TL_ERRNO_MSG ("tlDigitizerPXC200::open()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
      return -1;
    }
  }
  else {
    if (!imagenation_OpenLibrary(TL_PXC_DLL_9X,&(pxc),sizeof(pxc))) {
      TL_ERRNO_MSG ("tlDigitizerPXC200::open()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
      return -1;
    }
  }

  // initialize the frame processing library
  if (!imagenation_OpenLibrary(TL_FRAME_DLL,&(frame),sizeof(pxc))) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::open()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return -1;
  }

  // allocate frame grabber 
  fgHandle = pxc.AllocateFG((short)index);
  if(!(fgHandle)) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // wait for CCIR autodetect and/or pick video format
  input_format = AUTO_FORMAT;
  if (input & TL_SOURCE_AUTO) { 
    Sleep(2500);
  }
  else if (input & (TL_SOURCE_PAL || TL_SOURCE_PALM || TL_SOURCE_PALN)) input_format = PAL_FORMAT;
  else if (input & TL_SOURCE_NTSC) input_format = NTSC_FORMAT;

  // detect video format
  pxc.SetVideoDetect (fgHandle,input_format);

  // set initial channel
  setChannel (channel, input);

  // geometry
  pxc.SetWidth       (fgHandle,px_cols);
  pxc.SetHeight      (fgHandle,px_rows);
  pxc.SetLeft        (fgHandle,0);
  pxc.SetTop         (fgHandle,0);
  pxc.SetXResolution (fgHandle,px_cols);
  pxc.SetYResolution (fgHandle,px_rows);

  // allocate a frame buffer for right camera
  buffer = pxc.AllocateBuffer(px_cols, px_rows, TL_GRAB_PIXEL_TYPE);
  if(!buffer) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::open()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // first queued grab
  pxc.Grab( fgHandle, buffer, QUEUED );

  // save configuration
  this->width   = (int) px_cols;
  this->height  = (int) px_rows;
  this->channel = channel;
  this->input   = input;


  return 0; 
}


////////////////////////////////////////////////////////////////////////

int 
tlDigitizerPXC200::setSize (int width, 
                            int height)
{
  short px_cols = (short) width;
  short px_rows = (short) height;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::open()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // check input consistency
  if(px_cols < 1) px_cols = (short) TL_DEFAULT_COLS;
  if(px_rows < 1) px_rows = (short) TL_DEFAULT_ROWS;

  // geometry
  pxc.SetWidth  (fgHandle,px_cols);
  pxc.SetHeight (fgHandle,px_rows);
  pxc.SetLeft   (fgHandle,0);
  pxc.SetTop    (fgHandle,0);
  pxc.SetXResolution (fgHandle,px_cols);
  pxc.SetYResolution (fgHandle,px_rows);

  // free frame buffers
  if (buffer) frame.FreeFrame (buffer);

  // allocate a frame buffer
  buffer = pxc.AllocateBuffer(px_cols, px_rows, TL_GRAB_PIXEL_TYPE);
  if(!(buffer)) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::reset()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // first queued grab
  pxc.Grab( fgHandle, buffer, QUEUED );

  // store image size
  this->width  = (int) px_cols;
  this->height = (int) px_rows;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerPXC200::close ()
{
  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::close()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // wait for the queued ops to be over
  pxc.WaitFinished(fgHandle,0);

  // free frame buffers
  if (buffer) frame.FreeFrame (buffer);

  // free framegrabber handles
  if (fgHandle) pxc.FreeFG (fgHandle);

  // close libraries
  imagenation_CloseLibrary (&(pxc));
  imagenation_CloseLibrary (&(frame));

  // notify
  isOpen = false;


  return 0; 
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerPXC200::grab (tlImage *image)
{
  register tlPixel*       data;
  register unsigned char* p;

  int                     i,j;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // initialize image to grabber format
  image->init (RGB,width,height);

  // grab commands
  pxc.Grab( fgHandle, buffer, 0);

  // wait until grab is finished
  pxc.WaitFinished (fgHandle,0);

  // copy buffer onto globals data structure
  p = (unsigned char *) frame.FrameBuffer(buffer);

#ifdef TL_USE_x86_ASSEMBLY

  // assembly version
  int nbPix = height * width;
  data      = image->pixel + (nbPix-1)*TL_RGB_WIDTH-1;
  __asm {
    mov   esi, p      // move p to esi register
      mov   edi, data   // move data to edi register
      mov   ecx, nbPix  // move loop counter to ecx register
      std               // set direction flag
iploop:
    lodsd             // load dword esi into eax
      bswap eax         // swap byte orders in eax
      stosd             // store eax in dword edi
      inc   edi         // update pixel counter
      dec   ecx         // update loop counter
      jne   iploop
      cld
  }

#else

  // copy buffer onto globals data structure
  data = image->pixel;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }

#endif


  return 0;
}




////////////////////////////////////////////////////////////////////////

int 
tlDigitizerPXC200::grab (tlImage *image, 
                         tlRect  *rect)
{
  register tlPixel*       data;
  register unsigned char* p;

  int                     i,j;
  int                     x1,x2;
  int                     y1,y2;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // initialize image to grabber format
  image->init (RGB,width,height);

  // queue new grab commands
  pxc.Grab( fgHandle, buffer, 0);

  // wait until grab is finished
  pxc.WaitFinished (fgHandle,0);

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((rect->x)+(rect->width), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((rect->y)+(rect->height), height);

  // copy buffer onto globals data structure
  data = image->pixel;
  p = (unsigned char *) frame.FrameBuffer(buffer);
  for (j=y1; j<y2; j++) {
    for (i=x1; i<x2; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }


  return 0;
}




////////////////////////////////////////////////////////////////////////

int 
tlDigitizerPXC200::grabQueued (tlImage *image)
{
  register tlPixel*       data;
  register unsigned char* p;

  int                     i,j;



  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // initialize image to grabber format
  image->init (RGB,width,height);

  // wait until grab is finished
  pxc.WaitFinished (fgHandle,0);


  // copy buffer onto globals data structure
  p = (unsigned char *) frame.FrameBuffer(buffer);

#ifdef TL_USE_x86_ASSEMBLY

  // assembly version
  int nbPix = height * width;
  data      = image->pixel + (nbPix-1)*TL_RGB_WIDTH-1;
  __asm {
    mov   esi, p      // move p to esi register
      mov   edi, data   // move data to edi register
      mov   ecx, nbPix  // move loop counter to ecx register
      std               // set direction flag
iploop:
    lodsd             // load dword esi into eax
      bswap eax         // swap byte orders in eax
      stosd             // store eax in dword edi
      inc   edi         // update pixel counter
      dec   ecx         // update loop counter
      jne   iploop
      cld
  }

#else

  // copy buffer onto globals data structure
  data = image->pixel;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }

#endif

  // queue new grab commands
  pxc.Grab( fgHandle, buffer, QUEUED );


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerPXC200::grabQueued (tlImage *image, 
                               tlRect  *rect)
{
  register tlPixel*       data;
  register unsigned char* p;

  int                     i,j;
  int                     x1,x2;
  int                     y1,y2;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerPXC200::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (x1+(rect->width), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (y1+(rect->height), height);

  // initialize image to grabber format
  image->init (RGB,width,height);

  // wait until grab is finished */
  pxc.WaitFinished (fgHandle,0);

  // copy buffer onto globals data structure
  data = image->pixel;
  p = (unsigned char *) frame.FrameBuffer(buffer);
  for (j=y1; j<y2; j++) {
    for (i=x1; i<x2; i++) {
      *(data+2) = *(p++);   // B
      *(data+1) = *(p++);   // G
      *(data  ) = *(p++);   // R
      data += 3;
    }
  }

  // queue new grab commands
  pxc.Grab( fgHandle, buffer, QUEUED );


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerPXC200::setChannel (int            channel,
                               unsigned short input)
{
  short input_format = AUTO_FORMAT;


  // wait for CCIR autodetect and/or pick video format
  if (input & TL_SOURCE_AUTO) { 
    Sleep(2500);
    input_format = AUTO_FORMAT;
  }
  else if (input & (TL_SOURCE_PAL || TL_SOURCE_PALM || TL_SOURCE_PALN)) input_format = PAL_FORMAT;
  else if (input & TL_SOURCE_NTSC) input_format = NTSC_FORMAT;

  // detect video format
  pxc.SetVideoDetect(fgHandle,input_format);

  // composite input
  if (input & TL_SOURCE_COMPOSITE) {
    pxc.SetCamera        (fgHandle,(short)channel, 0);
    pxc.SetChromaControl (fgHandle, NOTCH_FILTER|BW_DETECT);
  }

  // svideo input
  else if (input & TL_SOURCE_SVIDEO) {
    pxc.SetCamera        (fgHandle,(short)channel, 0);
    pxc.SetChromaControl (fgHandle,SVIDEO);
  }

  // save parameter
  this->channel = channel;
  this->input   = input;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlDigitizerPXC200::setInput (unsigned short input)
{
  short input_format = AUTO_FORMAT;


  // wait for CCIR autodetect and/or pick video format
  if (input & TL_SOURCE_AUTO) { 
    Sleep(2500);
    input_format = AUTO_FORMAT;
  }
  else if (input & (TL_SOURCE_PAL || TL_SOURCE_PALM || TL_SOURCE_PALN)) input_format = PAL_FORMAT;
  else if (input & TL_SOURCE_NTSC) input_format = NTSC_FORMAT;

  // detect video format
  pxc.SetVideoDetect(fgHandle,input_format);

  // composite input
  if (input & TL_SOURCE_COMPOSITE) {
    pxc.SetCamera        (fgHandle,(short)channel, 0);
    pxc.SetChromaControl (fgHandle, NOTCH_FILTER|BW_DETECT);
  }

  // svideo input
  else if (input & TL_SOURCE_SVIDEO) {
    pxc.SetCamera        (fgHandle,(short)channel, 0);
    pxc.SetChromaControl (fgHandle,SVIDEO);
  }

  // save parameter
  this->input  = input;


  return 0;
}
