/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlDigitizerVRm.h"
#include "tlImage.h"



///  \file tlDigitizerVRm.h
///  \brief VRMagic digitizer interface, optimized for the VRmAVC-1 model.

///  \class tlDigitizerVRm
///  VRMagic digitizer interface for TLIB.





///  Prepare a digitizer for image acquisition using a given acquisition size.
///  A subsequent call to open() will open the digitizer with the given settings.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizerVRm::tlDigitizerVRm () : tlDigitizer (TL_DIGITIZERVRM_DEFAULT_WIDTH, TL_DIGITIZERVRM_DEFAULT_HEIGHT)
{
  if (tl_errno == TL_NO_ERROR) return;

  width  = TL_DEFAULT_COLS;
  height = TL_DEFAULT_ROWS;
  format = RGB;
  
  fps = 30;
}



///  \note
///  See \ref errors "error management" for details.

tlDigitizerVRm::~tlDigitizerVRm ()
{
  tl_errno = TL_NO_ERROR;
  close ();
}




////////////////////////////////////////////////////////////////////////

int 
tlDigitizerVRm::open (int index)
{
  VRmDWORD       device_list_size;
	VRmDeviceKey  *p_device_key;
  VRmImageFormat sourceFormat;

  

  // check for connected devices
  if (!VRmUsbCamGetDeviceKeyListSize(&device_list_size)) {
    // error
    return -1;
  }
  if (index > (int)device_list_size) {
    // error
    return -1;
  }

	// open device at index
  if (!VRmUsbCamGetDeviceKeyListEntry (index, &p_device_key)) {
    // error
    return -1;
  }
  if (!p_device_key->m_busy) {
    if (!VRmUsbCamOpenDevice (p_device_key, &device)) {
      // error
      return -1;
    }

    printf ("found %s: %s [serial %d]\n", p_device_key->mp_manufacturer_str, p_device_key->mp_product_str, p_device_key->m_serial);
    isOpen = true;

    // configure
    if (!VRmUsbCamSetSourceFormatIndex (device, 1)) {
      // error
      return -1;
    }
    VRmUsbCamGetSourceFormat (device, &sourceFormat);
    width  = sourceFormat.m_width;
    height = sourceFormat.m_height;

    // feuer frei
    if (!VRmUsbCamStart (device)) {
      // error
      return -1;
    }
  }



  return 0;
}




////////////////////////////////////////////////////////////////////////

int
tlDigitizerVRm::close ()
{
  // silently ignore unnecessary close calls
  if (!isOpen) {
    return 0;
  }

  // notify
  isOpen = false;
  if (!VRmUsbCamCloseDevice (device)) {
    return -1;
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerVRm::grab (tlImage *image)
{
  int              i,j;
  register BYTE    *buf;
  register tlPixel *dest;
  register int     C,D,E;
  int              upPitch, dnPitch;
  VRmImage         *srcimg;
  VRmBOOL          drp;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerVRm::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // reset image to RGB
  image->init (RGB, width, height);

  // get the captured image buffer
  if (!VRmUsbCamLockNextImage   (device, &srcimg, &drp)) { 
    TL_ERRNO_MSG ("tlDigitizerVRm::grab()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }


  // shortcuts
  dest    = image->pixel;
  buf     = srcimg->mp_buffer;
  upPitch = (height  )*width;
  dnPitch = (height+2)*width;

  // convert
  for (j=0; j<height/2; j++) {
    for (i=0; i<width; i+=2) {
      C = *(buf  ) -  16;
      D = *(buf+1) - 128;
      E = *(buf+3) - 128;
      *(dest  ) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+1) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+2) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      C = *(buf+2) -  16;
      *(dest+3) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+4) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+5) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      buf  += 4;
      dest += 6;
    }
    buf += upPitch;
    for (i=0; i<width; i+=2) {
      C = *(buf  ) -  16;
      D = *(buf+1) - 128;
      E = *(buf+3) - 128;
      *(dest  ) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+1) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+2) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      C = *(buf+2) -  16;
      *(dest+3) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+4) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+5) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      buf  += 4;
      dest += 6;
    }
    buf -= dnPitch;
  }

  if (!VRmUsbCamUnlockNextImage (device, &srcimg)) { return -1; }



  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerVRm::grab (tlImage *image, 
                      tlRect  *rect)
{
  int              i,j;
  long             size;
  register tlPixel *pixel;
  register int     C,D,E;
  VRmImage         *srcimg;
  VRmBOOL          drp;


  // check
  if (!isOpen) {
    TL_ERRNO_MSG ("tlDigitizerVRm::grab()", TL_ERROR_RESOURCE_NOT_AVAILABLE);
    return -1;
  }

  // reset image to RGB
  image->init (RGB, width, height);

  // shortcuts
  const int x1 = TL_MAX ((rect->x), 0);
  const int x2 = TL_MIN (((rect->x)+(rect->width)), width);
  const int y1 = TL_MAX ((rect->y), 0);
  const int y2 = TL_MIN (((rect->y)+(rect->height)), height);
  size         = width*height*3;
  pixel        = image->pixel;

  // get the captured image buffer
  if (!VRmUsbCamLockNextImage   (device, &srcimg, &drp)) { 
    TL_ERRNO_MSG ("tlDigitizerVRm::grab()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // convert
  for (j=y1; j<y2; j+=2) {
    register tlPixel *dest = pixel  + (j*width+x1)*TL_RGB_WIDTH;
    register BYTE    *buf  = srcimg->mp_buffer + (j/2*width+x1)*2;
    for (i=x1; i<x2; i+=2) {
      C = *(buf  ) -  16;
      D = *(buf+1) - 128;
      E = *(buf+3) - 128;
      *(dest  ) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+1) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+2) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      C = *(buf+2) -  16;
      *(dest+3) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+4) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+5) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      buf  += 4;
      dest += 6;
    }
    dest = pixel  + ((j+1)*width+x1)*TL_RGB_WIDTH;
    buf  = srcimg->mp_buffer + ((j/2+height/2)*width+x1)*2;
    for (i=x1; i<x2; i+=2) {
      C = *(buf  ) -  16;
      D = *(buf+1) - 128;
      E = *(buf+3) - 128;
      *(dest  ) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+1) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+2) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      C = *(buf+2) -  16;
      *(dest+3) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+4) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      *(dest+5) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      buf  += 4;
      dest += 6;
    }
  }
  
  if (!VRmUsbCamUnlockNextImage (device, &srcimg)) { return -1; }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerVRm::grabQueued (tlImage *image)
{
  return grab (image);
}



////////////////////////////////////////////////////////////////////////

int 
tlDigitizerVRm::grabQueued (tlImage *image, 
                            tlRect  *rect)
{
  return grab (image, rect);
}
