/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlSourceVideoFile.h"
#include "tlImage.h"



// DirectX headers
#include <qedit.h>
#include <streams.h>
#include <dshow.h>



// container
class DXFileObjects {
public:
  IGraphBuilder  *pGraph;
  IMediaControl  *pControl;
  IMediaEvent    *pEvent;
  ISampleGrabber *pGrabber;
  IMediaSeeking  *pSeek;
};



///  \file tlSourceVideoFile.h
///  \brief Video file image acquisition.

///  \class tlSourceVideoFile
///  Video file image source implementation for TLIB.



///  \fn tlSourceVideoFile::getLength()
///  Get the number of frames in the video sequence.
///
///  \return 
///  The video length [frames]



///  \fn tlSourceVideoFile::getIndex()
///  Get the current frame index.
///
///  \return 
///  The frame index [frames]



// DirectX pin connection utility

static HRESULT
GetUnconnectedPin(IBaseFilter   *pFilter,
                  PIN_DIRECTION PinDir,
                  IPin          **ppPin)
{
  HRESULT   hr;
  IEnumPins *pEnum  = 0;
  IPin      *pPin   = 0;


  *ppPin  = 0;
  hr = pFilter->EnumPins(&pEnum);
  if (FAILED(hr)) {
    return hr;
  }

  while (pEnum->Next(1, &pPin, NULL) == S_OK) {
    PIN_DIRECTION ThisPinDir;
    pPin->QueryDirection(&ThisPinDir);
    if (ThisPinDir == PinDir) {
      IPin *pTmp = 0;
      hr = pPin->ConnectedTo(&pTmp);
      if (SUCCEEDED(hr)) {
        pTmp->Release();
      }
      else {
        pEnum->Release();
        *ppPin = pPin;
        return S_OK;
      }
    }
    pPin->Release();
  }
  pEnum->Release();


  return E_FAIL;
}



// DirectX pin connection utility

static HRESULT
ConnectFilters(IGraphBuilder *pGraph,
               IPin          *pOut,
               IBaseFilter   *pDest)
{
  HRESULT hr;
  IPin    *pIn = 0;


  if ((pGraph == NULL) || (pOut == NULL) || (pDest == NULL)) {
    return E_POINTER;
  }

  // find an input pin on the downstream filter.
  hr = GetUnconnectedPin (pDest, PINDIR_INPUT, &pIn);
  if (FAILED(hr)) {
    return hr;
  }

  // try to connect them.
  hr = pGraph->Connect(pOut, pIn);
  pIn->Release();


  return hr;
}



// DirectX pin connection utility

static HRESULT 
ConnectFilters (IGraphBuilder *pGraph, 
                IBaseFilter   *pSrc, 
                IBaseFilter   *pDest)
{
  HRESULT hr;
  IPin    *pOut = 0;


  if ((pGraph == NULL) || (pSrc == NULL) || (pDest == NULL)) {
    return E_POINTER;
  }

  // find an output pin on the first filter.
  hr = GetUnconnectedPin(pSrc, PINDIR_OUTPUT, &pOut);
  if (FAILED(hr)) {
    return hr;
  }
  hr = ConnectFilters(pGraph, pOut, pDest);
  pOut->Release();


  return hr;
}



///  Open the given filename upon allocation.
///
///  \param filename    video file name
///
///  \note
///  See \ref errors "error management" for details.

tlSourceVideoFile::tlSourceVideoFile (char *filename) : tlSource ()
{
  if (tl_errno != TL_NO_ERROR) return;
  open (filename);
}



///  Open the given filename.
///
///  \param filename    video file name
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSourceVideoFile::open (char *filename)
{
  HRESULT         hr;
  IBaseFilter     *pGrabberF;
  IBaseFilter     *pSrc;
  IBaseFilter     *pNR;
  VIDEOINFOHEADER *pVih;
  AM_MEDIA_TYPE   mt;
  wchar_t         filestr[512];


  // convert filename
  mbstowcs (filestr, filename, 512);

  // initialize the COM library.
  hr = CoInitialize(NULL);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR);
    return -1;
  }

  // create the filter graph manager
  hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void **)&(grabber->pGraph));
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR);
    return -1;
  }

  // query for interfaces
  hr = grabber->pGraph->QueryInterface(IID_IMediaControl, (void **)&(grabber->pControl));
  hr = grabber->pGraph->QueryInterface(IID_IMediaEvent,   (void **)&(grabber->pEvent));
  hr = grabber->pGraph->QueryInterface(IID_IMediaSeeking, (void **)&(grabber->pSeek));

  // create the SampleGrabber
  hr = CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void**)&pGrabberF);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR);
    return -1;
  }
  hr = grabber->pGraph->AddFilter(pGrabberF, L"Sample Grabber");
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR);
    return -1;
  }
  pGrabberF->QueryInterface(IID_ISampleGrabber, (void**)&(grabber->pGrabber));

  // set output video format
  ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
  mt.majortype = MEDIATYPE_Video;
  mt.subtype   = MEDIASUBTYPE_RGB24;
  hr = grabber->pGrabber->SetMediaType(&mt);

  // create acquisition filter
  hr = grabber->pGraph->AddSourceFilter(filestr, L"Source", &pSrc);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return -1;
  }

  // create NULL renderer
  hr = CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (LPVOID *)&pNR);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  hr = grabber->pGraph->AddFilter(pNR, L"NullRenderer");

  // connect graph
  hr = ConnectFilters(grabber->pGraph, pSrc, pGrabberF);
  hr = ConnectFilters(grabber->pGraph, pGrabberF, pNR);

  // set one-shot mode and buffering.
  hr = grabber->pGrabber->SetOneShot(TRUE);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  hr = grabber->pGrabber->SetBufferSamples(TRUE);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // get video capabilities and seeking format
  DWORD dwCap = AM_SEEKING_CanSeekForwards;
  hr = grabber->pSeek->CheckCapabilities (&dwCap);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }
  const struct _GUID dwSup = TIME_FORMAT_FRAME;
  hr = grabber->pSeek->IsFormatSupported (&dwSup);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
  hr = grabber->pSeek->SetTimeFormat (&dwSup);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
  grabber->pSeek->GetDuration (&length);
  current = 0;
  if (tlVerbose) printf ("movie file %s (%d frames)\n", filename, length);

  // launch first acquisition
  hr = grabber->pControl->Run();
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile()", TL_ERROR_CONFIG_FAILED);
    return -1;
  }

  // wait until first image is ready
  grabber->pEvent->WaitForCompletion(INFINITE, &evCode);

  // init acquisition buffer
  hr = grabber->pGrabber->GetCurrentBuffer(&size, NULL);
  buffer = new char[size];
  if (!buffer) 
    if (FAILED(hr)) {
      TL_ERRNO_MSG ("tlSourceVideoFile::open()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }
    hr = grabber->pGrabber->GetCurrentBuffer(&size, (long *)buffer);
    if (FAILED(hr)) {
      TL_ERRNO_MSG ("tlSourceVideoFile()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }

    // get information on video format
    hr = grabber->pGrabber->GetConnectedMediaType(&mt);
    if ((mt.formattype == FORMAT_VideoInfo)        && 
      (mt.cbFormat   >= sizeof(VIDEOINFOHEADER)) &&
      (mt.pbFormat   != NULL) ) {
        pVih   = (VIDEOINFOHEADER*)mt.pbFormat;
        width  = pVih->bmiHeader.biWidth;
        height = pVih->bmiHeader.biHeight;
      }
    else {
      FreeMediaType(mt);
      close ();
      TL_ERRNO_MSG ("tlSourceVideoFile()", TL_ERROR_CONFIG_FAILED);
      return -1;
    }


    return 0;
}



///  Close the currently open file.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSourceVideoFile::close ()
{
  grabber->pControl->Release();
  grabber->pEvent->Release();
  grabber->pGraph->Release();
  CoUninitialize();


  return 0;
}



///  Goto a given frame in the movie. The next image acquired via the
///  grab() method will be at the given index in the video stream.
///
///  \param frame    image index in the video stream
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSourceVideoFile::setFrame (int frame)
{
  HRESULT hr;


  // seek index
  current = (LONGLONG)(frame-1);
  if (current++ >= length) return TL_ERROR_INDEX_OUT_OF_BOUNDS;
  hr = grabber->pSeek->SetPositions (&current, AM_SEEKING_IncrementalPositioning, &length, NULL);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
    return -1;
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlSourceVideoFile::grab (tlImage *image)
{
  register tlPixel *pixel;
  register char    *buf;

  int              i,j;
  HRESULT          hr;


  // end of stream
  if (current < 0) return TL_ERROR;

  // reset image to RGB
  image->init (RGB, width, height);

  // shortcuts
  pixel  = image->pixel;
  buf    = buffer;

  // seek index
  if (current++ >= length) current = -1;
  hr = grabber->pSeek->SetPositions (&current, AM_SEEKING_AbsolutePositioning, &length, NULL);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // launch grab and wait for buffer to be ready
  grabber->pControl->Run ();
  grabber->pEvent->WaitForCompletion (INFINITE, &evCode);

  // copy the captured image to the buffer
  hr = grabber->pGrabber->GetCurrentBuffer(&size, (long *)buffer);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // convert
  for (j=height-1; j>=0; j--) {
    register tlPixel *dest = pixel + j*width*TL_RGB_WIDTH;
    for (i=0; i<width; i++) {
      *(dest++) = *(buf+2);
      *(dest++) = *(buf+1);
      *(dest++) = *(buf  );
      buf += 3;
    }
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlSourceVideoFile::grab (tlImage *image,
                         tlRect  *rect)
{
  register tlPixel *pixel;
  register char    *buf;

  int              i,j;
  int              x1,x2,y1,y2;
  HRESULT          hr;


  // end of stream
  if (current < 0) return TL_ERROR;

  // reset image to RGB
  image->init (RGB, width, height);

  // shortcuts
  x1     = TL_MAX ((rect->x), 0);
  x2     = TL_MIN (((rect->x)+(rect->width)), width);
  y1     = TL_MAX ((rect->y), 0);
  y2     = TL_MIN (((rect->y)+(rect->height)), height);
  pixel  = image->pixel;
  buf    = buffer;

  // seek index
  if (current++ >= length) current = -1;
  hr = grabber->pSeek->SetPositions (&current, AM_SEEKING_AbsolutePositioning, &length, NULL);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // launch grab and wait for buffer to be ready
  grabber->pControl->Run ();
  grabber->pEvent->WaitForCompletion (INFINITE, &evCode);

  // copy the captured image to the buffer
  hr = grabber->pGrabber->GetCurrentBuffer(&size, (long *)buffer);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // convert
  for (j=y2-1; j>=y1; j--) {
    register tlPixel *dest = pixel + (j*width+x1)*TL_RGB_WIDTH;
    for (i=x1; i<x2; i++) {
      *(dest++) = *(buf+2);
      *(dest++) = *(buf+1);
      *(dest++) = *(buf  );
      buf += 3;
    }
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlSourceVideoFile::grabQueued (tlImage *image)
{
  register tlPixel *pixel;
  register char    *buf;

  int              i,j;
  HRESULT          hr;


  // end of stream
  if (current < 0) return TL_ERROR;

  // reset image to RGB
  image->init (RGB, width, height);

  // shortcuts
  pixel  = image->pixel;
  buf    = buffer;

  // wait for image buffer to be ready
  grabber->pEvent->WaitForCompletion (INFINITE, &evCode);

  // copy the captured image to the buffer
  hr = grabber->pGrabber->GetCurrentBuffer(&size, (long *)buffer);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // convert
  for (j=height-1; j>=0; j--) {
    register tlPixel *dest = pixel + j*width*TL_RGB_WIDTH;
    for (i=0; i<width; i++) {
      *(dest++) = *(buf+2);
      *(dest++) = *(buf+1);
      *(dest++) = *(buf  );
      buf += 3;
    }
  }

  // seek index
  if (current++ >= length) current = -1;
  hr = grabber->pSeek->SetPositions (&current, AM_SEEKING_AbsolutePositioning, &length, NULL );
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // launch grab and return immediately
  grabber->pControl->Run ();


  return 0;
}




////////////////////////////////////////////////////////////////////////

int
tlSourceVideoFile::grabQueued (tlImage *image,
                               tlRect  *rect)
{
  register tlPixel *pixel;
  register char    *buf;

  int              i,j;
  int              x1,x2,y1,y2;
  HRESULT          hr;


  // end of stream
  if (current < 0) return TL_ERROR;

  // reset image to RGB
  image->init (RGB, width, height);

  // shortcuts
  x1     = TL_MAX ((rect->x), 0);
  x2     = TL_MIN (((rect->x)+(rect->width)), width);
  y1     = TL_MAX ((rect->y), 0);
  y2     = TL_MIN (((rect->y)+(rect->height)), height);
  pixel  = image->pixel;
  buf    = buffer;

  // wait for image buffer
  grabber->pEvent->WaitForCompletion (INFINITE, &evCode);

  // copy the captured image to the buffer
  hr = grabber->pGrabber->GetCurrentBuffer(&size, (long *)buffer);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // convert
  for (j=y2-1; j>=y1; j--) {
    register tlPixel *dest = pixel + (j*width+x1)*TL_RGB_WIDTH;
    for (i=x1; i<x2; i++) {
      *(dest++) = *(buf+2);
      *(dest++) = *(buf+1);
      *(dest++) = *(buf  );
      buf += 3;
    }
  }

  // seek index
  if (current++ >= length) current = -1;
  hr = grabber->pSeek->SetPositions (&current, AM_SEEKING_AbsolutePositioning, &length, NULL);
  if (FAILED(hr)) {
    TL_ERRNO_MSG ("tlSourceVideoFile::grab()", TL_ERROR_GRAB_FAILED);
    return -1;
  }

  // launch grab and return immediately
  grabber->pControl->Run ();


  return 0;
}
