/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <string.h>

#include "tlVision.h"

#ifdef WIN32
#include <windows.h>
#ifdef BORLAND
#include <vcl.h>
#endif
#endif



///  \file tlVision.h
///  \brief Main TLIB header file.

///  \def TL_PI
///  Default value for the PI constant (defined here for portability reasons).

///  \def TL_DEFAULT_COLS
///  The default image width used in TLIB.

///  \def TL_DEFAULT_ROWS
///  The default image height used in TLIB.

///  \def TL_MAX_PIXEL_VAL
///  The highest value a pixel of type tlPixel can have.

///  \def TL_MIN_PIXEL_VAL
///  The smallest value a pixel of type tlPixel can have.

///  \def TL_NOT_CONNEX
///  Indicates that no connectivity is to be used.

///  \def TL_CONNEX_4
///  Indicates that 4-connex pixels are considered.

///  \def TL_CONNEX_8
///  Indicates that 8-connex pixels are considered.

///  \def TL_PIXEL(row, col, width)
///  Compute the address of a pixel in a given array of dimension 1.

///  \def TL_MIN(a, b)
///  Returns the smallest of two values.

///  \def TL_MAX(a, b)
///  Returns the greatest of two values.
 
///  \def TL_NOT_IMPLEMENTED()
///  Lazy seb emergency macro.

///  \def TL_PAUSE
///  Waits for the user to hit the ENTER key.

///  \def TL_NOT_IMPLEMENTED
///  \internal
///  Macro used to define routines that have not been implemented.

///  \def TL_RGB_WIDTH
///  Size (in bytes) of one pixel from a RGB image.

///  \def TL_NRG_WIDTH
///  Size (in bytes) of one pixel from a NRG image.

///  \def TL_HSI_WIDTH
///  Size (in bytes) of one pixel from a HSI image.
 
///  \def TL_YUV_WIDTH
///  Size (in bytes) of one pixel from a YUV image.

///  \def TL_GRAY_WIDTH
///  Size (in bytes) of one pixel from a GRAY image.

///  \def TL_BIN_WIDTH
///  Size (in bytes) of one pixel from a BIN image.

///  \def TL_STEREO_WIDTH
///  Size (in bytes) of one pixel from a STEREO image.

///  \def TL_MAX_PIXEL_WIDTH
///  Maximum size (in bytes) a single pixel can have in an image.

///  \def TL_CHANNEL_RED
///  Shortcut to the R channel of a RGB image.

///  \def TL_CHANNEL_GREEN
///  Shortcut to the G channel of a RGB image.

///  \def TL_CHANNEL_BLUE
///  Shortcut to the B channel of a RGB image.

///  \def TL_CHANNEL_H
///  Shortcut to the first H channel of a HSI image.

///  \def TL_CHANNEL_S
///  Shortcut to the S channel of a HSI image.

///  \def TL_CHANNEL_I
///  Shortcut to the first i of a HSI image.

///  \def TL_CHANNEL_NR
///  Shortcut to the normalized R channel of a NRG image.

///  \def TL_CHANNEL_NG
///  Shortcut to the normalized G channel of a NRG image.

///  \def TL_CHANNEL_GRAY
///  Shortcut to the intensity channel of a GRAY image.

///  \def TL_CHANNEL_BIN
///  Shortcut to the binary channel of a BIN image.

///  \def TL_CHANNEL_STEREO
///  Shortcut to the disparity channel of a STEREO image.

///  \typedef tlPixel
///  pixel element (8 bits,unsigned).

///  \typedef tlHistData
///  histogram data type (int,unsigned).

///  \var int tlVerbose
///  Verbosity flag, can be set via tl_set_verbose(int verbosity).

///  \def TL_HMHL_TO_HVAL(hm, hl)
///  Convert hue MSB and LSB to 16-bit HVAL.\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_HVAL_TO_HM(hval)
///  convert 16-bit HVAL to hue MSB.\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_HVAL_TO_HL(hval)
///  convert 16-bit HVAL to hue LSB.\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_DEG_TO_HVAL(deg)
///  Convert hue: [0,360] deg to scaled 16-bit.\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_HVAL_TO_DEG(hval)
///  Convert hue: scaled 16-bit to [0,360] deg.\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_PCT_TO_SVAL(pct)
///  Convert saturation: [0,100] percent to [0,TL_MAX_PIXEL_VAL].\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_SVAL_TO_PCT(sval)
///  Convert saturation: [0,TL_MAX_PIXEL_VAL] to [0,100] percent.\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_PCT_TO_IVAL(pct)
///  Convert intensity: [0,100] percent to [0,TL_MAX_PIXEL_VAL].\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \def TL_IVAL_TO_PCT(ival)
///  Convert intensity: [0,100] percent to [0,TL_MAX_PIXEL_VAL].\n
///  See \ref imageformats "image formats" for details on the HSI format.

///  \enum tl_format
///  See \ref imageformats "image formats" for more details on image formats implementation.

///  \var tl_format RGB
///  common RGB format [\ref TL_RGB_WIDTH bytes wide]\n

///  \var tl_format NRG
///  normalized RGB format (in the normalized red NR and normalized green NG plane) [\ref TL_NRG_WIDTH bytes wide]\n

///  \var tl_format YUV
///  YUV format (PAL color standard) [\ref TL_YUV_WIDTH bytes wide]\n
///  \note
///  The YUV format encodes RGB data in a way that requires less bandwidth for analogous transmission.
///  Y (luma) encapsulate the B/W information. In TLIB, intensity data (or GRAY conversion) from YUV
///  will always be luma. It is therefore NOT the same "intensity" as computed from RGB format.

///  \var tl_format HSI
///  HSI format [\ref TL_HSI_WIDTH bytes wide]\n
///  Hue is encoded on 2 bytes (H1 and H2) to preserve resolution. The following
///  macros can be used to manipulate the HSI format:\n
///  \ref TL_HMHL_TO_HVAL(hm, hl)\n
///  \ref TL_HVAL_TO_HM(hval)\n
///  \ref TL_HVAL_TO_HL(hval)\n
///  \ref TL_DEG_TO_HVAL(deg)\n
///  \ref TL_HVAL_TO_DEG(hval)\n
///  \ref TL_PCT_TO_SVAL(pct)\n
///  \ref TL_SVAL_TO_PCT(sval)\n
///  \ref TL_PCT_TO_IVAL(pct)\n
///  \ref TL_IVAL_TO_PCT(ival)\n

///  \var tl_format GRAY
///  common GRAY format [\ref TL_GRAY_WIDTH bytes wide]\n

///  \var tl_format BIN
///  binary image format [\ref TL_BIN_WIDTH bytes wide]\n
///  Any non-null pixel value is enabled/active, while null pixels are disabled/inactive.\n 

///  \var tl_format STEREO
///  stereo (disparity map) format [\ref TL_BIN_WIDTH bytes wide]\n
///  Negative pixel value indicate pixels whose disparity could not be computed.
///  The tlDisplay "hides" pixels with negative value in the case of stereo images for easthetic purposes.\n

///  \var tl_format CUSTOM
///  do-it-yourself format [width unspecified].\n
///  Programmers can use the routines in tlChannel.h to build their own image format as a 
///  combination of image channels from various sources/formats.\n

/// tl_verbose
/// HSI-RGB conversion macro

/// \def TL_PAUSE
/// Wait for the ENTER key...



// globals

char  *tl_app_name;
long  tl_timer_ref    = 0;
int   tlVerbose       = 0;



///  Set the level of verbosity displayed in the console.
///
///  \param val 0 for no output\n 1 for verbosity

void
tl_set_verbose (int val)
{
  tlVerbose = val;
}

static void
_tl_init (void)
{
  if (tlVerbose) {
    printf ("\nTLIB 2.0\n(C) 2002-2004 VRAI group, EPFL\n\n");
    printf ("initializing TLIB...\n");
  }

  // init timer
  tl_timer_init ();
}



///  \overload

int
tl_init_lib (void)
{
  _tl_init();

  // save application name
  tl_app_name = (char *) malloc ((strlen ("TL application") + 1) * sizeof (char));
  strcpy (tl_app_name, "TL application");


  return TL_NO_ERROR;
}



///  Initialize library components.\n
///  This consists in setting the application name, as well as
///  initializing the high-resolution timers used in TLIB.
///
///  \param argc    number of arguments
///  \param argv    arguments array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The arguments (argv,argc) should be the ones fed to
///  the main() call in a C ANSI application (console mode).

int
tl_init_lib (int  argc, 
             char **argv)
{
  _tl_init();

  // be tolerant with arg parsing for now...
  if (argc < 1) {
    tl_app_name = (char *) malloc ((strlen ("TL application") + 1) * sizeof (char));
    strcpy (tl_app_name, "TL application");
  }

  // save application name
  else {
    tl_app_name = (char *) malloc ((strlen (argv[0]) + 1) * sizeof (char));
    strcpy (tl_app_name, argv[0]);
  }


  return TL_NO_ERROR;
}


/**

  \mainpage documentation
  
  \image html tlib.png

    \section intro Introduction

This document is a basic overview of TLIB that exposes the underlying
concepts behind the design and organization of the software. It is not meant to be exhaustive but instead 
is just a general outline, explaining the key features of the library. A more thorough
description of TLIB classes and functionalities can be found in the reference pages of this on-line manual.

Please note that the /examples directory provides useful examples that are a 
good starting point to understand how to write an application using TLIB.

TLIB can be downloaded in multiple flavors and wrapping:
\htmlonly <li><a href="http://www.tuyphon.com/tlib/download/tlib-2.1-src.zip">tlib-2.1-src.zip</a> full source archive\endhtmlonly
\htmlonly <li><a href="http://www.tuyphon.com/tlib/download/tlib-2.1-msvc-2k8.zip">tlib-2.1-msvc-2k8.zip</a> pre-compiled for Microsoft Visual C++ 2008 (Express Edition)\endhtmlonly
\htmlonly <li><a href="http://www.tuyphon.com/tlib/download/tlib-2.1-msvc-2k5.zip">tlib-2.1-msvc-2k5.zip</a> pre-compiled for Microsoft Visual C++ 2005 (Express Edition)\endhtmlonly
\htmlonly <li><a href="http://www.tuyphon.com/tlib/download/tlib-2.1-msvc-2k3.zip">tlib-2.1-msvc-2k3.zip</a> pre-compiled for Microsoft Visual Studio 2003\endhtmlonly
\htmlonly <li><a href="http://www.tuyphon.com/tlib/download/tlib-2.1.tar.gz">tlib-2.1.tar.gz</a> pre-compiled for Linux (kernel 2.6)\endhtmlonly
\htmlonly <li><a href="http://www.tuyphon.com/tlib/download/tlib-2.1-msvc-6.0.zip">tlib-2.1-msvc-6.0.zip</a> pre-compiled for Microsoft Visual Studio 6.0\endhtmlonly
\htmlonly <li><a href="http://www.tuyphon.com/tlib/download/tlib-2.1-bbcp.zip">tlib-2.1-bbcp.zip</a> pre-compiled for Borland C++ Builder 6.0\endhtmlonly

The current version is \b tlib-2.1. Take a look at the \htmlonly <a href="http://www.tuyphon.com/tlib/download/releasenotes.txt">release notes</a> \endhtmlonly to find out about features. Feel free to contact seb@tuyphon.com for contributions and/or requests.

  \section installation Installation

After downloading the distribution from the \htmlonly <a href="http://www.tuyphon.com/tlib/download" target=_blank>download</a> \endhtmlonly page, 
uncompress it to the location of your choice, et voil ! Under Windows, you may be prompted for missing third-party DLLs, depending on what is already
installed on your system (see third-party run-time libraries requirements). The missing DLLs files must be placed in the same folder as your TLIB application, or in the system32 folder of your 
Windows installation (recommended).

  \b Third-party \b run-time \b libraries \b requirements
  \li (optionally) SVS 4.1 stereo processing library from VidereDesign \htmlonly <a href="http://www.videredesign.com" target=_blank>(website)</a> \endhtmlonly
  \li naturally, any DLL/modules related to your particular hardware

  \b Special \b instructions \b for \b Visual \b Studio \b Express \b 2005
  \li http://www.codeproject.com/wtl/WTLExpress.asp?df=100&forumid=255873&exp=0&select=1526308

  \section purpose Purpose

TLIB is an image processing library written specifically for real-time object tracking.
It is meant for people with basic C/C++ programming and image processing skills.
TLIB aims at fulfilling several goals:

  \li \b Efficiency \n
While real-time computer vision-based tracking usually requires highly optimized
code, writing optimized code is not a trivial task and can be quite time consuming. 
TLIB's intention is to provide a wrap-up to optimized low-level functions so
that the programmer does not have to worry about it.

  \li \b Ease-of-use \n
Other libraries exist that offer a set of optimized image processing functions, 
in particular the Intel OpenCV library. However, and even though they are very
powerful tools, these libraries make it hard to rapidly develop a simple code 
for people with basic programming skills. TLIB is designed to offer a high-level
interface that makes abstraction of the low-level optimized code, making it
very well suited for rapid image processing application development and student
training.

    \li \b Portability \n
TLIB is written in ANSI C/C++, mostly independant of OS specific libraries. Only
the image acquisition and display classes are OS dependant. Thus, TLIB can be ported
to any OS with little effort. Current distribution supports MS Windows and Linux.

  \li \b Educational \n
TLIB offers high-level and overloaded routines, along with simple and straightforward
data structures that make the code look compact and clean. It simplifies the dataflow
design and the understanding of the software for the programmer, and is a good 
introduction to more complex image processing software.

    \li <b> Use of multiple sources </b> \n
TLIB is designed to work with various image acquisition hardware and image formats,
in particular with color and stereoscopic images. The TLIB image formats support 
several standard and provide a wrapper to a stereo algorithm. It is also possible
to design your own image format from multiple sources, along with algorithms dedicated
to your custom format.

Thus, TLIB has been designed to make it easy for most programmers to include
efficient computer vision code into their applications rapidly, in particular 
for applications that require vision-based detection and tracking.

  \section features Features

This section covers some of the features and formats that TLIB supports.

    \subsection imageformats Image Formats

TLIB uses a simple pixel array type that can be used for virtually any image format. Pixels are
held in a unidimensional array, namely the \c pixel member or the tlPixelArray class. Depending on
the image format, an image will have a number of channels. Pixels from various channels are transparently
recorded in the \c pixel array. For instance, in the case of the RGB file format, \c pixel will contain an
array of pixels in the following sequence: RGBRGBRGBRGBRGB...\n\n
The \c pixel array starts at the upper left corner of the image and save pixels line after line. 
The \ref tl_format "tl_format" section describes the different image formats and color spaces that TLIB can handle.


  \section architecture Architecture

The TLIB library is a set of optimized ANSI C functions that are wrapped up
in a set of higher-level C++ objects. The benefits of this dual-level architecture
are a significant reduction of the amount and of the redundancy of code written 
by the programmer.

    \subsection objects Objects

This section describes the different high-level objects that TLIB is made of.

  \li \b tlPoint \b objects \n
Simple point coordinates are stored in tlPoint objects. tlPoint contains both 2D image
coordonates, a disparity value, and 3D world coordonates. The tlPoint object also provides
convenience routines to update and copy point coordinates. Details about the tlPoint 
object structure can be found in tlPoint.

  \li \b tlChain \b objects \n
A tlChain object is an arbitrary collection of tlPoints. tlChain objects are useful for
storing extracted points, but also for continuous edges extraction and other related 
operations such as angle detections. Details about the tlChain object structure can be 
found in tlChain.

  \li \b tlRect \b objects \n
One of the most commonly used objects in TLIB is the tlRect object. The tlRect object
contains simple rectangular data (position, size) and routines to duplicate,
move and resize it. It is used mostly to define a ROI (Region Of Interest, i.e. to
limit the image processing to a given portion of an image), leading to shorter execution times. 
Details about the tlRect object structure can be found in tlRect.

  \li \b tlColor \b objects \n
Color values of different color formats (described \ref imageformats "here") are stored in tlColor
objects. They are used as arguments to color filtering and color extraction routines. 
Details about the tlColor object structure can be found in tlColor.

  \li \b tlConst \b objects \n
Tracking routines often relies on some "magic" constants that vary over time according
to external conditions, or that simply cannot be determined anatically. The tlConst
object is a Kalman-filter based estimator of constants that can be dynamically updated.
Details about the tlConst object structure can be found in tlConst.

  \li \b tlPixelArray \b objects \n
The base class for every object that contains a pixel array. tlPixelArray contains a pixel array
with its dimension, format and properties. It also encapsulates all the pixel operator methods 
that can be applied to it. tlImage, toObject and tlMask inherit from tlPixelArray.
Details about the tlPixelArray object structure can be found in tlPixelArray.

  \li \b tlImage \b objects \n
Perhaps the most commonly used object in TLIB is the tlImage object. It contains the image
data (from its tlPixelArray parent), as well as useful routines such as drawing functions,
file read/write, etc. Details about the tlImage object structure can be found in tlImage.

  \li \b tlObject \b objects \n
Almost equally important in TLIB is the tlObject object. A tlObject is a set of 
data (pixel array, dimensions, velocity, ...) that describes an object in an image,
in a given image format. It contains the method necessary to retrieve the object position
in subsequent images, as well as the routines that update that object properties from an image.
Details about the tlObject object structure can be found in tlObject.

  \li \b tlMask \b objects \n
In a similar way, the tlMask object is a very simple structure that is used extensively
in TLIB routines. It is a basic binary mask with position and size information that can
be used to apply image processing routines to specific pixels only. A tlMask object is
also used to define which pixels are part of an object in a given area. Mask objects can
be combined, and support binary image processing routines such as morphological operators.
Details about the tlMask object structure can be found in tlMask.

  \li \b tlBlobs \b objects \n
A basic task in image processing, known as "connected compounds" extraction, consists in
managing and sorting blobs defined by a series of filters. The tlBlobs object is a
collection of tlMask objects that does just that. Given a binary image, it segments,
extracts and labels a set of tlMask, that can then be sorted with respect to different
criterions. This is extremely useful for object extraction. Details about
the tlBlobs object structure can be found in tlBlobs.

  \li \b tlHist \b objects \n
tlHist provides a histogram class compatible with TLIB images and objects.
It offers simple functions such as ranking, normalization and scaling,
building from images, comparison and data manipulation. 
Details about the tlHist object structure can be found in tlHist.

  \li \b tlHist2D \b objects \n
tlHist2D provides a 2D histogram class compatible with TLIB images and objects.
It offers simple functions such as ranking, normalization and scaling,
building from images, comparison and data manipulation. 
Details about the tlHist2D object structure can be found in tlHist2D.

  \li \b tlSpiral \b objects \n
A tlSpiral object provides a discrete spiral coordinates computation respectively
to a given starting point. This allows a routine to address image pixels in a spiral
sequence, which can be extremely useful for some correlation implementations or 
simply to find the closest filtered neighbor of a filtered pixel. Details about the tlSpiral 
object structure can be found in tlSpiral.

  \li \b tlHough \b objects \n
The tlHough class offers a quick an efficient way to extract geometric features from tlImage objects.
Currently, only straight lines are supported. Details about the tlHough object structure can
be found in tlHough.

The following objects are wrappers for hardware/platform dependant components such
as image digitizers and displays. They are provided as a convenience for commonly 
used hardware and operating systems, and can be used as templates for the programmer 
to write his/her own hardware drivers. The list below gives the currently
available implementations available in TLIB.
 
  \li \b tlDigitizer \b objects \n
Simple image acquisition should be performed through tlDigitizer objects.
Details about the tlDigitizer object structure can be found in tlDigitgizer. Under Windows, TLIB
currently supports the DirectShow generic interface grabber that can be used with most firewire and USB webcams. 
Dedicated implementations of tlDigitizer are also available for the ImageNation PXC200 and the Data Translation DT3130 framegrabbers.
Under Linux, TLIB supports Video4Linux2 compatible devices.

  \li \b tlDisplay \b objects \n
A tlDisplay object offers a simple image rendering window, along with platform-independant
mouse and keyboard events processing with respect to the windows. Details about the
tlDisplay object structure can be found in tlDisplay. Under Windows, TLIB currently uses DirectX
to display images. Under Linux, TLIB supports an XWindows implementation of tlDisplay.

  \li \b tlCamera \b objects \n
The tlCamera base class offers an encapsulation for image sources that are in a known location 
in the workspace, and possibly return calibrated (undistorted) images. An important child class
is the tlCameraStereo base class, which provides a generic interface to stereo engines.
Details about the tlCamera object structure can be found in tlCamera.

\li \b tlCameraStereo \b objects \n
The tlCameraStereo base class offers an encapsulation for stereo image sources and stereo
processing engines. Stereo cameras are defined here as "special" images that can 
acquire not only images but also disparity maps, registered to the source image. 
Details about the tlCameraStereo object structure can be found in tlCameraStereo.
 
   \subsection clike C functions

C functions are the core of TLIB. They implement optimized operators on specific data
types. Programmers should try to use objects structure, mostly because the objects
method performs all the necessary data consistency checks and conversion before calling
the low-level C function. However, experienced programmers that are familiar with TLIB
data types and architecture may find it useful to work with C function to enhance
performance or implement new object operators.

All low-level C functions have a \b tl_ prefix. They are grouped by categories, as 
described below.

  \li \b Timer \b functions \n
TLIB offers platform-independant timing routines with a worst resolution of approximately 1 ms.
On Windows machine, TLIB can usually use the resolution of the processor hardware clock.
Details about the tlTimer object structure can be found in tlTimer.h.

  \li \b Channel \b operators \n
The list of channel operators can be found in tlChannel.h.

  \li \b Convolution \b operators \n
The list of convolution operators can be found in tlConvolution.h.

  \li \b Edges \b operators \n
The list of edges operators can be found in tlEdges.h.

  \li \b Object \b extraction \b operators \n
The list of object extraction operators can be found in tlExtract.h.

  \li \b File \b operators \n
The list of file operators can be found in tlFile.h.

  \li \b Filter \b operators \n
The list of filter operators can be found in tlFilter.h.

  \li \b Find \b operators \n
The list of find operators can be found in tlFind.h.

  \li \b Format \b conversion \b operators \n
The list of format conversion operators can be found in tlFormatConversion.h.

  \li \b Memory \b operators \n
The list of memory operators can be found in tlMemory.h.

  \li \b Morphological \b operators \n
The list of morphological operators can be found in tlMorph.h.

  \li \b Pixel \b operators \n
The list of pixel operators can be found in tlPixel.h.
 
    \subsection datatype Data type

Here are the most important data types used in TLIB.

  \li \b tlPixel \b type \n
The tlPixel type is the typewith which all pixel arrays are built, whatever the
image format is. This uniformity is motivated by optimization requirements.
The programmer is encouraged to use the tlPixel type explicitely to guarantee 
compatibility with future releases, should the tlPixel typedef change in the future. 
Details on the tlPixel format can be found in tlVision.h.

    \subsection optimization Code optimization

A word is required here on code optimization. There are two kinds of code 
optimization techniques, described below.

  \li \b Hardware-independant \b optimization \n
Hardware-independant optimization include all optimization that do not take into
account specific hardware properties such as the physical memory cache management
algorithm and the memory alignment. It consists mostly of coding guidelines that
help the compiler optimization routines to do a better job. 

  \li \b Hardware-dependant \b optimization \n
Hardware-dependant optimization usually produces better results in terms of code
execution. However, the code becomes highly machine specific and might run much
slower (or not run at all) on another computer or another operating system.
  
For portability reasons, TLIB relies only on hardware-independant code optimization.

Below is a list of simple guidelines for programmers who use TLIB and who wish to make their code more 
efficient. Please note that, generally, optimization leads to less readable code that 
is more difficult to debug (quite often, optimizing code introduces bugs). The recommended 
approach is to  write "standard" code to test algorithms before applying optimization rules.

Also, please note that these techniques are here to help the compiler do a better job only.
Results will vary depending on the compiler, the optimization flags and the hardware.

  \li \b Loop \b unrolling \n
The easiest technique to optimize code consists in explicitely coding loops instead of using
\e for or \e while statements. \n \n
\e Example \n \n
instead of
\code
int i;
for (i=0; i<4; i++) {
  printf ("hello world\n");
}\endcode \n
prefer
\code
printf ("hello world\n");
printf ("hello world\n");
printf ("hello world\n");
printf ("hello world\n");\endcode

  \li \b Function \b overhead \n
Because of the way a compiler works, every time a function is called, some time
is lost. This time can be signiticant if the function is called numerous time (in a loop,
for instance). Also, the overhead of a C++ method is higher than a C function. \n \n
\e Example \n \n
instead of
\code
int i;
int val;

for (i=0; i<100; i++) {
  for (j=0; j<100; j++) {
    val = image->getPixel(i,j);
  }
}\endcode \n
prefer
\code
int i;
int val;
int width = image->width;

for (j=0; j<100; j++) {
  for (i=0; i<100; i++) {
    val = image->pixel[j*width+i];
  }
}\endcode

  \li \b Invariants \n
When the compiler comes accross an indirect reference, such as an array element or an
object field, it will explicitely compute the memory address of the reference even if
it has not changed. This can have an impact on the termination condition of loops. \n \n
\e Example \n \n
instead of
\code
int i,j;
int val;
int index = 0;

// in this case, the address of 
// (image->height), (image->width) and (image->pixel)
// are calculated at every iteration
for (j=0; j<(image->height); j++) {
  for (i=0; i<(image->width); i++) {
    val = image->pixel[index++];
  }
}\endcode \n
prefer
\code
int i,j;
int val;
int index = 0;

// in this case,
// (image->height), (image->width) and (image->pixel)
// are only calculated once
int height   = image->height;
int width    = image->width;
tlPixel *src = image->pixel;

for (j=0; j<height; j++) {
  for (i=0; i<width; i++) {
    val = src[index++];
  }
}\endcode

  \li \b Function \b arguments \n
Everytime a program calls a function, the function arguments are copied in the system
memory. Obviously, copying a structure containing 150 elements will take more resources
than a single byte. Therefore, when working with structures, programmer should try and 
use pointers.

  \li \b Function \b inlining \n
For the same reason, namely to avoid the overhead of function calls, it is possible to tell
the compiler to execute a function "in place" by using the keyword \e inlin when declaring
the function. Inlining can be tricky and sometimes impair performances. It is best used
with small functions that get called repeatedly from one single place in the code (singletons).

  \li \b Keywords \n
Some compilers (including Visual C++ compiler) will use the \e register keyword
in front of a variable declaration to try and force the variable in a processor
register. The effect will depend on the hardware (in particular on the size of
the register on the machine processor). It has close to no effects on Pentium 
class processors. \n \n
\e Example \n \n
\code
int j;
register int height = image->height;
register int width  = image->width;

for (j=0; j<height; j++) {
  do_something(j, width);
}\endcode


    \subsection support Support and contributions

TLIB is, by definition, an ever-evolving library. As such, it is also likely to
contain bugs and to display some design flaws. Questions and comments can be
addressed to seb@tuyphon.com (please mention TLIB in the subject line).

  \li \b How \b to \b contribute \n
Anybody who uses TLIB is welcome to contribute, either by providing code that
is compatible with the TLIB architecture and is relevant to real-time tracking
applications, or by sharing cool demo source code that makes use of TLIB.

  
*/



/**
 
  \page external External packages documentation
 
  \image html tuyphon-cvl.png " "
 
  TLIB makes use of several free and commercial software libraries.
  This section briefly explains what they are, what they do, and 
  where to find more information on them.
 
  \section svsDoc Small Vision System
 
  \subsection svsOverview Overview
  TLIB is compatible with the SVS (Small Vision System) stereo library, developed by Kurt Konolige from SRI international.
  SVS provides both internal and external calibration of stereo pairs, as well as real-time disparity map computation.
  Moreover, SVS is compatible with all digital stereo heads distributed by VidereDesign. TLIB provides dedicated
  classes based on tlCameraStereo to fully support the SVS features. In order to use the SVS features, users
  must have a valid license, available from VidereDesign.
 
  \subsection svsLinks Useful links
  \li \htmlonly <a href="http://www.videredesign.com" target=_blank>VidereDesign</a> \endhtmlonly
  \li \htmlonly <a href="http://www.ai.sri.com/~konolige" target=_blank>Kurt Konolige SRI page</a> \endhtmlonly
 
  \section tsai TSAI Camera Calibration
 
  \subsection tsaiOverview Overview
  TLIB includes an implementation of Roger Tsai's camera calibration in the tlCameraTSAI class, based on the
  Tsai calibration library currently distributed freely by CMU.
 
  \subsection tsaiLinks Useful links
  \li \htmlonly <a href="http://www-2.cs.cmu.edu/afs/cs.cmu.edu/user/rgw/www/TsaiDesc.html" target=_blank>CMU TSAI Calibration page</a> \endhtmlonly
  \li R. Tsai, "An Efficient and Accurate Camera Calibration Technique for 3D Machine Vision." IBM, T.J. Watson Research Center Yorktown Heights, NY 10598
 
*/
