/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2005 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlVision.h"
#include "tlBackground.h"
#include "tlRect.h"
#include "tlImage.h"
#include "tlCameraStereo.h"



/// \file tlBackground.h
/// \brief Background definition and learning.

///  \class tlBackground
///  \brief Background definition and learning class.
///
///  The tlBackground class makes it possible to define a background image
///  of any \ref tl_format "format", and optionally have it learn the statistical 
///  (gaussian) distribution of each pixel by providing it with additionnal images 
///  over time.
///
///  \section tl_background_parameters Background definition parameters
///  \b Vmin is the minimal allowed variance for each pixel.\n
///  \b Vmax is the maximal allowed variance for each pixel.\n
///  \b alpha is the gaussian mean averaging factor: mean = alpha*mean+(1-alpha)*new_value\n
///  \b K is the variance scaling factor: the background boundary is { mean-K*variance ; mean+K*variance }\n
///  \b Koffset is the foreground filter offset for STEREO format only : the foreground is defined as { [0 ; Koffset+mean-K*variance] && [ Koffset+mean+K*variance ; TL_MAX_VALUE] }.\n



///  Constructor.
///
///  \note
///  See \ref errors "error management" for details.

tlBackground::tlBackground ()
{
	mean      = NULL;
	variance  = NULL;
   
  learn     = 1;
  adapt     = 1;

  Vmin	    = 5.0f;
  Vmax      = 20.0f;
  alpha     = 0.99f;
  K         = 1.5f;
  Koffset   = 5.0f;
}



///  Constructor.
///  See \ref tl_background_parameters "background learning parameters" for details.
///
///  \note
///  See \ref errors "error management" for details.

tlBackground::tlBackground (float Vmin,
                            float Vmax,
                            float alpha, 
                            float K, 
                            float Koffset)
{
	mean      = NULL;
	variance  = NULL;
   
  learn     = 1;
  adapt     = 1;

  this->Vmin	    = Vmin;
  this->Vmax	    = Vmax;
  this->alpha     = alpha;
  this->K         = K;
  this->Koffset   = Koffset;
}



///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlBackground::~tlBackground ()
{
	if (mean)			free (mean);
	if (variance) free (variance);
}



///  Init or reset background computing
///
///  \param image			initial image
///  \param learn			0 for static background definition, 1 for dynamic background learning
///  \param adapt     0 if the background is static, 1 for adaptive background
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int
tlBackground::init (tlImage *image,
	  								int     learn,
                    int     adapt)
{
	int			   i,j;
	tlPixel    *pixel;
	float      *m;
	float      *v;
  int        x2,y2;


	// check arguments
  if (!image) {
    TL_ERRNO_MSG ("tlBackground::init()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // store image properties
  format     = image->format;
  width      = image->width;
  height     = image->height;
	pixel      = image->pixel;
  pixelWidth = image->pixelWidth;

	// save method
	this->learn = learn;
  this->adapt = adapt;

	// allocate background data
	mean			= m = (float *) realloc (mean, width*height*pixelWidth*sizeof(float));
	variance	= v = (float *) realloc (variance, width*height*pixelWidth*sizeof(float));
	if (!mean || !variance) {
    TL_ERRNO_MSG ("tlBackground::init()", TL_ERROR_ALLOCATION_FAILED);
		return -1;
	}

	// init vals
  x2 = pixelWidth*width;
  y2 = height;
	for (j=0; j<y2; j++) {
		for (i=0; i<x2; i++) {
			*(m++) = *(pixel++);
			*(v++) = Vmax;
		}
	}


	return 0;
}



///  Set the image learning parameters. 
///  See \ref tl_background_parameters "background learning parameters" for details.
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int 
tlBackground::setParam (float Vmin,
                        float Vmax, 
                        float alpha, 
                        float K, 
                        float Koffset)
{
  this->Vmin    = Vmin;
  this->Vmax    = Vmax;
  this->alpha   = alpha;
  this->K       = K;
  this->Koffset = Koffset;


  return 0;
}



///  Learn from a new image (next learning iteration).
///
///  \param image			new disparity image
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int
tlBackground::update (tlImage *image)
{
	int        i,j;
  int        x2, y2;
	float      *m;
	float      *v;


	// only if learning
	if (learn == 0) return 0;

  // check for format consistency
  if (image->format != format) {
    TL_ERRNO_MSG ("tlBackground::update()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

	// shortcuts
	m     = mean;
	v     = variance;

  // based on format
  switch (format) {
  
  case STEREO:
    {
      register tlHRPixel *pixel = image->hrpixel;

      // update values
      for (j=0; j<height; j++) {
        for (i=0; i<width; i++) {
          // update as necessary
          if (*(m) >= 0) {
            if (*(pixel) < (*(m) + K*(*(v)))) {
              *(v) = alpha*(*(v)) + (1.0f-alpha)*((float)*(pixel)-*(m))*((float)*(pixel)-*(m));
              *(m) = alpha*(*(m)) + (1.0f-alpha)*((float)*(pixel));
              if (*(v) < Vmin) *(v) = Vmin;
              if (*(v) > Vmax) *(v) = Vmax;
            }
          }
          else {
            if (*(pixel) >= 0) {
              *(m) = *(pixel);
              *(v) = Vmin;
            }
          }
          pixel++;
          m++;
          v++;
        }
      }
    }
    break;

  default:
    {
      tlPixel *pixel = image->pixel;

      // update values
      x2 = width;
      y2 = height;
      for (j=0; j<y2; j++) {
        for (i=0; i<x2; i++) {
          switch (pixelWidth) {
          case 1:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
            }
            pixel += 1;
            m     += 1;
            v     += 1;          
            break;
          case 2:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                          (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(v+1) = alpha*(*(v+1)) + (1.0f-alpha)*((float)*(pixel+1)-*(m+1))*((float)*(pixel+1)-*(m+1));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              *(m+1) = alpha*(*(m+1)) + (1.0f-alpha)*((float)*(pixel+1));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v+1) < Vmin) *(v+1) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
              if (*(v+1) > Vmax) *(v+1) = Vmax;
            }
            pixel += 2;
            m     += 2;
            v     += 2;          
            break;
          case 3:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                          (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                          (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2)))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(v+1) = alpha*(*(v+1)) + (1.0f-alpha)*((float)*(pixel+1)-*(m+1))*((float)*(pixel+1)-*(m+1));
              *(v+2) = alpha*(*(v+2)) + (1.0f-alpha)*((float)*(pixel+2)-*(m+2))*((float)*(pixel+2)-*(m+2));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              *(m+1) = alpha*(*(m+1)) + (1.0f-alpha)*((float)*(pixel+1));
              *(m+2) = alpha*(*(m+2)) + (1.0f-alpha)*((float)*(pixel+2));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v+1) < Vmin) *(v+1) = Vmin;
              if (*(v+2) < Vmin) *(v+2) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
              if (*(v+1) > Vmax) *(v+1) = Vmax;
              if (*(v+2) > Vmax) *(v+2) = Vmax;
            }
            pixel += 3;
            m     += 3;
            v     += 3;
            break;
          case 4:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                          (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                          (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2)))) &&
                          (*(pixel+3) < (*(m+3) + K*(*(v+3)))) && (*(pixel+3) > (*(m+3) - K*(*(v+3)))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(v+1) = alpha*(*(v+1)) + (1.0f-alpha)*((float)*(pixel+1)-*(m+1))*((float)*(pixel+1)-*(m+1));
              *(v+2) = alpha*(*(v+2)) + (1.0f-alpha)*((float)*(pixel+2)-*(m+2))*((float)*(pixel+2)-*(m+2));
              *(v+2) = alpha*(*(v+3)) + (1.0f-alpha)*((float)*(pixel+3)-*(m+3))*((float)*(pixel+3)-*(m+3));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              *(m+1) = alpha*(*(m+1)) + (1.0f-alpha)*((float)*(pixel+1));
              *(m+2) = alpha*(*(m+2)) + (1.0f-alpha)*((float)*(pixel+2));
              *(m+3) = alpha*(*(m+3)) + (1.0f-alpha)*((float)*(pixel+3));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v+1) < Vmin) *(v+1) = Vmin;
              if (*(v+2) < Vmin) *(v+2) = Vmin;
              if (*(v+3) < Vmin) *(v+3) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
              if (*(v+1) > Vmax) *(v+1) = Vmax;
              if (*(v+2) > Vmax) *(v+2) = Vmax;
              if (*(v+3) > Vmax) *(v+3) = Vmax;
            }
            pixel += 4;
            m     += 4;
            v     += 4;
            break;
          }
        }
      }
    }
  }


	return 0;
}



///  Learn part of the image from a new image (next learning iteration).
///
///  \param image			new disparity image
///  \param rect      the area of the image to consider
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int
tlBackground::update (tlImage *image,
                      tlRect  *rect)
{
	int        i,j;
	float      *m;
	float      *v;
  int        x1,x2,y1,y2;


	// only if learning
	if (learn == 0) return 0;

  // check for format consistency
  if (image->format != format) {
    TL_ERRNO_MSG ("tlBackground::update()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

	// shortcuts
  x1    = TL_MAX(rect->x, 0);
  x2    = TL_MIN((rect->x + rect->width), width);
  y1    = TL_MAX(rect->y, 0);
  y2    = TL_MIN((rect->y + rect->height), height);

  // based on format
  switch (format) {

  case STEREO:
    {
      tlHRPixel *pixel;

	    // update values
	    for (j=y1; j<y2; j++) {
        // first column
        pixel = image->hrpixel + (j*width + x1);
	      m     = &(mean[(j*width + x1)]);
	      v     = &(variance[(j*width + x1)]);
        for (i=x1; i<x2; i++) {
			    // update as necessary
          if (*(m) >= 0) {
            if (*(pixel) < (*(m) + K*(*(v)))) {
              *(v) = alpha*(*(v)) + (1.0f-alpha)*((float)*(pixel)-*(m))*((float)*(pixel)-*(m));
              *(m) = alpha*(*(m)) + (1.0f-alpha)*((float)*(pixel));
              if (*(v) < Vmin) *(v) = Vmin;
            }
          }
          else {
            if (*(pixel) >= 0) {
              *(m) = *(pixel);
              *(v) = Vmin;
            }
          }
          pixel++;
          m++;
          v++;
        }
	    }
    }
    break;

  default:
    {
      tlPixel *pixel;

      // update values
      x2 = pixelWidth * (x2-x1);
	    for (j=y1; j<y2; j++) {
        // first column
        pixel = image->pixel + pixelWidth * (j*width + x1);
	      m     = &(mean[pixelWidth*(j*width + x1)]);
	      v     = &(variance[pixelWidth*(j*width + x1)]);
        for (i=0; i<x2; i++) {
			    switch (pixelWidth) {
          case 1:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
            }
            pixel += 1;
            m     += 1;
            v     += 1;          
            break;
          case 2:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                          (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(v+1) = alpha*(*(v+1)) + (1.0f-alpha)*((float)*(pixel+1)-*(m+1))*((float)*(pixel+1)-*(m+1));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              *(m+1) = alpha*(*(m+1)) + (1.0f-alpha)*((float)*(pixel+1));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v+1) < Vmin) *(v+1) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
              if (*(v+1) > Vmax) *(v+1) = Vmax;
            }
            pixel += 2;
            m     += 2;
            v     += 2;          
            break;
          case 3:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                          (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                          (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2)))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(v+1) = alpha*(*(v+1)) + (1.0f-alpha)*((float)*(pixel+1)-*(m+1))*((float)*(pixel+1)-*(m+1));
              *(v+2) = alpha*(*(v+2)) + (1.0f-alpha)*((float)*(pixel+2)-*(m+2))*((float)*(pixel+2)-*(m+2));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              *(m+1) = alpha*(*(m+1)) + (1.0f-alpha)*((float)*(pixel+1));
              *(m+2) = alpha*(*(m+2)) + (1.0f-alpha)*((float)*(pixel+2));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v+1) < Vmin) *(v+1) = Vmin;
              if (*(v+2) < Vmin) *(v+2) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
              if (*(v+1) > Vmax) *(v+1) = Vmax;
              if (*(v+2) > Vmax) *(v+2) = Vmax;
            }
            pixel += 3;
            m     += 3;
            v     += 3;
            break;
          case 4:
            if (adapt || ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                          (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                          (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2)))) &&
                          (*(pixel+3) < (*(m+3) + K*(*(v+3)))) && (*(pixel+3) > (*(m+3) - K*(*(v+3)))))) {
              *(v  ) = alpha*(*(v  )) + (1.0f-alpha)*((float)*(pixel  )-*(m  ))*((float)*(pixel  )-*(m  ));
              *(v+1) = alpha*(*(v+1)) + (1.0f-alpha)*((float)*(pixel+1)-*(m+1))*((float)*(pixel+1)-*(m+1));
              *(v+2) = alpha*(*(v+2)) + (1.0f-alpha)*((float)*(pixel+2)-*(m+2))*((float)*(pixel+2)-*(m+2));
              *(v+2) = alpha*(*(v+3)) + (1.0f-alpha)*((float)*(pixel+3)-*(m+3))*((float)*(pixel+3)-*(m+3));
              *(m  ) = alpha*(*(m  )) + (1.0f-alpha)*((float)*(pixel  ));
              *(m+1) = alpha*(*(m+1)) + (1.0f-alpha)*((float)*(pixel+1));
              *(m+2) = alpha*(*(m+2)) + (1.0f-alpha)*((float)*(pixel+2));
              *(m+3) = alpha*(*(m+3)) + (1.0f-alpha)*((float)*(pixel+3));
              if (*(v  ) < Vmin) *(v  ) = Vmin;
              if (*(v+1) < Vmin) *(v+1) = Vmin;
              if (*(v+2) < Vmin) *(v+2) = Vmin;
              if (*(v+3) < Vmin) *(v+3) = Vmin;
              if (*(v  ) > Vmax) *(v  ) = Vmax;
              if (*(v+1) > Vmax) *(v+1) = Vmax;
              if (*(v+2) > Vmax) *(v+2) = Vmax;
              if (*(v+3) > Vmax) *(v+3) = Vmax;
            }
            pixel += 4;
            m     += 4;
            v     += 4;
            break;
          }
        }
      }
    }
  }


	return 0;
}



///  Get an image of the current learned background.
///
///  \param dest			\b [out] valid image pointer that will contain the background mask.
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int
tlBackground::get (tlImage *dest)
{
	int              i,j;
  int              x2;
	register tlPixel *ptr;
	register float   *m;


	// init
	dest->init (format, width, height);

	// shortcuts
	ptr = dest->pixel;
	m   = mean;

	// update values
  x2 = pixelWidth * width;
	for (j=0; j<height; j++) {
		for (i=0; i<x2; i++) {
			*(ptr++) = (tlPixel)*(m++);
		}
	}


	return 0;
}



///  Get an image of the current variance distribution.
///
///  \param dest			\b [out] valid image pointer that will contain the variance image.
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int
tlBackground::getVariance (tlImage *dest)
{
	int              i,j;
  int              x2;
	register tlPixel *ptr;
	register float   *v;


	// init
	dest->init (format, width, height);

	// shortcuts
	ptr = dest->pixel;
	v   = variance;

	// update values
  x2 = pixelWidth * width;
	for (j=0; j<height; j++) {
		for (i=0; i<x2; i++) {
			*(ptr++) = (tlPixel)*(v++);
		}
	}


	return 0;
}



///  Filter the background out of a given image.
///
///  \param image			the image to filter.
///  
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int 
tlBackground::filter (tlImage *image)
{
	int        i,j;
	float      *m;
	float      *v;


  // check for format consistency
  if (image->format != format) {
    TL_ERRNO_MSG ("tlBackground::update()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

	// shortcuts
	m     = mean;
	v     = variance;

  // based on format
  switch (format) {

  case STEREO:
    {
      tlHRPixel *pixel = image->hrpixel;

      // compute
      for (j=0; j<height; j++) {
        for (i=0; i<width; i++) {
          switch (learn) {
          case 0:
            if ((*m < 0) && (*(pixel) < 0)) *(pixel) = 0;
            else if ((*(m) >= 0) && (*(pixel) < (Koffset + *(m) + K*(*(v))))) *(pixel) = 0;
            break;
          case 1:
            if (*(pixel) >= 0) { if (*(pixel) < (Koffset + *(m) + K*(*(v)))) *(pixel) = 0; }
            else  *(pixel) = 0;
            break;
          }
          pixel++;
          m++;
          v++;
        }
      }
    }
    break;

  default:
    {
      tlPixel *pixel = image->pixel;

      // compute
      for (j=0; j<height; j++) {
        for (i=0; i<width; i++) {
          switch (pixelWidth) {
          case 1:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  ))))) {
              *(pixel) = 0;
            }
            pixel ++;
            v     ++;
            m     ++;
            break;
          case 2:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1))))) {
                *(pixel) = *(pixel+1) = 0;
            }
            pixel += 2;
            v     += 2;
            m     += 2;
            break;
          case 3:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2))))) {
              *(pixel) = *(pixel+1) = *(pixel+2) = 0;
            }
            pixel += 3;
            v     += 3;
            m     += 3;
            break;
          case 4:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2)))) &&
                (*(pixel+3) < (*(m+3) + K*(*(v+3)))) && (*(pixel+3) > (*(m+3) - K*(*(v+3))))) {
              *(pixel) = *(pixel+1) = *(pixel+2) = *(pixel+3) = 0;
            }
            pixel += 4;
            v     += 4;
            m     += 4;
            break;
          }
        }
      }
    }
  }
  

	return 0;
}



///  Filter the background out of a portion of a given image.
///
///  \param image			the image to filter.
///  \param rect      the portion of the image to filter.
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.

int 
tlBackground::filter (tlImage *image,
                      tlRect  *rect)
{
	int        i,j;
	float      *m;
	float      *v;
  int        x1,x2,y1,y2;


  // check for format consistency
  if (image->format != format) {
    TL_ERRNO_MSG ("tlBackground::update()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

	// shortcuts
  x1    = TL_MAX(rect->x, 0);
  x2    = TL_MIN((rect->x + rect->width), width);
  y1    = TL_MAX(rect->y, 0);
  y2    = TL_MIN((rect->y + rect->height), height);

  // based on format
  switch (format) {

  case STEREO:
    {
      tlHRPixel *pixel = image->hrpixel;

      // compute
	    for (j=y1; j<y2; j++) {
        // first column
        pixel = image->hrpixel + (j*width + x1)*sizeof(tlPixel);
	      m     = &(mean[(j*width + x1)]);
	      v     = &(variance[(j*width + x1)]);
		    for (i=x1; i<x2; i++) {
			    switch (learn) {
			    case 0:
				    if ((*m < 0) && (*(pixel) < 0)) *(pixel) = 0;
				    else if ((*(m) >= 0) && (*(pixel) < (Koffset + *(m) + K*(*(v))))) *(pixel) = 0;
				    break;
			    case 1:
				    if (*(pixel) >= 0) { if (*(pixel) < (Koffset + *(m) + K*(*(v)))) *(pixel) = 0; }
				    else  *(pixel) = 0;
				    break;
			    }
			    pixel++;
			    m++;
			    v++;
		    }
	    }
    }
    break;
  
  default:
    {
      tlPixel *pixel = image->pixel;

      // compute
	    for (j=y1; j<y2; j++) {
        // first column
        pixel = image->pixel + pixelWidth*(j*width + x1)*sizeof(tlPixel);
	      m     = &(mean[pixelWidth*(j*width + x1)]);
	      v     = &(variance[pixelWidth*(j*width + x1)]);
		    for (i=x1; i<x2; i++) {
          switch (pixelWidth) {
			    case 1:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  ))))) {
              *(pixel) = 0;
            }
            pixel ++;
            v     ++;
            m     ++;
            break;
          case 2:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1))))) {
                *(pixel) = *(pixel+1) = 0;
            }
            pixel += 2;
            v     += 2;
            m     += 2;
            break;
          case 3:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2))))) {
              *(pixel) = *(pixel+1) = *(pixel+2) = 0;
            }
            pixel += 3;
            v     += 3;
            m     += 3;
            break;
          case 4:
            if ((*(pixel  ) < (*(m  ) + K*(*(v  )))) && (*(pixel  ) > (*(m  ) - K*(*(v  )))) && 
                (*(pixel+1) < (*(m+1) + K*(*(v+1)))) && (*(pixel+1) > (*(m+1) - K*(*(v+1)))) &&
                (*(pixel+2) < (*(m+2) + K*(*(v+2)))) && (*(pixel+2) > (*(m+2) - K*(*(v+2)))) &&
                (*(pixel+3) < (*(m+3) + K*(*(v+3)))) && (*(pixel+3) > (*(m+3) - K*(*(v+3))))) {
              *(pixel) = *(pixel+1) = *(pixel+2) = *(pixel+3) = 0;
            }
            pixel += 4;
            v     += 4;
            m     += 4;
            break;
          }
		    }
      }
	  }
  }


	return 0;
}
