/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlBlobs.h"
#include "tlImage.h"
#include "tlRect.h"
#include "tlMask.h"



///  \file tlBlobs.h
///  \brief Blobs extraction, labeling, and filtering.

///  \class tlBlobs
///  \brief Blobs extraction, labeling, and filtering  class.
///
///  A basic task in image processing, known as "connected components extraction", consists in
///  managing and sorting blobs defined by a series of filters. The tlBlobs object is a
///  collection of tlMask objects that does just that. Given a binary image, it segments,
///  extracts and labels sets of connected pixels into tlMask objects, that can then be
///  sorted with respect to different criterions. This is extremely useful for object extraction.



///  \fn int tlBlobs::getCount()
///  Get the number of blobs currently in the list.
///
///  \return
///  the blob count.



///  Constructor, creates an empty tlBlobs object.
///  Call \ref extract(tlImage *image, int minArea) to perform the extraction.
///
///  \note
///  See \ref errors "error management" for details.

tlBlobs::tlBlobs ()
{
	this->count  = 0;
	this->tcount = 0;
	this->connex = TL_CONNEX_8;
	this->start  = NULL;
	this->end    = NULL;
	this->labels = NULL;

	tl_errno = TL_NO_ERROR;
}



///  Constructor, creates a tlBlobs object from an initial binary image.
///
///  \param image    binary image for blob extraction.
///  \param minArea  minimal area a blob must have to be considered; 
///                  smaller blobs are ignored.
///
///  \note
///  See \ref errors "error management" for details.

tlBlobs::tlBlobs (tlImage *image, 
									int     minArea)
{
	this->count  = 0;
	this->tcount = 0;
	this->connex = TL_CONNEX_8;
	this->start  = NULL;
	this->end    = NULL;
	this->labels = NULL;
	tlRect *rect = new tlRect (0, 0, image->width, image->height);

	processImage (image, rect, minArea);
	delete rect;

	tl_errno = TL_NO_ERROR;
}



///  Constructor, creates a tlBlobs object from a portion of a binary image.
///
///  \param image    binary image for blob extraction.
///  \param rect     portion of the image to process.
///  \param minArea  minimal area a blob must have to be considered; 
///                  smaller blobs are ignored.
///
///  \note
///  See \ref errors "error management" for details.

tlBlobs::tlBlobs (tlImage *image, 
									tlRect  *rect, 
									int     minArea)
{
	this->count  = 0;
	this->tcount = 0;
	this->connex = TL_CONNEX_8;
	this->start  = NULL;
	this->end    = NULL;
	this->labels = NULL;
	processImage (image, rect, minArea);

	tl_errno = TL_NO_ERROR;
}


///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlBlobs::~tlBlobs ()
{
	struct lblob *cur;
	struct lblob *rem;


	// play it safe
	cur = this->start;
	while ((cur != NULL) && (cur != (this->end))) {
		rem = cur;
		cur = cur->next;
		delete (rem->mask);
		free (rem);
	}

	// remove last point
	cur = this->end;
	if (cur != NULL) {
    if (cur->mask != NULL) {
      /// \todo fix blobs destructor crash
      // delete (cur->mask);
    }
		free (cur);
	}
	
  // remove label list
	if (labels)
		free (labels);

	tl_errno = TL_NO_ERROR;
}



///  Reset the blob list (remove all blobs).\n
///  Same as clear().
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::reset ()
{
	return this->clear ();
}



///  Reset the blob list (remove all blobs).\n
///  Same as reset().
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::clear ()
{
	struct lblob *cur;
	struct lblob *rem;


	// play it safe
	if (count > 0) {
		cur = this->start;
		while ((cur != NULL) && (cur != (this->end))) {
			rem = cur;
			cur = cur->next;
			delete (rem->mask);
			free (rem);
		}

		// remove last point
		cur = this->end;
		if (cur != NULL) {
			delete (cur->mask);
			free (cur);
		}
  }

	// reset params
	this->count = 0;
	this->connex = TL_CONNEX_8;
	this->start = NULL;
	this->end = NULL;

	// clean label list
	if (labels) {
		free (labels);
		labels = NULL;
	}
	
  
  return 0;
}



///  Populate the tlBlobs object from a portion of a binary image.
///
///  \param image    binary image for blob extraction.
///  \param rect     portion of the image to process.
///  \param minArea  minimal area a blob must have to enter the list.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::processImage (tlImage *image, 
											 tlRect  *rect, 
											 int     minArea)
{
  // check format
	if (image->pixelWidth != 1) {
		TL_ERRNO_MSG ("tlBlobs::processImage()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  // reset list
	reset ();

	// init
	tcount = findBlobs (image, rect);

	// no blobs found
	if (tcount <= 0) {
		reset ();
		return tcount;
	}

  // filter
	else {
		count = extractBlobs (image, rect);
		if (minArea > 1) {
			filterMinArea (minArea);
		}
		return 0;
	}
}



///  Finds the blobs (low-level). 
///
///  \param	image		the filtered binary image
///	 \param rect		the portion of the image to process
///
///  \return 
///  the blob count, or -1 on overflow error (\ref tl_error "tl_errno" is set).

int
tlBlobs::findBlobs (tlImage *image, 
										tlRect  *rect)
{
	int              i,j;
	int              x1,x2,y1,y2;
	int              label,total_label;
	int              width, height;
	int              *list;
	int              index;
	bool             old;
	bool             equiv[TL_MAX_LABELS][TL_MAX_LABELS];
	tlPixel          *pixel;
	register tlPixel *p;
	tlPixel          *q;


	// shortcuts
	label  = 0;
	width  = image->width;
	height = image->height;
	pixel  = image->pixel;
	x1     = TL_MAX ( (rect->x), 1);
	x2     = TL_MIN (((rect->x) + (rect->width)), width - 1);
	y1     = TL_MAX ( (rect->y), 1);
	y2     = TL_MIN (((rect->y) + (rect->height)), height - 1);

	// label equivalence
	for (i = 0; i < TL_MAX_LABELS; i++) {
		for (j = 0; j < TL_MAX_LABELS; j++) {
			equiv[i][j] = false;
		}
	}

	// deal with borders
	p = pixel + (y1 - 1) * width + (x1 - 1);
	q = pixel + (y2) * width + (x1 - 1);
	for (i = x1 - 1; i < x2 + 1; i++) {
		*(p++) = 0;
		*(q++) = 0;
	}
	p = pixel + (y1 - 1) * width + (x1 - 1);
	q = pixel + (y1 - 1) * width + (x2);
	for (j = y1 - 1; j < y2 + 1; j++) {
		*(p) = 0;
		*(q) = 0;
		p += width;
		q += width;
	}

	// work that magic
	for (j = y1; j < y2; j++) {
		p = pixel + j * width + x1;
		for (i = x1; i < x2; i++) {
			int	num = 0;
			tlPixel	n,nw,ne,w;

			if ((n  = *(p - width))     != 0) num++;
			if ((nw = *(p - width - 1)) != 0) num++;
			if ((ne = *(p - width + 1)) != 0) num++;
			if ((w  = *(p - 1))         != 0) num++;
			if (*(p)) {
				if (num == 0) {
					*(p) = (tlPixel)(++label);
					if (label >= TL_MAX_LABELS) {
						TL_ERRNO_MSG ("tlBlobs::findBlobs()", TL_ERROR_MAX_REACHED);
						return -1;
					}
				}

				else if (num == 1) {
					tlPixel val;
					if (n)       val = n;
					else if (nw) val = nw;
					else if (ne) val = ne;
					else         val = w;
					*(p) = val;
				}

				else {
					tlPixel val;
					if (n)       val = n;
					else if (nw) val = nw;
					else if (ne) val = ne;
					else         val = w;
					*(p) = val;
					if (n) {
						equiv[n][val] = true;
						equiv[val][n] = true;
					}
					if (nw) {
						equiv[nw][val] = true;
						equiv[val][nw] = true;
					}
					if (ne) {
						equiv[ne][val] = true;
						equiv[val][ne] = true;
					}
					if (w) {
						equiv[w][val] = true;
						equiv[val][w] = true;
					}
				}
			}
			p++;
		}
	}
	if (label == 0) {
		return 0;
	}

	// iii) F-W algorithm
	for (j = 0; j <= label; j++) {
		for (i = 0; i <= label; i++) {
			if (i == j)
				equiv[i][j] = true;
			if (equiv[i][j]) {
				for (int k = 0; k <= label; k++) {
					equiv[i][k] = equiv[i][k] || equiv[j][k];
				}
			}
		}
	}

	// build equivalence list
	index = 0;
	labels = (int *) malloc ((label + 1) * sizeof (int));
	list = (int *) malloc ((label + 1) * sizeof (int));
	total_label = label;
	for (j = 0; j <= label; j++) {
		int l = 0;
		while (!(equiv[j][l]))
			l++;
		list[j] = l;
		if (j != l) {
			total_label -= 1;
		}
		// normalization
		if (j > 0) {
			old = false;
			for (i = 0; i < index; i++) {
				if (labels[i] == l)
					old = true;
			}
			if (!old) {
				labels[index++] = l;
			}
		}
		else
			labels[0] = 0;
	}

	// replace class labels
	for (j = y1; j < y2; j++) {
		register tlPixel *p = pixel + j * width + x1;
		for (i = x1; i < x2; i++) {
			*(p) = (tlPixel) list[*(p++)];
		}
	}


	// cleanup
	free (list);
	image->format = GRAY;
	return total_label;
}



///  Extracts the blobs (low-level).
///
///  \param	tlImage *image    filtered binary image
///	 \param tlrect  *rect     portion of the image to process
/// 
///  \return	
///  The actual blob count on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::extractBlobs (tlImage *image, 
											 tlRect  *rect)
{
	int     b;
	int     i,j;
	int     x1, x2, y1, y2;
	int     sx, sy;
	int     width, height;
	int     xmin, xmax, ymin, ymax;
	int     area;
	int     label;
	tlPixel *data;
	tlPixel *p;
	tlPixel *mdata;
	tlMask *tmpMask;


	// check 
	if (!image) {
		TL_ERRNO_MSG ("tlBlobs::extractBlobs()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  // shortcuts
	tmpMask = new tlMask ();
	data = image->pixel;
	width = image->width;
	height = image->height;
	x1 = TL_MAX ((rect->x), 1);
	x2 = TL_MIN (((rect->x) + (rect->width)), width);
	y1 = TL_MAX ((rect->y), 1);
	y2 = TL_MIN (((rect->y) + (rect->height)), height);

	// fill in each blob
	for (b = 0; b < tcount; b++) {
		
    // init
    label = labels[b];
		area  = 0;
		sx    = 0;
		sy    = 0;
		xmin  = width;
		xmax  = 0;
		ymin  = height;
		ymax  = 0;

    // scan for boundaries
		for (j = y1; j < y2; j++) {
			p = data + j * width + x1;
			for (i = x1; i < x2; i++) {
				if (*(p) == label) {
					if (i > xmax) xmax = i;
					if (i < xmin) xmin = i;
					if (j > ymax) ymax = j;
					if (j < ymin)	ymin = j;
					area++;
					sx += i;
					sy += j;
				}
				p++;
			}
		}

		// adjust
		xmax += 1;
		ymax += 1;

		// create blob
		tmpMask->set ((xmax - xmin), (ymax - ymin));
		tmpMask->x = xmin;
		tmpMask->y = ymin;
		tmpMask->area = area;
		if (area != 0) {
			tmpMask->cx = sx / area - xmin;
			tmpMask->cy = sy / area - ymin;
		}
    else continue;

		// fill in blob
		mdata = tmpMask->pixel;
		for (j = ymin; j < ymax; j++) {
			p = data + j * width + xmin;
			for (i = xmin; i < xmax; i++) {
				if (*(p) == label) *(mdata) = 255;
				else               *(mdata) = 0;
				p++;
				mdata++;
			}
		}
		this->add (tmpMask);
	}


	// remove temporary vars
	delete tmpMask;
	return b;
}



///  Perform a blob extraction on the given image. This routine
///  ignores all blobs whose area is smaller than minArea.
///
///  \param image    binary image to perform extraction from
///  \param minArea  minimal blob area [pix]
///    
///  \return
///  The number of blobs found on success.\n
///  On error, -1 is returned and tl_errno is set.\n
///  See \ref tl_errors "TL_ERROR" for details.

int
tlBlobs::extract (tlImage *image, 
									int     minArea)
{
	// check
	if (image->hr) {
    TL_ERRNO_MSG ("tlBlobs::extract()", TL_ERROR_HIGHRES_INCOMPATIBLE);
		return -1;
  }

  clear ();
	processImage (image, (tlRect *) image, minArea);
	
  
  return count;
}



///  Perform a blob extraction on a portion of the given image. 
///  This routine ignores all blobs whose area is smaller than minArea.
///
///  \param image    binary image to perform extraction from
///  \param minArea  minimal blob area [pix]
///  \param rect     portion of the image to process
///    
///  \return
///  The number of blobs found on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::extract (tlImage *image, 
									tlRect  *rect, 
									int     minArea)
{
	// check
	if (image->hr) {
    TL_ERRNO_MSG ("tlBlobs::extract()", TL_ERROR_HIGHRES_INCOMPATIBLE);
		return -1;
  }

  clear ();
	processImage (image, rect, minArea);
	
  
  return count;
}



///  Sort the blobs by area so that the biggest blob is at index 0. 
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::sortByArea ()
{
	int done = 0;


	if (count < 2) return 0;

	while (!done) {
		done       = 1;
		lblob *cur = start;
		while (cur->next != NULL) {
			if (cur->mask->area < cur->next->mask->area) {
				done = 0;
				swap (cur, cur->next);
			}
			cur = cur->next;
		}
	}


	return 0;
}



///  Insert a blob at the end of the list.
///    
///  \param mask         the blob to add
///
///  \return
///  The index of the point if insertion on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::add (tlMask *mask)
{
	// first point case
	if (this->count == 0) {
		this->start = newBlob (mask);
		if (!this->start) {
			TL_ERRNO_MSG ("tlBlobs::add()", TL_ERROR_ALLOCATION_FAILED);
			return -1;
		}
    this->start->index = 0;
		this->start->next  = NULL;
		this->start->prev  = NULL;
		this->end          = this->start;
	}

	// otherwise
	else {
		this->end->next = newBlob (mask);
		if (!this->end->next) {
			TL_ERRNO_MSG ("tlBlobs::add()", TL_ERROR_ALLOCATION_FAILED);
			return -1;
		}
		this->end->next->prev = this->end;
		this->end             = this->end->next;
		this->end->index      = this->count;
		this->end->next       = NULL;
	}
	this->count += 1;


	return ((this->count) - 1);
}



///  Create an element to be inserted in the chained list.
///
///  \param mask     the blob to be added to the list
///    
///  \return
///  A pointer to the new lblob on success, NULL otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The element in the list is a copy of the mask,
///  so that the inserted blob can safely be deleted from the "outside".

struct lblob *
tlBlobs::newBlob (tlMask *mask)
{
	struct lblob *newb;


	newb = (struct lblob *) malloc (sizeof (struct lblob));
	if (!newb) {
		TL_ERRNO_MSG ("tlBlobs::newBlob()", TL_ERROR_ALLOCATION_FAILED);
		return NULL;
	}
	newb->mask = new tlMask (mask);
	newb->index = 0;
	newb->next = NULL;
	newb->prev = NULL;
	
  
  return newb;
}



///  Insert a blob at a specific index in the chained list.
///    
///  \param index     the index to insert to blob at
///  \param mask      the blob to insert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::insert (int    index, 
								 tlMask *mask)
{
	int          i        = 0;
	struct lblob *cur;
	struct lblob *newb;


	if (index >= (this->count)) {
		TL_ERRNO_MSG ("tlBlobs::insert()", TL_ERROR);
		return -1;
	}

	// find insert point
	cur = this->start;
	while (i < index) {
		cur = cur->next;
		i++;
	}

	// create new point
	newb = newBlob (mask);
	if (!newb) {
		TL_ERRNO_MSG ("tlBlobs::insert()", TL_ERROR_ALLOCATION_FAILED);
		return -1;
	}

	// insert new point
	if ((cur->prev) != NULL) cur->prev->next = newb;
	newb->index = index;
	newb->next  = cur;
	newb->prev  = cur->prev;
	cur->prev   = newb;

	// boundary
	if (index == 0) this->start = newb;

	// update following points index
	this->count += 1;
	i = index + 1;
	while (i < (this->count)) {
		cur->index += 1;
		cur = cur->next;
		i++;
	}


	return 0;
}



///  Remove a blob at a specific index from the chained list.
///    
///  @param index     the index of the blob to remove
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::remove (int index)
{
	int          i      = 0;
	struct lblob *cur;
	struct lblob *tmp;


	// check
	if (index >= (this->count)) {
		TL_ERRNO_MSG ("tlBlobs::remove()", TL_ERROR);
		return -1;
	}
	
  // start boundary 
	if (index == 0) {
		if ((this->count) == 1) {
			return (this->clear ());
		}
		cur = this->start;
		this->start = cur->next;
		this->start->prev = cur->prev;
	}
	
  // end boundary
	else if (index == ((this->count) - 1)) {
		cur = this->end;
		this->end = cur->prev;
		this->end->next = cur->next;
	}
	
  // any other case
	else {
		// retrieve point
		cur = this->start;
		while (i < index) {
			cur = cur->next;
			i++;
		}

		// redirect pointers
		cur->prev->next = cur->next;
		cur->next->prev = cur->prev;
	}

	// remove point
	tmp = cur;
	cur = cur->next;
	delete (tmp->mask);
	free (tmp);

	// restore index 
	this->count -= 1;
	i = index;
	while (i < (this->count)) {
		cur->index -= 1;
		cur = cur->next;
		i++;
	}


	return 0;
}



///  Remove a blob from the chained list.
///    
///  \param blob     the blob to remove
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::remove (lblob *blob)
{
	int          i      = 0;
	struct lblob *cur;
	struct lblob *tmp;


	// singleton
	if (count == 1) {
		count = 0;
		delete (start->mask);
		free (start);
		start = NULL;
		return 0;
	}
	
  // start boundary 
	if (blob == start) {
		cur = this->start;
		this->start = cur->next;
		this->start->prev = cur->prev;
	}
	
  // end boundary
	else if (blob->index == ((this->count) - 1)) {
		cur = this->end;
		this->end = cur->prev;
		this->end->next = cur->next;
	}
	
  // any other case
	else {
		cur = blob;

		// redirect pointers
		cur->prev->next = cur->next;
		cur->next->prev = cur->prev;
	}

	// remove point
	tmp = cur;
	cur = cur->next;
	i = tmp->index;
	delete (tmp->mask);
	free (tmp);

	// restore index 
	this->count -= 1;
	while (i < (this->count)) {
		cur->index -= 1;
		cur = cur->next;
		i++;
	}
	
  
  return 0;
}



///  Swap two blobs within the list.
///
///  \param blob1    the first blob
///  \param blob2    the second blob
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::swap (lblob * blob1, 
							 lblob * blob2)
{
	tlMask *tmp;


	// general case
	tmp = blob1->mask;
	blob1->mask = blob2->mask;
	blob2->mask = tmp;
	
  
  return 0;
}



///  Swap two blobs within the list.
///
///  \param index1    the first blob
///  \param index2    the second blob
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::swap (int index1, 
							 int index2)
{
	int          i, temp;
	struct lblob *cur;
	struct lblob *lblob1;
	struct lblob *lblob2;


	// shortcuts
	if (index1 > index2) {
		temp = index1;
		index1 = index2;
		index2 = temp;
	}
	
  // check
	if ((index1 >= (this->count)) || (index2 >= (this->count))) {
		TL_ERRNO_MSG ("tlBlobs::swap()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
		return -1;
	}
	
  // retrieve index1
	cur = this->start;
	for (i = 0; i < index1; i++) {
		cur = cur->next;
	}
	lblob1 = cur;

	// retrieve index2
	for (i = index1; i < index2; i++) {
		cur = cur->next;
	}
	lblob2 = cur;
	
  
  return this->swap (lblob1, lblob2);
}



///  Get the blob at a given index in the chain (if it exists).
///
///  \param index       point index
///    
///  \return
///  A pointer to the existing tlMask *  success, NULL otherwise. 
///  See \ref errors "error management" for details.
///
///  \remark
///  the pointer returned points to the actual mask in the chained list.
///  Thus, it should NOT be deleted. Use getBlobCopy() to get a copy
///  of the blob.

tlMask *
tlBlobs::getBlob (int index)
{
	int	         i;
	struct lblob *cur;


	// check
	if (index >= (this->count)) {
		TL_ERRNO_MSG ("tlBlobs::getBlob()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
		return NULL;
	}
	// retrieve
	cur = this->start;
	for (i = 0; i < index; i++) cur = cur->next;


	return cur->mask;
}



///  Fill an existing blob with data from the blob at a given 
///  index in the chain (if it exists).
///
///  \param index       point index
///  \param mask        mask to fill
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  getBlob (int index)
///  getBlobCopy (int index)

int
tlBlobs::getBlob (int index, tlMask * mask)
{
	int    i;
	struct lblob *cur;

	// check
	if (index >= (this->count)) {
		TL_ERRNO_MSG ("tlBlobs::getBlob()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
		return -1;
	}
	// retrieve
	cur = this->start;
	for (i = 0; i < index; i++)
		cur = cur->next;
	return cur->mask->copyTo (mask);
}



///  Get a copy of the blob at a given index in the chain (if it exists).
///
///  \param index       point index
///    
///  \return
///  A pointer to a copy of the existing tlMask on success, NULL otherwise. 
///  See \ref errors "error management" for details.
///
///  \remark
///  the pointer returned points to a new mask.
///  Thus, it should be deleted after use. 
///  Use getBlobCopy() to get a pointer to the actual blob.

tlMask *
tlBlobs::getBlobCopy (int index)
{
	int          i;
	struct lblob *cur;


	// check
	if (index >= (this->count)) {
		TL_ERROR_MSG ("tlBlobs::getBlob() : invalid index");
		return NULL;
	}
	
  // retrieve
	cur = this->start;
	for (i = 0; i < index; i++) cur = cur->next;
	
  
  return new tlMask (cur->mask);
}




///  Filters the blobs according to their apparent area
///  and deletes blob which do not satisfy the criterias
///
///  \param	minArea   minimal blob area [pix]
///  \param	maxArea   maximal blob area [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlBlobs::filterArea (int minArea,
                     int maxArea)
{
  lblob *cur = start;
  lblob *tmp;


  while (cur != NULL) {
    if ((cur->mask->area < minArea) || (cur->mask->area > maxArea)) {
      tmp = cur->next;
      remove (cur);
      cur = tmp;
    }
    else {
      cur = cur->next;
    }
  }


  return 0;
}



///  Filters the blobs according to their apparent area
///  and deletes blob which do not satisfy the criterias
///
///  \param	minArea		minimal blob area [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::filterMinArea (int minArea)
{
  lblob *cur = start;
  lblob *tmp;

  while (cur != NULL) {
    if ((cur->mask->area) < minArea) {
      tmp = cur->next;
      remove (cur);
      cur = tmp;
    }
    else {
      cur = cur->next;
    }
  }

  return 0;
}



///  Filters the blobs according to their apparent area
///  and deletes blob which do not satisfy the criterias
///
///  \param	maxArea		maximal blob area [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlBlobs::filterMaxArea (int maxArea)
{
  lblob *cur = start;
  lblob *tmp;


  while (cur != NULL) {
    if ((cur->mask->area) > maxArea) {
      tmp = cur->next;
      remove (cur);
      cur = tmp;
    }
    else {
      cur = cur->next;
    }
  }


  return 0;
}
