/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



// SVS for Linux broken for now (has to do with gcc version)
#ifndef LINUX


#include "tlCameraSVS.h"



///  \file tlCameraSVS.h
///  \brief \ref svsDoc "SVS" wrapper header for TLIB.

///  \class tlCameraSVS
///  SRI's Small Vision System wrapper for TLIB.

///  \def TL_CAMERA_SVS_FILENAME
///  Default parameter file for SVS configuration.

///  \fn tlCameraSVS::getLeftSource ()
///  Get the image source corresponding to the left image sensor.
///
///  \return 
///  A pointer to a valid tlSource.

///  \fn tlCameraSVS::getRightSource ()
///  Get the image source corresponding to the right image sensor.
///
///  \return 
///  A pointer to a valid tlSource.



///  Default constructor. Uses the default configuration file defined
///  in TL_CAMERA_SVS_FILENAME.
///  
///  \note
///  See \ref errors "error management" for details.

tlCameraSVS::tlCameraSVS () : tlCameraSVSBase ()
{
  if (tl_errno != TL_NO_ERROR) return;

  leftSource  = NULL;
  rightSource = NULL;

  readConfig (TL_CAMERA_SVS_FILENAME);
}



///  Specifies which configuration file to use.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraSVS::tlCameraSVS (char *filename) : tlCameraSVSBase ()
{
  if (tl_errno != TL_NO_ERROR) return;

  leftSource  = NULL;
  rightSource = NULL;

  readConfig (filename);
}



///  Default destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraSVS::~tlCameraSVS () 
{
  tl_errno = TL_NO_ERROR;
}



///  Specifies which image sources to use for image acquisition.
///  Both image sources must be valid and ready to provide images via
///  the grab() and grabQueued() methods.
///
///  \param leftSource    valid pointer to a source corresponding to the left image sensor.
///  \param rightSource   valid pointer to a source corresponding to the right image sensor.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlCameraSVS::setSource (tlSource *leftSource,
                        tlSource *rightSource)
{
  this->leftSource  = leftSource;
  this->rightSource = rightSource;


  return 0;
}



///  Specifies which image source to use for left image acquisition.
///  Both image sources must be valid and ready to provide images via
///  the grab() and grabQueued() methods.
///
///  \param source    valid pointer to a source corresponding to the left image sensor.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlCameraSVS::setLeftSource (tlSource *source)
{
  this->leftSource = source;


  return 0;
}



///  Specifies which image source to use for right image acquisition.
///  Both image sources must be valid and ready to provide images via
///  the grab() and grabQueued() methods.
///
///  \param source    valid pointer to a source corresponding to the right image sensor.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlCameraSVS::setRightSource (tlSource *source)
{
  this->rightSource = source;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grab (tlImage *image)
{
  return leftSource->grab (image);
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grab (tlImage *image,
                   tlRect  *rect) 
{
  return leftSource->grab (image, rect);
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grab (tlImage *image,
                   tlImage *disparity) 
{
  tlImage *right = new tlImage ();


  leftSource->grab  (image);
  rightSource->grab (right);
  getDisparity (image, right, disparity);
  delete right;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grab (tlImage *image,
                   tlImage *disparity,
                   tlRect  *rect) 
{
  tlImage *right = new tlImage ();


  leftSource->grab  (image, rect);
  rightSource->grab (right, rect);
  getDisparity (image, right, disparity, rect);
  delete right;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grabQueued (tlImage *image) 
{
  return grab (image);
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grabQueued (tlImage *image,
                         tlImage *disparity) 
{
  return grab (image, disparity);
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grabQueued (tlImage *image,
                         tlImage *disparity,
                         tlRect  *rect) 
{
  return grab (image, disparity, rect);
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVS::grabQueued (tlImage *image,
                         tlRect  *rect) 
{
  return grab (image, rect);
}


#endif
