/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////


// SVS broken under Linux for now (has to do with gcc version)
#ifndef LINUX


#include "tlCameraSVS.h"



///  \file tlCameraSVSBase.h
///  \brief Base class for \ref svsDoc "SVS"-based stereo cameras.

///  \class tlCameraSVSBase
///  Base class for SVS-based stereo cameras. Programmers should not use it, but
///  use tlCameraSVS or tlCameraSVSHead instead.



///  \fn tlCameraSVSBase::getOffX ()
///  Get the horopter offset from the SVS parameters structure.
///  
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.



///  \fn tlCameraSVSBase::getThreshold ()
///  Get the confidence threshold from the SVS parameters structure.
///  
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.



///  \fn tlCameraSVSBase::getNdisp ()
///  Get the number of disparities to compute from the SVS parameters structure.
///  
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.



///  \fn tlCameraSVSBase::getRight(tlImage *image)
///
///  This method returns the full image from the SVS right source acquired during
///  the last grab() call. The image will have the size and format the source
///  is currently configured for.
///
///  \param image   \b [out] a valid image pointer that will receive the image.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraSVSBase::getRight(tlImage *image, tlRect *rect)
///
///  This method returns a portion of the right image from the SVS right source
///  acquired during the last grab() call. The image will have the size and format the source
///  is currently configured for, but only the portion defined by the \e rect argument will contain valid pixel
///  values.
///
///  \param image   \b [out] a valid image pointer that will receive the image.
///  \param rect    a pointer to the portion of the image to acquire.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  Default constructor.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraSVSBase::tlCameraSVSBase ()
{
  // make a stereo processing object
  imageObject   = new svsStereoImage  ();
  processObject = new svsMultiProcess ();
  sourceObject  = NULL;

  // assign global legacy parameters to image
  sp = imageObject->GetSP ();

  // set default parameters
  color = imageObject->haveColor;


  tl_errno = TL_NO_ERROR;
}



///  Default constructor.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraSVSBase::~tlCameraSVSBase ()
{
  // flush it all
  if (imageObject)   delete imageObject;
  if (processObject) delete processObject;

  tl_errno = TL_NO_ERROR;
}



///  Set the number of disparities in the SVS parameters structure.
///
///  \param ndisp     number of disparities [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.

int
tlCameraSVSBase::setNdisp (int ndisp)
{
  // apply to image source, if exists
  if (sourceObject) {
    svsDisparityParams *dp = sourceObject->GetDP();
    dp->ndisp = ndisp;
    sourceObject->SetNDisp (ndisp);
  }


  // apply to current buffer
  imageObject->dp.ndisp = ndisp;
  return 0;
}




///  Set the horopter offset in the SVS parameters structure.
///
///  \param offX      horopter offset on X axis [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.

int
tlCameraSVSBase::setOffX (int offX)
{
  // apply to image source, if exists
  if (sourceObject) {
    svsDisparityParams *dp = sourceObject->GetDP();
    dp->offx = offX;
    sourceObject->SetHoropter (offX);
  }


  // apply to current buffer
  imageObject->dp.offx = offX;
  return 0;
}



///  Set the horopter offset in the SVS parameters structure.
///
///  \param offY      horopter offset on Y axis [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.

int
tlCameraSVSBase::setOffY (int offY)
{
  // apply to image source, if exists
  if (sourceObject) {
    svsDisparityParams *dp = sourceObject->GetDP();
    dp->offy = offY;
  }


  // apply to current buffer
  imageObject->dp.offy = offY;
  return 0;
}



///  Set the confidence threshold in the SVS parameters structure.
///  
///  \param threshold      confidence threshold [arbitrary]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.

int
tlCameraSVSBase::setThreshold (int threshold)
{
  // apply to image source, if exists
  if (sourceObject) {
    svsDisparityParams *dp = sourceObject->GetDP();
    dp->thresh = threshold;
    sourceObject->SetThresh (threshold);
  }


  // apply to current buffer
  imageObject->dp.thresh = threshold;
  return 0;
}



///  Read in the SVS calibration and parameters file.
///  
///  \param filename    configuration file name
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  Read the \ref svsDoc "SVS documentation" for more information.

int
tlCameraSVSBase::readConfig (char *filename)
{
  // read in rectification parameters
  if (!imageObject->ReadParams(filename))  {
    TL_ERRNO_MSG ("tlCameraSVSBase::readConfig()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return -1;
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::getDisparity (tlImage *disparity)
{
  int      i,j;
  short    *hrdata;
  short    *disp;


  // check parameters
  if (!disparity) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // reset image
  disparity->init (STEREO, width, height);

  // init high-res pixel buffer
  hrdata = disparity->hrpixel;
  hrdata = (short *) realloc (hrdata, width * height * sizeof (short));
  if (hrdata == NULL) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // shortcuts
  disparity->hrpixel = hrdata;

  // compute
  processObject->CalcStereo (imageObject);
  disp = imageObject->Disparity ();

  // compute disparity scaling factor
  switch (imageObject->dp.ndisp) {
  case 64: disparity->hrShift = 2; break;
  case 32: disparity->hrShift = 1; break;
  default: disparity->hrShift = 0; break;
  }

  // proceed to conversion
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      *(hrdata++) = *(disp++);
    }
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::getDisparity (tlImage *disparity,
                               tlRect  *rect)
{
  int      i,j;
  int      x1,x2,y1,y2;
  short    *hrdata;
  short    *disp;


  // check parameters
  if (!disparity) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // reset image
  disparity->init (STEREO, width, height);
  
  // init high-res pixel buffer
  hrdata = disparity->hrpixel;
  hrdata = (short *) realloc (hrdata, width * height * sizeof (short));
  if (hrdata == NULL) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // shortcuts
  disparity->hrpixel = hrdata;
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x)+(rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y)+(rect->height)), height);

  // compute
  processObject->CalcStereo (imageObject);
  color = imageObject->haveColor;
  disp = imageObject->Disparity ();

  // compute disparity scaling factor
  switch (imageObject->dp.ndisp) {
  case 64: disparity->hrShift = 2; break;
  case 32: disparity->hrShift = 1; break;
  default: disparity->hrShift = 0; break;
  }

  // proceed to conversion
  for (j=y1; j<y2; j++) {
    register short   *h = hrdata + j*width + x1;
    register short   *d = disp + j*width + x1;
    for (i=x1; i<x2; i++) {
      *(h++) = *(d++);
    }
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::getDisparity (tlImage *leftImage,
                               tlImage *rightImage,
                               tlImage *disparity)
{
  int      i,j;
  short    *hrdata;
  short    *disp;
  tlImage  *lim;
  tlImage  *rim;


  // check parameters
  if (!disparity || !leftImage || !rightImage) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check for calibration consistency
  if (leftImage->width  != width || leftImage->height  != height ||
      rightImage->width != width || rightImage->height != height) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // reset image
  disparity->init (STEREO, width, height);

  // init high-res pixel buffer
  hrdata = disparity->hrpixel;
  hrdata = (short *) realloc (hrdata, width * height * sizeof (short));
  if (hrdata == NULL) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // shortcuts
  disparity->hrpixel = hrdata;

  // set source images
  lim = new tlImage ();
  rim = new tlImage ();
  leftImage->toGray  (lim);
  rightImage->toGray (rim);
  imageObject->SetImages(lim->pixel, rim->pixel, NULL);

  // compute
  processObject->CalcStereo (imageObject);
  color = imageObject->haveColor;
  disp = imageObject->Disparity ();

  // compute disparity scaling factor
  switch (imageObject->dp.ndisp) {
  case 64: disparity->hrShift = 2; break;
  case 32: disparity->hrShift = 1; break;
  default: disparity->hrShift = 0; break;
  }

  // proceed to conversion
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      *(hrdata++) = *(disp++);
    }
  }


  return 0;
}




////////////////////////////////////////////////////////////////////////
///  \todo make it effectively compute within \e rect only

int
tlCameraSVSBase::getDisparity (tlImage *leftImage,
                               tlImage *rightImage,
                               tlImage *disparity,
                               tlRect  *rect)
{
  int      i,j;
  int      x1,x2,y1,y2;
  short    *hrdata;
  short    *disp;
  tlImage  *lim;
  tlImage  *rim;


  // check parameters
  if (!disparity || !leftImage || !rightImage) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check for calibration consistency
  if (leftImage->width  != width || leftImage->height  != height ||
      rightImage->width != width || rightImage->height != height) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // reset image
  disparity->init (STEREO, width, height);

    // init high-res pixel buffer
  hrdata = disparity->hrpixel;
  hrdata = (short *) realloc (hrdata, width * height * sizeof (short));
  if (hrdata == NULL) {
    TL_ERRNO_MSG ("tlCameraSVSBase::getDisparity()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // shortcuts
  disparity->hrpixel = hrdata;
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x)+(rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y)+(rect->height)), height);

  // set source images
  lim = new tlImage ();
  rim = new tlImage ();
  leftImage->toGray  (lim);
  rightImage->toGray (rim);
  imageObject->SetImages(lim->pixel, rim->pixel, NULL);

  // compute
  processObject->CalcStereo (imageObject);
  color = imageObject->haveColor;
  disp = imageObject->Disparity ();

  // compute disparity scaling factor
  switch (imageObject->dp.ndisp) {
  case 64: disparity->hrShift = 2; break;
  case 32: disparity->hrShift = 1; break;
  default: disparity->hrShift = 0; break;
  }

  // proceed to conversion
  for (j=y1; j<y2; j++) {
    register short   *h = hrdata + j*width + x1;
    register short   *d = disp + j*width + x1;
    for (i=x1; i<x2; i++) {
      *(h++) = *(d++);
    }
  }


  return 0;
}




////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::getRight (tlImage *image) 
{
  if (imageObject->haveColorRight) {
    image->init (RGB, width, height);
    delete image->pixel;
    if (!memcpy (image->pixel, imageObject->Right(), width*height*TL_RGB_WIDTH)) {
      TL_ERROR_MSG ("tlCameraSVSBase->getRight() : error copying data");
    }
  }
  
  // gray
  else {
    image->init (GRAY, width, height);
    delete image->pixel;
    if (!memcpy (image->pixel, imageObject->Right(), width*height*TL_GRAY_WIDTH)) {
      TL_ERROR_MSG ("tlCameraSVSBase->getRight() : error copying data");
    }
  }


  return 0;
}

////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::getRight (tlImage *image,
                           tlRect  *rect) 
{
  int           j;
  int           x1,x2,y1,y2;
  int           size;
  tlPixel       *data;
  unsigned char *rim;


  // initialize image
  image->init (GRAY, width, height);

  // shortcut
  x1   = TL_MAX ((rect->x), 0);
  x2   = TL_MIN (((rect->x)+(rect->width)), width);
  y1   = TL_MAX ((rect->y), 0);
  y2   = TL_MIN (((rect->y)+(rect->height)), height);
  size = (x2-x1);
  data = (image->pixel) + y1*width+x1;
  rim  = imageObject->Right() + y1*width+x1;

  // copy the thing
  for (j=y1; j<y2; j++) {
    data += width;
    rim  += width;
    memcpy (data, rim, size);
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::toWorld (tlLocation *location)
{
  float X,Y,Z;


  svsReconstruct3D (&X, &Y, &Z, 
    (float)(location->x), 
    (float)(location->y), 
    (float)(location->disp),
    sp);

  location->setWorldLocation (X, Y, Z);


  return 0;
}




////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::toWorld (tlLocation *location,
                          tlLocation *dest)
{
  float X,Y,Z;


  svsReconstruct3D (&X, &Y, &Z, 
    (float)(location->x), 
    (float)(location->y), 
    (float)(location->disp), 
    sp);

  location->toLocation (dest);
  dest->setWorldLocation (X, Y, Z);


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::toImage (tlLocation *location)
{
  float x,y;

  
  svsProject3D (&x, &y, 
    location->wx, 
    location->wy, 
    location->wz,
    sp,
    svsLEFT);
  
  location->setLocation ((int)x, (int)y);

  
  return 0;
}




////////////////////////////////////////////////////////////////////////

int
tlCameraSVSBase::toImage (tlLocation *location,
                          tlLocation *dest)
{
  float x,y;

  
  svsProject3D (&x, &y, 
    location->wx, 
    location->wy, 
    location->wz,
    sp,
    svsLEFT);
  
  location->toLocation (dest);
  dest->setLocation ((int)x, (int)y);

  
  return 0;
}



#endif
