/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlCameraStereo.h"



///  \file tlCameraStereo.h
///  \brief Stereo camera base class for TLIB.

///  \class tlCameraStereo
///  \brief Stereo camera base class.
/// 
///  The tlCameraStereo base class offers an encapsulation for stereo image sources and stereo
///  processing engines. Stereo cameras are defined here as "special" images that can 
///  acquire not only images but also disparity maps, registered to the source image. 



///  Default constructor.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraStereo::tlCameraStereo ():tlCamera ()
{
  tl_errno = TL_NO_ERROR;
}



///  Default destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraStereo::~tlCameraStereo ()
{
  tl_errno = TL_NO_ERROR;
}



///  \fn tlCameraStereo::grab (tlImage *image, tlImage *disparity)
///  This method acquires a full image and the related disparity map. 
///  The image will have the size and format determined by the camera settings.
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
///  \param disparity \b [out] a valid image pointer that will receive the  disparity map.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::grab (tlImage *image, tlImage *disparity, tlRect*rect)
///  This method acquires a portion of an image and the related disparity map. 
///  The image will have the size and format determined by the camera settings,
///  but only the portion defined by the \e rect argument will contain valid pixel
///  values. 
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
///  \param disparity \b [out] a valid image pointer that will receive the  disparity map.
///  \param rect a valid image pointer to the portion of the image to acquire.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::grabQueued (tlImage *image, tlImage *disparity)
///  This call acquires a full image and the related disparity map. 
///  The image will have the size and format determined by the camera settings.
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
///  \param disparity \b [out] a valid image pointer that will receive the  disparity map.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  This method is faster than grab() because image acquisition
///  happens in the background. However, since the grab command is sent 
///  to the grabber during the previous call to the grabQueued() 
///  routine, the image returned can be out-of-date.


///  \fn tlCameraStereo::grabQueued (tlImage *image, tlImage *disparity, tlRect *rect)
///  This method acquires a portion of an image and the related disparity map. 
///  The image will have the size and format determined by the camera settings, 
///  but only the portion defined by the \e rect argument will contain valid pixel
///  values.
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
///  \param disparity \b [out] a valid image pointer that will receive the  disparity map.
///  \param rect a valid image pointer to the portion of the image to acquire.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  This method is faster than grab() because image acquisition
///  happens in the background. However, since the grab command is sent 
///  to the grabber during the previous call to the grabQueued() 
///  routine, the image returned can be out-of-date.



///  \fn tlCameraStereo::getDisparity(tlImage *disparity)
///  This method computes a disparity map on the latest image set acquired  by grab.
///  The disparity map will have the size and format determined by the camera settings.
///  
///  \param disparity \b [out] a valid image pointer that will receive the  disparity map.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.


///  \fn tlCameraStereo::getDisparity(tlImage *disparity, tlRect *rect)
///  This method computes a partial disparity map on the latest image set acquired by grab.
///  The disparity map will have the size and format determined by the camera settings, 
///  but only the portion defined by the \e rect argument will contain valid pixel values.
///  
///  \param disparity \b [out] a valid image pointer that will receive the disparity map.
///  \param rect a valid image pointer to the portion of the disparity map to process.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::getDisparity (tlImage *leftImage, tlImage *rightImage, tlImage *disparity)
///  This method computes the disparity map from two images. The images are supposed to have been
///  acquired by the same tlCameraStereo object. The disparity map will have the size and format 
///  determined by the camera settings.
/// 
///  \param leftImage a pointer to an image taken from the left imager.
///  \param rightImage a pointer to an image taken from the right imager.
///  \param disparity \b [out] a valid image pointer that will receive the disparity map.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::getDisparity (tlImage *leftImage, tlImage *rightImage, tlImage *disparity, tlRect *rect)
///  This method computes the disparity map from a portion of two images. The images are supposed to have been
///  acquired by the same tlCameraStereo object. The disparity map will have the size and format 
///  determined by the camera settings, but only the portion defined by the \e rect argument will contain valid pixel values.
/// 
///  \param leftImage a pointer to an image taken from the left imager.
///  \param rightImage a pointer to an image taken from the right imager.
///  \param disparity \b [out] a valid image pointer that will receive the disparity map.
///  \param rect a valid image pointer to the portion of the disparity map to process.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::toWorld (tlLocation *location)
///  This method computes the 3D coordonates in the world frame of a given location, 
///  given the stereo camera calibration settings. The input location is defined
///  by its x,y position in the image frame and a disparity value. The resulting world coordonates
///  are written in the corresponding \e location fields.
/// 
///  \param location \b [int/out] location to convert
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::toWorld (tlLocation *location, tlLocation *dest)
///  This method computes the 3D coordonates in the world frame of a given location, 
///  given the stereo camera calibration settings. The input location is defined
///  by its x,y position in the image frame and a disparity value. The resulting world coordonates
///  are written in the \e dest location; image coordonates in \e dest are updated accordingly.
/// 
///  \param location location to convert
///  \param dest \b [out] resulting location (in both image and world frame)
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::toImage (tlLocation *location)
///  This method computes the coordonates in the image frame and disparity of a given location, 
///  given the stereo camera calibration settings. The input location is defined
///  by its wx,wy,wz position in the world frame. The resulting image coordonates
///  are written in the corresponding \e location fields.
/// 
///  \param location \b [int/out] location to convert
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlCameraStereo::toImage (tlLocation *location, tlLocation *dest)
///  This method computes the coordonates in the image frame and disparity of a given location, 
///  given the stereo camera calibration settings. The input location is defined
///  by its wx,wy,wz position in the world frame. The resulting image coordonates
///  are written in the \e dest location; world coordonates in \e dest are updated accordingly.
/// 
///  \param location location to convert
///  \param dest \b [out] resulting location (in both image and world frame)
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
