/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlCameraTSAI.h"
#include "tlPoint.h"
#include "tlImage.h"

extern "C" {
#include "Cal_main.h"
}



///  \file tlCameraTSAI.h
///  \brief TSAI camera calibration.

///  \class tlCameraTSAI
///  TSAI camera calibration for TLIB.

///  \def TL_CAMERA_TSAI_FILENAME
///  Default parameter file for TSAI camera calibration.

///  \def TL_CAMERA_TSAI_MAX_POINTS
///  Maximum amount of points allowed for TSAI camera calibration.



///  \fn tlCameraTSAI::getSource()
///  Return a pointer to the image source connected to this camera.
///
///  \return
///  A valid pointer to an image source if source is available, NULL otherwise.\n
///  See \ref errors "error management" for details.



///  Default constructor. Uses the defaults file specified in TL_CAMERA_TSAI_FILENAME,
///  or no calibration if file does not exist.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraTSAI::tlCameraTSAI () : tlCamera ()
{
  if (tl_errno != TL_NO_ERROR) return;

  this->source = NULL;
  if (0 != readConfig (TL_CAMERA_TSAI_FILENAME)) {
    calibrated = false;
  }
}



///  Specify an image source to use with these calibration settings.
///  The source must be valid and ready to provide images via the 
///  grab() or grabQueued() methods.
///
///  \param source    a valid pointer to an image source.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraTSAI::tlCameraTSAI (tlSource *source) : tlCamera ()
{
  if (tl_errno != TL_NO_ERROR) return;

  this->width  = source->getWidth  ();
  this->height = source->getHeight ();
  this->source = source;
}



///  Specify a calibration configuration file to read from.
///
///  \param filename      calibration configuration file
///
///  \note
///  See \ref errors "error management" for details.

tlCameraTSAI::tlCameraTSAI (char *filename) : tlCamera ()
{
  if (tl_errno != TL_NO_ERROR) return;

  // read camera parameters from file
  this->source = NULL;

  // read config
  if (0 != readConfig (filename)) {
    calibrated = false;
  }
}



///  Read calibration data from a given file and specify an image source to use with these calibration settings.
///  The source must be valid and ready to provide images via the 
///  grab() or grabQueued() methods. 
///
///  \param source      a valid pointer to an image source.
///  \param filename    calibration configuration file.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraTSAI::tlCameraTSAI (char     *filename,
                            tlSource *source) : tlCamera ()
{
  if (tl_errno != TL_NO_ERROR) return;

  // read camera parameters from file
  this->width  = source->getWidth  ();
  this->height = source->getHeight ();
  this->source = source;

  // read config
  if (0 != readConfig (filename)) {
    calibrated = false;
  }
}



///  Default destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlCameraTSAI::~tlCameraTSAI (void)
{
  tl_errno = TL_NO_ERROR;
}



///  Read calibration configuration from a given filename.
///
///  \param filename    calibration configuration file.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \todo
///  make sure the width and height params are consistent between
///  source and calibration data.
///
///  \todo
///  Detect invalid calibration data.

int
tlCameraTSAI::readConfig (char *filename)
{
  FILE *pFp; 


  // check
  if (!filename) {
    TL_ERRNO_MSG ("tlCameraTSAI::readFile()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // open file
  if (NULL == (pFp = fopen (filename, "r"))) {
    TL_ERRNO_MSG ("tlCameraTSAI::readFile()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return -1;
  }

  // read file...
  load_cp_cc_data (pFp, &cp, &cc);

  // close file
  fclose (pFp);

  // notify
  calibrated = true;


  return 0;
}



///  Write calibration configuration to a given filename.
///
///  \param filename    calibration configuration file.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlCameraTSAI::writeConfig (char *filename)
{
  FILE *pFp; 
  char readable[255];


  // check
  if (!filename) {
    TL_ERRNO_MSG ("tlCameraTSAI::writeFile()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // create human-readable calibration file
  sprintf (readable, "info-%s", filename);

  // open file
  pFp = fopen (filename, "w");
  if (!pFp) {
    TL_ERRNO_MSG ("tlCameraTSAI::writeFile()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return -1;
  }

  // write file
  dump_cp_cc_data (pFp, &cp, &cc);

  // close file
  fclose (pFp);

  // open file
  pFp = fopen (readable, "w");
  if (!pFp) {
    TL_ERRNO_MSG ("tlCameraTSAI::writeFile()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return -1;
  }

  // write human readable file
  print_cp_cc_data (pFp, &cp, &cc);

  // close file
  fclose (pFp);



  return 0;
}



///  Perform TSAI calibration given a set of data. See \ref tsai "TSAI calibration"
///  for more information on how to perform TSAI camera calibration.
///
///  \param data    a pointer to the calibration data structure
///
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlCameraTSAI::calibrate (tlCameraTSAICalibData *data)
{ 
  int i;


  // set calibration parameters
  if (data->Ncx != -1) {
    cp.Ncx  = data->Ncx;
    cp.Nfx  = data->Nfx;
    cp.dx   = data->dx;
    cp.dy   = data->dy;
    cp.dpx  = data->dpx;
    cp.dpy  = data->dpy;
    cp.Cx   = data->Cx;
    cp.Cy   = data->Cy;
    cp.sx   = data->sx;
  }
  else initialize_photometrics_parms ();

  // set calibration data
  cd.point_count = data->numPoints;
  for (i=0; i<data->numPoints; i++) {
    cd.xw[i] = data->xw[i];
    cd.yw[i] = data->yw[i];
    cd.zw[i] = data->zw[i];
    cd.Xf[i] = data->Xf[i];
    cd.Yf[i] = data->Yf[i];
  }

  // run calibration
  if (data->coplanar) coplanar_calibration_with_full_optimization ();
  else                noncoplanar_calibration_with_full_optimization ();

  // notify
  calibrated = true;


  return 0;
}



///  Specify an image source to use with these calibration settings.
///  The source must be valid and ready to provide images via the 
///  grab() or grabQueued() methods.
///
///  \param source    a valid pointer to an image source.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \todo
///  make sure the width and height params are consistent between
///  source and calibration data.

int
tlCameraTSAI::setSource (tlSource *source)
{
  this->width  = source->getWidth  ();
  this->height = source->getHeight ();
  this->source = source;


  return 0;
}



///  This routine performs an inverse perspective projection to determine
///  the position of a point in world coordinates that corresponds to a
///  given position in image coordinates.  To constrain the inverse
///  projection to a single point the routine requires a Z world
///  coordinate for the point in addition to the X and Y image coordinates.
///
///  \param point   the point to apply the transform to. Members x,y and wz are used as input, output is placed in members wx,wy.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlCameraTSAI::toWorld (tlPoint *point)
{
  double Xfd = (double)(point->x); 
  double Yfd = (double)(point->y);
  double zw  = (double)(point->wz*1000.0f);
  double xw, yw;


  image_coord_to_world_coord (Xfd, Yfd, zw, &xw, &yw);
  point->wx = (float)(xw)/1000.0f;
  point->wy = (float)(yw)/1000.0f;
  

  return 0;
}



///  This routine takes the position of a point in world coordinates [mm]
///  and determines the position of its image in image coordinates [pix].
///
///  \param point   the point to apply the transform to. Members wx,wy,wz are used as input, output is placed in members x,y.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlCameraTSAI::toImage (tlPoint *point)
{
  double xw = (double)(point->wx*1000.0f);
  double yw = (double)(point->wy*1000.0f);
  double zw = (double)(point->wz*1000.0f);
  double Xf, Yf;


  world_coord_to_image_coord (xw, yw, zw, &Xf, &Yf);
  point->x = (int)Xf;
  point->y = (int)Yf;


  return 0;
}



///  Unwarp point for the given camera. That is, return the undistorted
///  coordinate for a point in the image given the calibration settings.
///
///  \param point      the point to project
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  dewarp(tlPoint *point)

int
tlCameraTSAI::dewarp (tlPoint *point)
{
  double Xu, Yu;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if (!point) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // convert
  distorted_to_undistorted_image_coord ((double)(point->x), (double)(point->y), &Xu, &Yu);

  // store
  point->x = (int) Xu;
  point->y = (int) Yu;


  return 0;
}



///  Unwarp point for the given camera. That is, return the undistorted
///  coordinate for a point in the image given the calibration settings.
///  The result is placed in another point.
///
///  \param point      the point to project
///  \param dest       \b [out] the point that will contain the 
///                    undistorted coordinates
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  dewarp(tlPoint *point)

int
tlCameraTSAI::dewarp (tlPoint *point,
                      tlPoint *dest)
{
  double Xu, Yu;


  // make sure...
  if (!calibrated) {
    point->copyTo (dest);
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if ((!point) || (!dest)) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // convert
  distorted_to_undistorted_image_coord ((double)(point->x), (double)(point->y), &Xu, &Yu);

  // store
  dest->x = (int) Xu;
  dest->y = (int) Yu;


  return 0;
}





///  Warp point for the given camera. That is, return the distorted
///  coordinate for a point in a calibrated image given the calibration 
///  settings.
///
///  \param point       the point to project
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  the point coordinates are altered by this call. 
///  See warp(tlPoint *point, tlPoint *dest) to put the
///  warping result in another tlPoint object.
///
///  \sa
///  warp(tlPoint *point, tlPoint *dest)

int
tlCameraTSAI::warp (tlPoint *point)
{
  double Xd, Yd;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if (!point) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // convert
  undistorted_to_distorted_image_coord ((double)(point->x), (double)(point->y), &Xd, &Yd);

  // store
  point->x = (int) Xd;
  point->y = (int) Yd;


  return 0;
}



///  Warp point for the given camera. That is, return the distorted
///  coordinate for a point in a calibrated image given the calibration 
///  settings. The result is placed in another image.
///
///  \param point      the point to project
///  \param dest       \b [out] the point that will contain 
///                    the undistorted coordinates
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  warp(tlPoint *point)

int
tlCameraTSAI::warp (tlPoint *point,
                    tlPoint *dest)
{
  double Xd, Yd;


  // make sure...
  if (!calibrated) {
    point->copyTo (dest);
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if ((!point) || (!dest)) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // convert
  undistorted_to_distorted_image_coord ((double)(point->x), (double)(point->y), &Xd, &Yd);

  // store
  dest->x = (int) Xd;
  dest->y = (int) Yd;


  return 0;
}



///  Warp an image for the given camera. That is, return the 
///  distorted image given the calibration settings.
///
///  \param image       the image to project
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  the image is altered by this call. 
///  See warp(tlImage *image, tlImage *dest) to put the
///  warping result in another tlImage object.
///
///  \sa
///  warp(tlImage *image, tlRect *rect)\n
///  warp(tlImage *image, tlImage *dest)\n
///  warp(tlImage *image, tlRect *rect, tlImage *dest)

int
tlCameraTSAI::warp (tlImage *image)
{
  int     i,j,k;
  int     index, dindex;
  int     width, height;
  int     depth;
  int     byteCount;
  double  dx,dy;
  tlPixel *unwarped;
  tlPixel *warped;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if (!image) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcut
  unwarped   = image->pixel;
  depth      = image->pixelWidth;
  width      = image->width;
  height     = image->height;
  byteCount  = width*height*depth;

  // create temporary buffer
  warped = (tlPixel *) malloc(byteCount*sizeof(tlPixel));
  if (unwarped == NULL) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // warp
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      undistorted_to_distorted_image_coord (i, j, &dx, &dy);
      index  = depth*(j*width+i);
      dindex = depth*((int)dy*width+(int)dx);
      if ((dindex < 0) || (dindex >= byteCount)) 
        for (k=0; k<depth; k++)
          warped[dindex++] = unwarped[index++];
    }
  }

  // store result
  free (image->pixel);
  image->pixel = warped;


  return 0;
}



///  Warp a portion of an image for the given camera. That is, 
///  return the distorted image given the calibration settings.
///
///  \param image      the image to project
///  \param rect       portion of the image to project
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  the image is altered by this call. 
///  See warp(tlImage *image, tlImage *dest) to put the
///  dewarping result in another tlImage object.
///
///  \sa
///  warp(tlImage *image)\n
///  warp(tlImage *image, tlImage *dest)\n
///  warp(tlImage *image, tlRect *rect, tlImage *dest)
///
///  \todo
///  provide an option to use tl_bspline_interpolate() to retrieve best value
///  for real x,y source pixel in dewarped image.

int
tlCameraTSAI::warp (tlImage *image,
                    tlRect  *rect)
{
  int     i,j,k;
  int     index, dindex;
  int     width, height, imwidth;
  int     depth;
  int     byteCount;
  double  dx,dy;
  tlPixel *unwarped;
  tlPixel *warped;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if ((!image) || (!rect)) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcut
  unwarped   = image->pixel;
  depth      = image->pixelWidth;
  width      = rect->width;
  imwidth    = image->width;
  height     = rect->height;
  byteCount  = (image->width)*(image->height)*depth;

  // create temporary buffer
  warped = (tlPixel *) malloc(byteCount*sizeof(tlPixel));
  if (warped == NULL) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // dewarp
  for (j=(rect->y); j<height; j++) {
    for (i=(rect->x); i<width; i++) {
      undistorted_to_distorted_image_coord (i,j,&dx, &dy);
      dindex = depth*((int)dy*imwidth+(int)dx);
      index  = depth*(j*imwidth+i);
      if ((dindex < 0) || (dindex >= byteCount)) 
        for (k=0; k<depth; k++)
          warped[dindex++] = unwarped[index++];
    }
  }

  // store result
  free (image->pixel);
  image->pixel = unwarped;


  return 0;
}



///  Warp an image for the given camera. That is, return the 
///  distorted image given the calibration settings.
///  The result is placed in another image.
///
///  \param image      the image to project
///  \param dest       \b [out] projection result 
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  warp(tlImage *image)\n
///  warp(tlImage *image, tlRect *rect)\n
///  warp(tlImage *image, tlRect *rect, tlImage *dest)
///
///  \todo
///  provide an option to use tl_bspline_interpolate() to retrieve best value
///  for real x,y source pixel in dewarped image.

int
tlCameraTSAI::warp (tlImage *image,
                    tlImage *dest)
{  
  int     i,j,k;
  int     index, dindex;
  int     width, height;
  int     depth;
  int     byteCount;
  double  dx,dy;
  tlPixel *warped;
  tlPixel *unwarped;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NA);
    image->copyTo (dest);
    return -1;
  }

  // parameter check
  if ((!image) || (!dest)) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // adjust destination image
  dest->init (image->format, image->width, image->height);

  // shortcut
  unwarped   = image->pixel;
  warped     = dest->pixel;
  depth      = image->pixelWidth;
  width      = image->width;
  height     = image->height;
  byteCount  = width*height*depth;

  // dewarp
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      undistorted_to_distorted_image_coord (i,j,&dx, &dy);
      dindex = depth*((int)dy*width+(int)dx);
      index  = depth*(j*width+i);
      if ((dindex < 0) || (dindex >= byteCount)) 
        for (k=0; k<depth; k++)
          warped[dindex++] = unwarped[index++];
    }
  }


  return 0;
}



///  Warp a portion of an image for the given camera. That is, 
///  return the distorted image given the calibration settings.
///  The result is placed in another image.
///
///  \param image      the image to project
///  \param rect       portion of the image to project
///  \param dest       \b [out] projection result
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  warp(tlImage *image)\n
///  warp(tlImage *image, tlImage *dest)\n
///  warp(tlImage *image, tlImage *dest)
///
///  \todo
///  provide an option to use tl_bspline_interpolate() to retrieve best value
///  for real x,y source pixel in dewarped image.

int
tlCameraTSAI::warp (tlImage *image,
                    tlRect  *rect,
                    tlImage *dest)
{
  int     i,j,k;
  int     index, dindex;
  int     width, height, imwidth;
  int     depth;
  int     byteCount;
  double  dx,dy;
  tlPixel *warped;
  tlPixel *unwarped;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NA);
    image->copyTo (dest);
    return -1;
  }

  // parameter check
  if ((!image) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlCameraTSAI::warp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // adjust destination image
  dest->init (image->format, image->width, image->height);

  // shortcut
  unwarped   = image->pixel;
  warped     = dest->pixel;
  depth      = image->pixelWidth;
  width      = rect->width;
  imwidth    = image->width;
  height     = rect->height;
  byteCount  = (image->width)*(image->height)*depth;

  // dewarp
  for (j=(rect->y); j<height; j++) {
    for (i=(rect->x); i<width; i++) {
      undistorted_to_distorted_image_coord (i,j,&dx, &dy);
      dindex = depth*((int)dy*imwidth+(int)dx);
      index  = depth*(j*imwidth+i);
      if ((dindex < 0) || (dindex >= byteCount)) 
        for (k=0; k<depth; k++)
          warped[dindex++] = unwarped[index++];
    }
  }


  return 0;
}



///  Unwarp an image for the given camera. That is, return the 
///  undistorted image given the calibration settings.
///
///  \param image       the image to project
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  the image is altered by this call. 
///  See dewarp(tlImage *image, tlImage *dest) to put the
///  dewarping result in another tlImage object.
///
///  \sa
///  dewarp(tlImage *image, tlRect *rect)\n
///  dewarp(tlImage *image, tlImage *dest)\n
///  dewarp(tlImage *image, tlRect *rect, tlImage *dest)
///
///  \todo
///  provide an option to use tl_bspline_interpolate() to retrieve best value
///  for real x,y source pixel in warped image.

int
tlCameraTSAI::dewarp (tlImage *image)
{
  int     i,j,k;
  int     index, dindex;
  int     width, height;
  int     depth;
  int     byteCount;
  double  dx,dy;
  tlPixel *warped;
  tlPixel *unwarped;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if (!image) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcut
  warped     = image->pixel;
  depth      = image->pixelWidth;
  width      = image->width;
  height     = image->height;
  byteCount  = width*height*depth;

  // create temporary buffer
  unwarped = tl_create_array (width, height, depth);
  if (unwarped == NULL) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // dewarp
  index = 0;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      undistorted_to_distorted_image_coord (i,j,&dx, &dy);
      if (((int)dx >= width) || ((int)dx < 0) || ((int)dy >= height) || ((int)dy < 0)) continue;
      dindex = depth*((int)dy*width+(int)dx);
      for (k=0; k<depth; k++) {
        unwarped[index++] = warped[dindex++];
      }
    }
  }

  // store result
  free (image->pixel);
  image->pixel = unwarped;


  return 0;
}



///  Unwarp a portion of an image for the given camera. That is, 
///  return the undistorted image given the calibration settings.
///
///  \param image      the image to project
///  \param rect       portion of the image to project
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  the image is altered by this call. 
///  See dewarp(tlImage *image, tlImage *dest) to put the
///  dewarping result in another tlImage object.
///
///  \sa
///  dewarp(tlImage *image)\n
///  dewarp(tlImage *image, tlImage *dest)\n
///  dewarp(tlImage *image, tlRect *rect, tlImage *dest)
///
///  \todo
///  provide an option to use tl_bspline_interpolate() to retrieve best value
///  for real x,y source pixel in warped image.

int
tlCameraTSAI::dewarp (tlImage *image,
                      tlRect  *rect)
{
  int     i,j,k;
  int     index, uindex;
  int     width, height, imwidth;
  int     depth;
  int     byteCount;
  double  ux,uy;
  tlPixel *warped;
  tlPixel *unwarped;


  // make sure...
  if (!calibrated) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if ((!image) || (!rect)) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcut
  warped     = image->pixel;
  depth      = image->pixelWidth;
  width      = rect->width;
  imwidth    = image->width;
  height     = rect->height;
  byteCount  = (image->width)*(image->height)*depth;

  // create temporary buffer
  unwarped = (tlPixel *) malloc(byteCount*sizeof(tlPixel));
  if (unwarped == NULL) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // dewarp
  for (j=(rect->y); j<height; j++) {
    for (i=(rect->x); i<width; i++) {
      distorted_to_undistorted_image_coord (i,j,&ux, &uy);
      uindex = depth*((int)uy*imwidth+(int)ux);
      index  = depth*(j*imwidth+i);
      if ((uindex < 0) || (uindex >= byteCount)) 
        for (k=0; k<depth; k++)
          unwarped[uindex++] = 0;      
      else
        for (k=0; k<depth; k++)
          unwarped[uindex++] = warped[index++];
    }
  }

  // store result
  free (image->pixel);
  image->pixel = unwarped;


  return 0;
}



///  Unwarp an image for the given camera. That is, return the 
///  undistorted image given the calibration settings.
///  The result is placed in another image.
///
///  \param image      the image to project
///  \param dest       \b [out] projection result 
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  dewarp(tlImage *image)\n
///  dewarp(tlImage *image, tlRect *rect)\n
///  dewarp(tlImage *image, tlRect *rect, tlImage *dest)
///
///  \todo
///  provide an option to use tl_bspline_interpolate() to retrieve best value
///  for real x,y source pixel in source warped image.

int
tlCameraTSAI::dewarp (tlImage *image,
                      tlImage *dest)
{  
  int     i,j,k;
  int     index, uindex;
  int     width, height;
  int     depth;
  int     byteCount;
  double  ux,uy;
  tlPixel *warped;
  tlPixel *unwarped;


  // make sure...
  if (!calibrated) {
    image->copyTo (dest);
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if ((!image) || (!dest)) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // adjust destination image
  dest->init (image->format, image->width, image->height);

  // shortcut
  warped     = image->pixel;
  unwarped   = dest->pixel;
  depth      = image->pixelWidth;
  width      = image->width;
  height     = image->height;
  byteCount  = width*height*depth;

  // dewarp
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      distorted_to_undistorted_image_coord (i,j,&ux, &uy);
      uindex = depth*((int)uy*width+(int)ux);
      index  = depth*(j*width+i);
      if ((uindex < 0) || (uindex >= byteCount)) 
        for (k=0; k<depth; k++)
          unwarped[uindex++] = warped[index++];
    }
  }


  return 0;
}



///  Unwarp a portion of an image for the given camera. That is, 
///  return the undistorted image given the calibration settings.
///  The result is placed in another image.
///
///  \param image      the image to project
///  \param rect       portion of the image to project
///  \param dest       \b [out] projection result
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  dewarp(tlImage *image)\n
///  dewarp(tlImage *image, tlImage *dest)\n
///  dewarp(tlImage *image, tlImage *dest)
///
///  \todo
///  provide an option to use tl_bspline_interpolate() to retrieve best value
///  for real x,y source pixel in source warped image.

int
tlCameraTSAI::dewarp (tlImage *image,
                      tlRect  *rect,
                      tlImage *dest)
{
  int     i,j,k;
  int     index, uindex;
  int     width, height, imwidth;
  int     depth;
  int     byteCount;
  double  ux,uy;
  tlPixel *warped;
  tlPixel *unwarped;


  // make sure...
  if (!calibrated) {
    image->copyTo (dest);
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NA);
    return -1;
  }

  // parameter check
  if ((!image) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlCameraTSAI::dewarp()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // adjust destination image
  dest->init (image->format, image->width, image->height);

  // shortcut
  warped     = image->pixel;
  unwarped   = dest->pixel;
  depth      = image->pixelWidth;
  width      = rect->width;
  imwidth    = image->width;
  height     = rect->height;
  byteCount  = (image->width)*(image->height)*depth;

  // dewarp
  for (j=(rect->y); j<height; j++) {
    for (i=(rect->x); i<width; i++) {
      distorted_to_undistorted_image_coord (i,j,&ux, &uy);
      uindex = depth*((int)uy*imwidth+(int)ux);
      index  = depth*(j*imwidth+i);
      if ((uindex < 0) || (uindex >= byteCount)) 
        for (k=0; k<depth; k++)
          unwarped[uindex++] = warped[index++];
    }
  }


  return 0;
}



////////////////////////////////////////////////////////////////////////

int 
tlCameraTSAI::grab (tlImage *image)
{
  // acquire image
  if (source) return source->grab (image);
  else {
    TL_ERRNO_MSG ("tlCameraTSAI::grab()", TL_ERROR_NULL_POINTER);
    return -1;
  }
}



////////////////////////////////////////////////////////////////////////

int 
tlCameraTSAI::grab (tlImage *image, 
                    tlRect  *rect)
{
  // acquire image
  if (source) return source->grab (image, rect);
  else {
    TL_ERRNO_MSG ("tlCameraTSAI::grab()", TL_ERROR_NULL_POINTER);
    return -1;
  }
}



////////////////////////////////////////////////////////////////////////

int 
tlCameraTSAI::grabQueued (tlImage *image)
{
  // acquire image
  if (source) return source->grabQueued (image);
  else {
    TL_ERRNO_MSG ("tlCameraTSAI::grab()", TL_ERROR_NULL_POINTER);
    return -1;
  }
}


////////////////////////////////////////////////////////////////////////

int 
tlCameraTSAI::grabQueued (tlImage *image, 
                          tlRect  *rect)
{
  // acquire image
  if (source) return source->grabQueued (image, rect);
  else {
    TL_ERRNO_MSG ("tlCameraTSAI::grab()", TL_ERROR_NULL_POINTER);
    return -1;
  }
}
