/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#ifdef LINUX
#include <stdlib.h>
#endif

#include "tlChain.h"
#include "tlPoint.h"
#include "tlImage.h"
#include "tlRect.h"
#include "tlMask.h"



///  \file tlChain.h
///  \brief Dynamic point list management.

///  \class tlChain
///  \brief Point list management class.
///
///  A tlChain object is an arbitrary, dynamic collection of tlPoints. tlChain objects are useful for
///  storing extracted points, but also for continuous edges extraction and other related 
///  operations such as angle detections.
///
///  \sa
///  \htmlonly <a href="http://visual.ipan.sztaki.hu/corner/cornerweb.html">http://visual.ipan.sztaki.hu/corner/cornerweb.html</a> \endhtmlonly



///  \def TL_CHAIN_CORNER_DMIN
///  Smallest test triangle size for the default corner extraction.
///  See \htmlonly <a href="http://visual.ipan.sztaki.hu/corner/cornerweb.html">http://visual.ipan.sztaki.hu/corner/cornerweb.html</a> \endhtmlonly for details.

///  \def TL_CHAIN_CORNER_DMAX
///  Largest test triangle size for the default corner extraction.
///  See \htmlonly <a href="http://visual.ipan.sztaki.hu/corner/cornerweb.html">http://visual.ipan.sztaki.hu/corner/cornerweb.html</a> \endhtmlonly for details.

///  \def TL_CHAIN_CORNER_ALPHAMAX
///  Threshold angle for the default corner extraction.
///  See \htmlonly <a href="http://visual.ipan.sztaki.hu/corner/cornerweb.html">http://visual.ipan.sztaki.hu/corner/cornerweb.html</a> \endhtmlonly for details.



///  \fn tlPoint *tlChain::getStart()
///  Retrieve the first point of the chain.
///
///  \return
///  A pointer to the \c start point on success, NULL if the chain is empty.\n
///  See \ref errors "error management" for details.



///  \fn tlPoint *tlChain::getEnd()
///  Retrieve the last point of the chain.
///
///  \return
///  A pointer to the \c end point on success, NULL if the chain is empty.\n
///  See \ref errors "error management" for details.



///  \fn int tlChain::getLength()
///  Get the chain length.
///
///  \return
///  The chain length on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlChain::isClosed()
///  Test whether the chain is closed or not.
///
///  \return
///  1 if the chain is closed,\n
///  0 otherwise.



///  \fn int tlChain::getConnex()
///  Test whether the chain connectivity.
///
///  \return
///  The chain connectivity (TL_CONNEX_4 or TL_CONNEX_8).



///  Constructor, creates an empty tlChain object.
///
///  \note
///  See \ref errors "error management" for details.

tlChain::tlChain ()
{
	this->length = 0;
	this->closed = 0;
	this->connex = TL_NOT_CONNEX;
	this->start  = NULL;
	this->end    = NULL;

	tl_errno = TL_NO_ERROR;
}




///  Constructor, creates a tlChain with one starting point.
///
///  \param startPoint  starting point for chain extraction
///
///  \note
///  See \ref errors "error management" for details.

tlChain::tlChain (tlPoint *startPoint)
{
	this->length = 1;
	this->closed = 0;
	this->connex = TL_NOT_CONNEX;
	this->add (startPoint);

	tl_errno = TL_NO_ERROR;
}



///  Constructor, creates a tlChain with one starting point.
///
///  \param startx           starting point X coord
///  \param starty           starting point Y coord
///
///  \note
///  See \ref errors "error management" for details.

tlChain::tlChain (int startx, 
									int starty)
{
	this->length = 1;
	this->closed = 0;
	this->connex = TL_NOT_CONNEX;
	this->add (startx, starty);

	tl_errno = TL_NO_ERROR;
}



///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlChain::~tlChain ()
{
	struct lpoint *cur;
	struct lpoint *rem;


	// play it safe
	cur = this->start;
	while ((cur != NULL) && (cur != (this->end))) {
		rem = cur;
		cur = cur->next;
		delete (cur->point);
		free (rem);
	}

	// remove last point
	cur = this->end;
	if (cur != NULL) free (cur);

	tl_errno = TL_NO_ERROR;
}



///  Create a point to be inserted in the chained list.
///
///  \param point     the tlPoint data
///    
///  \return
///  A pointer to the new lpoint on success, NULL on failure.\n
///  See \ref errors "error management" for details.

struct lpoint *
tlChain::newPoint (tlPoint *p)
{
	struct lpoint *newp;
	
	
	newp = (struct lpoint *) malloc (sizeof (struct lpoint));
	if (!newp){
		TL_ERRNO_MSG ("tlChain::newPoint()", TL_ERROR_ALLOCATION_FAILED);
		return NULL;
	}

	newp->point = new tlPoint (p);
	newp->index = 0;
	newp->next  = NULL;
	newp->prev  = NULL;


	return newp;
}



///  \overload

struct lpoint *
tlChain::newPoint (int x, 
		   int y)
{
	struct lpoint *newp;


	newp = (struct lpoint *) malloc (sizeof (struct lpoint));
	if (!newp) {
		TL_ERRNO_MSG ("tlChain::newPoint()", TL_ERROR_ALLOCATION_FAILED);
		return NULL;
	}
	
  newp->point = new tlPoint (x, y);
	newp->index = 0;
	newp->next = NULL;
	newp->prev = NULL;


	return newp;
}



///  Recursive call for chain extraction.
///
///  \param data        binary image pixmap
///  \param width       image width
///  \param height      image height
///  \param x           point to process X coord
///  \param y           point to process Y coord
///  \param connex      connectivity (default is TL_CONNEX_8)
///
///  \return
///  0 in all cases.

int
tlChain::extractPoint (_tl_extraction_data *exdat)
{
	// geometric test
	if (exdat->x >= exdat->x2) return 0;
	if (exdat->y >= exdat->y2) return 0;
	if (exdat->x <  exdat->x1) return 0;
	if (exdat->y <  exdat->y1) return 0;
	
	// if point is invalid, or if we already have it
	if (exdat->data[TL_PIXEL (exdat->y, exdat->x, exdat->width)] == 0) return 0;
	if (this->isIn (exdat->x, exdat->y) > -1) return 0;
	
	// otherwise, add & recurse !
	this->add (exdat->x, exdat->y);
	exdat->x -= 1;
  extractPoint (exdat);
  exdat->x += 1;
  exdat->y -= 1;
  extractPoint (exdat);
  exdat->y += 2;
  extractPoint (exdat);
  exdat->x += 1;
  exdat->y -= 1;
  extractPoint (exdat);
  exdat->x -= 1;
	if (exdat->connex == TL_CONNEX_8) {
    exdat->x -= 1;
    exdat->y -= 1;
    extractPoint (exdat);
    exdat->y += 2;
    extractPoint (exdat);
    exdat->x += 2;
    exdat->y -= 2;
    extractPoint (exdat);
    exdat->y += 2;
    extractPoint (exdat);
    exdat->x -= 1;
    exdat->y -= 1;
	}
	
  
  return 0;
}



///  Get the point at a given index in the chain (if it exists).
///
///  \param index       point index
///    
///  \return
///  A pointer to a new tlPoint on success, NULL on failure.\n
///  See \ref errors "error management" for details.

tlPoint *
tlChain::getPoint (int index)
{
	int           i;
	struct lpoint *cur;


	// check
	if (index >= (this->length)) {
		TL_ERRNO_MSG ("tlChain::getPoint()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
		return NULL;
	}

	// retrieve
	cur = this->start;
	for (i = 0; i < index; i++) cur = cur->next;
	
	
	return new tlPoint (cur->point);
}



///  Reset the chain (remove all points), same as clear().
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::reset ()
{
	return this->clear ();
}



///  Reset the chain (remove all points), same as reset().
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::clear ()
{
	struct lpoint *cur;
	struct lpoint *rem;


	// play it safe
	cur = this->start;
	while ((cur != NULL) && (cur != (this->end))) {
		rem = cur;
		cur = cur->next;
		free (rem);
	}

	// remove last point
	cur = this->end;
	if (cur != NULL)
		free (cur);

	// reset params
	this->closed = 0;
	this->length = 0;
	this->connex = TL_NOT_CONNEX;
	this->start  = NULL;
	this->end    = NULL;


	return 0;
}



///  Remove close neighbors to avoid redundancies.
///
///  \param radius   the minimum allowed distance between 2 points
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::filter (int radius)
{
	int           i,j;
	struct lpoint *ref;
	struct lpoint *cmp;
	struct lpoint *tmp;
	int           count = 0;
	int           dist;


	ref = this->start;
	for (i = 0; i < this->length; i++) {
		cmp = this->start;
		for (j = 0; j < this->length; j++) {
			if (cmp == ref) {
				cmp = cmp->next;
				continue;
			}
			dist = cmp->point->dist (ref->point);
			if (dist < radius) {
				tmp = cmp;
				cmp = cmp->next;
				if (cmp == NULL)
					continue;
				this->remove (tmp->index);
				count++;
			}
			else cmp = cmp->next;
		}
		ref = ref->next;
	}


	return 0;
}



///  Close the chain (requires at least 2 points to be in the chain).
///    
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::close ()
{
	// check
	if ((this->length) < 2) {
		TL_ERRNO_MSG ("tlChain::close()", TL_ERROR);
		return -1;
	}

  // double-check
	if ((this->closed)) return 0;

	// close
	start->prev = end;
	end->next   = start;


	return 0;
}



///  Test for the presence of a point in the chain.
///    
///  \param point     point to test for
///
///  \return
///  The index of the point if it is in the chain, -1 otherwise.\n
///  See \ref errors "error management" for details.

inline int
tlChain::isIn (tlPoint *point)
{
	if (!point) {
		TL_ERRNO_MSG ("tlChain::isIn()", TL_ERROR_NULL_POINTER);
		return -1;
	}


	return this->isIn (point->x, point->y);
}



///  \overload

inline int
tlChain::isIn (int x, 
							 int y)
{
	int		        i;
	struct lpoint *cur;


	// check
	if (this->start == NULL) {
		return -1;
	}
	// locate point index
	cur = this->start;
	for (i = 0; i < (this->length); i++) {
		if (((cur->point->x) == x) && ((cur->point->y) == y)) return (cur->index);
		cur = cur->next;
	}


	return -1;
}



///  Insert a point at the end of the chain.
///    
///  \param x         the point to add X coord
///  \param y         the point to add Y coord
///
///  \return
///  The index of the point if insertion is successfull, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::add (int x, 
							int y)
{
	// first point case
	if (this->length == 0) {
		this->start = newPoint (x, y);
		if (!this->start) {
			TL_ERRNO_MSG ("tlChain::add()", TL_ERROR_ALLOCATION_FAILED);
			return -1;
		}
		this->start->index = 0;
		this->start->next = NULL;
		this->start->prev = NULL;
		this->end = this->start;
	}

	// otherwise
	else {
		this->end->next = newPoint (x, y);
		if (!this->end->next) {
			TL_ERRNO_MSG ("tlChain::add()", TL_ERROR_ALLOCATION_FAILED);
			return -1;
		}
		this->end->next->prev = this->end;
		this->end = this->end->next;
		this->end->index = this->length;
		this->end->next = NULL;
	}
	this->length += 1;

  
  return ((this->length) - 1);
}



///  \overload

int
tlChain::add (tlPoint *point)
{
	// check
	if (!point) {
		TL_ERRNO_MSG ("tlChain::add()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  
  return this->add (point->x, point->y);
}



///  Insert a point at a specific index in the chain.
///    
///  \param index     the index to insert to point at
///  \param x         the point to add X coord
///  \param y         the point to add Y coord
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::insert (int index, 
								 int x, 
								 int y)
{
	int	          i       = 0;
	struct lpoint *cur;
	struct lpoint *newp;


	if (index >= (this->length)) {
		TL_ERRNO_MSG ("tlChain->insert()", TL_ERROR);
		return -1;
	}

	// find insert point
	cur = this->start;
	while (i < index) {
		cur = cur->next;
		i++;
	}

	// create new point
	newp = newPoint (x, y);
	if (!newp) {
		TL_ERRNO_MSG ("tlChain->insert()", TL_ERROR_ALLOCATION_FAILED);
		return -1;
	}
	
  // insert new point
	if ((cur->prev) != NULL)
		cur->prev->next = newp;
	newp->index = index;
	newp->next = cur;
	newp->prev = cur->prev;
	cur->prev = newp;

	// boundary
	if (index == 0)
		this->start = newp;

	// update following points index
	this->length += 1;
	i = index + 1;
	while (i < (this->length)) {
		cur->index += 1;
		cur = cur->next;
		i++;
	}
	
  
  return 0;
}



///  \overload

int
tlChain::insert (int     index, 
								 tlPoint *point)
{
	// check
	if (!point) {
		TL_ERRNO_MSG ("tlChain->insert()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  
  return this->insert (index, point->x, point->y);
}



///  Remove a point at a specific index from the chain.
///    
///  \param index     the index of the point to remove
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::remove (int index)
{
	int	          i     = 0;
	struct lpoint *cur;
	struct lpoint *tmp;


	// check
	if (index >= (this->length)) {
		TL_ERRNO_MSG ("tlChain->remove()", TL_ERROR);
		return -1;
	}

	// start boundary 
	if (index == 0) {
		cur = this->start;
		this->start = cur->next;
		this->start->prev = cur->prev;
	}
	
  // end boundary
	else if (index == ((this->length) - 1)) {
		cur = this->end;
		this->end = cur->prev;
		this->end->next = cur->next;
	}
	
  // any other case
	else {
		// retrieve point
		cur = this->start;
		while (i < index) {
			cur = cur->next;
			i++;
		}

    // redirect pointers
		cur->prev->next = cur->next;
		cur->next->prev = cur->prev;
	}

	// remove point
	tmp = cur;
	cur = cur->next;
	free (tmp);

	// restore index 
	this->length -= 1;
	i = index;
	while (i < (this->length)) {
		cur->index -= 1;
		cur = cur->next;
		i++;
	}

  
  return 0;
}



///  Remove a point from the chain if it exists.
///    
///  \param x       point to remove X coord
///  \param y       point to remove Y coord
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::remove (int x, 
								 int y)
{
	int	          i;
	struct lpoint *cur;


	// check
	if (this->start == NULL) {
		TL_ERRNO_MSG ("tlChain->remove()", TL_ERROR);
		return -1;
	}

	// locate point index
	cur = this->start;
	for (i = 0; i < (this->length); i++) {
		if (((cur->point->x) == x) && ((cur->point->y) == y))
			return this->remove (cur->index);
		cur = cur->next;
	}
	
  
  return -1;
}



///  \overload

int
tlChain::remove (tlPoint *point)
{
	// check
	if (!point) {
		TL_ERRNO_MSG ("tlChain->remove()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  
  return this->remove (point->x, point->y);
}



///  \overload

int
tlChain::extract (tlImage *image, 
									tlPoint *point, 
									int     connex)
{
	if ((!point) || (!image)) {
		TL_ERRNO_MSG ("tlChain->extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	return extract (image, point->x, point->y, connex);
}


///  \overload

int
tlChain::extract (tlImage *image, 
									int     x, 
									int     y, 
									int     connex)
{
	int    res;
	// tlRect *rect;


	if (!image) {
		TL_ERRNO_MSG ("tlChain->extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	res  = extract (image, x, y, (tlRect*)image, connex);


	return res;
}



///  \overload

int
tlChain::extract (tlImage *image, 
									tlPoint *point, 
									tlRect  *rect,
									int     connex)
{
	if ((!point) || (!image)) {
		TL_ERRNO_MSG ("tlChain->extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}


	return extract (image, point->x, point->y, rect, connex);
}



///  Extract a chain from a binary image.
///    
///  \param image     binary image
///  \param x         X coord to start extraction from
///  \param y         Y coord to start extraction from
///  \param connex    connectivity (TL_CONNECT_4 or TL_CONNECT_8)
///  \param rect      bounding rectangle for extraction
///
///  \return
///  the length of the extracted chain if successfull,\n
///  0 if the starting point is not valid (0),\n
///  -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::extract (tlImage *image, 
                  int     x, 
                  int     y, 
                  tlRect  *rect, 
                  int     connex)
{
	int	                width, height;
	int	                x1, x2, y1, y2;
	int	                xs, ys, xe, ye;
	tlPixel             *data;
	_tl_extraction_data exdat;
	

	// check
	if (!image) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
	// check format
	if ((image->format) != BIN) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

	// shortcuts
	width  = image->width;
	height = image->height;
	data   = image->pixel;
	x1     = TL_MAX ((rect->x), 0);
	x2     = TL_MIN (((rect->x) + (rect->width)), width);
	y1     = TL_MAX ((rect->y), 0);
	y2     = TL_MIN (((rect->y) + (rect->height)), height);
	
	// check geometry
	if ((x >= x2) || (y >= y2) || (x < x1) || (y < y1)) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR_GEOMETRY);
		return -1;
	}

	// check that the starting point is valid
	if (data[TL_PIXEL (y, x, width)] == 0) {
		return 0;
	}

	// reset
	this->reset ();
	
	// process
  exdat.data    = data;
  exdat.width   = width;
  exdat.height  = height;
  exdat.x       = x;
  exdat.y       = y;
  exdat.x1      = x1;
  exdat.x2      = x2;
  exdat.y1      = y1;
  exdat.y2      = y2;
  exdat.connex  = connex;
	extractPoint (&exdat);
	
	// detect simple closed chains
	if (this->length > 1) {
		xs = this->start->point->x;
		ys = this->start->point->y;
		xe = this->end->point->x;
		ye = this->end->point->y;
		if (connex == TL_CONNEX_8) {
			if (((xe - 1) <= xs) && (xs <= (xe + 1)) && ((ye - 1) <= ys) && (ys <= (ye + 1))) {
				this->close ();
			}
		}
		else {
			if ((xs == (xe - 1)) || (xs == (xe + 1))) {
			  if (ys == ye) this->close ();
			}		
			else if (xs == xe) {
				if (((ye - 1) <= ys) && (ys <= (ye + 1))) this->close ();
			}
		}
	}


	// store connectivity
	this->connex = connex;
	return (this->length);
}



///  \overload

int
tlChain::blindExtract (tlImage *image, 
											 tlPoint *point, 
											 int     connex)
{
	return this->blindExtract (image, point->x, point->y, connex);
}



///  \overload

int
tlChain::blindExtract (tlImage *image, 
											 int     x, 
											 int     y, 
											 int     connex)
{
	int	   res;
	tlRect *rect;


	if (!image) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  rect = new tlRect (0, 0, image->width, image->height);
	res = this->blindExtract (image, x, y, rect, connex);
	delete rect;


	return res;
}



///  \overload

int
tlChain::blindExtract (tlImage *image, 
											 tlPoint *point, 
											 tlRect  *rect,
											 int     connex)
{
	if ((!point) || (!image)) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR_NULL_POINTER);
		return -1;
	}


	return this->blindExtract (image, point->x, point->y, rect, connex);
}



///  Extract a chain from a binary image. 
///  Looks around a given point for a valid starting point.
///    
///  \param image     binary image
///  \param x         starting point X coord
///  \param y         starting point Y coord
///  \param rect      bounding rectangle to limit extraction area
///  \param connex    connectivity (TL_CONNECT_4 or TL_CONNECT_8)
///
///  \return
///  The length of the extracted chain on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The starting point can be anywhere within the rectangle.
///  The search algorithm first searches on a vertical and horizontal
///  line for the presence of a valid starting point. If none, it then 
///  spirals around (\c x,\c y) until a valid point is found.

int
tlChain::blindExtract (tlImage *image, 
											 int     x, 
											 int     y, 
											 tlRect  *rect,
											 int     connex)
{
	int	    i, j;
	int	    i_act, j_act;
	int	    i_incr, j_incr, cur_incr;
	int	    x1, x2, y1, y2;
	int	    width, height;
	char    l_flag, r_flag, t_flag, b_flag;
	tlPixel *data;


	// check
	if (!image) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  // check format
	if ((image->format) != BIN) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
	
  // shortcuts
	width  = image->width;
	height = image->height;
	data   = image->pixel;
	x1     = TL_MAX ((rect->x), 0);
	x2     = TL_MIN (((rect->x) + (rect->width)), width);
	y1     = TL_MAX ((rect->y), 0);
	y2     = TL_MIN (((rect->y) + (rect->height)), height);

	// check geometry
	if ((x >= x2) || (y >= y2) || (x < x1) || (y < y1)) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR);
		return -1;
	}

	// reset
	this->reset ();

	// fast loop - try direct vertical and horizontal
	i = x;
	while (i >= x1) {
		if (data[TL_PIXEL (y, i, width)] != 0)
			return this->extract (image, i, y, rect, connex);
		i--;
	}
	i = x;
	while (i < x2) {
		if (data[TL_PIXEL (y, i, width)] != 0)
			return this->extract (image, i, y, rect, connex);
		i++;
	}
	j = y;
	while (j >= y1) {
		if (data[TL_PIXEL (j, x, width)] != 0)
			return this->extract (image, x, j, rect, connex);
		j--;
	}
	j = y;
	while (j < y2) {
		if (data[TL_PIXEL (j, x, width)] != 0)
			return this->extract (image, x, j, rect, connex);
		j++;
	}

	// otherwise, go for the long way
	i = x;
	j = y;
	l_flag = r_flag = t_flag = b_flag = 0;
	cur_incr = i_incr = j_incr = 0;
	while ((!l_flag) || (!r_flag) || (!t_flag) || (!b_flag)) {
		// check for validity
		i_act = TL_MAX (i, x1);
		i_act = TL_MIN (i_act, x2 - 1);
		j_act = TL_MAX (j, y1);
		j_act = TL_MIN (j_act, y2 - 1);
		if (data[TL_PIXEL (j_act, i_act, width)] != 0)
			return this->extract (image, i_act, j_act, rect, connex);

		// update vars (safe spiral)
		if (i_incr < abs (cur_incr)) {
			if (cur_incr > 0) i++;
			else              i--;
			i_incr++;
			if (i < x1) {
				l_flag = 1;
			}
			else if (i >= x2) {
				r_flag = 1;
			}
		}
		else if (j_incr < abs (cur_incr)) {
			if (cur_incr > 0) j++;
			else              j--;
			j_incr++;
			if (j < y1) {
				t_flag = 1;
			}
			else if (j >= y2) {
				b_flag = 1;
			}
		}
		else {
			if (cur_incr < 0) cur_incr = (-1) * (cur_incr - 1);
			else              cur_incr = (-1) * (cur_incr + 1);
			i_incr = j_incr = 0;
		}
	}


	return -1;
}



///  \overload

int
tlChain::extract (tlMask  *mask, 
									tlPoint *point, 
									int     connex)
{
	if ((!point) || (!mask)) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}


	return this->extract (mask, point->x, point->y, connex);
}



///  \overload

int
tlChain::extract (tlMask *mask, 
									int    x, 
									int    y, 
									int    connex)
{
	int	   res;
	tlRect *rect;


	if (!mask) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	rect = new tlRect (0, 0, mask->width, mask->height);
	res = this->extract (mask, x, y, rect, connex);
	delete rect;


	return res;
}



///  \overload

int
tlChain::extract (tlMask  *mask, 
									tlPoint *point, 
									tlRect  *rect, 
									int     connex)
{
	if ((!point) || (!mask)) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}


	return this->extract (mask, point->x, point->y, rect, connex);
}



///  Extract a chain from a mask.
///    
///  \param mask      mask to extract from
///  \param x         X coord to start extraction from
///  \param y         Y coord to start extraction from
///  \param connex    connectivity (TL_CONNECT_4 or TL_CONNECT_8)
///  \param rect      bounding rectangle for extraction
///
///  \return
///  The length of the extracted chain if successfull,\n
///  0 if the starting point is not valid (0),\n
///  -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::extract (tlMask *mask, 
									int    x, 
									int    y, 
									tlRect *rect, 
									int    connex)
{
	int	                width, height;
	int	                x1, x2, y1, y2;
	int	                xs, ys, xe, ye;
	tlPixel             *data;
  _tl_extraction_data exdat;


	// check
	if (!mask) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// shortcuts
	width  = mask->width;
	height = mask->height;
	data   = mask->pixel;
	x1     = TL_MAX ((rect->x), 0);
	x2     = TL_MIN (((rect->x) + (rect->width)), width);
	y1     = TL_MAX ((rect->y), 0);
	y2     = TL_MIN (((rect->y) + (rect->height)), height);

	// check geometry
	if ((x >= x2) || (y >= y2) || (x < x1) || (y < y1)) {
		TL_ERRNO_MSG ("tlChain::extract()", TL_ERROR);
		return -1;
	}

	// check that the starting point is valid
	if (data[TL_PIXEL (y, x, width)] == 0) {
		return 0;
	}
	
  // reset
	this->reset ();

	// process
  exdat.data    = data;
  exdat.width   = width;
  exdat.height  = height;
  exdat.x       = x;
  exdat.y       = y;
  exdat.x1      = x1;
  exdat.x2      = x2;
  exdat.y1      = y1;
  exdat.y2      = y2;
  exdat.connex  = connex;
	extractPoint (&exdat);

	// detect simple closed chains
	if (this->length > 1) {
		xs = this->start->point->x;
		ys = this->start->point->y;
		xe = this->end->point->x;
		ye = this->end->point->y;
		if (connex == TL_CONNEX_8) {
			if (((xe - 1) <= xs) && (xs <= (xe + 1)) && ((ye - 1) <= ys) && (ys <= (ye + 1))) {
					this->close ();
				}
		}
		else {
			if ((xs == (xe - 1)) || (xs == (xe + 1))) {
				if (ys == ye) this->close ();
			}
			else if (xs == xe) {
				if (((ye - 1) <= ys) && (ys <= (ye + 1))) this->close ();
			}
		}
	}


	// store connectivity
	this->connex = connex;
	return (this->length);
}



///  \overload

int
tlChain::blindExtract (tlMask  *mask, 
											 tlPoint *point, 
											 int     connex)
{
	return this->blindExtract (mask, point->x, point->y, connex);
}



///  \overload

int
tlChain::blindExtract (tlMask *mask, 
											 int    x, 
											 int    y, 
											 int    connex)
{
	int	   res;
	tlRect *rect;


	if (!mask) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  rect = new tlRect (0, 0, mask->width, mask->height);
	res  = this->blindExtract (mask, x, y, rect, connex);
	delete rect;


	return res;
}



///  \overload

int
tlChain::blindExtract (tlMask  *mask, 
											 tlPoint *point, 
											 tlRect  *rect,
											 int     connex)
{
	if ((!point) || (!mask)) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  
  return this->blindExtract (mask, point->x, point->y, rect, connex);
}



///  Extract a chain from a mask.
///  Looks around a given point for a valid starting point.
///    
///  \param mask      mask to extract from
///  \param x         starting point X coord
///  \param y         starting point Y coord
///  \param rect      bounding rectangle to limit extraction area
///  \param connex    connectivity (TL_CONNECT_4 or TL_CONNECT_8)
///
///  \return
///  The length of the extracted chain on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The starting point can be anywhere within the rectangle.
///  The search algorithm first searches on a vertical and horizontal
///  line for the presence of a valid starting point. If none, it then 
///  spirals around (\c x,\c y) until a valid point is found.

int
tlChain::blindExtract (tlMask *mask, 
											 int    x, 
											 int    y, 
											 tlRect *rect,
											 int    connex)
{
	int	    i, j;
	int	    i_act,j_act;
	int	    i_incr,	j_incr,	cur_incr;
	int	    x1,	x2,	y1,	y2;
	int	    width, height;
	char    l_flag, r_flag, t_flag, b_flag;
	tlPixel *data;


	// check
	if (!mask) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// shortcuts
	width  = mask->width;
	height = mask->height;
	data   = mask->pixel;
	x1     = TL_MAX ((rect->x), 0);
	x2     = TL_MIN (((rect->x) + (rect->width)), width);
	y1     = TL_MAX ((rect->y), 0);
	y2     = TL_MIN (((rect->y) + (rect->height)), height);

	// check geometry
	if ((x >= x2) || (y >= y2) || (x < x1) || (y < y1)) {
		TL_ERRNO_MSG ("tlChain::blindExtract()", TL_ERROR);
		return -1;
	}

	// reset
	this->reset ();

	// do the spiral search
	i        = x;
	j        = y;
	l_flag   = r_flag = t_flag = b_flag = 0;
	cur_incr = i_incr = j_incr = 0;

	while ((!l_flag) || (!r_flag) || (!t_flag) || (!b_flag)) {
		// check for validity
		i_act = TL_MAX (i, x1);
		i_act = TL_MIN (i_act, x2);
		j_act = TL_MAX (j, y1);
		j_act = TL_MIN (j_act, y2);
		if (data[TL_PIXEL (j_act, i_act, width)] != 0)
			return this->extract (mask, i_act, j_act, rect, connex);

		// update vars (safe spiral)
		if (i_incr < abs (cur_incr)) {
			if (cur_incr > 0) i++;
			else              i--;
			i_incr++;
			if (i < x1) {
				l_flag = 1;
			}
			else if (i >= x2) {
				r_flag = 1;
			}
		}
		else if (j_incr < abs (cur_incr)) {
			if (cur_incr > 0) j++;
			else              j--;
			j_incr++;
			if (j < y1) {
				t_flag = 1;
			}
			else if (j >= y2) {
				b_flag = 1;
			}
		}
		else {
			if (cur_incr < 0) cur_incr = (-1) * (cur_incr - 1);
			else              cur_incr = (-1) * (cur_incr + 1);
			i_incr = j_incr = 0;
		}
	}


	return -1;
}



///  Extract chain corners (points of high curvature)
///  using default parameters.
///
///  \return
///  The number of corners found on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::corners (tlChain *corners)
{
	if (!corners) {
		TL_ERRNO_MSG ("tlChain::corners()", TL_ERROR_NULL_POINTER);
		return -1;
	}


	return this->corners (TL_CHAIN_CORNER_DMIN, TL_CHAIN_CORNER_DMAX, TL_CHAIN_CORNER_ALPHAMAX, corners);
}



///  Extract chain corners (points of high curvature).
///    
///  \param res         triangle segment length [chain elements]
///  \param alphaMax    max allowed angle (for normal curvature)
///  \param corners     \b [out] a chain containing the corner points
///
///  \return
///  The number of corners found on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::corners (int     res, 
									int     alphaMax, 
									tlChain *corners)
{
	if (!corners) {
		TL_ERRNO_MSG ("tlChain::corners()", TL_ERROR_NULL_POINTER);
		return -1;
	}


	return this->corners (res, res + 2, alphaMax, corners);
}



///  Extract chain corners (points of high curvature).
///    
///  \param dmin        min triangle segment length [chain elements]
///  \param dmax        max triangle segment length [chain elements]
///  \param alphaMax    max allowed angle (for normal curvature)
///  \param corners     \b [out] a chain containing the corner points
///
///  \return
///  The number of corners found on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlChain::corners (int     dmin, 
									int     dmax, 
									int     alphaMax, 
									tlChain *corners)
{
	int    p;
	int    d;
	int    k;
	int    dx1,dy1,dx2,dy2;
	double a,b,dalpha;
	double alphaThres;
	struct lpoint *p0     = NULL;
	struct lpoint *p1     = NULL;
	struct lpoint *p2     = NULL;
	struct lpoint *lmaxp  = NULL;


	if (!corners) {
		TL_ERRNO_MSG ("tlChain::corners()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  // threshold
	alphaThres = TL_PI * alphaMax / 180.00;

	// first pass - retrieve all potential candidates
	p0 = this->start;
	for (p = 0; p < (this->length); p++) {

		// initialize curvature
		p0->curvature = TL_PI;

		// retrieve initial extremities
		p1 = p0;
		p2 = p0;
		for (k = 0; k < dmin; k++) {
			if (p1->next != NULL) p1 = p1->next;
			if (p2->prev != NULL) p2 = p2->prev;
		}

		// successive attempts at different (consecutive) resolutions
		for (d = dmin; d <= dmax; d++) {

			// compute vectors
			dx1 = (p1->point->x) - (p0->point->x);
			dy1 = (p1->point->y) - (p0->point->y);
			dx2 = (p2->point->x) - (p0->point->x);
			dy2 = (p2->point->y) - (p0->point->y);

			// compute scalar values
			a = sqrt ((double)(dx1 * dx1 + dy1 * dy1));
			b = sqrt ((double)(dx2 * dx2 + dy2 * dy2));

			// compute angle
			if ((a == 0) || (b == 0)) dalpha = TL_PI;
			else                      dalpha = fabs (acos ((dx2 * dx1 + dy2 * dy1) / (a * b)));

			// store if necessary
			if (dalpha < p0->curvature) p0->curvature = dalpha;

			// get next/prev points
			if (p1->next != NULL) p1 = p1->next;
			if (p2->prev != NULL) p2 = p2->prev;
		}

		// go to next point
		p0 = p0->next;
	}

	// second pass - eliminates redundancy
	p0 = this->start;
	for (p = 0; p < (this->length); p++) {

		// consider valid curvatures
		if (p0->curvature <= alphaThres) {

			// sweep neighbors to find local minima
			p1    = p0;
			p2    = p0;
			lmaxp = p0;
			for (k = 0; k < dmax; k++) {
				if (p1->next != NULL) {
					p1 = p1->next;
					if (p1->curvature < lmaxp->curvature) lmaxp = p1;
				}
				if (p2->prev != NULL) {
					p2 = p2->prev;
					if (p2->curvature < lmaxp->curvature) lmaxp = p2;
				}
			}
		}
		// if another local minima exist, discard point
		// otherwise, it is elected as a corner
		if (lmaxp == p0) {
			corners->add (p0->point);
		}
		// go to next point
		p0 = p0->next;
	}


	return corners->getLength ();
}
