/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlChannel.h"
#include "tlRect.h"



///  \def TL_MAX_CHANNEL
///  Number of channels tolerated in an image.
///  Used in tlImage, tlObject and tlColor.

///  \file tlChannel.h
///  \brief Internal channel manipulation routines.
///  \internal



///  \overload

int
tl_to_channel (tlPixel *src, 
               int     width, 
               int     height, 
               int     pixel_width,
               int     channel, 
               tlRect  *rect, 
               tlPixel *dest)
{
  register int index;
  register int dindex;

  int i, j;
  int x1, x2;
  int y1, y2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // proceed
  for (j = y1; j < y2; j++) {
    dindex = TL_PIXEL (j, x1, width);
    index = pixel_width * dindex + channel;
    for (i = x1; i < x2; i++) {
      dest[dindex] = src[index];
      dindex++;
      index += pixel_width;
    }
  }


  return TL_NO_ERROR;
}



///  Extract a channel from an image.
///  The result is placed in a pixel array.
///
///  \param src          pixel array containing source image
///  \param width        image width
///  \param height       image height
///  \param pixel_width  bytes per pixel
///  \param channel      channel to extract
///  \param mask         conversion mask (only converts active pixels)
///  \param rect         portion of the image to process
///  \param dest         \b [out] pixel array containing the converted image
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_to_channel (tlPixel *src, 
               int     width, 
               int     height, 
               int     pixel_width,
               int     channel, 
               tlPixel *mask, 
               tlRect  *rect, 
               tlPixel *dest)
{
  register int index;
  register int dindex;

  int i, j;
  int x1, x2;
  int y1, y2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // proceed
  for (j = y1; j < y2; j++) {
    dindex = TL_PIXEL (j, x1, width);
    index = pixel_width * dindex + channel;
    for (i = x1; i < x2; i++) {
      if (mask[dindex]) {
        dest[dindex] = src[index];
      }
      dindex++;
      index += pixel_width;
    }
  }


  return TL_NO_ERROR;
}



///  \overload

int
tl_channel_put (tlPixel *src, 
                int     width, 
                int     height, 
                int     pixel_width,
                int     channel, 
                tlRect  *rect, 
                tlPixel *dest)
{
  register int index;
  register int dindex;

  int i, j;
  int x1, x2;
  int y1, y2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // proceed
  for (j = y1; j < y2; j++) {
    index = TL_PIXEL (j, x1, width);
    dindex = pixel_width * index + channel;
    for (i = x1; i < x2; i++) {
      dest[dindex] = src[index];
      dindex += pixel_width;
      index++;
    }
  }


  return TL_NO_ERROR;
}



///  Fill a given channel of a pixel array with another channel 
///  contained in another pixel array.
///
///  \param src           pixel array containing source image
///  \param width         image width
///  \param height        image height
///  \param pixel_width   destination image bytes per pixel
///  \param channel       channel to put data into
///  \param mask          conversion mask (only fills active pixels)
///  \param rect          portion of the image to process
///  \param dest          \b [out] pixel array containing the converted image
///
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_channel_put (tlPixel *src, 
                int     width, 
                int     height, 
                int     pixel_width,
                int     channel, 
                tlPixel *mask, 
                tlRect  *rect, 
                tlPixel *dest)
{
  register int index;
  register int dindex;

  int i, j;
  int x1, x2;
  int y1, y2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // proceed
  for (j = y1; j < y2; j++) {
    index = TL_PIXEL (j, x1, width);
    dindex = pixel_width * index + channel;
    for (i = x1; i < x2; i++) {
      if (mask[index]) {
        dest[dindex] = src[index];
      }
      dindex += pixel_width;
      index++;
    }
  }


  return TL_NO_ERROR;
}



///  
///  Swap channels in a pixel array
///
///  \param src           pixel array containing source image
///  \param width         image width
///  \param height        image height
///  \param pixel_width   destination image bytes per pixel
///  \param channel1      channel to swap
///  \param channel2      channel to swap
///  \param rect          portion of the image to process
///  \param dest          \b [out] pixel array containing the converted image
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_channel_swap (tlPixel *src, 
                 int     width, 
                 int     height, 
                 int     pixel_width,
                 int     channel1, 
                 int     channel2, 
                 tlRect  *rect, 
                 tlPixel *dest)
{
  int i, j;
  int x1, x2, y1, y2;
  int index;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);
  for (j = y1; j < y2; j++) {
    index = pixel_width * (j * width + x1);
    for (i = x1; i < x2; i++) {
      dest[index + channel1] = src[index + channel2];
      dest[index + channel2] = src[index + channel1];
      index += pixel_width;
    }
  }


  return TL_NO_ERROR;
}
