/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#ifdef LINUX
#include <stdlib.h>
#endif

#include "tlColor.h"
#include "tlPixelArray.h"
#include "tlPoint.h"



///  \file tlColor.h
///  \brief Color object description.

///  \class tlColor
///  \brief Color format description class.
///
///  Color values of different color formats (described \ref imageformats "here") are stored in tlColor
///  objects. They are used as arguments to color filtering and color extraction routines.  

///  \var tl_format tlColor::format
///  Color format.
///  \sa tl_format

///  \var tlPixel tlColor::raw[TL_MAX_CHANNEL]
///  Raw (unspecified) format.
///  \sa CUSTOM

///  \var tlPixel tlColor::r
///  R (red) component.
///  \sa RGB

///  \var tlPixel tlColor::g
///  G (green) component.
///  \sa RGB

///  \var tlPixel tlColor::b
///  B (blue) component.
///  \sa RGB

///  \var tlPixel tlColor::y
///  Y (luma) component.
///  \sa YUV

///  \var tlPixel tlColor::u
///  U (R-Y) component.
///  \sa YUV

///  \var tlPixel tlColor::v
///  V (B-Y) component.
///  \sa YUV

///  \var tlPixel tlColor::h1
///  H1 (hue, coarse) component.
///  \sa HSI

///  \var tlPixel tlColor::h2
///  H2 (hue, fine) component.
///  \sa HSI

///  \var tlPixel tlColor::s
///  S (saturation) component.
///  \sa HSI 

///  \var tlPixel tlColor::i
///  I (intensity) component.
///  \sa HSI

///  \var tlPixel tlColor::nr
///  NR (normalized red) component.
///  \sa NRG

///  \var tlPixel tlColor::ng
///  NG (normalized green) component.
///  \sa NRG

///  \var tlPixel tlColor::intensity
///  GRAY (intensity) component.
///  \sa GRAY

///  \def TL_COLOR_RED
///  Red color, useful for the overlay routines defined in tlImage. 

///  \def TL_COLOR_GREEN
///  Green color, useful for the overlay routines defined in tlImage 

///  \def TL_COLOR_BLUE
///  Blue color, useful for the overlay routines defined in tlImage. 

///  \def TL_COLOR_WHITE
///  White color, useful for the overlay routines defined in tlImage.

///  \def TL_COLOR_BLACK
///  Black color, useful for the overlay routines defined in tlImage.

///  \def TL_COLOR_ORANGE
///  Orange color, useful for the overlay routines defined in tlImage.

///  \def TL_COLOR_GOLD
///  Gold color, useful for the overlay routines defined in tlImage.

///  \def TL_COLOR_PINK
///  Pink color, useful for the overlay routines defined in tlImage.

///  \def TL_COLOR_YELLOW
///  Yellow color, useful for the overlay routines defined in tlImage.



///  Fast inverse cosine based on polynomial approximation,
///  maximum absolute error = 6.7626e-05.
/// 
///  \param a input (0 <= a <= 1)
/// 
///  \note
///  Adapted from MgcFastFunction.cpp::InvCos0() by Magic Software, Inc.\n
///  \htmlonly <a href="http://www.magic-software.com">http://www.magic-software.com</a>\endhtmlonly\n
///  Copyright (c) 2000-2003.  All Rights Reserved.\n

inline static float
acosFast (float a)
{
  float fResult = -0.0187293f;


  fResult *= a;
  fResult += 0.0742610f;
  fResult *= a;
  fResult -= 0.2121144f;
  fResult *= a;
  fResult += 1.5707288f;
  fResult *= (float) sqrt (1.0f - a);
  
  
  return fResult;
}



///  Fast cosine based on polynomial approximation,
///  maximum absolute error = 6.7626e-05.
/// 
///  \param a input (0 <= a <= PI/2)
/// 
///  \note
///  Adapted from MgcFastFunction.cpp::Cos0() by Magic Software, Inc.\n
///  \htmlonly <a href="http://www.magic-software.com">http://www.magic-software.com</a>\endhtmlonly\n
///  Copyright (c) 2000-2003.  All Rights Reserved.\n

inline static float
cosFast (float a)
{
  float aSqr      = a * a;
  float fResult   = 3.705e-02f;


  fResult *= aSqr;
  fResult -= 4.967e-01f;
  fResult *= aSqr;
  fResult += 1.0f;


  return fResult;
}



///  Constructor, initializes color format to RGB.
///
///  \note
///  See \ref errors "error management" for details.

tlColor::tlColor ()
{
	int i;


	this->format    = RGB;
	this->r         = 0;
	this->g         = 0;
	this->b         = 0;
	this->intensity = 0;
	for (i = 0; i < TL_MAX_CHANNEL; i++) {
		this->raw[i] = 0;
	}

	tl_errno = TL_NO_ERROR;
}



///  Constructor, initializes color format to RGB.
///
///  \param r           R color values [0..255]
///  \param g           G color values [0..255]
///  \param b           B color values [0..255]
///
///  \note
///  See \ref errors "error management" for details.

tlColor::tlColor (tlPixel r, 
									tlPixel g, 
									tlPixel b)
{
	int i;


	for (i = 0; i < TL_MAX_CHANNEL; i++) {
		this->raw[i] = 0;
	}
	this->format    = RGB;
	this->r         = this->raw[0] = r;
	this->g         = this->raw[1] = g;
	this->b         = this->raw[2] = b;
	this->intensity = (tlPixel) (((int) r + (int) g + (int) b) / 3);

	tl_errno = TL_NO_ERROR;
}



///  Constructor, initialises color format to HSI.
///
///  \param h1          H1 (coarse) color values [0..255]
///  \param h2          H2 (fine) color values   [0..255]
///  \param s           S color values           [0..255]
///  \param i           I color values           [0..255]
///
///  \note
///  See \ref errors "error management" for details.

tlColor::tlColor (tlPixel h1, 
									tlPixel h2, 
									tlPixel s, 
									tlPixel i)
{
	int k;


	for (k = 0; k < TL_MAX_CHANNEL; k++) {
		this->raw[k] = 0;
	}
	this->format    = HSI;
	this->h1        = this->raw[0] = h1;
	this->h2        = this->raw[1] = h2;
	this->s         = this->raw[2] = s;
	this->i         = this->raw[3] = i;
	this->intensity = i;

	tl_errno = TL_NO_ERROR;
}



///  Constructor, initializes color format to NRG.
///
///  \param nr          NR color value [0..255]
///  \param ng          NG color value [0..255]
///
///  \note
///  See \ref errors "error management" for details.

tlColor::tlColor (tlPixel nr, 
									tlPixel ng)
{
	int i;


	for (i = 0; i < TL_MAX_CHANNEL; i++) {
		this->raw[i] = 0;
	}
	this->format    = NRG;
	this->nr        = this->raw[0] = nr;
	this->ng        = this->raw[1] = ng;
	this->intensity = (tlPixel) (((int) nr + (int) ng) / 2);

	tl_errno = TL_NO_ERROR;
}



///  Constructor, initializes color format to GRAY.
///
///  \param intensity   GRAY color values [0..255]
///
///  \note
///  See \ref errors "error management" for details.

tlColor::tlColor (tlPixel intensity)
{
	int i;


	for (i = 0; i < TL_MAX_CHANNEL; i++) {
		this->raw[i] = 0;
	}
	this->format = GRAY;
	this->intensity = this->raw[0] = intensity;

	tl_errno = TL_NO_ERROR;
}



///  Constructor, creates a copy of an existing color.  
///
///  \param src      color to copy
///
///  \note
///  See \ref errors "error management" for details.

tlColor::tlColor (tlColor *src)
{
	int i;
	

	if (!src) {
		TL_ERRNO_MSG ("tlColor::tlColor()", TL_ERROR_NULL_POINTER);
		return;
	}

	this->format    = src->format;
	this->r         = src->r;
	this->g         = src->g;
	this->b         = src->b;
	this->intensity = src->intensity;
  this->y         = src->y;
  this->u         = src->u;
  this->v         = src->v;
  this->h1        = src->h1;
  this->h2        = src->h2;
  this->s         = src->s;
  this->i         = src->i;
  this->nr        = src->nr;
  this->ng        = src->ng;
	for (i = 0; i < TL_MAX_CHANNEL; i++) {
		this->raw[i] = src->raw[i];
	}

	tl_errno = TL_NO_ERROR;
}



///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlColor::~tlColor ()
{
	tl_errno = TL_NO_ERROR;
}



///  Copy data from another color.
///
///  \param color     color to copy
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::copy (tlColor *color)
{
	int	k;


	// check
	if (!color) {
		TL_ERRNO_MSG ("tlColor::copy()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  // proceed
	this->format    = color->format;
	this->intensity = color->intensity;
	this->r         = color->r;
	this->g         = color->g;
	this->b         = color->b;
  this->y         = color->y;
  this->u         = color->u;
  this->v         = color->v;
	this->nr        = color->nr;
	this->ng        = color->ng;
	this->h1        = color->h1;
	this->h2        = color->h2;
	this->s         = color->s;
	this->i         = color->i;
	for (k = 0; k < TL_MAX_CHANNEL; k++) {
		this->raw[k] = color->raw[k];
	}


	return 0;
}



///  Copy data to another color.
///
///  \param color     \b [out] color to copy to
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::copyTo (tlColor *color)
{
	int	k;


	// check
	if (!color) {
		TL_ERRNO_MSG ("tlColor::copyTo()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// proceed
	color->format     = this->format;
	color->intensity  = this->intensity;
	color->r          = this->r;
	color->g          = this->g;
	color->b          = this->b;
  color->y          = this->y;
  color->u          = this->u;
  color->v          = this->v;
	color->nr         = this->nr;
	color->ng         = this->ng;
	color->h1         = this->h1;
	color->h2         = this->h2;
	color->s          = this->s;
	color->i          = this->i;
	for (k = 0; k < TL_MAX_CHANNEL; k++) {
		color->raw[k] = this->raw[k];
	}


	return 0;
}



///  Set color value for a particular channel.
///
///  \param channel     channel to be affected
///  \param val         new channel value
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setChannel (int     channel, 
										 tlPixel val)
{
	this->raw[channel] = val;
	return 0;
}



///  Set the color argument to the pixel value of a given pixel array.
///
///  \param pixelArray   source pixel array to get color from
///  \param point        pixel to consider in the pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::set (tlPixelArray *pixelArray, 
							tlPoint      *point)
{
	return set (pixelArray, point->x, point->y);
}



///  Set the color argument to the pixel value of a given pixel array.
///
///  \param pixelArray   source pixel array to get color from
///  \param x            X coord of the pixel to consider
///  \param y            Y coord of the pixel to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::set (tlPixelArray *pixelArray, 
							int          x, 
							int          y)
{
	int	k;


	// check
	if (!pixelArray) {
		TL_ERRNO_MSG ("tlColor::set()", TL_ERROR_NULL_POINTER);
		return -1;
	}
	
  // check
	if (pixelArray->hr) {
    TL_ERRNO_MSG ("tlBlobs::extract()", TL_ERROR_HIGHRES_INCOMPATIBLE);
		return -1;
  }

  // init color
	format = pixelArray->format;

	// copy
	for (k = 0; k < pixelArray->pixelWidth; k++) {
		raw[k] = (tlPixel)pixelArray->getPixel (x, y, k);
	}
	
  switch (format) {
	case RGB:
		r = raw[0];
		g = raw[1];
		b = raw[2];
		intensity = (tlPixel)((r + g + b) / 3);
		break;
  case YUV:
    y = raw[0];
    u = raw[1];
		v = raw[2];
		intensity = raw[0];
    break;
	case HSI:
		h1 = raw[0];
		h2 = raw[1];
		s = raw[2];
		i = raw[3];
		intensity = i;
		break;
	case NRG:
		nr = raw[0];
		ng = raw[1];
		intensity = (tlPixel)((nr + ng) / 2);
		break;
	case GRAY:
		intensity = raw[0];
		break;
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::set()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}


	return 0;
}



///  Swap two channel values.
///
///  \param channel1    channel to be swapped
///  \param channel2    channel to be swapped
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::swapChannels (int channel1, 
											 int channel2)
{
	if ((channel1 >= TL_MAX_CHANNEL) || (channel1 >= TL_MAX_CHANNEL)) {
		TL_ERRNO_MSG ("tlColor::swapChannels()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
		return -1;
	}

	tlPixel	temp        = this->raw[channel1];
	this->raw[channel1] = this->raw[channel2];
	this->raw[channel2] = temp;


	return 0;
}



///  Set the RGB color value.
///
///  \param r      R color value [0..255]
///  \param g      G color value [0..255]
///  \param b      B color value [0..255]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setRgb (tlPixel r, 
								 tlPixel g, 
								 tlPixel b)
{
	this->format = RGB;
	this->r = this->raw[0] = r;
	this->g = this->raw[1] = g;
	this->b = this->raw[2] = b;
	this->intensity = (tlPixel) (((int) r + (int) g + (int) b) / 3);


	return 0;
}



///  Set the YUV color value.
///
///  \param y     Y color value [0..255]
///  \param u     U color value [0..255]
///  \param v     V color value [0..255]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setYuv (tlPixel y, 
								 tlPixel u, 
								 tlPixel v)
{
	this->format = YUV;
	this->y = this->raw[0] = y;
	this->u = this->raw[1] = u;
	this->v = this->raw[2] = v;
	this->intensity = y;


	return 0;
}



///  Set NRG color value.
///
///  \param nr      NR color value [0..255]
///  \param ng      NG color value [0..255]; 
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setNrg (tlPixel nr, 
								 tlPixel ng)
{
	this->format = NRG;
	this->nr = this->raw[0] = nr;
	this->ng = this->raw[1] = ng;
	this->intensity = (tlPixel) (((int) nr + (int) ng) / 2);


	return 0;
}



///  Set HSI color value.
///
///  \param h1     H (coarse) color values [0..255]
///  \param h2     H (fine) color values [0..255]
///  \param s      S color values [0..255]
///  \param i      I color values [0..255]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setHsi (tlPixel h1, 
								 tlPixel h2, 
								 tlPixel s, 
								 tlPixel i)
{
	this->format = HSI;
	this->h1 = this->raw[0] = h1;
	this->h2 = this->raw[1] = h2;
	this->s  = this->raw[2] = s;
	this->i  = this->raw[3] = i;
	this->intensity = i;


	return 0;
}



///  Set GRAY color value.
///
///  \param intensity   GRAY color value [0..255]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setGray (tlPixel intensity)
{
	this->intensity = this->raw[0] = intensity;


	return 0;
}


///  Set color to the lowest value in a portion of a given pixel array;
///  color takes the format of the input pixel array.
///
///  \param pixelArray  the pixel array to extract the color from
///  \param rect        the portion of the pixel array to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setMin (tlPixelArray *pixelArray, 
								 tlRect       *rect)
{
	// check
	if ((!pixelArray) || (!rect)) {
		TL_ERRNO_MSG ("tlColor::setMin()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// check
	if (pixelArray->hr) {
    TL_ERRNO_MSG ("tlBlobs::extract()", TL_ERROR_HIGHRES_INCOMPATIBLE);
		return -1;
  }

  // init color
	this->format = pixelArray->format;
	switch (format) {
	case RGB:
		return tl_extract_color_min_rgb  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
  case YUV:
		return tl_extract_color_min_yuv  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case HSI:
		return tl_extract_color_min_hsi  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case NRG:
		return tl_extract_color_min_nrg  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case GRAY:
		return tl_extract_color_min_gray (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::setMin()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
}



///  Set color to the highest value in a portion of a given pixel array;
///  color takes the format of the input pixel array.
///
///  \param pixelArray  the pixel array to extract the color from
///  \param rect        the portion of the pixel array to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setMax (tlPixelArray *pixelArray, 
								 tlRect       *rect)
{
	// check
	if ((!pixelArray) || (!rect)) {
		TL_ERRNO_MSG ("tlColor::setMax()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// check
	if (pixelArray->hr) {
    TL_ERRNO_MSG ("tlBlobs::extract()", TL_ERROR_HIGHRES_INCOMPATIBLE);
		return -1;
  }

  // init color
	this->format = pixelArray->format;
	switch (format) {
	case RGB:
		return tl_extract_color_max_rgb  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
  case YUV:
		return tl_extract_color_max_yuv  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case HSI:
		return tl_extract_color_max_hsi  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case NRG:
		return tl_extract_color_max_nrg  (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case GRAY:
		return tl_extract_color_max_gray (pixelArray->pixel, pixelArray->width, pixelArray->height, this, rect);
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::setMax()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
}



///  Set color to the maximum variation in a portion of a given pixel array;
///  color takes the format of the input pixel array.
///
///  \param pixelArray  the pixel array to extract the color from
///  \param rect        the portion of the pixel array to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::setVar (tlPixelArray *pixelArray, 
								 tlRect       *rect)
{
  tlColor	minColor;
	tlColor	maxColor;

  
  // check
	if ((!pixelArray) || (!rect)) {
		TL_ERRNO_MSG ("tlColor::setVar()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// check
	if (pixelArray->hr) {
    TL_ERRNO_MSG ("tlBlobs::extract()", TL_ERROR_HIGHRES_INCOMPATIBLE);
		return -1;
  }

  // init color
	this->format = pixelArray->format;
	switch (format) {
	case RGB:
		tl_extract_color_min_rgb (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &minColor, rect);
		tl_extract_color_max_rgb (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &maxColor, rect);
		break;
  case YUV:
		tl_extract_color_min_yuv (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &minColor, rect);
		tl_extract_color_max_yuv (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &maxColor, rect);
		break;
	case HSI:
		tl_extract_color_min_hsi (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &minColor, rect);
		tl_extract_color_max_hsi (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &maxColor, rect);
		break;
	case NRG:
		tl_extract_color_min_nrg (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &minColor, rect);
		tl_extract_color_max_nrg (pixelArray->pixel, pixelArray->width,
		                          pixelArray->height, &maxColor, rect);
		break;
	case GRAY:
		tl_extract_color_min_gray (pixelArray->pixel, pixelArray->width,
		                           pixelArray->height, &minColor, rect);
		tl_extract_color_max_gray (pixelArray->pixel, pixelArray->width,
		                           pixelArray->height, &maxColor, rect);
		break;
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::setVar()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
	
  
  maxColor.sub (&minColor);
  return this->copy (&maxColor);
}



///  Convert color to RGB format (if possible).
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::toRgb ()
{
	switch (format) {
	case RGB:
		TL_ERRNO_MSG ("tlColor::toRgb()", TL_ERROR_NOT_REQUIRED);
		return -1;
  case YUV:
    {
      format = RGB;
      int C = y -  16;
      int D = u - 128;
      int E = v - 128;
      r = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      g = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
      b = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
    }
    return 0;
	case HSI:
		{
			int   R,G,B;
			float H, S;
			float TWOPI = (float) (2 * TL_PI);
			float PI3   = (float) (TL_PI / 3);
			float H1    = (float) (2 * TL_PI / 3);
			float H2    = (float) (4 * TL_PI / 3);

			// scaled hue conversion (16-bit)
			unsigned short int sMax = TL_MAX_PIXEL_VAL * TL_MAX_PIXEL_VAL;

      format = RGB;

			// scale hue 16-bit to [0,2PI] 
			H = ((h1 << 8) | h2) * TWOPI / sMax;

			// scale saturation [0,TL_MAX_PIXEL_VAL] to [0,1] 
			S = s / (float) TL_MAX_PIXEL_VAL;

			if (H <= H1) {		/* [0, 2PI/3] */
				B = (int) (intensity * (1 - S));
				R = (int) (intensity * (1 + (S * cosFast (H) / cosFast (PI3 - H))));
				G = 3 * intensity - R - B;
			}
			else if (H <= H2) {	/* (2PI/3, 4PI/3) */
				float h = H - H1;
				R = (int) (intensity * (1 - S));
				G = (int) (intensity * (1 + (S * cosFast (h) / cosFast (PI3 - h))));
				B = 3 * intensity - R - G;
			}

			else {			/* (4PI/3, 2PI) */
				float h = H - H2;
				G = (int) (intensity * (1 - S));
				B = (int) (intensity * (1 + (S * cosFast (h) / cosFast (PI3 - h))));
				R = 3 * intensity - G - B;
			} 
			R =           TL_MIN (R, TL_MAX_PIXEL_VAL);
			r = (tlPixel) TL_MAX (R, TL_MIN_PIXEL_VAL);
			G =           TL_MIN (G, TL_MAX_PIXEL_VAL);
			g = (tlPixel) TL_MAX (G, TL_MIN_PIXEL_VAL);
			B =           TL_MIN (B, TL_MAX_PIXEL_VAL);
			b = (tlPixel) TL_MAX (B, TL_MIN_PIXEL_VAL);
		}
		return 0;
	default:
		TL_ERRNO_MSG ("tlColor::toRgb()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
}



///  Convert color to NRG format (if possible).
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::toNrg ()
{
	int   sum;
  float fsum;


	switch (format) {
	case RGB:
		sum       = r + g + b;
		format    = NRG;
		nr        = (tlPixel) (r * TL_MAX_PIXEL_VAL / sum);
		ng        = (tlPixel) (g * TL_MAX_PIXEL_VAL / sum);
		intensity = (tlPixel) (sum/3);
		return 0;
  case YUV:
    fsum      = (3.492f*(y-16) + 1.627f*(u-128) + 0.783f*(v-128));
    format    = NRG;
    nr        = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (int)(TL_MAX_PIXEL_VAL * (1.164f*(y-16) + 1.596f*(v-128)                 )/fsum), TL_MAX_PIXEL_VAL);
    ng        = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (int)(TL_MAX_PIXEL_VAL * (1.164f*(y-16) - 0.813f*(v-128) - 0.391f*(u-128))/fsum), TL_MAX_PIXEL_VAL);
    intensity = y;
    return 0;
	case HSI:
		if (0 == toRgb ()) return toNrg ();
		else return -1;
	case NRG:
		TL_ERRNO_MSG ("tlColor::toNrg()", TL_ERROR_NOT_REQUIRED);
		return -1;
	default:
		TL_ERRNO_MSG ("tlColor::toNrg()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
}



///  Convert color to HSI format (if possible).
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::toHsi ()
{
	switch (format) {
	case RGB:
		{
      int rg, rb, gb;		// must be at least signed short!
			float H, TWOPI = (float) (2 * TL_PI);
			tlPixel I, S;
			int min, denom;

			// scaled hue conversion (16-bit)
			unsigned short int sHue, sMax = TL_MAX_PIXEL_VAL * TL_MAX_PIXEL_VAL;

      format = HSI;
			rg = r - g;
			rb = r - b;
			gb = g - b;

			// H = [0,2PI] 
			denom = rg * rg + rb * gb;
			if (denom == 0) {	
				H = 0;
			}
			else {
				H = (float) acosFast ((rg + rb) / (2.0f * (float) sqrt ((float)denom)));
				if (b > g) {
					H = TWOPI - H;
				}
			}

			// I = [0,255] 
			I = (tlPixel)((r + g + b) / 3);

			// scale S [0,1] to [0,TL_MAX_PIXEL_VAL] 
			if (I == 0) {
				S = 0;
			}
			else {
				min = TL_MIN (r, g);
				min = TL_MIN (min, b);
				S = (tlPixel) (TL_MAX_PIXEL_VAL - (min * TL_MAX_PIXEL_VAL / I));
			}

			// scale hue [0,2PI] to 16-bit 
			sHue      = (unsigned short int) (sMax * H / TWOPI);
			h1        = (tlPixel) (sHue >> 8);
			h2        = (tlPixel) (sHue & 0xff);
			s         = (tlPixel) S;
			intensity = (tlPixel) I;
		}
		return 0;
  case YUV:
    if (0 > toRgb ()) return -1;
    if (0 > toHsi ()) return -1;
    break;
	case HSI:
		TL_ERRNO_MSG ("tlColor::toHsi()", TL_ERROR_NOT_REQUIRED);
		return -1;
	default:
		TL_ERRNO_MSG ("tlColor::toHsi()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}


  return 0;
}



///  Convert color to GRAY format if (possible).
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::toGray ()
{
	switch (format) {
	case RGB:
		format = GRAY;
		intensity = (tlPixel) (((int) r + (int) g + (int) b) / 3);
		return 0;
  case YUV:
    format = GRAY;
    intensity = y;
    return 0;
	case HSI:
		format = GRAY;
		intensity = i;
		return 0;
	case NRG:
		format = GRAY;
		intensity = (tlPixel) (nr + ng);
		return 0;
	case GRAY:
		TL_ERRNO_MSG ("tlColor::toGray()", TL_ERROR_NOT_REQUIRED);
		return -1;
	default:
		TL_ERRNO_MSG ("tlColor::toGray()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
}



///  Convert color to BIN format if (possible).
///
///  \param threshold   threshold value for pixel (de)activation
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlColor::toBin (int threshold)
{
	switch (format) {
	case RGB:
  case YUV:
	case NRG:
	case HSI:
	case GRAY:
		format = BIN;
		if (intensity >= threshold)
			intensity = TL_MAX_PIXEL_VAL;
		else
			intensity = TL_MIN_PIXEL_VAL;
		return 0;
	case BIN:
		TL_ERRNO_MSG ("tlColor::toBin()", TL_ERROR_NOT_REQUIRED);
		return -1;
	default:
		TL_ERRNO_MSG ("tlColor::toBin()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}
}



///  Perform color addition.
///
///  \param color     color to add to current color
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  Result is cast to \ref TL_MAX_PIXEL_VAL in case of an overflow.

int
tlColor::add (tlColor * color)
{
	// check
	if (!color) {
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  // test format
	if (color->format != format) {
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  // perform addition
	switch (format) {
	case RGB:
		r = (tlPixel)(TL_MIN (r + color->r, TL_MAX_PIXEL_VAL));
		g = (tlPixel)(TL_MIN (g + color->g, TL_MAX_PIXEL_VAL));
		b = (tlPixel)(TL_MIN (b + color->b, TL_MAX_PIXEL_VAL));
		intensity = (tlPixel)((r + g + b) / 3);
		break;
  case YUV:
  	y = (tlPixel)(TL_MIN (y + color->y, TL_MAX_PIXEL_VAL));
		u = (tlPixel)(TL_MIN (u + color->u, TL_MAX_PIXEL_VAL));
		v = (tlPixel)(TL_MIN (v + color->v, TL_MAX_PIXEL_VAL));
		intensity = y;
		break;
	case HSI:
		h1 = (tlPixel)(TL_MIN (h1 + color->h1, TL_MAX_PIXEL_VAL));
		h2 = (tlPixel)(TL_MIN (h2 + color->h2, TL_MAX_PIXEL_VAL));
		s = (tlPixel)(TL_MIN (s + color->s, TL_MAX_PIXEL_VAL));
		i = (tlPixel)(TL_MIN (i + color->i, TL_MAX_PIXEL_VAL));
		intensity = i;
		break;
	case NRG:
		nr = (tlPixel)(TL_MIN (nr + color->nr, TL_MAX_PIXEL_VAL));
		ng = (tlPixel)(TL_MIN (ng + color->ng, TL_MAX_PIXEL_VAL));
		intensity = (tlPixel)((nr + ng) / 2);
		break;
	case GRAY:
		intensity = (tlPixel)(TL_MIN (intensity + color->intensity, TL_MAX_PIXEL_VAL));
		break;
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  
  return 0;
}



///  Perform color addition; the result is placed in another color object.
///
///  \param color     color to add to current color
///  \param dest      \b [out] destination color
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  \a dest must be allocated.
///  Result is cast to \ref TL_MAX_PIXEL_VAL in case of an overflow.

int
tlColor::add (tlColor *color, 
              tlColor *dest)
{
	// check
	if ((!color) || (!dest)) {
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  // test format
	if (color->format != format) {
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  // init format
	dest->copy (this);

	// perform addition
	switch (format) {
	case RGB:
		dest->r = (tlPixel)(TL_MIN (r + color->r, TL_MAX_PIXEL_VAL));
		dest->g = (tlPixel)(TL_MIN (g + color->g, TL_MAX_PIXEL_VAL));
		dest->b = (tlPixel)(TL_MIN (b + color->b, TL_MAX_PIXEL_VAL));
		dest->intensity = (tlPixel)((dest->r + dest->g + dest->b) / 3);
		break;
  case YUV:
		dest->y = (tlPixel)(TL_MIN (y + color->y, TL_MAX_PIXEL_VAL));
		dest->u = (tlPixel)(TL_MIN (u + color->u, TL_MAX_PIXEL_VAL));
		dest->v = (tlPixel)(TL_MIN (v + color->v, TL_MAX_PIXEL_VAL));
		dest->intensity = y;
		break;
	case HSI:
		dest->h1 = (tlPixel)(TL_MIN (h1 + color->h1, TL_MAX_PIXEL_VAL));
		dest->h2 = (tlPixel)(TL_MIN (h2 + color->h2, TL_MAX_PIXEL_VAL));
		dest->s = (tlPixel)(TL_MIN (s + color->s, TL_MAX_PIXEL_VAL));
		dest->i = (tlPixel)(TL_MIN (i + color->i, TL_MAX_PIXEL_VAL));
		dest->intensity = dest->i;
		break;
	case NRG:
		dest->nr = (tlPixel)(TL_MIN (nr + color->nr, TL_MAX_PIXEL_VAL));
		dest->ng = (tlPixel)(TL_MIN (ng + color->ng, TL_MAX_PIXEL_VAL));
		dest->intensity = (tlPixel)((dest->nr + dest->ng) / 2);
		break;
	case GRAY:
		dest->intensity = (tlPixel)(TL_MIN (intensity + color->intensity, TL_MAX_PIXEL_VAL));
		break;
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  
  return 0;
}



///  Perform color subtraction.
///
///  \param color     color to sub from current color
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remarks
///  Result is the absolute value of the difference.

int
tlColor::sub (tlColor *color)
{
	// check
	if (!color) {
		TL_ERRNO_MSG ("tlColor::sub()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  // test format
	if (color->format != format) {
		TL_ERRNO_MSG ("tlColor::sub()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  // perform addition
	switch (format) {
	case RGB:
		r = (tlPixel) abs (r - color->r);
		g = (tlPixel) abs (g - color->g);
		b = (tlPixel) abs (b - color->b);
		intensity = (tlPixel)((r + g + b) / 3);
		break;
	case YUV:
		y = (tlPixel) abs (y - color->y);
		u = (tlPixel) abs (u - color->u);
		v = (tlPixel) abs (v - color->v);
		intensity = y;
		break;
	case HSI:
		h1 = (tlPixel) abs (h1 - color->h1);
		h2 = (tlPixel) abs (h2 - color->h2);
		s = (tlPixel) abs (s - color->s);
		i = (tlPixel) abs (i - color->i);
		intensity = i;
		break;
	case NRG:
		nr = (tlPixel) abs (nr - color->nr);
		ng = (tlPixel) abs (ng - color->ng);
		intensity = (tlPixel)((nr + ng) / 2);
		break;
	case GRAY:
		intensity = (tlPixel) abs (intensity - color->intensity);
		break;
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  
  return 0;
}



///  Perform color subtraction; the result is placed in another color object.
///
///  \param color     color to sub from current color
///  \param dest      destination color
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  \a dest must be allocated.\n
///  Result is the absolute value of the difference. 

int
tlColor::sub (tlColor *color, 
              tlColor *dest)
{
	// check
	if ((!color) || (!dest)) {
		TL_ERRNO_MSG ("tlColor::sub()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  // test format
	if (color->format != format) {
		TL_ERRNO_MSG ("tlColor::sub()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  // init format
	dest->copy (this);

	// perform addition
	switch (format) {
	case RGB:
		dest->r = (tlPixel) abs (r - color->r);
		dest->g = (tlPixel) abs (g - color->g);
		dest->b = (tlPixel) abs (b - color->b);
		dest->intensity = (tlPixel)((dest->r + dest->g + dest->b) / 3);
		break;
  case YUV:
		dest->y = (tlPixel) abs (y - color->y);
		dest->u = (tlPixel) abs (u - color->u);
		dest->v = (tlPixel) abs (v - color->v);
		dest->intensity = y;
		break;
	case HSI:
		dest->h1 = (tlPixel) abs (h1 - color->h1);
		dest->h2 = (tlPixel) abs (h2 - color->h2);
		dest->s = (tlPixel) abs (s - color->s);
		dest->i = (tlPixel) abs (i - color->i);
		dest->intensity = dest->i;
		break;
	case NRG:
		dest->nr = (tlPixel) abs (nr - color->nr);
		dest->ng = (tlPixel) abs (ng - color->ng);
		dest->intensity = (tlPixel)((dest->nr + dest->ng) / 2);
		break;
	case GRAY:
		dest->intensity = (tlPixel) abs (intensity - color->intensity);
		break;
	case BIN:
	default:
		TL_ERRNO_MSG ("tlColor::add()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

  
  return 0;
}
