/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlColorModel.h"
#include "tlChain.h"
#include "tlHist2D.h"
#include "tlImage.h"



///  \file tlColorModel.h
///  \brief Color model object definition.

///  \class tlColorModel
///  \brief Statistical color model class.
///
///  This is a color definition class. It requires training and supports both normalized RGB 
///  and HSI format. 2D histograms of the color distribution in the given format are built 
///  based on the training data. Can be used to define any (and even multiple) color.

///  \var tlHist2D  *tlColorModel::NRGHist
///  Normalized RGB \ref tl_format "color model"

///  \var tlHist2D  *tlColorModel::HSHist
///  HSI \ref tl_format "color model"

///  \def TL_COLOR_MODEL_DETECTION_THRESHOLD
///  Default contribution of a color to the color model [%]. 




///  Constructor.

tlColorModel::tlColorModel()
{
	NRGHist = new tlHist2D(256,256);
	HSHist  = new tlHist2D(256,256);
}



///  Constructor.
///  \param filebase        basename for color model file to upload.

tlColorModel::tlColorModel (char *filebase)
{
	NRGHist = new tlHist2D(256,256);
	HSHist  = new tlHist2D(256,256);
	load (filebase);
}



///  Destructor.

tlColorModel::~tlColorModel(void)
{
	delete NRGHist;
	delete HSHist;
}



///  Reset color model.

int
tlColorModel::reset ()
{
	NRGHist->init (256,256);
	HSHist->init  (256,256);

	return 0;
}



///  Once training is finished, call this method to scale the histogram correctly.
///  The threshold argument for the filter() method will then be expressed in percent,
///  as opposed to pixel count.

int
tlColorModel::finalize ()
{
	NRGHist->normalize ();
	HSHist->normalize  ();
	return 0;
}




///  Update the histograms based on a segmented image of either format.
///
///  \param image     segmented image, only pixels of the desired color are non-zero.
///
///  \note
///  Only the histogram of the image format is updated.
///  User should call finalize() once training is finished.

int
tlColorModel::train (tlImage *image)
{
	int        i,j;
	tlPixel    *pixel   = image->pixel;
	tlHistData **data;


	switch (image->format) {
	case NRG:
		data = NRGHist->data;
		for (j=0; j<(image->height); j++) {
			for (i=0; i<(image->width); i++) {
				if ((*(pixel)) || (*(pixel+1))) {
					data[*(pixel)][*(pixel+1)] += 1;
				}
				pixel += 2;
			}
		}
		NRGHist->update ();
		break;
	case HSI:
		data = HSHist->data;
		for (j=0; j<(image->height); j++) {
			for (i=0; i<(image->width); i++) {
				if ((*(pixel)) || (*(pixel+1)) || (*(pixel+2)) || (*(pixel+3))) {
					data[*(pixel)][*(pixel+2)] += 1;
				}
				pixel += 4;
			}
		}
		HSHist->update ();
		break;
  default:
    TL_ERRNO_MSG ("tlColorModel::train()", TL_ERROR_INVALID_FORMAT);
	}


	return 0;
}



///  Add the contribution of a single pixel from a training image to the color histogram.
///
///  \param image     the image containing the pixel of interest.
///  \param point     the location of the pixel to train on.
///
///  \note
///  Only the histogram of the image format is updated. A better (and faster) 
///  way to build the color distribution is to use a pre-segmented image 
///  with train(tlImage *image).
///  User should call finalize() once training is finished.

int
tlColorModel::train (tlImage *image, 
									   tlPoint *point)
{
	switch (image->format) {
	case NRG:
		NRGHist->incVal (image->getPixel (point,0), image->getPixel (point, 1));
		break;
	case HSI:
		HSHist->incVal (image->getPixel (point,0), image->getPixel (point, 2));
		break;
  default:
    TL_ERRNO_MSG ("tlColorModel::train()", TL_ERROR_INVALID_FORMAT);
	}


	return 0;
}



///  Add the contribution of a number of pixels from a training image to the color histogram.
///
///  \param image     the image containing the pixel of interest.
///  \param chain     the locations of the pixels to train on.
///
///  \note
///  Only the histogram of the image format is updated. A better (and faster) 
///  way to build the color distribution is to use a pre-segmented image 
///  with train(tlImage *image).
///  User should call finalize() once training is finished.

int
tlColorModel::train (tlImage *image, 
									   tlChain *chain)
{
  int i;

	switch (image->format) {
	case NRG:
		for (i=0; i<chain->getLength(); i++) {
			NRGHist->incVal (image->getPixel (chain->getPoint(i),0), image->getPixel (chain->getPoint(i), 1));
		}
		break;
	case HSI:
		for (i=0; i<chain->getLength(); i++) {
			HSHist->incVal (image->getPixel (chain->getPoint(i),0), image->getPixel (chain->getPoint(i), 2));
		}
		break;
  default:
    TL_ERRNO_MSG ("tlColorModel::train()", TL_ERROR_INVALID_FORMAT);
	}

	return 0;
}



///  Filter a color image using the (thresholded) color distribution contained in the histograms.
///  That is, only pixels with color value that contribute for more than 'threshold' percent to
///  the histogram will be kept.
///
///  \param image       the image to be filtered.
///  \param threshold   percentage or minimal pixel count.
///
///  \note
///  Threshold represents a percentage only if finalize() is called after the training is over.
///  Otherwise, threshold must be expressed in pixel counts.

int
tlColorModel::filter (tlImage *image, 
										  int     threshold)
{
	int			   i,j;
	tlPixel	   *pixel;
	tlHistData **data;
	int        width, height;

	
	// shortcuts
	width		= image->width;
	height	= image->height;
	pixel		= image->pixel;

	// filtering
	switch (image->format) {
	case NRG:
		data = NRGHist->data;
		for (j=0; j<height; j++) {
			for (i=0; i<width; i++) {
				if (data[*(pixel)][*(pixel+1)] < threshold) {
					*(pixel  ) = 0;
					*(pixel+1) = 0;
				}
				pixel += 2;
			}
		}
		break;
	case HSI:
		data = HSHist->data;
		for (j=0; j<height; j++) {
			for (i=0; i<width; i++) {
				if (data[*(pixel)][*(pixel+2)] < threshold) {
					*(pixel  ) = 0;
					*(pixel+1) = 0;
					*(pixel+2) = 0;
					*(pixel+3) = 0;
				}
				pixel += 4;
			}
		}
		break;
  default:
    TL_ERRNO_MSG ("tlColorModel::filter()", TL_ERROR_INVALID_FORMAT);
	}


	return 0;
}



///  Filter a portion of a color image using the (thresholded) color distribution 
///  contained in the histograms. That is, only pixels with color value that contribute 
///  for more than 'threshold' percent to the histogram will be kept.
///
///  \param image       the image to be filtered.
///  \param rect        portion of the image to process
///  \param threshold   percentage or minimal pixel count.
///
///  \note
///  Threshold represents a percentage only if finalize() is called after the training is over.
///  Otherwise, threshold must be expressed in pixel counts.

int
tlColorModel::filter	(tlImage *image, 
									  	 tlRect  *rect, 
									  	 int     threshold)
{
	int			   i,j;
	int        x1,x2,y1,y2;
	int        width, height, pixelWidth;
	tlPixel	   *pixel;
	tlHistData **data;

	
	// shortcuts
	width		   = image->width;
	height	   = image->height;
	pixelWidth = image->pixelWidth;
	pixel		   = image->pixel;
	x1         = TL_MAX ((rect->x), 0);
	x2         = TL_MIN ((rect->x)+(rect->width), width);
	y1         = TL_MAX ((rect->y), 0);
	y2         = TL_MIN ((rect->y)+(rect->height), height);

	// filtering
	switch (image->format) {
	case NRG:
		data = NRGHist->data;
		for (j=y1; j<y2; j++) {
			register tlPixel *ptr = pixel + (j*width+x1)*pixelWidth;
			for (i=x1; i<x2; i++) {
				if (data[*(ptr)][*(ptr+1)] < threshold) {
					*(ptr  ) = 0;
					*(ptr+1) = 0;
				}
				ptr += 2;
			}
		}
		break;
	case HSI:
		data = HSHist->data;
		for (j=y1; j<y2; j++) {
			register tlPixel *ptr = pixel + (j*width+x1)*pixelWidth;
			for (i=x1; i<x2; i++) {
				if (data[*(ptr)][*(ptr+2)] < threshold) {
					*(ptr  ) = 0;
					*(ptr+1) = 0;
					*(ptr+2) = 0;
					*(ptr+3) = 0;
				}
				ptr += 4;
			}
		}
		break;
  default:
    TL_ERRNO_MSG ("tlColorModel::filter()", TL_ERROR_INVALID_FORMAT);
	}


	return 0;
}



///  Test whether a given color accounts for more than a given threshold
///  in the color histograms.
///
///  \param color       the color to test for
///  \param threshold   the histogram threshold value

int
tlColorModel::test (tlColor *color,
                    int     threshold)
{
	switch (color->format) {
	case NRG:
		if (NRGHist->data[color->nr][color->ng] < threshold) return 0;
    else                                                 return 1;
    break;
  case HSI:
		if (HSHist->data[color->h1][color->s] < threshold)   return 0;
    else                                                 return 1;
    break;
  default:
    TL_ERRNO_MSG ("tlColorModel::test()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Save histograms to disk.
///
///  \param filebase   base name for the hitogram files.

int
tlColorModel::save (char *filebase)
{
	int  res = 0;
	char NRGFile[255];
	char HSFile [255];


	sprintf (NRGFile, "%sNRG.hst", filebase);
	sprintf (HSFile,  "%sHS.hst",  filebase);

	if (TL_NO_ERROR != NRGHist->save (NRGFile)) res = -1;
	if (TL_NO_ERROR != HSHist->save  (HSFile))  res = -1;


	return res;
}




///  Load histograms from disk.
///
///  \param filebase   base name for the hitogram files.

int
tlColorModel::load (char *filebase)
{
	int  res = 0;
	char NRGFile[255];
	char HSFile [255];


	sprintf (NRGFile, "%sNRG.hst", filebase);
	sprintf (HSFile,  "%sHS.hst",  filebase);

	if (TL_NO_ERROR != NRGHist->load (NRGFile)) res = -1;
	if (TL_NO_ERROR != HSHist->load  (HSFile))  res = -1;


	return res;
}
