/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlConst.h"



///  \file tlConst.h
///  \brief Adaptive constants based on Kalman filtering.

///  \class tlConst
///  \brief Adaptive constants based on Kalman filtering.
///
///  Tracking routines often relies on some "magic" constants that vary over time according
///  to external conditions, or that simply cannot be determined analytically. The tlConst
///  object is a Kalman-filter based estimator of constants that can be dynamically updated.
///  It provides an easy way to perform least-square approximation of
///  a set of data without memorization.



///  Constructor, initialize constant to 0.
///
///  \note
///  See \ref errors "error management" for details.

template < typename T >
tlConst < T >::tlConst ()
{
  val = pred = (T) 0;
  s_  = s    = 0.0f;
  e_  = e    = 0.0f;
  p   = 0.0f;
  m   = 1.0f;
  tol = -1;

  tl_errno = TL_NO_ERROR;
}



///  Constructor, initialize the most significant Kalman filter parameter (process covariance).
/// 
///  \param init    initial value of the constant
///  \param pCv     process covariance
///  \param tol     tolerance for input values [%],\n-1 for no input filtering
///
///  \note
///  See \ref errors "error management" for details.

template < typename T >
tlConst < T >::tlConst (T     init, 
                        float pCv, 
                        int   tol)
{
  val = pred = (T) init;
  s_  = s    = (float) init;
  e_  = e    = 1.0f;
  p   = pCv;
  m   = 1.0f;
  this->tol = tol;

  tl_errno = TL_NO_ERROR;
}



///  Constructor, initialize all Kalman filter parameters.
/// 
///  \param init    initial value of the constant
///  \param eCv     initial estimate covariance
///  \param pCv     process covariance
///  \param mCv     measurment covariance
///  \param tol     tolerance for input values [%],\n-1 for no input filtering
///
///  \note
///  See \ref errors "error management" for details.

template < typename T >
tlConst < T >::tlConst (T     init, 
                        float eCv, 
                        float pCv,
                        float mCv, 
                        int   tol)
{
  val = pred = (T) init;
  s_  = s    = (float) init;
  e_  = e    = eCv;
  p   = pCv;
  m   = mCv;
  this->tol = tol;

  tl_errno = TL_NO_ERROR;
}



///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

template < typename T >
tlConst < T >::~tlConst ()
{
  tl_errno = TL_NO_ERROR;
}



///  (Re)initialize the constant, that is, set the most significant 
///  Kalman filter parameter (process covariance).
/// 
///  \param init    initial value of the constant
///  \param pCv     process covariance
///  \param tol     tolerance for input values [%],\n-1 for no input filtering
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

template < typename T >
int tlConst < T >::init (T     init, 
                         float pCv, 
                         int   tol)
{
  val = pred = (T) init;
  s_  = s    = (float) init;
  e_  = e    = 1.0f;
  p   = pCv;
  m   = 1.0f;
  this->tol = tol;


  return 0;
}



///  (Re)initialize the constant, that is, set all Kalman filter parameters.
/// 
///  \param init    initial value of the constant
///  \param eCv     initial estimate covariance
///  \param pCv     process covariance
///  \param mCv     measurment covariance
///  \param tol     tolerance for input values [%],\n-1 for no input filtering
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

template < typename T >
int tlConst < T >::init (T     init, 
                         float eCv, 
                         float pCv,
                         float mCv, 
                         int tol)
{
  val = pred = (T) init;
  s_  = s    = (float) init;
  e_  = e    = eCv;
  p   = pCv;
  m   = mCv;
  this->tol = tol;


  return 0;
}



///  Update the constant with new data.
/// 
///  \param input   new input to the constant
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

template < typename T >
int tlConst < T >::update (T input)
{
  static float k;


  // check for tolerance
  if (tol != -1) {
    if (100 * (int) fabs ((double) (input - val)) > tol) {
      return 0;
    }
  }

  // compute blending factor
  k = e_ / (e_ + m);

  // compute current best state (in the least mean-square sense)
  s = s_ + k * ((float) input - s_);
  e = (1 - k) * e_;

  // compute estimate (next time step state prediction)
  s_ = s;
  e_ = e + p;

  // store and cast results
  val = (T) s;
  pred = (T) s_;


  return 0;
}



///  Set the current best estimate for the constant value;
///  the value is cast to the template type.
/// 
///  \param init      value to reset the constant to
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

template < typename T >
int tlConst < T >::setVal (T init)
{
  val = pred = (T) init;
  s_  = s = (float) init;


  return 0;
}



///  Set the most significant Kalman filter parameter (process covariance).
/// 
///  \param pCv     process covariance
///  \param tol     tolerance for input values [%],\n-1 for no input filtering
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

template < typename T >
int tlConst < T >::setParam (float pCv, 
                             int   tol)
{
  e_ = e = 1.0f;
  p  = pCv;
  m  = 1.0f;
  this->tol = tol;


  return 0;
}



///  Set all Kalman filter parameters.
/// 
///  \param eCv     initial estimate covariance
///  \param pCv     process covariance
///  \param mCv     measurment covariance
///  \param tol     tolerance for input values [%],\n-1 for no input filtering
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

template < typename T >
int tlConst < T >::setParam (float eCv, 
                             float pCv,
                             float mCv, 
                             int   tol)
{
  e_ = e = eCv;
  p  = pCv;
  m  = mCv;
  this->tol = tol;


  return 0;
}



///  Set the tolerance on input values.
///  Values that do are not within tolerance [%] of the current value
///  will not be taken into account.
/// 
///  \param tol     tolerance for input values [%]\n-1 disables input filtering
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

template < typename T >
int tlConst < T >::setTol (int tol)
{
  this->tol = tol;
  return 0;
}



///  \fn int tlConst<T>::getTol()
///  Return the tolerance filter for input values [%].
/// 
///  \return
///  Tolerance [%]\n
///  -1 means that no filtering is currently in use
///
///  \sa
///  setTol()



///  \fn int tlConst<T>::getVal()
///  Return the current best estimate for the constant value;
///  the value is cast to the template type.
/// 
///  \return
///  The current best estimate for the constant value.



///  \fn int tlConst<T>::getNext()
///  Return the current best prediction for the next constant value;
///  the value is cast to the template type.
/// 
///  \return
///  The current best prediction for the next constant value.



///  \fn template<typename T> float tlConst<T>::getState()
///  Return the actual best estimate for the constant value;
///  the value is NOT cast to the template type.
///
///  \return
///  the actual best estimate for the constant value
///
///  \remarks
///  The purpose of this routine is only to get access to untruncated
///  value of the constant.
///
///  \sa
///  getVal()



///  \fn float tlConst<T>::getEstimate()
///  Return the actual best prediction for the next constant value;
///  the value is NOT cast to the template type.
/// 
///  \return
///  The actual best prediction for the next constant value.
///
///  \remarks
///  the purpose of this routine is only to get access to untruncated
///  value of the prediction.
///
///  \sa
///  getNext()




// templates instantiation
// (required for library export)

template class tlConst < int >;
template class tlConst < float >;
template class tlConst < double >;
template class tlConst < char >;
template class tlConst < long >;
template class tlConst < short >;
template class tlConst < unsigned int >;
template class tlConst < unsigned char >;
template class tlConst < unsigned long >;
template class tlConst < unsigned short >;
