/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#define _USE_MATH_DEFINES
#include <math.h>
#ifndef M_PI 
#define M_PI 3.1415926535897932384626433832795
#endif
#include <memory.h>

#ifdef LINUX
#include <stdlib.h>
#endif

#include "tlConvolution.h"
#include "tlRect.h"



///  \file tlConvolution.h
///  \brief Internal convolution operators.
///  \internal



///  Perform linear convolution with any square kernel (with integer values).
///  
///  \param src           image pixel array
///  \param width         image width
///  \param height        image height
///  \param pixel_width   pixel width
///  \param kernel        convolution kernel (square)
///  \param kernel_width  width of the convolution kernel 
///  \param kernel_height height of the convolution kernel
///  \param rect          portion of the image to process
///  \param dest          \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution (tlPixel *src, 
                int     width, 
                int     height, 
                int     pixel_width,
                char    *kernel, 
                int     kernel_width, 
                int     kernel_height,
                tlRect  *rect, 
                tlPixel *dest)
{
  int i, j, k;
  int ii, jj;
  int index;
  int x1, x2, y1, y2;
  int kwmin, kwmax;
  int khmin, khmax;
  int sum[TL_MAX_PIXEL_WIDTH];


  // shortcuts
  kwmin = -kernel_width / 2;
  kwmax = kernel_width + kwmin;
  khmin = -kernel_height / 2;
  khmax = kernel_height + khmin;
  x1    = TL_MAX ((rect->x), -kwmin);
  x2    = TL_MIN (((rect->x) + (rect->width)), width - kwmax);
  y1    = TL_MAX ((rect->y), -khmin);
  y2    = TL_MIN (((rect->y) + (rect->height)), height - khmax);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0;
  }

  // proceed
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      index = pixel_width * (j * width + i);
      for (jj = khmin; jj < khmax; jj++) {
        for (ii = kwmin; ii < kwmax; ii++) {
          for (k = 0; k < pixel_width; k++) {
            sum[k] +=	kernel[(jj - khmin) * kernel_width + (ii - kwmin)] * src[pixel_width * ((j + jj) * width + (i + ii)) + k];
          }
        }
      }
      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (int) TL_MAX_PIXEL_VAL) sum[k] = (int) TL_MAX_PIXEL_VAL;
        if (sum[k] < (int) TL_MIN_PIXEL_VAL) sum[k] = (int) TL_MIN_PIXEL_VAL;
        dest[index + k] = (tlPixel) (sum[k]);
        sum[k] = 0;
      }
    }
  }


  return 0;
}



///  Perform linear convolution with any square kernel (with floating-point values).
///  
///  \param src           image pixel array
///  \param width         image width
///  \param height        image height
///  \param pixel_width   pixel width
///  \param kernel        convolution kernel (square)
///  \param kernel_width  width of the convolution kernel 
///  \param kernel_height height of the convolution kernel
///  \param rect          portion of the image to process
///  \param dest          \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution (tlPixel * src, 
                int width, 
                int height, 
                int pixel_width,
                float *kernel, 
                int kernel_width, 
                int kernel_height,
                tlRect * rect, 
                tlPixel * dest)
{
  int   i, j, k;
  int   ii, jj;
  int   index;
  int   x1, x2, y1, y2;
  int   kwmin, kwmax;
  int   khmin, khmax;
  float sum[TL_MAX_PIXEL_WIDTH];


  // shortcuts
  kwmin = -kernel_width / 2;
  kwmax = kernel_width + kwmin;
  khmin = -kernel_height / 2;
  khmax = kernel_height + khmin;
  x1    = TL_MAX ((rect->x), -kwmin);
  x2    = TL_MIN (((rect->x) + (rect->width)), width - kwmax);
  y1    = TL_MAX ((rect->y), -khmin);
  y2    = TL_MIN (((rect->y) + (rect->height)), height - khmax);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0.0f;
  }

  // proceed
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      index = pixel_width * (j * width + i);
      for (jj = khmin; jj < khmax; jj++) {
        for (ii = kwmin; ii < kwmax; ii++) {
          for (k = 0; k < pixel_width; k++) {
            sum[k] += kernel[(jj - khmin) * kernel_width + (ii - kwmin)] * src[pixel_width * ((j + jj) * width + (i + ii)) + k];
          }
        }
      }
      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (float) TL_MAX_PIXEL_VAL) sum[k] = (float) TL_MAX_PIXEL_VAL;
        if (sum[k] < (float) TL_MIN_PIXEL_VAL) sum[k] = (float) TL_MIN_PIXEL_VAL;
        dest[index + k] = (tlPixel) (sum[k]);
        sum[k] = 0.0f;
      }
    }
  } 


  return 0;
}



///  Perform linear convolution with any square kernel (with integer values);
///  this routine performs monodimensional convolution along all lines of an image with a separable kernel
///  
///  \param src image pixel array
///  \param width image width
///  \param height image height
///  \param pixel_width pixel width
///  \param kernel convolution kernel (line)
///  \param kernel_size size of the convolution kernel 
///  \param rect portion of the image to process
///  \param dest \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution_x (tlPixel *src, 
                  int     width, 
                  int     height, 
                  int     pixel_width,
                  char    *kernel, 
                  int     kernel_size, 
                  tlRect  *rect, 
                  int     *dest)
{
  int i, j, k;
  int ii;
  int index;
  int i_index, k_index;
  int x1, x2, y1, y2;
  int kmin, kmax;
  int sum[TL_MAX_PIXEL_WIDTH];


  // shortcuts
  kmin = -kernel_size / 2;
  kmax = kernel_size + kmin;
  x1   = TL_MAX ((rect->x), -kmin);
  x2   = TL_MIN (((rect->x) + (rect->width)), width - kmax);
  y1   = TL_MAX ((rect->y), 0);
  y2   = TL_MIN (((rect->y) + (rect->height)), height);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0;
  }

  // proceed
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      index = pixel_width * (j * width + i);
      i_index = index + pixel_width * kmin;
      k_index = 0;
      for (ii = kmin; ii < kmax; ii++) {
        for (k = 0; k < pixel_width; k++) {
          sum[k] += kernel[k_index] * src[i_index + k];
        }
        k_index++;
        i_index += pixel_width;
      }
      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (int) TL_MAX_PIXEL_VAL) sum[k] = (int) TL_MAX_PIXEL_VAL;
        if (sum[k] < (int) TL_MIN_PIXEL_VAL) sum[k] = (int) TL_MIN_PIXEL_VAL;
        dest[index + k] = (tlPixel) (sum[k]);
        sum[k] = 0;
      }
    }
  }


  return 0;
}



///  Perform linear convolution with any square kernel (with floating-point values);
///  this routine performs monodimensional convolution along all lines of an image with a separable kernel
///  
///  \param src image pixel array
///  \param width image width
///  \param height image height
///  \param pixel_width pixel width
///  \param kernel convolution kernel (line)
///  \param kernel_size size of the convolution kernel 
///  \param rect portion of the image to process
///  \param dest \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution_x (tlPixel *src, 
                  int     width, 
                  int     height, 
                  int     pixel_width,
                  float   *kernel, 
                  int     kernel_size, 
                  tlRect  *rect, 
                  float   *dest)
{
  int   i, j, k;
  int   ii;
  int   index;
  int   i_index, k_index;
  int   x1, x2, y1, y2;
  int   kmin, kmax;
  float sum[TL_MAX_PIXEL_WIDTH];


  // shortcuts
  kmin = -kernel_size / 2;
  kmax = kernel_size + kmin;
  x1   = TL_MAX ((rect->x), -kmin);
  x2   = TL_MIN (((rect->x) + (rect->width)), width - kmax);
  y1   = TL_MAX ((rect->y), 0);
  y2   = TL_MIN (((rect->y) + (rect->height)), height);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0.0f;
  }

  // proceed
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      index = pixel_width * (j * width + i);
      i_index = index + pixel_width * kmin;
      k_index = 0;
      for (ii = kmin; ii < kmax; ii++) {
        for (k = 0; k < pixel_width; k++) {
          sum[k] += kernel[k_index] * src[i_index + k];
        }
        k_index++;
        i_index += pixel_width;
      }
      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (float) TL_MAX_PIXEL_VAL) sum[k] = (float) TL_MAX_PIXEL_VAL;
        if (sum[k] < (float) TL_MIN_PIXEL_VAL) sum[k] = (float) TL_MIN_PIXEL_VAL;
        dest[index + k] = (tlPixel) (sum[k]);
        sum[k] = 0.0f;
      }
    }
  }


  return 0;
}



///  Perform linear convolution with any square kernel (with integer values);
///  this routine performs monodimensional convolution along all columns of an image with a separable kernel
///  
///  \param src           image pixel array
///  \param width         image width
///  \param height        image height
///  \param pixel_width   pixel width
///  \param kernel        convolution kernel (line)
///  \param kernel_size   size of the convolution kernel 
///  \param rect          portion of the image to process
///  \param dest          \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution_y (int     *src, 
                  int     width, 
                  int     height, 
                  int     pixel_width,
                  char    *kernel, 
                  int     kernel_size, 
                  tlRect  *rect,
                  tlPixel *dest)
{
  int i, j, k;
  int jj;
  int index;
  int i_index, k_index;
  int x1, x2, y1, y2;
  int kmin, kmax;
  int sum[TL_MAX_PIXEL_WIDTH];


  // shortcuts
  kmin = -kernel_size / 2;
  kmax = kernel_size + kmin;
  x1   = TL_MAX ((rect->x), 0);
  x2   = TL_MIN (((rect->x) + (rect->width)), width);
  y1   = TL_MAX ((rect->y), -kmin);
  y2   = TL_MIN (((rect->y) + (rect->height)), height - kmax);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0;
  }

  // proceed
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      index = pixel_width * (j * width + i);
      i_index = index + (pixel_width * width * kmin);
      k_index = 0;
      for (jj = kmin; jj < kmax; jj++) {
        for (k = 0; k < pixel_width; k++) {
          sum[k] += kernel[k_index] * src[i_index + k];
        }
        k_index++;
        i_index += pixel_width * width;
      }
      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (int) TL_MAX_PIXEL_VAL) sum[k] = (int) TL_MAX_PIXEL_VAL;
        if (sum[k] < (int) TL_MIN_PIXEL_VAL) sum[k] = (int) TL_MIN_PIXEL_VAL;
        dest[index + k] = (tlPixel) (sum[k]);
        sum[k] = 0;
      }
    }
  }


  return 0;
}



///  Perform linear convolution with any square kernel (with floating-point values);
///  this routine performs monodimensional convolution along all columns of an image with a separable kernel
///  
///  \param src image pixel array
///  \param width image width
///  \param height image height
///  \param pixel_width pixel width
///  \param kernel convolution kernel (line)
///  \param kernel_size size of the convolution kernel 
///  \param rect portion of the image to process
///  \param dest \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution_y (float   *src, 
                  int     width, 
                  int     height, 
                  int     pixel_width,
                  float   *kernel, 
                  int     kernel_size, 
                  tlRect  *rect,
                  tlPixel *dest)
{
  int   i, j, k;
  int   jj;
  int   index;
  int   i_index, k_index;
  int   x1, x2, y1, y2;
  int   kmin, kmax;
  float sum[TL_MAX_PIXEL_WIDTH];


  // shortcuts
  kmin = -kernel_size / 2;
  kmax = kernel_size + kmin;
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), -kmin);
  y2 = TL_MIN (((rect->y) + (rect->height)), height - kmax);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0.0f;
  }

  // proceed
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {

      // compute
      index = pixel_width * (j * width + i);
      i_index = index + (pixel_width * width * kmin);
      k_index = 0;
      for (jj = kmin; jj < kmax; jj++) {
        for (k = 0; k < pixel_width; k++) {
          sum[k] += kernel[k_index] * src[i_index + k];
        }
        k_index++;
        i_index += pixel_width * width;
      }

      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (float) TL_MAX_PIXEL_VAL) sum[k] = (float) TL_MAX_PIXEL_VAL;
        if (sum[k] < (float) TL_MIN_PIXEL_VAL) sum[k] = (float) TL_MIN_PIXEL_VAL;
        dest[index + k] = (tlPixel) (sum[k]);
        sum[k] = 0.0f;
      }
    }
  }


  return 0;
}



///  Compute gaussian kernel of a given size. The kernel is a line
///  and must be applied by using the convolution in separate mode.
///  \see tlPixelArray::convolution(float *xkernel, int xksize, 
///  float *ykernel, int yksize)
///  
///  \param ksize kernel size (line)
///  \param sigma desired sigma
///  \param kernel \b [out] destination gaussian kernel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \todo
///  check why output is different from Gaetan's version

int
tl_create_gaussian_kernel (int   ksize, 
                           float sigma, 
                           float *kernel)
{
  int   i,j;
  int   size        = ksize + ksize%2;
  int   halfsize    = (size-1)/2;
  float sigmasquare = -2*sigma*sigma;
  float sigmaroot   = (float)(sigma*sqrt(2.0f*M_PI));


  // odd size
  if (ksize%2) {
    for (i=-halfsize,j=0; i<=halfsize; i++,j++) {
      kernel[j] = (1/sigmaroot) * exp(i*i/sigmasquare);
    }
  }

  // even size
  else {
    for (i=-halfsize,j=0; i<=0; i++,j++) {
      kernel[size-1-j] = kernel[j] = (1/sigmaroot) * exp(i*i/sigmasquare);
    }
  }


  return size;
}



///  Perform linear convolution with any square kernel (with floating-point values);
///  this routine performs monodimensional convolution along all rows of an image with a separable kernel
///  
///  \param src image pixel array
///  \param width image width
///  \param height image height
///  \param pixel_width pixel width
///  \param kernel convolution kernel (line)
///  \param kernel_size size of the convolution kernel 
///  \param rect portion of the image to process
///  \param mask Mask the object of the src (must be of the same size of the source)
///  \param dest \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution_x (tlPixel *src, 
                  int     width, 
                  int     height, 
                  int     pixel_width,
                  float   *kernel, 
                  int     kernel_size, 
                  tlRect  *rect,
                  tlPixel *mask,
                  float   *dest)
{
  int   i, j, k;
  int   ii;
  int   i_index, k_index, m_index;
  int   x1, x2, y1, y2;
  int   kmin, kmax;
  float sum[TL_MAX_PIXEL_WIDTH];

  // shortcuts
  kmin = -kernel_size / 2;
  kmax = kernel_size + kmin;
  x1   = TL_MAX ((rect->x) -kmin, -kmin);
  x2   = TL_MIN (((rect->x) + (rect->width)) -kmax, width -kmax);
  y1   = TL_MAX ((rect->y), 0);
  y2   = TL_MIN (((rect->y) + (rect->height)), height);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0.0f;
  }

  // proceed
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {

      // mask index (TL_BIN_WIDTH)
      m_index = j*width + i;

      // if the current pixel of the mask is null (black) we skip it
      if (!mask[m_index])
        continue;

      for (ii=i+kmin, k_index=0; ii<i+kmax; ii++, k_index++) {
        i_index = pixel_width*(j*width + ii);
        // if one of the pixel covered by the kernel belongs to the inactive part
        // of the mask, the whole sum is set to null
        if (mask[j*width + ii]) {
          for (k = 0; k < pixel_width; k++)
            sum[k] += kernel[k_index] * src[i_index + k];
        }
        else {
          for (k = 0; k < pixel_width; k++)
            sum[k] = 0.0f;
          break;
        }
      }

      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (float) TL_MAX_PIXEL_VAL) sum[k] = (float) TL_MAX_PIXEL_VAL;
        if (sum[k] < (float) TL_MIN_PIXEL_VAL) sum[k] = (float) TL_MIN_PIXEL_VAL;
        dest[pixel_width * m_index + k] = sum[k];
        sum[k] = 0.0f;
      }
    }
  }


  return 0;
}




///  Perform linear convolution with any square kernel (with floating-point values);
///  this routine performs monodimensional convolution along all columns of an image with a separable kernel
///  
///  \param src image pixel array
///  \param width image width
///  \param height image height
///  \param pixel_width pixel width
///  \param kernel convolution kernel (line)
///  \param kernel_size size of the convolution kernel 
///  \param rect portion of the image to process
///  \param mask mask the object of the src (must be of the same size of the source)
///  \param dest \b [out] destination pixel array 
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_convolution_y (float   *src, 
                  int     width, 
                  int     height, 
                  int     pixel_width,
                  float   *kernel, 
                  int     kernel_size, 
                  tlRect  *rect,
                  tlPixel *mask,
                  tlPixel *dest)
{
  int   i, j, k;
  int   jj;
  int   i_index, k_index, m_index;
  int   x1, x2, y1, y2;
  int   kmin, kmax;
  float sum[TL_MAX_PIXEL_WIDTH];


  // shortcuts
  kmin = -kernel_size / 2;
  kmax = kernel_size + kmin;
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y) -kmin, -kmin);
  y2 = TL_MIN (((rect->y) + (rect->height)) -kmax, height -kmax);

  // init
  for (k = 0; k < pixel_width; k++) {
    sum[k] = 0.0f;
  }

  // proceed
  for (i = x1; i < x2; i++) {
    for (j = y1; j < y2; j++) {

      // mask index (TL_BIN_WIDTH)
      m_index = j*width + i;

      // if the current pixel of the mask is null (black) we skip it
      if (!mask[m_index])
        continue;

      for (jj=j+kmin, k_index=0; jj<j+kmax; jj++, k_index++) {
        i_index = pixel_width*(jj*width + i);
        // if one of the pixel covered by the kernel belongs to the inactive part
        // of the mask, the whole sum is set to null
        if (mask[jj*width + i]) {
          for (k = 0; k < pixel_width; k++)
            sum[k] += kernel[k_index] * src[i_index + k];
        }
        else {
          for (k = 0; k < pixel_width; k++)
            sum[k] = 0.0f;
          break;
        }
      }

      // store and reset
      for (k = 0; k < pixel_width; k++) {
        if (sum[k] > (float) TL_MAX_PIXEL_VAL) sum[k] = (float) TL_MAX_PIXEL_VAL;
        if (sum[k] < (float) TL_MIN_PIXEL_VAL) sum[k] = (float) TL_MIN_PIXEL_VAL;
        dest[pixel_width * m_index + k] = (tlPixel) (sum[k]);
        sum[k] = 0.0f;
      }
    }
  }


  return 0;
}
