/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlDigitizer.h"



///  \file tlDigitizer.h
///  \brief Digitizer base class.

///  \class tlDigitizer
///  \brief Image acquisition base class for hardware digitizers.
///
///  Simple image acquisition should be performed through tlDigitizer objects.
///  Details about the tlDigitizer object structure can be found in tlDigitgizer.h. Under Windows, TLIB
///  currently supports the DirectShow generic interface grabber that can be used with most firewire and USB webcams. 
///  Dedicated implementations of tlDigitizer are also available for the ImageNation PXC200 and the Data Translation DT3130 framegrabbers.
///  Under Linux, TLIB supports Video4Linux2 compatible devices.
///
///  \anchor inputFormat 
///  \par Input format
///  Input can be either \ref TL_SOURCE_COMPOSITE "composite" or \ref TL_SOURCE_SVIDEO "Svideo".\n
///  Some digitizers might support \ref TL_SOURCE_AUTO "auto-detection" of the input format.\n
///  The video encoding can be any of the following:
///  \li \ref TL_SOURCE_DIGITAL "digital input"
///  \li \ref TL_SOURCE_NTSC "NTSC"
///  \li \ref TL_SOURCE_NTSCJ "NTSCJ (Japanese)"
///  \li \ref TL_SOURCE_PAL "PAL (European)"
///  \li \ref TL_SOURCE_PALM "PALM"
///  \li \ref TL_SOURCE_PALN "PALN"
///  \li \ref TL_SOURCE_SECAM "SECAM (France)"
///
///  Some digitizer are also able to perform \ref TL_SOURCE_AUTO "auto-detection"
///  of the video format.\n
///
///  Video input format can be combined as follows when passed as argument to tlDigitizer::open() and tlDigitizer::setInput()
///  \code input = TL_SOURCE_NTSC | TL_SOURCE_SVIDEO; \endcode


///  \def TL_SOURCE_COMPOSITE
///  Composite video input.

///  \def TL_SOURCE_SVIDEO
///  SVideo video input.

///  \def TL_SOURCE_AUTO
///  Automatic input and/or video format configuration (if available).
///
///  \remark
///  This feature is hardware-specific. If a framegrabber
///  allocation fails, try to explicitly specify the video format.

///  \sa 
///  \ref inputFormat "input formats"

///  \def TL_SOURCE_DIGITAL
///  Digital video format.

///  \def TL_SOURCE_NTSC
///  NTSC (North America) video format.

///  \def TL_SOURCE_NTSCJ
///  NTSC-J (Japan) video format.

///  \def TL_SOURCE_PAL
///  PAL (Europe + Pacific) video format.

///  \def TL_SOURCE_PALM
///  PAL-M video format.

///  \def TL_SOURCE_PALN
///  PAL-N video format.

///  \def TL_SOURCE_SECAM
///  SECAM (France) video format.



///  Default constructor.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizer::tlDigitizer () : tlSource ()
{
  isOpen       = false;
  this->width  = TL_DEFAULT_COLS;
  this->height = TL_DEFAULT_ROWS;
  format       = RGB;
  input        = TL_SOURCE_AUTO;
  channel      = 0;
  fps          = TL_SOURCE_AUTO;
	tl_errno     = TL_NO_ERROR;
}



///  Default constructor.
///
///  \param width   acquisition width
///  \param height  acquisition height
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizer::tlDigitizer (int width,
                          int height)  : tlSource ()
{
  isOpen       = false;
  this->width  = width;
  this->height = height;
  format       = RGB;
  input        = TL_SOURCE_AUTO;
  channel      = 0;
  fps          = TL_SOURCE_AUTO;
	tl_errno     = TL_NO_ERROR;
}



///  Default destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlDigitizer::~tlDigitizer ()
{
	tl_errno = TL_NO_ERROR;
	return;
}



///  \fn int tlDigitizer::open (int index)
///  Open a digitizer for image acquisition using default settings.
///
///  \param index   \b [default=0] the index of the digitizer in the system 
///                 (for use of multiple digitizers at the same time)
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDigitizer::open (int width, int height, int index)
///  Open a digitizer for image acquisition using a given acquisition size.
///
///  \param width   new acquisition width
///  \param height  new acquisition height
///  \param index   \b [default=0] the index of the digitizer in the system 
///                 (for use of multiple digitizers at the same time)
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
int 
tlDigitizer::open (int width,
                   int height,
                   int index)
{ 
  TL_ERRNO_MSG ("tlDigitizer::open()", TL_ERROR_NOT_AVAILABLE); 
  return -1; 
}


///  \fn tlDigitizer::close ()
///  Clean up digitizer buffers and close hardware connection.
///  Configuration is retained, so that a call to open() will
///  restore the digitizer state to the last used mode.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlDigitizer::isOpen()
///  Return the current state of the digitizer.
///
///  \return
///  \e True if digitizer is currently open,\n
///  \e False otherwise.



///  \fn tlDigitizer::getChannel()
///  Get currently selected input channel.
///
///  \return
///  The input channel in use.



///  \fn tlDigitizer::getInput()
///  Get the input format used on the currently open channel.
///
///  \return
///  The \ref inputFormat "input format" in use.



///  Set the acquisition channel input format.
///    
///  \param input    input format
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  \ref tlDigitizer.h "TLIB input formats"

int 
tlDigitizer::setInput (unsigned short input)
{ 
  TL_ERRNO_MSG ("tlDigitizer::setVideoFormat()", TL_ERROR_NOT_AVAILABLE); 
  return -1; 
}



///  Set the acquisition channel.
///    
///  \param channel    channel index on the digitizer
///  \param input      input format
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  \ref tlDigitizer.h "TLIB input formats"

int 
tlDigitizer::setChannel (int            channel, 
                         unsigned short input)
{ 
  TL_ERRNO_MSG ("tlDigitizer::setInputChannel()", TL_ERROR_NOT_AVAILABLE); 
  return -1; 
}



///  Set the acquisition frame rate.
///    
///  \param fps     desired frames per second
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDigitizer::setFrameRate (int fps)
{ 
  TL_ERRNO_MSG ("tlDigitizer::setFrameRate()", TL_ERROR_NOT_AVAILABLE); 
  return -1;
}



///  \fn int tlDigitizer::setSize (int width, int height)
///  Set the digitizer acquisition size. All other
///  parameters are left untouched. Depending on the hardware,
///  this call might close and reopen the device, possibly leading to
///  some latency.
///
///  \param width   new acquisition width
///  \param height  new acquisition height
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlDigitizer::setSize (int width, 
                      int height)
{ 
  TL_ERRNO_MSG ("tlDigitizer::setSize()", TL_ERROR_NOT_AVAILABLE); 
  return -1; 
}
