/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlDisplay.h"



///  \file tlDisplay.h
///  \brief Image display and event management.

///  \class tlDisplay
///  \brief Image display and event management class.
///
///  Provides a window to render images, as well as simple event management 
///  (mouse and keyboard input).

///  \def TL_MAX_DISPLAY
///  Maximum number of displays opened simultaneously.

///  \def TL_DISPLAY_NAME
///  String that will be displayed on the window by default.  

///  \fn int tlDisplay::getID()
///  Return window ID tag.

///  \fn int tlDisplay::getWidth()
///  Return window height (displayable area without borders).

///  \fn int tlDisplay::getHeight()
///  Return window width (displayable area without borders).

///  \fn bool tlDisplay::getVisible()
///  Return window state (visible or hidden).

///  \def TL_NO_EVENT
///  No event occured.

///  \def TL_EVENT
///  Some event occured.

///  \def TL_EVENT_EXIT
///  Program exit event.

///  \def TL_EVENT_MOUSE
///  Mouse event

///  \def TL_EVENT_MOUSEDOWN
///  Mouse bouton pressed event.

///  \def TL_EVENT_MOUSEUP
///  Mouse button released event.

///  \def TL_EVENT_MOUSERECT
///  Mouse button released event after rectangle selection.

///  \def TL_EVENT_MOUSEDBLCLK
///  Mouse button double-clicked event.

///  \def TL_EVENT_KEY
///  Keyboard event.

///  \def TL_EVENT_KEYDOWN
///  Keyboard key pressed event.

///  \def TL_EVENT_KEYUP
///  Keyboard key released event.



///  \fn tlDisplay::tlDisplay  ()
///  Default constructor.
/// 
///  \note
///  See \ref errors "error management" for details.



///  \fn tlDisplay::tlDisplay  (char *name)
///  Constructor.
/// 
///  \param name        window name (displayed in title bar)
/// 
///  \note
///  See \ref errors "error management" for details.



///  \fn tlDisplay::tlDisplay  (int x, int y)
///  Constructor, specifies window position on the screen.
/// 
///  \param x       initial window X position  [pix]
///  \param y       initial window Y position  [pix]
/// 
///  \note
///  See \ref errors "error management" for details.



///  \fn tlDisplay::tlDisplay  (char *name, int  x, int  y)
///  Constructor, specifies both position and window name.
/// 
///  \param name    window name (displayed in title bar)
///  \param x       initial window X position  [pix]
///  \param y       initial window Y position  [pix]
/// 
///  \note
///  See \ref errors "error management" for details.



///  \fn tlDisplay::~tlDisplay ()
///  
///  Destructor.
/// 
///  \note
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::update ()
///  internal method used to update window size and position following user interaction.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::setPos (int x, int y)
///  Reposition the window and the display area.
/// 
///  \param x    new display area (upper-left corner) X position
///  \param y    new display area (upper-left corner) Y position
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::resize (int width, int height)
///  Resize the window and the display area.
/// 
///  \param width     new display area width
///  \param height    new display area height
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::close ()
///  Close and destroys the window.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::show ()
///  Make window visible.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::hide ()
///  Make window invisible.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::display (tlHist *hist)
///  Display a histogram in the display area.
///
///  \param hist     the histogram to display
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::display (tlHist2D *hist)
///  Display a 2D histogram in the display area.
/// 
///  \param hist     the 2D histogram to display
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn int tlDisplay::queryEvent (void)
///  Check if an event has occured and is in the event queue.
/// 
///  \return
///  TL_EVENT if an event occured,\n
///  TL_NO_EVENT otherwise.


///  \fn int tlDisplay::event (void)
///  Process TL library windows events;
///  result is stored in TL specific data structure.
///  This is a blocking function.
/// 
///  \return
///  TL_EVENT status (see tlDisplay.h for details)



///  \fn int tlDisplay::getLastEvent ()
///  Return last received event. Is set to TL_NO_EVENT once called.
/// 
///  \return
///  TL_EVENT status (see tlDisplay.h for details)\n
///  TL_NO_EVENT if there is no new event



///  \fn tlDisplay* tlDisplay::getLastDisplay ()
///  Return a pointer to the display that generated the last processed event.
/// 
///  \return
///  a pointer to the display that received the last event (as given by tl_event_get_last),\n
///  NULL if no display is associated with the event.



///  \fn tlPoint* tlDisplay::getMouse ()
///  Get mouse click coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \return
///  A pointer to a copy of a tlPoint object.
/// 
///  \note
///  It is the responsibility of the programmer to delete the point
///  after use.
/// 
///  \sa
///  tlDisplay::getMouse (tlPoint *point)




///  \fn tlDisplay::getMouse (tlPoint *point)
///  Get mouse click coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \param point    \b [out] a valid pointer to a point structure to receive the coordonates.
/// 
///  \return
///  A pointer to a \ref tlPoint object.



///  \fn tlPoint* tlDisplay::getMouseDown ()
///  Get mouse coordonates when button is pressed. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \return
///  A pointer to a copy of a tlPoint object.
/// 
///  \note
///  It is the responsibility of the programmer to delete the point
///  after use.
/// 
///  \sa
///  tlDisplay::getMouseDown (tlPoint *point)



///  \fn tlDisplay::getMouseDown (tlPoint *point)
///  Get mouse coordonates when button is pressed. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \param point    \b [out] a valid pointer to a point structure to receive the coordonates.
/// 
///  \return
///  A pointer to a \ref tlPoint object.



///  \fn tlPoint* tlDisplay::getMouseUp ()
///  Get mouse coordonates when button is released. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \return
///  A pointer to a copy of a tlPoint object.
/// 
///  \note
///  It is the responsibility of the programmer to delete the point
///  after use.
/// 
///  \sa
///  tlDisplay::getMouseUp (tlPoint *point)



///  \fn tlDisplay::getMouseUp (tlPoint *point)
///  Get mouse coordonates when button is released. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \param point    \b [out] a valid pointer to a point structure to receive the coordonates.
/// 
///  \return
///  A pointer to a \ref tlPoint object.



///  \fn tlPoint* tlDisplay::getMouseDouble ()
///  Get mouse double click coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \return
///  A pointer to a copy of a tlPoint object.
/// 
///  \note
///  It is the responsibility of the programmer to delete the point after use.
/// 
///  \sa
///  tlDisplay::getMouseDouble (tlPoint *point)



///  \fn tlDisplay::getMouseDouble (tlPoint *point)
///  Get mouse double click coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \param point    \b [out] a valid pointer to a point structure to receive the coordonates.
/// 
///  \return
///  A pointer to a \ref tlPoint object.



///  \fn tlRect* tlDisplay::getMouseRect ()
///  Get mouse defined rectangle coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \return
///  A pointer to a copy of a tlRect object.
/// 
///  \note
///  It is the responsibility of the programmer to delete the rect after use.
/// 
///  \sa
///  tlDisplay::getMouseRect (tlRect *rect)



///  \fn tlDisplay::getMouseRect (tlRect *rect)
///  Get mouse defined rectangle coordonates. User can be in *any* TL window. 
///  Use tlDisplay::getLastDisplay() to retrieve the window 
///  that currently "has" the mouse.
/// 
///  \param rect    \b [out] a valid pointer to a point structure to receive the coordonates.
/// 
///  \return
///  A pointer to a \ref tlRect object.



///  \fn char tlDisplay::getKeyDown ()
///  Get keycode for keyboard downstroke.
/// 
///  \return
///  The character ASCII code.



///  \fn char tlDisplay::getKeyUp ()
///  Get keycode for keyboard uptroke.
/// 
///  \return
///  The character ASCII code.



///  \fn int tlDisplay::setMousePos (int x, int y)
///  Warp the mouse pointer to an absolute screen position. 
///  (x, y) = (0, 0) = top-left corner
/// 
///  \param x    absolute screen X position
///  \param y    absolute screen Y position
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn char tlDisplay::waitKey ()
///  Wait for a key to be pressed on the keyboard.
/// 
///  \return
///  The character entered by the user.



///   \fn void tlDisplay::waitMouse ()
///   Wait for the user to click in a window.
