/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlEdges.h"
#include "tlRect.h"



///  \file tlEdges.h
///  \brief Internal edges operators.
///  \internal
///
///  \todo
///  implement Canny edge detector.

///  \def TL_EDGES_BIN
///  Method tag for edges extraction.\n
///  BIN image edge extraction.
///
///  \sa
///  tlImage::edges()

///  \def TL_EDGES_MORPH
///  Method tag for edges extraction.\n
///  Morphological-based edge extraction.
///
///  \sa
///  tlImage::edges()

///  \def TL_EDGES_SOBEL
///  Method tag for edges extraction.\n
///  Sobel convolution edge extraction.
///
///  \sa
///  tlImage::edges()

///  \def TL_EDGES_PREWITT
///  Method tag for edges extraction.\n
///  Prewitt convolution image edge extraction.
///
///  \sa
///  tlImage::edges()



///  Binary image edge detection.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the edge map
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_edges_bin (tlPixel *src, 
              int     width, 
              int     height, 
              tlRect  *rect,
              tlPixel *dest)
{
  register int index;

  int i, j;
  int x1, x2;
  int y1, y2;


  // check
  if ((!src) || (!dest)) {
    TL_ERRNO_MSG ("tl_rgb2hsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check rect
  if (rect == NULL) {
    // shortcuts
    x1 = 0;
    y1 = 0;
    x2 = width;
    y2 = height;
  }
  else {
    // shortcuts
    x1 = TL_MAX ((rect->x), 0);
    x2 = TL_MIN (((rect->x) + (rect->width)), width) - 1;
    y1 = TL_MAX ((rect->y), 0);
    y2 = TL_MIN (((rect->y) + (rect->height)), height) - 1;
  }

  // upper limit
  index = TL_PIXEL (y1, (x1 + 1), width);
  for (i = x1 + 1; i < (x2); i++) {
    if (src[index] != 0) {
      if ((src[index - 1] == 0) || (src[index + 1] == 0) || (src[index + width] == 0)) dest[index] = TL_MAX_PIXEL_VAL;
      else dest[index] = 0;
    }
    else dest[index] = 0;
    index++;
  }

  // lower limit
  index = TL_PIXEL (y2, (x1 + 1), width);
  for (i = x1 + 1; i < (x2); i++) {
    if (src[index] != 0) {
      if ((src[index - 1] == 0) || (src[index + 1] == 0) || (src[index - width] == 0)) dest[index] = TL_MAX_PIXEL_VAL;
      else dest[index] = 0;
    }
    else dest[index] = 0;
    index++;
  }

  // left edge
  index = TL_PIXEL ((y1 + 1), x1, width);
  for (j = (y1 + 1); j < y2; j++) {
    if (src[index] != 0) {
      if ((src[index + 1] == 0) || (src[index - width] == 0) || (src[index + width] == 0)) dest[index] = TL_MAX_PIXEL_VAL;
      else dest[index] = 0;
    }
    else dest[index] = 0;
    index += width;
  }

  // right edge
  index = TL_PIXEL ((y1 + 1), x2, width);
  for (j = (y1 + 1); j < y2; j++) {
    if (src[index] != 0) {
      if ((src[index - 11] == 0) || (src[index - width] == 0) || (src[index + width] == 0)) dest[index] = TL_MAX_PIXEL_VAL;
      else dest[index] = 0;
    }
    else dest[index] = 0;
    index += width;
  }

  // process middle area
  for (j = y1 + 1; j < (y2 - 1); j++) {
    index = TL_PIXEL (j, (x1 + 1), width);
    for (i = x1 + 1; i < (x2 - 1); i++) {
      if (src[index] != 0) {
        if (
          (src[index - 1]     == 0) || (src[index + 1]     == 0) || 
          (src[index - width] == 0) || (src[index + width] == 0))	
          dest[index] = TL_MAX_PIXEL_VAL;
        else 
          dest[index] = 0;
      }
      else dest[index] = 0;
      index++;
    }
  }


  return TL_NO_ERROR;
}



///  Vertical image edge detection using the Sobel operator.
///  
///  \param src         pixel array containing image to convert
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the edge map
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_edges_vert_sobel (tlPixel *src, 
                     int     width, 
                     int     height, 
                     int     pixel_width,
                     tlRect  *rect, 
                     tlPixel *dest)
{
  int  res;
  char kernel_x[3] = { 1, 2, 1 };
  char kernel_y[3] = { -1, 0, 1 };
  int  *tmp;


  tmp = (int *) malloc (width * height * pixel_width * sizeof (int));
  res =  tl_convolution_x (src, width, height, pixel_width, kernel_x, 3, rect, tmp);
  if (res == 0) {
    res = tl_convolution_y (tmp, width, height, pixel_width, kernel_y, 3, rect, dest);
  }


  free (tmp);
  return res;
}



///  Horizontal image edge detection using the Sobel operator.
///  
///  \param src           pixel array containing image to convert
///  \param width         image width 
///  \param height        image height 
///  \param pixel_width   pixel width
///  \param rect          portion of the image to process
///  \param dest          \b [out] pixel array containing the edge map
///  
///  \return
///  0 on success, -1 otherwise. 
///  See \ref errors "error management" for details.

int
tl_edges_hor_sobel (tlPixel *src, 
                    int     width, 
                    int     height, 
                    int     pixel_width,
                    tlRect  *rect, 
                    tlPixel *dest)
{
  int  res;
  char kernel_x[3] = { -1, 0, 1 };
  char kernel_y[3] = { 1, 2, 1 };
  int  *tmp;


  tmp = (int *) malloc (width * height * pixel_width * sizeof (int));
  res = tl_convolution_x (src, width, height, pixel_width, kernel_x, 3, rect, tmp);
  if (res == 0) {
    res = tl_convolution_y (tmp, width, height, pixel_width, kernel_y, 3, rect,	dest);
  }


  free (tmp);
  return res;
}



///  Vertical image edge detection using the Prewitt operator.
///  
///  \param src         pixel array containing image to convert
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the edge map
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_edges_vert_prewitt (tlPixel *src, 
                       int     width, 
                       int     height, 
                       int     pixel_width,
                       tlRect  *rect, 
                       tlPixel *dest)
{
  int  res;
  char kernel_x[3] = { 1, 1, 1 };
  char kernel_y[3] = { 1, 0, -1 };
  int  *tmp;


  tmp = (int *) malloc (width * height * pixel_width * sizeof (int));
  res =  tl_convolution_x (src, width, height, pixel_width, kernel_x, 3, rect,  tmp);
  if (res == 0) {
    res = tl_convolution_y (tmp, width, height, pixel_width, kernel_y, 3, rect,	dest);
  }


  free (tmp);
  return res;
}



// Horizontal image edge detection using the Prewitt operator.
// 
// \param src         pixel array containing image to convert
// \param width       image width 
// \param height      image height 
// \param pixel_width pixel width
// \param rect        portion of the image to process
// \param dest        \b [out] pixel array containing the edge map
// 
// \return
// 0 on success, -1 otherwise.\n
// See \ref errors "error management" for details.

int
tl_edges_hor_prewitt (tlPixel *src, 
                      int     width, 
                      int     height, 
                      int     pixel_width,
                      tlRect  *rect, 
                      tlPixel *dest)
{
  int  res;
  char kernel_x[3] = { 1, 0, -1 };
  char kernel_y[3] = { 1, 1, 1 };
  int  *tmp;


  tmp = (int *) malloc (width * height * pixel_width * sizeof (int));
  res = tl_convolution_x (src, width, height, pixel_width, kernel_x, 3, rect,  tmp);
  if (res == 0) {
    res = tl_convolution_y (tmp, width, height, pixel_width, kernel_y, 3, rect, dest);
  }


  free (tmp);
  return res;
}
