/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlError.h"
#include <string.h>

#ifdef LINUX
#include <pthread.h>
#endif



///  \file tlError.h
///  \brief Thread-safe error management.
///
///  \anchor errors \b Error \b management
///
///  TLIB uses a thread-safe global, \c tl_errno, to store the last error that occured in the thread currently executing.
///  Most functions and methods will return either 0 or a valid, positive value on success, and -1 (or NULL) in case of
///  failure. On failure, programmers can check the value of \c tl_errno against the \ref tl_errors "error values".\n\n
///  To help identify the errror, the \ref tlError() functions return a string that describes the error. If tl_set_verbose() 
///  is enabled, TLIB will spontaneously produce error messages to the standard output as they occur.\n\n
///  An important feature of the TLIB error management is its ability to detect errors in constructors/destructors.
///  Every constructor and destructore of TLIB classes sets the \c tl_errno global to TL_NO_ERROR on success, or an
///  \ref tl_errors "error value" otherwise.\n

///  \enum tl_errors
///  Valid error values used by all TLIB methods and functions.

///  \var tl_errors TL_NO_ERROR
///  Indicates that no error occured during processing. This is the default
///  return value for most TLIB functions.

///  \var tl_errors TL_ERROR
///  Indicates that a generic error occured.

///  \var tl_errors TL_ERROR_INVALID_FORMAT
///  Indicates that an operation cannot be performed on the given image or object format.
///  Either the \ref imageformats "image format" conflicts with another argument, or the 
///  operation cannot be performed on this particular format. In the case of tlDigitizer,
///  it also indicates that a particular acquisition format is not available.

///  \var tl_errors TL_ERROR_GEOMETRY
///  Indicates that there is a geometric mismatch between the entities being processed.
///  This typically happens when two TLIB objects with different sizes are applied a
///  transformation.

///  \var tl_errors TL_ERROR_NULL_POINTER
///  Indicates that an invalid pointer was used.
///  This is typically returned when a NULL argument is given as an argument to a TLIB method.

///  \var tl_errors TL_ERROR_INDEX_OUT_OF_BOUNDS
///  Indicates that an array is being referenced at a position bigger than its size.

///  \var tl_errors TL_ERROR_MAX_REACHED
///  Indicates that some data has reached its maximum allowed value. This can be the number of
///  physical devices opened at the same time, or the maximume number of features detected in a
///  blob extraction performed by tlBlob.

///  \var tl_errors TL_ERROR_ALLOCATION_FAILED
///  Indicates that memory allocations failed.

///  \var tl_errors TL_ERROR_NOT_IMPLEMENTED
///  Indicates that a method or function is not currently implemented.

///  \var tl_errors TL_ERROR_FILE_IO
///  Indicates that an error occured in a file read/write operation.

///  \var tl_errors TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT
///  Indicates that a necessary file (an image file, a library or a configuration file)
///  is either missing or corrupt. Check the filename.

///  \var tl_errors TL_ERROR_NOT_REQUIRED
///  Indicates that a given operation will leave the data unmodified.
///  While it is more a warning than an actual error, it should be avoided
///  in real-time vision programming.

///  \var tl_errors TL_ERROR_NOT_FOUND
///  Indicates that an object or feature was not found by the correlation routines.

///  \var tl_errors TL_ERROR_NOT_AVAILABLE
///  Indicates that an operation or feature is not available in the current context.
///  This usually means that the function is not available for a particular implementation of
///  a hardware interface.

///  \var tl_errors TL_ERROR_NOT_IN_IMAGE
///  Indicates that a point is not within the image boundaries. This error can be safely ignored
///  in most TLIB high-level methods, as these check the validity of all arguments before applying them.

///  \var tl_errors TL_ERROR_STACK_OVERFLOW
///  Indicates that a stack overflow exception was caught. This error typically
///  happen during a recursive call to the object recursive extraction routines (see tlObject).
///  It is possible to prevent it by asking your compiler to increase the default stack size.

///  \var tl_errors TL_ERROR_TIMEOUT
///  Indicates that an operation timed out. This is usually thread related (see tlThread.h)

///  \var tl_errors TL_ERROR_THREAD
///  Indicates a thread related error (see tlThread.h)

///  \var tl_errors TL_ERROR_TIMER
///  Indicates a timer related error (see tlTimer.h).

///  \var tl_errors TL_ERROR_RESOURCE_NOT_AVAILABLE 
///  Indicates that a hardware resource is not available. This can mean that the device is not
///  present, is already opened, or that there is a conflict.

///  \var tl_errors TL_ERROR_CONFIG_FAILED
///  Indicates that a hardware resource could not be configured with the given parameters.

///  \var tl_errors TL_ERROR_GRAB_FAILED 
///  Indicates that image acquisition failed. Depending on the source, it could mean that the hardware
///  was not open or has been disconnected, or that an image (or video) file is missing.

///  \var tl_errors TL_ERROR_NA
///  Indicates that the requested operation cannot be applied to the given arguments.

///  \def TL_ERRNO_MSG(msg,errno)
///  Sets the \c tl_errno global and displays an error message to the standard output if tl_set_verbose() is enabled.



static char *tlErrorStr[24] = 
{
  "no error has occured yet",                   // TL_NO_ERROR
  "undocumented error",                         // TL_ERROR
  "invalid format",                             // TL_ERROR_INVALID_FORMAT
  "invalid geometry",                           // TL_ERROR_GEOMETRY
  "null pointer",                               // TL_ERROR_NULL_POINTER
  "index out of bounds",                        // TL_ERROR_INDEX_OUT_OF_BOUNDS
  "max value reached",                          // TL_ERROR_MAX_REACHED
  "memory allocation failed",                   // TL_ERROR_ALLOCATION_FAILED
  "not implemented",                            // TL_ERROR_NOT_IMPLEMENTED
  "file I/O error",                             // TL_ERROR_FILE_IO
  "file not found",                             // TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT
  "operation not required",                     // TL_ERROR_NOT_REQUIRED
  "object or shape not found",                  // TL_ERROR_NOT_FOUND
  "operation not available",                    // TL_ERROR_NOT_AVAILABLE
  "location not within image boundaries",       // TL_ERROR_NOT_IN_IMAGE
  "stack overflow"                              // TL_ERROR_STACK_OVERFLOW
  "operation timed out",                        // TL_ERROR_TIMEOUT
  "thread-related error",                       // TL_ERROR_THREAD
  "timer-related error",                        // TL_ERROR_TIMER
  "resource not available or busy",             // TL_ERROR_RESOURCE_NOT_AVAILABLE
  "configuration failed",                       // TL_ERROR_CONFIG_FAILED
  "image acquisition failed",                   // TL_ERROR_GRAB_FAILED
  "operation not supported for high-res image", // TL_ERROR_HIGHRES_INCOMPATIBLE
  "operation not applicable"                    // TL_ERROR_NA
};



// Windows thread keys 
#ifdef WIN32
static DWORD tl_errno_key = 0xFFFFFFFF;
#endif

// Linux thread keys 
#ifdef LINUX
static pthread_key_t tl_errno_key;
#  ifdef PTHREAD_ONCE_INIT
pthread_once_t _tl_errno_once = PTHREAD_ONCE_INIT;
#  else
#    ifdef pthread_once_init
pthread_once_t _tl_errno_once = pthread_once_init;
#    else
pthread_once_t _tl_errno_once = 0;
#    endif
#  endif
#endif


///  Display an error message corresponding to a given error. The error
///  is printed on the standard output, preceded by the str argument.
///
///  \param errno    the error number 
///  \param str      preceding string

void
tlError (int        errno, 
         const char *str)
{
  if (str == NULL) {
    return;
  }

  if (strlen (str) > 1) printf ("%s: %s\n", str, tlErrorStr[errno]);
  else                  printf ("TLIB: %s\n", tlErrorStr[errno]);
}



///  Display an error message corresponding to the last error that 
///  occured in the calling thread. The error is printed on the standard 
///  output, preceded by the str argument.
///
///  \param str      preceding string

void
tlError (const char *str)
{
  if (strlen (str) > 1) printf ("%s: %s\n", str, tlErrorStr[tl_errno]); 
  else                  printf ("TLIB: %s\n", tlErrorStr[tl_errno]);
}



///  Display an error message corresponding to a given error. The error
///  is printed on the standard output.
///
///  \param errno    the error number 

void
tlError (int errno)
{
  printf ("TLIB: %s\n", tlErrorStr[errno]);
}



///  Display an error message corresponding to the last error that 
///  occured in the calling thread.
///  The message is printed on the standard output.

void
tlError ()
{
  printf ("TLIB: %s\n", tlErrorStr[tl_errno]);
}


///
/// \internal



///  Function that returns the address of the tl_errno variable.

int *
tl_errno_location ()
{
  int *output;


#ifdef WIN32
  if (tl_errno_key == 0xFFFFFFFF)	tl_errno_once ();
  output = (int *) TlsGetValue (tl_errno_key);
  if (output == NULL) {
    output = (int *) calloc (1, sizeof (int));
    TlsSetValue (tl_errno_key, output);
  }
#endif

#ifdef LINUX
  pthread_once (&_tl_errno_once, tl_errno_once);
  output = (int *) pthread_getspecific (tl_errno_key);
  if (output == NULL) {
    output = (int *) calloc (1, sizeof (int));
    pthread_setspecific (tl_errno_key, output);
  }
#endif


  return (output);
}



/// errno initializer

void
tl_errno_once ()
{

#ifdef WIN32
  tl_errno_key = TlsAlloc ();
#endif

#ifdef LINUX
  pthread_key_create (&tl_errno_key, tl_errno_destructor);
#endif
}



/// errno dectructor

void
tl_errno_destructor (void *ptr)
{
#ifdef WIN32
  ptr = TlsGetValue (tl_errno_key);
  if (ptr != NULL) free (ptr);
#endif

#ifdef LINUX
  if (ptr != NULL) free (ptr);
#endif
}
