/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#ifdef LINUX
#include <stdlib.h>
#endif

#include "tlExtract.h"



///  \file tlExtract.h
///  \brief Internal segmentation operators.
///  \internal

///  \def TL_EXTRACT_DATA_DEPTH
///  Maximum image depth (i.e. channels such as R,G,B) that is supported by the extraction routine. 




///  Recursive call to filter pixel values
///  in an image with arbitrary depth.
///  
///  \param data  structure containing the search params
///  
///  \return
///  0 on success, -1 otherwise. 
///  See \ref errors "error management" for details.

int
tl_extract (tl_extract_data *data)
{
  int     k;
  int     px          = data->px;
  int     py          = data->py;
  int     width       = data->width;
  int     map_index   = py * width + px;
  int     index       = sizeof (tlPixel) * map_index;
  int     depth       = data->depth;
  tlPixel delta;
  tlPixel *val;
  tlPixel min[TL_EXTRACT_DATA_DEPTH];
  tlPixel max[TL_EXTRACT_DATA_DEPTH];
  tlPixel tol[TL_EXTRACT_DATA_DEPTH];


  // check for point validity
  if ((px < (data->lx1)) || (px >= (data->lx2)) || (py < (data->ly1)) || (py >= (data->ly2))) {
    return 0;
  }

  // shortcuts
  val = &(data->src[(data->depth) * index]);
  for (k = 0; k < depth; k++) {
    min[k] = (tlPixel)data->min[k];
    max[k] = (tlPixel)data->max[k];
    tol[k] = (tlPixel)data->tol[k];
  }

  // test current point
  if ((data->objmap[map_index]) != -1) return 0;

  // test point
  for (k = 0; k < depth; k++) {
    if ((val[k] < (min[k] - tol[k])) || (val[k] > (max[k] + tol[k]))) {
      data->objmap[map_index] = 0;
      return 0;
    }
  }

  // 
  // valid point
  // 

  // tag
  data->objmap[map_index] = 1;

  // update boundaries
  if (px < *(data->x1)) *(data->x1) = px;
  if (px > *(data->x2))	*(data->x2) = px;
  if (py < *(data->y1))	*(data->y1) = py;
  if (py > *(data->y2))	*(data->y2) = py;

  // recenter validity gap
  for (k = 0; k < depth; k++) {
    if (val[k] < min[k]) {
      delta = (tlPixel)(min[k] - val[k]);
      data->min[k] = (tlPixel)(TL_MAX (0, (min[k] - delta)));
      data->max[k] = (tlPixel)(TL_MAX (0, (max[k] - delta)));
    }
    if (val[k] > max[k]) {
      delta = (tlPixel)(val[k] - max[k]);
      data->min[k] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (min[k] + delta)));
      data->max[k] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (max[k] + delta)));
    }
  }

  // 
  // recurse (reset to current boundary between each call)
  // 

  // to the right
  data->px = px + 1;
  tl_extract (data);
  for (k = 0; k < depth; k++) {
    data->min[k] = min[k];
    data->max[k] = max[k];
  }

  // to the left
  data->px = px - 1;
  tl_extract (data);
  for (k = 0; k < depth; k++) {
    data->min[k] = min[k];
    data->max[k] = max[k];
  }

  // recenter
  data->px = px;

  // up
  data->py = py + 1;
  tl_extract (data);
  for (k = 0; k < depth; k++) {
    data->min[k] = min[k];
    data->max[k] = max[k];
  }

  // down
  data->py = py - 1;
  tl_extract (data);
  for (k = 0; k < depth; k++) {
    data->min[k] = min[k];
    data->max[k] = max[k];
  }

  // recenter
  data->py = py;

  // case of 8 connect object extraction
  if (data->connectivity == TL_CONNEX_8) {

    // up right
    data->px = px + 1;
    data->py = py + 1;
    tl_extract (data);
    for (k = 0; k < depth; k++) {
      data->min[k] = min[k];
      data->max[k] = max[k];
    }

    // down right
    data->px = px + 1;
    data->py = py - 1;
    tl_extract (data);
    for (k = 0; k < depth; k++) {
      data->min[k] = min[k];
      data->max[k] = max[k];
    }

    // up left
    data->px = px - 1;
    data->py = py + 1;
    tl_extract (data);
    for (k = 0; k < depth; k++) {
      data->min[k] = min[k];
      data->max[k] = max[k];
    }

    // down left
    data->px = px - 1;
    data->py = py - 1;
    tl_extract (data);
    for (k = 0; k < depth; k++) {
      data->min[k] = min[k];
      data->max[k] = max[k];
    }

    // recenter
    data->px = px;
    data->py = py;
  }


  return 0;
}



///  Non-recursive call to filter pixel values
///  in an image with arbitrary depth.
///  
///  \param data structure containing the search params
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_nr (tl_extract_data *data)
{
  int     i, j, k;
  int     x               = data->px;
  int     y               = data->py;
  int     x1, x2, y1, y2;
  int     width           = data->width;
  int     index           = sizeof (tlPixel) * (y * width + x);
  int     depth           = data->depth;
  char    *objmap         = data->objmap;
  int     *minmap;
  int     *maxmap;
  tlPixel *val;
  tlPixel delta;
  tlPixel tol[TL_EXTRACT_DATA_DEPTH];


  // shortcuts
  val = &(data->src[(data->depth) * index]);
  for (k = 0; k < depth; k++) {
    tol[k] = (tlPixel)data->tol[k];
  }
  x1 = data->lx1;
  x2 = data->lx2;
  y1 = data->ly1;
  y2 = data->ly2;

  // initialize maps
  minmap = (int *) malloc (depth * width * (data->height) * sizeof (int));
  maxmap = (int *) malloc (depth * width * (data->height) * sizeof (int));
  for (j = y1; j < y2; j++) {
    index = depth * (j * width + x1);
    for (i = x1; i < x2; i++) {
      minmap[index] = -1;
      maxmap[index] = -1;
      index += depth;
    }
  }

  // initialize filter values
  index = depth * (y * width + x);
  for (k = 0; k < depth; k++) {
    minmap[index + k] = data->min[k];
    maxmap[index + k] = data->max[k];
  }

  // 
  // spiral browsing of rectangular surface
  // 

  int ii, jj;
  int l_flag, r_flag, t_flag, b_flag;
  int cur_incr, i_incr, j_incr;
  int i_act, j_act;
  int ii_act, jj_act;
  int map_index, obj_index;
  int valid;

  i = x;
  j = y;
  l_flag = r_flag = t_flag = b_flag = 0;
  cur_incr = i_incr = j_incr = 0;
  while ((!l_flag) || (!r_flag) || (!t_flag) || (!b_flag)) {

    // 
    // perform check for local point
    // 

    i_act = TL_MAX (i, x1);
    i_act = TL_MIN (i_act, x2);
    j_act = TL_MAX (j, y1);
    j_act = TL_MIN (j_act, y2);

    // compute index
    index = j_act * width + i_act;

    // for each neighbour
    for (jj = -1; jj < 2; jj++) {
      for (ii = -1; ii < 2; ii++) {

        // get safe pixel address
        ii_act = TL_MAX ((i + ii), x1);
        ii_act = TL_MIN (ii_act, x2);
        jj_act = TL_MAX ((j + jj), y1);
        jj_act = TL_MIN (jj_act, y2);

        // don't test current pixel itself
        if ((ii_act == 0) && (jj_act == 0))
          continue;

        // compute index and get value
        obj_index = jj_act * width + ii_act;
        map_index = depth * obj_index;

        // test validity
        if (objmap[obj_index] <= 0)
          continue;

        // test value
        for (k = 0, valid = 1; (k < depth) && (valid); k++) {
          if ((val[k] < (minmap[map_index + k] - tol[k]))	|| (val[k] > (maxmap[map_index + k] + tol[k]))) {
            objmap[obj_index] = 0;
            valid = 0;
          }
        }

        // if valid, compute new boundaries
        if (valid) {

          // announce good news
          objmap[obj_index] = 1;

          // update boundaries
          if (i < *(data->x1)) *(data->x1) = i;
          if (i > *(data->x2)) *(data->x2) = i;
          if (j < *(data->y1)) *(data->y1) = j;
          if (j > *(data->y2)) *(data->y2) = j;

          // recenter validity gap
          for (k = 0; k < depth; k++) {
            if (val[k] < minmap[map_index + k]) {
              delta                     = (tlPixel) (minmap[index + k] - val[k]);
              minmap[depth * index + k] =	TL_MAX (0, (minmap[map_index + k] - delta));
              maxmap[depth * index + k] =	TL_MAX (0, (maxmap[map_index + k] - delta));
            }
            if (val[k] > maxmap[map_index + k]) {
              delta                     = (tlPixel) (val[k] - maxmap[map_index + k]);
              minmap[depth * index + k] =	TL_MIN (TL_MAX_PIXEL_VAL, (minmap[map_index + k] + delta));
              maxmap[depth * index + k] =	TL_MIN (TL_MAX_PIXEL_VAL, (maxmap[map_index + k] + delta));
            }
          }
        }
      }
    }

    // 
    // update vars (safe spiral)
    // 

    if (i_incr < abs (cur_incr)) {
      if (cur_incr > 0)	i++;
      else              i--;
      i_incr++;
      if (i < x1) {
        l_flag = 1;
      }
      else if (i >= x2) {
        r_flag = 1;
      }
    }

    else if (j_incr < abs (cur_incr)) {
      if (cur_incr > 0) j++;
      else              j--;
      j_incr++;
      if (j < y1) {
        t_flag = 1;
      }
      else if (j >= y2) {
        b_flag = 1;
      }
    }
    else {
      if (cur_incr < 0) cur_incr = (-1) * (cur_incr - 1);
      else              cur_incr = (-1) * (cur_incr + 1);
      i_incr = j_incr = 0;
    }
  }


  // free arrays
  free (minmap);
  free (maxmap);
  return 0;
}



///  Recursive call to filter NRG pixel values
///  in an image with arbitrary depth.
///  
///  \param data  structure containing the search params
///  
///  \return
///  0 on success, -1 otherwise. 
///  See \ref errors "error management" for details.

int
tl_extract_cm_nrg (tl_extract_data_cm *data)
{
  int     px          = data->px;
  int     py          = data->py;
  int     width       = data->width;
  int     index       = py * width + px;
  tlPixel *val;


  // check for point validity
  if ((px < (data->lx1)) || (px >= (data->lx2)) || (py < (data->ly1)) || (py >= (data->ly2))) {
    return 0;
  }

  // shortcuts
  val = &(data->src[2*index]);

  // test if current point already invalidated
  if ((data->objmap[index]) != -1) return 0;

  // test point vs. color model
  if (data->cmHist->data[val[0]][val[1]] < data->threshold) {
    data->objmap[index] = 0;
    return 0;
  }


  // 
  // valid point
  // 

  // tag
  data->objmap[index] = 1;

  // update boundaries
  if (px < *(data->x1)) *(data->x1) = px;
  if (px > *(data->x2))	*(data->x2) = px;
  if (py < *(data->y1))	*(data->y1) = py;
  if (py > *(data->y2))	*(data->y2) = py;


  // 
  // recurse (reset to current boundary between each call)
  // 

  // to the right
  data->px = px + 1;
  tl_extract_cm_nrg (data);

  // to the left
  data->px = px - 1;
  tl_extract_cm_nrg (data);

  // recenter
  data->px = px;

  // up
  data->py = py + 1;
  tl_extract_cm_nrg (data);

  // down
  data->py = py - 1;
  tl_extract_cm_nrg (data);

  // recenter
  data->py = py;

  // case of 8 connect object extraction
  if (data->connectivity == TL_CONNEX_8) {

    // up right
    data->px = px + 1;
    data->py = py + 1;
    tl_extract_cm_nrg (data);

    // down right
    data->px = px + 1;
    data->py = py - 1;
    tl_extract_cm_nrg (data);

    // up left
    data->px = px - 1;
    data->py = py + 1;
    tl_extract_cm_nrg (data);

    // down left
    data->px = px - 1;
    data->py = py - 1;
    tl_extract_cm_nrg (data);

    // recenter
    data->px = px;
    data->py = py;
  }


  return 0;
}



///  Recursive call to filter HSI pixel values
///  in an image with arbitrary depth.
///  
///  \param data  structure containing the search params
///  
///  \return
///  0 on success, -1 otherwise. 
///  See \ref errors "error management" for details.

int
tl_extract_cm_hsi (tl_extract_data_cm *data)
{
  int     px          = data->px;
  int     py          = data->py;
  int     width       = data->width;
  int     index       = py * width + px;
  tlPixel *val;


  // check for point validity
  if ((px < (data->lx1)) || (px >= (data->lx2)) || (py < (data->ly1)) || (py >= (data->ly2))) {
    return 0;
  }

  // shortcuts
  val = &(data->src[4*index]);

  // test if current point already invalidated
  if ((data->objmap[index]) != -1) return 0;

  // test point vs. color model
  if (data->cmHist->data[val[0]][val[2]] < data->threshold) {
    data->objmap[index] = 0;
    return 0;
  }


  // 
  // valid point
  // 

  // tag
  data->objmap[index] = 1;

  // update boundaries
  if (px < *(data->x1)) *(data->x1) = px;
  if (px > *(data->x2))	*(data->x2) = px;
  if (py < *(data->y1))	*(data->y1) = py;
  if (py > *(data->y2))	*(data->y2) = py;


  // 
  // recurse (reset to current boundary between each call)
  // 

  // to the right
  data->px = px + 1;
  tl_extract_cm_hsi (data);

  // to the left
  data->px = px - 1;
  tl_extract_cm_hsi (data);

  // recenter
  data->px = px;

  // up
  data->py = py + 1;
  tl_extract_cm_hsi (data);

  // down
  data->py = py - 1;
  tl_extract_cm_hsi (data);

  // recenter
  data->py = py;

  // case of 8 connect object extraction
  if (data->connectivity == TL_CONNEX_8) {

    // up right
    data->px = px + 1;
    data->py = py + 1;
    tl_extract_cm_hsi (data);

    // down right
    data->px = px + 1;
    data->py = py - 1;
    tl_extract_cm_hsi (data);

    // up left
    data->px = px - 1;
    data->py = py + 1;
    tl_extract_cm_hsi (data);

    // down left
    data->px = px - 1;
    data->py = py - 1;
    tl_extract_cm_hsi (data);

    // recenter
    data->px = px;
    data->py = py;
  }


  return 0;
}
