/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange, Terry Fong
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlFile.h"
#include "tlImage.h"

#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include <fstream>
#include <math.h>
#include <malloc.h>
#include <assert.h>

using std::ifstream;
using std::ofstream;
using std::ios;

#ifdef WIN32
#include <windows.h>
#endif



// 
// JPEG stuff
//

#ifndef NO_DOC
extern "C" {
#include "jpeglib.h"
}
struct my_error_mgr {
  struct jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
};
typedef struct my_error_mgr *my_error_ptr;
METHODDEF (void)
my_error_exit (j_common_ptr cinfo);
METHODDEF (void)
my_error_exit (j_common_ptr cinfo) {
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  char buffer[JMSG_LENGTH_MAX];
  (*cinfo->err->format_message) (cinfo, buffer);
  longjmp (myerr->setjmp_buffer, 1);
}
#endif



///  \file tlFile.h
///  \brief Internal file I/O routines.
///  \internal



///  Write to TLIB image format.
///  
///  image image to save
///  filename the file name. If NULL, write to stdout.
///  
///  \return 
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_write_tlb (tlImage *image, 
              char    *filename)
{
  register int i, j, k;

  int           width, height, pixel_width;
  FILE          *pFp;
  bool          hr;
  unsigned int  bytesLine;


  // the following only valid for 1 byte per channel !!!
  assert (TL_MAX_PIXEL_VAL == 255);

  // shortcut
  hr          = false;
  width       = image->width;
  height      = image->height;
  pixel_width = image->pixelWidth;

  // sanity check
  if (!image->pixel && !image->hrpixel) {
    TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR_NULL_POINTER);
    return (-1);
  }

  // status message
  if (tlVerbose) {
    if (filename) {
      printf ("writing TLIB image to %s\n", filename);
    }
    else {
      printf ("writing TLIB image to stdout\n");
    }
  }

  // open file for binary write
  if (filename) {
    if (NULL == (pFp = fopen (filename, "wb"))) {
      TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR_FILE_IO);
      return (-1);
    }
  }
  else {
    pFp = stdout;
  }
  bytesLine = width * pixel_width;

  // write TLB header 
  char *format;
  switch (image->format) {
  case RGB:
    format = "RGB";
    break;
  case NRG:
    format = "NRG";
    break;
  case HSI:
    format = "HSI";
    break;
  case GRAY:
    format = "GRAY";
    break;
  case BIN:
    format = "BIN";
    break;
  case STEREO:
    hr = true;
    format = "STEREO";
    break;
  default:
    TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR);
    return (-1);
  }

  fprintf (pFp, "TLB2\n%d %d\n%s\n%d\n", width, height, format, pixel_width);
  
  // standard image
  if (!hr) {
    tlPixel *buffer;
    tlPixel *src = image->pixel;
    if (NULL == (buffer = (tlPixel *) malloc (bytesLine*sizeof(tlPixel)))) {
      TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR);
      return (-1);
    }

    // write TLIB data one line at a time (this is MUCH faster than one RGB
    // pixel at a time and slightly faster than the whole image at once)
    for (i = 0; i < height; i++) {
      tlPixel *p = buffer;
      for (j = 0; j < width; j++) {
        for (k = 0; k < pixel_width; k++) {
          *p++ = (tlPixel) *src++;
        }
      } 

      if (bytesLine != fwrite ((void *) buffer, sizeof (tlPixel), bytesLine, pFp)) {
        TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR_FILE_IO);
        if (filename) {
          fclose (pFp);
        }
        free (buffer);
        return (-1);
      }
    }
    free (buffer);
  }

  // high-resolution image
  else {
    tlHRPixel *buffer;
    tlHRPixel *src = image->hrpixel;
    if (NULL == (buffer = (tlHRPixel *) malloc (bytesLine*sizeof(tlHRPixel)))) {
      TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR);
      return (-1);
    }

    // write TLIB data one line at a time (this is MUCH faster than one RGB
    // pixel at a time and slightly faster than the whole image at once)
    for (i = 0; i < height; i++) {
      tlHRPixel *p = buffer;
      for (j = 0; j < width; j++) {
        for (k = 0; k < pixel_width; k++) {
          *p++ = (tlHRPixel) (*(src++));
        }
      } 

      if (bytesLine != fwrite ((void *) buffer, sizeof (tlHRPixel), bytesLine, pFp)) {
        TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR_FILE_IO);
        if (filename) {
          fclose (pFp);
        }
        free (buffer);
        return (-1);
      }
    }
    free (buffer);
  }

  // clean up
  if (filename) {
    if (0 > fclose (pFp)) {
      TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR_FILE_IO);
      return (-1);
    }
  }


  return (0);
}



///  Write to Portable Pixmap format (binary format).
///  
///  \param src       RGB pixel array to save
///  \param width     image width
///  \param height    image height
///  \param filename  the file name; if NULL, write to stdout
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_write_ppm (tlPixel *src, 
              int     width, 
              int     height,
              char    *filename)
{
  register int i, j;

  FILE          *pFp;
  unsigned char *buffer;
  unsigned int  bytesLine;


  // sanity check
  if (!src) {
    TL_ERRNO_MSG ("tl_write_ppm()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // status message
  if (tlVerbose) {
    if (filename) {
      printf ("writing PPM image to %s\n", filename);
    }
    else {
      printf ("writing PPM image to stdout\n");
    }
  }

  // open file for binary write
  if (filename) {
    if (NULL == (pFp = fopen (filename, "wb"))) {
      TL_ERRNO_MSG ("tl_write_ppm()", TL_ERROR_FILE_IO);
      return (-1);
    }
  }
  else {
    pFp = stdout;
  }

  bytesLine = width * 3;
  if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
    TL_ERRNO_MSG ("tl_write_ppm()", TL_ERROR);
    return (-1);
  }

  // write PPM header (P6 format)
  fprintf (pFp, "P6\n%d %d\n255\n", width, height);

  // write PPM data one line at a time (this is MUCH faster than one RGB
  // pixel at a time and slightly faster than the whole image at once)
  for (i = 0; i < height; i++) {
    unsigned char *p = buffer;
    for (j = 0; j < width; j++) {
      *p++ = (unsigned char) *src++;
      *p++ = (unsigned char) *src++;
      *p++ = (unsigned char) *src++;
    }

    if (bytesLine != fwrite ((void *) buffer, sizeof (unsigned char), bytesLine, pFp)) {
      TL_ERRNO_MSG ("tl_write_ppm()", TL_ERROR_FILE_IO);
      if (filename) fclose (pFp);
      free (buffer);
      return (-1);
    }
  }
  free (buffer);

  // close file
  if (filename) {
    if (0 > fclose (pFp)) {
      TL_ERRNO_MSG ("tl_write_ppm()", TL_ERROR_FILE_IO);
      return (-1);
    }
  }


  return (0);
}



///  Write to Bitmap format.
///  
///  \param src       RGB pixel array to save
///  \param width     image width
///  \param height    image height
///  \param format    image format
///  \param filename  the file name. If NULL, write to stdout
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \warning
///  this code is NOT portable! Do not attempt to use this
///  on anything other than 32-bit little-endian machines!

int
tl_write_bmp (tlPixel   *src, 
              int       width, 
              int       height, 
              tl_format format,
              char      *filename)
{
  register int i, j;

  int               index;
  BITMAPINFOHEADER  bmih;
  BITMAPFILEHEADER  bmfh;
  unsigned char     *buffer;
  unsigned int      bytesLine;
  int               align;


  // sanity check
  if (!src) {
    TL_ERRNO_MSG ("tl_write_bmp()", TL_ERROR_NULL_POINTER);
    return (-1);
  }

  // status message
  if (tlVerbose) {
    if (filename) {
      printf ("writing BMP image to %s\n", filename);
    }
    else {
      printf ("writing BMP image to stdout\n");
    }
  }

  switch (format) {
case RGB:
  {
    // set header
    memset (&bmih, 0, sizeof (BITMAPINFOHEADER));
    bmih.biSize         = sizeof (BITMAPINFOHEADER);
    bmih.biWidth        = width;
    bmih.biHeight       = height;
    bmih.biPlanes       = 1;
    bmih.biBitCount     = 24;
    bmih.biCompression  = 0L;
    bmih.biClrUsed      = 0;
    bmih.biSizeImage    = width * height;
    memset (&bmfh, 0, sizeof (BITMAPFILEHEADER));
    bmfh.bfType         = (WORD) 0x4D42;
    bmfh.bfSize         = (DWORD) (sizeof (BITMAPFILEHEADER) + bmih.biSize + bmih.biClrUsed * sizeof (RGBQUAD) + bmih.biSizeImage);
    bmfh.bfOffBits      = (DWORD) (sizeof (BITMAPFILEHEADER) + bmih.biSize +  bmih.biClrUsed * sizeof (RGBQUAD));

    // write header
    FILE *pFp;
    if (filename) pFp = fopen (filename, "wb");
    else          pFp = stdout;
    fwrite ((char *) &bmfh.bfType, sizeof (short), 1, pFp);	// 0-1
    fwrite ((char *) &bmfh.bfSize, sizeof (long), 1, pFp);	// 2-5
    fwrite ((char *) &bmfh.bfReserved1, sizeof (short), 1, pFp);	// 6-7
    fwrite ((char *) &bmfh.bfReserved2, sizeof (short), 1, pFp);	// 8-9
    fwrite ((char *) &bmfh.bfOffBits, sizeof (long), 1, pFp);	// 10-13
    fwrite ((char *) &bmih, sizeof (BITMAPINFOHEADER),1, pFp);

    // write data
    align = (4 - ((width * 3) % 4)) & 3;	// align within 32 bits
    bytesLine = width * 3 + align;
    if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
      TL_ERRNO_MSG ("tl_write_bmp()", TL_ERROR);
      return (-1);
    }

    int stride = TL_RGB_WIDTH * width;
    for (j = height - 1; j >= 0; j--) {
      unsigned char *p = buffer;
      unsigned char r, g, b;
      index = stride * j;
      for (i = 0; i < width; i++) {
        b = (unsigned char) src[index++];
        g = (unsigned char) src[index++];
        r = (unsigned char) src[index++];
        *p++ = r;
        *p++ = g;
        *p++ = b;
      } 
      fwrite ((void *)buffer, sizeof (unsigned char), bytesLine, pFp);
    } 
    if (filename) {
      fclose (pFp);
    }
    free (buffer);
  } 
  break;

case GRAY:
  {
    // set header
    LPBYTE pPalette = new BYTE[1024];

    for (i = 0; i < 256; i++) {
      pPalette[4 * i    ] = (BYTE) i;
      pPalette[4 * i + 1] = (BYTE) i;
      pPalette[4 * i + 2] = (BYTE) i;
      pPalette[4 * i + 3] = 0;
    }

    memset (&bmih, 0, sizeof (BITMAPINFOHEADER));
    bmih.biSize         = sizeof (BITMAPINFOHEADER);
    bmih.biWidth        = width;
    bmih.biHeight       = height;
    bmih.biPlanes       = 1;
    bmih.biBitCount     = 8;
    bmih.biCompression  = 0L;
    bmih.biClrUsed      = 256;
    bmih.biSizeImage    = width * height;
    memset (&bmfh, 0, sizeof (BITMAPFILEHEADER));
    bmfh.bfType         = (WORD) 0x4D42;
    bmfh.bfSize         = (DWORD) (sizeof (BITMAPFILEHEADER) + bmih.biSize + bmih.biClrUsed * sizeof (RGBQUAD) + bmih.biSizeImage);
    bmfh.bfOffBits      = (DWORD) (sizeof (BITMAPFILEHEADER) + bmih.biSize + bmih.biClrUsed * sizeof (RGBQUAD));

    // write header
    FILE *pFp;
    if (filename) pFp = fopen (filename, "wb");
    else          pFp = stdout;
    fwrite ((void *) &bmfh.bfType, sizeof (short), 1, pFp);	// 0-1
    fwrite ((void *) &bmfh.bfSize, sizeof (long), 1, pFp);	// 2-5
    fwrite ((void *) &bmfh.bfReserved1, sizeof (short), 1, pFp);	// 6-7
    fwrite ((void *) &bmfh.bfReserved2, sizeof (short), 1, pFp);	// 8-9
    fwrite ((void *) &bmfh.bfOffBits, sizeof (long), 1, pFp);	// 10-13
    fwrite ((void *) &bmih, sizeof (BITMAPINFOHEADER),1, pFp);
    fwrite ((void *) pPalette, 1024, 1, pFp);

    // write data
    bytesLine = width;
    if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
      TL_ERRNO_MSG ("tl_write_bmp()", TL_ERROR);
      return (-1);
    }

    for (j = height - 1; j >= 0; j--) {
      unsigned char *p = buffer;
      index = j * width;
      for (i = 0; i < width; i++) {
        *p++ = (src[index++]);
      }
      fwrite ((void *)buffer, sizeof (unsigned char), bytesLine, pFp);
    }
    if (filename) {
      fclose (pFp);
    }
    free (buffer);
    delete (pPalette);
  } 
  break;

default:
  TL_ERRNO_MSG ("tl_write_bmp()", TL_ERROR);
  return (-1);
  }


  return (0);
}



///  Write to JPEG format.
///  
///  \param src         RGB pixel array to save
///  \param width       image width
///  \param height      image height
///  \param filename    the file name; if NULL, write to stdout
///  \param pixel_width the number of bytes in each pixel
///  \param quality     the quality of the output jpg file [%]
///  
///  \return
///  0 on success, -1 otherwise. 
///  See \ref errors "error management" for details.

int
tl_write_jpg (tlPixel *src, 
              int     width, 
              int     height, 
              int     pixel_width,
              int     quality, 
              char    *filename)
{
  unsigned int                uwidth, uheight;
  struct jpeg_compress_struct cinfo;
  FILE                        *pFp;
  struct my_error_mgr         jerr;


  // sanity check
  if (!src) {
    TL_ERRNO_MSG ("tl_write_jpg()", TL_ERROR_NULL_POINTER);
    return (-1);
  }

  // status message
  if (tlVerbose) {
    if (filename) printf ("writing JPEG image to %s\n", filename);
    else          printf ("writing JPEG image to stdout\n");
  }

  // explicit type casting
  uwidth  = (unsigned int) width;
  uheight = (unsigned int) height;

  // allocate and initialize JPEG compression object
  cinfo.err           = jpeg_std_error (&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  if (setjmp (jerr.setjmp_buffer)) {
    TL_ERRNO_MSG ("tl_write_jpg()", TL_ERROR);
    jpeg_destroy_compress (&cinfo);
    return (-1);
  }
  jpeg_create_compress (&cinfo);

  // open file for binary write
  if (filename) {
    if (NULL == (pFp = fopen (filename, "wb"))) {
      TL_ERRNO_MSG ("tl_write_jpg()", TL_ERROR_FILE_IO);
      return (-1);
    }
  }
  else {
    pFp = stdout;
  }
  jpeg_stdio_dest (&cinfo, pFp);

  // set parameters for compression
  cinfo.image_width  = uwidth;
  cinfo.image_height = uheight;
  if (pixel_width == 3) {
    cinfo.input_components = 3;
    cinfo.in_color_space   = JCS_RGB;
  }
  else {
    cinfo.input_components = 1;
    cinfo.in_color_space   = JCS_GRAYSCALE;
  }
  jpeg_set_defaults (&cinfo);
  jpeg_set_quality (&cinfo, quality, TRUE);

  // start compressor
  jpeg_start_compress (&cinfo, TRUE);

  // while (scan lines remain to be written)
  while (cinfo.next_scanline < cinfo.image_height) {
    unsigned char *outRow;
    if (pixel_width == 3) {
      outRow = src + (cinfo.next_scanline * uwidth * 3);
    }
    else {
      outRow = src + (cinfo.next_scanline * uwidth);
    }
    (void) jpeg_write_scanlines (&cinfo, &outRow, 1);
  }

  // finish compression
  jpeg_finish_compress (&cinfo);

  // close file
  if (filename) {
    if (0 > fclose (pFp)) {
      TL_ERRNO_MSG ("tl_write_tlb()", TL_ERROR_FILE_IO);
      return (-1);
    }
  }

  // release JPEG compression object
  jpeg_destroy_compress (&cinfo);


  return (0);
}



///  Read an image in TLIB format.
///  
///  \param filename      file name
///  \param width         \b [out] image width
///  \param height        \b [out] image height
///  \param pixel_width   \b [out] pixel width
///  \param format        \b [out] image format
///  
///  \return
///  A pointer to the image pixel array read on success, NULL otherwise.\n
///  See \ref errors "error management" for details.

void *
tl_read_tlb (char      *filename, 
             int       *width, 
             int       *height, 
             int       *pixel_width,
             tl_format *format)
{
  register int i, j, k;

  char          header[5];
  void*         pixel;
  FILE          *pFp;
  bool          hr;
  char          theFormat[20];
  unsigned int  bytesLine;


  // sanity check
  if (!filename) {
    TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_NULL_POINTER);
    return (NULL);
  }

  // status message
  if (tlVerbose) {
    printf ("reading TLIB from %s\n", filename);
  }

  // open file
  if (NULL == (pFp = fopen (filename, "rb"))) {
    TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_FILE_IO);
    return (NULL);
  }

  // read TLIB header
  fscanf (pFp, "%4s", header);
  if (strcmp (header, "TLB2")) {
    TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_INVALID_FORMAT);
    fclose (pFp);
    return (NULL);
  }

  // skip any lines not whitespace & comments
  while (1) {
    // read discard up to first non-whitespace in the line
    do {
      i = fgetc (pFp);
    }
    while (isspace (i));
    if (isdigit (i)) {
      ungetc (i, pFp);
      break;
    }
    // discard the line
    else {
      do {
        i = fgetc (pFp);
      }
      while (i != '\n');
    }
  }

  // read image parameters
  hr = false;
  if (EOF == fscanf (pFp, "%d %d %s %d", width, height, theFormat, pixel_width)) {
    TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_FILE_IO);
    fclose (pFp);
    return (NULL);
  }
  if (!strcmp (theFormat, "RGB")) {
    *format = RGB;
  }
  else if (!strcmp (theFormat, "NRG")) {
    *format = NRG;
  }
  else if (!strcmp (theFormat, "HSI")) {
    *format = HSI;
  }
  else if (!strcmp (theFormat, "GRAY")) {
    *format = GRAY;
  }
  else if (!strcmp (theFormat, "BIN")) {
    *format = BIN;
  }
  else if (!strcmp (theFormat, "STEREO")) {
    *format = STEREO;
    hr = true;
  }
  else {
    TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_INVALID_FORMAT);
    return (NULL);
  }

  // read the whitespace character
  i = fgetc (pFp);
  if (!isspace (i)) {
    TL_ERRNO_MSG ("tl_read_tlib()", TL_ERROR_INVALID_FORMAT);
    return (NULL);
  }

  // standard file
  if (!hr) {
    tlPixel *data;
    tlPixel *dest;
    tlPixel *buffer;

    // create pixel array
    data = (tlPixel *) malloc ((*width) * (*height) * (*pixel_width) * sizeof (tlPixel));
    if (data == NULL) {
      TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_ALLOCATION_FAILED);
      return (NULL);
    }
    bytesLine = (*width) * (*pixel_width);
    if (NULL == (buffer = (tlPixel *) malloc (bytesLine*sizeof(tlPixel)))) {
      TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_ALLOCATION_FAILED);
      return (NULL);
    }
    dest = data;
    for (j = 0; j < (*height); j++) {
      // read TLIB data one line at a time (this is MUCH faster than one 
      // RGB
      // pixel at a time and slightly faster than the whole image at
      // once)
      if (bytesLine != fread ((void *) buffer, sizeof (tlPixel), bytesLine, pFp)) {
        TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_FILE_IO);
        fclose (pFp);
        free (buffer);
        return (NULL);
      }
      tlPixel *p = buffer;	// start of line buffer
      for (i = 0; i < (*width); i++) {
        for (k = 0; k < (*pixel_width); k++) {
          *dest++ = *p++;
        }
      }
    }
    free (buffer);

    pixel = (void *)data;
  }

  // high-resolution file
  else {
    tlHRPixel *data;
    tlHRPixel *dest;
    tlHRPixel *buffer;

    // create pixel array
    data = (tlHRPixel *) malloc ((*width) * (*height) * (*pixel_width) * sizeof (tlHRPixel));
    if (data == NULL) {
      TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_ALLOCATION_FAILED);
      return (NULL);
    }
    bytesLine = (*width) * (*pixel_width);
    if (NULL == (buffer = (tlHRPixel *) malloc (bytesLine*sizeof(tlHRPixel)))) {
      TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_ALLOCATION_FAILED);
      return (NULL);
    }
    dest = data;
    for (j = 0; j < (*height); j++) {
      // read TLIB data one line at a time (this is MUCH faster than one 
      // RGB
      // pixel at a time and slightly faster than the whole image at
      // once)
      if (bytesLine != fread ((void *) buffer, sizeof (tlHRPixel), bytesLine, pFp)) {
        TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_FILE_IO);
        fclose (pFp);
        free (buffer);
        return (NULL);
      }
      tlHRPixel *p = buffer;	// start of line buffer
      for (i = 0; i < (*width); i++) {
        for (k = 0; k < (*pixel_width); k++) {
          *dest++ = *p++;
        }
      }
    }
    free (buffer);

    pixel = (void *)data;
  }

  // close file
  if (0 > fclose (pFp)) {
    TL_ERRNO_MSG ("tl_read_tlb()", TL_ERROR_FILE_IO);
    return (NULL);
  }


  return (pixel);
}



///  Read an image from a Portable Pixmap format (binary format).
///  
///  \param filename  file name
///  \param width     \b [out] image width
///  \param height    \b [out] image height
///  \param format    \b [out] image format
///  
///  \return
///  A pointer to the image pixel array read on success, NULL otherwise.\n
///  See \ref errors "error management" for details.

tlPixel *
tl_read_ppm (char      *filename, 
             int       *width, 
             int       *height, 
             tl_format *format)
{
  register int i, j;

  int           maxval;
  char          header[3];
  FILE          *pFp;
  tlPixel       *data, *dest;
  unsigned char *buffer;
  unsigned int  bytesLine;


  // sanity check
  if (!filename) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return (NULL);
  }

  // status message
  if (tlVerbose) {
    printf ("reading PPM from %s\n", filename);
  }

  // open file
  if (NULL == (pFp = fopen (filename, "rb"))) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_FILE_IO);
    return (NULL);
  }

  // read PPM header
  if ((1 != fscanf (pFp, "%2s", &header[0])) || (strcmp (header, "P6"))) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_INVALID_FORMAT);
    fclose (pFp);
    return (NULL);
  }

  // skip any lines not whitespace & comments
  while (1) {
    // read discard up to first non-whitespace in the line
    do {
      i = fgetc (pFp);
    }
    while (isspace (i));
    if (isdigit (i)) {
      ungetc (i, pFp);
      break;
    }
    // discard the line
    else {
      do {
        i = fgetc (pFp);
      }
      while (i != '\n');
    }
  }

  // read image parameters
  if (3 != fscanf (pFp, "%d %d %d", width, height, &maxval)) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_FILE_IO);
    fclose (pFp);
    return (NULL);
  }

  // check ceiling value
  if (maxval != 255) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_INVALID_FORMAT);
    fclose (pFp);
    return (NULL);
  }
  // create pixel array
  data = (tlPixel *) malloc ((*width) * (*height) * TL_RGB_WIDTH * sizeof (tlPixel));
  if (data == NULL) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_ALLOCATION_FAILED);
    return (NULL);
  }

  // read the whitespace character
  i = fgetc (pFp);
  if (!isspace (i)) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_INVALID_FORMAT);
    return (NULL);
  }

  // read PPM data
  *format = RGB;
  bytesLine = (*width) * 3;
  if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_ALLOCATION_FAILED);
    return (NULL);
  }

  dest = data;
  for (j = 0; j < (*height); j++) {
    // read PPM data one line at a time (this is MUCH faster than one
    // RGB
    // pixel at a time and slightly faster than the whole image at
    // once)
    if (bytesLine != fread ((void *) buffer, sizeof (unsigned char), bytesLine, pFp)) {
      TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_FILE_IO);
      fclose (pFp);
      free (buffer);
      return (NULL);
    }
    unsigned char *p = buffer;	// start of line buffer
    for (i = 0; i < (*width); i++) {
      *dest++ = (tlPixel) * p++;
      *dest++ = (tlPixel) * p++;
      *dest++ = (tlPixel) * p++;
    }
  }
  free (buffer);

  // close file
  if (0 > fclose (pFp)) {
    TL_ERRNO_MSG ("tl_read_ppm()", TL_ERROR_FILE_IO);
    return (NULL);
  }


  return (data);
}



///  Read an image from a BMP format.
///  
///  \param filename  file name
///  \param width     \b [out] image width
///  \param height    \b [out] image height
///  \param format    \b [out] image format
///  
///  \return
///  A pointer to the image pixel array read on success, NULL otherwise.\n
///  See \ref errors "error management" for details.

tlPixel *
tl_read_bmp (char      *filename, 
             int       *width, 
             int       *height, 
             tl_format *format)
{
  register int i, j;

  unsigned char    r, g, b;
  int              index;
  tlPixel          *pixel;
  tlPixel          *src;
  BITMAPINFOHEADER bmih;
  BITMAPFILEHEADER bmfh;
  unsigned char    *buffer;
  unsigned int     bytesLine;
  int              align;


  // sanity check
  if (!filename) {
    TL_ERRNO_MSG ("tl_read_bmp()", TL_ERROR_NULL_POINTER);
    return (NULL);		// failure
  }
  // open file
#ifdef LINUX
  /*
  * gcc 3.2 implements C++ standard libraries, which does NOT contain
  * ios::nocreate 
  */
  std::ifstream bild (filename, ios::binary);
  if (!bild) {
    perror ("tl_read_bmp");
    return (NULL);		// failure
  }
#else /* */
  /*
  * ios::nocreate is supported by MSVC 
  */
  ifstream bild (filename, ifstream::binary | ifstream::in);

  if (!bild) {
    TL_ERRNO_MSG ("tl_read_bmp()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return (NULL);		// failure
  }
#endif /* */
  bild.read ((char *) &bmfh.bfType, sizeof (short));
  bild.read ((char *) &bmfh.bfSize, sizeof (long));
  bild.read ((char *) &bmfh.bfReserved1, sizeof (short));
  bild.read ((char *) &bmfh.bfReserved2, sizeof (short));
  bild.read ((char *) &bmfh.bfOffBits, sizeof (long));
  bild.read ((char *) &bmih, sizeof (BITMAPINFOHEADER));

  // initialize image
  *width = bmih.biWidth;
  *height = bmih.biHeight;
  switch (bmih.biClrUsed) {
case 0:
  switch (bmih.biBitCount) {
case 8:			// 8-bit per pixel with color palette
  {
    LPBYTE palette = new BYTE[1024];

    // create pixel array
    *format = RGB;
    pixel = tl_create_array (*width, *height, TL_RGB_WIDTH);
    src = pixel;

    // read-in palette
    bild.read ((char *) palette, 1024);

    bytesLine = (*width);
    if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
      TL_ERRNO_MSG ("tl_read_bmp()", TL_ERROR_ALLOCATION_FAILED);
      return (NULL);	// failure
    }
    // read data 
    bild.seekg (bmfh.bfOffBits, ios::beg);
    for (j = (*height) - 1; j >= 0; j--) {
      index = TL_RGB_WIDTH * (j * (*width));
      bild.read ((char *) buffer, bytesLine);
      unsigned char *p = buffer;

      for (i = 0; i < (*width); i++) {
        r = *p++;
        pixel[index++] = (tlPixel) palette[4 * r + 2];
        pixel[index++] = (tlPixel) palette[4 * r + 1];
        pixel[index++] = (tlPixel) palette[4 * r];
      }
    }
    free (buffer);
  }
  break;

case 24:			// 24-bit per pixel (no palette)
  // create pixel array
  *format = RGB;
  pixel = tl_create_array (*width, *height, TL_RGB_WIDTH);
  src = pixel;

  // a bit tricky: align within 32 bits
  align = (4 - ((*width) % 4)) & 3;
  bytesLine = (*width + align) * 3;
  if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
    TL_ERRNO_MSG ("tl_read_bmp()", TL_ERROR_ALLOCATION_FAILED);
    return (NULL);		// failure
  }

  // read data
  bild.seekg (bmfh.bfOffBits, ios::beg);
  for (j = (*height) - 1; j >= 0; j--) {
    index = TL_RGB_WIDTH * (j * (*width));
    bild.read ((char *) buffer, bytesLine);
    unsigned char *p = buffer;
    for (i = 0; i < (*width); i++) {
      b = *p++;
      g = *p++;
      r = *p++;
      pixel[index++] = (tlPixel) r;
      pixel[index++] = (tlPixel) g;
      pixel[index++] = (tlPixel) b;
    }
  }
  free (buffer);
  break;

case 32:			// 32-bit per pixel
  // create pixel array
  *format = RGB;
  pixel = tl_create_array (*width, *height, TL_RGB_WIDTH);
  src = pixel;
  bytesLine = (*width) * 4;
  if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
    TL_ERRNO_MSG ("tl_read_bmp()", TL_ERROR_ALLOCATION_FAILED);
    return (NULL);		// failure
  }

  // read data
  bild.seekg (bmfh.bfOffBits, ios::beg);
  for (j = (*height) - 1; j >= 0; j--) {
    index = TL_RGB_WIDTH * (j * (*width));
    bild.read ((char *) buffer, bytesLine);
    unsigned char *p = buffer;
    for (i = 0; i < (*width); i++) {
      b = *p++;
      g = *p++;
      r = *p++;
      p++;			// a
      pixel[index++] = (tlPixel) r;
      pixel[index++] = (tlPixel) g;
      pixel[index++] = (tlPixel) b;
    }
  }
  free (buffer);
  break;

default:
  {
    // monochrom and 4bits (8 colors)
    // unsupported yet
    TL_ERRNO_MSG ("tl_read_bmp()", TL_ERROR_INVALID_FORMAT);
    return (NULL);		// failure
  }
  }
  break;
default:			// indexed color
  {
    LPBYTE palette = new BYTE[4 * bmih.biClrUsed];

    // create pixel array
    *format = RGB;
    pixel = tl_create_array (*width, *height, TL_RGB_WIDTH);
    src = pixel;

    // read-in palette
    bild.read ((char *) palette, 4 * bmih.biClrUsed);

    bytesLine = (*width);
    if (NULL == (buffer = (unsigned char *) malloc (bytesLine))) {
      TL_ERRNO_MSG ("tl_read_bmp()", TL_ERROR_ALLOCATION_FAILED);
      return (NULL);		// failure
    }

    // read data
    bild.seekg (bmfh.bfOffBits, ios::beg);
    for (j = (*height) - 1; j >= 0; j--) {
      index = TL_RGB_WIDTH * (j * (*width));
      bild.read ((char *) buffer, bytesLine);
      unsigned char *p = buffer;
      for (i = 0; i < (*width); i++) {
        r = *p++;
        pixel[index++] = (tlPixel) palette[4 * r + 2];
        pixel[index++] = (tlPixel) palette[4 * r + 1];
        pixel[index++] = (tlPixel) palette[4 * r];
      }
    }
    delete (palette);
  }
  }
  bild.close ();


  return (pixel);
}



///  Read an image from a JPEG format.
///  
///  \param filename  file name
///  \param width     \b [out] image width
///  \param height    \b [out] image height
///  \param format    \b [out] image format
///  
///  \return
///  A pointer to the image pixel array read on success, NULL otherwise.\n
///  See \ref errors "error management" for details.

tlPixel *
tl_read_jpg (char      *filename, 
             int       *width, 
             int       *height, 
             tl_format *format)
{
  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr           jerr;

  FILE       *infile;
  JSAMPARRAY buffer;
  int        row_stride;
  tlPixel    *src;


  // check
  if (!filename) {
    TL_ERRNO_MSG ("tl_read_jpg()", TL_ERROR_NULL_POINTER);
    return (NULL);
  }

  // open file
  if ((infile = fopen (filename, "rb")) == NULL) {
    TL_ERRNO_MSG ("tl_read_jpg()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return (NULL);
  }

  // step 1: allocate and initialize JPEG decompression object
  cinfo.err = jpeg_std_error (&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  if (setjmp (jerr.setjmp_buffer)) {
    jpeg_destroy_decompress (&cinfo);
    if (infile != NULL)
      fclose (infile);
    return (NULL);
  }
  jpeg_create_decompress (&cinfo);

  // specify data source (the file)
  jpeg_stdio_src (&cinfo, infile);

  // read file parameters with jpeg_read_header()
  (void) jpeg_read_header (&cinfo, TRUE);

  // start decompressor
  (void) jpeg_start_decompress (&cinfo);
  *width = (int) (cinfo.output_width);
  *height = (int) (cinfo.output_height);
  row_stride = cinfo.output_width * cinfo.output_components;
  buffer =(*cinfo.mem->alloc_sarray) ((j_common_ptr) & cinfo, JPOOL_IMAGE, row_stride, 1);
  src = tl_create_array (*width, *height, 3);
  if (src == NULL) {
    TL_ERRNO_MSG ("tl_read_jpg()", TL_ERROR_ALLOCATION_FAILED);
    return (NULL);
  }

  // while (scan lines remain to be read)
  while (cinfo.output_scanline < cinfo.output_height) {
    (void) jpeg_read_scanlines (&cinfo, buffer, 1);
    if (cinfo.out_color_components == 3) {
      *format = RGB;
      int offset = (cinfo.output_scanline - 1) * (*width) * 3;
      int count;
      for (count = 0; count < (*width); count++) {
        *(src + offset + count * 3 + 0) = *(buffer[0] + count * 3 + 0);
        *(src + offset + count * 3 + 1) = *(buffer[0] + count * 3 + 1);
        *(src + offset + count * 3 + 2) = *(buffer[0] + count * 3 + 2);
      }
    }
    else if (cinfo.out_color_components == 1) {
      *format = GRAY;
      int offset = (cinfo.output_scanline - 1) * (*width) * 3;
      int count;
      for (count = 0; count < (*width); count++) {
        unsigned char iGray;
        iGray = *(buffer[0] + count);
        *(src + offset + count * 3 + 0) = iGray;
        *(src + offset + count * 3 + 1) = iGray;
        *(src + offset + count * 3 + 2) = iGray;
      }
    }
  }

  // finish decompression
  (void) jpeg_finish_decompress (&cinfo);

  // release JPEG decompression object
  jpeg_destroy_decompress (&cinfo);
  fclose (infile);


  return src;
}
