/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlFilter.h"
#include "tlRect.h"
#include "tlColor.h"



///  \file tlFilter.h
///  \brief Internal filter routines.
///  \internal



///  Filter an RGB image using two color boundaries.
///  
///  \param src     RGB pixel array to be filtered
///  \param width   image width
///  \param height  image height
///  \param r_min   lower channel boundaries [0..255]
///  \param r_max   upper channel boundaries [0..255]
///  \param g_min   lower channel boundaries [0..255]
///  \param g_max   upper channel boundaries [0..255]
///  \param b_min   lower channel boundaries [0..255]
///  \param b_max   upper channel boundaries [0..255]
///  \param rect    portion of the image to process
///  \param mask    mask for the given area (must be same size as image size)
///  \param dest    \b [out] filtered pixel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_filter_rgb (tlPixel *src, 
               int     width, 
               int     height, 
               tlPixel r_min,
               tlPixel r_max, 
               tlPixel g_min, 
               tlPixel g_max, 
               tlPixel b_min,
               tlPixel b_max, 
               tlRect  *rect, 
               tlPixel *dest, 
               tlPixel *mask)
{
  register int index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel r, g, b;
  int     mask_index;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_filter_rgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  if (mask == NULL) {
    // filter
    for (j = y1; j < y2; j++) {
      index = TL_RGB_WIDTH * TL_PIXEL (j, x1, width);
      for (i = x1; i < x2; i++) {
        r = src[index];
        g = src[index + 1];
        b = src[index + 2];
        if ((r > r_max) || (r < r_min) || (g > g_max)
          || (g < g_min)
          || (b > b_max) || (b < b_min)) {
            dest[index] = 0;
            dest[index + 1] = 0;
            dest[index + 2] = 0;
          }
        else {
          dest[index] = r;
          dest[index + 1] = g;
          dest[index + 2] = b;
        }
        index += TL_RGB_WIDTH;
      }
    }
  }

  else {
    // filter
    for (j = y1; j < y2; j++) {
      mask_index = TL_PIXEL (j, x1, width);
      index = TL_RGB_WIDTH * mask_index;
      for (i = x1; i < x2; i++) {
        if (mask[mask_index] != 0) {
          r = src[index];
          g = src[index + 1];
          b = src[index + 2];
          if ((r > r_max) || (r < r_min) || (g > g_max)
            || (g < g_min)
            || (b > b_max) || (b < b_min)) {
              dest[index] = 0;
              dest[index + 1] = 0;
              dest[index + 2] = 0;
              mask[mask_index] = 0;
            }
          else {
            dest[index] = r;
            dest[index + 1] = g;
            dest[index + 2] = b;
          }
        }
        index += TL_RGB_WIDTH;
        mask_index += 1;
      }
    }
  }


  return 0;
}



///  Filter an HSI image using two color boundaries.
///  
///  \param src     HSI pixel array to be filtered
///  \param width   image width
///  \param height  image height
///  \param h1_min  lower channel boundaries [0..255]
///  \param h1_max  upper channel boundaries [0..255]
///  \param h2_min  lower channel boundaries [0..255]
///  \param h2_max  upper channel boundaries [0..255]
///  \param s_min   lower channel boundaries [0..255]
///  \param s_max   upper channel boundaries [0..255]
///  \param i_min   lower channel boundaries [0..255]
///  \param i_max   upper channel boundaries [0..255]
///  \param rect    portion of the image to process
///  \param mask    mask for the given area (must be same size as image size)
///  \param dest    \b [out] filtered pixel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_filter_hsi (tlPixel *src, 
               int     width, 
               int     height, 
               tlPixel h1_min,
               tlPixel h1_max, 
               tlPixel h2_min, 
               tlPixel h2_max, 
               tlPixel s_min,
               tlPixel s_max, 
               tlPixel i_min, 
               tlPixel i_max, 
               tlRect  *rect,
               tlPixel *dest, 
               tlPixel *mask)
{
  register int index;

  int i, j;
  int x1, x2;
  int y1, y2;
  tlPixel h1, h2, s, i_;
  int mask_index;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_filter_hsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  if (mask == NULL) {
    // filter
    for (j = y1; j < y2; j++) {
      index = TL_HSI_WIDTH * TL_PIXEL (j, x1, width);
      for (i = x1; i < x2; i++) {
        h1 = src[index  ];
        h2 = src[index+1];
        s  = src[index+2];
        i_ = src[index+3];
        if ((h1 > h1_max) || (h1 < h1_min) || (h2 > h2_max)
          || (h2 < h2_min)
          || (s > s_max) || (s < s_min) || (i_ > i_max)
          || (i_ < i_min)) {
            dest[index  ] = 0;
            dest[index+1] = 0;
            dest[index+2] = 0;
            dest[index+3] = 0;
          }
        else {
          dest[index  ] = h1;
          dest[index+1] = h2;
          dest[index+2] = s;
          dest[index+3] = i_;
        }
        index += TL_HSI_WIDTH;
      }
    }
  }

  else {
    // filter
    for (j = y1; j < y2; j++) {
      mask_index = TL_PIXEL (j, x1, width);
      index = TL_HSI_WIDTH * mask_index;
      for (i = x1; i < x2; i++) {
        if (mask[mask_index] != 0) {
          h1 = src[index  ];
          h2 = src[index+1];
          s  = src[index+2];
          i_ = src[index+3];
          if ((h1 > h1_max) || (h1 < h1_min) || (h2 > h2_max)
            || (h2 < h2_min)
            || (s > s_max) || (s < s_min) || (i_ > i_max)
            || (i_ < i_min)) {
              dest[index  ] = 0;
              dest[index+1] = 0;
              dest[index+2] = 0;
              dest[index+3] = 0;
              mask[mask_index] = 0;
            }
          else {
            dest[index  ] = h1;
            dest[index+1] = h2;
            dest[index+2] = s;
            dest[index+3] = i_;
          }
        }
        index      += TL_HSI_WIDTH;
        mask_index += 1;
      }
    }
  }


  return 0;
}



///  Filter an NRG image using two color boundaries.
///  
///  src      NRG pixel array to be filtered
///  nr_min   lower channel boundaries [0..255]
///  nr_max   upper channel boundaries [0..255]
///  ng_min   lower channel boundaries [0..255]
///  ng_max   upper channel boundaries [0..255]
///  rect     portion of the image to process
///  mask     mask for the given area (must be same size as image size)
///  dest     \b [out] filtered pixel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_filter_nrg (tlPixel *src, 
               int     width, 
               int     height, 
               tlPixel nr_min,
               tlPixel nr_max, 
               tlPixel ng_min, 
               tlPixel ng_max, 
               tlRect  *rect,
               tlPixel *dest, 
               tlPixel *mask)
{
  register int index;

  int i, j;
  int x1, x2;
  int y1, y2;
  tlPixel nr, ng;
  int mask_index;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_filter_nrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  if (mask == NULL) {
    // filter
    for (j = y1; j < y2; j++) {
      index = TL_NRG_WIDTH * TL_PIXEL (j, x1, width);
      for (i = x1; i < x2; i++) {
        nr = src[index];
        ng = src[index + 1];
        if ((nr > nr_max) || (nr < nr_min) || (ng > ng_max)
          || (ng < ng_min)) {
            dest[index  ] = 0;
            dest[index+1] = 0;
          }
        else {
          dest[index  ] = nr;
          dest[index+1] = ng;
        }
        index += TL_NRG_WIDTH;
      }
    }
  }

  else {
    // filter
    for (j = y1; j < y2; j++) {
      mask_index = TL_PIXEL (j, x1, width);
      index = TL_NRG_WIDTH * mask_index;
      for (i = x1; i < x2; i++) {
        if (mask[mask_index] != 0) {
          nr = src[index  ];
          ng = src[index+1];
          if ((nr > nr_max) || (nr < nr_min) || (ng > ng_max)
            || (ng < ng_min)) {
              dest[index  ]    = 0;
              dest[index+1]    = 0;
              mask[mask_index] = 0;
            }
          else {
            dest[index  ] = nr;
            dest[index+1] = ng;
          }
        }
        index      += TL_NRG_WIDTH;
        mask_index += 1;
      }
    }
  }


  return 0;
}



///  Filter a GRAY image using two color boundaries.
///  
///  \param src             GRAY pixel array to be filtered
///  \param width           image width
///  \param height          image height
///  \param intensity_min   lower channel boundaries [0..255]
///  \param intensity_max   upper channel boundaries [0..255]
///  \param rect            portion of the image to process
///  \param mask            mask for the given area (must be same size as image size)
///  \param dest            \b [out] filtered pixel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_filter_gray (tlPixel *src, 
                int     width, 
                int     height, 
                tlPixel intensity_min,
                tlPixel intensity_max, 
                tlRect  *rect, 
                tlPixel *dest,
                tlPixel *mask)
{
  register int index;

  int i, j;
  int x1, x2;
  int y1, y2;
  tlPixel intensity;
  int mask_index;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_filter_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  if (mask == NULL) {
    // filter
    for (j = y1; j < y2; j++) {
      index = TL_GRAY_WIDTH * TL_PIXEL (j, x1, width);
      for (i = x1; i < x2; i++) {
        intensity = src[index];
        if ((intensity > intensity_max)
          || (intensity < intensity_min))
          dest[index] = 0;

        else
          dest[index] = intensity;
        index += TL_GRAY_WIDTH;
      }
    }
  }

  else {
    // filter
    for (j = y1; j < y2; j++) {
      mask_index = TL_PIXEL (j, x1, width);
      index = TL_GRAY_WIDTH * mask_index;
      for (i = x1; i < x2; i++) {
        if (mask[mask_index] != 0) {
          intensity = src[index];
          if ((intensity > intensity_max)
            || (intensity < intensity_min)) {
              dest[index] = 0;
              mask[mask_index] = 0;
            }
          else {
            dest[index] = intensity;
          }
        }
        index += TL_GRAY_WIDTH;
        mask_index += 1;
      }
    }
  }


  return 0;
}



///  Filter a HRGRAY image using two color boundaries.
///  
///  \param src             GRAY pixel array to be filtered
///  \param width           image width
///  \param height          image height
///  \param intensity_min   lower channel boundaries [0..255]
///  \param intensity_max   upper channel boundaries [0..255]
///  \param rect            portion of the image to process
///  \param mask            mask for the given area (must be same size as image size)
///  \param dest            \b [out] filtered pixel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_filter_hrgray (tlHRPixel *src, 
                  int       width, 
                  int       height, 
                  tlHRPixel intensity_min,
                  tlHRPixel intensity_max, 
                  tlRect    *rect, 
                  tlHRPixel *dest,
                  tlPixel   *mask)
{
  register int index;

  int i, j;
  int x1, x2;
  int y1, y2;
  tlHRPixel intensity;
  int mask_index;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_filter_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  if (mask == NULL) {
    // filter
    for (j = y1; j < y2; j++) {
      index = TL_PIXEL (j, x1, width);
      for (i = x1; i < x2; i++) {
        intensity = src[index];
        if ((intensity > intensity_max)
          || (intensity < intensity_min))
          dest[index] = 0;

        else dest[index] = intensity;
        index++;
      }
    }
  }

  else {
    // filter
    for (j = y1; j < y2; j++) {
      mask_index = TL_PIXEL (j, x1, width);
      index = mask_index;
      for (i = x1; i < x2; i++) {
        if (mask[mask_index] != 0) {
          intensity = src[index];
          if ((intensity > intensity_max) || (intensity < intensity_min)) {
              dest[index] = 0;
              mask[mask_index] = 0;
            }
          else dest[index] = intensity;
        }
        index++;
        mask_index++;
      }
    }
  }


  return 0;
}



///  Extract the color boundaries of a portion of an RGB image;
///  color boundaries take the same format as the input image.
///  
///  \param src       RGB pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param maxColor  \b [out] upper color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_rgb (tlPixel *src, 
                      int     width, 
                      int     height,
                      tlColor *minColor, 
                      tlColor *maxColor, 
                      tlRect  *rect,
                      tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel r, g, b;
  tlPixel min_r, max_r;
  tlPixel min_g, max_g;
  tlPixel min_b, max_b;


  // check
  if ((!src) || (!minColor) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_rgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_r = min_g = min_b = TL_MAX_PIXEL_VAL;
  max_r = max_g = max_b = TL_MIN_PIXEL_VAL;
  x1    = TL_MAX ((rect->x), 0);
  x2    = TL_MIN ((x1 + (rect->width)), width);
  y1    = TL_MAX ((rect->y), 0);
  y2    = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_RGB_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        r = src[index];
        g = src[index + 1];
        b = src[index + 2];
        max_r = (r > max_r) ? r : max_r;
        min_r = (r < min_r) ? r : min_r;
        max_g = (g > max_g) ? g : max_g;
        min_g = (g < min_g) ? g : min_g;
        max_b = (b > max_b) ? b : max_b;
        min_b = (b < min_b) ? b : min_b;
      }
      index += TL_RGB_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->r = min_r;
  minColor->g = min_g;
  minColor->b = min_b;
  maxColor->r = max_r;
  maxColor->g = max_g;
  maxColor->b = max_b;


  return 0;
}



///  Extract the lower color boundary of a portion of an RGB image;
///  color boundaries take the same format as the input image.
///  
///  \param src       RGB pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_min_rgb (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *minColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel r, g, b;
  tlPixel min_r, min_g, min_b;


  // check
  if ((!src) || (!minColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_min_rgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_r = min_g = min_b = TL_MAX_PIXEL_VAL;
  x1    = TL_MAX ((rect->x), 0);
  x2    = TL_MIN ((x1 + (rect->width)), width);
  y1    = TL_MAX ((rect->y), 0);
  y2    = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_RGB_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        r = src[index];
        g = src[index + 1];
        b = src[index + 2];
        min_r = (r < min_r) ? r : min_r;
        min_g = (g < min_g) ? g : min_g;
        min_b = (b < min_b) ? b : min_b;
      }
      index      += TL_RGB_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->r = min_r;
  minColor->g = min_g;
  minColor->b = min_b;


  return 0;
}


///  Extract the upper color boundary of a portion of an RGB image;
///  color boundaries take the same format as the input image.
/// 
///  \param src          RGB pixel array to be processed
///  \param width        input image width
///  \param height       input image height
///  \param rect         portion of the image to process
///  \param maxColor     upper color boundary
///  \param mask         binary mask that indicates which pixels to take into account
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_max_rgb (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *maxColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel r, g, b;
  tlPixel max_r, max_g, max_b;


  // check
  if ((!src) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_max_rgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  max_r = max_g = max_b = TL_MIN_PIXEL_VAL;
  x1    = TL_MAX ((rect->x), 0);
  x2    = TL_MIN ((x1 + (rect->width)), width);
  y1    = TL_MAX ((rect->y), 0);
  y2    = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_RGB_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        r = src[index];
        g = src[index + 1];
        b = src[index + 2];
        max_r = (r > max_r) ? r : max_r;
        max_g = (g > max_g) ? g : max_g;
        max_b = (b > max_b) ? b : max_b;
      }
      index      += TL_RGB_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  maxColor->r = max_r;
  maxColor->g = max_g;
  maxColor->b = max_b;


  return 0;
}



///  Extract the color boundaries of a portion of an YUV image;
///  color boundaries take the same format as the input image.
///  
///  \param src       YUV pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param maxColor  \b [out] upper color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_yuv (tlPixel *src, 
                      int     width, 
                      int     height,
                      tlColor *minColor, 
                      tlColor *maxColor, 
                      tlRect  *rect,
                      tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel y, u, v;
  tlPixel min_y, max_y;
  tlPixel min_u, max_u;
  tlPixel min_v, max_v;


  // check
  if ((!src) || (!minColor) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_yuv()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_y = min_u = min_v = TL_MAX_PIXEL_VAL;
  max_y = max_u = max_v = TL_MIN_PIXEL_VAL;
  x1    = TL_MAX ((rect->x), 0);
  x2    = TL_MIN ((x1 + (rect->width)), width);
  y1    = TL_MAX ((rect->y), 0);
  y2    = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_YUV_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        y = src[index];
        u = src[index + 1];
        v = src[index + 2];
        max_y = (y > max_y) ? y : max_y;
        min_y = (y < min_y) ? y : min_y;
        max_u = (u > max_u) ? u : max_u;
        min_u = (u < min_u) ? u : min_u;
        max_v = (v > max_v) ? v : max_v;
        min_v = (v < min_v) ? v : min_v;
      }
      index += TL_YUV_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->y = min_y;
  minColor->u = min_u;
  minColor->v = min_v;
  maxColor->y = max_y;
  maxColor->u = max_u;
  maxColor->v = max_v;


  return 0;
}



///  Extract the lower color boundary of a portion of an YUV image;
///  color boundaries take the same format as the input image.
///  
///  \param src       YUV pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_min_yuv (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *minColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel y, u, v;
  tlPixel min_y, min_u, min_v;


  // check
  if ((!src) || (!minColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_min_yuv()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_y = min_u = min_v = TL_MAX_PIXEL_VAL;
  x1    = TL_MAX ((rect->x), 0);
  x2    = TL_MIN ((x1 + (rect->width)), width);
  y1    = TL_MAX ((rect->y), 0);
  y2    = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_YUV_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        y = src[index];
        u = src[index + 1];
        v = src[index + 2];
        min_y = (y < min_y) ? y : min_y;
        min_u = (u < min_u) ? u : min_u;
        min_v = (v < min_v) ? v : min_v;
      }
      index      += TL_YUV_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->y = min_y;
  minColor->u = min_u;
  minColor->v = min_v;


  return 0;
}


///  Extract the upper color boundary of a portion of an YUV image;
///  color boundaries take the same format as the input image.
/// 
///  \param src          YUV pixel array to be processed
///  \param width        input image width
///  \param height       input image height
///  \param rect         portion of the image to process
///  \param maxColor     upper color boundary
///  \param mask         binary mask that indicates which pixels to take into account
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_max_yuv (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *maxColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel y, u, v;
  tlPixel max_y, max_u, max_v;


  // check
  if ((!src) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_max_yuv()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  max_y = max_u = max_v = TL_MIN_PIXEL_VAL;
  x1    = TL_MAX ((rect->x), 0);
  x2    = TL_MIN ((x1 + (rect->width)), width);
  y1    = TL_MAX ((rect->y), 0);
  y2    = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_YUV_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        y = src[index];
        u = src[index + 1];
        v = src[index + 2];
        max_y = (y > max_y) ? y : max_y;
        max_u = (u > max_u) ? u : max_u;
        max_v = (v > max_v) ? v : max_v;
      }
      index      += TL_YUV_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  maxColor->y = max_y;
  maxColor->u = max_u;
  maxColor->v = max_v;


  return 0;
}



///  Extract the color boundaries of a portion of an HSI image;
///  color boundaries take the same format as the input image.
///  
///  \param src       HSI pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param maxColor  \b [out] upper color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_hsi (tlPixel *src, 
                      int     width, 
                      int     height,
                      tlColor *minColor, 
                      tlColor *maxColor, 
                      tlRect  *rect,
                      tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel h1, h2, s, i_;
  tlPixel min_h1, max_h1;
  tlPixel min_h2, max_h2;
  tlPixel min_s, max_s;
  tlPixel min_i, max_i;


  // check
  if ((!src) || (!minColor) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_hsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_h1 = min_h2 = min_s = min_i = TL_MAX_PIXEL_VAL;
  max_h1 = max_h2 = max_s = max_i = TL_MIN_PIXEL_VAL;
  x1    = TL_MAX ((rect->x), 0);
  x2    = TL_MIN ((x1 + (rect->width)), width);
  y1    = TL_MAX ((rect->y), 0);
  y2    = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_HSI_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        h1     = src[index];
        h2     = src[index + 1];
        s      = src[index + 2];
        i_     = src[index + 3];
        max_h1 = (h1 > max_h1) ? h1 : max_h1;
        min_h1 = (h1 < min_h1) ? h1 : min_h1;
        max_h2 = (h2 > max_h2) ? h2 : max_h2;
        min_h2 = (h2 < min_h2) ? h2 : min_h2;
        max_s  = (s > max_s) ? s : max_s;
        min_s  = (s < min_s) ? s : min_s;
        max_i  = (i_ > max_i) ? i_ : max_i;
        min_i  = (i_ < min_i) ? i_ : min_i;
      }
      index      += TL_HSI_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->h1 = min_h1;
  minColor->h2 = min_h2;
  minColor->s  = min_s;
  minColor->i  = min_i;
  maxColor->h1 = max_h1;
  maxColor->h2 = max_h2;
  maxColor->s  = max_s;
  maxColor->i  = max_i;


  return 0;
}



///  Extract the lower color boundary of a portion of an HSI image;
///  color boundaries take the same format as the input image.
///  
///  \param src       HSI pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_min_hsi (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *minColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel h1, h2, s, i_;
  tlPixel min_h1, min_h2, min_s, min_i;


  // check
  if ((!src) || (!minColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_min_hsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_h1 = min_h2 = min_s = min_i = TL_MAX_PIXEL_VAL;
  x1     = TL_MAX ((rect->x), 0);
  x2     = TL_MIN ((x1 + (rect->width)), width);
  y1     = TL_MAX ((rect->y), 0);
  y2     = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index      = TL_HSI_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        h1     = src[index];
        h2     = src[index + 1];
        s      = src[index + 2];
        i_     = src[index + 3];
        min_h1 = (h1 < min_h1) ? h1 : min_h1;
        min_h2 = (h2 < min_h2) ? h2 : min_h2;
        min_s  = (s < min_s) ? s : min_s;
        min_i  = (i_ < min_i) ? i_ : min_i;
      }
      index      += TL_HSI_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->h1 = min_h1;
  minColor->h2 = min_h2;
  minColor->s  = min_s;
  minColor->i  = min_i;


  return 0;
}



///  Extract the upper color boundary of a portion of an HSI image;
///  color boundaries take the same format as the input image.
///  
///  \param src         HSI pixel array to be processed
///  \param width       input image width
///  \param height      input image height
///  \param rect        portion of the image to process
///  \param maxColor    \b [out] upper color boundary
///  \param mask        binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_max_hsi (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *maxColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel h1, h2, s, i_;
  tlPixel max_h1, max_h2, max_s, max_i;


  // check
  if ((!src) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_max_hsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  max_h1 = max_h2 = max_s = max_i = TL_MIN_PIXEL_VAL;
  x1     = TL_MAX ((rect->x), 0);
  x2     = TL_MIN ((x1 + (rect->width)), width);
  y1     = TL_MAX ((rect->y), 0);
  y2     = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index      = TL_HSI_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        h1     = src[index];
        h2     = src[index + 1];
        s      = src[index + 2];
        i_     = src[index + 3];
        max_h1 = (h1 > max_h1) ? h1 : max_h1;
        max_h2 = (h2 > max_h2) ? h2 : max_h2;
        max_s  = (s > max_s) ? s : max_s;
        max_i  = (i_ > max_i) ? i_ : max_i;
      }
      index      += TL_HSI_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  maxColor->h1 = max_h1;
  maxColor->h2 = max_h2;
  maxColor->s  = max_s;
  maxColor->i  = max_i;


  return 0;
}



///  Extract the color boundaries of a portion of an NRG image;
///  color boundaries take the same format as the input image.
///  
///  \param src       NRG pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param maxColor  \b [out] upper color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_nrg (tlPixel *src, 
                      int     width, 
                      int     height,
                      tlColor *minColor, 
                      tlColor *maxColor, 
                      tlRect  *rect,
                      tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel nr, ng;
  tlPixel min_nr, max_nr;
  tlPixel min_ng, max_ng;


  // check
  if ((!src) || (!minColor) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_nrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_nr = min_ng = TL_MAX_PIXEL_VAL;
  max_nr = max_ng = TL_MIN_PIXEL_VAL;
  x1     = TL_MAX ((rect->x), 0);
  x2     = TL_MIN ((x1 + (rect->width)), width);
  y1     = TL_MAX ((rect->y), 0);
  y2     = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index      = TL_NRG_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        nr     = src[index];
        ng     = src[index + 1];
        max_nr = (nr > max_nr) ? nr : max_nr;
        min_nr = (nr < min_nr) ? nr : min_nr;
        max_ng = (ng > max_ng) ? ng : max_ng;
        min_ng = (ng < min_ng) ? ng : min_ng;
      }
      index      += TL_NRG_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->nr = min_nr;
  minColor->ng = min_ng;
  maxColor->nr = max_nr;
  maxColor->ng = max_ng;


  return 0;
}



///  Extract the lower color boundary of a portion of an NRG image;
///  color boundaries take the same format as the input image.
///  
///  \param src       NRG pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] minColor lower color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_min_nrg (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *minColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel nr, ng;
  tlPixel min_nr, min_ng;


  // check
  if ((!src) || (!minColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_min_nrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_nr = min_ng = TL_MAX_PIXEL_VAL;
  x1     = TL_MAX ((rect->x), 0);
  x2     = TL_MIN ((x1 + (rect->width)), width);
  y1     = TL_MAX ((rect->y), 0);
  y2     = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = TL_NRG_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        nr     = src[index];
        ng     = src[index + 1];
        min_nr = (nr < min_nr) ? nr : min_nr;
        min_ng = (ng < min_ng) ? ng : min_ng;
      }
      index      += TL_NRG_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->nr = min_nr;
  minColor->ng = min_ng;


  return 0;
}



///  Extract the upper color boundary of a portion of an NRG image;
///  color boundaries take the same format as the input image.
/// 
///  \param src       NRG pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param maxColor  \b [out] upper color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_max_nrg (tlPixel *src, 
                          int     width, 
                          int     height,
                          tlColor *maxColor, 
                          tlRect  *rect, 
                          tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel nr, ng;
  tlPixel max_nr, max_ng;


  // check
  if ((!src) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_max_nrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  max_nr = max_ng = TL_MIN_PIXEL_VAL;
  x1     = TL_MAX ((rect->x), 0);
  x2     = TL_MIN ((x1 + (rect->width)), width);
  y1     = TL_MAX ((rect->y), 0);
  y2     = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index      = TL_NRG_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        nr     = src[index];
        ng     = src[index + 1];
        max_nr = (nr > max_nr) ? nr : max_nr;
        max_ng = (ng > max_ng) ? ng : max_ng;
      }
      index      += TL_NRG_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  maxColor->nr = max_nr;
  maxColor->ng = max_ng;


  return 0;
}



///  Extract the color boundaries of a portion of an GRAY image;
///  color boundaries take the same format as the input image.
///  
///  \param src        GRAY pixel array to be processed
///  \param width      input image width
///  \param height     input image height
///  \param rect       portion of the image to process
///  \param minColor   \b [out] lower color boundary
///  \param maxColor   \b [out] upper color boundary
///  \param mask       binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_gray (tlPixel *src, 
                       int     width, 
                       int     height,
                       tlColor *minColor, 
                       tlColor *maxColor, 
                       tlRect  *rect,
                       tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel intensity;
  tlPixel min_intensity, max_intensity;


  // check
  if ((!src) || (!minColor) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_intensity = TL_MAX_PIXEL_VAL;
  max_intensity = TL_MIN_PIXEL_VAL;
  x1            = TL_MAX ((rect->x), 0);
  x2            = TL_MIN ((x1 + (rect->width)), width);
  y1            = TL_MAX ((rect->y), 0);
  y2            = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index      = TL_GRAY_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        intensity     = src[index];
        max_intensity =	(intensity > max_intensity) ? intensity : max_intensity;
        min_intensity =	(intensity < min_intensity) ? intensity : min_intensity;
      }
      index      += TL_GRAY_WIDTH;
      mask_index += 1;
    }
  }

  // save color data
  minColor->intensity = min_intensity;
  maxColor->intensity = max_intensity;


  return 0;
}



///  Extract the lower color boundary of a portion of an GRAY image;
///  color boundaries take the same format as the input image.
///  
///  \param src       GRAY pixel array to be processed
///  \param width     input image width
///  \param height    input image height
///  \param rect      portion of the image to process
///  \param minColor  \b [out] lower color boundary
///  \param mask      binary mask that indicates which pixels to take into account
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_min_gray (tlPixel *src, 
                           int     width, 
                           int     height,
                           tlColor *minColor, 
                           tlRect  *rect, 
                           tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel intensity;
  tlPixel min_intensity;


  // check
  if ((!src) || (!minColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_min_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  min_intensity = TL_MAX_PIXEL_VAL;
  x1            = TL_MAX ((rect->x), 0);
  x2            = TL_MIN ((x1 + (rect->width)), width);
  y1            = TL_MAX ((rect->y), 0);
  y2            = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index      = TL_GRAY_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        intensity     = src[index];
        min_intensity =	(intensity < min_intensity) ? intensity : min_intensity;
      }
      index += TL_GRAY_WIDTH;
    }
  }


  // save color data
  minColor->intensity = min_intensity;
  return 0;
}



///  Extract the upper color boundary of a portion of an GRAY image;
///  color boundaries take the same format as the input image.
///
///  \param src           GRAY pixel array to be processed
///  \param width         input image width
///  \param height        input image height
///  \param maxColor      upper color boundary
///  \param rect          portion of the image to process
///  \param mask          binary mask that indicates which pixels to take into account
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_extract_color_max_gray (tlPixel *src, 
                           int     width, 
                           int     height,
                           tlColor *maxColor, 
                           tlRect  *rect, 
                           tlPixel *mask)
{
  register int index;
  register int mask_index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel intensity;
  tlPixel max_intensity;


  // check
  if ((!src) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tl_extract_color_max_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  // shortcuts
  max_intensity = TL_MIN_PIXEL_VAL;
  x1            = TL_MAX ((rect->x), 0);
  x2            = TL_MIN ((x1 + (rect->width)), width);
  y1            = TL_MAX ((rect->y), 0);
  y2            = TL_MIN ((y1 + (rect->height)), height);

  // filter
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index      = TL_GRAY_WIDTH * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        intensity     = src[index];
        max_intensity = (intensity > max_intensity) ? intensity : max_intensity;
      }
      index      += TL_GRAY_WIDTH;
      mask_index += 1;
    }
  }


  // save color data
  maxColor->intensity = max_intensity;
  return 0;
}
