/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#ifdef LINUX
#include <stdlib.h>
#include <errno.h>
#endif

#include "tlFind.h"
#include "tlImage.h"
#include "tlMask.h"
#include "tlObject.h"
#include "tlConst.h"



///  \def TL_DEFAULT_INCREMENT
///  Default increment in [pix] used for correlation routines.
///  The correlation routines will compare one pixel in every TL_DEFAULT_INCREMENT.

///  \file tlFind.h
///  \brief Internal template matching operators.
///  \internal



///  Find a match for a particular shape in an image,
///  given a minimum confidence threshold.
///  
///  \param image       the BIN image to search
///  \param shape       the shape mask
///  \param confidence  min match threshold [%]
///  \param rect        image portion to process
///  \param increment   correlation increment [pix]
///  \param shape       shape location is updated
///  
///  \return
///  confidence [%] on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_find_shape (tlPixelArray *image, 
               tlMask       *shape, 
               int          confidence,
               tlRect       *rect, 
               int          increment)
{
  register int      index;
  register int      shape_index;
  register tlPixel  *src;
  register tlPixel  *shape_src;

  int i = 0, j;
  int ii, jj;
  int x1, x2;
  int y1, y2;
  int width, height;
  int shape_width, shape_height, shape_area;
  int count, conf = 0;


  // check
  if ((!image) || (!shape) || (!rect)) {
    TL_ERRNO_MSG ("tl_find_shape()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if ((image->pixelWidth) != 1) {
    TL_ERRNO_MSG ("tl_find_shape()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // shortcuts
  width         = image->width;
  height        = image->height;
  shape_width   = shape->width;
  shape_height  = shape->height;
  shape_area    = shape->area;
  src           = image->pixel;
  shape_src     = shape->pixel;
  x1            = TL_MAX ((rect->x) - ((rect->x) % increment), 0);
  x2            = TL_MIN ((x1 + (rect->width) - shape_width + 1), (width - shape_width - increment));
  y1            = TL_MAX ((rect->y) - ((rect->y) % increment), 0);
  y2            = TL_MIN ((y1 + (rect->height) - shape_height + 1), (height - shape_height - increment));

  // browse rectangle
  for (j = y1; j < y2; j += increment) {
    for (i = x1; i < x2; i += increment) {
      // browse object
      count = 0;
      for (jj = 0; jj < shape_height; jj++) {
        index = TL_PIXEL ((j + jj), i, width);
        shape_index = TL_PIXEL (jj, 0, shape_width);
        for (ii = 0; ii < shape_width; ii++) {
          // if shape match at this location,
          // add to overlapping area count
          if (shape_src[shape_index])
            if (src[index])
              count++;
          // move to next point
          index       += 1;
          shape_index += 1;
        }
      }
      // if we're confident enough, return
      conf = 100 * count / shape_area;
      if (conf >= confidence) {
        break;
      }
    }
  }

  // validate
  if (conf >= confidence) {
    shape->setLocation (i, j);
    return conf;
  }

  // or admit being defeated
  else {
    errno = TL_ERROR_NOT_FOUND;
    return -1;
  }
}



///  Find the best match for a particular shape in a given image.
///  
///  \param image       the BIN image to search
///  \param shape       the shape mask
///  \param rect        image portion to process
///  \param increment   correlation increment [pix]
///  \param shape       \b [out] shape location is updated
///  
///  \return
///  confidence [%] on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_find_best_shape (tlPixelArray *image, 
                    tlMask       *shape, 
                    tlRect       *rect,
                    int          increment)
{
  register int     index;
  register int     shape_index;
  register tlPixel *src;
  register tlPixel *shape_src;

  int i, j;
  int ii, jj;
  int x1, x2;
  int y1, y2;
  int width, height;
  int shape_width, shape_height;
  int count;
  int best_count, best_x = 0, best_y = 0;


  // check
  if ((!image) || (!shape) || (!rect)) {
    TL_ERRNO_MSG ("tl_find_best_shape()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if ((image->pixelWidth) != 1) {
    TL_ERRNO_MSG ("tl_find_best_shape()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // shortcuts
  width        = image->width;
  height       = image->height;
  shape_width  = shape->width;
  shape_height = shape->height;
  src          = image->pixel;
  shape_src    = shape->pixel;
  x1           = TL_MAX ((rect->x) - ((rect->x) % increment), 0);
  x2           = TL_MIN (((rect->x) + (rect->width) - shape_width + 1), (width - shape_width - increment));
  y1           = TL_MAX ((rect->y) - ((rect->y) % increment), 0);
  y2           = TL_MIN (((rect->y) + (rect->height) - shape_height + 1), (height - shape_height - increment));

  // browse rectangle
  best_count = 0;
  for (j = y1; j < y2; j += increment) {
    for (i = x1; i < x2; i += increment) {
      // browse object
      count = 0;
      for (jj = 0; jj < shape_height; jj++) {
        index = TL_PIXEL ((j + jj), i, width);
        shape_index = TL_PIXEL (jj, 0, shape_width);
        for (ii = 0; ii < shape_width; ii++) {
          // if shape match at this location,
          // add to overlapping area count 
          if (shape_src[shape_index])
            if (src[index])
              count++;
          // move to next point
          index += 1;
          shape_index += 1;
        }
      }
      // evaluate
      if (count > best_count) {
        best_count = count;
        best_x = i;
        best_y = j;
      }
    }
  }

  // save result and return confidence
  if (best_count != 0) {
    shape->x = best_x;
    shape->y = best_y;
    return 100 * best_count / (shape->area);
  }

  // or admit being defeated
  else {
    errno = TL_ERROR_NOT_FOUND;
    return -1;
  }
}



///  Find a match for a particular object in an image,
///  given a minimum confidence threshold;
///  object and image have to be of the same format.
///  
///  \param image       the image to search
///  \param object      the object to find
///  \param confidence  min match threshold [%]
///  \param rect        image portion to process
///  \param increment   correlation increment [pix]
///  \param object      \b [out] object location is updated
///  
///  \return
///  confidence [%] on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_find_object (tlPixelArray *image, 
                tlObject     *object, 
                int          confidence,
                tlRect       *rect, 
                int          increment)
{
  register int     index;
  register int     shape_index;
  register tlPixel *src;
  register tlPixel *object_src;
  register tlPixel *shape_src;

  int i = 0, j, k;
  int ii, jj;
  int x1, x2;
  int y1, y2;
  int width, height;
  int object_width, object_height;
  int pixel_width;
  int ref_conf, conf = 0, lssum;


  // check
  if ((!image) || (!object) || (!rect)) {
    TL_ERRNO_MSG ("tl_find_object()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check formats
  if ((image->format) != (object->format)) {
    TL_ERRNO_MSG ("tl_find_object()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // shortcuts
  width         = image->width;
  height        = image->height;
  pixel_width   = image->pixelWidth;
  object_width  = object->width;
  object_height = object->height;
  src           = image->pixel;
  shape_src     = object->shape->pixel;
  object_src    = object->pixel;
  x1            = TL_MAX ((rect->x) - ((rect->x) % increment), 0);
  x2            = TL_MIN ((x1 + (rect->width) - object_width + 1), (width - object_width - increment));
  y1            = TL_MAX ((rect->y) - ((rect->y) % increment), 0);
  y2            = TL_MIN ((y1 + (rect->height) - object_height + 1), (height - object_height - increment));

  // compute worse possible square difference (arbitrary)...
  ref_conf = (object->shape->area) * pixel_width * (TL_MAX_PIXEL_VAL - TL_MIN_PIXEL_VAL) * (TL_MAX_PIXEL_VAL - TL_MIN_PIXEL_VAL);

  // browse rectangle
  for (j = y1; j < y2; j += increment) {
    for (i = x1; i < x2; i += increment) {
      index = TL_PIXEL (j, i, width);
      // browse object
      lssum = 0;
      for (jj = 0; jj < object_height; jj++) {
        index = pixel_width * TL_PIXEL ((j + jj), i, width);
        shape_index = TL_PIXEL (jj, 0, object_width);
        for (ii = 0; ii < object_width; ii++) {
          if (shape_src[shape_index])
            for (k = 0; k < pixel_width; k++) {
              int tsum = (int) (src[index + k]) - (int) (object_src[pixel_width * shape_index]);
              lssum += tsum*tsum;
            }
          index += pixel_width;
          shape_index += 1;
        }
      }
      // evaluate success
      conf = (int) (100.0f * (1.0f - (float) (lssum) / (float) (pixel_width * (object->shape->area) *	TL_MAX_PIXEL_VAL * 20)));
      if (conf >= confidence) break;
    }
  }

  // victory...
  if (conf >= confidence) {
    float vx = (float) ((i - (object->x)) / (float) (tl_timer_get_time () - object->timeStamp));
    float vy = (float) ((j - (object->y)) / (float) (tl_timer_get_time () - object->timeStamp));

    // compute velocity
    object->vx->update (vx);
    object->vy->update (vy);

    // save position found
    object->setLocation (i, j);

    // return confidence
    return TL_MAX (conf, 1);
  }

  // or admit being defeated
  else {
    errno = TL_ERROR_NOT_FOUND;
    return -1;
  }
}



///  Find the best match for a particular object in a given image;
///  object and image have to be of the same format.
///  
///  \param image     the image to search
///  \param object    the object to find
///  \param rect      image portion to process
///  \param increment correlation increment [pix]
///  \param object    \b [out] object location is updated
///  
///  \return
///  confidence [%] on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_find_best_object (tlPixelArray *image, 
                     tlObject     *object, 
                     tlRect       *rect,
                     int          increment)
{
  register int     index;
  register int     shape_index;
  register tlPixel *src;
  register tlPixel *object_src;
  register tlPixel *shape_src;

  int i, j, k;
  int ii, jj;
  int x1, x2;
  int y1, y2;
  int width, height;
  int object_width, object_height;
  int pixel_width;
  int lssum;
  int best_conf, best_x = 0, best_y = 0;


  // check
  if ((!image) || (!object) || (!rect)) {
    TL_ERRNO_MSG ("tl_find_best_object()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check formats
  if ((image->format) != (object->format)) {
    TL_ERRNO_MSG ("tl_find_best_object()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // shortcuts
  width         = image->width;
  height        = image->height;
  pixel_width   = image->pixelWidth;
  object_width  = object->width;
  object_height = object->height;
  src           = image->pixel;
  shape_src     = object->shape->pixel;
  object_src    = object->pixel;
  x1            = TL_MAX ((rect->x) - ((rect->x) % increment), 0);
  x2            = TL_MIN ((x1 + (rect->width) - object_width + 1), (width - object_width - increment));
  y1            = TL_MAX ((rect->y) - ((rect->y) % increment), 0);
  y2            = TL_MIN ((y1 + (rect->height) - object_height + 1), (height - object_height - increment));

  // browse rectangle
  best_conf = 100000000;
  for (j = y1; j < y2; j += increment) {
    for (i = x1; i < x2; i += increment) {
      // browse object
      lssum = 0;
      for (jj = 0; jj < object_height; jj++) {
        index = pixel_width * TL_PIXEL ((j + jj), i, width);
        shape_index = TL_PIXEL (jj, 0, object_width);
        for (ii = 0; ii < object_width; ii++) {
          if (shape_src[shape_index])
            for (k = 0; k < pixel_width; k++) {
              int tsum = (int) (src[index + k]) - (int) (object_src[pixel_width * shape_index]);
              lssum += tsum*tsum;
            }
            index += pixel_width;
            shape_index += 1;
        }
      }
      // evaluate success
      if (lssum < best_conf) {
        best_conf = lssum;
        best_x = i;
        best_y = j;
      }
    }
  }

  // save result and return confidence
  if (best_conf != 100000000) {
    // compute velocity
    float vx = (float) ((best_x - (object->x)) / (float) (tl_timer_get_time () - object->timeStamp));
    float vy = (float) ((best_y - (object->y)) / (float) (tl_timer_get_time () - object->timeStamp));

    // update object velocity
    object->vx->update (vx);
    object->vy->update (vy);

    // save position found
    object->setLocation (best_x, best_y);
    return TL_MAX(0, (int) (100.0f * (1.0f - (float)(best_conf) / (float) (pixel_width * (object->shape->area) * TL_MAX_PIXEL_VAL * 20))));
  }

  // or admit being defeated
  else {
    errno = TL_ERROR_NOT_FOUND;
    return -1;
  }
}
