/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>
#include <assert.h>

#include "tlCameraStereo.h"
#include "tlFormatConversion.h"
#include "tlRect.h"



///  \file tlFormatConversion.h
///  \brief Internal format conversion operators.
///  \internal




///  Fast inverse cosine based on polynomial approximation,
///  maximum absolute error = 6.7626e-05.
///  
///  \param a   input (0 <= a <= 1)
///  
///  \note
///  Adapted from MgcFastFunction.cpp::InvCos0() by Magic Software, Inc.\n
///  \htmlonly <a href="http://www.magic-software.com">http://www.magic-software.com</a>\endhtmlonly
///  Copyright (c) 2000-2003.  All Rights Reserved.\n

inline static float
acosFast (float a)
{
  float fResult = -0.0187293f;


  fResult *= a;
  fResult += 0.0742610f;
  fResult *= a;
  fResult -= 0.2121144f;
  fResult *= a;
  fResult += 1.5707288f;
  fResult *= (float) sqrt (1.0f - a);


  return fResult;
}



///  Fast cosine based on polynomial approximation
///  maximum absolute error = 1.1880e-03
///  
///  \param a input (0 <= a <= PI/2)
///  
///  \note
///  Adapted from MgcFastFunction.cpp::Cos() by Magic Software, Inc.\n
///  \htmlonly <a href="http://www.magic-software.com">http://www.magic-software.com</a>\endhtmlonly
///  Copyright (c) 2000-2003.  All Rights Reserved.\n

inline static float
cosFast (float a)
{
  float aSqr    = a * a;
  float fResult = 3.705e-02f;


  fResult *= aSqr;
  fResult -= 4.967e-01f;
  fResult *= aSqr;
  fResult += 1.0f;


  return fResult;
}



///  RGB to NRG format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2nrg (tlPixel *src, 
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_rgb2nrg (src, width, height, NULL, rect, dest));
}



///  RGB to NRG format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    binary conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2nrg (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  int     sum;
  tlPixel R, G, B;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_rgb2nrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_RGB_WIDTH == 3);
  assert (TL_NRG_WIDTH == 2);
  for (j = y1; j < y2; j++) {
    index = TL_PIXEL (j, x1, width);
    sindex = TL_RGB_WIDTH * index;
    dindex = TL_NRG_WIDTH * index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        R = src[sindex++];
        G = src[sindex++];
        B = src[sindex++];
        sum = R + G + B;
        if (sum) {
          dest[dindex++] = (tlPixel) (TL_MAX_PIXEL_VAL * R / sum);
          dest[dindex++] = (tlPixel) (TL_MAX_PIXEL_VAL * G / sum);
        }
        else {
          dest[dindex++] = 0;
          dest[dindex++] = 0;
        }
      }
      else {
        sindex += TL_RGB_WIDTH;
        dindex += TL_NRG_WIDTH;
      }
      index++;
    }
  }
  return 0;
}



///  RGB to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2gray (tlPixel *src, 
             int     width, 
             int     height, 
             tlRect  *rect,
             tlPixel *dest)
{
  return (tl_rgb2gray (src, width, height, NULL, rect, dest));
}



///  RGB to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2gray (tlPixel *src, 
             int     width, 
             int     height, 
             tlPixel *mask,
             tlRect  *rect, 
             tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel R, G, B;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_rgb2gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_RGB_WIDTH  == 3);
  assert (TL_GRAY_WIDTH == 1);
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = TL_RGB_WIDTH * index;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        R = src[sindex++];
        G = src[sindex++];
        B = src[sindex++];
        dest[dindex++] = (tlPixel)((R + G + B) / 3);
      }
      else {
        sindex += TL_RGB_WIDTH;
        dindex++;
      }
      index++;
    }
  }


  return 0;
}



///  RGB to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2bin (tlPixel *src, 
            int     width, 
            int     height, 
            int     min, 
            int     max,
            tlRect  *rect, 
            tlPixel *dest)
{
  return (tl_rgb2bin (src, width, height, min, max, NULL, rect, dest));
}



///  RGB to BIN format conversion.
///  
///  \param src pixel array containing image to convert
///  \param width image width
///  \param height image height
///  \param min lower binary threshold [0..255]
///  \param max upper binary threshold [0..255]
///  \param mask conversion mask (only converts active pixels)
///  \param rect portion of the image to process
///  \param dest \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2bin (tlPixel *src, 
            int     width, 
            int     height, 
            int     min, 
            int     max,
            tlPixel *mask, 
            tlRect  *rect, 
            tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  int     sum;
  tlPixel R, G, B;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_rgb2bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_RGB_WIDTH == 3);
  assert (TL_BIN_WIDTH == 1);
  min *= 3;
  max *= 3;
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = TL_RGB_WIDTH * index;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        R = src[sindex++];
        G = src[sindex++];
        B = src[sindex++];
        sum = R + G + B;
        if ((sum >= min) && (sum <= max)) {
          dest[dindex++] = TL_MAX_PIXEL_VAL;
        }
        else {
          dest[dindex++] = TL_MIN_PIXEL_VAL;
        }
      }
      else {
        sindex += TL_RGB_WIDTH;
        dindex++;
      }
      index++;
    }
  }


  return 0;
}



///  NRG to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_nrg2gray (tlPixel *src, 
             int     width, 
             int     height, 
             tlRect  *rect,
             tlPixel *dest)
{
  return (tl_nrg2gray (src, width, height, NULL, rect, dest));
}



///  NRG to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_nrg2gray (tlPixel *src, 
             int     width, 
             int     height, 
             tlPixel *mask,
             tlRect  *rect, 
             tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel NR, NG;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_nrg2gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert 
  assert (TL_NRG_WIDTH  == 2);
  assert (TL_GRAY_WIDTH == 1);
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = TL_NRG_WIDTH * index;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        NR = src[sindex++];
        NG = src[sindex++];
        dest[dindex++] = (tlPixel)(NR + NG);
      }
      else {
        sindex += TL_NRG_WIDTH;
        dindex++;
      }
      index++;
    }
  }


  return 0;
}



///  NRG to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_nrg2bin (tlPixel *src, 
            int     width, 
            int     height, 
            int     min, 
            int     max,
            tlRect  *rect, 
            tlPixel *dest)
{
  return (tl_nrg2bin (src, width, height, min, max, NULL, rect, dest));
}



///  NRG to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_nrg2bin (tlPixel *src, 
            int     width, 
            int     height, 
            int     min, 
            int     max,
            tlPixel *mask, 
            tlRect  *rect, 
            tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  int     sum;
  tlPixel NR, NG;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_nrg2bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert 
  assert (TL_NRG_WIDTH == 2);
  assert (TL_BIN_WIDTH == 1);
  min *= 2;
  max *= 2;
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = TL_NRG_WIDTH * index;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        NR = src[sindex++];
        NG = src[sindex++];
        sum = NR + NG;
        if ((sum >= min) && (sum <= max)) {
          dest[dindex++] = TL_MAX_PIXEL_VAL;
        }
        else {
          dest[dindex++] = TL_MIN_PIXEL_VAL;
        }
      }
      else {
        sindex += TL_NRG_WIDTH;
        dindex++;
      }
      index++;
    }
  }


  return 0;
}



///  GRAY to RGB format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_gray2rgb (tlPixel *src, 
             int     width, 
             int     height, 
             tlRect  *rect, 
             tlPixel *dest)
{
  return (tl_gray2rgb (src, width, height, NULL, rect, dest));
}



///  GRAY to RGB format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_gray2rgb (tlPixel *src, 
             int     width, 
             int     height, 
             tlPixel *mask, 
             tlRect  *rect, 
             tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel gray;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_gray2rgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert 
  assert (TL_GRAY_WIDTH == 1);
  assert (TL_RGB_WIDTH  == 3);
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = index;
    dindex = TL_RGB_WIDTH * index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        gray = src[sindex++];
        dest[dindex++] = gray;
        dest[dindex++] = gray;
        dest[dindex++] = gray;
      }
      else {
        sindex++;
        dindex += 3;
      }
      index++;
    }
  }


  return 0;
}



///  GRAY to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_gray2bin (tlPixel *src, 
             int     width, 
             int     height, 
             int     min, 
             int     max,
             tlRect  *rect, 
             tlPixel *dest)
{
  return (tl_gray2bin (src, width, height, min, max, NULL, rect, dest));
}



///  GRAY to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_gray2bin (tlPixel *src, 
             int     width, 
             int     height, 
             int     min, 
             int     max,
             tlPixel *mask, 
             tlRect  *rect, 
             tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel gray;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_gray2bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert 
  assert (TL_GRAY_WIDTH == 1);
  assert (TL_BIN_WIDTH  == 1);
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = index;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        gray = src[sindex++];
        if ((gray >= min) && (gray <= max)) {
          dest[dindex++] = TL_MAX_PIXEL_VAL;
        }
        else {
          dest[dindex++] = TL_MIN_PIXEL_VAL;
        }
      }
      else {
        sindex++;
        dindex++;
      }
      index++;
    }
  }


  return 0;
}



///  GRAY to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hrgray2bin (tlHRPixel     *src, 
               int           width, 
               int           height, 
               int           min, 
               int           max,
               tlRect        *rect, 
               tlPixel       *dest)
{
  return (tl_hrgray2bin (src, width, height, min, max, NULL, rect, dest));
}



///  HRGRAY to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hrgray2bin (tlHRPixel *src, 
               int       width, 
               int       height, 
               int       min, 
               int       max,
               tlPixel   *mask, 
               tlRect    *rect, 
               tlPixel   *dest)
{
  register int index, sindex, dindex;

  int       i, j;
  int       x1, x2;
  int       y1, y2;
  tlHRPixel gray;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_gray2bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert 
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = index;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        gray = src[sindex++];
        if ((gray >= min) && (gray <= max)) {
          dest[dindex++] = TL_MAX_PIXEL_VAL;
        }
        else {
          dest[dindex++] = TL_MIN_PIXEL_VAL;
        }
      }
      else {
        sindex++;
        dindex++;
      }
      index++;
    }
  }


  return 0;
}



///  RGB to HSI format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2hsi (tlPixel *src, 
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_rgb2hsi (src, width, height, NULL, rect, dest));
}



///  RGB to HSI format conversion
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \note
///  For transform details, see:\n
///  Gonzalez and Woods. 1992. Digital Image Processing. Addison-Wesley.

int
tl_rgb2hsi (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel R, G, B;
  int     rg, rb, gb;
  float   H, TWOPI      = (float) (2 * TL_PI);
  tlPixel I, S;
  int     min, denom;


  // scaled hue conversion (16-bit)
  unsigned short int sHue, sMax = TL_MAX_PIXEL_VAL * TL_MAX_PIXEL_VAL;

  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_rgb2hsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_RGB_WIDTH == 3);
  assert (TL_HSI_WIDTH == 4);
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = TL_RGB_WIDTH * index;
    dindex = TL_HSI_WIDTH * index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        R = src[sindex++];
        G = src[sindex++];
        B = src[sindex++];
        rg = R - G;
        rb = R - B;
        gb = G - B;

        // H = [0,2PI] 
        denom = rg * rg + rb * gb;
        if (denom == 0) {
          H = 0;
        }
        else {
          H = (float) acosFast ((rg + rb) / (2.0f * (float) sqrt ((float)denom)));
          if (B > G) {
            H = TWOPI - H;
          }
        }

        // I = [0,255]
        I = (tlPixel)((R + G + B) / 3);

        // scale S [0,1] to [0,TL_MAX_PIXEL_VAL] 
        if (I == 0) {
          S = 0;
        }
        else {
          min = TL_MIN (R, G);
          min = TL_MIN (min, B);
          S = (tlPixel) (TL_MAX_PIXEL_VAL - (min * TL_MAX_PIXEL_VAL / I));
        }

        // scale hue [0,2PI] to 16-bit 
        sHue           = (unsigned short int) (sMax * H / TWOPI);
        dest[dindex++] = (tlPixel) (sHue >> 8);
        dest[dindex++] = (tlPixel) (sHue & 0xff);
        dest[dindex++] = (tlPixel) S;
        dest[dindex++] = (tlPixel) I;
      }
      else {
        sindex += TL_RGB_WIDTH;
        dindex += TL_HSI_WIDTH;
      }
      index++;
    }
  }


  return 0;
}



///  HSI to RGB format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \note
///  For transform details, see:\n
///  Gonzalez and Woods. 1992. Digital Image Processing. Addison-Wesley.

int
tl_hsi2rgb (tlPixel *src,
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_hsi2rgb (src, width, height, NULL, rect, dest));
}



///  HSI to RGB format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2rgb (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  register int index, sindex, dindex;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  tlPixel hl, hm, I;
  int     R, G, B;
  float   H, S;
  float   TWOPI       = (float) (2 * TL_PI);
  float   PI3         = (float) (TL_PI / 3);
  float   H1          = (float) (2 * TL_PI / 3);
  float   H2          = (float) (4 * TL_PI / 3);


  // scaled hue conversion (16-bit)
  unsigned short int sMax = TL_MAX_PIXEL_VAL * TL_MAX_PIXEL_VAL;

  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_hsi2rgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_HSI_WIDTH == 4);
  assert (TL_RGB_WIDTH == 3);
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = TL_HSI_WIDTH * index;
    dindex = TL_RGB_WIDTH * index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        hm = src[sindex++];
        hl = src[sindex++];

        // scale hue 16-bit to [0,2PI]
        H = ((hm << 8) | hl) * TWOPI / sMax;

        // scale saturation [0,TL_MAX_PIXEL_VAL] to [0,1]
        S = src[sindex++] / (float) TL_MAX_PIXEL_VAL;

        // I = [0,255]
        I = src[sindex++];
        if (H <= H1) {
          B = (int) (I * (1 - S));
          R = (int) (I * (1 + (S * cosFast (H) / cosFast (PI3 - H))));
          G = 3 * I - R - B;
        }
        else if (H <= H2) {
          float h = H - H1;
          R = (int) (I * (1 - S));
          G = (int) (I * (1 + (S * cosFast (h) / cosFast (PI3 - h))));
          B = 3 * I - R - G;
        }
        else {
          float h = H - H2;
          G = (int) (I * (1 - S));
          B = (int) (I * (1 + (S * cosFast (h) / cosFast (PI3 - h))));
          R = 3 * I - G - B;
        } 

        R = TL_MIN (R, TL_MAX_PIXEL_VAL);
        R = TL_MAX (R, TL_MIN_PIXEL_VAL);
        G = TL_MIN (G, TL_MAX_PIXEL_VAL);
        G = TL_MAX (G, TL_MIN_PIXEL_VAL);
        B = TL_MIN (B, TL_MAX_PIXEL_VAL);
        B = TL_MAX (B, TL_MIN_PIXEL_VAL);
        dest[dindex++] = (tlPixel) R;
        dest[dindex++] = (tlPixel) G;
        dest[dindex++] = (tlPixel) B;
      }
      else {
        sindex += TL_HSI_WIDTH;
        dindex += TL_RGB_WIDTH;
      }
      index++;
    }
  }


  return 0;
}



///  HSI to NRG format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2nrg (tlPixel *src, 
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_hsi2nrg (src, width, height, NULL, rect, dest));
}



///  HSI to NRG format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2nrg (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  if (0 == tl_hsi2rgb (src, width, height, mask, rect, dest)) 
    return tl_rgb2nrg (dest, width, height, mask, rect, dest);
  else return -1;
}



///  HSI to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2gray (tlPixel *src, 
             int     width, 
             int     height, 
             tlRect  *rect,
             tlPixel *dest)
{
  return (tl_hsi2gray (src, width, height, NULL, rect, dest));
}



///  HSI to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2gray (tlPixel *src, 
             int     width, 
             int     height, 
             tlPixel *mask,
             tlRect  *rect, 
             tlPixel *dest)
{
  register int index, sindex, dindex;

  int i, j;
  int x1, x2;
  int y1, y2;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_hsi2gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert 
  assert (TL_HSI_WIDTH  == 4);
  assert (TL_GRAY_WIDTH == 1);
  for (j = y1; j < y2; j++) {
    index = TL_PIXEL (j, x1, width);
    sindex = TL_HSI_WIDTH * index;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        dest[dindex] = src[sindex];
      }
      index++;
      sindex += TL_HSI_WIDTH;
      dindex++;
    }
  }


  return 0;
}



///  HSI to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2bin (tlPixel *src, 
            int     width, 
            int     height, 
            int     min, 
            int     max,
            tlRect  *rect, 
            tlPixel *dest)
{
  return (tl_hsi2bin (src, width, height, min, max, NULL, rect, dest));
}



///  HSI to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2bin (tlPixel *src, 
            int     width, 
            int     height, 
            int     min, 
            int     max,
            tlPixel *mask, 
            tlRect  *rect, 
            tlPixel *dest)
{
  register int index, sindex, dindex;

  int i, j;
  int x1, x2;
  int y1, y2;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_hsi2bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert 
  assert (TL_HSI_WIDTH == 4);
  assert (TL_BIN_WIDTH == 1);

  // proceed
  for (j = y1; j < y2; j++) {
    index  = TL_PIXEL (j, x1, width);
    sindex = TL_HSI_WIDTH * index + 3;
    dindex = index;
    for (i = x1; i < x2; i++) {
      if (!mask || (mask[index])) {
        if ((src[sindex] >= min) && (src[sindex] <= max)) {
          dest[dindex] = TL_MAX_PIXEL_VAL;
        }
        else {
          dest[dindex] = TL_MIN_PIXEL_VAL;
        }
      }
      index++;
      sindex += TL_HSI_WIDTH;
      dindex++;
    }
  }


  return 0;
}



///  HRGRAY to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  \param shift   \b [default=0] amount of right shift (read "divisions by 2") to perform 
///                 on each pixel to cast high-res image into single-byte image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hrgray2gray (tlHRPixel     *src, 
                int           width, 
                int           height, 
                tlRect        *rect, 
                tlPixel       *dest,
                unsigned char shift)
{
  return (tl_hrgray2gray (src, width, height, NULL, rect, dest, shift));
}



///  HRGRAY to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  \param shift   \b [default=0] amount of right shift (read "divisions by 2") to perform 
///                 on each pixel to cast high-res image into single-byte image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hrgray2gray (tlHRPixel     *src, 
                int           width, 
                int           height, 
                tlPixel       *mask,
                tlRect        *rect, 
                tlPixel       *dest,
                unsigned char shift)
{
  register tlHRPixel *sptr;
  register tlPixel   *dptr;
  register tlPixel   *mptr;

  int i, j;
  int x1, x2;
  int y1, y2;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_hrgray2gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);
  for (j = y1; j < y2; j++) {
    sptr = src  + j * width;
    dptr = dest + j * width;
    mptr = mask + j * width;
    for (i = x1; i < x2; i++) {
      if (!mask || (*mptr)) {
        if (*sptr > 0) *dptr = (tlPixel)((*sptr) >> shift);
        else           *dptr = 0;
      }
      sptr++;
      dptr++;
      mptr++;
    }
  }


  return 0;
}


///  RGB to YUV format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_rgb2yuv (tlPixel *src, 
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_rgb2yuv (src, width, height, NULL, rect, dest));
}



///  RGB to YUV format conversion
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \note
///  For transform details, see:\n
///  http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  For a fast integer implementation, see:\n
///  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/wceddraw/html/_dxce_converting_between_yuv_and_rgb.asp

int
tl_rgb2yuv (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  register tlPixel *sptr, *dptr, *mptr;

  int i, j;
  int x1, x2;
  int y1, y2;
  int offset;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_rgb2yuv()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_RGB_WIDTH == 3);
  assert (TL_YUV_WIDTH == 3);
  for (j=y1; j<y2; j++) {
    offset = j*width+x1;
    sptr = src  + TL_RGB_WIDTH*offset;
    dptr = dest + TL_YUV_WIDTH*offset;
    mptr = mask + offset;
    for (i=x1; i<x2; i++) {
      if (!mask || (*mptr++)) {
        *(dptr)   = (tlPixel) TL_BOUND(TL_MIN_PIXEL_VAL, ( (  66 * (*(sptr)) + 129 * (*(sptr+1)) +  25 * (*(sptr+2)) + 128) >> 8) +  16, TL_MAX_PIXEL_VAL);
        *(dptr+1) = (tlPixel) TL_BOUND(TL_MIN_PIXEL_VAL, ( ( -38 * (*(sptr)) -  74 * (*(sptr+1)) + 112 * (*(sptr+2)) + 128) >> 8) + 128, TL_MAX_PIXEL_VAL);
        *(dptr+2) = (tlPixel) TL_BOUND(TL_MIN_PIXEL_VAL, ( ( 112 * (*(sptr)) -  94 * (*(sptr+1)) -  18 * (*(sptr+2)) + 128) >> 8) + 128, TL_MAX_PIXEL_VAL);
      }
      sptr += TL_RGB_WIDTH;
      dptr += TL_YUV_WIDTH;
    }
  }


  return 0;
}



///  YUV to RGB format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_yuv2rgb (tlPixel *src,
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_yuv2rgb (src, width, height, NULL, rect, dest));
}



///  HSI to RGB format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  For transform details, see:\n
///  http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  For a fast integer implementation, see:\n
///  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/wceddraw/html/_dxce_converting_between_yuv_and_rgb.asp

int
tl_yuv2rgb (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  register tlPixel *sptr, *dptr, *mptr;
  register int     C,D,E;

  int i, j;
  int x1, x2;
  int y1, y2;
  int offset;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_yuv2rgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_RGB_WIDTH == 3);
  assert (TL_YUV_WIDTH == 3);
  for (j=y1; j<y2; j++) {
    offset = j*width+x1;
    sptr = src  + TL_YUV_WIDTH*offset;
    dptr = dest + TL_RGB_WIDTH*offset;
    mptr = mask + offset;
    for (i=x1; i<x2; i++) {
      if (!mask || (*mptr++)) {
        C = *(sptr  ) -  16;
        D = *(sptr+1) - 128;
        E = *(sptr+2) - 128;
        *(dptr  ) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C           + 409 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
        *(dptr+1) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C - 100 * D - 208 * E + 128) >> 8), TL_MAX_PIXEL_VAL);
        *(dptr+2) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (( 298 * C + 516 * D           + 128) >> 8), TL_MAX_PIXEL_VAL);
      }
      sptr += TL_YUV_WIDTH;
      dptr += TL_RGB_WIDTH;
    }
  }


  return 0;
}



///  HSI to YUV format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_hsi2yuv (tlPixel *src, 
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_hsi2yuv (src, width, height, NULL, rect, dest));
}



///  HSI to YUV format conversion
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \note
///  For transform details, see:\n
///  http://www.answers.com/topic/yuv-rgb-conversion-formulas

int
tl_hsi2yuv (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  tlPixel *tmp;


  tmp = tl_create_array (width, height, TL_RGB_WIDTH);
  if (0 > tl_hsi2rgb (src, width, height, mask, rect, tmp)) {
    free (tmp);
    return -1;
  }
  if (0 > tl_rgb2yuv (tmp, width, height, mask, rect, dest)) {
    free (tmp);
    return -1;
  }

  free (tmp);


  return 0;
}



///  YUV to HSI format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_yuv2hsi (tlPixel *src, 
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_yuv2hsi (src, width, height, NULL, rect, dest));
}



///  YUV to HSI format conversion
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \note
///  For transform details, see:\n
///  http://www.answers.com/topic/yuv-rgb-conversion-formulas

int
tl_yuv2hsi (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  tlPixel *tmp;


  tmp = tl_create_array (width, height, TL_RGB_WIDTH);
  if (0 > tl_yuv2rgb (src, width, height, mask, rect, tmp)) {
    free (tmp);
    return -1;
  }
  if (0 > tl_rgb2hsi (tmp, width, height, mask, rect, dest)) {
    free (tmp);
    return -1;
  }

  free (tmp);


  return 0;
}




///  YUV to NRG format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_yuv2nrg (tlPixel *src,
            int     width, 
            int     height, 
            tlRect  *rect,
            tlPixel *dest)
{
  return (tl_yuv2nrg (src, width, height, NULL, rect, dest));
}



///  YUV to NRG format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  For transform details, see:\n
///  http://www.answers.com/topic/yuv-rgb-conversion-formulas


int
tl_yuv2nrg (tlPixel *src, 
            int     width, 
            int     height, 
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  register tlPixel *sptr, *dptr, *mptr;

  int   i, j;
  int   x1, x2;
  int   y1, y2;
  int   offset;
  float sum;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_yuv2nrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_YUV_WIDTH == 3);
  assert (TL_NRG_WIDTH == 2);
  for (j=y1; j<y2; j++) {
    offset = j*width+x1;
    sptr = src  + TL_YUV_WIDTH*offset;
    dptr = dest + TL_NRG_WIDTH*offset;
    mptr = mask + offset;
    for (i=x1; i<x2; i++) {
      if (!mask || (*mptr++)) {
        sum = (3.492f*(*(sptr)-16) + 1.627f*(*(sptr+1)-128) + 0.783f*(*(sptr+2)-128));
        *(dptr)   = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (int)(TL_MAX_PIXEL_VAL * (1.164f*(*(sptr)-16) + 1.596f*(*(sptr+2)-128)                         )/sum), TL_MAX_PIXEL_VAL);
        *(dptr+1) = (tlPixel) TL_BOUND (TL_MIN_PIXEL_VAL, (int)(TL_MAX_PIXEL_VAL * (1.164f*(*(sptr)-16) - 0.813f*(*(sptr+2)-128) - 0.391f*(*(sptr+1)-128))/sum), TL_MAX_PIXEL_VAL);

      }
      sptr += TL_YUV_WIDTH;
      dptr += TL_NRG_WIDTH;
    }
  }


  return 0;
}



///  YUV to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_yuv2gray (tlPixel *src,
             int     width, 
             int     height, 
             tlRect  *rect,
             tlPixel *dest)
{ 
  return (tl_yuv2gray (src, width, height, NULL, rect, dest));
}



///  YUV to GRAY format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  For transform details, see:\n
///  http://www.answers.com/topic/yuv-rgb-conversion-formulas


int
tl_yuv2gray (tlPixel *src, 
             int     width, 
             int     height, 
             tlPixel *mask,
             tlRect  *rect, 
             tlPixel *dest)
{
  register tlPixel *sptr, *dptr, *mptr;

  int   i, j;
  int   x1, x2;
  int   y1, y2;
  int   offset;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_yuv2gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_YUV_WIDTH  == 3);
  assert (TL_GRAY_WIDTH == 1);
  for (j=y1; j<y2; j++) {
    offset = j*width+x1;
    sptr = src  + TL_YUV_WIDTH*offset;
    dptr = dest + TL_GRAY_WIDTH*offset;
    mptr = mask + offset;
    for (i=x1; i<x2; i++) {
      if (!mask || (*mptr++)) {
        *(dptr) = *(sptr);
      }
      sptr += TL_YUV_WIDTH;
      dptr += TL_GRAY_WIDTH;
    }
  }


  return 0;
}



///  YUV to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_yuv2bin (tlPixel *src,
            int     width, 
            int     height, 
            int     min,
            int     max,
            tlRect  *rect,
            tlPixel *dest)
{ 
  return (tl_yuv2bin (src, width, height, min, max, NULL, rect, dest));
}



///  YUV to BIN format conversion.
///  
///  \param src     pixel array containing image to convert
///  \param width   image width
///  \param height  image height
///  \param min     lower binary threshold [0..255]
///  \param max     upper binary threshold [0..255]
///  \param mask    conversion mask (only converts active pixels)
///  \param rect    portion of the image to process
///  \param dest    \b [out] pixel array containing the converted image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  For transform details, see:\n
///  http://www.answers.com/topic/yuv-rgb-conversion-formulas


int
tl_yuv2bin (tlPixel *src, 
            int     width, 
            int     height, 
            int     min,
            int     max,
            tlPixel *mask,
            tlRect  *rect, 
            tlPixel *dest)
{
  register tlPixel *sptr, *dptr, *mptr;

  int   i, j;
  int   x1, x2;
  int   y1, y2;
  int   offset;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_yuv2bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN (((rect->x) + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN (((rect->y) + (rect->height)), height);

  // convert
  assert (TL_YUV_WIDTH == 3);
  assert (TL_BIN_WIDTH == 1);
  for (j=y1; j<y2; j++) {
    offset = j*width+x1;
    sptr = src  + TL_YUV_WIDTH*offset;
    dptr = dest + TL_BIN_WIDTH*offset;
    mptr = mask + offset;
    for (i=x1; i<x2; i++) {
      if (!mask || (*mptr++)) {
        if ((*(sptr) >= min) && (*(sptr) <= max)) *(dptr) = *(sptr);
        else                                      *(dptr) = 0;
      }
      sptr += TL_YUV_WIDTH;
      dptr += TL_BIN_WIDTH;
    }
  }


  return 0;
}
