/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <stdlib.h>

#include "tlHist.h"
#include "tlRect.h"



#ifndef NO_DOC

// quicksort data struct
struct _tl_hist_elem {
  tlHistData val;
  int index;
};

// quicksort comparison function for hist ranking
int
_tl_hist_compare (const void *arg1, const void *arg2)
{
  if (((_tl_hist_elem *) (arg1))->val > ((_tl_hist_elem *) (arg2))->val)	return 1;
  if (((_tl_hist_elem *) (arg1))->val == ((_tl_hist_elem *) (arg2))->val)	return 0;
  return -1;
}

#endif 



///  \file tlHist.h
///  \brief Histogram management.

///  \class tlHist
///  \brief Histogram implementation and tools.
///
///  tlHist provides a histogram class compatible with TLIB images and objects.
///  It offers simple functions such as ranking, normalization and scaling,
///  building from images and data manipulation.

///  \var int tlHist::channel
///  Desribe the color channel the histogram was built from (R, G, B, etc.)

///  \var int *tlHist::data
///  Contain the histogram data.

///  \var int *tlHist::rank
///  Contain the index of the histogram values sorted in reversed order.
///  (rank[0] = index of the highest peak in the histogram)



///  \fn int tlHist::getMaxVal()
///  Return the highest value in the histogram.
///
///  \return 
///  The highest value in the histogram.



///  \fn int tlHist::getMinVal()
///  Return the lowest value in the histogram.
///
///  \return 
///  The lowest value in the histogram.



///  \fn int tlHist::getMaxIndex()
///  Return the index of the highest value in the histogram.
///
///  \return 
///  The index of the highest value in the histogram.



///  \fn int tlHist::getMinIndex()
///  Return the index of the lowest value in the histogram.
///
///  \return 
///  The index of the lowest value in the histogram.



///  \fn int tlHist::getLength()
///  Return the length of the histogram.
///
///  \return 
///  The length of the histogram.



///  \fn int tlHist::getOffset()
///  Return the offset of the first histogram index.
///
///  \return 
///  The offset of the first histogram value.



///  \fn int tlHist::getMean()
///  Return the mean of the data in the histogram.
///
///  \return 
///  The mean of the data in the histogram.



///  \fn int tlHist::getStdDev()
///  Return the standard deviation of the data in the histogram.
///
///  \return 
///  The standard deviation of the data in the histogram.



///  Constructor, creates an empty histogram.
/// 
///  \note
///  See \ref errors "error management" for details.

tlHist::tlHist ()
{
  this->length  = 0;
  this->offset  = 0;
  this->max     = 0;
  this->min     = 0;
  this->minVal  = 0;
  this->maxVal  = 0;
  this->channel = 0;
  this->data    = NULL;
  this->rank    = NULL;

  tl_errno = TL_NO_ERROR;
}



///  Constructor, create a histogram with a given length, where the first element
///  starts at a given offset. For example, to create a histogram ranging between 
///  10 and 30, call tlHist(20,10).
///
///  \param length      number of values in histogram
///  \param offset      first index of the histogram
/// 
///  \note
///  See \ref errors "error management" for details.

tlHist::tlHist (int length, 
                int offset)
{
  tl_errno = TL_NO_ERROR;
  create (length, offset);
}



///  Destructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlHist::~tlHist ()
{
  tl_errno = TL_NO_ERROR;
  deleteData ();
}



///  Create an histogram object. This call allocates the histogram memory
///  array and sets the addressing variables (offset,etc.) properly.
///
///  \param length      number of values in histogram
///  \param offset      first index of the histogram
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::create (int length, 
                int offset)
{
  int i;


  this->length  = length;
  this->offset  = offset;
  this->min     = 0;
  this->max     = 0;
  this->minVal  = 0;
  this->maxVal  = 0;
  this->channel = 0;

  // create hist array
  if (NULL == (data = (tlHistData *) malloc (length * sizeof (tlHistData)))) {
    TL_ERRNO_MSG ("tlHistogram->create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  // create ranking array
  if (NULL == (rank = (tlHistData *) malloc (length * sizeof (tlHistData)))) {
    TL_ERRNO_MSG ("tlHistogram->create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  // init data
  for (i = 0; i < length; i++) {
    data[i] = 0;
    rank[i] = i;
  }


  return 0;
}



///  Initialize an existing histogram object to a different size. 
///  Memory array is reallocated, every index is set to 0.
///
///  \param length      number of values in histogram
///  \param offset      first index of the histogram
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::init (int length, 
              int offset)
{
  // free current hist
  deleteData ();
  return create (length, offset);
}



///  Clear an existing histogram object. That is, set all index to 0.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::clear ()
{
  int i;


  if (!data) {
    TL_ERRNO_MSG ("tlHist::clear()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // clear all index
  for (i = 0; i < length; i++) {
    data[i] = 0;
    rank[i] = i;
  }

  update ();


  return 0;
}



///  Delete the histogram data array. Memory array is deallocated, any
///  attempt to read it directly will fail.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::deleteData ()
{
  if (data) free (data);
  if (rank) free (rank);
  return 0;
}



///  Force recomputation of all internal data such as minima, maxima, 
///  and ranking information.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::update ()
{
  int i;


  // update
  mean   = 0;
  weight = 0;
  stdDev = 0;

  // build elements, compute temp vars
  _tl_hist_elem *sortlist;

  sortlist = (_tl_hist_elem *) malloc (length * sizeof (_tl_hist_elem));
  for (i = 0; i < length; i++) {
    register int val = data[i];
    sortlist[i].val = val;
    sortlist[i].index = i;
    mean += i * val;
    weight += val;
  }

  // mean
  if (weight != 0) mean /= weight;
  else mean = 0;

  // sort
  qsort (sortlist, length, sizeof (_tl_hist_elem), _tl_hist_compare);

  // get result
  minVal = sortlist[0].val;
  min = sortlist[0].index;
  maxVal = sortlist[length - 1].val;
  max = sortlist[length - 1].index;
  for (i = length - 1; i >= 0; i--) {
    rank[i] = sortlist[i].index;
    stdDev += (float) (data[i] * ((i - mean) * (i - mean)));
  }

  // standard deviation
  if (weight != 0) stdDev /= (float) weight;
  else stdDev = 0.0f;
  mean += offset;

  // cleanup
  free (sortlist);
  
  
  return 0;
}



///  Set color channel contained in the histogram. This is provided as
///  as an reference tool and does not affect the content or the
///  processing of the histogram.
///
///  \param channel     the desired channel as described in tlVision.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::setChannel (int channel)
{
  this->channel = channel;
  return 0;
}



///  Get the value of a given histogram index.
///
///  \param index     the index to poll
///
///  \return
///  The histogram value for the given element  on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::getVal (int index)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist::getVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }


  return data[index - offset];
}



///  Get the rank of a given histogram index.
///  getRank(0) returns the index of the highest histogram value.
///  getRank(length) returns the index of the smallest histogram value.
///
///  \param index     the rank index to poll
///
///  \return
///  The index at the given rank  on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::getRank (int index)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist::getRank()", TL_ERROR_NULL_POINTER);
    return -1;
  }


  return rank[index];
}



///  Set the value of a given histogram index.
///
///  \param index     the index to set
///  \param val       the new value of the index
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::setVal (int index, 
                int val)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist::setVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // set value
  data[index - offset] = val;
  update ();


  return 0;
}



///  Increment the value of a given histogram index.
///
///  \param index     the index to increment
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::incVal (int index)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist::incVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // set value
  data[index - offset] += 1;
  update ();


  return 0;
}



///  Decrement the value of a given histogram index
///
///  \param index     the index to decrement
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::decVal (int index)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist::decVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // set value
  data[index - offset] -= 1;
  update ();


  return 0;
}



///  Scale all histogram values proportionally so that the greatest value
///  matches a desired value. The default (newMax=100) sets all index 
///  to their weight in the histogram in [%].
///
///  \param newMax       the desired greatest value
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::normalize (int newMax)
{
  int   i;
  float fact = (float) newMax / (float) maxVal;


  // process
  for (i = 0; i < length; i++) {
    data[i] = (tlHistData) ((float) (data[i]) * fact);
  }

  // update
  maxVal = newMax;
  minVal = (tlHistData) (minVal * fact);


  return 0;
}



///  Scale all histogram values proportionally so that the sum of all 
///  values matches a desired value (as closely as possible).
///
///  \param sum       the desired sum value
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::integrate (int sum)
{
  int        i;
  int        hsum     = 0;
  tlHistData newMax;


  // compute current sum
  for (i = 0; i < length; i++) {
    hsum += data[i];
  }

  // compute new max value
  newMax = maxVal * sum / hsum;


  // normalize to new value
  return this->normalize (newMax);
}



///  Build a histogram from a tlPixel array.
///
///  \param src         pixel data set to build upon
///  \param len         pixel data set length
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  This routine first goes through the data array once
///  to determine the length and offset of the data, which are then 
///  set automatically. See build(tlPixel *src, int len, int length, int offset)
///  to manually set the offset.
///
///  \sa
///  build(tlPixel *src, int len, int length, int offset)
///  build(tlPixel *src, int width, int height, int pixelWidth, tlPixel *mask, int channel, tlRect *rect)

int
tlHist::build (tlPixel *src, 
               int     len)
{
  int  i;
  int  lmin, lmax;


  // set initial values
  lmin = 99999999;
  lmax = -99999999;

  // compute offset and length
  for (i = 0; i < len; i++) {
    register int val = src[i];
    if (val > lmax) lmax = val;
    if (val < lmin)	lmin = val;
  }


  // build histogram
  return build (src, len, (lmax - lmin), lmin);
}



///  Build a histogram from a tlPixel array.
///
///  \param src         pixel data set to build upon
///  \param len         pixel data set length
///  \param length      number of values in histogram
///  \param offset      first index of the histogram
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  See build(tlPixel *src, int len) to set the offset and length
///  parameters automatically.
///
///  \sa
///  build(tlPixel *src, int len, int offset)
///  build(tlPixel *src, int width, int height, int pixelWidth, tlPixel *mask, int channel, tlRect *rect)

int
tlHist::build (tlPixel *src, 
               int     len, 
               int     length, 
               int     offset)
{
  int i;


  // init data
  init (length, offset);

  // update
  for (i = 0; i < len; i++) {
    register int val = src[i];
    if ((val >= (length + offset)) || (val < offset)) {
      TL_ERRNO_MSG ("tlHist::build()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
      return -1;
    }
    data[val - offset] += 1;
  }


  // set min/max
  update ();
  return 0;
}



///  Build a histogram from the given channel of a portion 
///  of a given pixel array.
///
///  \param src         the pixel array
///  \param width       image width
///  \param height      image height
///  \param pixelWidth  bytes per pixel
///  \param mask        pixel mask
///  \param channel     the channel to consider
///  \param rect        the portion of the image to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::build (tlPixel *src, 
               int     width, 
               int     height, 
               int     pixelWidth, 
               tlPixel *mask, 
               int     channel, 
               tlRect  *rect)
{
  register int index;
  register int mask_index;

  int i, j;
  int x1, x2;
  int y1, y2;


  // check
  if ((!src) || (!rect)) {
    TL_ERRNO_MSG ("tlHist::build()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // reset histogram
  init (TL_MAX_PIXEL_VAL - TL_MIN_PIXEL_VAL + 1);

  // process
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = pixelWidth * mask_index + channel;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        data[src[index]] += 1;
      }
      mask_index += 1;
      index += pixelWidth;
    }
  }


  // set min/max
  update ();
  return 0;
}



///  Compare the histogram with another histogram.
///
///  \param hist the histogram to compare with
///
///  \return
///  A value representing the similitude of the histograms [0;1]\n
///  (1 meaning similar, 0 meaning radically different).\n
///  Returns -1 if the argument histogram is empty.
///
///  \note
///  Both histogram must be of the same length.

float
tlHist::compare (tlHist * hist)
{
  int        i;
  int        sum;
  int        coeff;
  tlHistData *hdata;


  // test
  if (length != hist->getLength ()) {
    TL_ERRNO_MSG ("tlHist::compare()", TL_ERROR);
    return -1;
  }

  // init
  sum   = 0;
  coeff = 0;

  // shortcuts
  hdata = hist->data;

  // compute intersection
  for (i = 0; i < length; i++) {
    sum += hdata[i];
    coeff += TL_MIN (hdata[i], data[i]);
  }

  if (sum != 0) return (float) (coeff) / (float) (sum);
  else          return -1.0f;
}



///  Build a histogram from a high-resolution pixel array.
///
///  \param src         pixel data set to build upon
///  \param len         pixel data set length
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  This routine first goes through the data array once
///  to determine the length and offset of the data, which are then 
///  set automatically. See build(tlPixel *src, int len, int length, int offset)
///  to manually set the offset.
///
///  \sa
///  build(tlHRPixel *src, int len, int length, int offset)
///  build(tlHRPixel *src, int width, int height, int pixelWidth, tlPixel *mask, int channel, tlRect *rect)

int
tlHist::build (tlHRPixel *src, 
               int       len)
{
  register tlHRPixel *psrc;

  int       i;
  tlHRPixel lmin, lmax;


  // shortcuts
  psrc = src;

  // set initial values
  lmin =  32767;
  lmax = -32767;

  // compute offset and length
  for (i = 0; i < len; i++) {
    register tlHRPixel val = *(psrc++);
    if (val > lmax) lmax = val;
    if (val < lmin)	lmin = val;
  }


  // build histogram
  return build (src, len, (lmax - lmin), lmin);
}



///  Build a histogram from a high-resolution pixel array.
///
///  \param src         pixel data set to build upon
///  \param len         pixel data set length
///  \param length      number of values in histogram
///  \param offset      first index of the histogram
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  See build(tlHRPixel *src, int len) to set the offset and length
///  parameters automatically.
///
///  \sa
///  build(tlHRPixel *src, int len, int offset)
///  build(tlHRPixel *src, int width, int height, int pixelWidth, tlPixel *mask, int channel, tlRect *rect)

int
tlHist::build (tlHRPixel *src, 
               int       len, 
               int       length, 
               int       offset)
{
  register tlHRPixel *psrc;

  int i;


  // shortcuts
  psrc = src;

  // init data
  init (length, offset);

  // update
  for (i = 0; i < len; i++) {
    register tlHRPixel val = *(psrc++);
    if ((val >= (length + offset)) || (val < offset)) {
      TL_ERRNO_MSG ("tlHist::build()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
      return -1;
    }
    data[val-offset] += 1;
  }


  return update ();
}



///  Build a histogram from the given channel of a portion 
///  of a given high-resolution pixel array.
///
///  \param src         the pixel array
///  \param width       image width
///  \param height      image height
///  \param pixelWidth  bytes per pixel
///  \param mask        pixel mask
///  \param channel     the channel to consider
///  \param rect        the portion of the image to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist::build (tlHRPixel *src, 
               int       width, 
               int       height, 
               int       pixelWidth, 
               tlPixel   *mask, 
               int       channel, 
               tlRect    *rect)
{
  register int index;
  register int mask_index;

  int i, j;
  int x1, x2;
  int y1, y2;


  // check
  if ((!src) || (!rect)) {
    TL_ERRNO_MSG ("tlHist::build()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // reset histogram
  init (TL_MAX_PIXEL_VAL - TL_MIN_PIXEL_VAL + 1);

  // process
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = pixelWidth * mask_index + channel;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        data[src[index]] += 1;
      }
      mask_index += 1;
      index += pixelWidth;
    }
  }


  return update ();
}
