/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <stdlib.h>

#ifdef LINUX
#include <string.h>
#include <ctype.h>
#endif

#include "tlHist2D.h"
#include "tlHist.h"
#include "tlRect.h"
#include "tlImage.h"



#ifndef NO_DOC

// quicksort data struct
struct _tl_hist2D_elem {
  tlHistData val;
  int x;
  int y;
};

// quicksort comparison function for hist ranking
int
_tl_hist2D_compare (const void *arg1, const void *arg2)
{
  if (((_tl_hist2D_elem *) (arg1))->val > ((_tl_hist2D_elem *) (arg2))->val)       return 1;
  else if (((_tl_hist2D_elem *) (arg1))->val == ((_tl_hist2D_elem *) (arg2))->val) return 0;
  else return -1;
}

#endif



///  \file tlHist2D.h
///  \brief 2D histogram management.

///  \class tlHist2D
///  \brief 2D Histogram implementation and tools.
///
///  tlHist2D provides a 2D histogram class compatible with TLIB images and objects.
///  It offers simple functions such as ranking, normalization and scaling,
///  building from images, comparison and data manipulation. 

///  \var int tlHist2D::xChannel
///  Desribe the first color channel the histogram was built from (R, G, B, etc.)

///  \var int tlHist2D::yChannel
///  Desribe the second color channel the histogram was built from (R, G, B, etc.)

///  \var int **tlHist2D::data
///  Contain the histogram data.

///  \var int *tlHist2D::xRank
///  Contain the X index of the histogram values sorted in reversed order.
///  (rank[0] = X index of the highest peak in the histogram)

///  \var int *tlHist2D::yRank
///  Contain the Y index of the histogram values sorted in reversed order.
///  (rank[0] = Y index of the highest peak in the histogram)



///  \fn int tlHist2D::getMaxVal()
///  Return the highest value in the histogram.
///
///  \return 
///  The highest value in the histogram.



///  \fn int tlHist2D::getMinVal()
///  Return the lowest value in the histogram.
///
///  \return 
///  The lowest value in the histogram.



///  \fn int tlHist2D::getXMaxIndex()
///  Return the X index of the highest value in the histogram.
///
///  \return 
///  The X index of the highest value in the histogram.



///  \fn int tlHist2D::getXMinIndex()
///  Return the X index of the lowest value in the histogram.
///
///  \return 
///  The X index of the lowest value in the histogram.



///  \fn int tlHist2D::getYMaxIndex()
///  Return the Y index of the highest value in the histogram.
///
///  \return 
///  The Y index of the highest value in the histogram.



///  \fn int tlHist2D::getYMinIndex()
///  Return the Y index of the lowest value in the histogram.
///
///  \return 
///  The Y index of the lowest value in the histogram.



///  \fn int tlHist2D::getXLength()
///  Return the length of the histogram on the X axis.
///
///  \return 
///  The length of the histogram on the X axis.



///  \fn int tlHist2D::getXOffset()
///  Return the offset of the first histogram index on the X axis.
///
///  \return 
///  The offset of the first histogram value on the X axis.



///  \fn int tlHist2D::getYLength()
///  Return the length of the histogram on the Y axis.
///
///  \return 
///  The length of the histogram on the Y axis.



///  \fn int tlHist2D::getYOffset()
///  Return the offset of the first histogram index on the Y axis.
///
///  \return 
///  The offset of the first histogram value on the Y axis.



///  Constructor, creates an empty histogram.
/// 
///  \note
///  See \ref errors "error management" for details.

tlHist2D::tlHist2D ()
{
  this->xLength   = 0;
  this->xOffset   = 0;
  this->xMax      = 0;
  this->xMin      = 0;
  this->xChannel  = 0;
  this->yLength   = 0;
  this->yOffset   = 0;
  this->yMax      = 0;
  this->yMin      = 0;
  this->yChannel  = 0;
  this->minVal    = 0;
  this->maxVal    = 0;
  this->data      = NULL;
  this->xRank     = NULL;
  this->yRank     = NULL;

  tl_errno = TL_NO_ERROR;
}



///  Constructor, create a square 2D histogram with a given number of bins.
///
///  \param length      number of values in histogram on each axis
/// 
///  \note
///  See \ref errors "error management" for details.

tlHist2D::tlHist2D (int length)
{
  tl_errno = TL_NO_ERROR;
  create (length, length, 0, 0);
}



///  Create a 2D histogram with a given number of bins, where the first
///  element of each dimension starts at a given offset.
///
///  For example, to create a histogram with both axis ranging between 10 and 30,
///  call tlHist(20,20,10,10).
///
///  \param xlength      number of values in histogram on the X axis
///  \param xoffset      first index of the histogram on the X axis
///  \param ylength      number of values in histogram on the Y axis
///  \param yoffset      first index of the histogram on the Y axis
/// 
///  \note
///  See \ref errors "error management" for details.

tlHist2D::tlHist2D (int xlength, int ylength, int xoffset, int yoffset)
{
  tl_errno = TL_NO_ERROR;
  create (xlength, ylength, xoffset, yoffset);
}



///  Destructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlHist2D::~tlHist2D ()
{
  tl_errno = TL_NO_ERROR;
  deleteData ();
}



///  Create a 2D histogram object. This call allocates the histogram memory
///  array and sets the addressing variables (offset,etc.) properly.
///
///  \param xlength      number of values in histogram on the X axis
///  \param xoffset      first index of the histogram on the X axis
///  \param ylength      number of values in histogram on the Y axis
///  \param yoffset      first index of the histogram on the Y axis
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::create (int xlength, 
                  int ylength, 
                  int xoffset, 
                  int yoffset)
{
  int i, j;


  this->minVal    = 0;
  this->maxVal    = 0;
  this->xLength   = xlength;
  this->xOffset   = xoffset;
  this->xMin      = 0;
  this->xMax      = 0;
  this->xChannel  = 0;
  this->yLength   = ylength;
  this->yOffset   = yoffset;
  this->yMin      = 0;
  this->yMax      = 0;
  this->yChannel  = 0;

  // create hist array
  if (NULL == (data = (tlHistData **) malloc (xlength * sizeof (tlHistData *)))) {
    TL_ERRNO_MSG ("tlHistogram->create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  for (i = 0; i < ylength; i++) {
    if (NULL == (data[i] = (tlHistData *) malloc (xlength * sizeof (tlHistData)))) {
      TL_ERRNO_MSG ("tlHistogram->create()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }

  // create X ranking array
  if (NULL == (xRank = (tlHistData *) malloc (xlength * ylength * sizeof (tlHistData)))) {
    TL_ERRNO_MSG ("tlHistogram->create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // create Y ranking array
  if (NULL == (yRank = (tlHistData *) malloc (xlength * ylength * sizeof (tlHistData)))) {
    TL_ERRNO_MSG ("tlHistogram->create()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // init data
  for (i = 0; i < xlength; i++) {
    for (j = 0; j < ylength; j++) {
      data[i][j] = 0;
    }
    xRank[i] = 0;
  }
  for (j = 0; j < ylength; j++) {
    yRank[j] = 0;
  }


  return 0;
}



///  Initialize an existing 2D histogram object to a different square size. 
///  Memory array is reallocated, every index is set to 0.
///
///  \param length      number of values in histogram on each axis
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::init (int length)
{
  // free current hist
  deleteData ();
  return create (length, length, 0, 0);
}



///  Initialize an existing 2D histogram object to a different size. 
///  Memory array is reallocated, every index is set to 0.
///
///  \param xlength      number of values in histogram on the X axis
///  \param xoffset      first index of the histogram on the X axis
///  \param ylength      number of values in histogram on the Y axis
///  \param yoffset      first index of the histogram on the Y axis
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::init (int xlength, 
                int ylength, 
                int xoffset, 
                int yoffset)
{
  // free current hist
  deleteData ();
  return create (xlength, ylength, xoffset, yoffset);
}



///  Clear an existing histogram object. That is, set all index to 0.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::clear ()
{
  int i,j;


  if (!data) {
    TL_ERRNO_MSG ("tlHist2D::clear()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // clear all index
  for (i = 0; i < xLength; i++) {
    for (j = 0; j < yLength; j++) {
      data[i][j] = 0;
    }
    xRank[i] = 0;
  }
  for (j = 0; j < yLength; j++) {
    yRank[j] = 0;
  }
  update ();


  return 0;
}



///  Delete the histogram data array. Memory array is deallocated, any
///  attempt to read it directly will fail.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::deleteData ()
{
  int i;


  if (data) {
    for (i = 0; i < xLength; i++) {
      if (data[i])
        free (data[i]);
    }
    free (data);
  }

  if (xRank) free (xRank);
  if (yRank) free (yRank);


  return 0;
}



///  Force recomputation of all internal data such as minima, maxima.
///  It is only necessary to call this routine if the histogram values
///  have been change manually (i.e. directly in the tlHist2D::data field).
///  Calls to setVal(), incVal() and decVal() update this information
///  automatically.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::update ()
{
  int i,j;


  minVal = 999999999;
  maxVal = -999999999;
  for (i = 0; i < xLength; i++) {
    for (j = 0; j < yLength; j++) {
      register int val = data[i][j];
      if (val < minVal) {
        xMin = i;
        yMin = j;
        minVal = val;
      }
      else if (val > maxVal) {
        xMax = i;
        yMax = j;
        maxVal = val;
      }
    }
  }


  return 0;
}


///  Compute ranking information. Depending on the size of the histogram,
///  this operation can be very time consuming !
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::updateRank ()
{
  int i,j;


  // build elements
  _tl_hist2D_elem * sortlist;
  sortlist = (_tl_hist2D_elem *) malloc (xLength * yLength * sizeof (_tl_hist2D_elem));
  int index = 0;
  for (i = 0; i < xLength; i++) {
    for (j = 0; j < yLength; j++) {
      sortlist[index].val = data[i][j];
      sortlist[index].x = i;
      sortlist[index++].y = j;
    }
  }

  // sort
  qsort (sortlist, xLength * yLength, sizeof (_tl_hist2D_elem),
    _tl_hist2D_compare);

  // get result
  for (i = xLength * yLength - 1; i >= 0; i--) {
    xRank[i] = sortlist[i].x;
    yRank[i] = sortlist[i].y;
  }


  // cleanup
  free (sortlist);
  return 0;
}



///  Set color channel contained in the histogram X axis. This is provided as
///  as an reference and does not affect the content or the
///  processing of the histogram.
///
///  \param channel     the desired channel as described in tlVision.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::setXChannel (int channel)
{
  this->xChannel = channel;
  return 0;
}



///  Set color channel contained in the histogram Y axis. This is provided as
///  as an reference and does not affect the content or the
///  processing of the histogram.
///
///  \param channel     the desired channel as described in tlVision.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::setYChannel (int channel)
{
  this->yChannel = channel;
  return 0;
}



///  Get the value of a given histogram index.
///
///  \param x     the X index to poll
///  \param y     the Y index to poll
///
///  \return
///  The histogram value for the given element on success, -1 otherwise. 
///  See \ref errors "error management" for details.

int
tlHist2D::getVal (int x, 
                  int y)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist2D::getVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }


  return data[x - xOffset][y - yOffset];
}



///  Get the X axis index of a histogram value at a given rank.
///  getRank(0) returns the X index of the highest histogram value.
///  getRank(xLength) returns the X index of the smallest histogram value.
///
///  \param index     the X rank index to poll
///
///  \return
///  The index at the given rank on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  The rank information must be computed by calling updateRank() first.

int
tlHist2D::getXRank (int index)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist2DgetXRank()", TL_ERROR_NULL_POINTER);
    return -1;
  }


  return xRank[index];
}



///  Get the Y axis index of a histogram value at a given rank.
///  getRank(0) returns the Y index of the highest histogram value.
///  getRank(xLength) returns the Y index of the smallest histogram value.
///
///  \param index     the Y rank index to poll
///
///  \return
///  The index at the given rank  on success, -1 otherwise. 
///  See \ref errors "error management" for details.
///
///  \note
///  The rank information must be computed by calling updateRank() first.

int
tlHist2D::getYRank (int index)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist2DgetYRank()", TL_ERROR_NULL_POINTER);
    return -1;
  }


  return yRank[index];
}



///  Set the value of a given histogram index.
///
///  \param x         the X index to set
///  \param y         the Y index to set
///  \param val       the new value of the index
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::setVal (int        x, 
                  int        y, 
                  tlHistData val)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist2D::setVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // set value
  data[x - xOffset][y - yOffset] = val;
  update ();


  return 0;
}



///  Increment the value of a given histogram index.
///
///  \param x         the X index to increment
///  \param y         the Y index to increment
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::incVal (int x, 
                  int y)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist2D::incVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  // set value
  data[x - xOffset][y - yOffset] += 1;
  update ();


  return 0;
}



///  Decrement the value of a given histogram index
///
///  \param x         the X index to decrement
///  \param y         the Y index to decrement
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::decVal (int x, 
                  int y)
{
  if (!data) {
    TL_ERRNO_MSG ("tlHist2D::decVal()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  // set value
  data[x - xOffset][y - yOffset] -= 1;
  update ();


  return 0;
}



///  Scale all histogram values proportionally so that the greatest value
///  matches a desired value. The default (newMax=100) sets all index 
///  to their weight in the histogram in [%].
///
///  \param newMax       the desired greatest value
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::normalize (int newMax)
{
  int   i, j;
  float fact  = (float) newMax / (float) maxVal;


  // process
  for (i = 0; i < xLength; i++) {
    for (j = 0; j < yLength; j++) {
      data[i][j] = (tlHistData) ((float) (data[i][j]) * fact);
    }
  }

  // update
  maxVal = newMax;
  minVal = (tlHistData) ((float) minVal * fact);


  return 0;
}



///  Scale all histogram values proportionally so that the sum of all 
///  values matches a desired value (as closely as possible).
///
///  \param sum       the desired sum value
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::integrate (int sum)
{
  register int hsum = 0;

  int        i, j;
  tlHistData newMax;


  // compute current sum
  for (i = 0; i < xLength; i++) {
    for (j = 0; j < yLength; j++) {
      hsum += data[i][j];
    }
  }

  // compute new max value
  newMax = maxVal * sum / hsum;


  // normalize to new value
  return this->normalize (newMax);
}



///  Build a 2D histogram from two tlPixel array.
///
///  \param xdata       data set (first dimension)
///  \param ydata       data set (second dimension)
///  \param len         data sets length
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  This routine first goes through the data array once
///  to determine the length and offset of each dimension, which are then 
///  set automatically. See build(tlPixel *xdata, tlPixel *ydata, int len, int xlength, int ylength, int xoffset, int yoffset)
///  to manually set the offset.
///
///  \sa
///  build(tlPixel *xdata, tlPixel *ydata, int len, int xlength, int ylength, int xoffset, int yoffset)
///  build (tlPixel *src, int width, int height, int pixelWidth, tlPixel *mask, int xchannel, int ychannel, tlRect *rect)

int
tlHist2D::build (tlPixel *xdata, 
                 tlPixel *ydata, 
                 int     len)
{
  int  i;
  int  xlmin, xlmax;
  int  ylmin, ylmax;


  // set initial values
  xlmin = 99999999;
  xlmax = -99999999;
  ylmin = 99999999;
  ylmax = -99999999;

  // compute offset and length
  for (i = 0; i < len; i++) {
    if (xdata[i] > xlmax) xlmax = xdata[i];
    if (xdata[i] < xlmin) xlmin = xdata[i];
    if (ydata[i] > ylmax) ylmax = ydata[i];
    if (ydata[i] < ylmin) ylmin = ydata[i];
  }


  // build histogram
  return build (xdata, ydata, len, (xlmax - xlmin), (ylmax - ylmin), xlmin, ylmin);
}



///  Build a histogram from two tlPixel array.
///
///  \param xdata       data set (first dimension)
///  \param ydata       data set (second dimension)
///  \param len         data sets length
///  \param xlength     number of values in histogram on the X axis
///  \param xoffset     first index of the histogram on the X axis
///  \param ylength     number of values in histogram on the Y axis
///  \param yoffset     first index of the histogram on the Y axis
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  See build(tlPixel *xdata, tlPixel *ydata, int len) to set the offset and length
///  parameters automatically.
///
///  \sa
///  build(tlPixel *xdata, tlPixel *ydata, int len)
///  build (tlPixel *src, int width, int height, int pixelWidth, tlPixel *mask, int xchannel, int ychannel, tlRect *rect)

int
tlHist2D::build (tlPixel *xdata, 
                 tlPixel *ydata, 
                 int     len, 
                 int     xlength,
                 int     ylength, 
                 int     xoffset, 
                 int     yoffset)
{
  int i;
  int xi, yi;


  // init data
  init (xlength, ylength, xoffset, yoffset);

  // update
  for (i = 0; i < len; i++) {
    xi = xdata[i];
    yi = ydata[i];
    if ((xi >= (xlength + xoffset)) || (xi < xoffset) || (yi >= (ylength + yoffset)) || (yi < xoffset)) {
      TL_ERRNO_MSG ("tlHist2D::build()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
      return -1;
    }
    data[xi - xoffset][yi - yoffset] += 1;
  }


  // set min/max
  update ();
  return 0;
}



///  Build a histogram from the given channel of a portion 
///  of a given pixel array.
///
///  \param src         the pixel array
///  \param width       image width
///  \param height      image height
///  \param pixelWidth  bytes per pixel
///  \param mask        pixel mask
///  \param xchannel    the channel to consider (first dimension)
///  \param ychannel    the channel to consider (second dimension)
///  \param rect        the portion of the image to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::build (tlPixel *src, 
                 int     width, 
                 int     height, 
                 int     pixelWidth,
                 tlPixel *mask, 
                 int     xchannel, 
                 int     ychannel, 
                 tlRect  *rect)
{
  register int i,j;
  register int index;

  int x1, x2;
  int y1, y2;
  int mask_index;


  // check
  if ((!src) || (!rect)) {
    TL_ERRNO_MSG ("tlHist2D::build()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // reset histogram
  init (TL_MAX_PIXEL_VAL - TL_MIN_PIXEL_VAL + 1);

  // process
  for (j = y1; j < y2; j++) {
    mask_index = TL_PIXEL (j, x1, width);
    index = pixelWidth * mask_index;
    for (i = x1; i < x2; i++) {
      if (!mask || mask[mask_index]) {
        data[src[index + xchannel]][src[index + ychannel]] += 1;
      }
      mask_index += 1;
      index += pixelWidth;
    }
  }


  // set min/max
  update ();
  return 0;
}



///  Compare the histogram with another histogram.
///
///  \param hist the histogram to compare with
///
///  \return
///  A value representing the similitude of the histograms [0;1]\n
///  (1 meaning similar, 0 meaning radically different).\n
///  Return -1 if argument histogram is empty.
///
///  \note
///  Both histogram must have the same dimension.

float
tlHist2D::compare (tlHist2D *hist)
{
  int        i, j;
  int        sum        = 0;
  int        coeff      = 0;
  tlHistData **hdata;


  // test
  if ((xLength != hist->getXLength ()) || (yLength != hist->getYLength ())) {
    TL_ERRNO_MSG ("tlHist2D->compare()", TL_ERROR);
    return -1;
  }

  // shortcuts
  hdata = hist->data;

  // compute intersection
  for (i = 0; i < xLength; i++) {
    for (j = 0; j < yLength; j++) {
      sum += hdata[i][j];
      coeff += TL_MIN (hdata[i][j], data[i][j]);
    }
  }


  // compute coeff
  if (sum != 0) return (float) (coeff) / (float) sum;
  else          return -1;
}



///  Project the histogram on the X axis and build the resulting unidimensional
///  histogram.
///
///  \param hist \b [out] the resulting projected histogram
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::getXHist (tlHist *hist)
{
  int i,j;
  int sum;


  hist->init (xLength, xOffset);
  for (i = 0; i < xLength; i++) {
    sum = 0;
    for (j = 0; j < yLength; j++) {
      sum += data[i][j];
    }
    hist->data[i] = sum;
  }


  return hist->update ();
}



///  Project the histogram on the Y axis and build the resulting unidimensional
///  histogram.
///
///  \param hist \b [out] the resulting projected histogram
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::getYHist (tlHist * hist)
{
  int i,j;
  int sum;


  hist->init (yLength, yOffset);
  for (j = 0; j < yLength; j++) {
    sum = 0;
    for (i = 0; i < xLength; i++) {
      sum += data[i][j];
    }
    hist->data[j] = sum;
  }


  return hist->update ();
}



///  Project the histogram on the X axis and build the resulting unidimensional
///  histogram.
///
///  \return
///  A pointer to the resulting projected histogram.
///
///  \note
///  It is the responsability of the programmer to delete the tlHist object
///  after use.

tlHist *
tlHist2D::getXHist ()
{
  tlHist *hist = new tlHist ();
  getXHist (hist);
  return hist;
}



///  Project the histogram on the Y axis and build the resulting unidimensional
///  histogram.
///
///  \return
///  A pointer to the resulting projected histogram.
///
///  \note
///  It is the responsability of the programmer to delete the tlHist object
///  after use.

tlHist *
tlHist2D::getYHist ()
{
  tlHist *hist = new tlHist ();
  getYHist (hist);
  return hist;
}



///  Converts the 2D histogram into a gray scale tlImage.
///
///  \param image \b [out] resulting image
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::toImage (tlImage *image)
{
  int     i, j;
  int     index   = 0;
  tlPixel *dest;


  if (!image) {
    TL_ERRNO_MSG ("tlHist2D->toImage()", TL_ERROR);
    return -1;
  }

  image->init (GRAY, xLength, yLength);
  dest = image->pixel;
  for (i = 0; i < xLength; i++) {
    for (j = 0; j < yLength; j++) {
      dest[index++] = (tlPixel) (data[i][j]);
    }
  }


  return 0;
}



///  Converts the 2D histogram into a gray scale tlImage.
///
///  \return
///  A pointer to the resulting tlImage object.
///
///  \note
///  It is the responsability of the programmer to delete the tlImage object
///  after use.

tlImage *
tlHist2D::toImage ()
{
  tlImage *image = new tlImage ();
  toImage (image);
  return image;
}



///  Read a 2D histogram from a file.
///
///  \param filename file name
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::load (char *filename)
{
  register int i;

  char header[7];
  FILE *pFp;


  // sanity check
  if (!filename) {
    TL_ERRNO_MSG ("tlHist2D->load()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // status message
  if (tlVerbose) {
    printf ("reading HIST2D from %s\n", filename);
  }

  // open file
  if (NULL == (pFp = fopen (filename, "rb"))) {
    TL_ERRNO_MSG ("tlHist2D->load()", TL_ERROR);	// failur)e
    return -1;
  }

  // read TLIB header
  fscanf (pFp, "%6s", header);
  if (strcmp (header, "HIST2D")) {
    TL_ERRNO_MSG ("tlHist2D->load()", TL_ERROR);
    fclose (pFp);
    return -1;
  }

  // skip any lines not whitespace & comments
  while (1) {
    // read discard up to first non-whitespace in the line
    do {
      i = fgetc (pFp);
    }
    while (isspace (i));
    if (isdigit (i)) {
      ungetc (i, pFp);
      break;
    }
    // discard the line
    else {
      do {
        i = fgetc (pFp);
      }
      while (i != '\n');
    }
  }

  // delete existing array
  for (i = 0; i < yLength; i++) {
    free (data[i]);
  }
  free (data);

  // read image parameters
  if (EOF == fscanf (pFp, "%d %d %d %d", &xLength, &yLength, &xOffset, &yOffset)) {
    TL_ERRNO_MSG ("tlHist2D->load()", TL_ERROR);
    fclose (pFp);
    return -1;
  }

  // read the whitespace character
  i = fgetc (pFp);
  if (!isspace (i)) {
    TL_ERRNO_MSG ("tlHist2D::load() ", TL_ERROR);
    return -1;
  }

  // create hist array
  if (NULL == (data = (tlHistData **) malloc (xLength * sizeof (tlHistData *)))) {
    TL_ERRNO_MSG ("tlHist2D::load()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  for (i = 0; i < yLength; i++) {
    if (NULL == (data[i] = (tlHistData *) malloc (xLength * sizeof (tlHistData)))) {
      TL_ERRNO_MSG ("tlHist2D::load()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }

  for (i = 0; i < xLength; i++) {
    // read TLIB data one line at a time (this is MUCH faster than one 
    // RGB
    // pixel at a time and slightly faster than the whole image at
    // once)
    if ((unsigned int) yLength != fread ((void *) data[i], sizeof (tlHistData), yLength, pFp)) {
      TL_ERRNO_MSG ("tlHist2D::load()", TL_ERROR_FILE_IO);
      fclose (pFp);
      return -1;
    }
  }

  // close file
  if (0 > fclose (pFp)) {
    TL_ERRNO_MSG ("tlHist2D::load()", TL_ERROR);
    return -1;
  }


  return 0;
}



///  Save a 2D histogram to a file.
///
///  \param filename file name
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHist2D::save (char *filename)
{
  register int i;

  FILE         *pFp;
  unsigned int bytesLine;


  // sanity check
  if ((!filename)) {
    TL_ERRNO_MSG ("tlHist2D::save()", (TL_ERROR_NULL_POINTER));
    return -1;
  }

  // open file for binary write
  if (NULL == (pFp = fopen (filename, "wb"))) {
    TL_ERRNO_MSG ("tlHist2D::save()", (TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT));
    return -1;
  }
  bytesLine = yLength;

  // write TLB header 
  fprintf (pFp, "HIST2D\n%d %d\n%d\n%d\n", xLength, yLength, xOffset, yOffset);

  for (i = 0; i < xLength; i++) {
    if (bytesLine != fwrite ((void *) data[i], sizeof (tlHistData), bytesLine, pFp)) {
      TL_ERRNO_MSG ("tlHist2D::save()", TL_ERROR_FILE_IO);
      fclose (pFp);
      return -1;
    }
  }

  // close file
  if (0 > fclose (pFp)) {
    TL_ERRNO_MSG ("tlHist2D::save()", (TL_ERROR_FILE_IO));
    return -1;
  }
  return (0);
}



///  Build a 2D histogram from two high-resolution pixel arrays.
///
///  \param xdata       data set (first dimension)
///  \param ydata       data set (second dimension)
///  \param len         data sets length
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  This routine first goes through the data array once
///  to determine the length and offset of each dimension, which are then 
///  set automatically. See build(tlPixel *xdata, tlPixel *ydata, int len, int xlength, int ylength, int xoffset, int yoffset)
///  to manually set the offset.
///
///  \sa
///  build(tlHRPixel *xdata, tlHRPixel *ydata, int len, int xlength, int ylength, int xoffset, int yoffset)

int
tlHist2D::build (tlHRPixel *xdata, 
                 tlHRPixel *ydata, 
                 int       len)
{
  int  i;
  int  xlmin, xlmax;
  int  ylmin, ylmax;


  // set initial values
  xlmin = 99999999;
  xlmax = -99999999;
  ylmin = 99999999;
  ylmax = -99999999;

  // compute offset and length
  for (i = 0; i < len; i++) {
    if (xdata[i] > xlmax) xlmax = xdata[i];
    if (xdata[i] < xlmin) xlmin = xdata[i];
    if (ydata[i] > ylmax) ylmax = ydata[i];
    if (ydata[i] < ylmin) ylmin = ydata[i];
  }


  // build histogram
  return build (xdata, ydata, len, (xlmax - xlmin), (ylmax - ylmin), xlmin, ylmin);
}



///  Build a histogram from two high-resolution pixel arrays.
///
///  \param xdata       data set (first dimension)
///  \param ydata       data set (second dimension)
///  \param len         data sets length
///  \param xlength     number of values in histogram on the X axis
///  \param xoffset     first index of the histogram on the X axis
///  \param ylength     number of values in histogram on the Y axis
///  \param yoffset     first index of the histogram on the Y axis
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  See build(tlPixel *xdata, tlPixel *ydata, int len) to set the offset and length
///  parameters automatically.
///
///  \sa
///  build(tlPixel *xdata, tlPixel *ydata, int len)
///  build (tlPixel *src, int width, int height, int pixelWidth, tlPixel *mask, int xchannel, int ychannel, tlRect *rect)

int
tlHist2D::build (tlHRPixel *xdata, 
                 tlHRPixel *ydata, 
                 int       len, 
                 int       xlength,
                 int       ylength, 
                 int       xoffset, 
                 int       yoffset)
{
  int i;
  int xi, yi;


  // init data
  init (xlength, ylength, xoffset, yoffset);

  // update
  for (i = 0; i < len; i++) {
    xi = xdata[i];
    yi = ydata[i];
    if ((xi >= (xlength + xoffset)) || (xi < xoffset) || (yi >= (ylength + yoffset)) || (yi < xoffset)) {
      TL_ERRNO_MSG ("tlHist2D::build()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
      return -1;
    }
    data[xi-xoffset][yi-yoffset] += 1;
  }


  // set min/max
  update ();
  return 0;
}


