/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://vrai-group.epfl.ch/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlHough.h"
#include "tlLine.h"
#include "tlImage.h"
#include "tlPoint.h"
#include "tlHist.h"



///  \file tlHough.h
///  \brief Hough transform for geometric feature extraction.

///  \class tlHough
///  This class implements a Hough transform of a tlImage in order to identify straight line segments.
///  The number of lines to extract and the angular resolution of the line construction can be adjusted
///  to suit the application and/or to optimize performance.



///  Constructor.
///
///  \param sWidth    input image width.
///  \param sHeight   input image height.
///  \param angIncr   the angular resolution in [deg]. The default is optimized with respect to the image input size.
///
///  \return
///  See \ref errors "error management" for details.

tlHough::tlHough (int   sWidth,
                  int   sHeight,
                  float angIncr)
{
  accumulator = NULL;
  sinTheta    = NULL;
  cosTheta    = NULL;
  
  if (setParam (sWidth, sHeight, angIncr) != 0) tl_errno = TL_ERROR;
	else                                          tl_errno = TL_NO_ERROR;
}



///  Destructor.
///
///  \return
///  See \ref errors "error management" for details.

tlHough::~tlHough ()
{
  if (accumulator) delete [] accumulator;
  if (sinTheta)    delete [] sinTheta;
  if (cosTheta)    delete [] cosTheta;
}



///  Configure the angular resolution of the Hough line representation.
///
///  \param angIncr    the angular resolution in [deg].
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHough::setAngularIncrement (float angIncr)
{
  if (angIncr <= 0.0f) {
    angularIncrement = (float)(TL_PI/(2.0f*srcWidth));
    hHeight = 2*srcWidth;
  }
  else {
    angularIncrement = (float)(TL_PI*angIncr/180.0f);
    hHeight = (int)(180.0f/angIncr);
  }
  
  
  return makeAccumulator ();
}



///  Initialize the Hough accumulators for a given input image size.
///
///  \param sWidth    input image width.
///  \param sHeight   input image height.
///  \param angIncr   the angular resolution in [deg]. The default is optimized with respect to the image input size.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHough::setParam (int   sWidth,
                   int   sHeight,
                   float angIncr)
{
  srcWidth  = sWidth;
  srcHeight = sHeight;
  hOffset   = (int)(sqrt((float)(sWidth*sWidth+sHeight*sHeight)));
  hWidth    = 2*hOffset;
  
  return setAngularIncrement (angIncr);
}



///  Extract a given number of lines from a single-channel image.
///  The lines are sorted by weight (i.e. the first one in the list
///  has the more pixels contributing to it).
///
///  \param image       single-channel image to process.
///  \param lineCount   number of lines to extract.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHough::extract (tlImage *image,
                  int     lineCount)
{
  int              i,j;
  register tlPixel *val;
  register tlPixel *pixel;
  int              aIndex;


  // check
  if (!image) {
    TL_ERRNO_MSG ("tlHough::extract()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  
  // more check
  if (image->hr) {
    TL_ERRNO_MSG ("tlHough::extract()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check format
  if (image->pixelWidth != 1) {
    TL_ERRNO_MSG ("tlHough::extract()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // shortcuts
  pixel = image->pixel;

	// init destination
  clearAccumulator();

  // convert
	for (j=1; j<srcHeight-2; j++) {
    val = pixel + j*srcWidth + 1;
		for (i=1; i<srcWidth-2; i++) {
			if (*(val++)) {
        for (int n=0; n<hHeight; n++) {
          int r = ((i*cosTheta[n] + j*sinTheta[n]) >> 10) + hOffset;
          aIndex = n*hWidth + r;
          accumulator[aIndex] += 1;
          if (accumulator[aIndex] > maxAcc) maxAcc = accumulator[aIndex];
        }
      }
		}
	}

  extractLines (lineCount);
	
	
	return 0;
}



///  Extract a given number of lines from a Hough accumulator.
///  The lines are sorted by weight (i.e. the first one in the list
///  has the more pixels contributing to it).
///
///  \param lineCount   number of lines to extract.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHough::extractLines (int lineCount)
{
  tlHist        *hist = new tlHist (maxAcc+1);
  short         *acc;
  int           i,j;
  unsigned int  k;
  short         m;
  int           length = hWidth * hHeight;
  vector<short> scoreList;
  int           count;
  tlLine        *line;


  // sort accumulators
  for (i=0; i<length; i++) {
    if (accumulator[i] > 2) hist->data[accumulator[i]] += 1;
  }

  // get requested number of lines
  count = 0;
  for (m=maxAcc; m>=0; m--) {
    if (hist->data[m] > 0) {
      count += hist->data[m];
      scoreList.push_back(m);
    }
    if (count > lineCount) {
      break;
    }
  }

  // detect them and put them in a line
  for (k=0; k<scoreList.size(); k++) {
    short val = scoreList[k];
    acc = accumulator;
    for (j=0; j<length; j++) {
      if ((*acc) == val) {
        float theta = (float)(j/hWidth);
        float r     = (float)((j%hWidth)-hOffset);
        theta *= angularIncrement;
        line = new tlLine (theta, r);
        line->clip (srcWidth, srcHeight);
        lineList.push_back (line);
      }
      acc++;
      if (lineList.size() == (unsigned int)lineCount) break;
    }
    if (lineList.size() == (unsigned int)lineCount) break;
  }

  delete hist;


  return 0;
}



///  Return a pointer to a copy of the detected line at a given index.
///
///  \param index   the ranking of the line to return.
///
///  \return
///  a pointer to a \ref tlLine object on success, NULL otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  It is the responsibility of the programmer to delete the returned \ref tlLine object.

tlLine *
tlHough::getLine (int index)
{
  if (index < 0 || index >= (int)(lineList.size())) {
    TL_ERRNO_MSG ("tlHough::getLine()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
    return NULL;
  }


  return new tlLine(lineList[index]);
}



///  Return a pointer to the detected line at a given index.
///
///  \param index   the ranking of the line to retrieve.
///
///  \return
///  a pointer to a \ref tlLine object on success, NULL otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  The pointer is managed by the \ref tlHough object, and the memory block
///  it references will be freed by the \ref tlHough object eventually. Use
///  at your own risk !

tlLine *
tlHough::getLinePtr (int index)
{
  if (index < 0 || index >= (int)(lineList.size())) {
    TL_ERRNO_MSG ("tlHough::getLinePtr()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
    return NULL;
  }


  return lineList[index];
}



///  Return a copy of the detected line at a given index.
///
///  \param line    a pointer to the destination line object.
///  \param index   the ranking of the line to return.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlHough::getLineCopy (tlLine *line, 
                      int     index)
{
  // check
  if (index < 0 || index >= (int)(lineList.size())) {
    TL_ERRNO_MSG ("tlHough::getLineCopy()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
    return -1;
  }


  return lineList[index]->copyTo (line);
}



///  Return the actual number of lines detected.
///
///  \return
///  The number of lines.\n
///  See \ref errors "error management" for details.

int
tlHough::getLineCount ()
{
  return (int)(lineList.size());
}



///  Allocate the Hough accumulator and the sin/cos LUTs 
///  used to speed up projection computation.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlHough::makeAccumulator ()
{
  float ang = 0.0f;


  // reallocate arrays
  accumulator = (short *)realloc(accumulator, hWidth*hHeight*sizeof(short));
  sinTheta    = (short *)realloc(sinTheta,    hHeight*sizeof(short));
  cosTheta    = (short *)realloc(cosTheta,    hHeight*sizeof(short));
  
  // compute LUT content
  for (int a=0; a<hHeight; a++) {
    sinTheta[a] = (short)(1024.0f*sin(ang));
    cosTheta[a] = (short)(1024.0f*cos(ang));
    ang += angularIncrement;
  }


  return 0;
}



///  Empty the Hough accumulators.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.


int 
tlHough::clearAccumulator ()
{
  unsigned int i;


  memset (accumulator, 0, hWidth*hHeight*sizeof(short));
  maxAcc = 0;
  for (i=0; i<lineList.size(); i++) {
    delete lineList[i];
  }
  lineList.clear();


  return 0;
}



///  Fill-in a tlImage object with the Hough accumulator.
///  The resuling tlImage is fit for display with tlDisplay.
///
///  \param image    image object to fill-in.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlHough::getHoughImage (tlImage *image)
{
  tlPixel *ipt;
  short   *acc;
  int     length;


  // check format
	if (!image) {
		TL_ERRNO_MSG ("tlHough::getHoughImage()", TL_ERROR_NULL_POINTER);
		return -1;
	}

  // init destination
  image->init (GRAY, hWidth, hHeight);

  // shortcuts
  ipt    = image->pixel;
  acc    = accumulator;
  length = hWidth * hHeight;

  // convert
  for (int i=0; i<length; i++) {
    *(ipt++) = (tlPixel)(*(acc++)*255/maxAcc);
  }


  return 0;
}
