/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <memory.h>

#include "tlImage.h"



///  Crop an image to a given area
/// 
///  \param rect area to keep
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::crop (tlRect * rect)
{
  return crop (rect->x, rect->y, rect->width, rect->height);
}



///  Crop an image to a given area
/// 
///  \param x       crop area starting point on X axis [pix]
///  \param y       crop area starting point on Y axis [pix]
///  \param cwidth  crop area width [pix]
///  \param cheight crop area height [pix]
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::crop (int x, 
               int y, 
               int cwidth, 
               int cheight)
{
  int     j;
  int     y1, y2;
  int     old_width;
  tlPixel *temp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::crop()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // make sure it all makes sense
  if ((x < 0) || (y < 0) ||
    ((x+cwidth) > width) || ((y+cheight) > height)) {
      TL_ERRNO_MSG ("tlImage::crop()", TL_ERROR_GEOMETRY);
      return -1;
    }

    // shortcuts
    y1        = TL_MAX (0, y);
    y2        = TL_MIN (height, (y + cheight));
    old_width = width;
    temp      = tl_copy_array (pixel, width, height, pixelWidth);
    if (temp == NULL) {
      TL_ERRNO_MSG ("tlImage::crop()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }

    // reset
    this->init (format, cwidth, cheight);

    // reset
    for (j = 0; j < (y2 - y1); j++) {
      memcpy (pixel + j * width * pixelWidth, temp + ((j + y1) * old_width + x) * pixelWidth, cwidth * pixelWidth);
    }

    // cleanup
    free (temp);


    return 0;
}



///  Crop an image to a given area
///  Result is placed in another image (original is unmodified).
/// 
///  \param rect    area to keep
///  \param dest    destination image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::crop (tlRect  *rect,
               tlImage *dest)
{
  return this->crop (rect->x, rect->y, rect->width, rect->height, dest);
}



///  Crop an image to a given area.
///  Result is placed in another image (original is unmodified).
/// 
///  \param x       crop area starting point on X axis [pix]
///  \param y       crop area starting point on Y axis [pix]
///  \param cwidth  crop area width [pix]
///  \param cheight crop area height [pix]
///  \param dest    destination image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
//////////////////////////////////////////////////////////////////////// 

int
tlImage::crop (int     x, 
               int     y, 
               int     cwidth, 
               int     cheight,
               tlImage *dest)
{
  register tlPixel *src;
  register tlPixel *ptr;

  int j;
  int y1, y2;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::crop()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // make sure it all makes sense
  if ((x < 0) || (y < 0) ||
    ((x+cwidth) > width) || ((y+cheight) > height)) {
      TL_ERRNO_MSG ("tlImage::crop()", TL_ERROR_GEOMETRY);
      return -1;
    }

    // shortcuts
    y1 = TL_MAX (0, y);
    y2 = TL_MIN (height, (y + cheight));

    // reset
    dest->init (format, cwidth, cheight);
    src = pixel + (y1*width + x)*pixelWidth;
    ptr = dest->pixel;

    // reset
    for (j = 0; j < (y2 - y1); j++) {
      memcpy (ptr, src, cwidth * pixelWidth);
      ptr += cwidth * pixelWidth;
      src += width  * pixelWidth;
    }


    // cleanup
    return 0;
}



///  Add black pixels around current image
/// 
///  \param b border thickness [pix]
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderBlack (int b)
{
  return borderBlack (b, b);
}



///  Add black pixels around current image
/// 
///  \param bx border thickness on the X axis [pix]
///  \param by border thickness on the Y axis [pix]
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderBlack (int bx, 
                      int by)
{
  int     j;
  int     index, old_index;
  int     old_width, old_height;
  tlPixel *old_pixel;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::borderBlack()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // shortcuts
  old_pixel = tl_copy_array (this->pixel, width, height, pixelWidth);
  if (!old_pixel) {
    TL_ERRNO_MSG ("tlImage::borderBlack()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  old_width  = width;
  old_height = height;

  // reinitialize image with new size
  this->init (format, width + 2 * bx, height + 2 * by);

  // clear desired area
  this->clear ();

  // put pixels back
  index = pixelWidth * (by * width + bx);
  old_index = 0;
  for (j = 0; j < old_height; j++) {
    index += pixelWidth * width;
    old_index += pixelWidth * old_width;
    memcpy (pixel + index * sizeof (tlPixel), old_pixel + old_index * sizeof (tlPixel), old_width * pixelWidth * sizeof (tlPixel));
  }


  // clean up
  free (old_pixel);
  return 0;
}



///  Add black pixels around current image
///  place result in a new image
/// 
///  \param b     border thickness [pix]
///  \param dest  \b [out] destination image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderBlack (int     b, 
                      tlImage *dest)
{
  return this->borderBlack (b, b, dest);
}



///  Add black pixels around current image
///  place result in a new image
/// 
///  \param bx    border thickness on the X axis [pix]
///  \param by    border thickness on the Y axis [pix]
///  \param dest  \b [out] destination image
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderBlack (int     bx, 
                      int     by, 
                      tlImage *dest)
{
  int     j;
  int     new_index, index;
  int     new_width, new_height;
  tlPixel *new_pixel;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::borderBlack()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // shortcuts
  new_width  = width  + 2 * bx;
  new_height = height + 2 * by;

  // reinitialize image with new size
  dest->init (format, new_width, new_height);
  new_pixel = dest->pixel;

  // clear desired area
  dest->clear ();

  // put pixels back
  new_index = pixelWidth * (by * new_width + bx);
  index = 0;
  for (j = 0; j < height; j++) {
    new_index += pixelWidth * new_width;
    index += pixelWidth * width;
    memcpy (new_pixel + new_index * sizeof (tlPixel), pixel + index * sizeof (tlPixel), width * pixelWidth * sizeof (tlPixel));
  }


  return 0;
}



///  Add mirrored pixels around current image
/// 
///  \param b border thickness [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderMirror (int b)
{
  return this->borderMirror (b, b);
}



///  Add mirrored pixels around current image.
/// 
///  \param bx  border thickness on the X axis [pix]
///  \param by  border thickness on the Y axis [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderMirror (int bx, 
                       int by)
{
  int     i, j, k;
  int     index, old_index;
  int     old_width, old_height;
  tlPixel *old_pixel;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::borderMirror()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check geometry
  if ((bx > width) || (by > height)) {
    TL_ERRNO_MSG ("tlImage::borderMirror()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // shortcuts
  old_pixel = tl_copy_array (this->pixel, width, height, pixelWidth);
  if (!old_pixel) {
    TL_ERRNO_MSG ("tlImage::borderMirror()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }
  old_width  = width;
  old_height = height;

  // reinitialize image with new size
  this->init (format, width + 2 * bx, height + 2 * by);

  // clear desired area
  this->clear ();

  // put pixels back
  index     = pixelWidth * (by * width + bx);
  old_index = 0;
  for (j = 0; j < old_height; j++) {
    index += pixelWidth * width;
    old_index += pixelWidth * old_width;
    memcpy (pixel + index * sizeof (tlPixel), old_pixel + old_index * sizeof (tlPixel), old_width * pixelWidth * sizeof (tlPixel));
  }

  // mirror upper and lower borders
  for (j = 0; j <= by; j++) {

    // upper border
    index = pixelWidth * (j * width + bx);
    old_index = pixelWidth * ((by - j) * old_width);
    memcpy (pixel + index * sizeof (tlPixel), old_pixel + old_index * sizeof (tlPixel), old_width * pixelWidth * sizeof (tlPixel));

    // lower border
    index = pixelWidth * ((height - 1 - j) * width + bx);
    old_index = pixelWidth * ((old_height - 1 - (by - j + 1)) * old_width);
    memcpy (pixel + index * sizeof (tlPixel),  old_pixel + old_index * sizeof (tlPixel), old_width * pixelWidth * sizeof (tlPixel));
  }

  // mirror left and right borders
  for (j = 0; j < old_height; j++) {
    for (i = 0; i < bx; i++) {

      // upper border
      index = pixelWidth * ((j + by) * width + i);
      old_index = pixelWidth * (j * old_width + (bx - i));
      for (k = 0; k < pixelWidth; k++) {
        pixel[index + k] = old_pixel[old_index + k];
      }

      // lower border
      index = pixelWidth * ((j + by) * width + (width - 1 - i));
      old_index = pixelWidth * (j * old_width + (old_width - 1 - bx + i));
      for (k = 0; k < pixelWidth; k++) {
        pixel[index + k] = old_pixel[old_index + k];
      }
    }
  }


  // clean up
  free (old_pixel);
  return 0;
}



///  Add mirrored pixels around current image
///  place result in a new image
/// 
///  \param b border thickness [pix]
///  \param dest \b [out] destination image
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderMirror (int     b, 
                       tlImage *dest)
{
  return borderMirror (b, b, dest);
}



///  Add mirrored pixels around current image
///  place result in a new image
/// 
///  \param bx border thickness on the X axis [pix]
///  \param by border thickness on the Y axis [pix]
///  \param dest \b [out] destination image
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::borderMirror (int     bx, 
                       int     by, 
                       tlImage *dest)
{
  int     i, j, k;
  int     index, new_index;
  int     new_width, new_height;
  tlPixel *new_pixel;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::borderMirror()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check geometry
  if (!dest) {
    TL_ERRNO_MSG ("tlImage::borderMirror()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check geometry
  if ((bx > width) || (by > height)) {
    TL_ERRNO_MSG ("tlImage::borderMirror()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // shortcuts
  new_width  = width  + 2 * bx;
  new_height = height + 2 * by;

  // reinitialize image with new size
  dest->init (format, new_width, new_height);
  new_pixel = dest->pixel;

  // put pixels back
  new_index = pixelWidth * (by * new_width + bx);
  index     = 0;
  for (j = 0; j < height; j++) {
    new_index += pixelWidth * new_width;
    index += pixelWidth * width;
    memcpy (new_pixel + new_index * sizeof (tlPixel),  pixel + index * sizeof (tlPixel),  width * pixelWidth * sizeof (tlPixel));
  }

  // mirror upper and lower borders
  for (j = 0; j <= by; j++) {

    // upper border
    new_index = pixelWidth * (j * new_width + bx);
    index = pixelWidth * ((by - j) * width);
    memcpy (new_pixel + new_index * sizeof (tlPixel), pixel + index * sizeof (tlPixel), width * pixelWidth * sizeof (tlPixel));

    // lower border
    new_index = pixelWidth * ((new_height - 1 - j) * new_width + bx);
    index = pixelWidth * ((height - 1 - (by - j + 1)) * width);
    memcpy (new_pixel + new_index * sizeof (tlPixel),  pixel + index * sizeof (tlPixel), width * pixelWidth * sizeof (tlPixel));
  }

  // mirror left and right borders
  for (j = 0; j < height; j++) {
    for (i = 0; i < bx; i++) {

      // upper border
      new_index = pixelWidth * ((j + by) * new_width + i);
      index = pixelWidth * (j * width + (bx - i));
      for (k = 0; k < pixelWidth; k++) {
        new_pixel[new_index + k] = pixel[index + k];
      }

      // lower border
      new_index = pixelWidth * ((j + by) * new_width + (new_width - 1 - i));
      index = pixelWidth * (j * width + (width - 1 - bx + i));
      for (k = 0; k < pixelWidth; k++) {
        new_pixel[new_index + k] = pixel[index + k];
      }
    }
  }


  return 0;
}
