/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlImage.h"



///  Extract edges using the given method.
/// 
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edges (int method)
{
  return edges ((tlRect *)this, method);
}



///  Extract edges using the given method in a portion of the image.
/// 
///  \param rect    portion of the image to process
///  \param method  edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edges (tlRect *rect, 
                int    method)
{
  int     res     = 0;
  tlPixel *tmp;

  
  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlImage::edges()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // more check
  if (hr) {
    TL_ERRNO_MSG ("tlImage::edges()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // create temporary array
  tmp = tl_copy_array (pixel, width, height, pixelWidth, rect);

  // dispatch
  switch (format) {
  case BIN:
    res = tl_edges_bin (tmp, width, height, this, pixel);
    break;
  default:
    switch (method) {
    case TL_EDGES_MORPH:
      {
        tlImage *temp = new tlImage (this);
        temp->erode  (2);
        this->dilate (2);
        this->sub (temp);
        delete temp;
      }
      break;
    case TL_EDGES_PREWITT:
      {
        tlPixel *temp_hor  = tl_create_array (width, height, pixelWidth);
        tlPixel *temp_vert = tl_create_array (width, height, pixelWidth);
        tl_edges_hor_prewitt  (tmp, width, height, pixelWidth, this, temp_hor);
        tl_edges_vert_prewitt (tmp, width, height, pixelWidth, this, temp_vert);
        tl_pixel_add (temp_hor, temp_vert, width, height, pixelWidth, this, pixel);
        free (temp_hor);
        free (temp_vert);
      }
      break;
    case TL_EDGES_SOBEL:
      {
        tlPixel *temp_hor  = tl_create_array (width, height, pixelWidth);
        tlPixel *temp_vert = tl_create_array (width, height, pixelWidth);
        tl_edges_hor_sobel  (tmp, width, height, pixelWidth, this, temp_hor);
        tl_edges_vert_sobel (tmp, width, height, pixelWidth, this, temp_vert);
        tl_pixel_add (temp_hor, temp_vert, width, height, pixelWidth, this, pixel);
        free (temp_hor);
        free (temp_vert);
      }
      break;
    default:
      TL_ERRNO_MSG ("tlImage::edges()", TL_ERROR_INVALID_FORMAT);
      res = -1;
    }
  }

  // clean-up
  free (tmp);


  return res;
}



///  Extract edges using the given method;
///  the result is put in another image.
/// 
///  \param dest    \b [out] destination image
///  \param method  edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edges (tlImage *dest, 
                int     method)
{
  return edges (this, dest, method);
}



///  Extract edges using the given method in a portion of the image;
///  the result is put in another image.
/// 
///  \param rect    portion of the image to process
///  \param dest    \b [out] destination image
///  \param method  edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edges (tlRect  *rect, 
                tlImage *dest, 
                int     method)
{
  int res = 0;


  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlImage::edges()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  
  // more check
  if (hr) {
    TL_ERRNO_MSG ("tlImage::edges()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // init destination
  dest->init (this);

  // dispatch
  switch (format) {
  case BIN:
    res = tl_edges_bin (pixel, width, height, rect, dest->pixel);
    break;  
  default:
    switch (method) {
    case TL_EDGES_MORPH:
      {
        tlImage *temp = new tlImage ();
        temp->copy (this, rect);
        temp->erode (2, rect);
        this->dilate (2, dest);
        dest->sub (temp);
        delete temp;
      }
      break;
    case TL_EDGES_PREWITT:
      {
        tlPixel *temp_hor  = tl_create_array (width, height, pixelWidth);
        tlPixel *temp_vert = tl_create_array (width, height, pixelWidth);
        tl_edges_hor_prewitt  (pixel, width, height, pixelWidth, rect, temp_hor);
        tl_edges_vert_prewitt (pixel, width, height, pixelWidth, rect, temp_vert);
        tl_pixel_add (temp_hor, temp_vert, width, height, pixelWidth, rect, dest->pixel);
        free (temp_hor);
        free (temp_vert);
      }
      break;
    case TL_EDGES_SOBEL:
      {
        tlPixel *temp_hor  = tl_create_array (width, height, pixelWidth);
        tlPixel *temp_vert = tl_create_array (width, height, pixelWidth);
        tl_edges_hor_sobel  (pixel, width, height, pixelWidth, rect, temp_hor);
        tl_edges_vert_sobel (pixel, width, height, pixelWidth, rect, temp_vert);
        tl_pixel_add (temp_hor, temp_vert, width, height, pixelWidth, rect, dest->pixel);
        free (temp_hor);
        free (temp_vert);
      }
      break;
    default:
      TL_ERRNO_MSG ("tlImage::edges()", TL_ERROR);
      res = -1;
    }
  }


  return res;
}



///  Extract horizontal edges using the given method.
/// 
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesHor (int method)
{
  return edgesHor (this, method);
}



///  Extract horizontal edges using the given method in a portion of the image.
/// 
///  \param rect portion of the image to process
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesHor (tlRect *rect, 
                   int    method)
{
  int     res     = 0;
  tlPixel *tmp;


  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlImage::edgesHor()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // more check
  if (hr) {
    TL_ERRNO_MSG ("tlImage::edgesHor()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // create temporary array
  tmp = tl_copy_array (pixel, width, height, pixelWidth, rect);

  // dispatch
  switch (format) {
  case BIN:
    res = tl_edges_bin (tmp, width, height, this, pixel);
    break;
  default:
    switch (method) {
    case TL_EDGES_MORPH:
      {
        tlImage *temp = new tlImage (this);
        temp->erode (2);
        this->dilate (2);
        this->sub (temp);
        delete temp;
      }
      break;
    case TL_EDGES_PREWITT:
      tl_edges_hor_prewitt (tmp, width, height, pixelWidth, this, pixel);
      break;
    case TL_EDGES_SOBEL:
      tl_edges_hor_sobel (tmp, width, height, pixelWidth, this, pixel);
      break;
    default:
      TL_ERRNO_MSG ("tlImage::edgesHor()", TL_ERROR_INVALID_FORMAT);
      res = -1;
      break;
    }
  }


  // clean-up
  free (tmp);
  return res;
}



///  Extract horizontal edges using the given method;
///  the result is put in another image.
/// 
///  \param dest \b [out] destination image
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesHor (tlImage *dest, 
                   int     method)
{
  return edgesHor (this, dest, method);
}


///  Extract horizontal edges using the given method in a portion of the image;
///  the result is put in another image.
/// 
///  \param rect portion of the image to process
///  \param dest \b [out] destination image
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesHor (tlRect  *rect, 
                   tlImage *dest, 
                   int     method)
{
  int res = 0;


  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlImage::edgesHor()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // more check
  if (hr) {
    TL_ERRNO_MSG ("tlImage::edgesHor()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // init destination
  dest->init (this);

  // dispatch
  switch (format) {
  case BIN:
    res = tl_edges_bin (pixel, width, height, rect, dest->pixel);
    break;
  default:
    switch (method) {
    case TL_EDGES_MORPH:
      {
        tlImage *temp = new tlImage ();
        temp->copy (this, rect);
        temp->erode (2, rect);
        this->dilate (2, dest);
        dest->sub (temp);
        delete temp;
      }
      break;
    case TL_EDGES_PREWITT:
      tl_edges_hor_prewitt (pixel, width, height, pixelWidth, rect, dest->pixel);
      break;
    case TL_EDGES_SOBEL:
      tl_edges_hor_sobel (pixel, width, height, pixelWidth, rect,	dest->pixel);
      break;
    default:
      TL_ERRNO_MSG ("tlImage::edgesHor()", TL_ERROR_INVALID_FORMAT);
      res = -1;
      break;
    }
  }


  return res;
}



///  Extract vertical edges using the given method.
/// 
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesVert (int method)
{
  return edgesVert (this, method);
}



///  Extract vertical edges using the given method in a portion of the image.
/// 
///  \param rect portion of the image to process
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesVert (tlRect *rect, 
                    int    method)
{
  int     res   = 0;
  tlPixel *tmp;


  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlImage::edgesVert()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  
  // more check
  if (hr) {
    TL_ERRNO_MSG ("tlImage::edgesVert()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // create temporary array
  tmp = tl_copy_array (pixel, width, height, pixelWidth, rect);

  // dispatch
  switch (format) {
  case BIN:
    res = tl_edges_bin (tmp, width, height, this, pixel);
    break;
  default:
    switch (method) {
    case TL_EDGES_MORPH:
      {
        tlImage *temp = new tlImage (this);
        temp->erode (2);
        this->dilate (2);
        this->sub (temp);
        delete temp;
      }
      break;
    case TL_EDGES_PREWITT:
      tl_edges_vert_prewitt (tmp, width, height, pixelWidth, this, pixel);
      break;
    case TL_EDGES_SOBEL:
      tl_edges_vert_sobel (tmp, width, height, pixelWidth, this, pixel);
      break;
    default:
      TL_ERRNO_MSG ("tlImage::edgesVert()", TL_ERROR_INVALID_FORMAT);
      res = -1;
      break;
    }
  }

  // clean-up
  free (tmp);


  return res;
}



///  Extract vertical edges using the given method;
///  the result is put in another image.
/// 
///  \param dest \b [out] destination image
///  \param method edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesVert (tlImage *dest, 
                    int     method)
{
  return edgesVert (this, dest, method);
}



///  Extract vertical edges using the given method in a portion of the image;
///  the result is put in another image.
/// 
///  \param rect    portion of the image to process
///  \param dest    \b [out] destination image
///  \param method  edge extraction \ref tlEdges.h "method"
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::edgesVert (tlRect  *rect, 
                    tlImage *dest, 
                    int     method)
{
  int res = 0;


  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlImage::edgesVert()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // more check
  if (hr) {
    TL_ERRNO_MSG ("tlImage::edgesVert()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // init destination
  dest->init (this);

  // dispatch
  switch (format) {
  case BIN:
    res = tl_edges_bin (pixel, width, height, rect, dest->pixel);
    break;
  default:
    switch (method) {
    case TL_EDGES_MORPH:
      {
        tlImage *temp = new tlImage ();
        temp->copy (this, rect);
        temp->erode (2, rect);
        this->dilate (2, dest);
        dest->sub (temp);
        delete temp;
      }
      break;
    case TL_EDGES_PREWITT:
      tl_edges_vert_prewitt (pixel, width, height, pixelWidth, rect, dest->pixel);
      break;
    case TL_EDGES_SOBEL:
      tl_edges_vert_sobel (pixel, width, height, pixelWidth, rect, dest->pixel);
      break;
    default:
      TL_ERRNO_MSG ("tlImage::edgesVert()", TL_ERROR_INVALID_FORMAT);
      res = -1;
      break;
    }
  }


  return res;
}
