/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlImage.h"



///  Write the image to a file of the default format 
///  (which is a proprietary TLIB format)
/// 
///  \param filename the file name. If NULL, write to stdout
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::writeToFile (char *filename)
{
  // proceed
  return (tl_write_tlb (this, filename));
}


///  Write the image to a PPM (binary) file 
/// 
///  \param filename the file name. If NULL, write to stdout
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
int
tlImage::writeToPPM (char *filename)
{
  // proceed
  if (format == RGB) return (tl_write_ppm (pixel, width, height, filename));
  else {
    TL_ERRNO_MSG ("tlImage->writeToPPM()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Write the image to a BMP file 
/// 
///  \param filename the file name. If NULL, write to stdout
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::writeToBMP (char *filename)
{
  // proceed
  switch (format) {
  case RGB:
  case GRAY:
    return (tl_write_bmp (pixel, width, height, format, filename));
  default:
    TL_ERRNO_MSG ("tlImage->writeToBMP()", TL_ERROR_INVALID_FORMAT);
    return (-1);
  }
}



///  Write the image to a JPG file 
/// 
///  \param filename the file name. If NULL, write to stdout
///  \param quality quality of JPEG compression [%]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \remark
///  JPEG is a lossy compression format that's very handy for
///  web distribution of images; it is not recommended to use JPEG
///  images as a source for image processing, because of the dramatic
///  data loss (not visually perceptible, as JPEG compression is 
///  optimized to minimize loss in human perception of pictures)

int
tlImage::writeToJPG (char *filename, 
                     int  quality)
{
  // proceed
  if ((pixelWidth == 1) || (pixelWidth == 3)) {
    return (tl_write_jpg(pixel, width, height, pixelWidth, quality, filename));
  }
  else {
    TL_ERRNO_MSG ("tlImage->writeToJPG()", TL_ERROR_INVALID_FORMAT);
    return (-1);
  }
}



///  Read an image from a given file, 
///  expecting the default format (which is a proprietary TLIB format)
/// 
///  \param filename the file name
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::readFromFile (char *filename)
{
  void *tmppixel;

  // check
  if (!filename) {
    TL_ERRNO_MSG ("tlImage->readFromFile()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  deletePixel ();
  tmppixel = tl_read_tlb (filename, &width, &height, &pixelWidth, &format);

  // select resolution
  switch (format) {
  case STEREO:
    hr = true;
    hrpixel = (tlHRPixel *)tmppixel;
    break;
  default:
    hr = false;
    pixel = (tlPixel *)tmppixel;
  }

  // detect errors
  if (!pixel && !hrpixel) {
    TL_ERRNO_MSG ("tlImage->readFromFile()", TL_ERROR);
    return -1;
  }


  return 0;
}



///  Read an image from a PPM file.
/// 
///  \param filename the file name
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::readFromPPM (char *filename)
{
  // check
  if (!filename) {
    TL_ERRNO_MSG ("tlImage->readFromPPM()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  // proceed
  deletePixel ();
  pixel = tl_read_ppm (filename, &width, &height, &format);

  // detect errors
  if (!pixel) {
    TL_ERRNO_MSG ("tlImage->readFromPPM()", TL_ERROR);
    return -1;
  }

  // reset image dimension
  pixelWidth = tl_pixel_width (format);
  return 0;
}



///  Read an image from a BMP file.
/// 
///  \param filename the file name
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::readFromBMP (char *filename)
{
  // check
  if (!filename) {
    TL_ERRNO_MSG ("tlImage->readFromBMP()", TL_ERROR_NULL_POINTER);
    return -1;
  }
  // proceed
  deletePixel ();
  pixel = tl_read_bmp (filename, &width, &height, &format);

  // detect errors
  if (!pixel) {
    TL_ERRNO_MSG ("tlImage->readFromBMP()", TL_ERROR);
    return -1;
  }

  // reset image dimension
  pixelWidth = tl_pixel_width (format);
  return 0;
}



///  Read an image from a JPG file.
/// 
///  \param filename the file name
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \remark
///  JPEG is a lossy compression format that's very handy for
///  web distribution of images; it is not recommended to use JPEG
///  images as a source for image processing, because of the dramatic
///  data loss (not visually perceptible, as JPEG compression is 
///  optimized to minimize loss in human perception of pictures)

int
tlImage::readFromJPG (char *filename)
{
  // check
  if (!filename) {
    TL_ERRNO_MSG ("tlImage->readFromJPG()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  deletePixel ();
  pixel = tl_read_jpg (filename, &width, &height, &format);

  // detect errors
  if (!pixel) {
    TL_ERRNO_MSG ("tlImage->readFromJPG()", TL_ERROR);
    return -1;
  }

  // reset image dimension
  pixelWidth = tl_pixel_width (format);
  return 0;
}
