/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlImage.h"



///  Normalize pixel values in a portion of the image. The pixel value 
///  range of the image portion is projected to [0,TL_MAX_PIXEL].
///
///  \param rect  portion of the image to process
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.

int
tlImage::normalize (tlRect *rect)
{
  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  switch (format) {
  case GRAY:
    return tl_normalize_gray (pixel, width, height, rect, pixel);
  default:
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Normalize pixel values in an image. The pixel value 
///  range used in the image is projected to [0,TL_MAX_PIXEL].
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.
///
////////////////////////////////////////////////////////////////////////

int
tlImage::normalize ()
{
  return normalize (this);
}



///  Normalize pixel values in a masked portion of the image. 
///  The pixel value range of the image portion is projected 
///  to [0,TL_MAX_PIXEL].
/// 
///  \param mask  mask that defines which pixels are taken into account
///  \param rect  portion of the image to process
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.

int
tlImage::normalize (tlMask *mask, 
                    tlRect *rect)
{
  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  switch (format) {
  case GRAY:
    return tl_normalize_gray (pixel, width, height, mask, rect, pixel);
  default:
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Normalize pixel values in a masked image. The pixel value range 
///  of the image portion is projected to [0,TL_MAX_PIXEL].
/// 
///  \param mask mask that defines which pixels are taken into account
/// 
///  \return
///  0 on success, -1 otherwise. 
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.

int
tlImage::normalize (tlMask *mask)
{
  return normalize (mask, (tlRect *) this);
}



///  Normalize pixel values in a portion of the image. The pixel value 
///  range of the image portion is projected to [0,TL_MAX_PIXEL].
///  The result is put in another image.
/// 
///  \param rect  portion of the image to process
///  \param dest  \b [out] destination image
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.

int
tlImage::normalize (tlRect  *rect, 
                    tlImage *dest)
{
  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  switch (format) {
  case GRAY:
    return tl_normalize_gray (pixel, width, height, rect, dest->pixel);
  default:
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Normalize pixel values in an image. The pixel value 
///  range of the image portion is projected to [0,TL_MAX_PIXEL].
///  The result is put in another image.
/// 
///  \param dest  \b [out] destination image
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.

int
tlImage::normalize (tlImage *dest)
{
  return normalize (this, dest);
}



///  Normalize pixel values in a masked portion of the image. 
///  The pixel value range of the image portion is projected 
///  to [0,TL_MAX_PIXEL]. The result is put in another image.
/// 
///  \param mask  mask that defines which pixels are taken into account
///  \param rect  portion of the image to process
///  \param dest  destination image
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.

int
tlImage::normalize (tlMask  *mask, 
                    tlRect  *rect, 
                    tlImage *dest)
{
  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  switch (format) {
  case GRAY:
    return tl_normalize_gray (pixel, width, height, mask, rect, dest->pixel);
  default:
    TL_ERRNO_MSG ("tlImage::normalize()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Normalize pixel values in a masked image. The pixel value range 
///  of the image portion is projected to [0,TL_MAX_PIXEL]. 
///  The result is put in another image.
/// 
///  \param mask  mask that defines which pixels are taken into account
///  \param dest  destination image
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  Only GRAY images can be normalized.

int
tlImage::normalize (tlMask  *mask, 
                    tlImage *dest)
{
  return normalize (mask, this, dest);
}
