/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <memory.h>

#include "tlImage.h"
#include "tlTransforms.h"



///  Mirror the image to itself
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::mirror ()
{
  register tlPixel *in;
  register tlPixel *out;
  register tlPixel pix;
	

  // do the mirror thing
  out = pixel;
  for (int j=0; j<height; j++) {
    in  = &(pixel[pixelWidth*(width + width*j)]);
    out = &(pixel[pixelWidth*(        width*j)]);
		for (int i=0; i<width/2; i++) {
      in -= pixelWidth;
      for (int k=0; k<pixelWidth; k++) {
        pix = *(in);
        *(in) = *(out);
        *(out) = pix;
        in  ++;
        out ++;
      }
      in -= pixelWidth;
		}
	}


  return 0;
}



///  Mirror the image and put the result into another image.
///  The original is untouched.
/// 
///  \param dest  destination image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::mirror (tlImage *dest)
{
  register tlPixel *in;
  register tlPixel *out;
  register tlPixel *buf;
	

  // check consistency
  if (!dest) {
    TL_ERRNO_MSG ("tlImage::mirror()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // init dest image
  dest->init (format, width, height);

  // shortcuts
  buf = dest->pixel;

  // do the mirror thing
	for (int j=0; j<height; j++) {
    in  = &(pixel[pixelWidth*(width + width*j)]);
    out = &(buf  [pixelWidth*(        width*j)]);
		for (int i=0; i<width; i++) {
      in -= pixelWidth;
      for (int k=0; k<pixelWidth; k++) {
        *(out) = *(in);
        in  ++;
        out ++;
      }
      in -= pixelWidth;
		}
	}


  return 0;
}



///  Rotate the image clockwise.
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::rotateRight ()
{
  int     res;
  tlPixel *src;


  src = tl_copy_array  (pixel, width, height, pixelWidth);
  init (format, height, width);
  res =	tl_rotate_right (src, height, width, pixelWidth, pixel);
  free (src);


  return res;
}



///  Rotate the image clockwise.
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::rotateLeft ()
{
  int     res;
  tlPixel *src;


  src = tl_copy_array  (pixel, width, height, pixelWidth);
  init (format, height, width);
  res =	tl_rotate_left (src, height, width, pixelWidth, pixel);
  free (src);


  return res;
}



///  Rotate the image clockwise.
///  
///  \param dest    destination image
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::rotateRight (tlImage *dest)
{
  dest->init (format, height, width);
  return tl_rotate_right (pixel, width, height, pixelWidth, dest->pixel);
}



///  Rotate the image counter-clockwise.
///  
///  \param dest    destination image
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::rotateLeft (tlImage *dest)
{
  dest->init (format, height, width);
  return tl_rotate_left (pixel, width, height, pixelWidth, dest->pixel);
}



///  Make a negative of a given image.
///  The original is untouched.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::invert ()
{
  register tlPixel *buf;
	

  // shortcuts
  buf = pixel;

  // do the mirror thing
	for (int j=0; j<height; j++) {
		for (int i=0; i<width; i++) {
      for (int k=0; k<pixelWidth; k++) {
        *(buf) = TL_MAX_PIXEL_VAL - *(buf++);
      }
    }
  }


  return 0;
}



///  Make a negative of a given image.
///  The original is untouched.
/// 
///  \param dest  destination image
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlImage::invert (tlImage *dest)
{
  register tlPixel *in;
  register tlPixel *out;
	

  // check consistency
  if (!dest) {
    TL_ERRNO_MSG ("tlImage::mirror()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // init dest image
  dest->init (format, width, height);

  // shortcuts
  in  = pixel;
  out = dest->pixel;

  // do the mirror thing
	for (int j=0; j<height; j++) {
		for (int i=0; i<width; i++) {
      for (int k=0; k<pixelWidth; k++) {
        *(out++) = TL_MAX_PIXEL_VAL - *(in++);
      }
    }
  }


  return 0;
}
