/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://vrai-group.epfl.ch/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlLine.h"
#include "tlPoint.h"
#include "tlRect.h"



///  \file tlLine.h
///  \brief Line representation and operators.

///  \class tlLine
///  This class contains the representation of a line segment. It is used by the tlHough
///  line extractor class.



///  Constructor.
///
///  \return
///  See \ref errors "error management" for details.

tlLine::tlLine ()
{
  p1 = new tlPoint ();
  p2 = new tlPoint ();
  set (0, 0, 1, 1);


  return;
}



///  Constructor.
///
///  \param ang     line angle in polar representation (in [rad]).
///  \param rad     line radius in polar representation (in [pix]).
///
///  \return
///  See \ref errors "error management" for details.

tlLine::tlLine (float ang, 
                float rad)
{
  p1 = new tlPoint ();
  p2 = new tlPoint ();
  set (ang, rad);
}



///  Constructor.
///
///  \param x1     first segment extremity X coord in [pix].
///  \param y1     first segment extremity Y coord in [pix].
///  \param x2     second segment extremity X coord in [pix].
///  \param y2     second segment extremity Y coord in [pix].
///
///  \return
///  See \ref errors "error management" for details.

tlLine::tlLine (int x1, 
                int y1, 
                int x2, 
                int y2)
{
  p1 = new tlPoint ();
  p2 = new tlPoint ();
  set (x1,y1, x2,y2);
}



///  Constructor.
///
///  \param pt1     first segment extremity.
///  \param pt2     second segment extremity.
///
///  \return
///  See \ref errors "error management" for details.

tlLine::tlLine (tlPoint *pt1, 
                tlPoint *pt2)
{
  p1 = new tlPoint ();
  p2 = new tlPoint ();
  set (pt1, pt2);
}



///  Constructor.
///
///  \param line   pointer to another tlLine to copy parameters from.
///
///  \return
///  See \ref errors "error management" for details.

tlLine::tlLine (tlLine *line)
{
  p1 = new tlPoint ();
  p2 = new tlPoint ();
  set (line);
}



///  Destructor.
///
///  \return
///  See \ref errors "error management" for details.

tlLine::~tlLine ()
{
  delete p1;
  delete p2;
}



///  Define the line parameters in polar coordonate system.
///
///  \param ang     line angle in polar representation (in [rad]).
///  \param rad     line radius in polar representation (in [pix]).
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlLine::set (float ang, 
             float rad)
{
  theta = ang;
  r     = rad;

  p1->setLocation (0, (int)(r/cos(theta)));
  p2->setLocation ((int)(r/sin(theta)), 0);
  
  
  return 0;
}



///  Define the line segment parameters from the location of its extremities.
///
///  \param x1     first segment extremity X coord in [pix].
///  \param y1     first segment extremity Y coord in [pix].
///  \param x2     second segment extremity X coord in [pix].
///  \param y2     second segment extremity Y coord in [pix].
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlLine::set (int x1, 
             int y1, 
             int x2, 
             int y2)
{
  p1->setLocation (x1, y1);
  p2->setLocation (x2, y2);

  theta = atan2(-(float)(x2-x1),(float)(y2-y1));
  r     = x1*cos(theta) + y1*sin(theta);


  return 0;
}



///  Define the line segment parameters from the location of its extremities.
///
///  \param pt1     first segment extremity.
///  \param pt2     second segment extremity.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlLine::set (tlPoint *pt1, 
             tlPoint *pt2)
{
  p1->setLocation (pt1);
  p2->setLocation (pt2);

  theta = atan2(-(float)(p2->x-p1->x),(float)(p2->y-p1->y));
  r     = p1->x*cos(theta) + p1->y*sin(theta);


  return 0;
}



///  Define the line as the best fit from a series of points.
///
///  \param chain     list of tlPoints to match.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlLine::fit (tlChain *chain)
{
  return 0;
}



///  Clip a line of arbitrary length to a bounding rectangle.
///
///  \param rect     bounding rectangle that define the line clipping area.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLine::clip (tlRect *rect)
{
  return clip (rect->width, rect->height);
}



///  Clip a line of arbitrary length to a given dimension.
///
///  \param width     clipping width.
///  \param height    clipping height.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLine::clip (int width,
              int height)
{
  tlPoint *p;
  int wx1,wy1,wx2,wy2;
  int hx1,hy1,hx2,hy2;


  if (theta == 0.0 || theta == TL_PI) {
    hy1 = 0;
    hx1 = (int)r;
    hy2 = height-1;
    hx2 = (int)r;
    p1->setLocation (hx1,hy1);
    p2->setLocation (hx2,hy2);
  }
  else if (theta == TL_PI/2.0f) {
    wx1 = 0;
    wy1 = (int)r;
    wx2 = width-1;
    wy2 = (int)r;
    p1->setLocation (wx1,wy1);
    p2->setLocation (wx2,wy2);
  }
  else {
    wx1 = 0;
    wy1 = (int)(r/sin(theta));
    wx2 = width-1;
    wy2 = (int)((r-wx2*cos(theta))/sin(theta));
    hy1 = 0;
    hx1 = (int)(r/cos(theta));
    hy2 = height-1;
    hx2 = (int)((r-hy2*sin(theta))/cos(theta));

    p = p1;
    if (wy1 >= 0 && wy1 < height) {
      p->setLocation (wx1,wy1);
      p = p2;
    }
    if (wy2 >= 0 && wy2 < height) {
      p->setLocation (wx2,wy2);
      if (p == p2) return 0;
      else p = p2;
    }
    if (hx1 >= 0 && hx1 < width) {
      p->setLocation (hx1,hy1);
      if (p == p2) return 0;
      else p = p2;
    }
    if (hx2 >= 0 && hx2 < width) {
      p->setLocation (hx2,hy2);
      if (p == p2) return 0;
      else return -1;
    }
  }


  return 0;
}



///  Copy parameters from an existing tlLine object.
///
///  \param line   pointer to another tlLine to copy parameters from.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLine::set (tlLine *line)
{
  return copy (line);
}



///  Copy parameters from an existing tlLine object.
///
///  \param line   pointer to another tlLine to copy parameters from.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLine::copy (tlLine *line)
{
  p1->setLocation (line->p1);
  p2->setLocation (line->p2);

  theta = line->theta;
  r     = line->r;


  return 0;
}



///  Copy parameters to an existing tlLine object.
///
///  \param line   pointer to another tlLine to copy parameters to.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLine::copyTo (tlLine *line)
{
  return line->copy (this);
}
