/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlLocation.h"
#include "tlPoint.h"



///  \file tlLocation.h
///  \brief Geometric location object.

///  \class tlLocation
///  \brief Location object

///  \var int tlLocation::x
///  X location of the object [pix].

///  \var int tlLocation::y
///  Y location of the object [pix].

///  \var int tlLocation::disp
///  Disparity of the object [disp].

///  \var int tlLocation::wx
///  X location of the object in real-world coordonates [m].

///  \var int tlLocation::wy
///  Y location of the object in real-world coordonates [m].

///  \var int tlLocation::wz
///  Z location of the object in real-world coordonates [m].

///  \fn tlLocation::~tlLocation
///  Destructor.



///  Constructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlLocation::tlLocation ()
{
  x    = 0;
  y    = 0;
  disp = 0;
  wx   = 0.0f;
  wy   = 0.0f;
  wz   = 0.0f;

  tl_errno = TL_NO_ERROR;
}



///  Constructor.
///
///  \param x      x location of the object [pix]
///  \param y      y location of the object [pix]
///  \param disp   disparity of the object  [disp]
/// 
///  \note
///  See \ref errors "error management" for details.

tlLocation::tlLocation (int x, 
                        int y, 
                        int disp)
{
  this->x    = x;
  this->y    = y;
  this->disp = disp;
  this->wx   = 0.0f;
  this->wy   = 0.0f;
  this->wz   = 0.0f;

  tl_errno = TL_NO_ERROR;
}



///  Destructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlLocation::~tlLocation ()
{
  tl_errno = TL_NO_ERROR;
}



///  Set the location object parameters.
/// 
///  \param x     x location of the object [pix]
///  \param y     y location of the object [pix]
///  \param disp  disparity of the object [disp]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLocation::setLocation (int x, 
                         int y, 
                         int disp)
{
  this->x    = x;
  this->y    = y;
  this->disp = disp;

  return 0;
}



///  Set the location object parameters.
/// 
///  \param point tlPoint object
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLocation::setLocation (tlPoint *point)
{
  this->x    = point->x;
  this->y    = point->y;
  this->disp = point->disp;

  return 0;
}



///  Set the world location object parameters.
/// 
///  \param wx  world X location
///  \param wy  world Y location
///  \param wz  world Z location
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLocation::setWorldLocation (float wx, 
                              float wy, 
                              float wz)
{
  this->wx = wx;
  this->wy = wy;
  this->wz = wz;

  return 0;
}



///  Shift the object location.
/// 
///  \param dx shift along x axis [pix]
///  \param dy shift along y axis [pix]
///  \param ddisp disparity shift [disp]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLocation::shift (int dx, 
                   int dy, 
                   int ddisp)
{
  x    += dx;
  y    += dy;
  disp += ddisp;

  return 0;
}



///  Get the location of the object.
/// 
///  \param location \b [out] tlLocation object that will receive the data
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlLocation::toLocation (tlLocation *location)
{
  location->x    = this->x;
  location->y    = this->y;
  location->disp = this->disp;
  location->wx   = this->wx;
  location->wy   = this->wy;
  location->wz   = this->wz;

  return 0;
}



///  Get a copy of the location of the object.
/// 
///  \return
///  A pointer to a tlLocation object containing the data.
/// 
///  \note
///  It is the responsibility of the programmer to delete the object returned.

tlLocation *
tlLocation::toLocation ()
{
  tlLocation *location = new tlLocation ();


  location->setLocation (x, y, disp);
  location->setWorldLocation (wx, wy, wz);


  return location;
}
