/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <malloc.h>
#include <memory.h>
#include <math.h>

#include "tlMask.h"
#include "tlRect.h"
#include "tlImage.h"
#include "tlObject.h"



///  \file tlMask.h
///  \brief Binary mask for image processing routines.

///  \class tlMask
///  \brief Binary mask for image processing routines.
///
///  The tlMask object is a very simple structure that is used extensively
///  in TLIB routines. It is a basic binary mask with position and size information that can
///  be used to apply image processing routines to specific pixels only. A tlMask object is
///  also used to define which pixels are part of an object in a given area. Mask objects can
///  be combined, and support binary image processing routines such as morphological operators.

///  \var int tlMask::cx
///  X position of the mask reference point relatively to the upper left corner.

///  \var int tlMask::cy
///  Y position of the mask reference point relatively to the upper left corner.

///  \var int tlMask::area
///  Number of active pixels in the mask.



///  Constructor, build and empty mask.
/// 
///  \note
///  See \ref errors "error management" for details.

tlMask::tlMask () : tlPixelArray (BIN, TL_DEFAULT_COLS, TL_DEFAULT_ROWS)
{
  if (tl_errno != TL_NO_ERROR) return;

  this->cx = TL_DEFAULT_COLS / 2;
  this->cy = TL_DEFAULT_ROWS / 2;

  // set mask to transparent
  set ();
}



///  Constructor, build a mask of a given size at a given position.
///
///  \param cx        mask center x coord (w/ respect to UL corner)
///  \param cy        mask center y coord (w/ respect to UL corner)
///  \param width     mask width
///  \param height    mask height
/// 
///  \note
///  See \ref errors "error management" for details.

tlMask::tlMask (int cx, 
                int cy, 
                int width, 
                int height) : tlPixelArray (BIN, width, height)
{
  if (tl_errno != TL_NO_ERROR) return;

  this->cx = cx;
  this->cy = cy;

  // set mask to transparent
  set ();
}



///  Constructor, create a copy of an existing mask.
///
///  \param mask      mask to copy
/// 
///  \note
///  See \ref errors "error management" for details.

tlMask::tlMask (tlMask *mask)
{
  int size;


  // shortcuts
  size             = (mask->width) * (mask->height) * sizeof (tlPixel);
  this->cx         = mask->cx;
  this->cy         = mask->cy;
  this->x          = mask->x;
  this->y          = mask->y;
  this->width      = mask->width;
  this->height     = mask->height;
  this->area       = mask->area;
  this->pixelWidth = 1;
  this->format     = BIN;

  // mask allocation
  pixel = (tlPixel *) realloc (pixel, size);
  if (!pixel) {
    TL_ERRNO_MSG ("tlMask()", TL_ERROR_ALLOCATION_FAILED);
    return;
  }

  // copy image data
  memcpy (pixel, mask->pixel, size);
  tl_errno = TL_NO_ERROR;
}



///  Constructor, create a a mask from a binary image.
///
///  \param image      binary image to copy as a mask
/// 
///  \note
///  See \ref errors "error management" for details.

tlMask::tlMask (tlImage *image)
{
  register tlPixel *ptr;

  int i, j;
  int size;


  // test pixel width
  if (image->pixelWidth != 1) {
    TL_ERRNO_MSG ("tlMask()", TL_ERROR_INVALID_FORMAT);
    return;
  }

  // shortcuts
  size             = (image->width) * (image->height) * sizeof (tlPixel);
  this->cx         = 0;
  this->cy         = 0;
  this->x          = 0;
  this->y          = 0;
  this->width      = image->width;
  this->height     = image->height;
  this->pixelWidth = 1;
  this->format     = BIN;

  // mask allocation
  pixel = (tlPixel *) realloc (pixel, size);
  if (!pixel) {
    TL_ERRNO_MSG ("tlMask()", TL_ERROR_ALLOCATION_FAILED);
    return;
  }

  // copy image data
  memcpy (pixel, image->pixel, size);

  // compute area
  area = 0;
  ptr = pixel;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      if (*(ptr++)) {
        area++;
      }
    }
  }

  tl_errno = TL_NO_ERROR;
}



///  Destructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlMask::~tlMask ()
{
  tl_errno = TL_NO_ERROR;
}



///  Set mask to fully active (transparent).
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set ()
{
  memset (pixel, 1, width * height);
  return 0;
}



///  Change mask size and set mask to fully active (transparent).
///
///  \param width     new mask width  [pix]
///  \param height    new mask height [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (int width, 
             int height)
{
  int  size;


  // shortcuts
  this->width  = width;
  this->height = height;
  this->area   = width * height;
  size         = area * sizeof (tlPixel);

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // set mask to transparent
  set ();


  return 0;
}



///  Copy mask attributes and data from another mask.
///
///  \param mask      mask to copy data from
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (tlMask *mask)
{
  int	size;


  // check
  if ((!mask)) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  this->width  = mask->width;
  this->height = mask->height;
  this->cx     = mask->cx;
  this->cy     = mask->cy;
  this->x      = mask->x;
  this->y      = mask->y;
  this->area   = mask->area;
  size         = width * height * sizeof (tlPixel);

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }


  // copy mask data
  memcpy (pixel, mask->pixel, size);
  return 0;
}



///  Set up a mask from a byte array.
///  Non-zero elements in the byte array will be translated in active
///  (transparent) pixels in the mask. Null elements in the byte array
///  will turn into inactive (opaque) mask pixels.
///
///  \param mask      byte array
///  \param width     byte array width  [pix]
///  \param height    byte array height [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (char *mask, 
             int  width, 
             int  height)
{
  register tlPixel *ptr;

  int i,j;
  int	size;


  // check
  if (!mask) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  size         = width * height * sizeof (tlPixel);
  this->width  = width;
  this->height = height;

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // compute area
  area = 0;
  ptr  = pixel;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      if (*(mask++)) {
        *(ptr++) = 1;
        area++;
      }
      else *(ptr++) = 0;
    }
  }


  return 0;
}



///  Set up a mask from a portion of a byte array.
///  Works like set(char *mask, int width, int height), only on 
///  a portion of a byte array. The resulting mask has the size of 
///  the \a rect argument.
///
///  \param mask      byte array
///  \param width     byte array width  [pix]
///  \param height    byte array height [pix]
///  \param rect      portion of the mask to use
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (char   *mask, 
             int    width, 
             int    height, 
             tlRect *rect)
{
  int	i,j;
  int	x1,x2,y1,y2;
  int	size;


  // check
  if (!mask) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1           = TL_MAX (0, (rect->x));
  x2           = TL_MIN (width, ((rect->x) + (rect->width)));
  y1           = TL_MAX (0, (rect->y));
  y2           = TL_MIN (height, ((rect->y) + (rect->height)));
  this->width  = (x2 - x1);
  this->height = (y2 - y1);
  size         = (this->width) * (this->height) * sizeof (tlPixel);

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // assign and compute area
  area = 0;
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      if (mask[TL_PIXEL (j, i, width)] > 0) {
        pixel[TL_PIXEL ((j - y1), (i - x1), (this->width))] = 1;
        area++;
      }
      else pixel[TL_PIXEL ((j - y1), (i - x1), (this->width))] = 0;
    }
  }


  return 0;
}



///  Set up a mask from a pixel array.
///  Non-zero elements in the pixel array will be translated in active
///  (transparent) pixels in the mask. Null elements in the pixel array
///  will turn into inactive (opaque) mask pixels.
///
///  \param mask      pixel array
///  \param width     pixel array width  [pix]
///  \param height    pixel array height [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (tlPixel *mask, 
             int     width, 
             int     height)
{
  register tlPixel *ptr;

  int	i,j;
  int	size;


  // check
  if (!mask) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  size         = width * height * sizeof (tlPixel);
  this->width  = width;
  this->height = height;

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // copy mask data
  memcpy (this->pixel, mask, size);

  // compute area
  ptr = pixel;
  area = 0;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      if (*(ptr++))	area++;
    }
  }


  return 0;
}



///  Set up a mask from a portion of a pixel array.
///  Works like set(tlPixel *mask, int width, int height), only on 
///  a portion of a pixel array. The resulting mask has the size of 
///  the \a rect argument.
///
///  \param mask      pixel array
///  \param width     pixel array width  [pix]
///  \param height    pixel array height [pix]
///  \param rect      portion of the pixel array to use
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (tlPixel *mask, 
             int     width, 
             int     height, 
             tlRect  *rect)
{
  int i,j;
  int x1,x2,y1,y2;
  int	size;


  // check
  if (!mask) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1           = TL_MAX (0, (rect->x));
  x2           = TL_MIN (width, ((rect->x) + (rect->width)));
  y1           = TL_MAX (0, (rect->x));
  y2           = TL_MIN (height, ((rect->x) + (rect->height)));
  this->width  = (x2 - x1);
  this->height = (y2 - y1);
  size         = (this->width) * (this->height) * sizeof (tlPixel);

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // assign and compute area
  area = 0;
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      if (mask[TL_PIXEL (j, i, width)] != 0) {
        pixel[TL_PIXEL ((j - y1), (i - x1), (this->width))] = 1;
        area++;
      }
      else	pixel[TL_PIXEL ((j - y1), (i - x1), (this->width))] = 0;
    }
  }


  return 0;
}



///  Set up a mask from a \ref tlObject pixel array.
///  Works like set(tlPixel *pixel, int width, int height). Takes a
///  \ref tlObject pixel array as an input array.
///
///  \param object    the object to get the mask from
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (tlObject *object)
{
  int     i, j, k;
  int	    index, o_index;
  int     sum;
  int	    o_pixelWidth;
  tlPixel *data;


  // check
  if (!object) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  this->width  = object->width;
  this->height = object->height;
  data         = object->pixel;
  o_pixelWidth = object->pixelWidth;

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, width * height * sizeof (tlPixel));
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // go through object pixels
  index   = 0;
  o_index = 0;
  area    = 0;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      sum = 0;
      for (k = 0; k < o_pixelWidth; k++) {
        sum += data[o_index++];
      }
      if (sum == 0)	pixel[index++] = 0;
      else {
        pixel[index++] = 1;
        area++;
      }
    }
  }


  return 0;
}



///  Set up a mask from a \ref tlImage pixel array.
///  Works like set(tlPixel *pixel, int width, int height). Takes a
///  \ref tlImage pixel array as an input array.
///
///  \param image    the image to get the mask from
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (tlImage * image)
{
  int	   res;
  tlRect *rect = image->toRect ();


  res = set (image, rect);
  delete rect;


  return res;
}


///  Set up a mask from a portion of a \ref tlImage pixel array.
///  Works like set(tlPixel *pixel, int width, int height). Takes a
///  portion \ref tlImage pixel array as an input array. The resulting
///  mask has the size of the \a rect argument.
///
///  \param image	the image to get the mask from
///  \param rect		subimage to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::set (tlImage *image, 
             tlRect  *rect)
{
  int	    i,j,k;
  int	    x1,x2,y1,y2;
  int	    index,i_index;
  int	    sum;
  int	    i_width;
  int	    i_pixelWidth;
  tlPixel *data;


  // check
  if ((!image) || (!rect)) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1           = TL_MAX (0, (rect->x));
  x2           = TL_MIN ((image->width), ((rect->x) + (rect->width)));
  y1           = TL_MAX (0, (rect->y));
  y2           = TL_MIN ((image->height), ((rect->y) + (rect->height)));
  this->width  = (x2 - x1);
  this->height = (y2 - y1);
  data         = image->pixel;
  i_width      = image->width;
  i_pixelWidth = image->pixelWidth;

  // reset dimension
  x  = x1;
  y  = y1;
  cx = 0;
  cy = 0;

  // reallocate pixel map
  pixel =	(tlPixel *) realloc (pixel, (this->width) * (this->height) * sizeof (tlPixel));
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }

  // go through object pixels
  area  = 0;
  index = 0;
  for (j = y1; j < y2; j++) {
    i_index = (j * i_width + x1) * i_pixelWidth;
    for (i = x1; i < x2; i++) {
      sum = 0;
      for (k = 0; k < i_pixelWidth; k++) {
        sum += data[i_index++];
      }
      if (sum == 0)	pixel[index++] = 0;
      else {
        pixel[index++] = 1;
        area++;
      }
    }
  }


  return 0;
}



///  Inverse all the pixels of a mask.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::invert ()
{
  register tlPixel *ptr;

  int i,j;


  ptr = pixel;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      if (!(*(ptr)))	*(ptr++) = 1;
      else
        *(ptr++) = 0;
    }
  }


  return 0;
}


///  Inverse all the pixels in a portion of a mask.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::invert (tlRect *rect)
{
  int	i,j;
  int	index         = 0;
  int	x1,x2,y1,y2;


  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlMask::invert()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX (0, (rect->x));
  x2 = TL_MIN (width, ((rect->x) + (rect->width)));
  y1 = TL_MAX (0, (rect->x));
  y2 = TL_MIN (height, ((rect->x) + (rect->height)));

  // invert
  for (j = y1; j < y2; j++) {
    index = TL_PIXEL (j, x1, width);
    for (i = x1; i < x2; i++) {
      pixel[index] = (tlPixel)((pixel[index++]) ? TL_MIN_PIXEL_VAL : TL_MAX_PIXEL_VAL);
    }
  }


  return 0;
}



///  Set up the mask center coordinates to be the barycenter.
///  The barycenter (cx,cy) is expressed in pixels, respectively to
///  the mask UL corner.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::setCenter ()
{
  register tlPixel *ptr;

  int i,j;
  int	Snx,Sny,Sn;


  // shortcuts
  ptr = pixel;
  Snx = Sny = Sn = 0;

  // compute
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      if (*(ptr++)) {
        Snx += i;
        Sny += j;
        Sn += 1;
      }
    }
  }

  // register
  if (Sn != 0) {
    cx = Snx / Sn;
    cy = Sny / Sn;
  }
  else {
    cx = width / 2;
    cy = height / 2;
  }


  return 0;
}



///  Set up the mask center coordinates.
///  The center (cx,cy) is expressed in pixels, respectively to
///  the mask UL corner.
///
///  \param cx        mask center x coord (w/ respect to mask UL corner)
///  \param cy        mask center y coord (w/ respect to mask UL corner)
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::setCenter (int cx, 
                   int cy)
{
  this->cx = cx;
  this->cy = cy;
  return 0;
}



///  Set mask to fully inactive (opaque).
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::clear ()
{
  memset (pixel, 0, width * height);
  return 0;
}



///  Set mask to fully inactive and change mask size.
///
///  \param width       mask width
///  \param height      mask height
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::clear (int width, 
               int height)
{
  int	size;


  // shortcuts
  this->width  = width;
  this->height = height;
  this->area   = width * height;
  size         = area * sizeof (tlPixel);

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::clear()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }


  // set mask to transparent
  return clear ();
}



///  Superimpose two masks and merges their data.
///
///  \param mask       mask to merge with
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::merge (tlMask *mask)
{
  tlMask *temp = new tlMask ();
  int	   w, h;
  int	   index;
  int	   i, j;
  int	   i1, j1, i2, j2, index1, index2;


  // check
  if (!mask) {
    TL_ERRNO_MSG ("tlMask::merge()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  int	w1  = this->width;
  int	h1  = this->height;
  int	w2  = mask->width;
  int	h2  = mask->height;
  int	x11 = this->x;
  int	x12 = x11 + w1;
  int	y11 = this->y;
  int	y12 = y11 + h1;
  int	x21 = mask->x;
  int	x22 = x21 + w2;
  int	y21 = mask->y;
  int	y22 = y21 + h2;

  // compute final width new x position
  if (x11 < x21) {
    if (x12 > x22)    w = w1;
    else              w = x22 - x11;
    temp->x = x11;
  } else {
    if (x12 < x22)	w = w2;
    else              w = x12 - x21;
    temp->x = x21;
  }

  // compute final height new y position
  if (y11 < y21) {
    if (y12 > y22)	h = h1;
    else              h = y22 - y11;
    temp->y = y11;
  } else {
    if (y12 < y22)	h = h2;
    else              h = y12 - y21;
    temp->y = y21;
  }

  // set temp mask
  temp->set (w, h);
  temp->clear ();

  // merge
  index = 0;
  int xmin = temp->x,
    ymin = temp->y;
  int xmax = xmin + w,
    ymax = ymin + h;
  for (j = ymin; j < ymax; j++) {
    for (i = xmin; i < xmax; i++) {
      // check first mask
      if ((j < y11) || (j >= y12)) j1 = -1;
      else {
        j1 = j - y11;
        if ((i < x11) || (i >= x12)) i1 = -1;
        else {
          i1 = i - x11;
          index1 = j1 * w1 + i1;
          if ((index1 >= 0) && (this->pixel[index1] != 0)) {
            temp->pixel[index] = this->pixel[index1];
          }
        }
      }

      // check second mask
      if ((j < y21) || (j >= y22)) j2 = -1;
      else {
        j2 = j - y21;
        if ((i < x21) || (i >= x22)) i2 = -1;
        else {
          i2 = i - x21;
          index2 = j2 * w2 + i2;
          if ((index2 >= 0) && (mask->pixel[index2] != 0)) {
            temp->pixel[index] = mask->pixel[index2];
          }
        }
      }

      // increment
      index++;
    }
  }

  // update blob
  this->set (temp);
  this->setLocation (temp->x, temp->y);
  this->area = 0;
  index = 0;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      if (pixel[index]) {
        (this->area)++;
      }
      index++;
    }
  }


  // delete temporary mask
  delete temp;
  return 0;
}



///  Fill all "holes" in the mask structure. That is, make the mask convex.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::fill (void)
{
  int	i, j;
  int	index;
  int start, end;


  // process lines
  for (j = 0; j < height; j++) {
    start = -1;
    end   = -1;

    // get borders
    index = j * width;
    for (i = 0; i < width; i++) {
      if (pixel[index] != 0) {
        if (start == -1) start = end = i;
        else end = i;
      }
      index += 1;
    }

    // fill lines
    index = j * width + start;
    for (i = start; i < end; i++) {
      pixel[index] = 255;
      index += 1;
    }
  }

  // process columns
  for (i = 0; i < width; i++) {
    start = -1;
    end = -1;

    // get borders
    index = i;
    for (j = 0; j < height; j++) {
      if (pixel[index] != 0) {
        if (start == -1) start = end = j;
        else end = j;
      }
      index += width;
    }

    // fill lines
    index = start * width + i;
    for (j = start; j < end; j++) {
      pixel[index] = 255;
      index += width;
    }
  }


  return 0;
}



///  Remove a given "layer" of pixels around the mask.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::crop (int val)
{
  int i,j;
  int index;
  int start, end;


  // process lines
  if (width > 2 * val) {
    for (j = 0; j < height; j++) {
      start = -1;
      end   = -1;

      // get borders
      index = j * width;
      for (i = 0; i < width; i++) {
        if (pixel[index] != 0) {
          if (start == -1) start = end = i;
          else end = i;
        }
        index += 1;
      }

      // fill lines
      if (end - start > 2 * val) {

        // remove left layer
        index = j * width + start;
        for (i = start; i < (start + val); i++) {
          pixel[index] = 0;
          index += 1;
        }

        // remove right layer
        index = j * width + end;
        for (i = end; i >= (end - val); i--) {
          pixel[index] = 0;
          index -= 1;
        }
      }
    }
  }

  // process columns
  if (height > 2 * val) {
    for (i = 0; i < width; i++) {
      start = -1;
      end   = -1;

      // get borders
      index = i;
      for (j = 0; j < height; j++) {
        if (pixel[index] != 0) {
          if (start == -1) start = end = j;
          else end = j;
        }
        index += width;
      }
      if (end - start > 2 * val) {

        // remove left layer
        index = start * width + i;
        for (j = start; j < (start + val); j++) {
          pixel[index] = 0;
          index += width;
        }

        // remove right layer
        index = (end) * width + i;
        for (j = (end); j >= (end - val); j--) {
          pixel[index] = 0;
          index -= width;
        }
      }
    }
  }


  return 0;
}



///  Copy mask attributes and data from another mask.
///
///  \param mask      mask to copy data from
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::copy (tlMask *mask)
{
  int size;


  // check
  if ((!mask)) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  this->width  = mask->width;
  this->height = mask->height;
  this->cx     = mask->cx;
  this->cy     = mask->cy;
  this->x      = mask->x;
  this->y      = mask->y;
  this->area   = mask->area;
  size         = width * height * sizeof (tlPixel);

  // reallocate pixel map
  pixel = (tlPixel *) realloc (pixel, size);
  if (pixel == NULL) {
    TL_ERRNO_MSG ("tlMask::set()", TL_ERROR_ALLOCATION_FAILED);
    return -1;
  }


  // copy mask data
  memcpy (pixel, mask->pixel, size);
  return 0;
}



///  Copy mask attributes and data to another mask.
///
///  \param mask      mask to copy data to
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::copyTo (tlMask *mask)
{
  return mask->copy (this);
}



///  Adjust to any rectangular component. The resulting mask will be
///  the size of the rect argument. The mask pixel will be translated
///  according to the mask x,y position in the rect coordonates.
///
///  \param rect pointer to a valid rectangular object to adjust to
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::adjust (tlRect *rect)
{
  register tlPixel *oldptr;
  register tlPixel *newptr;

  int i,j;
  int	x1,x2,y1,y2;
  int	oldwidth;


  // check args
  if (!rect) {
    TL_ERRNO_MSG ("tlMask::adjust()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1       = TL_MAX (0, x);
  x2       = TL_MIN ((rect->width), (x + width));
  y1       = TL_MAX (0, y);
  y2       = TL_MIN ((rect->height), (y + height));
  oldwidth = width;

  // reset params
  x      = 0;
  y      = 0;
  width  = rect->width;
  height = rect->height;

  // create new array
  tlPixel *tmp = tl_create_array (width, height, 1);
  memset (tmp, '\0', width * height);

  // copy old array at correct location in the new array
  for (j = y1; j < y2; j++) {
    newptr = tmp + j * width + x1;
    oldptr = pixel + (j - y1) * oldwidth;
    for (i = x1; i < x2; i++) {
      *(newptr++) = *(oldptr++);
    }
  }


  // finish the job
  free (pixel);
  pixel = tmp;
  return 0;
}



///  Minimize the mask size so that it is just big enough to contain
///  active pixels. This method gets rid of the "empty edges" of the mask.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlMask::minimize ()
{
  register tlPixel *oldptr;
  register tlPixel *newptr;

  int	    i, j;
  int	    x1, x2, y1, y2;
  int	    oldwidth;
  tlPixel *tmp;


  // determine minimal parameters
  y1     = height;
  y2     = 0;
  x1     = width;
  x2     = 0;
  newptr = pixel;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      if (*(newptr++)) {
        if (y1 > j) y1 = j;
        if (y2 < j)	y2 = j;
        if (x1 > i)	x1 = i;
        if (x2 < i)	x2 = i;
      }
    }
  }

  // reset params
  oldwidth = width;
  x       += x1;
  y       += y1;
  width    = (x2 - x1);
  height   = (y2 - y1);

  // create new temporary array
  tmp = tl_create_array (width, height, 1);
  memset (tmp, '\0', width * height);

  // copy old array at correct location in the new array
  for (j = 0; j < height; j++) {
    newptr = tmp + j * width;
    oldptr = pixel + (j + y1) * oldwidth + x1;
    for (i = 0; i < width; i++) {
      *(newptr++) = *(oldptr++);
    }
  }


  // finish the job
  free (pixel);
  pixel = tmp;
  return 0;
}
