/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <assert.h>
#include <memory.h>

#include "tlMemory.h"
#include "tlRect.h"



///  \file tlMemory.h
///  \brief Internal memory allocation routines.
///  \internal


///  Allocate a new pixel array.
///  
///  \param width         image width  [pix]
///  \param height        image height [pix]
///  \param pixel_width   byte per pixel
///  
///  \return
///  A pointer to the new array, NULL if allocation failed.

tlPixel *
tl_create_array (int width, 
                 int height, 
                 int pixel_width)
{
  int     count;
  tlPixel *new_array;


  count = width * height * pixel_width * sizeof (tlPixel);
  count = (count == 0) ? 1 : count;
  new_array = (tlPixel *) malloc (count);


  return new_array;
}



///  Allocate a new high-resolution pixel array.
///  
///  \param width         image width  [pix]
///  \param height        image height [pix]
///  \param pixel_width   byte per pixel
///  
///  \return
///  A pointer to the new array, NULL if allocation failed.

tlHRPixel *
tl_create_hrarray (int width, 
                   int height, 
                   int pixel_width)
{
  int       count;
  tlHRPixel *new_array;


  count = width * height * pixel_width * sizeof (tlHRPixel);
  count = (count == 0) ? 1 : count;
  new_array = (tlHRPixel *) malloc (count);


  return new_array;
}



///  Create a copy of a given array containing an image.
///  
///  \param src           source array
///  \param width         image width  [pix]
///  \param height        image height [pix]
///  \param pixel_width   byte per pixel
///  
///  \return
///  A pointer to the new (copied) array, NULL if allocation failed.
///  
///  \note
///  for optimization reasons, no verification of the validity
///  of the input array is made; in case an invalid pointer is
///  passed, results are impredictable.  

tlPixel *
tl_copy_array (tlPixel *src, 
               int     width, 
               int     height, 
               int     pixel_width)
{
  tlPixel *new_array;
  size_t  count;


  // prepare
  count = width * height * pixel_width * sizeof (tlPixel);
  if (count == 0) {
    return NULL;
  }
  new_array = (tlPixel *) malloc (count);
  if (new_array == NULL) {
    TL_ERRNO_MSG ("tl_copy_array()", TL_ERROR_ALLOCATION_FAILED);
    return NULL;
  }
  assert (src != NULL);


  // copy
  memcpy (new_array, src, count);
  return new_array;
}



///  Create a copy of a given high-resolution array containing an image.
///  
///  \param src           source array
///  \param width         image width  [pix]
///  \param height        image height [pix]
///  \param pixel_width   channel per pixel
///  
///  \return
///  A pointer to the new (copied) array, NULL if allocation failed.
///  
///  \note
///  for optimization reasons, no verification of the validity
///  of the input array is made; in case an invalid pointer is
///  passed, results are impredictable.  

tlHRPixel *
tl_copy_hrarray (tlHRPixel *src, 
                 int       width, 
                 int       height, 
                 int       pixel_width)
{
  tlHRPixel *new_array;
  size_t    count;


  // prepare
  count = width * height * pixel_width * sizeof (tlHRPixel);
  if (count == 0) {
    return NULL;
  }
  new_array = (tlHRPixel *) malloc (count);
  if (new_array == NULL) {
    TL_ERRNO_MSG ("tl_copy_hrarray()", TL_ERROR_ALLOCATION_FAILED);
    return NULL;
  }
  assert (src != NULL);

  // copy
  memcpy (new_array, src, count);


  return new_array;
}



///  Create a copy of a portion of a given array containing an image;
///  new (copied) array has the same size as the original.
///  
///  \param src         source array
///  \param width       image width  [pix]
///  \param height      image height [pix]
///  \param pixel_width byte per pixel
///  \param rect        portion of the image to copy
///  
///  \return
///  A pointer to the new (copied) array, NULL if allocation failed.
///  
///  \note
///  for optimization reasons, no verification of the validity
///  of the input array is made; in case an invalid pointer is
///  passed, results are impredictable.  

tlPixel *
tl_copy_array (tlPixel *src, 
               int     width, 
               int     height, 
               int     pixel_width,
               tlRect  *rect)
{
  int     i, index;
  int     x1;
  int     y1, y2;
  size_t  count;
  tlPixel *new_array;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);
  count = (rect->width) * pixel_width * sizeof (tlPixel);

  // proceed
  new_array = (tlPixel *) malloc (width * height * pixel_width * sizeof (tlPixel));
  for (i = y1; i < y2; i++) {
    index = TL_PIXEL (i, x1, width) * pixel_width;
    memcpy (&new_array[index], &src[index], count);
  }


  return new_array;
}



///  Create a copy of a portion of a given high-resolution array containing an image;
///  new (copied) array has the same size as the original.
///  
///  \param src         source array
///  \param width       image width  [pix]
///  \param height      image height [pix]
///  \param pixel_width channels per pixel
///  \param rect        portion of the image to copy
///  
///  \return
///  A pointer to the new (copied) array, NULL if allocation failed.
///  
///  \note
///  for optimization reasons, no verification of the validity
///  of the input array is made; in case an invalid pointer is
///  passed, results are impredictable.  

tlHRPixel *
tl_copy_hrarray (tlHRPixel *src, 
                 int       width, 
                 int       height, 
                 int       pixel_width,
                 tlRect    *rect)
{
  int       i, index;
  int       x1;
  int       y1, y2;
  size_t    count;
  tlHRPixel *new_array;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);
  count = (rect->width) * pixel_width * sizeof (tlHRPixel);

  // proceed
  new_array = (tlHRPixel *) malloc (width * height * pixel_width * sizeof (tlHRPixel));
  for (i = y1; i < y2; i++) {
    index = TL_PIXEL (i, x1, width) * pixel_width;
    memcpy (&new_array[index], &src[index], count);
  }


  return new_array;
}



///  Create a copy of a portion of a given array containing an image;
///  new (copied) array has the same size as the portion of the image to duplicate.
///  
///  \param src         source array
///  \param width       image width  [pix]
///  \param height      image height [pix]
///  \param pixel_width byte per pixel
///  \param rect        portion of the image to copy
///  
///  \return
///  A pointer to the new (copied) array, NULL if allocation failed.
///  
///  \note
///  for optimization reasons, no verification of the validity
///  of the input array is made; in case an invalid pointer is
///  passed, results are impredictable.  

tlPixel *
tl_copy_subarray (tlPixel *src, 
                  int     width, 
                  int     height, 
                  int     pixel_width,
                  tlRect  *rect)
{
  int     i;
  int     index;
  int     x1;
  int     y1, y2;
  int     new_index;
  size_t  count;
  tlPixel *new_array;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);
  count = (rect->width) * pixel_width * sizeof (tlPixel);

  // proceed
  new_array = (tlPixel *) malloc (count * (rect->height) * sizeof (tlPixel));
  new_index = 0;
  for (i = y1; i < y2; i++) {
    index = TL_PIXEL (i, x1, width) * pixel_width;
    memcpy (&new_array[new_index], &src[index], count);
    new_index += count;
  }


  return new_array;
}



///  Create a copy of a portion of a given high-resolution array containing an image;
///  new (copied) array has the same size as the portion of the image to duplicate.
///  
///  \param src         source array
///  \param width       image width  [pix]
///  \param height      image height [pix]
///  \param pixel_width byte per pixel
///  \param rect        portion of the image to copy
///  
///  \return
///  A pointer to the new (copied) array, NULL if allocation failed.
///  
///  \note
///  for optimization reasons, no verification of the validity
///  of the input array is made; in case an invalid pointer is
///  passed, results are impredictable.  

tlHRPixel *
tl_copy_hrsubarray (tlHRPixel *src, 
                    int       width, 
                    int       height, 
                    int       pixel_width,
                    tlRect    *rect)
{
  int       i;
  int       index;
  int       x1;
  int       y1, y2;
  int       new_index;
  size_t    count;
  tlHRPixel *new_array;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);
  count = (rect->width) * pixel_width * sizeof (tlHRPixel);

  // proceed
  new_array = (tlHRPixel *) malloc ((rect->width) * (rect->height) * pixel_width * sizeof (tlHRPixel));
  new_index = 0;
  for (i = y1; i < y2; i++) {
    index = TL_PIXEL (i, x1, width) * pixel_width;
    memcpy (&new_array[new_index], &src[index], count);
    new_index += count;
  }


  return new_array;
}
