/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <memory.h>

#include "tlMorph.h"
#include "tlRect.h"

#define TL_MORPH_HIT          0x01
#define TL_MORPH_MISS         0x02
#define TL_MORPH_NUMKERNELS   8

static tlPixel tl_morph_kernels[][9] = {
	{  	2,  2,  2,  
     	0,  1,  0,  
     	1,  1,  1 },
	{  	0,  2,  2,  
     	1,  1,  2,  
     	0,  1,  0 },
	{  	1,  0,  2,  
     	1,  1,  2,  
     	1,  0,  2 },
	{  	0,  1,  0,  
     	1,  1,  2,  
     	0,  2,  2 },
	{  	1,  1,  1,  
     	0,  1,  0,  
     	2,  2,  2 },
	{  	0,  1,  0,  
    	2,  1,  1,  
     	2,  2,  0 },
	{  	2,  0,  1,  
     	2,  1,  1,  
     	2,  0,  1 },
	{  	2,  2,  0,  
     	2,  1,  1,  
     	0,  1,  0 }
};



///  \file tlMorph.h
///  \brief Internal morphological routines.
///  \internal



///  Perform morphological erosion on each channel of a pixel array
///  using a square pattern as an eroding shape.
///  
///  \param src         pixel array (of the appropriate format)
///  \param width       image width [pix]
///  \param height      image height [pix]
///  \param pixel_width pixel width
///  \param size        size of the square mask to use
///  \param rect        portion of the image to erode
///  \param dest        \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_erode_all (tlPixel *src, 
              int     width, 
              int     height, 
              int     pixel_width, 
              int     size,
              tlRect  *rect, 
              tlPixel *dest)
{
  register int index;

  int     i, j, k;
  int     s1, s2;
  int     x1, x2;
  int     y1, y2;
  int     tmp_index;
  tlPixel new_pixel;
  tlPixel min[TL_MAX_CHANNEL];
  tlPixel *tmp;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_erode_all()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  s2 = size / 2;
  s1 = -(size - s2);
  x1 = TL_MAX ((rect->x), -s1);
  x2 = TL_MIN (((rect->x) + (rect->width)), width - s2);
  y1 = TL_MAX ((rect->y), -s1);
  y2 = TL_MIN (((rect->y) + (rect->height)), height - s2);

  // create a temp array to store intermediary results
  tmp = (tlPixel *) malloc (width * height * pixel_width);
  memcpy (tmp, src, width * height * pixel_width);

  // horizontal filter 
  // for each point within filtering window
  for (j = y1; j < y2; j++) {
    int ii;

    // beginning of a line
    for (k = 0; k < pixel_width; k++) {
      min[k] = TL_MAX_PIXEL_VAL;
    }
    index = pixel_width * TL_PIXEL (j, x1, width);
    tmp_index = index + pixel_width * s1;
    for (ii = s1; ii < s2; ii++) {
      for (k = 0; k < pixel_width; k++) {
        min[k] = TL_MIN (src[tmp_index + k], min[k]);
      }
      tmp_index += pixel_width;
    }

    // if we are within a line 
    for (i = x1 + 1; i < x2; i++) {
      index = pixel_width * TL_PIXEL (j, i, width);
      for (k = 0; k < pixel_width; k++) {

        // get new channel min
        new_pixel = src[index + pixel_width * (s2 - 1) + k];
        if (new_pixel <= min[k]) min[k] = new_pixel;
        else {
          min[k] = TL_MAX_PIXEL_VAL;
          for (ii = s1; ii < s2; ii++)
            min[k] = TL_MIN (src[index + pixel_width * ii + k], min[k]);
        }
      }

      // set output
      for (k = 0; k < pixel_width; k++) {
        tmp[index + k] = min[k];
      }
    }
  }

  // vertical filter
  // for each point within filtering window
  for (i = x1; i < x2; i++) {
    int ii;

    // beginning of a column
    index = pixel_width * TL_PIXEL (y1, i, width);
    for (k = 0; k < pixel_width; k++) {
      min[k] = TL_MAX_PIXEL_VAL;
    }
    tmp_index = index + pixel_width * width * s1;
    for (ii = s1; ii < s2; ii++) {
      for (k = 0; k < pixel_width; k++) {
        min[k] = TL_MIN (tmp[tmp_index], min[k]);
      }
      tmp_index += pixel_width;
    }

    // if we are within a column 
    for (j = y1 + 1; j < y2; j++) {
      index = pixel_width * TL_PIXEL (j, i, width);

      // get new channel min
      for (k = 0; k < pixel_width; k++) {
        new_pixel = tmp[index + pixel_width * width * (s2 - 1) + k];
        if (new_pixel <= min[k]) min[k] = new_pixel;
        else {
          min[k] = TL_MAX_PIXEL_VAL;
          for (ii = s1; ii < s2; ii++)
            min[k] =
            TL_MIN (tmp[index + pixel_width * width * ii + k], min[k]);
        }
      }

      // set output
      for (k = 0; k < pixel_width; k++) {
        dest[index + k] = min[k];
      }
    }
  }


  // get rid of temporary image
  free (tmp);
  return 0;
}



///  Perform morphological erosion on a GRAY pixel array
///  using a square pattern as an eroding shape.
///  
///  \param src     pixel array (of the appropriate format)
///  \param width   image width [pix]
///  \param height  image height [pix]
///  \param size    size of the square mask to use
///  \param rect    portion of the image to erode
///  \param dest    \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_erode_gray (tlPixel *src, 
               int     width, 
               int     height, 
               int     size, 
               tlRect  *rect,
               tlPixel *dest)
{
  register int index;

  int     i, j;
  int     s1, s2;
  int     x1, x2;
  int     y1, y2;
  int     tmp_index;
  tlPixel min;
  tlPixel new_pixel;
  tlPixel *tmp;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_erode_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // extract useful data
  s2 = size / 2;
  s1 = -(size - s2);
  x1 = TL_MAX ((rect->x), -s1);
  x2 = TL_MIN (((rect->x) + (rect->width)), width - s2);
  y1 = TL_MAX ((rect->y), -s1);
  y2 = TL_MIN (((rect->y) + (rect->height)), height - s2);

  // create a temp array to store intermediary results
  tmp = (tlPixel *) malloc (TL_GRAY_WIDTH * width * height);
  memcpy (tmp, src, TL_GRAY_WIDTH * width * height);

  // horizontal filter 
  // for each point within filtering window
  for (j = y1; j < y2; j++) {
    int ii;

    // beginning of a line
    min = TL_MAX_PIXEL_VAL;
    index = TL_PIXEL (j, x1, width);
    tmp_index = index + s1;
    for (ii = s1; ii < s2; ii++) {
      min = TL_MIN (src[tmp_index], min);
      tmp_index += TL_GRAY_WIDTH;
    }

    // within a line
    for (i = x1 + 1; i < x2; i++) {
      index = TL_PIXEL (j, i, width);
      new_pixel = src[index + (s2 - 1)];
      if (new_pixel <= min)	min = new_pixel;
      else {
        min = TL_MAX_PIXEL_VAL;
        for (ii = s1; ii < s2; ii++)
          min = TL_MIN (src[index + ii], min);
      }

      // set output
      tmp[index] = min;
    }
  }

  // vertical filter
  // for each point within filtering window
  for (i = x1; i < x2; i++) {
    int ii;

    // beginning of a column
    min = TL_MAX_PIXEL_VAL;
    index = TL_PIXEL (y1, i, width);
    tmp_index = index + width * s1;
    for (ii = s1; ii < s2; ii++) {
      min = TL_MIN (tmp[tmp_index], min);
      tmp_index += TL_GRAY_WIDTH;
    }

    // within a column
    for (j = y1 + 1; j < y2; j++) {
      index = TL_PIXEL (j, i, width);
      new_pixel = tmp[index + width * (s2 - 1)];
      if (new_pixel <= min)	min = new_pixel;
      else {
        min = TL_MAX_PIXEL_VAL;
        for (ii = s1; ii < s2; ii++)
          min = TL_MIN (tmp[index + width * ii], min);
      }

      // set output
      dest[index] = min;
    }
  }


  // get rid of temporary image
  free (tmp);
  return 0;
}



///  Perform morphological erosion on a BIN pixel array
///  using a square pattern as an eroding shape.
///  
///  \param src     pixel array (of the appropriate format)
///  \param width   image width [pix]
///  \param height  image height [pix]
///  \param size    size of the square mask to use
///  \param rect    portion of the image to erode
///  \param dest    \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_erode_bin (tlPixel *src, 
              int     width, 
              int     height, 
              int     size, 
              tlRect  *rect,
              tlPixel *dest)
{
  register int index;

  int i, j;
  int x1, x2;
  int y1, y2;
  int s1, s2;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_erode_bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  s2 = size / 2;
  s1 = -(size - s2);
  x1 = TL_MAX ((rect->x), -s1);
  x2 = TL_MIN (((rect->x) + (rect->width)), width - s2);
  y1 = TL_MAX ((rect->y), -s1);
  y2 = TL_MIN (((rect->y) + (rect->height)), height - s2);

  // process
  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      int ii, jj;

      // if original is set, ignore what follows
      if (src[TL_PIXEL (j, i, width)] != 0)
        continue;

      // else padd the slider area around current pixel in image
      for (jj = s1; jj < s2; jj++) {
        index = (j + jj) * width + (i + s1);
        for (ii = s1; ii < s2; ii++) {
          dest[index++] = 0;
        }
      }
    }
  }


  return 0;
}



///  Perform morphological dilatation on all channels of a pixel array
///  using a square pattern as an dilatating shape.
///  
///  \param src         pixel array (of the appropriate format)
///  \param width       image width [pix]
///  \param height      image height [pix]
///  \param pixel_width pixel width
///  \param size        size of the square mask to use
///  \param rect        portion of the image to dilate
///  \param dest        \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_dilate_all (tlPixel *src, 
               int     width, 
               int     height, 
               int     pixel_width,
               int     size, 
               tlRect  *rect, 
               tlPixel *dest)
{
  register int index;

  int     i, j, k;
  int     s1, s2;
  int     x1, x2;
  int     y1, y2;
  int     tmp_index;
  tlPixel new_pixel;
  tlPixel max[TL_MAX_CHANNEL];
  tlPixel *tmp;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_dilate_all()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  s2 = size / 2;
  s1 = -(size - s2);
  x1 = TL_MAX ((rect->x), -s1);
  x2 = TL_MIN (((rect->x) + (rect->width)), width - s2);
  y1 = TL_MAX ((rect->y), -s1);
  y2 = TL_MIN (((rect->y) + (rect->height)), height - s2);

  // create a temp array to store intermediary results
  tmp = (tlPixel *) malloc (pixel_width * width * height);
  memcpy (tmp, src, pixel_width * width * height);

  // horizontal filter 
  // for each point within filtering window
  for (j = y1; j < y2; j++) {
    int ii;

    // beginning of a line
    for (k = 0; k < pixel_width; k++) {
      max[k] = TL_MIN_PIXEL_VAL;
    }
    index = pixel_width * TL_PIXEL (j, x1, width);
    tmp_index = index + pixel_width * s1;
    for (ii = s1; ii < s2; ii++) {
      for (k = 0; k < pixel_width; k++) {
        max[k] = TL_MAX (src[tmp_index + k], max[k]);
      }
      tmp_index += pixel_width;
    }

    // within a line
    for (i = x1 + 1; i < x2; i++) {
      index = pixel_width * TL_PIXEL (j, i, width);
      for (k = 0; k < pixel_width; k++) {
        new_pixel = src[index + pixel_width * (s2 - 1) + k];
        if (new_pixel >= max[k])	max[k] = new_pixel;
        else {
          max[k] = TL_MIN_PIXEL_VAL;
          for (ii = s1; ii < s2; ii++)
            max[k] = TL_MAX (src[index + pixel_width * ii + k], max[k]);
        }
      }

      // set output
      for (k = 0; k < pixel_width; k++) {
        tmp[index + k] = (tlPixel) max[k];
      }
    }
  }

  // vertical filter
  // for each point within filtering window
  for (i = x1; i < x2; i++) {
    int ii;

    // beginning of a column
    for (k = 0; k < pixel_width; k++) {
      max[k] = TL_MIN_PIXEL_VAL;
    }
    index = pixel_width * TL_PIXEL (y1, i, width);
    tmp_index = index + pixel_width * width * s1;
    for (ii = s1; ii < s2; ii++) {
      for (k = 0; k < pixel_width; k++) {
        max[k] = TL_MAX (tmp[tmp_index + k], max[k]);
      }
      tmp_index += pixel_width;
    }

    // within a column
    for (j = y1 + 1; j < y2; j++) {
      index = pixel_width * TL_PIXEL (j, i, width);
      for (k = 0; k < pixel_width; k++) {
        new_pixel = tmp[index + pixel_width * width * (s2 - 1) + k];
        if (new_pixel >= max[k])	max[k] = new_pixel;
        else {
          max[k] = TL_MIN_PIXEL_VAL;
          for (ii = s1; ii < s2; ii++)
            max[k] =
            TL_MAX (tmp[index + pixel_width * width * ii + k], max[k]);
        }
      }

      // set output
      for (k = 0; k < pixel_width; k++) {
        dest[index + k] = max[k];
      }
    }
  }


  // get rid of temporary image
  free (tmp);
  return 0;
}



///  Perform morphological dilatation on a GRAYpixel array
///  using a square pattern as an dilatating shape.
///  
///  \param src     pixel array (of the appropriate format)
///  \param width   image width [pix]
///  \param height  image height [pix]
///  \param size    size of the square mask to use
///  \param rect    portion of the image to dilate
///  \param dest    \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_dilate_gray (tlPixel *src, 
                int     width, 
                int     height, 
                int     size, 
                tlRect  *rect,
                tlPixel *dest)
{
  register int index;

  int     i, j;
  int     s1, s2;
  int     x1, x2;
  int     y1, y2;
  int     tmp_index;
  tlPixel new_pixel;
  tlPixel max;
  tlPixel *tmp;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_dilate_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // extract useful data
  s2 = size / 2;
  s1 = -(size - s2);
  x1 = TL_MAX ((rect->x), -s1);
  x2 = TL_MIN ((rect->x) + (rect->width), width - s2);
  y1 = TL_MAX ((rect->y), -s1);
  y2 = TL_MIN ((rect->y) + (rect->height), height - s2);

  // create a temp array to store intermediary results
  tmp = (tlPixel *) malloc (TL_GRAY_WIDTH * width * height);
  memcpy (tmp, src, width * height * TL_GRAY_WIDTH);

  // horizontal filter 
  // for each point within filtering window
  for (j = y1; j < y2; j++) {
    int ii;

    // beginning of a line
    max = TL_MIN_PIXEL_VAL;
    index = j * width + x1;
    tmp_index = index + s1;
    for (ii = s1; ii < s2; ii++) {
      max = TL_MAX (src[tmp_index], max);
      tmp_index += TL_GRAY_WIDTH;
    }

    // within a line
    for (i = x1 + 1; i < x2; i++) {
      index = j * width + i;
      new_pixel = src[index + (s2 - 1)];
      if (new_pixel >= max)	max = new_pixel;
      else {
        max = TL_MIN_PIXEL_VAL;
        for (ii = s1; ii < s2; ii++)
          max = TL_MAX (src[index + ii], max);
      }

      // set output
      tmp[index] = max;
    }
  }

  // vertical filter
  // for each point within filtering window
  for (i = x1; i < x2; i++) {
    int ii;

    // beginning of a column
    max = TL_MIN_PIXEL_VAL;
    index = y1 * width + i;
    tmp_index = index + width * s1;
    for (ii = s1; ii < s2; ii++) {
      max = TL_MAX (tmp[tmp_index], max);
      tmp_index += TL_GRAY_WIDTH;
    }

    // within a column
    for (j = y1 + 1; j < y2; j++) {
      index = j * width + i;
      new_pixel = tmp[index + width * (s2 - 1)];
      if (new_pixel >= max)	max = new_pixel;
      else {
        max = TL_MIN_PIXEL_VAL;
        for (ii = s1; ii < s2; ii++)
          max = TL_MAX (tmp[index + width * ii], max);
      }

      // set output
      dest[index] = max;
    }
  }


  // get rid of temporary image
  free (tmp);
  return 0;
}



///  Perform morphological dilatation on a BIN pixel array
///  using a square pattern as an dilatating shape.
///  
///  \param src     pixel array (of the appropriate format)
///  \param width   image width [pix]
///  \param height  image height [pix]
///  \param size    size of the square mask to use
///  \param rect    portion of the image to dilate
///  \param dest    \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_dilate_bin (tlPixel *src, 
               int     width, 
               int     height, 
               int     size, 
               tlRect  *rect,
               tlPixel *dest)
{
  register int index;

  int i, j;
  int x1, x2;
  int y1, y2;
  int s1, s2;


  // verify parameters
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_dilate_bin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // temp variables to speed up conversion
  s2 = size / 2;
  s1 = -(size - s2);
  x1 = TL_MAX ((rect->x) + s1, -s1);
  x2 = TL_MIN ((rect->x) + (rect->width), width - s2);
  y1 = TL_MAX ((rect->y) + s1, -s1);
  y2 = TL_MIN ((rect->y) + (rect->height), height - s2);

  for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      int ii, jj;

      // if original is set, ignore what follows
      if (src[j * width + i] == 0)	continue;

      // else padd the slider area around current pixel in image
      for (jj = s1; jj < s2; jj++) {
        index = (j + jj) * width + (i + s1);
        for (ii = s1; ii < s2; ii++) {
          dest[index] = TL_MAX_PIXEL_VAL;
          index += 1;
        }
      }
    }
  }


  return 0;
}



///  Perform morphological thinning on a BIN pixel array.
///  
///  \param src     pixel array (of the appropriate format)
///  \param width   image width [pix]
///  \param height  image height [pix]
///  \param kernel  thinning kernel
///  \param rect    portion of the image to dilate
///  \param dest    \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

inline int 
tl_thin_hitandmiss (tlPixel *src, 
                    int     width, 
                    int     height, 
                    tlPixel *kernel, 
                    tlRect  *rect, 
                    tlPixel *dest)
{
  register tlPixel *sptr;
  register tlPixel *dptr;
  register tlPixel *ptr;

  int  i, j, k, l;
  int  x1, x2;
  int  y1, y2;
  int  count;
  int  index;
  bool match;


  // verify parameters
  if ((!src) || (!rect) || (!dest) || (!kernel)) {
    TL_ERRNO_MSG ("tl_thin_hitandmiss()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // temp variables to speed up conversion
  x1 = TL_MAX ((rect->x), 1);
  x2 = TL_MIN ((rect->x) + (rect->width), width-1);
  y1 = TL_MAX ((rect->y), 1);
  y2 = TL_MIN ((rect->y) + (rect->height), height-1);


  count = 0;
  for (j=y1; j<y2; j++) {
    sptr = src  + j*width+x1-1;
    dptr = dest + j*width+x1-1;
    for (i=x1; i<x2; i++) {
      match = true;
      ptr   = sptr-width-1;
      index = 0;
      for (k=0; match && k<3; k++) {
        for(l=0; match && l<3; l++) {
          match = (!kernel[index] || ( *ptr && (kernel[index]&TL_MORPH_HIT )) || (!*ptr && (kernel[index]&TL_MORPH_MISS)));
          ++index;
          ++ptr;
        }
        ptr += width-3;
      }
      if (match) *dptr = 0;
      else *dptr = *sptr;
      if (*dptr) ++count;
      ++sptr;
      ++dptr;
    }
  }


  return count;
}



///  Perform skeletonization on a BIN pixel array.
///  
///  \param src     pixel array (of the appropriate format)
///  \param width   image width [pix]
///  \param height  image height [pix]
///  \param rect    portion of the image to dilate
///  \param dest    \b [out] destination array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_thin (tlPixel *src, 
         int     width, 
         int     height, 
         tlRect  *rect, 
         tlPixel *dest)
{
  int res, lastres;
  int iteration = 1;
  int lastmod   = 0;
  int k         = 0;
  

  res = tl_thin_hitandmiss (src, width, height, tl_morph_kernels[k], rect, dest);
  while( iteration < lastmod + TL_MORPH_NUMKERNELS + 1) {
    if (++k == TL_MORPH_NUMKERNELS) k = 0;
    lastres = res;
    res = tl_thin_hitandmiss (dest, width, height, tl_morph_kernels[k], rect, dest);
    if (res != lastres) lastmod = iteration;
    iteration++;
  }


  return 0;
}
