/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlNormalize.h"
#include "tlRect.h"
#include "tlMask.h"



///  \file tlNormalize.h
///  \brief Internal normalization routines.
///  \internal



///  Normalize a GRAY picture so that it uses to full range
///  as defined by TL_MIN_PIXEL_VAL and TL_MAX_PIXEL_VAL.
///  
///  \param src     the pixel array to normalize
///  \param width   image width
///  \param height  image height
///  \param rect    portion of the image to process
///  \param dest    \b [out] destination pixel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \note
///  src and dest can be pointing to the same pixel array.
///  When normalizing a subimage, the normalization occurs
///  in the subimage ONLY (the rest of the image is ignored);
///  therefore, the same area of an image can render differently
///  if normalized within the whole image or a subimage.

int
tl_normalize_gray (tlPixel *src, 
                   int     width, 
                   int     height, 
                   tlRect  *rect,
                   tlPixel *dest)
{
  register int index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  double  scale;
  tlPixel val, min_val, max_val;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_normalize_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // extract current minima and maxima
  min_val = TL_MAX_PIXEL_VAL;
  max_val = TL_MIN_PIXEL_VAL;
  for (j = y1; j < y2; j++) {
    index = TL_GRAY_WIDTH * TL_PIXEL (j, x1, width);
    for (i = x1; i < x2; i++) {
      val = src[index];
      max_val = (val > max_val) ? val : max_val;
      min_val = (val < min_val) ? val : min_val;
      index += TL_GRAY_WIDTH;
    }
  }

  // compute scaling factor
  scale = ((double) (TL_MAX_PIXEL_VAL - TL_MIN_PIXEL_VAL)) / ((double) (max_val - min_val));

  // scale image
  for (j = y1; j < y2; j++) {
    index = TL_GRAY_WIDTH * TL_PIXEL (j, x1, width);
    for (i = x1; i < x2; i++) {
      dest[index] = (tlPixel) ((double) (src[index] - min_val) * scale);
      index += TL_GRAY_WIDTH;
    }
  } 


  return 0;
}



///  Normalize a GRAY picture so that it uses to full range
///  as defined by TL_MIN_PIXEL_VAL and TL_MAX_PIXEL_VAL.
///  The normlization is limited to the masked area.
///  
///  \param src     the pixel array to normalize
///  \param width   image width
///  \param height  image height
///  \param mask    data mask
///  \param rect    portion of the image to process
///  \param dest    \b [out] destination pixel array
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///  
///  \note
///  src and dest can be pointing to the same pixel array.
///  When normalizing a subimage, the normalization occurs
///  in the subimage ONLY (the rest of the image is ignored);
///  therefore, the same area of an image can render differently
///  if normalized within the whole image or a subimage.

int
tl_normalize_gray (tlPixel *src, 
                   int     width, 
                   int     height, 
                   tlMask  *mask,
                   tlRect  *rect, 
                   tlPixel *dest)
{
  register int index;

  int     i, j;
  int     x1, x2;
  int     y1, y2;
  double  scale;
  tlPixel val, min_val, max_val;
  tlPixel *mask_data;


  // check
  if ((!src) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tl_normalize_gray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  x1        = TL_MAX ((rect->x), 0);
  x2        = TL_MIN ((x1 + (rect->width)), width);
  y1        = TL_MAX ((rect->y), 0);
  y2        = TL_MIN ((y1 + (rect->height)), height);
  mask_data = mask->pixel;

  // extract current minima and maxima
  min_val = TL_MAX_PIXEL_VAL;
  max_val = TL_MIN_PIXEL_VAL;
  for (j = y1; j < y2; j++) {
    index = TL_GRAY_WIDTH * TL_PIXEL (j, x1, width);
    for (i = x1; i < x2; i++) {
      if (mask_data[index]) {
        val = src[index];
        max_val = (val > max_val) ? val : max_val;
        min_val = (val < min_val) ? val : min_val;
      }
      index += TL_GRAY_WIDTH;
    }
  }

  // compute scaling factor
  scale = ((double) (TL_MAX_PIXEL_VAL - TL_MIN_PIXEL_VAL)) / ((double) (max_val - min_val));

  // scale image
  for (j = y1; j < y2; j++) {
    index = TL_GRAY_WIDTH * TL_PIXEL (j, x1, width);
    for (i = x1; i < x2; i++) {
      if (mask_data[index]) dest[index] = (tlPixel) ((double) (src[index] - min_val) * scale);
      else dest[index] = TL_MIN_PIXEL_VAL;
      index += TL_GRAY_WIDTH;
    }
  }


  return 0;
}
