/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlObject.h"
#include "tlPoint.h"
#include "tlConst.h"
#include "tlMask.h"



///  \file tlObject.h
///  \brief Tracking object description.

///  \class tlObject
///  \brief Tracking object, contain object description and tracking methods.
///
///  A tlObject is a set of data (pixel array, dimensions, velocity, ...) that 
///  describes an object in an image, in a given image format. It contains the 
///  methods necessary to retrieve the object position in subsequent images, 
///  as well as the routines that update that object properties from an image.

///  \var tlMask *tlObject::shape
///  Binary mask of the size of the pixel field.
///  Active pixels in the mask indicate pixels that belong to the object.

///  \var float tlObject::vx
///  Velocity of the object along the X axis [pixels/sec].

///  \var float tlObject::vy
///  Velocity of the object along the Y axis [pixels/sec].



///  Constructor; creates an empty object.
/// 
///  \note
///  See \ref errors "error management" for details.

tlObject::tlObject () : tlPixelArray ()
{
  if (tl_errno != TL_NO_ERROR) return;

  // init location
  setLocation (0, 0);

  // create veloctiy const
  vx = new tlConst < float >(0.0, 1.0f);
  vy = new tlConst < float >(0.0, 1.0f);

  // create mask
  shape = new tlMask (1, 1, width, height);
  if (!shape) {
    TL_ERRNO_MSG ("tlObject::create()", TL_ERROR_ALLOCATION_FAILED);
    return;
  }

  tl_errno = TL_NO_ERROR;
}



///  Constructor; creates an empty object.
///
///  \param format   object pixel array format
///  \param x        object x position in image [pix]
///  \param y        object y position in image [pix]
///  \param width    object width  [pix]
///  \param height   object height [pix]
/// 
///  \note
///  See \ref errors "error management" for details.

tlObject::tlObject (tl_format format, 
                    int       x, 
                    int       y, 
                    int       width, 
                    int       height) : tlPixelArray (format, width, height)
{
  if (tl_errno != TL_NO_ERROR) return;

  // init location
  setLocation (x, y);
  setSize (width, height);

  // create veloctiy const
  vx = new tlConst < float >(0.0, 1.0f);
  vy = new tlConst < float >(0.0, 1.0f);

  // create mask
  shape = new tlMask (width / 2, height / 2, width, height);
  if (!shape) {
    TL_ERRNO_MSG ("tlObject::create()", TL_ERROR_ALLOCATION_FAILED);
    return;
  }

  tl_errno = TL_NO_ERROR;
}



///  Destructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlObject::~tlObject ()
{
  if (shape) delete shape;
  if (vx) delete vx;
  if (vy) delete vy;

  tl_errno = TL_NO_ERROR;
}



///  Return the predicted current position.
///
///  \return
///  A point located at the predicted object position.
///
///  \sa
///  getLocation()

tlPoint *
tlObject::getPredictedPosition ()
{
  int           px, py;
  unsigned long dt;

  dt = tl_timer_get_time () - timeStamp;
  px = x + (int) (vx->getVal () * dt);
  py = y + (int) (vy->getVal () * dt);
  return new tlPoint (px, py);
}



///  Return the predicted current X position.
///
///  \return
///  The predicted object X position.
///
///  \sa
///  getPosition()\n
///  getPredictedPosition()

int
tlObject::getPredictedX ()
{
  return (x + (int) (vx->getVal () * (tl_timer_get_time () - timeStamp)));
}



///  Return the predicted current Y position.
///
///  \return
///  The predicted object Y position.
///
///  \sa
///  getPosition()\n
///  getPredictedPosition()

int
tlObject::getPredictedY ()
{
  return (y + (int) (vy->getVal () * (tl_timer_get_time () - timeStamp)));
}
