/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlObject.h"
#include "tlImage.h"
#include "tlPoint.h"
#include "tlColor.h"
#include "tlColorModel.h"
#include "tlMask.h"



///  Extract (i.e. segment) an object from an image at a given location.
/// 
///  \param image     source image
///  \param point     extraction starting point
///  \param gapColor  color tolerance around color value at given pixel (if NULL, a zero tolerance will be use around the given pixel color)
///  \param varColor  color variation tolerance (if NULL, no color variation will be tolerated)
///  \param connex    connectivity
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  The object extraction routine takes the colour at the 'point' location
///  in the image. It then recursively checks neighbour pixels. If a tested
///  pixel is within a 'gapColor' interval around the original color, it
///  is accepted as part of the object and the recursive testing continues.
///  The 'varColor' argument specifies by how much the reference color is
///  allowed to drift spatially. This makes it possible to detect object
///  boundaries, while tolerating color changes due to shadows, or non-linearity
///  of the acquisition system.

int
tlObject::extract (tlImage *image, 
                   tlPoint *point, 
                   tlColor *gapColor,
                   tlColor *varColor, 
                   int     connex)
{
  int             i, j;
  int             x1, x2, y1, y2;
  int             index;
  int             width, height;
  char            *objmap;
  tl_extract_data data;


  // check
  if ((!image) || (!point)) {
    TL_ERRNO_MSG ("tlObject::extract()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // save format
  format = image->format;

  // shortcuts
  x1     = point->x;
  y1     = point->y;
  x2     = point->x;
  y2     = point->y;
  width  = image->width;
  height = image->height;

  // create pattern buffer
  index = 0;
  objmap = (char *) malloc (width * height * sizeof (char));
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      objmap[index++] = -1;
    }
  }

  // init recursive data struct
  data.width        = width;
  data.height       = height;
  data.px           = point->x;
  data.py           = point->y;
  data.src          = image->pixel;
  data.objmap       = objmap;
  data.x1           = &x1;
  data.x2           = &x2;
  data.y1           = &y1;
  data.y2           = &y2;
  data.lx1          = 0;
  data.lx2          = width;
  data.ly1          = 0;
  data.ly2          = height;
  data.connectivity = connex;

  // 
  // retrieve matching pixels
  // 

  try {
    switch (image->format) {
    case RGB:
      {
        // assign params
        data.depth = 3;
        if (gapColor != NULL) {
          data.min[0] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 0) - (gapColor->r) / 2)));
          data.max[0] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[0] + gapColor->r)));
          data.min[1] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 1) - (gapColor->g) / 2)));
          data.max[1] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[1] + gapColor->g)));
          data.min[2] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 2) - (gapColor->b) / 2)));
          data.max[2] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[2] + gapColor->b)));
        }
        else {
          data.min[0] = data.max[0] = (tlPixel)image->getPixel (point, 0);
          data.min[1] = data.max[1] = (tlPixel)image->getPixel (point, 1);
          data.min[2] = data.max[2] = (tlPixel)image->getPixel (point, 2);
        }
        if (varColor != NULL) {
          data.tol[0] = varColor->r;
          data.tol[1] = varColor->g;
          data.tol[2] = varColor->b;
        }
        else {
          data.tol[0] = 0;
          data.tol[1] = 0;
          data.tol[2] = 0;
        }

        // launch recursive call
        tl_extract (&data);
      }
      break;
    case NRG:
      {
        // assign params
        data.depth = 2;
        if (gapColor != NULL) {
          data.min[0] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 0) - (gapColor->nr) / 2)));
          data.max[0] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[0] + gapColor->nr)));
          data.min[1] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 1) - (gapColor->ng) / 2)));
          data.max[1] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[1] + gapColor->ng)));
        }
        else {
          data.min[0] = data.max[0] = (tlPixel)image->getPixel (point, 0);
          data.min[1] = data.max[1] = (tlPixel)image->getPixel (point, 1);
        }
        if (varColor != NULL) {
          data.tol[0] = varColor->nr;
          data.tol[1] = varColor->ng;
        }
        else {
          data.tol[0] = 0;
          data.tol[1] = 0;
        }

        // launch recursive call
        tl_extract (&data);
      }
      break;
    case GRAY:
      {
        // assign params
        data.depth = 1;
        if (gapColor != NULL) {
          data.min[0] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 0) - (gapColor->intensity) / 2)));
          data.max[0] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[0] + gapColor->intensity)));
        }
        else {
          data.min[0] = data.max[0] = (tlPixel)image->getPixel (point, 0);
        }
        if (varColor != NULL) data.tol[0] = varColor->intensity;
        else                  data.tol[0] = 0;

        // launch recursive call
        tl_extract (&data);
      }
      break;
    case BIN:
      {
        data.depth = 1;
        if (image->getPixel (point) != 0) {
          data.min[0] = 1;
          data.max[0] = TL_MAX_PIXEL_VAL;
        }
        else {
          data.min[0] = 0;
          data.max[0] = 1;
        }
        data.tol[0] = 0;
        tl_extract (&data);
      }
      break;
    default:
      {
        TL_ERRNO_MSG ("tlObject::extract()", TL_ERROR_INVALID_FORMAT);
        return -1;
      }
    }

    // check for consistency
    if (((x2 - x1) < 1) || ((y2 - y1) < 1)) {
      free (objmap);
      return TL_ERROR_NOT_FOUND;
    }
  }

  // stack overflow : 
  // create object with appropriate properties
  catch (...) {

    // object size
    tlRect *rect = new tlRect (x1, y1, (x2 - x1), (y2 - y1));

    // object image
    this->define (image, rect);

    // object shape
    this->shape->set (objmap, image->width, image->height, rect);

    // clean up
    free (objmap);
    delete rect;

    TL_ERRNO_MSG ("tlObject::extract()", TL_ERROR_STACK_OVERFLOW);
    return -1;
  }

  // 
  // create object with appropriate properties
  // 

  // object size
  tlRect *rect = new tlRect (x1, y1, (x2 - x1), (y2 - y1));

  // object image
  this->define (image, rect);

  // object shape
  this->shape->set (objmap, image->width, image->height, rect);

  // clean up
  free (objmap);
  delete rect;


  return 0;
}



///  Extract (i.e. segment) an object from a portion of an image, 
///  starting at a given location.
/// 
///  \param image     source image
///  \param point     extraction starting point
///  \param gapColor  color tolerance around color value at given pixel (if NULL, a zero tolerance will be use around the given pixel color)
///  \param varColor  color variation tolerance (if NULL, no color variation will be tolerated)
///  \param rect      bounding rectangle for object extraction
///  \param connex    connectivity
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  The object extraction routine takes the colour at the 'point' location
///  in the image. It then recursively checks neighbour pixels. If a tested
///  pixel is within a 'gapColor' interval around the original color, it
///  is accepted as part of the object and the recursive testing continues.
///  The 'varColor' argument specifies by how much the reference color is
///  allowed to drift spatially. This makes it possible to detect object
///  boundaries, while tolerating color changes due to shadows, or
///  non-linearity of the acquisition system.

int
tlObject::extract (tlImage *image, 
                   tlPoint *point, 
                   tlColor *gapColor,
                   tlColor *varColor, 
                   tlRect  *rect, 
                   int     connex)
{
  int             i, j;
  int             x1, x2, y1, y2;
  int             index;
  int             width, height;
  char            *objmap;
  tl_extract_data data;


  // check
  if ((!image) || (!point)) {
    TL_ERRNO_MSG ("tlObject::extract()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // save format
  format = image->format;

  // shortcuts
  x1     = point->x;
  y1     = point->y;
  x2     = point->x;
  y2     = point->y;
  width  = image->width;
  height = image->height;

  // create pattern buffer
  index = 0;
  objmap = (char *) malloc (width * height * sizeof (char));
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      objmap[index++] = -1;
    }
  }

  // init recursive data struct
  data.width        = width;
  data.height       = height;
  data.px           = point->x;
  data.py           = point->y;
  data.src          = image->pixel;
  data.objmap       = objmap;
  data.x1           = &x1;
  data.x2           = &x2;
  data.y1           = &y1;
  data.y2           = &y2;
  data.lx1          = TL_MAX (rect->x, 0);
  data.lx2          = TL_MIN ((rect->x) + (rect->width), width);
  data.ly1          = TL_MAX (rect->y, 0);
  data.ly2          = TL_MIN ((rect->y) + (rect->height), height);
  data.connectivity = connex;

  // 
  // retrieve matching pixels
  // 

  try {
    switch (image->format) {
    case RGB:
      // assign params
      data.depth = 3;
      if (gapColor != NULL) {
        data.min[0] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 0) - (gapColor->r) / 2)));
        data.max[0] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[0] + gapColor->r)));
        data.min[1] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 1) - (gapColor->g) / 2)));
        data.max[1] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[1] + gapColor->g)));
        data.min[2] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 2) - (gapColor->b) / 2)));
        data.max[2] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[2] + gapColor->b)));
      }
      else {
        data.min[0] = data.max[0] = (tlPixel)image->getPixel (point, 0);
        data.min[1] = data.max[1] = (tlPixel)image->getPixel (point, 1);
        data.min[2] = data.max[2] = (tlPixel)image->getPixel (point, 2);
      }
      if (varColor != NULL) {
        data.tol[0] = varColor->r;
        data.tol[1] = varColor->g;
        data.tol[2] = varColor->b;
      }
      else {
        data.tol[0] = 0;
        data.tol[1] = 0;
        data.tol[2] = 0;
      }

      // launch recursive call
      tl_extract (&data);
      break;

    case NRG:
      // assign params
      data.depth = 2;
      if (gapColor != NULL) {
        data.min[0] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 0) - (gapColor->nr) / 2)));
        data.max[0] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[0] + gapColor->nr)));
        data.min[1] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 1) - (gapColor->ng) / 2)));
        data.max[1] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[1] + gapColor->ng)));
      }
      else {
        data.min[0] = data.max[0] = (tlPixel)image->getPixel (point, 0);
        data.min[1] = data.max[1] = (tlPixel)image->getPixel (point, 1);
      }
      if (varColor != NULL) {
        data.tol[0] = varColor->nr;
        data.tol[1] = varColor->ng;
      }
      else {
        data.tol[0] = 0;
        data.tol[1] = 0;
      }

      // launch recursive call
      tl_extract (&data);
      break;

    case GRAY:
      // assign params
      data.depth = 1;
      if (gapColor != NULL) {
        data.min[0] = (tlPixel)(TL_MAX (0, (image->getPixel (point, 0) - (gapColor->intensity) / 2)));
        data.max[0] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (data.min[0] + gapColor->intensity)));
      }
      else {
        data.min[0] = data.max[0] = (tlPixel)image->getPixel (point, 0);
      }
      if (varColor != NULL)
        data.tol[0] = varColor->intensity;
      else
        data.tol[0] = 0;

      // launch recursive call
      tl_extract (&data);
      break;
    default:
      TL_ERRNO_MSG ("tlObject::extract()", TL_ERROR_INVALID_FORMAT);
      return -1;
    }

    // check for consistency
    if (((x2 - x1) < 1) || ((y2 - y1) < 1)) {
      return TL_ERROR_NOT_FOUND;
    }
  }

  // stack overflow : 
  // create object with appropriate properties
  catch (...) {

    // object size
    tlRect *rect = new tlRect (x1, y1, (x2 - x1), (y2 - y1));

    // object image
    this->define (image, rect);

    // object shape
    this->shape->set (objmap, image->width, image->height, rect);

    // clean up
    free (objmap);
    delete rect;

    return TL_ERROR_STACK_OVERFLOW;
  }

  // 
  // create object with appropriate properties
  // 

  // object size
  tlRect *trect = new tlRect (x1, y1, (x2 - x1), (y2 - y1));

  // object image
  this->define (image, trect);

  // object shape
  this->shape->set (objmap, image->width, image->height, trect);

  // clean up
  free (objmap);
  delete trect;


  return 0;
}



///  Extract (i.e. segment) an object from a portion of an image, 
///  starting at a given location.
/// 
///  \param image       source image
///  \param point       extraction starting point
///  \param colorModel  color tolerance around color value at given pixel (if NULL, a zero tolerance will be use around the given pixel color)
///  \param rect        bounding rectangle for object extraction
///  \param threshold   color model application threshold (see \ref tlColorModel "color model" for more details)
///  \param connex      connectivity
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  The object extraction routine checks the colour at the 'point' location
///  in the image. It then recursively checks neighbour pixels. If a tested
///  pixel matches the color model at the given threshold, it is accepted as 
///  part of the object and the recursive testing continues.

int
tlObject::extract (tlImage      *image, 
                   tlPoint      *point, 
                   tlColorModel *colorModel,
                   tlRect       *rect, 
                   int          threshold,
                   int          connex)
{
  int                 i, j;
  int                 x1, x2, y1, y2;
  int                 index;
  int                 width, height;
  char                *objmap;
  tl_extract_data_cm  data;


  // check
  if ((!image) || (!point)) {
    TL_ERRNO_MSG ("tlObject::extract()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // save format
  format = image->format;

  // shortcuts
  x1     = point->x;
  y1     = point->y;
  x2     = point->x;
  y2     = point->y;
  width  = image->width;
  height = image->height;

  // create pattern buffer
  index = 0;
  objmap = (char *) malloc (width * height * sizeof (char));
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      objmap[index++] = -1;
    }
  }

  // init recursive data struct
  data.width        = width;
  data.height       = height;
  data.px           = point->x;
  data.py           = point->y;
  data.src          = image->pixel;
  data.objmap       = objmap;
  data.x1           = &x1;
  data.x2           = &x2;
  data.y1           = &y1;
  data.y2           = &y2;
  data.lx1          = TL_MAX (rect->x, 0);
  data.lx2          = TL_MIN ((rect->x) + (rect->width), width);
  data.ly1          = TL_MAX (rect->y, 0);
  data.ly2          = TL_MIN ((rect->y) + (rect->height), height);
  data.connectivity = connex;

  // 
  // retrieve matching pixels
  // 

  try {
    switch (image->format) {
    case NRG:
      // assign params
      data.cmHist    = colorModel->getNRGHist ();
      data.threshold = threshold;

      // launch recursive call
      tl_extract_cm_nrg (&data);
      break;
    case HSI:
      // assign params
      data.cmHist    = colorModel->getHSHist ();
      data.threshold = threshold;

      // launch recursive call
      tl_extract_cm_hsi (&data);
      break;
    default:
      TL_ERRNO_MSG ("tlObject::extract()", TL_ERROR_INVALID_FORMAT);
      return -1;
    }

    // check for consistency
    if (((x2 - x1) < 1) || ((y2 - y1) < 1)) {
      return TL_ERROR_NOT_FOUND;
    }
  }

  // stack overflow : 
  // create object with appropriate properties
  catch (...) {

    // object size
    tlRect *rect = new tlRect (x1, y1, (x2 - x1), (y2 - y1));

    // object image
    this->define (image, rect);

    // object shape
    this->shape->set (objmap, image->width, image->height, rect);

    // clean up
    free (objmap);
    delete rect;

    return TL_ERROR_STACK_OVERFLOW;
  }

  // 
  // create object with appropriate properties
  // 

  // object size
  tlRect *trect = new tlRect (x1, y1, (x2 - x1), (y2 - y1));

  // object image
  this->define (image, trect);

  // object shape
  this->shape->set (objmap, image->width, image->height, trect);

  // clean up
  free (objmap);
  delete trect;


  return 0;
}
