/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <assert.h>

#include "tlObject.h"
#include "tlImage.h"
#include "tlMask.h"
#include "tlColor.h"



///  Define an object parameters using an image portion.
///  
///  \param image reference image for object definition
///  \param rect  portion of the image to assign to object
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlObject::define (tlImage *image, 
									tlRect  *rect)
{
	// check
	if ((!image) || (!rect)) {
		TL_ERRNO_MSG ("tlObject::define()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// reset object
	deletePixel ();
	format = image->format;

	// extract image data
	x = rect->x;
	y = rect->y;
	width = rect->width;
	height = rect->height;
	pixelWidth = image->pixelWidth;
	pixel =	tl_copy_subarray (image->pixel, image->width, image->height, tl_pixel_width (format), rect);

	// set shape
	assert (shape != NULL);
	shape->set (width, height);
	shape->setLocation (x, y);


	return 0;
}



///  Define object parameters using an image 
///  and appropriate color boundaries
///  
///  \param image reference image for object definition
///  \param minColor lower color boundary
///  \param maxColor upper color boundary
///  \param rect portion of the image to process
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlObject::define (tlImage *image, 
									tlColor *minColor, 
									tlColor *maxColor,
									tlRect  *rect)
{
	// check
	if ((!image) || (!minColor) || (!maxColor) || (!rect)) {
		TL_ERRNO_MSG ("tlObject::define()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// check format
	if ((image->format) != (minColor->format) || (image->format) != (maxColor->format)) {
		TL_ERRNO_MSG ("tlObject::define()", TL_ERROR_INVALID_FORMAT);
		return -1;
	}

	// reset object
	deletePixel ();
	format = image->format;

	// extract image data
	x          = rect->x;
	y          = rect->y;
	width      = rect->width;
	height     = rect->height;
	pixelWidth = image->pixelWidth;
	pixel      =	tl_copy_subarray (image->pixel, image->width, image->height, tl_pixel_width (format), rect);

	// filter pixel array
	filter (minColor, maxColor);

	// set shape
	shape->set (this);


	return 0;
}



///  Copy object data from another object.
/// 
///  \param object object to copy from
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlObject::copy (tlObject *object)
{
	// check
	if (!object) {
		TL_ERRNO_MSG ("tlObject::copy()", TL_ERROR_NULL_POINTER);
		return -1;
	}

	// proceed
	x          = object->x;
	y          = object->y;
	width      = object->width;
	height     = object->height;
	timeStamp  = object->timeStamp;
	format     = object->format;
	pixelWidth = object->pixelWidth;

	// copy data
	deletePixel ();
	pixel = tl_copy_array (object->pixel, width, height, pixelWidth);
	shape->copy (object->shape);


	return 0;
}



///  Copy object data to another object
/// 
///  \param object object to copy to
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlObject::copyTo (tlObject *object)
{
	return object->copy (this);
}
