/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#ifdef LINUX
#include <stdlib.h>
#endif

#include "tlPixel.h"
#include "tlRect.h"



///  \file tlPixel.h
///  \brief Internal pixel manipulation operators.
///  \internal 



///  Elementary pixel addition on pixel array.
///  
///  \param src         pixel array containing initial image
///  \param operand     pixel array containing image to add
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the operation result
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_pixel_add (tlPixel *src, 
              tlPixel *operand, 
              int     width, 
              int     height,
              int     pixel_width, 
              tlRect  *rect, 
              tlPixel *dest)
{
  int j, k;
  int x1, x2, y1, y2;
  int index1, index2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // process
  for (j = y1; j < y2; j++) {
    index1 = pixel_width * TL_PIXEL (j, x1, width);
    index2 = pixel_width * TL_PIXEL (j, x2, width);
    for (k = index1; k < index2; k++) {
      dest[k] = (tlPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (src[k] + operand[k])));
    }
  }


  return 0;
}



///  Elementary pixel subtraction on pixel array.
///  
///  \param src         pixel array containing initial image
///  \param operand     pixel array containing image to subtract
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the operation result
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_pixel_sub (tlPixel *src, 
              tlPixel *operand, 
              int     width, 
              int     height,
              int     pixel_width, 
              tlRect  *rect, 
              tlPixel *dest)
{
  int j, k;
  int x1, x2, y1, y2;
  int index1, index2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // process
  for (j = y1; j < y2; j++) {
    index1 = pixel_width * TL_PIXEL (j, x1, width);
    index2 = pixel_width * TL_PIXEL (j, x2, width);
    for (k = index1; k < index2; k++) {
      dest[k] = (tlPixel) abs ((int) src[k] - (int) operand[k]);
    }
  } 


  return 0;
}


///  Elementary pixel scaling on pixel array.
///  
///  \param src         pixel array containing initial image
///  \param factor      scaling factor
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the operation result
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_pixel_scale (tlPixel *src, 
                float   factor, 
                int     width, 
                int     height,
                int     pixel_width, 
                tlRect  *rect, 
                tlPixel *dest)
{
  int j, k;
  int x1, x2, y1, y2;
  int index1, index2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // process
  for (j = y1; j < y2; j++) {
    index1 = pixel_width * TL_PIXEL (j, x1, width);
    index2 = pixel_width * TL_PIXEL (j, x2, width);
    for (k = index1; k < index2; k++) {
      dest[k] =	(tlPixel) TL_MIN (TL_MAX_PIXEL_VAL, factor * (float) (src[k]));
    }
  }


  return 0;
}



///  Elementary pixel addition on high-resolution pixel array.
///  
///  \param src         pixel array containing initial image
///  \param operand     pixel array containing image to add
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the operation result
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_pixel_hradd (tlHRPixel *src, 
                tlHRPixel *operand, 
                int       width, 
                int       height,
                int       pixel_width, 
                tlRect    *rect, 
                tlHRPixel *dest)
{
  int j, k;
  int x1, x2, y1, y2;
  int index1, index2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // process
  for (j = y1; j < y2; j++) {
    index1 = pixel_width * TL_PIXEL (j, x1, width);
    index2 = pixel_width * TL_PIXEL (j, x2, width);
    for (k = index1; k < index2; k++) {
      dest[k] = (tlHRPixel)(TL_MIN (TL_MAX_PIXEL_VAL, (src[k] + operand[k])));
    }
  }


  return 0;
}



///  Elementary pixel subtraction on high-resolution pixel array.
///  
///  \param src         pixel array containing initial image
///  \param operand     pixel array containing image to subtract
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the operation result
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_pixel_hrsub (tlHRPixel *src, 
                tlHRPixel *operand, 
                int       width, 
                int       height,
                int       pixel_width, 
                tlRect    *rect, 
                tlHRPixel *dest)
{
  int j, k;
  int x1, x2, y1, y2;
  int index1, index2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // process
  for (j = y1; j < y2; j++) {
    index1 = pixel_width * TL_PIXEL (j, x1, width);
    index2 = pixel_width * TL_PIXEL (j, x2, width);
    for (k = index1; k < index2; k++) {
      dest[k] = (tlHRPixel) abs ((int) src[k] - (int) operand[k]);
    }
  } 


  return 0;
}


///  Elementary pixel scaling on high-resolution pixel array.
///  
///  \param src         pixel array containing initial image
///  \param factor      scaling factor
///  \param width       image width 
///  \param height      image height 
///  \param pixel_width pixel width
///  \param rect        portion of the image to process
///  \param dest        \b [out] pixel array containing the operation result
///  
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_pixel_hrscale (tlHRPixel *src, 
                  float     factor, 
                  int       width, 
                  int       height,
                  int       pixel_width, 
                  tlRect    *rect, 
                  tlHRPixel *dest)
{
  int j, k;
  int x1, x2, y1, y2;
  int index1, index2;


  // shortcuts
  x1 = TL_MAX ((rect->x), 0);
  x2 = TL_MIN ((x1 + (rect->width)), width);
  y1 = TL_MAX ((rect->y), 0);
  y2 = TL_MIN ((y1 + (rect->height)), height);

  // process
  for (j = y1; j < y2; j++) {
    index1 = pixel_width * TL_PIXEL (j, x1, width);
    index2 = pixel_width * TL_PIXEL (j, x2, width);
    for (k = index1; k < index2; k++) {
      dest[k] =	(tlHRPixel) TL_MIN (TL_MAX_PIXEL_VAL, factor * (float) (src[k]));
    }
  }


  return 0;
}
