/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlPixelArray.h"



///  \file tlPixelArray.h
///  \brief Pixel array object description.

///  \class tlPixelArray
///  \brief Pixel array object, provides access to pixel array data and processing.
///
///  The base class for every object that contains a pixel array. tlPixelArray contains a pixel array
///  with its dimension, format and properties. It also encapsulates all the pixel operator methods 
///  that can be applied to it. tlImage, toObject and tlMask inherit from tlPixelArray.


///  \var tlPixel *tlPixelArray::pixel
///  Pointer to the pixel array. The array size is (width * height * pixelWidth) bytes.

///  \var int *tlPixelArray::pixelWidth
///  The number of bytes per pixel (depends on the image format).

///  \var int tlPixelArray::format
///  Image format (see tl_format for more information).

///  \var long tlPixelArray::timeStamp
///  Time at which the pixel array was updated (from a hardware digitizer).




///  Constructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlPixelArray::tlPixelArray ()
{
  tl_errno = TL_NO_ERROR;

  pixel   = NULL;
  hrpixel = NULL;
  hrShift = 0;
  create (RGB, TL_DEFAULT_COLS, TL_DEFAULT_ROWS);
}



///  Constructor.
///
///  \param format      pixel array format
///  \param width       pixel array width  [pix]
///  \param height      pixel array height [pix]
/// 
///  \note
///  See \ref errors "error management" for details.

tlPixelArray::tlPixelArray (tl_format format, 
                            int       width, 
                            int       height)
{
  tl_errno = TL_NO_ERROR;

  pixel   = NULL;
  hrpixel = NULL;
  hrShift = 0;
  create (format, width, height);
}



///  Constructor.
///
///  \param pixelArray   pixel array to copy from
/// 
///  \note
///  See \ref errors "error management" for details.

tlPixelArray::tlPixelArray (tlPixelArray *pixelArray)
{
  tl_errno = TL_NO_ERROR;

  pixel   = NULL;
  hrpixel = NULL;
  hrShift = 0;
  create (pixelArray->format, pixelArray->width, pixelArray->height);
  timeStamp = pixelArray->timeStamp;
  pixel = tl_copy_array (pixelArray->pixel, width, height, pixelWidth);
}



///  Destructor.
/// 
///  \note
///  See \ref errors "error management" for details.

tlPixelArray::~tlPixelArray ()
{
  tl_errno = TL_NO_ERROR;
  deletePixel ();
}
