/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"



///  Add a channel to a pixel array. The new channel is added after the 
///  current last channel.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::addChannel (void)
{
  int	    i, j, k;
  int	    new_index, old_index;
  tlPixel *old;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::addChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // backup old pixel array
  old = tl_copy_array (pixel, width, height, pixelWidth);

  // change properties
  format = CUSTOM;
  pixelWidth += 1;
  pixel = (tlPixel *) realloc (pixel, (size_t) (width * height * pixelWidth));

  // shortcuts
  new_index = 0;
  old_index = 0;

  // insert old data back into new pixel array
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      for (k = 0; k < (pixelWidth - 1); k++) {
        pixel[new_index++] = old[old_index++];
      }
      pixel[new_index++] = 0;
    }
  }


  // free
  free (old);
  return 0;
}



///  Remove a channel from a pixel array.
///
///  \param channel channel to remove
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::remChannel (int channel)
{
  int     i, j, k;
  int     new_index, old_index;
  tlPixel *old;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::remChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::remChannel()", TL_ERROR);
    return -1;
  }

  // backup old pixel array
  old = tl_copy_array (pixel, width, height, pixelWidth);

  // change properties
  format = CUSTOM;
  pixelWidth -= 1;

  // reallocate pixel array
  pixel = (tlPixel *) realloc (pixel, width * height * pixelWidth);

  // shortcuts
  new_index = 0;
  old_index = 0;

  // insert old data back into new pixel array
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      for (k = 0; k < pixelWidth; k++) {
        if (k == channel) old_index++;
        pixel[new_index++] = old[old_index++];
      }
      if (channel >= pixelWidth) old_index++;
    }
  }


  // free
  free (old);
  return 0;
}



///  Add a channel to a pixel array at a given index.
///
///  \param pixelArray pixel array to insert in the pixel array
///  \param channel channel where pixel array is to be inserted
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::putChannel (tlPixel *pixelArray, 
                          int     channel)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }


  return tl_channel_put (pixelArray, width, height, pixelWidth, channel, this,  pixel);
}



///  Add a channel to part of a pixel array at a given index.
///
///  \param pixelArray pixel array to insert in the pixel array
///  \param channel channel where pixel array is to be inserted
///  \param rect portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::putChannel (tlPixel *pixelArray, 
                          tlRect  *rect, 
                          int     channel)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!pixelArray) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }


  return tl_channel_put (pixelArray, width, height, pixelWidth, channel, rect, pixel);
}



///  Add a channel to a pixel array at a given index.
///
///  \param pixelArray pixel array to insert
///  \param channel channel where pixel array is to be inserted
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The pixel array must contain a single channel.

int
tlPixelArray::putChannel (tlPixelArray *pixelArray, 
                          int          channel)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // check pixel width
  if (pixelArray->pixelWidth > 1) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }


  return tl_channel_put (pixelArray->pixel, width, height, pixelWidth, channel, (tlRect *) this, pixel);
}



///  Add a channel to part of a pixel array at a given index.
///
///  \param pixelArray pixel array to insert
///  \param channel channel where pixel array is to be inserted
///  \param rect portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The pixel array must contain a single channel.

int
tlPixelArray::putChannel (tlPixelArray *pixelArray, 
                          tlRect       *rect,
                          int          channel)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // check pixel width
  if (pixelArray->pixelWidth > 1) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }


  return tl_channel_put (pixelArray->pixel, width, height, pixelWidth, channel, rect, pixel);
}



///  Swap two channels in a pixel array.
///
///  \param channel1 first channel to swap
///  \param channel2 second channel to swap
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::swapChannels (int channel1, 
                            int channel2)
{
  int     res;
  tlPixel *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::swapChannels()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((channel1 >= pixelWidth) || (channel2 >= pixelWidth)) {
    TL_ERRNO_MSG ("tlPixelArray::swapChannels()", TL_ERROR);
    return -1;
  }

  // create copy of source pixel array
  tmp = tl_copy_array (pixel, width, height, pixelWidth);

  // swap
  res =	tl_channel_swap (tmp, width, height, pixelWidth, channel1, channel2, this, pixel);


  // free
  free (tmp);
  return res;
}



///  Remove a channel from a pixel array and store the result in a new pixel array.
///
///  \param channel channel to remove
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::remChannel (int          channel, 
                          tlPixelArray *dest)
{
  int	    i, j, k;
  int     new_index, old_index;
  int	    pixel_width;
  tlPixel *new_pixel;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::remChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::remChannel()", TL_ERROR);
    return -1;
  }

  // change properties
  dest->init (width, height, pixelWidth - 1);
  pixel_width = dest->pixelWidth;
  new_pixel = dest->pixel;

  // shortcuts
  new_index = 0;
  old_index = 0;

  // insert old data back into new pixel array
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      for (k = 0; k < pixel_width; k++) {
        if (k == channel) old_index++;
        new_pixel[new_index++] = pixel[old_index++];
      }
      if (channel >= pixel_width) old_index++;
    }
  }


  return 0;
}



///  Add a channel to a pixel array at a given index, and store 
///  the result in a new pixel array
///
///  \param pixelArray pixel array to insert in the pixel array
///  \param channel channel where pixel array is to be inserted
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::putChannel (tlPixel      *pixelArray, 
                          int          channel,
                          tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!pixelArray) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // init
  dest->init (this);
  return tl_channel_put (pixelArray, width, height, pixelWidth, channel, this, dest->pixel);
}



///  Add a channel to part of an pixel array at a given index, and
///  store the result in a new pixel array.
///
///  \param pixelArray pixel array to insert in the pixel array
///  \param channel channel where pixel array is to be inserted
///  \param rect portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::putChannel (tlPixel      *pixelArray, 
                          tlRect       *rect, 
                          int          channel,
                          tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!pixelArray) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // init
  dest->init (this);
  return tl_channel_put (pixelArray, width, height, pixelWidth, channel, rect, dest->pixel);
}



///  Add a channel to a pixel array at a given index and
///  store the result in a new pixel array
///
///  \param pixelArray pixel array to insert
///  \param channel channel where pixel array is to be inserted
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The pixel array must contain a single channel.

int
tlPixelArray::putChannel (tlPixelArray *pixelArray, 
                          int          channel,
                          tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // check pixel width
  if (pixelArray->pixelWidth > 1) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // init
  dest->init (this);
  return tl_channel_put (pixelArray->pixel, width, height, pixelWidth, channel, (tlRect *) this, pixel);
}



///  Add a channel to part of an pixel array at a given index and
///  store the result in a new pixel array
///
///  \param pixelArray pixel array to insert
///  \param channel channel where pixel array is to be inserted
///  \param rect portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \remark
///  The pixel array must contain a single channel.

int
tlPixelArray::putChannel (tlPixelArray *pixelArray, 
                          tlRect       *rect,
                          int          channel, 
                          tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::putChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check channel 
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // check pixel width
  if (pixelArray->pixelWidth > 1) {
    TL_ERRNO_MSG ("tlPixelArray::putChannel()", TL_ERROR);
    return -1;
  }

  // init
  dest->init (dest);
  return tl_channel_put (pixelArray->pixel, width, height, pixelWidth, channel, rect, pixel);
}



///  Swap two channels in a pixel array and 
///  store the result in a new pixel array.
///
///  \param channel1 first channel to swap
///  \param channel2 second channel to swap
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::swapChannels (int channel1, 
                            int channel2, 
                            tlPixelArray * dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG("tlPixelArray::swapChannels()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::swapChannels()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check
  if ((channel1 >= pixelWidth) || (channel2 >= pixelWidth)) {
    TL_ERRNO_MSG ("tlPixelArray::swapChannels()", TL_ERROR);
    return -1;
  }

  // init
  dest->init (this);

  // swap
  return tl_channel_swap (pixel, width, height, pixelWidth, channel1, channel2, this, dest->pixel);
}
