/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"
#include "tlColor.h"
#include "tlMask.h"
#include "tlPoint.h"



///  Retrieve the color of a given pixel.
///
///  \param point   point in the pixel array to retrieve
///  \param color   \b [out] color object with expected color and format
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColor (tlPoint *point, 
                        tlColor *color)
{
  return color->set (this, point);
}



///  Retrieve the color of a given pixel.
///
///  \param x   X coord of the pixel to retrieve
///  \param y   Y coord of the pixel to retrieve
///  \param color \b [out] color object with expected color and format
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColor (int     x, 
                        int     y, 
                        tlColor *color)
{
  return color->set (this, x, y);
}



///  Extract color maxima/minima from a pixel array.
///  The color objects are set to the same format as the input pixel array.
///
///  \param minColor  \b [out] lower color boundary
///  \param maxColor  \b [out] upper color boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColors (tlColor *minColor, 
                         tlColor *maxColor)
{
  return getColors (minColor, maxColor, this);
}



///  Extract color maxima/minima from a portion of a pixel array.
///  The color objects are set to the same format as the input pixel array.
///
///  \param rect      portion of the pixel array to process
///  \param minColor  \b [out] lower color boundary
///  \param maxColor  \b [out] upper color boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColors (tlColor *minColor, 
                         tlColor *maxColor,
                         tlRect  *rect)
{
  // check
  if ((!minColor) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::getColors()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // init colors
  minColor->format = format;
  maxColor->format = format;

  // extract accordingly
  switch (format) {
  case RGB:
    return tl_extract_color_rgb (pixel, width, height, minColor, maxColor, rect);
  case YUV:
    return tl_extract_color_yuv (pixel, width, height, minColor, maxColor, rect);
  case HSI:
    return tl_extract_color_hsi (pixel, width, height, minColor, maxColor, rect);
  case NRG:
    return tl_extract_color_nrg (pixel, width, height, minColor, maxColor, rect);
  case GRAY:
    return tl_extract_color_gray (pixel, width, height, minColor, maxColor, rect);
  case BIN:
  default:
    TL_ERRNO_MSG ("tlPixelArray::getColors()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Extract color minima from a pixel array.
///  The color object is set to the same format as the input pixel array.
///
///  \param minColor \b [out] lower color boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColorMin (tlColor *minColor)
{
  return getColorMin (minColor, this);
}



///  Extract color minima from a portion of a pixel array.
///  The color object is set to the same format as the input pixel array.
///
///  \param rect      portion of the pixel array to process
///  \param minColor  \b [out] lower color boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColorMin (tlColor *minColor, 
                           tlRect  *rect)
{
  // check
  if ((!minColor) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::getColorMin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // init colors
  minColor->format = format;

  // extract accordingly
  switch (format) {
  case RGB:
    return tl_extract_color_min_rgb (pixel, width, height, minColor, rect);
  case YUV:
    return tl_extract_color_min_yuv (pixel, width, height, minColor, rect);
  case HSI:
    return tl_extract_color_min_hsi (pixel, width, height, minColor, rect);
  case NRG:
    return tl_extract_color_min_nrg (pixel, width, height, minColor, rect);
  case GRAY:
    return tl_extract_color_min_gray (pixel, width, height, minColor, rect);
  case BIN:
  default:
    TL_ERRNO_MSG ("tlPixelArray::getColorMin()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Extract color maxima from a pixel array.
///  The color object is set to the same format as the input pixel array.
///
///  \param maxColor  \b [out] upper color boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColorMax (tlColor *maxColor)
{
  return getColorMax (maxColor, this);
}



///  Extract color maxima from a portion of a pixel array.
///  The color object is set to the same format as the input pixel array.
///
///  \param rect      portion of the pixel array to process
///  \param maxColor  \b [out] upper color boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getColorMax (tlColor *maxColor, 
                           tlRect  *rect)
{
  // check
  if ((!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::getColorMax()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // init colors
  maxColor->format = format;

  // extract accordingly
  switch (format) {
  case RGB:
    return tl_extract_color_max_rgb (pixel, width, height, maxColor, rect);
  case YUV:
    return tl_extract_color_max_yuv (pixel, width, height, maxColor, rect);
  case HSI:
    return tl_extract_color_max_hsi (pixel, width, height, maxColor, rect);
  case NRG:
    return tl_extract_color_max_nrg (pixel, width, height, maxColor, rect);
  case GRAY:
    return tl_extract_color_max_gray (pixel, width, height, maxColor, rect);
  case BIN:
  default:
    TL_ERRNO_MSG ("tlPixelArray::getColorMax()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}




///  Extract maxima from a portion of a pixel array.
///
///  \param mask      portion of the pixel array to process
///  \param channel   channel of the pixel array to process
///  \param point     location of maxima in image
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getMax (tlMask  *mask,
                      tlPoint *point,
                      int     channel)
{
  int     maxVal = TL_MIN_PIXEL_VAL;
  tlPixel *mpixel;
  int     mwidth;
  int     i,j;
  int     x1, y1, x2, y2;


  // check
  if (!mask || !point) {
    TL_ERRNO_MSG ("tlPixelArray::getMax()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  mpixel = mask->pixel;
  mwidth = mask->width;
  x1 = TL_MAX((mask->x),0); 
  x2 = TL_MIN(((mask->x)+(mwidth)), width); 
  y1 = TL_MAX((mask->y),0); 
  y2 = TL_MIN(((mask->y)+(mask->height)), height);

  // process
  if (!hr) {
    register tlPixel *pi;
    register tlPixel *pm;
    for (j=y1; j<y2; j++) {
      pi = pixel + (j*width+x1)*pixelWidth + channel;
      pm = mpixel + (j-y1)*mwidth;
      for (i=x1; i<x2; i++) {
        if (*(pm)) {
          if (*pi > maxVal) {
            maxVal = *pi;
            point->x = i;
            point->y = j;
          }
        }
        pm++;
        pi += pixelWidth;
      }
    }
  }
  else {
    register tlHRPixel *pi;
    register tlPixel   *pm;
    for (j=y1; j<y2; j++) {
      pi = hrpixel + (j*width+x1)*pixelWidth + channel;
      pm = mpixel  + (j-y1)*mwidth;
      for (i=x1; i<x2; i++) {
        if (*(pm)) {
          if (*pi > maxVal) {
            maxVal = *pi;
            point->x = i;
            point->y = j;
          }
        }
        pm++;
        pi += pixelWidth;
      }
    }
  }


  return maxVal;
}



///  Extract minima from a portion of a pixel array.
///
///  \param mask      portion of the pixel array to process
///  \param channel   channel of the pixel array to process
///  \param point     location of minima in image
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getMin (tlMask  *mask,
                      tlPoint *point,
                      int     channel)
{
  int     minVal = TL_MAX_PIXEL_VAL;
  tlPixel *mpixel;
  int     mwidth;
  int     i,j;
  int     x1, y1, x2, y2;


  // check
  if (!mask || !point) {
    TL_ERRNO_MSG ("tlPixelArray::getMin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // shortcuts
  mpixel = mask->pixel;
  mwidth = mask->width;
  x1 = TL_MAX((mask->x),0); 
  x2 = TL_MIN(((mask->x)+(mwidth)), width); 
  y1 = TL_MAX((mask->y),0); 
  y2 = TL_MIN(((mask->y)+(mask->height)), height);


  // process
  if (!hr) {
    register tlPixel *pi;
    register tlPixel *pm;
    for (j=y1; j<y2; j++) {
      pi = pixel  + (j*width+x1)*pixelWidth + channel;
      pm = mpixel + (j-y1)*mwidth;
      for (i=x1; i<x2; i++) {
        if (*(pm)) {
          if (*pi < minVal) {
            minVal = *pi;
            point->x = i;
            point->y = j;
          }
        }
        pm++;
        pi += pixelWidth;
      }
    }
  }
  else {
    register tlHRPixel *pi;
    register tlPixel   *pm;
    for (j=y1; j<y2; j++) {
      pi = hrpixel + (j*width+x1)*pixelWidth + channel;
      pm = mpixel  + (j-y1)*mwidth;
      for (i=x1; i<x2; i++) {
        if (*(pm)) {
          if (*pi < minVal) {
            minVal = *pi;
            point->x = i;
            point->y = j;
          }
        }
        pm++;
        pi += pixelWidth;
      }
    }
  }


  return minVal;
}
