/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"



///  Convolution with an integer square kernel.
///
///  \param kernel      convolution kernel (integer values)
///  \param kernel_size square kernel dimension
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char *kernel, 
                           int  kernel_size)
{
  int     res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  src = tl_copy_array (pixel, width, height, pixelWidth);
  res =	tl_convolution (src, width, height, pixelWidth, kernel, kernel_size, kernel_size, this, pixel);
  free (src);


  return res;
}



///  Convolution with an integer rectangular kernel.
///
///  \param kernel        convolution kernel (integer values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char *kernel, 
                           int  kernel_width,
                           int  kernel_height)
{
  int	    res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  src = tl_copy_array (pixel, width, height, pixelWidth);
  res =	tl_convolution (src, width, height, pixelWidth, kernel, kernel_width,	kernel_height, this, pixel);
  free (src);


  return res;
}



///  Convolution with a floating point square convolution kernel.
///
///  \param kernel      convolution kernel (floating-point values)
///  \param kernel_size square kernel dimension
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float *kernel, 
                           int   kernel_size)
{
  int	    res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  src = tl_copy_array (pixel, width, height, pixelWidth);
  res =	tl_convolution (src, width, height, pixelWidth, kernel, kernel_size, kernel_size, this, pixel);
  free (src);


  return res;
}



///  Convolution with a floating point rectangular convolution kernel.
///
///  \param kernel        convolution kernel (floating-point values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float *kernel, 
                           int   kernel_width,
                           int   kernel_height)
{
  int     res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  src = tl_copy_array (pixel, width, height, pixelWidth);
  res =	tl_convolution (src, width, height, pixelWidth, kernel, kernel_width,	kernel_height, this, pixel);
  free (src);


  return res;
}



///  Convolution of part of the pixel array with 
///  an integer square convolution kernel.
///
///  \param kernel      convolution kernel (integer values)
///  \param kernel_size square kernel dimension
///  \param rect        portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char   *kernel, 
                           int    kernel_size, 
                           tlRect *rect)
{
  int     res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  src = tl_copy_array (pixel, width, height, pixelWidth, rect);
  res =	tl_convolution (src, width, height, pixelWidth, kernel, kernel_size, kernel_size, rect, pixel);
  free (src);


  return res;
}



///  Convolution of part of the pixel array with 
///  an integer rectangular convolution kernel.
///
///  \param kernel        convolution kernel (integer values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///  \param rect          portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char   *kernel, 
                           int    kernel_width,
                           int    kernel_height, 
                           tlRect *rect)
{
  int	    res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  src = tl_copy_array (pixel, width, height, pixelWidth, rect);
  res = tl_convolution (src, width, height, pixelWidth, kernel, kernel_width,	kernel_height, rect, pixel);
  free (src);


  return res;
}



///  Convolution of part of the pixel array with 
///  a floating point square convolution kernel.
///
///  \param kernel      convolution kernel (floating-point values)
///  \param kernel_size square kernel dimension
///  \param rect        portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float  *kernel, 
                           int    kernel_size, 
                           tlRect *rect)
{
  int     res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  src = tl_copy_array (pixel, width, height, pixelWidth, rect);
  res =	tl_convolution (src, width, height, pixelWidth, kernel, kernel_size, kernel_size, rect, pixel);
  free (src);


  return res;
}



///  Convolution of part of the pixel array with 
///  a floating point rectangular convolution kernel.
///
///  \param kernel        convolution kernel (floating-point values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///  \param rect          portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float  *kernel, 
                           int    kernel_width,
                           int    kernel_height, 
                           tlRect *rect)
{
  int     res;
  tlPixel *src;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  src = tl_copy_array (pixel, width, height, pixelWidth, rect);
  res = tl_convolution (src, width, height, pixelWidth, kernel, kernel_width,	kernel_height, rect, pixel);
  free (src);


  return res;
}



///  Convolution with a separable integer kernel.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char *xkernel, 
                           int  xksize, 
                           char *ykernel,
                           int  yksize)
{
  int *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  tmp = (int *) malloc (width * height * pixelWidth * sizeof (int));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, this, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, this, pixel);
  free (tmp);


  return 0;
}



///  Convolution of part of the pixel array with a separable integer kernel.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///  \param rect      portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char   *xkernel, 
                           int    xksize, 
                           char   *ykernel,
                           int    yksize, 
                           tlRect *rect)
{
  int *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  tmp = (int *) malloc (width * height * pixelWidth * sizeof (int));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, rect, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, rect, pixel);
  free (tmp);


  return 0;
}



///  Convolution with a separable floating point kernel.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float *xkernel, 
                           int   xksize, 
                           float *ykernel,
                           int   yksize)
{
  float *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  tmp = (float *) malloc (width * height * pixelWidth * sizeof (float));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, this, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, this,	pixel);
  free (tmp);


  return 0;
}



///  Convolution of part of the pixel array with a 
///  separable floating point kernel.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///  \param rect      portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float  *xkernel, 
                           int    xksize, 
                           float  *ykernel,
                           int    yksize, 
                           tlRect *rect)
{
  float *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  tmp = (float *) malloc (width * height * pixelWidth * sizeof (float));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, rect, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, rect, pixel);
  free (tmp);


  return 0;
}



///  Convolution with an integer square kernel.
///  Put the result in another pixel array.
///
///  \param kernel      convolution kernel (integer values)
///  \param kernel_size square kernel dimension
///  \param dest        \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char         *kernel, 
                           int          kernel_size,
                           tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_size, kernel_size, this, dest->pixel);
}



///  Convolution with an integer rectangular kernel.
///  Put the result in another pixel array.
///
///  \param kernel        convolution kernel (integer values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///  \param dest          \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char         *kernel, 
                           int          kernel_width,
                           int          kernel_height, 
                           tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_width, kernel_height, this, dest->pixel);
}



///  Convolution with a floating point square convolution kernel.
///  Put the result in another pixel array.
///
///  \param kernel      convolution kernel (floating-point values)
///  \param kernel_size square kernel dimension
///  \param dest        \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float        *kernel, 
                           int          kernel_size,
                           tlPixelArray *dest)
{

  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_size, kernel_size, this, dest->pixel);
}



///  Convolution with a floating point rectangular convolution kernel.
///  Put the result in another pixel array.
///
///  \param kernel        convolution kernel (floating-point values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///  \param dest          \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float        *kernel, 
                           int          kernel_width,
                           int          kernel_height, 
                           tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_width, kernel_height, this, dest->pixel);
}



///  Convolution of part of the pixel array with 
///  an integer square convolution kernel.
///  Put the result in another pixel array.
///
///  \param kernel        convolution kernel (integer values)
///  \param kernel_size   square kernel dimension
///  \param rect          portion of the pixel array to process
///  \param dest          \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char         *kernel, 
                           int          kernel_size, 
                           tlRect       *rect,
                           tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_size, kernel_size, rect, dest->pixel);
}



///  Convolution of part of the pixel array with 
///  an integer rectangular convolution kernel.
///  Put the result in another pixel array.
///
///  \param kernel        convolution kernel (integer values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///  \param rect          portion of the pixel array to process
///  \param dest          \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char         *kernel, 
                           int          kernel_width,
                           int          kernel_height, 
                           tlRect       *rect,
                           tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_width, kernel_height, rect, dest->pixel);
}



///  Convolution of part of the pixel array with 
///  a floating point square convolution kernel.
///  Put the result in another pixel array.
///
///  \param kernel      convolution kernel (floating-point values)
///  \param kernel_size square kernel dimension
///  \param rect        portion of the pixel array to process
///  \param dest        \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float        *kernel, 
                           int          kernel_size, 
                           tlRect       *rect,
                           tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_size, kernel_size, rect, dest->pixel);
}



///  Convolution of part of the pixel array with 
///  a floating point rectangular convolution kernel.
///  Put the result in another pixel array.
///
///  \param kernel        convolution kernel (floating-point values)
///  \param kernel_width  convolution kernel width
///  \param kernel_height convolution kernel height
///  \param rect          portion of the pixel array to process
///  \param dest          \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float        *kernel, 
                           int          kernel_width,
                           int          kernel_height, 
                           tlRect       *rect,
                           tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  return tl_convolution (pixel, width, height, pixelWidth, kernel, kernel_width, kernel_height, rect, dest->pixel);
}



///  Convolution with a separable integer kernel.
///  Put the result in another pixel array.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///  \param dest      \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char         *xkernel, 
                           int          xksize, 
                           char         *ykernel,
                           int          yksize, 
                           tlPixelArray *dest)
{
  int *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  tmp = (int *) malloc (width * height * pixelWidth * sizeof (int));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, this, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, this, dest->pixel);
  free (tmp);


  return 0;
}



///  Convolution of part of the pixel array with a separable integer kernel.
///  Put the result in another pixel array.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///  \param rect      portion of the pixel array to process
///  \param dest      \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (char         *xkernel, 
                           int          xksize, 
                           char         *ykernel,
                           int          yksize, 
                           tlRect       *rect, 
                           tlPixelArray *dest)
{
  int *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!dest) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  tmp = (int *) malloc (width * height * pixelWidth * sizeof (int));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, rect, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, rect, dest->pixel);
  free (tmp);


  return 0;
}



///  Convolution with a separable floating point kernel.
///  Put the result in another pixel array.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///  \param dest      \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float        *xkernel, 
                           int          xksize, 
                           float        *ykernel,
                           int          yksize, 
                           tlPixelArray *dest)
{
  float *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  tmp = (float *) malloc (width * height * pixelWidth * sizeof (float));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, this, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, this, dest->pixel);
  free (tmp);


  return 0;
}



///  Convolution of part of the pixel array with a 
///  separable floating point kernel.
///  Put the result in another pixel array.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///  \param rect      portion of the pixel array to process
///  \param dest      \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::convolution (float        *xkernel, 
                           int          xksize, 
                           float        *ykernel,
                           int          yksize, 
                           tlRect       *rect, 
                           tlPixelArray *dest)
{
  float *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!dest) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  dest->init (format, width, height);
  tmp = (float *) malloc (width * height * pixelWidth * sizeof (float));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, rect, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, rect, dest->pixel);
  free (tmp);


  return 0;
}




///  Convolution of part of the pixel array with a 
///  separable floating point kernel.
///  Put the result in another pixel array.
///
///  \param xkernel   convolution kernel along the X axis
///  \param xksize    X convolution kernel size
///  \param ykernel   convolution kernel along the Y axis
///  \param yksize    Y convolution kernel size
///  \param rect      portion of the pixel array to process
///  \param mask      actual pixels to process
///  \param dest      \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlPixelArray::convolution (float        *xkernel,
                           int          xksize,
                           float        *ykernel, 
                           int          yksize, 
                           tlRect       *rect,
                           tlPixelArray *mask, 
                           tlPixelArray *dest)
{
  float *tmp;

  
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!dest) || (!rect) || (!mask)) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check if mask is valid
  if (mask->width != width || mask->height != height || mask->format != BIN) {
    TL_ERRNO_MSG ("tlPixelArray::convolution()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // process
  dest->init (format, width, height);
  tmp = (float *) malloc (width * height * pixelWidth * sizeof (float));
  tl_convolution_x (pixel, width, height, pixelWidth, xkernel, xksize, rect, mask->pixel, tmp);
  tl_convolution_y (tmp, width, height, pixelWidth, ykernel, yksize, rect, mask->pixel, dest->pixel);
  free (tmp);


  return 0;
}
