/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlPixelArray.h"



///  Add a pixel array to another.
///
///  \param pixelArray the pixel array to add to the current pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::add (tlPixelArray *pixelArray)
{
  return add (pixelArray, (tlRect *) this);
}



///  Add part of a pixel array to another.
///
///  \param pixelArray the pixel array to add to the current pixel array
///  \param rect portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::add (tlPixelArray *pixelArray, 
                   tlRect       *rect)
{
  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray->add()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (format != (pixelArray->format)) {
    TL_ERRNO_MSG ("tlPixelArray->add()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // check geometry
  if ((width != (pixelArray->width)) || (height != (pixelArray->height))) {
    TL_ERRNO_MSG ("tlPixelArray->add()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // process
  if (!hr) return tl_pixel_add   (pixel,   pixelArray->pixel,   width, height, pixelWidth, rect, pixel);
  else     return tl_pixel_hradd (hrpixel, pixelArray->hrpixel, width, height, pixelWidth, rect, hrpixel);
}



///  Add a pixel array to another; the result is put into a third pixel array.
///
///  \param pixelArray the pixel array to add to the current pixel array
///  \param dest \b [out] the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::add (tlPixelArray *pixelArray, 
                   tlPixelArray *dest)
{
  return add (pixelArray, this, dest);
}



///  Add part of a pixel array to another; the result is put into a third pixel array
///
///  \param pixelArray the pixel array to add to the current pixel array
///  \param rect portion of the pixel array to process
///  \param dest \b [out] the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::add (tlPixelArray *pixelArray, 
                   tlRect       *rect,
                   tlPixelArray *dest)
{
  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray->add()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (format != (pixelArray->format)) {
    TL_ERRNO_MSG ("tlPixelArray->add()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // check geometry
  if ((width != (pixelArray->width)) || (height != (pixelArray->height))) {
    TL_ERRNO_MSG ("tlPixelArray->add()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // init destination pixel array
  dest->init (this);

  // add pixel arrays
  if (!hr) return tl_pixel_add   (pixel,   pixelArray->pixel,   width, height, pixelWidth, rect, dest->pixel);
  else     return tl_pixel_hradd (hrpixel, pixelArray->hrpixel, width, height, pixelWidth, rect, dest->hrpixel);
}



///  Subtract a pixel array from another.
///
///  \param pixelArray the pixel array to subtract from the current pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::sub (tlPixelArray *pixelArray)
{
  return sub (pixelArray, (tlRect *) this);
}



///  Subtract part of a pixel array from another
///
///  \param pixelArray the pixel array to subtract from the current pixel array
///  \param rect portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::sub (tlPixelArray *pixelArray, 
                   tlRect       *rect)
{
  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray->sub()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (format != (pixelArray->format)) {
    TL_ERRNO_MSG ("tlPixelArray->sub()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // check geometry
  if ((width != (pixelArray->width)) || (height != (pixelArray->height))) {
    TL_ERRNO_MSG ("tlPixelArray->sub()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // process
  if (!hr) return tl_pixel_sub   (pixel,   pixelArray->pixel,   width, height, pixelWidth, rect, pixel);
  else     return tl_pixel_hrsub (hrpixel, pixelArray->hrpixel, width, height, pixelWidth, rect, hrpixel);
}



///  Subtract a pixel array from another;
///  the result is put into a third pixel array
///
///  \param pixelArray the pixel array to subtract from the current pixel array
///  \param dest \b [out] the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::sub (tlPixelArray *pixelArray, 
                   tlPixelArray *dest)
{
  return sub (pixelArray, this, dest);
}



///  Subtract part of a pixel array from another;
///  the result is put into a third pixel array
///
///  \param pixelArray the pixel array to subtract from the current pixel array
///  \param rect portion of the pixel array to process
///  \param dest \b [out] the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::sub (tlPixelArray *pixelArray, 
                   tlRect       *rect,
                   tlPixelArray *dest)
{
  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray->sub()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // format check
  if (format != (pixelArray->format)) {
    TL_ERRNO_MSG ("tlPixelArray::sub()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // geometry check
  if ((width != (pixelArray->width)) || (height != (pixelArray->height))) {
    TL_ERRNO_MSG ("tlPixelArray::sub()", TL_ERROR_GEOMETRY);
    return -1;
  }

  // init destination pixel array
  dest->init (format, width, height);

  // process
  if (!hr) return tl_pixel_sub   (pixel,   pixelArray->pixel,   width, height, pixelWidth, rect, dest->pixel);
  else     return tl_pixel_hrsub (hrpixel, pixelArray->hrpixel, width, height, pixelWidth, rect, dest->hrpixel);
}
