/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <stdlib.h>
#include <math.h>

#include "tlPixelArray.h"
#include "tlColor.h"



///  Filter a portion of a monochrome pixel array using arbitrary boundaries.
///
///  \param minVal lower pixel value boundary
///  \param maxVal upper pixel value boundary
///  \param rect portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (int    minVal, 
                      int    maxVal, 
                      tlRect *rect)
{
  int res;


  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (pixelWidth != 1) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // proceed
  switch (format) {
  case STEREO:
    res = tl_filter_hrgray (hrpixel, width, height, (tlHRPixel)minVal, (tlHRPixel)maxVal, rect, hrpixel);
    break;
  default:
    res = tl_filter_gray (pixel, width, height, (tlPixel)minVal, (tlPixel)maxVal, rect, pixel);
    break;
  }


  return res;
}



///  Filter a monochrome pixel array using arbitrary boundaries.
///
///  \param minVal lower pixel value boundary
///  \param maxVal upper pixel value boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (int minVal, 
                      int maxVal)
{
  return filter (minVal, maxVal, (tlRect *) this);
}



///  Filter a portion of a monochrome pixel array using arbitrary boundaries.
///  The result is put in another pixel array.
///
///  \param minVal lower pixel value boundary
///  \param maxVal upper pixel value boundary
///  \param rect portion of the pixel array to process
///  \param dest \b [out] the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (int          minVal, 
                      int          maxVal, 
                      tlRect       *rect,
                      tlPixelArray *dest)
{
  int res;


  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (pixelWidth != 1) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // init dest
  dest->init (this);

  // proceed
  switch (format) {
  case STEREO:
    res = tl_filter_hrgray (hrpixel, width, height, (tlHRPixel)minVal, (tlHRPixel)maxVal, rect, dest->hrpixel);
    break;
  default:
    res = tl_filter_gray (pixel, width, height, (tlPixel)minVal, (tlPixel)maxVal, rect, dest->pixel);
    break;
  }


  return res;
}



///  Filter a monochrome pixel array using arbitrary boundaries.
///  The result is put in another pixel array.
///
///  \param minVal lower pixel value boundary
///  \param maxVal upper pixel value boundary
///  \param dest \b [out] the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (int          minVal, 
                      int          maxVal, 
                      tlPixelArray *dest)
{
  return filter (minVal, maxVal, this, dest);
}



///  Filter a portion of a pixel array using color boundaries.
///
///  \param minColor lower pixel value boundary
///  \param maxColor upper pixel value boundary
///  \param rect portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (tlColor *minColor, 
                      tlColor *maxColor, 
                      tlRect  *rect)
{
  int	res;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }
  
  // check
  if ((!minColor) || (!maxColor) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (((minColor->format) != format) || ((maxColor->format) != format)) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // proceed
  switch (format) {
  case RGB:
    res = tl_filter_rgb (pixel, width, height, minColor->r, maxColor->r, minColor->g, maxColor->g, minColor->b, maxColor->b, rect,	pixel);
    break;
  case HSI:
    res = tl_filter_hsi (pixel, width, height, minColor->h1, maxColor->h1,	minColor->h2, maxColor->h2, minColor->s, maxColor->s,	minColor->i, maxColor->i, rect, pixel);
    break;
  case NRG:
    res = tl_filter_nrg (pixel, width, height, minColor->nr, maxColor->nr,	minColor->ng, maxColor->ng, rect, pixel);
    break;
  case GRAY:
    res = tl_filter_gray (pixel, width, height, minColor->intensity, maxColor->intensity, rect, pixel);
    break;
  case BIN:
  default:
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }


  return res;
}



///  Filter a pixel array using color boundaries.
///
///  \param minColor lower pixel value boundary
///  \param maxColor upper pixel value boundary
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (tlColor *minColor, 
                      tlColor *maxColor)
{
  return filter (minColor, maxColor, (tlRect *) this);
}



///  Filter a portion of a pixel array using color boundaries.
///  The result is put in another pixel array.
///
///  \param minColor lower pixel value boundary
///  \param maxColor upper pixel value boundary
///  \param rect portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (tlColor      *minColor, 
                      tlColor      *maxColor, 
                      tlRect       *rect,
                      tlPixelArray *dest)
{
  int res;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }
  
  // check
  if ((!minColor) || (!maxColor) || (!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (((minColor->format) != format) || ((maxColor->format) != format)) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // init dest
  dest->init (format, width, height);

  // proceed
  switch (format) {
  case RGB:
    res = tl_filter_rgb (pixel, width, height, minColor->r, maxColor->r, minColor->g, maxColor->g, minColor->b, maxColor->b, rect, dest->pixel);
    break;
  case HSI:
    res = tl_filter_hsi (pixel, width, height, minColor->h1, maxColor->h1, minColor->h2, maxColor->h2, minColor->s, maxColor->s, minColor->i, maxColor->i, rect, dest->pixel);
    break;
  case NRG:
    res = tl_filter_nrg (pixel, width, height, minColor->nr, maxColor->nr, minColor->ng, maxColor->ng, rect, dest->pixel);
    break;
  case GRAY:
    res = tl_filter_gray (pixel, width, height, minColor->intensity, maxColor->intensity, rect, dest->pixel);
    break;
  case BIN:
  default:
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }


  return res;
}



///  Filter a pixel array using color boundaries.
///  The result is put in another pixel array.
///
///  \param minColor lower pixel value boundary
///  \param maxColor upper pixel value boundary
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::filter (tlColor      *minColor, 
                      tlColor      *maxColor,
                      tlPixelArray *dest)
{
  return filter (minColor, maxColor, this, dest);
}



///  Median filter helper function.

int 
_tl_med_cmp (const void *a, 
             const void *b) { 
  if (*((tlPixel *)a) >  *((tlPixel *)b)) return  1;
	if (*((tlPixel *)a) == *((tlPixel *)b)) return  0;
	return -1;
}



///  Apply a median filter to the pixel array.
///
///  \param size    size of the median kernel
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.


int
tlPixelArray::median (int size)
{
  return median (size, this);
}



///  Apply a median filter to a portion of the pixel array.
///
///  \param size    size of the median kernel
///  \param rect    portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.


int
tlPixelArray::median (int    size,
                      tlRect *rect)
{
  int     i, j, k, m, n;
  int     kmin, kmax;
  int     x1, x2, y1, y2;
  int     sqsize = size*size;
  tlPixel *array = new tlPixel[sqsize];

  register tlPixel *in, *atmp;
  register tlPixel *newpix, *out;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::filter()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }
  
  // allocate new pixel array
  newpix = tl_create_array (width, height, pixelWidth);

	// shortcuts
  kmin  = size / 2;
	kmax  = size - kmin;
  in    = pixel  + kmin*pixelWidth + kmin*pixelWidth*width;
  out   = newpix + kmin*pixelWidth + kmin*pixelWidth*width;
  x1    = TL_MAX ((rect->x), kmin);
	x2    = TL_MIN (((rect->x) + (rect->width)), width - kmax);
	y1    = TL_MAX ((rect->y), kmin);
	y2    = TL_MIN (((rect->y) + (rect->height)), height - kmax);

	// proceed
	for (j = y1; j < y2; j++) {
    for (i = x1; i < x2; i++) {
      for (k=0; k<pixelWidth; k++) {
        atmp = in - kmin*pixelWidth - kmin*pixelWidth*width;
        for (m = 0; m < size; m++) {
          for (n = 0; n < size; n++) {
            // fill in array
            *(array++) = *(atmp);
            atmp += pixelWidth;
          }
          // next line
          atmp += pixelWidth*(width - size);
        }

        // retract array counter
        array -= sqsize;

        // sort line
        qsort (array, sqsize, sizeof(tlPixel), _tl_med_cmp);

        // store result
        *(out) = array[sqsize/2];

        // move on
        in   += 1;
        out  += 1;
      }
    }
    in  += size*pixelWidth;
    out += size*pixelWidth;
  }

  // clean up
  delete [] array;
  in    = pixel;
  pixel = newpix;
  free  (in);


  return 0;
}
