/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"
#include "tlObject.h"
#include "tlMask.h"



///  Find an object in a portion of an image 
///  given a minimum confidence threshold.
///  This call updates the \b object argument.
///
///  \param object \b [out] object to find
///  \param confidence min match threshold [%]
///  \param rect portion of the image to process
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::find (tlObject *object, 
                    int      confidence, 
                    tlRect   *rect,
                    int      increment)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::find()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check format
  if (format != object->format) {
    TL_ERRNO_MSG ("tlPixelArray::find()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // check geometry
  if ((object->width > width) || (object->height > height)) {
    TL_ERRNO_MSG ("tlPixelArray::find()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // call correlation function
  return tl_find_object (this, object, confidence, rect, increment);
}



///  Find an object in an image 
///  given a minimum confidence threshold.
///  This call updates the \b object argument.
///
///  \param object \b [out] object to find
///  \param confidence min match threshold [%]
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::find (tlObject *object, 
                    int      confidence, 
                    int      increment)
{
  return find (object, confidence, this, increment);
}



///  Find the best match for an object in a portion of the image.
///  This call updates the \b object argument.
///
///  \param object \b [out] object to find
///  \param rect portion of the image to process
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findBest (tlObject *object, 
                        tlRect   *rect, 
                        int      increment)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::findBest()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check format
  if (format != object->format) {
    TL_ERRNO_MSG ("tlPixelArray::findBest()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // check geometry
  if ((object->width > width) || (object->height > height)) {
    TL_ERRNO_MSG ("tlPixelArray::findBest()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // call correlation function
  return tl_find_best_object (this, object, rect, increment);
}



///  Find the best match for an object in the image.
///  This call updates the \b object argument.
///
///  \param object \b [out] object to find
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findBest (tlObject *object, 
                        int      increment)
{
  return findBest (object, this, increment);
}



///  Find an object shape in a portion of an image 
///  given a minimum confidence threshold.
///  This call updates the \b shape argument.
///
///  \param shape shape to find
///  \param confidence min match threshold [%]
///  \param rect portion of the image to process
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::find (tlMask *shape, 
                    int    confidence, 
                    tlRect *rect,
                    int    increment)
{
  // check format
  if (format != BIN) {
    TL_ERRNO_MSG ("tlPixelArray::find()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // check geometry
  if ((shape->width > width) || (shape->height > height)) {
    TL_ERRNO_MSG ("tlPixelArray::find()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // call correlation function
  return tl_find_shape (this, shape, confidence, rect, increment);
}



///  Find an object shape in an image 
///  given a minimum confidence threshold.
///  This call updates the \b shape argument.
///
///  \param shape shape to find
///  \param confidence min match threshold [%]
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::find (tlMask *shape, 
                    int    confidence, 
                    int    increment)
{
  return find (shape, confidence, this, increment);
}



///  Find the best match for an object shape in a portion of the image.
///  This call updates the \b shape argument.
///
///  \param shape shape to find
///  \param rect portion of the image to process
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findBest (tlMask *shape, 
                        tlRect *rect, 
                        int    increment)
{
  // check format
  if (format != BIN) {
    TL_ERRNO_MSG ("tlPixelArray::findBest()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // check geometry
  if ((shape->width > width) || (shape->height > height)) {
    TL_ERRNO_MSG ("tlPixelArray::findBest()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // call correlation function
  return tl_find_best_shape (this, shape, rect, increment);
}



///  Find the best match for an object shape in the image.
///  This call updates the \b shape argument.
///
///  \param shape shape to find
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findBest (tlMask *shape, 
                        int    increment)
{
  return findBest (shape, this, increment);
}



///  Find an object shape in an image 
///  given a minimum confidence threshold.
///  This call updates the \b shape argument.
///
///  \param object shape to find
///  \param confidence min match threshold [%]
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findShape (tlObject *object, 
                         int      confidence, 
                         int      increment)
{
  int res;


  res = find (object->shape, confidence, this, increment);
  object->setLocation (object->shape->x, object->shape->y);


  return res;
}



///  Find an object shape in a portion of an image 
///  given a minimum confidence threshold.
///  This call updates the \b shape argument.
///
///  \param object shape to find
///  \param confidence min match threshold [%]
///  \param rect portion of the image to process
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findShape (tlObject *object, 
                         int      confidence, 
                         tlRect   *rect,
                         int      increment)
{
  int res;


  res = find (object->shape, confidence, rect, increment);
  object->setLocation (object->shape->x, object->shape->y);


  return res;
}



///  Find the best match for an object shape in the image.
///  This call updates the \b shape argument.
///
///  \param object shape to find
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findBestShape (tlObject *object, 
                             int      increment)
{
  int res;


  res = findBest (object->shape, this, increment);
  object->setLocation (object->shape->x, object->shape->y);


  return res;
}



///  Find the best match for an object shape in a portion of the image.
///  This call updates the \b shape argument.
///
///  \param object shape to find
///  \param rect portion of the image to process
///  \param increment correlation increment
///
///  \return
///  An arbitraty confidence index in [%], -1 if not found.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::findBestShape (tlObject *object, 
                             tlRect   *rect,
                             int      increment)
{
  int res;


  res = findBest (object->shape, rect, increment);
  object->setLocation (object->shape->x, object->shape->y);


  return res;
}
