/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"



///  Extract one of the pixel array channel.
///
///  \param channel channel to extract
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toChannel (int channel)
{
  return toChannel (channel, (tlRect *) this);
}



///  Extract a portion of one of the pixel array channel.
///
///  \param channel channel to extract
///  \param rect    portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toChannel (int    channel, 
                         tlRect *rect)
{
  int     res;
  tlPixel *tmp;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::toChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  tmp = tl_copy_array (pixel, width, height, pixelWidth);
  res =	tl_to_channel (pixel, width, height, pixelWidth, channel, rect, pixel);

  // change attribs
  pixelWidth = 1;
  format     = GRAY;
  free (tmp);


  return res;
}



///  Extract one of the pixel array channel. The result is put in another pixel array.
///
///  \param channel channel to extract
///  \param dest    \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toChannel (int          channel, 
                         tlPixelArray *dest)
{
  return toChannel (channel, (tlRect *) this, dest);
}



///  Extract a portion of one of the pixel array channel.
///  The result is put in another pixel array.
///
///  \param channel channel to extract
///  \param rect    portion of the pixel array to process
///  \param dest    \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toChannel (int          channel, 
                         tlRect       *rect, 
                         tlPixelArray *dest)
{
  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::toChannel()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // check
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toChannel()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // init dest pixel array
  dest->init (GRAY, width, height);
  return tl_to_channel (pixel, width, height, pixelWidth, channel, rect, dest->pixel);
}



///  Convert the pixel array to RGB format.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toRgb ()
{
  return toRgb ((tlRect *) this);
}



///  Convert part of the pixel array to RGB format
///
///  \param rect portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toRgb (tlRect * rect)
{
  int	    res;
  tlPixel *tmp;


  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toRgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  tmp = tl_copy_array (pixel, width, height, pixelWidth, rect);
  switch (format) {
  case HSI:
    init (RGB, width, height);
    res = tl_hsi2rgb (tmp, width, height, rect, pixel);
    break;
  case YUV:
    init (RGB, width, height);
    res = tl_yuv2rgb (tmp, width, height, rect, pixel);
    break;
  case GRAY:
    init (RGB, width, height);
    res = tl_gray2rgb (tmp, width, height, rect, pixel);
    break;
  case RGB:
    TL_ERRNO_MSG ("tlPixelArray::toRgb()", TL_ERROR_NOT_REQUIRED);
    res = -1;
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::toRgb()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }
  free (tmp);


  return res;
}



///  Convert the pixel array to RGB format. The result is put in another pixel array.
///
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toRgb (tlPixelArray *dest)
{
  return toRgb ((tlRect *) this, dest);
}



///  Convert part of the pixel array to RGB format.
///  The result is put in another pixel array.
///
///  \param rect portion of the pixel array to convert
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toRgb (tlRect       *rect, 
                     tlPixelArray *dest)
{
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toRgb()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  dest->init (RGB, width, height);
  switch (format) {
  case RGB:
    return dest->copy (this);
  case HSI:
    return tl_hsi2rgb (pixel, width, height, rect, dest->pixel);
  default:
    TL_ERRNO_MSG ("tlPixelArray::toRgb()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Convert the pixel array to YUV format.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toYuv ()
{
  return toYuv ((tlRect *) this);
}



///  Convert part of the pixel array to YUV format
///
///  \param rect portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toYuv (tlRect * rect)
{
  int	    res;
  tlPixel *tmp;


  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toYuv()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  tmp = tl_copy_array (pixel, width, height, pixelWidth, rect);
  switch (format) {
  case RGB:
    init (YUV, width, height);
    res = tl_rgb2yuv (tmp, width, height, rect, pixel);
    break;
  case HSI:
    init (YUV, width, height);
    res = tl_hsi2yuv (tmp, width, height, rect, pixel);
    break;
  case YUV:
    TL_ERRNO_MSG ("tlPixelArray::toYuv()", TL_ERROR_NOT_REQUIRED);
    res = -1;
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::toYuv()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }
  free (tmp);


  return res;
}



///  Convert the pixel array to YUV format. The result is put in another pixel array.
///
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toYuv (tlPixelArray *dest)
{
  return toYuv ((tlRect *) this, dest);
}



///  Convert part of the pixel array to YUV format.
///  The result is put in another pixel array.
///
///  \param rect portion of the pixel array to convert
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toYuv (tlRect       *rect, 
                     tlPixelArray *dest)
{
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toYuv()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  dest->init (YUV, width, height);
  switch (format) {
  case YUV:
    return dest->copy (this);
  case RGB:
    return tl_rgb2yuv (pixel, width, height, rect, dest->pixel);
  case HSI:
    return tl_hsi2yuv (pixel, width, height, rect, dest->pixel);
  default:
    TL_ERRNO_MSG ("tlPixelArray::toYuv()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Convert the pixel array to HSI format.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toHsi ()
{
  return toHsi ((tlRect *) this);
}



///  Convert part of the pixel array to HSI format.
///
///  \param rect portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toHsi (tlRect *rect)
{
  int     res;
  tlPixel *tmp;


  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toHsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  tmp = tl_copy_array (pixel, width, height, pixelWidth, rect);
  switch (format) {
  case RGB:
    init (HSI, width, height);
    res = tl_rgb2hsi (tmp, width, height, rect, pixel);
    break;
  case YUV:
    init (HSI, width, height);
    res = tl_yuv2hsi (tmp, width, height, rect, pixel);
    break;
  case HSI:
    TL_ERRNO_MSG ("tlPixelArray::toHsi()", TL_ERROR_NOT_REQUIRED);
    res = -1;
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::toHsi()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }
  free (tmp);


  return res;
}



///  Convert the pixel array to HSI format.
///  The result is put in another pixel array.
///
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toHsi (tlPixelArray *dest)
{
  return toHsi ((tlRect *) this, dest);
}



///  Convert part of the pixel array to HSI format.
///  The result is put in another pixel array.
///
///  \param rect portion of the pixel array to convert
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toHsi (tlRect       *rect, 
                     tlPixelArray *dest)
{
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toHsi()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  dest->init (HSI, width, height);
  switch (format) {
  case HSI:
    return dest->copy (this);
  case RGB:
    return tl_rgb2hsi (pixel, width, height, rect, dest->pixel);
  case YUV:
    return tl_yuv2hsi (pixel, width, height, rect, dest->pixel);
  default:
    TL_ERRNO_MSG ("tlPixelArray::toHsi()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Convert the pixel array to NRG format.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toNrg ()
{
  return toNrg ((tlRect *) this);
}



///  Convert part of the pixel array to NRG format.
///
///  \param rect portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toNrg (tlRect * rect)
{
  int     res;
  tlPixel *tmp;


  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toNrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  tmp = tl_copy_array (pixel, width, height, pixelWidth, rect);
  switch (format) {
  case RGB:
    init (NRG, width, height);
    res = tl_rgb2nrg (tmp, width, height, rect, pixel);
    break;
  case YUV:
    init (NRG, width, height);
    res = tl_yuv2nrg (tmp, width, height, rect, pixel);
    break;
  case HSI:
    init (NRG, width, height);
    res = tl_hsi2nrg (tmp, width, height, rect, pixel);
    break;
  case NRG:
    TL_ERRNO_MSG ("tlPixelArray::toNrg()", TL_ERROR_NOT_REQUIRED);
    res = -1;
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::toNrg()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
  free (tmp);


  return res;
}



///  Convert the pixel array to NRG format.
///  The result is put in another pixel array.
///
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toNrg (tlPixelArray *dest)
{
  return toNrg ((tlRect *) this, dest);
}



///  Convert part of the pixel array to NRG format.
///  The result is put in another pixel array.
///
///  \param rect portion of the pixel array to convert
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toNrg (tlRect       *rect, 
                     tlPixelArray *dest)
{
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toNrg()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  dest->init (NRG, width, height);
  switch (format) {
  case RGB:
    return tl_rgb2nrg (pixel, width, height, rect, dest->pixel);
  case YUV:
    return tl_yuv2nrg (pixel, width, height, rect, dest->pixel);
  case HSI:
    return tl_hsi2nrg (pixel, width, height, rect, dest->pixel);
  case NRG:
    return dest->copy (this);
  default:
    TL_ERRNO_MSG ("tlPixelArray::toNrg()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Convert the pixel array to GRAY format.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to GRAY simply saves the luma (Y) channel
///  into the GRAY image - it is therefore NOT equivalent to the
///  RGB to GRAY conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toGray ()
{
  return toGray ((tlRect *) this);
}



///  Convert part of the pixel array to GRAY format.
///
///  \param rect portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to GRAY simply saves the luma (Y) channel
///  into the GRAY image - it is therefore NOT equivalent to the
///  RGB to GRAY conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toGray (tlRect *rect)
{
  int	      res;
  tlPixel   *tmp   = NULL;
  tlHRPixel *hrtmp = NULL;


  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toGray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  if (!hr) tmp   = tl_copy_array   (pixel,   width, height, pixelWidth, rect);
  else     hrtmp = tl_copy_hrarray (hrpixel, width, height, pixelWidth, rect);
  switch (format) {
  case RGB:
    init (GRAY, width, height);
    res = tl_rgb2gray (tmp, width, height, rect, pixel);
    break;
  case YUV:
    init (GRAY, width, height);
    res = tl_yuv2gray (tmp, width, height, rect, pixel);
    break;
  case HSI:
    init (GRAY, width, height);
    res = tl_hsi2gray (tmp, width, height, rect, pixel);
    break;
  case NRG:
    init (GRAY, width, height);
    res = tl_nrg2gray (tmp, width, height, rect, pixel);
    break;
  case GRAY:
    TL_ERRNO_MSG ("tlPixelArray::toGray()", TL_ERROR_NOT_REQUIRED);
    res = -1;
    break;
  case STEREO:
    init (GRAY, width, height);
    res = tl_hrgray2gray (hrtmp, width, height, rect, pixel, hrShift);
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::toGray()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
  if (tmp)   free (tmp);
  if (hrtmp) free (hrtmp);


  return res;
}



///  Convert the pixel array to GRAY format.
///  The result is put in another pixel array.
///
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to GRAY simply saves the luma (Y) channel
///  into the GRAY image - it is therefore NOT equivalent to the
///  RGB to GRAY conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toGray (tlPixelArray *dest)
{
  return toGray ((tlRect *) this, dest);
}



///  Convert part of the pixel array to GRAY format.
///  The result is put in another pixel array.
///
///  \param rect portion of the pixel array to convert
///  \param dest \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to GRAY simply saves the luma (Y) channel
///  into the GRAY image - it is therefore NOT equivalent to the
///  RGB to GRAY conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toGray (tlRect       *rect, 
                      tlPixelArray *dest)
{
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toGray()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  dest->init (GRAY, width, height);
  switch (format) {
  case RGB:
    return tl_rgb2gray (pixel, width, height, rect, dest->pixel);
  case YUV:
    return tl_yuv2gray (pixel, width, height, rect, dest->pixel);
  case HSI:
    return tl_hsi2gray (pixel, width, height, rect, dest->pixel);
  case NRG:
    return tl_nrg2gray (pixel, width, height, rect, dest->pixel);
  case GRAY:
    return dest->copy (this);
  case STEREO:
    return tl_hrgray2gray (hrpixel, width, height, rect, dest->pixel, hrShift);
  default:
    TL_ERRNO_MSG ("tlPixelArray::toGray()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Convert the pixel array to BINARY format.
///
///  \param threshold binary threshold [0...255]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int threshold)
{
  return toBin (threshold, (tlRect *) this);
}



///  Convert part of the pixel array to BINARY format.
///
///  \param threshold binary threshold [0...255]
///  \param rect      portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int    threshold, 
                     tlRect *rect)
{
  int	      res;
  tlPixel   *tmp   = NULL;
  tlHRPixel *hrtmp = NULL;


  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  if (!hr) tmp   = tl_copy_array   (pixel,   width, height, pixelWidth, rect);
  else     hrtmp = tl_copy_hrarray (hrpixel, width, height, pixelWidth, rect);
  switch (format) {
  case RGB:
    init (BIN, width, height);
    res = tl_rgb2bin (tmp, width, height, threshold, 255, rect, pixel);
    break;
  case YUV:
    init (BIN, width, height);
    res = tl_yuv2bin (tmp, width, height, threshold, 255, rect, pixel);
    break;
  case HSI:
    init (BIN, width, height);
    res = tl_hsi2bin (tmp, width, height, threshold, 255, rect, pixel);
    break;
  case NRG:
    init (BIN, width, height);
    res = tl_nrg2bin (tmp, width, height, threshold, 255, rect, pixel);
    break;
  case GRAY:
    init (BIN, width, height);
    res = tl_gray2bin (tmp, width, height, threshold, 255, rect, pixel);
    break;
  case STEREO:
    init (BIN, width, height);
    res = tl_hrgray2bin (hrtmp, width, height, threshold, 255, rect, pixel);
    break;
  case BIN:
    TL_ERRNO_MSG ("tlPixelArray::toBin()", 0);
    res = -1;
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }
  if (tmp)   free (tmp);
  if (hrtmp) free (hrtmp);


  return res;
}



///  Convert the pixel array to BINARY format.
///  The result is put in another pixel array.
///
///  \param min   binary lower threshold [0...255]
///  \param max   binary upper threshold [0...255]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int min, 
                     int max)
{
  return toBin (min, max, (tlRect *) this);
}



///  Convert part of the pixel array to BINARY format.
///  The result is put in another pixel array.
///
///  \param min   binary lower threshold [0...255]
///  \param max   binary upper threshold [0...255]
///  \param rect  portion of the pixel array to convert
///
///  \return
///  0 on success, -1 otherwise. 
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int    min, 
                     int    max, 
                     tlRect *rect)
{
  int	      res;
  tlPixel   *tmp   = NULL;
  tlHRPixel *hrtmp = NULL;


  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  if (!hr) tmp   = tl_copy_array   (pixel,   width, height, pixelWidth, rect);
  else     hrtmp = tl_copy_hrarray (hrpixel, width, height, pixelWidth, rect);
  switch (format) {
  case RGB:
    init (BIN, width, height);
    res = tl_rgb2bin (tmp, width, height, min, max, rect, pixel);
    break;
  case YUV:
    init (BIN, width, height);
    res = tl_yuv2bin (tmp, width, height, min, max, rect, pixel);
    break;
  case HSI:
    init (BIN, width, height);
    res = tl_hsi2bin (tmp, width, height, min, max, rect, pixel);
    break;
  case NRG:
    init (BIN, width, height);
    res = tl_nrg2bin (tmp, width, height, min, max, rect, pixel);
    break;
  case GRAY:
    init (BIN, width, height);
    res = tl_gray2bin (tmp, width, height, min, max, rect, pixel);
    break;
  case STEREO:
    init (BIN, width, height);
    res = tl_hrgray2bin (hrtmp, width, height, min, max, rect, pixel);
    break;
  case BIN:
    TL_ERRNO_MSG ("tlPixelArray::toBin()", 0);
    res = -1;
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }
  if (tmp)   free (tmp);
  if (hrtmp) free (hrtmp);


  return res;
}



///  Convert the pixel array to GRAY format.
///  The result is put in another pixel array.
///
///  \param threshold binary threshold [0...255]
///  \param dest      \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int          threshold, 
                     tlPixelArray *dest)
{
  return toBin (threshold, (tlRect *) this, dest);
}



///  Convert part of the pixel array to BINARY format.
///  The result is put in another pixel array.
///
///  \param threshold binary threshold [0...255]
///  \param rect      portion of the pixel array to process
///  \param dest      \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int          threshold, 
                     tlRect       *rect, 
                     tlPixelArray *dest)
{
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  if (this == dest)	return this->toBin (threshold, (tlRect *) rect);
  dest->init (BIN, width, height);
  switch (format) {
  case RGB:
    return tl_rgb2bin (pixel, width, height, threshold, 255, rect, dest->pixel);
  case YUV:
    return tl_yuv2bin (pixel, width, height, threshold, 255, rect, dest->pixel);
  case HSI:
    return tl_hsi2bin (pixel, width, height, threshold, 255, rect, dest->pixel);
  case NRG:
    return tl_nrg2bin (pixel, width, height, threshold, 255, rect, dest->pixel);
  case GRAY:
    return tl_gray2bin (pixel, width, height, threshold, 255, rect, dest->pixel);
  case STEREO:
    return tl_hrgray2bin (hrpixel, width, height, threshold, 255, rect, dest->pixel);
  case BIN:
    return dest->copy (this);
  default:
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}



///  Convert the pixel array to BINARY format.
///  The result is put in another pixel array.
///
///  \param min   binary lower threshold [0...255]
///  \param max   binary upper threshold [0...255]
///  \param dest  \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int          min, 
                     int          max, 
                     tlPixelArray *dest)
{
  return toBin (min, max, (tlRect *) this, dest);
}



///  Convert part of the pixel array to BINARY format.
///  The result is put in another pixel array.
///
///  \param min   binary lower threshold [0...255]
///  \param max   binary upper threshold [0...255]
///  \param rect  portion of the pixel array to process
///  \param dest  \b [out] convert to destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Conversion from YUV to BIN simply thresholds the luma (Y) channel
///  into the BIN image - it is therefore NOT equivalent to the
///  RGB to BIN conversion. See http://www.answers.com/topic/yuv-rgb-conversion-formulas
///  for details.

int
tlPixelArray::toBin (int          min, 
                     int          max, 
                     tlRect       *rect, 
                     tlPixelArray *dest)
{
  if ((!rect) || (!dest)) {
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  if (this == dest) return this->toBin (min, max, (tlRect *) rect);
  dest->init (BIN, width, height);
  switch (format) {
  case RGB:
    return tl_rgb2bin (pixel, width, height, min, max, rect, dest->pixel);
  case YUV:
    return tl_yuv2bin (pixel, width, height, min, max, rect, dest->pixel);
  case HSI:
    return tl_hsi2bin (pixel, width, height, min, max, rect, dest->pixel);
  case NRG:
    return tl_nrg2bin (pixel, width, height, min, max, rect, dest->pixel);
  case GRAY:
    return tl_gray2bin (pixel, width, height, min, max, rect, dest->pixel);
  case STEREO:
    return tl_hrgray2bin (hrpixel, width, height, min, max, rect, dest->pixel);
  case BIN:
    return dest->copy (this);
  default:
    TL_ERRNO_MSG ("tlPixelArray::toBin()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }
}
