/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"
#include "tlMask.h"



///  Hide all the pixels that are not in a given portion of the image
///  by setting their value to 0 on every channel. 
///
///  \param rect  the portion of the pixel array to keep 
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlPixelArray::mask (tlRect *rect) 
{ 
  int     i,j,k; 
  int     x1,x2,y1,y2;


  // check 
  if (!rect) { 
    TL_ERRNO_MSG ("tlPixelArray::mask()", TL_ERROR_NULL_POINTER); 
    return -1; 
  }

  // shortcuts 
  x1 = TL_MAX((rect->x),0); 
  x2 = TL_MIN(((rect->x)+(rect->width)), width); 
  y1 = TL_MAX((rect->y),0); 
  y2 = TL_MIN(((rect->y)+(rect->height)), height);

  // masking
  if (!hr) {
    register tlPixel *sptr;
    register tlPixel *dptr;
    tlPixel *tmp;
    tmp = tl_create_array (width, height, pixelWidth); 
    memset (tmp, 0, width*height*pixelWidth*sizeof(tlPixel));
    for (j=y1; j<y2; j++) { 
      sptr = pixel + pixelWidth*(j*width+x1); 
      dptr = tmp + pixelWidth*(j*width+x1); 
      for (i=x1; i<x2; i++) { 
        for (k=0; k<pixelWidth; k++) { 
          *dptr++ = *sptr++; 
        } 
      } 
    }
  }

  // masking high-resolution
  else {
    register tlHRPixel *sptr;
    register tlHRPixel *dptr;
    tlHRPixel *tmp;
    tmp = tl_create_hrarray (width, height, pixelWidth); 
    memset (tmp, 0, width*height*pixelWidth*sizeof(tlHRPixel));
    for (j=y1; j<y2; j++) { 
      sptr = hrpixel + pixelWidth*(j*width+x1); 
      dptr = tmp + pixelWidth*(j*width+x1); 
      for (i=x1; i<x2; i++) { 
        for (k=0; k<pixelWidth; k++) { 
          *dptr++ = *sptr++; 
        } 
      } 
    }
  }


  return 0; 
}



///  Hide all the pixels that are not in a given portion of the image
///  by setting their value to 0 on every channel. The result is placed 
///  in the dest pixel array. 
///
///  \param rect  the portion of the pixel array to keep
///  \param dest  the destination pixel array 
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int 
tlPixelArray::mask (tlRect       *rect, 
                    tlPixelArray *dest) 
{ 
  int     i,j,k;
  int     x1,x2,y1,y2; 


  // check 
  if (!rect) { 
    TL_ERRNO_MSG ("tlPixelArray::mask()", TL_ERROR_NULL_POINTER); 
    return -1; 
  }

  // create temporary copy 
  dest->init (this); 

  // shortcuts 
  x1  = TL_MAX((rect->x),0); 
  x2  = TL_MIN(((rect->x)+(rect->width)), width); 
  y1  = TL_MAX((rect->y),0); 
  y2  = TL_MIN(((rect->y)+(rect->height)), height); 

  // masking
  if (!hr) {
    register tlPixel *sptr; 
    register tlPixel *dptr; 
    tlPixel *tmp;
    tmp = dest->pixel;
    memset (tmp, 0, width*height*pixelWidth*sizeof(tlHRPixel));
    for (j=y1; j<y2; j++) { 
      sptr = pixel + pixelWidth*(j*width+x1); 
      dptr = tmp + pixelWidth*(j*width+x1); 
      for (i=x1; i<x2; i++) { 
        for (k=0; k<pixelWidth; k++) { 
          *dptr++ = *sptr++; 
        } 
      } 
    }
  }

  // masking high-resolution
  else {
    register tlHRPixel *sptr; 
    register tlHRPixel *dptr; 
    tlHRPixel *tmp;
    tmp = dest->hrpixel;
    memset (tmp, 0, width*height*pixelWidth*sizeof(tlHRPixel));
    for (j=y1; j<y2; j++) { 
      sptr = hrpixel + pixelWidth*(j*width+x1); 
      dptr = tmp + pixelWidth*(j*width+x1); 
      for (i=x1; i<x2; i++) { 
        for (k=0; k<pixelWidth; k++) { 
          *dptr++ = *sptr++; 
        } 
      } 
    }
  }


  return 0; 
}



///  Hide the pixel array behind a given mask (provided by a binary pixel array)
///  by setting all inactive pixels to value 0 on every channel.
///
///  \param pixelArray  the binary pixel array to apply as a mask to the pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Only the portion of the image that is overlapped by mask will be processed.

int
tlPixelArray::mask (tlPixelArray *pixelArray)
{
  return mask (pixelArray, (tlRect *) this);
}



///  Hide a portion of the pixel array behind a given mask (provided by a 
///  binary pixel array) by setting all inactive pixels to value 0 on every 
///  channel.
///
///  \param mask  the binary pixel array to apply as a mask to the pixel array
///  \param rect  the portion of the pixel array to mask
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Only the portion of the image that is overlapped by mask will be processed.

int
tlPixelArray::mask (tlPixelArray *mask, 
                    tlRect       *rect)
{
  register tlPixel *mptr;
  register int     sinc;
  register int     minc;

  int i, j, k;
  int x1, x2, y1, y2;
  int	mx1, mx2, my1, my2;
  int	Dxmr,	Dymr;
  int	mask_width;


  // check
  if ((!mask) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::mask()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (mask->pixelWidth != 1) {
    TL_ERRNO_MSG ("tlPixelArray::mask()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // compute mask image coordonates
  mx1 = TL_MAX ((mask->x), 0);
  mx2 = TL_MIN (((mask->x) + (mask->width)), width);
  my1 = TL_MAX ((mask->y), 0);
  my2 = TL_MIN (((mask->y) + (mask->height)), height);

  // compute mask/rect intersection
  x1 = TL_MAX ((rect->x), mx1);
  x2 = TL_MIN (((rect->x) + (rect->width)), mx2);
  y1 = TL_MAX ((rect->y), my1);
  y2 = TL_MIN (((rect->y) + (rect->height)), my2);

  // shortcuts
  mask_width = mask->width;
  Dymr       = TL_MAX (y1 - my1, 0);
  Dxmr       = TL_MAX (x1 - mx1, 0);
  sinc       = pixelWidth * (width - (x2 - x1));
  minc       = mask_width - (x2 - x1);

  // if the mask is hr
  tlPixelArray *tmpmask = NULL;
  if (mask->hr) {
    tmpmask = new tlPixelArray();
    mask->toBin (1, rect, tmpmask);
    mptr = tmpmask->pixel + TL_PIXEL ((Dymr), Dxmr, mask_width);
  }
  else {
    mptr = mask->pixel + TL_PIXEL ((Dymr), Dxmr, mask_width);
  }

  // masking
  if (!hr) {
    register tlPixel *sptr = pixel + pixelWidth * TL_PIXEL ((y1), x1, width);
    for (j = 0; j < (y2 - y1); j++) {
      for (i = x1; i < x2; i++) {
        if (*(mptr++) == 0) {
          for (k = 0; k < pixelWidth; k++) {
            *(sptr++) = 0;
          }
        }
        else sptr += pixelWidth;
      }
      sptr += sinc;
      mptr += minc;
    }
  }

  // masking high-resolution
  else {
    register tlHRPixel *sptr = hrpixel + pixelWidth * TL_PIXEL ((y1), x1, width);
    for (j = 0; j < (y2 - y1); j++) {
      for (i = x1; i < x2; i++) {
        if (*(mptr++) == 0) {
          for (k = 0; k < pixelWidth; k++) {
            *(sptr++) = 0;
          }
        }
        else sptr += pixelWidth;
      }
      sptr += sinc;
      mptr += minc;
    }
  }

  // if the mask was hr
  if (mask->hr) {
    delete tmpmask;
  }


  return 0;
}



///  Hide the pixel array behind a given mask (provided by a binary pixel array)
///  by setting all inactive pixels to value 0 on every channel.
///  The result is placed in the dest pixel array.
///
///  \param pixelArray  the binary pixel array to apply as a mask to the pixel array
///  \param dest        the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Only the portion of the image that is overlapped by mask will be processed.

int
tlPixelArray::mask (tlPixelArray *pixelArray, 
                    tlPixelArray *dest)
{
  return mask (pixelArray, (tlRect *) this, dest);
}



///  Hide a portion of the pixel array behind a given mask (provided by a 
///  binary pixel array) by setting all inactive pixels to value 0 on every 
///  channel. The result is placed in the dest pixel array.
///
///  \param mask  the binary pixel array to apply as a mask to the pixel array
///  \param rect  the portion of the pixel array to mask
///  \param dest  the destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Only the portion of the image that is overlapped by mask will be processed.

int
tlPixelArray::mask (tlPixelArray *mask, 
                    tlRect       *rect, 
                    tlPixelArray *dest)
{
  register tlPixel *mptr;
  register int	   sinc;
  register int     minc;

  int	i, j, k;
  int	x1,	x2,	y1,	y2;
  int	mx1, mx2,	my1, my2;
  int	mask_width;
  int	Dxmr,	Dymr;


  // check
  if ((!mask) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::mask()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // check format
  if (mask->pixelWidth != 1) {
    TL_ERRNO_MSG ("tlPixelArray::mask()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  // reset dest image
  dest->init (this);
  dest->x = x;
  dest->y = y;

  // compute mask image coordonates
  mx1 = TL_MAX ((mask->x), 0);
  mx2 = TL_MIN (((mask->x) + (mask->width)), width);
  my1 = TL_MAX ((mask->y), 0);
  my2 = TL_MIN (((mask->y) + (mask->height)), height);

  // compute mask/rect intersection
  x1 = TL_MAX ((rect->x), mx1);
  x2 = TL_MIN (((rect->x) + (rect->width)), mx2);
  y1 = TL_MAX ((rect->y), my1);
  y2 = TL_MIN (((rect->y) + (rect->height)), my2);

  // shortcuts
  mask_width = mask->width;
  Dymr       = TL_MAX (y1 - my1, 0);
  Dxmr       = TL_MAX (x1 - mx1, 0);
  sinc       = pixelWidth * (width - (x2 - x1));
  minc       = mask_width - (x2 - x1);

  // if the mask is hr
  tlPixelArray *tmpmask = NULL;
  if (mask->hr) {
    tmpmask = new tlPixelArray();
    mask->toBin (1, rect, tmpmask);
    mptr = tmpmask->pixel + TL_PIXEL ((Dymr), Dxmr, mask_width);
  }
  else {
    mptr = mask->pixel + TL_PIXEL ((Dymr), Dxmr, mask_width);
  }

  // masking
  if (!hr) {
    register tlPixel *sptr = pixel + pixelWidth * TL_PIXEL ((y1), x1, width);;
    register tlPixel *dptr = dest->pixel + pixelWidth * TL_PIXEL ((y1), x1, width);;
    for (j = 0; j < (y2 - y1); j++) {
      for (i = x1; i < x2; i++) {
        if (*(mptr++) == 0) {
          for (k = 0; k < pixelWidth; k++) {
            *(dptr++) = 0;
          }
          sptr += pixelWidth;
        }
        else {
          for (k = 0; k < pixelWidth; k++) {
            *(dptr++) = *(sptr++);
          }
        }
      }
      sptr += sinc;
      dptr += sinc;
      mptr += minc;
    }
  }

  // masking high-resolution
  else {
    register tlHRPixel *sptr = hrpixel + pixelWidth * TL_PIXEL ((y1), x1, width);
    register tlHRPixel *dptr = dest->hrpixel + pixelWidth * TL_PIXEL ((y1), x1, width);
    for (j = 0; j < (y2 - y1); j++) {
      for (i = x1; i < x2; i++) {
        if (*(mptr++) == 0) {
          for (k = 0; k < pixelWidth; k++) {
            *(dptr++) = 0;
          }
          sptr += pixelWidth;
        }
        else {
          for (k = 0; k < pixelWidth; k++) {
            *(dptr++) = *(sptr++);
          }
        }
      }
      sptr += sinc;
      dptr += sinc;
      mptr += minc;
    }
  }

  if (mask->hr) {
    delete tmpmask;
  }


  return 0;
}
