/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <memory.h>

#include "tlPixelArray.h"
#include "tlPoint.h"



///  Image initialization copies attributes from another pixel array.
///  The pixel array, however, is allocated but not copied.
///
///  \param pixelArray reference pixel array
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::init (tlPixelArray *pixelArray)
{
  // make sure we really need to do it
  if ((pixelArray->format == format) && (pixelArray->width == width) && (pixelArray->height == height)) return 0;

  // geometry
  this->format     = pixelArray->format;
  this->hr         = pixelArray->hr;
  this->hrShift    = pixelArray->hrShift;
  this->width      = pixelArray->width;
  this->height     = pixelArray->height;
  this->pixelWidth = pixelArray->pixelWidth;
  this->timeStamp  = 0;


  // pixel reallocation
  if (!hr) {
    pixel = (tlPixel *) realloc (pixel, width * height * pixelWidth * sizeof (tlPixel));
    if (pixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::init()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }
  else {
    hrpixel = (tlHRPixel *) realloc (hrpixel, width * height * pixelWidth * sizeof (tlHRPixel));
    if (hrpixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::init()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }


  return 0;
}



///  Image initialization reset the pixel array format and size. The pixel
///  array is reallocated accordingly.
///
///  \param format pixel array format
///  \param width pixel array width [pix]
///  \param height pixel array height [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::init (tl_format format, 
                    int       width, 
                    int       height)
{
  // make sure we really need to do it
  if ((this->format == format) && (this->width == width) && (this->height == height)) return 0;

  // decide on resolution
  switch (format) {
  case STEREO:
    hr = true;
    hrShift = 0;
    break;
  default:
    hr = false;
  }

  // geometry
  this->format     = format;
  this->width      = width;
  this->height     = height;
  this->pixelWidth = tl_pixel_width (format);
  this->timeStamp  = 0;

  // pixel reallocation
  if (!hr) {
    pixel = (tlPixel *) realloc (pixel, width * height * pixelWidth * sizeof (tlPixel));
    if (pixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::init()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }
  else {
    hrpixel = (tlHRPixel *) realloc (hrpixel, width * height * pixelWidth * sizeof (tlHRPixel));
    if (hrpixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::init()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }


  return 0;
}



///  Image initialization reset the pixel array format and size. The pixel
///  array is reallocated accordingly.
///
///  \param width pixel array width [pix]
///  \param height pixel array height [pix]
///  \param pixelWidth pixel width [bytes]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::init (int width, 
                    int height, 
                    int pixelWidth)
{
  // this is going to be a special one
  this->format     = CUSTOM;

  // make sure we really need to do it
  if ((this->pixelWidth == pixelWidth) && (this->width == width) && (this->height == height)) return 0;

  // geometry
  this->hr         = false;
  this->width      = width;
  this->height     = height;
  this->pixelWidth = pixelWidth;
  this->timeStamp  = 0;

  // pixel reallocation
  if (!hr) {
    pixel = (tlPixel *) realloc (pixel, width * height * pixelWidth * sizeof (tlPixel));
    if (pixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::init()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }
  else {
    hrpixel = (tlHRPixel *) realloc (hrpixel, width * height * pixelWidth * sizeof (tlHRPixel));
    if (hrpixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::init()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
  }


  return 0;
}



///  Deallocate the pixel array pixel array.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::deletePixel (void)
{
  if (pixel) {
    free (pixel);
    pixel = NULL;
  }
  if (hrpixel) {
    free (hrpixel);
    hrpixel = NULL;
  }


  return 0;
}



///  Perform pixel array creation and allocation.
///
///  \param format  pixel array format
///  \param width   pixel array width [pix]
///  \param height  pixel array height [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::create (tl_format format, 
                      int       width, 
                      int       height)
{
  // geometry
  this->format     = format;
  this->width      = width;
  this->height     = height;
  this->pixelWidth = tl_pixel_width (format);
  this->timeStamp  = 0;

  // pixel allocation
  switch (format) {
  case STEREO:
    hr = true;
    hrShift = 0;
    hrpixel = (tlHRPixel *) realloc (hrpixel, width * height * pixelWidth * sizeof (tlHRPixel));
    if (hrpixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::create()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
    break;
  default:
    hr = false;
    pixel = (tlPixel *) realloc (pixel, width * height * pixelWidth * sizeof (tlPixel));
    if (pixel == NULL) {
      TL_ERRNO_MSG ("tlPixelArray::create()", TL_ERROR_ALLOCATION_FAILED);
      return -1;
    }
    break;
  }

  return 0;
}



///  Set all pixels in the pixel array to black.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::clear (void)
{
  if (!hr) memset (pixel,   0, width * height * pixelWidth * sizeof (tlPixel));
  else     memset (hrpixel, 0, width * height * pixelWidth * sizeof (tlHRPixel));
  
  
  return 0;
}



///  Set all pixels in a portion of the pixel array to black.
///
///  \param rect portion of the pixel array to clear
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::clear (tlRect *rect)
{
  int	j;
  int	length,	inc;
  int	x1,	x2,	y1,	y2;


  // shortcurs
  x1     = TL_MAX (0, (rect->x));
  x2     = TL_MIN (width, ((rect->x) + (rect->width)));
  y1     = TL_MAX (0, (rect->y));
  y2     = TL_MIN (height, ((rect->y) + (rect->height)));
  inc    = width * pixelWidth;

  // erase
  if (!hr) {
    register tlPixel *p;
    p      = pixel + (y1 * width + x1) * pixelWidth;
    length = (x2 - x1) * pixelWidth * sizeof (tlPixel);
    for (j = y1; j < y2; j++) {
      memset (p, 0, length);
      p += inc;
    }
  }
  else {
    register tlHRPixel *p;
    p      = hrpixel + (y1 * width + x1) * pixelWidth;
    length = (x2 - x1) * pixelWidth * sizeof (tlHRPixel);
    for (j = y1; j < y2; j++) {
      memset (p, 0, length);
      p += inc;
    }
  }


  return 0;
}



///  Set the pixel value at the given position,
///  and for the given channel.
///
///  \param val        new pixel value
///  \param x          X position in pixel array
///  \param y          Y position in pixel array
///  \param channel    channel index
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::setPixel (int val, 
                        int x, 
                        int y, 
                        int channel)
{
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::setPixel()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  if (!hr) return pixel  [pixelWidth * TL_PIXEL (y, x, width) + channel] = (tlPixel)val;
  else     return hrpixel[pixelWidth * TL_PIXEL (y, x, width) + channel] = (tlHRPixel)val;
}



///  Set the pixel value at the given position,
///  and for the given channel.
///
///  \param val        new pixel value
///  \param point      pixel position in pixel array
///  \param channel    channel index
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::setPixel (int     val, 
                        tlPoint *point, 
                        int     channel)
{
  if (channel >= pixelWidth) {
    TL_ERRNO_MSG ("tlPixelArray::setPixel()", TL_ERROR_INVALID_FORMAT);
    return -1;
  }

  if (!hr) return pixel  [pixelWidth * TL_PIXEL (point->y, point->x, width) + channel] = (tlPixel)val;
  else     return hrpixel[pixelWidth * TL_PIXEL (point->y, point->x, width) + channel] = (tlHRPixel)val;
}



///  Return the pixel value at the given position.
///
///  \param x position in pixel array
///  \param y position in pixel array
///  \param channel specific channel
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getPixel (int x, 
                        int y, 
                        int channel)
{
  if (!hr) return (int)(pixel  [pixelWidth * TL_PIXEL (y, x, width) + channel]);
  else     return (int)(hrpixel[pixelWidth * TL_PIXEL (y, x, width) + channel]);
}



///  Return the pixel value at a given position for a given channel.
///
///  \param point pixel location
///  \param channel specific channel (default is 0)
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::getPixel (tlPoint *point, 
                        int     channel)
{
  if (!hr) return (int)(pixel  [pixelWidth * TL_PIXEL (point->y, point->x, width) + channel]);
  else     return (int)(hrpixel[pixelWidth * TL_PIXEL (point->y, point->x, width) + channel]);
}



///  Copy the pixel array content to another pixel array.
///
///  \param dest \b [out] the pixel array to copy content to
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::copyTo (tlPixelArray *dest)
{
  // check
  if (!dest) {
    TL_ERRNO_MSG ("tlPixelArray::copyTo()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  return dest->copy (this);
}



///  Copy the pixel array content to another pixel array.
///
///  \param rect \b part of the pixel array to copy
///  \param dest \b [out] the pixel array to copy content to
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::copyTo (tlPixelArray *dest, 
                      tlRect       *rect)
{
  // check
  if ((!dest) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::copyTo()", TL_ERROR_NULL_POINTER);
    return -1;
  }


  return dest->copy (this, rect);
}



///  Copy the pixel array content from another pixel array.
///
///  \param pixelArray the pixel array to copy from
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::copy (tlPixelArray *pixelArray)
{
  // check
  if (!pixelArray) {
    TL_ERRNO_MSG ("tlPixelArray::Copy()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // initialize
  deletePixel ();
  format     = pixelArray->format;
  hr         = pixelArray->hr;
  width      = pixelArray->width;
  height     = pixelArray->height;
  pixelWidth = pixelArray->pixelWidth;
  timeStamp  = pixelArray->timeStamp;

  // copy pixel array
  if (!hr) {
    pixel = tl_copy_array (pixelArray->pixel, width, height, pixelWidth);
    if (!pixel) {
      TL_ERRNO_MSG ("tlPixelArray::copy()", TL_ERROR_NULL_POINTER);
      return -1;
    }
  }
  else {
    hrpixel = tl_copy_hrarray (pixelArray->hrpixel, width, height, pixelWidth);
    if (!hrpixel) {
      TL_ERRNO_MSG ("tlPixelArray::copy()", TL_ERROR_NULL_POINTER);
      return -1;
    }
  }


  return 0;
}



///  Copy the pixel array content from another pixel array.
///
///  \param pixelArray the pixel array to copy from
///  \param rect portion of the pixel array to copy from
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::copy (tlPixelArray *pixelArray, 
                    tlRect       *rect)
{
  int	    j;
  int	    x1, x2, y1, y2;


  // check
  if ((!pixelArray) || (!rect)) {
    TL_ERRNO_MSG ("tlPixelArray::Copy()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // initialize
  init (pixelArray);
  timeStamp = pixelArray->timeStamp;

  // shortcuts
  x1     = TL_MAX (0, (rect->x));
  x2     = TL_MIN (((rect->x) + (rect->width)), width);
  y1     = TL_MAX (0, (rect->y));
  y2     = TL_MIN (((rect->y) + (rect->height)), height);

  // copy pixel array
  if (!hr) {
    tlPixel *src   = pixelArray->pixel;
    int     rwidth = (x2 - x1) * pixelWidth * sizeof (tlPixel);
    int     iwidth = width * pixelWidth;
    int     index  = y1 * iwidth + x1 * pixelWidth;
    for (j = y1; j < y2; j++) {
      memcpy (&(pixel[index]), &(src[index]), rwidth);
      index += iwidth;
    }
  }
  else {
    tlHRPixel *src   = pixelArray->hrpixel;
    int       rwidth = (x2 - x1) * pixelWidth * sizeof (tlHRPixel);
    int       iwidth = width * pixelWidth;
    int       index  = y1 * iwidth + x1 * pixelWidth;
    for (j = y1; j < y2; j++) {
      memcpy (&(hrpixel[index]), &(src[index]), rwidth);
      index += iwidth;
    }
  }


  return 0;
}



///  Reset a tlRect object with the pixel array dimensions.
/// 
///  \param rect \b [out] the bounding rectangle
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::toRect (tlRect *rect)
{
  // check 
  if (!rect) {
    TL_ERRNO_MSG ("tlPixelArray::toRect()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // proceed
  rect->x      = x;
  rect->y      = y;
  rect->width  = width;
  rect->height = height;


  return 0;
}



///  Create a tlRect object that contains the pixel array.
///
///  \return
///  A pointer to a tlRect object, NULL on error.

tlRect *
tlPixelArray::toRect ()
{
  tlRect *rect = new tlRect (x, y, width, height);
  return rect;
}
