/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"



///  Perform morphological erosion on the pixel array using a square mask.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::erode (int    size, 
                     tlRect *rect)
{
  int       res;
  tlPixel   *tmp;


  // check
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::erode()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  tmp   = tl_copy_array   (pixel,   width, height, pixelWidth);
  switch (format) {
  case GRAY:
    res = tl_erode_gray (tmp, width, height, size, rect, pixel);
    break;
  case BIN:
    res = tl_erode_bin (tmp, width, height, size, rect, pixel);
    break;
  default:
    res = tl_erode_all (tmp, width, height, pixelWidth, size, rect, pixel);
  }
  free (tmp);
  

  return res;
}



/// \overload

int
tlPixelArray::erode (int size)
{
  return erode (size, (tlRect *) this);
}



///  Perform morphological dilatation on the pixel array using a square mask.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::dilate (int    size, 
                      tlRect *rect)
{
  int       res;
  tlPixel   *tmp;


  // check
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::erode()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  tmp   = tl_copy_array   (pixel,   width, height, pixelWidth);
  switch (format) {
  case GRAY:
    res = tl_dilate_gray (tmp, width, height, size, rect, pixel);
    break;
  case BIN:
    res = tl_dilate_bin (tmp, width, height, size, rect, pixel);
    break;
  default:
    res = tl_dilate_all (tmp, width, height, pixelWidth, size, rect, pixel);
  }
  free (tmp);


  return res;
}



/// \overload

int
tlPixelArray::dilate (int size)
{
  return dilate (size, (tlRect *) this);
}



///  Perform morphological opening on the pixel array using a square mask.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::opening (int    size, 
                       tlRect *rect)
{
  int	res;

  if ((res = dilate (size, rect)) != 0) return res;
  return erode (size, rect);
}



/// \overload

int
tlPixelArray::opening (int size)
{
  int	res;

  if ((res = dilate (size, (tlRect *) this)) != 0) return res;
  return erode (size, (tlRect *) this);
}



///  Perform morphological closing on the pixel array using a square mask.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::closing (int    size, 
                       tlRect *rect)
{
  int	res;

  if ((res = erode (size, rect)) != 0) return res;
  return dilate (size, rect);
}



/// \overload

int
tlPixelArray::closing (int size)
{
  int	res;

  if ((res = erode (size, (tlRect *) this)) != 0) return res;
  return dilate (size, (tlRect *) this);
}



///  Perform morphological erosion on the pixel array using a square mask; the
///  result is put in another pixel array.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::erode (int          size, 
                     tlRect       *rect, 
                     tlPixelArray *dest)
{
  // check
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::erode()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  dest->init (this);
  switch (format) {
  case GRAY:
    return tl_erode_gray (pixel, width, height, size, rect, dest->pixel);
  case BIN:
    return tl_erode_bin (pixel, width, height, size, rect, dest->pixel);
  default:
    return tl_erode_all (pixel, width, height, pixelWidth, size, rect,
      dest->pixel);
  }
}



/// \overload

int
tlPixelArray::erode (int          size, 
                     tlPixelArray *dest)
{
  return erode (size, (tlRect *) this, dest);
}



///  Perform morphological dilatation on the pixel array using a square mask; the
///  result is put in another pixel array.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::dilate (int          size, 
                      tlRect       *rect, 
                      tlPixelArray *dest)
{
  // check
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::erode()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  dest->init (this);
  switch (format) {
  case GRAY:
    return tl_dilate_gray (pixel, width, height, size, rect, dest->pixel);
  case BIN:
    return tl_dilate_bin (pixel, width, height, size, rect, dest->pixel);
  default:
    return tl_dilate_all (pixel, width, height, pixelWidth, size, rect, dest->pixel);
  }
}



/// \overload

int
tlPixelArray::dilate (int          size, 
                      tlPixelArray *dest)
{
  return dilate (size, (tlRect *) this, dest);
}



///  Perform morphological opening on the pixel array using a square mask; the
///  result is put in another pixel array.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::opening (int          size, 
                       tlRect       *rect, 
                       tlPixelArray *dest)
{
  int	res;

  if ((res = erode (size, rect, dest)) != 0) return res;
  return dest->dilate (size, rect);
}



/// \overload

int
tlPixelArray::opening (int          size, 
                       tlPixelArray *dest)
{
  int	res;

  if ((res = erode (size, (tlRect *) this, dest)) != TL_NO_ERROR) return res;
  return dest->dilate (size, (tlRect *) this);
}



///  Perform morphological closing on the pixel array using a square mask; the
///  result is put in another pixel array.
///  The operation is performed on every pixel array channel independantly. 
///
///  \param size the square mask size [pix]
///  \param rect the portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::closing (int          size, 
                       tlRect       *rect, 
                       tlPixelArray *dest)
{
  int	res;

  if ((res = dilate (size, rect, dest)) != 0) return res;
  return dest->erode (size, rect);
}



///  \overload

int
tlPixelArray::closing (int          size, 
                       tlPixelArray *dest)
{
  int	res;

  if ((res = dilate (size, (tlRect *) this, dest)) != 0) return res;
  return dest->erode (size, (tlRect *) this);
}



///  Perform morphological (hit and miss) thinning on the pixel array using a square mask.
///  The operation is performed on BIN type only. 
///
///  \param rect the portion of the pixel array to process
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::thin (tlRect *rect)
{
  tlPixel *tmp;
  int     res;


  tmp = tl_copy_array (pixel, width, height, pixelWidth);
  switch (format) {
  case BIN:
    res = tl_thin (tmp, width, height, rect, pixel);
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::thin()", TL_ERROR_INVALID_FORMAT);
    res = -1;
  }

  free (tmp);


  return res;
}



// \overload

int
tlPixelArray::thin ()
{
  return thin ((tlRect *) this);
}



///  Perform morphological (hit and miss) thinning on the pixel array using a square mask; 
///  the result is put in another pixel array.
///  The operation is performed on BIN type only. 
///
///  \param rect the portion of the pixel array to process
///  \param dest \b [out] destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::thin (tlRect       *rect, 
                    tlPixelArray *dest)
{
  dest->init (format, width, height);
  switch (format) {
  case BIN:
    return tl_thin (pixel, width, height, rect, dest->pixel);
  default:
    TL_ERRNO_MSG ("tlPixelArray::thin()", TL_ERROR_INVALID_FORMAT);
  }


  return -1;
}



/// \overload

int
tlPixelArray::thin (tlPixelArray *dest)
{
  return thin ((tlRect *) this, dest);
}
