/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlPixelArray.h"



///  Scale the pixel array.
///
///  \param factor the fp scaling factor
///  \param method scaling method
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::scale (float factor, 
                     int   method)
{
  tlPixel *output;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // get scaled pixel array
  switch (method) {
  case TL_SCALE_DECIMATE:
    output = tl_scale_decimate (pixel, factor, width, height, pixelWidth);
    break;
  case TL_SCALE_BSPLINE:
    tl_scale_bspline_init (pixel, width, height, pixelWidth);
    output = tl_scale_bspline (factor);
    tl_scale_bspline_cleanup ();
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR);
    return -1;
  }

  // replace current
  deletePixel ();
  width = (int) (factor * width);
  height = (int) (factor * height);
  pixel = output;


  return 0;
}



///  Copy a scaled version of the pixel array to another pixel array.
///
///  \param factor the fp scaling factor
///  \param method scaling method
///  \param dest destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::scale (float        factor, 
                     tlPixelArray *dest, 
                     int          method)
{
  tlPixel *output;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // get scaled pixel array
  switch (method) {
  case TL_SCALE_DECIMATE:
    output = tl_scale_decimate (pixel, factor, width, height, pixelWidth);
    break;
  case TL_SCALE_BSPLINE:
    tl_scale_bspline_init (pixel, width, height, pixelWidth);
    output = tl_scale_bspline (factor);
    tl_scale_bspline_cleanup ();
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR);
    return -1;
  }

  // init destination pixel array
  dest->init (format, (int) (factor * width), (int) (factor * height));

  // replace current
  dest->deletePixel ();
  dest->pixel = output;


  return 0;
}



///  Scale the pixel array.
///
///  \param newWidth desired width of the scaled image
///  \param newHeight desired height of the scaled image
///  \param method scaling method
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::scale (int newWidth, 
                     int newHeight, 
                     int method)
{
  tlPixel *output;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // get scaled pixel array
  switch (method) {
  case TL_SCALE_DECIMATE:
    output = tl_scale_decimate (pixel, newWidth, newHeight, width, height, pixelWidth);
    break;
  case TL_SCALE_BSPLINE:
    tl_scale_bspline_init (pixel, width, height, pixelWidth);
    output = tl_scale_bspline (newWidth, newHeight);
    tl_scale_bspline_cleanup ();
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR);
    return -1;
  }

  // replace current
  deletePixel ();
  width = newWidth;
  height = newHeight;
  pixel = output;


  return 0;
}



///  Copy a scaled version of the pixel array to another pixel array.
///
///  \param newWidth desired width of the scaled image
///  \param newHeight desired height of the scaled image
///  \param method scaling method
///  \param dest destination pixel array
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPixelArray::scale (int          newWidth, 
                     int          newHeight, 
                     tlPixelArray *dest,
                     int          method)
{
  tlPixel *output;


  // only for single-byte image
  if (hr) {
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR_HIGHRES_INCOMPATIBLE);
    return -1;
  }

  // get scaled pixel array
  switch (method) {
  case TL_SCALE_DECIMATE:
    output = tl_scale_decimate (pixel, newWidth, newHeight, width, height, pixelWidth);
    break;
  case TL_SCALE_BSPLINE:
    tl_scale_bspline_init (pixel, width, height, pixelWidth);
    output = tl_scale_bspline (newWidth, newHeight);
    tl_scale_bspline_cleanup ();
    break;
  default:
    TL_ERRNO_MSG ("tlPixelArray::scale()", TL_ERROR);
    return -1;
  }

  // init destination pixel array
  dest->init (format, newWidth, newHeight);

  // replace current
  dest->deletePixel ();
  dest->pixel = output;


  return 0;
}
