/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#include "tlPoint.h"


///  \file tlPoint.h
///  \brief Image point object description.

///  \class tlPoint
///  \brief Point object used for image processing.
///
///  Simple point coordinates are stored in tlPoint objects. tlPoint contains both 2D image
///  coordonates, a disparity value, and 3D world coordonates. The tlPoint object also provides
///  convenience routines to update and copy point coordinates.



///  Constructor
/// 
///  \note
///  See \ref errors "error management" for details.

tlPoint::tlPoint () : tlLocation ()
{
  return;
}



///  Constructor
///
///  \param x      initial x location [pix]
///  \param y      initial y location [pix]
///  \param disp   initial disparity  [disp]
/// 
///  \note
///  See \ref errors "error management" for details.

tlPoint::tlPoint (int x, 
                  int y, 
                  int disp) : tlLocation (x, y, disp)
{
  return;
}



///  Constructor
///  \param point     point location
/// 
///  \note
///  See \ref errors "error management" for details.

tlPoint::tlPoint (tlPoint *point) : tlLocation (point->x, point->y)
{
  return;
}



///  Destructor
/// 
///  \note
///  See \ref errors "error management" for details.

tlPoint::~tlPoint ()
{
  tl_errno = TL_NO_ERROR;
}



///  Create a copy of the point.
///
///  \return
///  A pointer to a copy of the current point on success, -1 otherwise. 
///  See \ref errors "error management" for details.
///
///  \note
///  It is the responsibility of the programmer to delete the
///  point returned.

tlPoint *
tlPoint::copy ()
{
  return new tlPoint (x, y, disp);
}


///  Copy the point data from an existing tlPoint object.
///
///  \param point \b [out] destination point
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPoint::copy (tlPoint *point)
{
  // check
  if (!point) {
    TL_ERRNO_MSG ("tlPoint::copy()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  this->x    = point->x;
  this->y    = point->y;
  this->disp = point->disp;

  return 0;
}



///  Copy point location to a new tlPoint object.
///
///  \param point \b [out] destination point
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPoint::copyTo (tlPoint *point)
{
  // check
  if (!point) {
    TL_ERRNO_MSG ("tlPoint::copyTo()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  point->x    = this->x;
  point->y    = this->y;
  point->disp = this->disp;

  return 0;
}



///  Compute the distance to another point
///
///  \param point other extremity of distance to compute
///
///  \return
///  The distance between the point on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPoint::dist (tlPoint *point)
{
  if (!point) {
    TL_ERRNO_MSG ("tlPoint::dist()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  return this->dist (point->x, point->y);
}



///  Compute the distance to another point
///
///  \param x  x coordinate of other extremity of distance to compute
///  \param y  y coordinate of other extremity of distance to compute
///
///  \return
///  The distance between the point on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlPoint::dist (int x, 
               int y)
{
  return (int) (sqrt (pow ((float)((this->x) - (x)), 2.0f) + pow ((float)((this->y) - (y)), 2.0f)));
}
