/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#ifdef LINUX
#include <stdlib.h>
#endif

#include "tlRect.h"
#include "tlPoint.h"


///  \file tlRect.h
///  \brief Rectangle object description.


/// \class tlRect
/// \brief Rectangle object.
///
/// One of the most commonly used objects in TLIB is the tlRect object. The tlRect object
/// contains simple rectangular data (position, size) and routines to duplicate,
/// move and resize it. It is used mostly to define a ROI (Region Of Interest, i.e. to
/// limit the image processing to a given portion of an image), leading to shorter execution times. 



///  Constructor, initialize all members to 0.
///
///  \note
///  See \ref errors "error management" for details.

tlRect::tlRect () : tlSize (), tlLocation ()
{
  return;
}



///  Constructor, initialize all members to the desired value.
///
///  \param x       rectangle upper-left corner x position [pix]
///  \param y       rectangle upper-left corner y position [pix]
///  \param width   rectangle width                        [pix]
///  \param height  rectangle height                       [pix]
///
///  \note
///  See \ref errors "error management" for details.

tlRect::tlRect (int x, 
                int y, 
                int width, 
                int height) : tlSize (width, height), tlLocation (x, y)
{
  return;
}



///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlRect::~tlRect ()
{
  return;
}



///  Set rectangle member values.
///
///  \param x       rectangle upper-left corner x position [pix]
///  \param y       rectangle upper-left corner y position [pix]
///  \param width   rectangle width                        [pix]
///  \param height  rectangle height                       [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlRect::set (int x, 
             int y, 
             int width, 
             int height)
{
  this->x      = x;
  this->y      = y;
  this->width  = width;
  this->height = height;

  return 0;
}



///  \overload

int
tlRect::set (tlPoint *p, 
             int     width, 
             int     height)
{
  // check
  if (!p) {
    TL_ERRNO_MSG ("tlRect::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  this->x      = p->x;
  this->y      = p->y;
  this->width  = width;
  this->height = height;

  return 0;
}



///  \overload

int
tlRect::set (tlPoint * p1, 
             tlPoint * p2)
{
  // check
  if ((!p1) || (!p2)) {
    TL_ERRNO_MSG ("tlRect::set()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  this->x      = TL_MIN ((p1->x), (p2->x));
  this->y      = TL_MIN ((p1->y), (p2->y));
  this->width  = abs ((p1->x) - (p2->x));
  this->height = abs ((p1->y) - (p2->y));

  return 0;
}



///  Evenly shrinks the rectangle around its center.
///
///  \param dsize     width/height decrease [pix]
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlRect::collapse (int dsize)
{
  x      += (dsize / 2);
  y      += (dsize / 2);
  width  -= dsize;
  height -= dsize;

  return 0;
}



///  \overload

int
tlRect::collapse (int dwidth, 
                  int dheight)
{
  x      += (dwidth  / 2);
  y      += (dheight / 2);
  width  -= dwidth;
  height -= dheight;

  return 0;
}



///  Evenly expands the rectangle around its center  
///
///  \param dsize     width/height decrease [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlRect::expand (int dsize)
{
  x      -= (dsize / 2);
  y      -= (dsize / 2);
  width  += dsize;
  height += dsize;

  return 0;
}



///  \overload

int
tlRect::expand (int dwidth, 
                int dheight)
{
  x      -= (dwidth  / 2);
  y      -= (dheight / 2);
  width  += dwidth;
  height += dheight;

  return 0;
}


///  Set the location of the rectangle center.  
///
///  \param x     rectangle center x position [pix]
///  \param y     rectangle center corner y position [pix]
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlRect::setCenter (int x, 
                   int y)
{
  this->x = x - (width  / 2);
  this->y = y - (height / 2);

  return 0;
}



///  \overload

int
tlRect::setCenter (tlPoint * p)
{
  // check
  if (!p) {
    TL_ERRNO_MSG ("tlRect::setCenter()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  this->x = p->x - (width  / 2);
  this->y = p->y - (height / 2);

  return 0;
}



///  Create a copy of the rectangle.
///
///  \return
///  A pointer to the new (copied) rectangle.
///
///  \warning
///  The new rectangle is allocated. Do not forget to deallocate it
///  using \c delete when done with it.
///
///  \sa
///  copy(tlRect *rect), copyTo (tlRect *rect)

tlRect *
tlRect::copy ()
{
  return new tlRect (x, y, width, height);
}



///  Copy the rectangle parameters from another tlRect object.
///
///  \param rect      the rectangle that should be copied
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  copy(), copyTo(tlRect *rect)

int
tlRect::copy (tlRect * rect)
{
  return rect->copyTo (this);
}



///  Copy rectangle parameters to another rect object
///
///  \param rect      the rectangle that should contained copied parameters
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \sa
///  copy (), copy (tlRect *rect)

int
tlRect::copyTo (tlRect * rect)
{
  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlRect::copyTo()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  rect->x      = x;
  rect->y      = y;
  rect->width  = width;
  rect->height = height;

  return 0;
}



///  Check if a given rectangle intesects with the current object.
///
///  \param rect      the rectangle to check for intersection
///
///  \return
///  1 if rectangles intersect, 0 if not, -1 on error.\n
///  See \ref errors "error management" for details.

int
tlRect::intersect (tlRect * rect)
{
  // check
  if (!rect) {
    TL_ERRNO_MSG ("tlRect::intersect()", TL_ERROR_NULL_POINTER);
    return -1;
  }

  // process
  if (
    (rect->x+rect->width  < x) ||
    (rect->y+rect->height < y) ||
    (rect->x > x + width)      ||
    (rect->y > y + height)) 
  {
    return 1;
  }


  return 0;
}
