/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>
#include <assert.h>

#include "tlScale.h"



///  \file tlScale.h
///  \brief Internal scaling routines.
///  \internal


///  \def TL_SCALE_DECIMATE
///  Scaling method that uses decimation (pick the nearest pixel). This is a fast,
///  noisy (inaccurate) scaling method.

///  \def TL_SCALE_BSPLINE
///  Scaling method that uses bicubic splines to interpolate scaled pixel values.
///  Slower, yet best method (minimizes scaling error). The version used here
///  uses order 2 polynomials (NOT bicubic) for optimization purposes.


#define TL_SCALE_BSPLINE_TOL                0.0001f
#define Z0                                 -0.1715728752538099023966225515806f
#define Z1                                 -0.26794919243112270647255365849413f

// internal 
static float *coeff2[TL_MAX_CHANNEL];
static long _tl_width, _tl_height, _tl_pixel_width;



///  Compute initial value for causal coefficient filter.
///
///  \param c       array containing data to interpolate
///  \param l       length of array
///  \param tol     interpolation tolerance
///  \param z       pole to use
///
///  \return
///  initial value

float
tl_scale_bspline_init_causal (float *c, 
                              long  l, 
                              float tol, 
                              float z)
{
  float sum, zn, z2n, iz;
  long  n, horizon;


  // this initialization corresponds to mirror boundaries 
  horizon = l;
  if (tol > 0.0)
    horizon = (long) ceil (log (tol) / log (fabs (z)));

  // accelerated loop 
  if (horizon < l) {
    zn = z;
    sum = c[0];
    for (n = 1; n < horizon; n++) {
      sum += zn * c[n];
      zn *= z;
    }
    return sum;
  }

  // full loop 
  else {
    zn = z;
    iz = 1.0f / z;
    z2n = (float) pow (z, (float) (l - 1));
    sum = c[0] + z2n * c[l - 1];
    z2n *= z2n * iz;
    for (n = 1; n <= l - 2; n++) {
      sum += (zn + z2n) * c[n];
      zn *= z;
      z2n *= iz;
    }
    return (sum / (1.0f - zn * zn));
  }
}



///  Compute initial value for anticausal coefficient filter.
///
///  \param c       array containing data to interpolate
///  \param l       length of array
///  \param z       pole to use
///
///  \return
///  initial value

float
tl_scale_bspline_init_anticausal (float *c, 
                                  long  l, 
                                  float z)
{
  // this initialization corresponds to mirror boundaries 
  return (float) ((z / (z * z - 1.0)) * (z * c[l - 2] + c[l - 1]));
}



///  Initialize interpolation bspline coeff for a single line or column.
///
///  \param i       array of data to interpolate
///  \param l       length of array
///  \param tol     <default=BSPLINE2D_TOL> tolerance
///  \param z       pole to use

void
tl_scale_bspline_convert (float *coeff, 
                          long  length, 
                          float tol, 
                          float z)
{
  float lambda = 1.0;
  long  n;


  // special case required by mirror boundaries 
  if (length == 1) {
    return;
  }

  // compute the overall gain 
  lambda = lambda * (1.0f - Z1) * (1.0f - 1.0f / Z1);

  // apply the gain 
  for (n = 0; n < length; n++) {
    coeff[n] *= lambda;
  }

  // causal initialization 
  coeff[0] = tl_scale_bspline_init_causal (coeff, length, tol, z);

  // causal recursion 
  for (n = 1; n < length; n++) {
    coeff[n] += Z1 * coeff[n - 1];
  }

  // anticausal initialization 
  coeff[length - 1] = tl_scale_bspline_init_anticausal (coeff, length, z);

  // anticausal recursion 
  for (n = length - 2; 0 <= n; n--) {
    coeff[n] = Z1 * (coeff[n + 1] - coeff[n]);
  }
}



///  Retrieve a column from an image.
///
///  \param array   image data array
///  \param width   image width
///  \param height  image height
///  \param col     index of column to retrieve
///  \param line    \b [out] array containing the given column
void
tl_scale_bspline_getcol (float *array, 
                         float *line, 
                         long  width, 
                         long  height,
                         long  col)
{
  long j;

  for (j = 0; j < height; j++) {
    line[j] = array[j * width + col];
  }
}



///  Replace a column in an image.
///
///  \param array   \b [out] array containing the column to be replaced
///  \param line    column containing values to replace
///  \param width   image width
///  \param height  image height
///  \param col     index of column to retrieve

void
tl_scale_bspline_putcol (float *array, 
                         float *line, 
                         long  width, 
                         long  height,
                         long  col)
{
  long j;

  for (j = 0; j < height; j++) {
    array[j * width + col] = line[j];
  }
}



///  Compute interpolation with bspline of degree 2.
///
///  \param x         X position at which to interpolate
///  \param y         Y position at which to interpolate
///  \param channel   the given image channel
///
///  \return
///  The interpolated value at the given position.

inline float
tl_scale_bspline_fast_interpolate (float x, 
                                   float y, 
                                   int   channel)
{
  float xweight0, xweight1, xweight2;
  float yweight0, yweight1, yweight2;
  float w;
  float interpolated;
  float *p;
  int   xindex0, xindex1, xindex2;
  int   yindex0, yindex1, yindex2;
  int   iwidth;
  int   iheight;
  int   i, j;
  float X, Y;


  // shortcuts 
  iwidth  = 2 * _tl_width - 2;
  iheight = 2 * _tl_height - 2;

  // compute the interpolation indexes 
  X = x + 0.5f;
  i = (int) X - 1;
  if (X < 0) {
    i--;
  }
  Y = y + 0.5f;
  j = (int) Y - 1;
  if (Y < 0) {
    j--;
  }
  xindex0 = i++;
  xindex1 = i++;
  xindex2 = i++;
  yindex0 = j++;
  yindex1 = j++;
  yindex2 = j++;

  // compute the X interpolation weights 
  w = x - xindex1;
  xweight1 = 0.75f - w * w;
  xweight2 = 0.5f * (w - xweight1 + 1.0f);
  xweight0 = 1.0f - xweight1 - xweight2;

  // compute the Y interpolation weights 
  w = y - yindex1;
  yweight1 = 0.75f - w * w;
  yweight2 = 0.5f * (w - yweight1 + 1.0f);
  yweight0 = 1.0f - yweight1 - yweight2;

#if 0
  // currently disabled... see note in tl_scale_bspline_init 

  // apply the mirror boundary conditions 
  if (_tl_width == 1) {
    xindex0 = xindex1 = xindex2 = 0;
  }
  if (_tl_height == 1) {
    yindex0 = yindex1 = yindex2 = 0;
  }
#endif

  if (xindex0 < 0) {
    xindex0 = 0;
  }
  else if (_tl_width <= xindex0) {
    xindex0 = iwidth - xindex0;
  }
  if (yindex0 < 0) {
    yindex0 = 0;
  }
  else if (_tl_height <= yindex0) {
    yindex0 = iheight - yindex0;
  }
  if (xindex1 < 0) {
    xindex1 = 0;
  }
  else if (_tl_width <= xindex1) {
    xindex1 = iwidth - xindex1;
  }
  if (yindex1 < 0) {
    yindex1 = 0;
  }
  else if (_tl_height <= yindex1) {
    yindex1 = iheight - yindex1;
  }
  if (xindex2 < 0) {
    xindex2 = 0;
  }
  else if (_tl_width <= xindex2) {
    xindex2 = iwidth - xindex2;
  }
  if (yindex2 < 0) {
    yindex2 = 0;
  }
  else if (_tl_height <= yindex2) {
    yindex2 = iheight - yindex2;
  }

  // perform interpolation 
  p = &(coeff2[channel][yindex0 * _tl_width]);
  w = xweight0 * p[xindex0] + xweight1 * p[xindex1] + xweight2 * p[xindex2];
  interpolated = yweight0 * w;
  p = &(coeff2[channel][yindex1 * _tl_width]);
  w = xweight0 * p[xindex0] + xweight1 * p[xindex1] + xweight2 * p[xindex2];
  interpolated += yweight1 * w;
  p = &(coeff2[channel][yindex2 * _tl_width]);
  w = xweight0 * p[xindex0] + xweight1 * p[xindex1] + xweight2 * p[xindex2];
  interpolated += yweight2 * w;

  return interpolated;
}



///  Fast bspline interpolation cleanup

int
tl_scale_bspline_cleanup ()
{
  int k;

  for (k = 0; k < _tl_pixel_width; k++) {
    free (coeff2[k]);
  }

  return 0;
}



///  Fast bspline interpolation initialization

int
tl_scale_bspline_init (tlPixel *src, 
                       int     width, 
                       int     height, 
                       int     pixel_width)
{
  int   i, j, k;
  int   index;
  int   src_index;
  float *line, *lineY;
  int   coeff2Size      = height * width * sizeof (float);
  int   lineSize        = height * sizeof (float);


  // if width or height == 1, then need to apply mirror boundary
  // conditions in tl_scale_bspline_fast_interpolate... (currently disabled) 
  assert (width != 1);
  assert (height != 1);

  // store locally 
  _tl_width = width;
  _tl_height = height;
  _tl_pixel_width = pixel_width;

  // create coeff arrays 
  lineY = (float *) malloc (lineSize);
  for (k = 0; k < pixel_width; k++) {
    coeff2[k] = (float *) malloc (coeff2Size);
    index = 0;
    src_index = 0;
    for (j = 0; j < height; j++) {
      for (i = 0; i < width; i++) {
        coeff2[k][index++] = (float) (src[src_index + k]);
        src_index += pixel_width;
      }
    }

    // perform conversion along x 
    for (j = 0; j < height; j++) {
      line = &(coeff2[k][j * width]);
      tl_scale_bspline_convert (line, width, TL_SCALE_BSPLINE_TOL, Z0);
    }

    // perform conversion along y 
    for (i = 0; i < width; i++) {
      tl_scale_bspline_getcol (coeff2[k], lineY, width, height, i);
      tl_scale_bspline_convert (lineY, height, TL_SCALE_BSPLINE_TOL, Z0);
      tl_scale_bspline_putcol (coeff2[k], lineY, width, height, i);
    }
  }
  free (lineY);


  return 0;
}



///  Fast bspline interpolation

tlPixel *
tl_scale_bspline (float factor)
{
  register int i, j, k;
  register int index = 0;

  float x           = 0.0f;
  float y           = 0.0f;
  float dx          = 1.0f / factor;
  float dy          = 1.0f / factor;
  int   new_width   = (int) (factor * _tl_width);
  int   new_height  = (int) (factor * _tl_height);


  // buffer allocation
  tlPixel *dest = tl_create_array (new_width, new_height, _tl_pixel_width);

  // unrolled loop: only valid up to 4
  assert (TL_MAX_PIXEL_WIDTH == 4);

  // scaling operation
  for (j = 0; j < new_height; j++) {
    for (i = 0; i < new_width; i++) {
      for (k = 0; k < _tl_pixel_width; k++) {
        int val = (int) tl_scale_bspline_fast_interpolate (x, y, k);
        if (val > TL_MAX_PIXEL_VAL) {
          dest[index++] = TL_MAX_PIXEL_VAL;
        }
        else if (val < TL_MIN_PIXEL_VAL) {
          dest[index++] = TL_MIN_PIXEL_VAL;
        }
        else {
          dest[index++] = (tlPixel) val;
        }
      }
      x += dx;
    }
    x = 0.0f;
    y += dy;
  }


  return dest;
}



///  Fast bspline interpolation

tlPixel *
tl_scale_bspline (int new_width, 
                  int new_height)
{
  int   i, j, k;
  int   index     = 0;
  float x         = 0.0f;
  float y         = 0.0f;
  float dx        = (float) _tl_width  / (float) new_width;
  float dy        = (float) _tl_height / (float) new_height;


  // buffer allocation
  tlPixel *dest = tl_create_array (new_width, new_height, _tl_pixel_width);

  // scaling operation
  for (j = 0; j < new_height; j++) {
    for (i = 0; i < new_width; i++) {
      for (k = 0; k < _tl_pixel_width; k++) {
        float val = tl_scale_bspline_fast_interpolate (x, y, k);
        if (val > (float) TL_MAX_PIXEL_VAL) 
          dest[index + k] = TL_MAX_PIXEL_VAL;
        else if (val < (float) TL_MIN_PIXEL_VAL)
          dest[index + k] = TL_MIN_PIXEL_VAL;
        else
          dest[index + k] = (tlPixel) val;
      }
      index += _tl_pixel_width;
      x += dx;
    }
    x = 0.0f;
    y += dy;
  }


  return dest;
}



///  Decimation interpolation

tlPixel *
tl_scale_decimate (tlPixel *src, 
                   float   factor, 
                   int     width, 
                   int     height,
                   int     pixel_width)
{
  int     i, j, k;
  int     new_width, new_height;
  int     new_index, index;
  int     src_row;
  tlPixel *output;


  // compute new size
  new_width = (int) (factor * width);
  new_height = (int) (factor * height);

  // create new pixel array
  output = (tlPixel *) malloc (new_width * new_height * pixel_width * sizeof (tlPixel));

  // scaling operation
  new_index = 0;
  for (j = 0; j < new_height; j++) {
    src_row = (int) (j * height / new_height * width);
    for (i = 0; i < new_width; i++) {
      index = (src_row + (int) (i * width / new_width)) * pixel_width;
      // copy the pixel
      for (k = 0; k < pixel_width; k++) {
        output[new_index++] = src[index++];
      }
    }
  }


  return output;
}



///  Decimation interpolation

tlPixel *
tl_scale_decimate (tlPixel *src, 
                   int     new_width, 
                   int     new_height, 
                   int     width,
                   int     height, 
                   int     pixel_width)
{
  int     i, j, k;
  int     new_index, index;
  int     src_row;
  tlPixel *output;


  // create new pixel array
  output = (tlPixel *) malloc (new_width * new_height * pixel_width *	sizeof (tlPixel));

  // scaling operation
  new_index = 0;
  for (j = 0; j < new_height; j++) {
    src_row = (int) (j * height / new_height * width);
    for (i = 0; i < new_width; i++) {
      index = (src_row + (int) (i * width / new_width)) * pixel_width;
      // copy the pixel
      for (k = 0; k < pixel_width; k++) {
        output[new_index++] = src[index++];
      }
    }
  }


  return output;
}
