/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlSource.h"

// "dirty" globals
static int  list[TL_MAX_SOURCE];
static bool first = true;

// static initialization
int tlSource::count = 0;



///  \class tlSource
///  \brief Image acquisition base class.
///
///  This is the base class for any image acquisition for TLIB.
///  It is used by tlDigitizer, tlCamera, tlSourceVideoFile and tlSourceFiles.
///  You should only use tlSource as a base class to implement your own image acquisition interface.


///  \def TL_MAX_SOURCE
///  The maximum allowed number of image sources.



///  \fn tlSource::tlSource()
///  Base class constructor.
///
///  \note
///  See \ref errors "error management" for details.



///  \fn tlSource::~tlSource()
///  Base class destructor.
///
///  \note
///  See \ref errors "error management" for details.



///  \fn tlSource::getWidth()
///  Returns the width of the image source. Any image acquired through this tlSource object
///  will have the current width.
///
///  \return 
///  The width in pixels of the image source.\n
///  See \ref errors "error management" for details.



///  \fn tlSource::getHeight()
///  Returns the height of the image source. Any image acquired through this tlSource object
///  will have the current height.
///
///  \return 
///  The height in pixels of the image source.\n
///  See \ref errors "error management" for details.



///  \fn tlSource::getID()
///  Return a unique identifier for this source. Identifiers are allocated dynamically and are
///  only valid during application run-time.
///
///  \return A unique numerical identifier on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlSource::getFormat()
///  Returns the \ref imageformats "format" of the image source. Any image acquired through this tlSource object
///  will have the current format.
///
///  \return 
///  The format of the image source.



///  \fn tlSource::grab(tlImage *image)
///  This call acquires a full image from an image source. The image will have the size and format the source
///  is currently configured for.
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlSource::grab(tlImage *image, tlRect *rect)
///  This call acquires a portion of an image from an image source. The image will have the size and format the source
///  is currently configured for, but only the portion defined by the \e rect argument will contain valid pixel
///  values.
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
///  \param rect a pointer to the portion of the image to acquire.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.



///  \fn tlSource::grabQueued(tlImage *image)
///  This call acquires a full image from an image source. The image will have the size and format the source
///  is currently configured for.
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  This routine is faster than grab() because image acquisition
///  happens in the background. However, since the grab command is sent 
///  to the grabber during the previous call to the grabQueued() 
///  routine, the image returned can be out-of-date.



///  \fn tlSource::grabQueued(tlImage *image, tlRect *rect)
///  This call acquires a portion of an image from an image source. The image will have the size and format the source
///  is currently configured for, but only the portion defined by the \e rect argument will contain valid pixel
///  values.
/// 
///  \param image \b [out] a valid image pointer that will receive the image.
///  \param rect a pointer to the portion of the image to acquire.
/// 
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
/// 
///  \note
///  This routine is faster than grab() because image acquisition
///  happens in the background. However, since the grab command is sent 
///  to the grabber during the previous call to the grabQueued() 
///  routine, the image returned can be out-of-date.



////////////////////////////////////////////////////////////////////////

tlSource::tlSource ()
{
  int k;

  // if this is the first source
  if (first == 1) {

    // initialize device list
    for (k = 0; k < TL_MAX_SOURCE; k++)
      list[k] = 0;

    // finish initialization
    first = false;
  }

  // find first available source index
  this->ID = -1;
  for (k = 0; k < TL_MAX_SOURCE; k++) {
    if (list[k] == 0) {
      this->ID = k;
      list[k] = 1;
      break;
    }
  }

  // check for availability
  if ((this->ID) == -1) {
    TL_ERRNO_MSG ("tlSource::open()", TL_ERROR_MAX_REACHED);
    return;
  }

  // init default fields
  this->width = TL_DEFAULT_COLS;
  this->height = TL_DEFAULT_ROWS;
  this->format = RGB;

  tl_errno = TL_NO_ERROR;

  // update count;
  count++;
}



////////////////////////////////////////////////////////////////////////

tlSource::~tlSource ()
{
  list[ID] = 0;
  count--;
}


// documentation

/// \file tlSource.h
/// \brief Image acquisition base class.
