/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <string.h>
#include <sys/stat.h>

#include "tlSourceFiles.h"
#include "tlImage.h"
#include "tlError.h"


#ifndef NO_DOC
static int
fileExists (char *filename)
{
  struct stat fileStat;

  if (-1 == stat (filename, &fileStat)) {
    return (0);
  }
  return (1);
}
#endif



///  \file tlSourceFiles.h
///  \brief Filesystem "digitizer" class.

///  \class tlSourceFiles
///  Filesystem digitizer class.



///  \fn tlSourceFiles::getLength()
///  Return the length of the image if it is known. The full set must be
///  grabbed once before the length is known.
///
///  \return
///  The file set length if it is known, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn tlSourceFiles::getIndex()
///  Return the current index in the file set.
///
///  \return
///  The position in the file set.\n
///  See \ref errors "error management" for details.



///  Constructor.
///
///  \note
///  See \ref errors "error management" for details.

tlSourceFiles::tlSourceFiles () : tlSource ()
{
  this->filenameBase = NULL;
  if (tl_errno == TL_NO_ERROR) return;
  open ((char *) TL_SOURCEFILES_DEFAULT_FILENAME);
}



///  Constructor.
///
///  \return
///  See \ref errors "error management" for details.

tlSourceFiles::tlSourceFiles (char *filenameBase) :	tlSource ()
{
  this->filenameBase = NULL;
  if (tl_errno == TL_NO_ERROR) return;
  open (filenameBase);
}



///  Open the first file of the serie using the default basename.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSourceFiles::open ()
{
  return (open ((char *) TL_SOURCEFILES_DEFAULT_FILENAME));
}



///  Open the first file of the serie using the default basename.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSourceFiles::open (char *filenameBase)
{
  char filename[255];


  if (this->filenameBase) free (this->filenameBase);
  this->filenameBase = (char *)malloc(strlen(filenameBase)+1);
  strcpy (this->filenameBase, filenameBase);

  current = 0;
  length = -1;
  delay = 0;
  lastGrabTime = tl_timer_get_time ();

  // check if first file exists, if not return error
  sprintf (filename, "%s%d.tlb", filenameBase, current);
  if (!fileExists (filename)) {
    TL_ERRNO_MSG ("tlSourceFiles::open()", TL_ERROR_FILE_NOT_FOUND_OR_CORRUPT);
    return -1;
  }


  return 0;
}



///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlSourceFiles::~tlSourceFiles ()
{
  if (filenameBase) {
    free (filenameBase);
  }
  tl_errno = TL_NO_ERROR;
}



///  Set the "frame rate" of the filesystem digitizer by "throttling"
///  the acquisition.
///
///  \param fps frame rate in frame per second.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.
///
///  \note
///  Throttle code depends on the main TLIB application timer. So, if the
///  user calls tl_timer_init(), the frame rate will be messed up.

int
tlSourceFiles::setFrameRate (int fps)
{
  if (fps <= 0) {
    delay = 0;
  } else {
    delay = 1000 / fps;
  }


  return 0;
}



///  Set the current frame index. The image at the given index will
///  be returned at the next grab() call.
///
///  \param index desired frame index.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSourceFiles::setIndex (int index)
{
  char filename[255];


  // if length currently unknown
  if (length == -1) {
    sprintf (filename, "%s%d.tlb", filenameBase, index);
    if (!fileExists (filename)) {
      TL_ERRNO_MSG ("tlSourceFiles->setIndex()", TL_ERROR);
      return -1;
    }
  } else if (index >= length) {
    TL_ERRNO_MSG ("tlSourceFiles->setIndex()", TL_ERROR_INDEX_OUT_OF_BOUNDS);
    return -1;
  }
  // if length is known
  current = index;


  return 0;
}



////////////////////////////////////////////////////////////////////////

int
tlSourceFiles::grab (tlImage *image)
{
  char filename[255];


  // throttle to fps if setFrameRate has been called. timing accuracy
  // is limited by tl_timer_get_time() resolution and by tlThreadSleep
  if (delay) {
    long int
      curTime = tl_timer_get_time ();
    long int
      dt = curTime - lastGrabTime;
    if (dt < delay) {
      tlThreadSleep (delay - dt);
    }
    lastGrabTime = tl_timer_get_time ();
  }

  // check if next file exists, if not reset index
  sprintf (filename, "%s%d.tlb", filenameBase, current++);
  if (!fileExists (filename)) {
    if (tlVerbose) {
      printf ("tlSourceFiles::grab: %s does not exist, resetting curFileNum\n", filename);
    }
    length = current - 1;
    current = 0;
    sprintf (filename, "%s%d.tlb", filenameBase, current);
  }


  return image->readFromFile (filename);
}



////////////////////////////////////////////////////////////////////////

int
tlSourceFiles::grab (tlImage *image, 
                     tlRect  *rect)
{
  return grab (image);
}



////////////////////////////////////////////////////////////////////////

int
tlSourceFiles::grabQueued (tlImage *image)
{
  return grab (image);
}



////////////////////////////////////////////////////////////////////////

int
tlSourceFiles::grabQueued (tlImage *image, 
                           tlRect  *rect)
{
  return grab (image);
}
