/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include <math.h>

#ifdef LINUX
#include <stdlib.h>
#endif

#include "tlSpiral.h"
#include "tlPoint.h"
#include "tlRect.h"



///  \file tlSpiral.h
///  \brief Spiral coordinates for pixel browsing.

///  \class tlSpiral
///  \brief Spiral coordinates for pixel browsing.
///
///  A tlSpiral object provides a discrete spiral coordinates computation respectively
///  to a given starting point. This allows a routine to address image pixels in a spiral
///  sequence, which can be extremely useful for some correlation implementations or 
///  simply to find the closest filtered neighbor of a filtered pixel.



///  \fn tlPoint *tlSpiral::get()
///  Return a tlPoint to the current position.
///
///  \return
///  A pointer to a tlPoint on success, NULL otherwise.
///  See \ref errors "error management" for details.
///
///  \remark
///  The return value points to a copy of the tlPoint. Therefore,
///  you should explicitely delete the point after use to avoid memory leak.
///  See get(tlPoint *point) to avoid allocating new memory.



///  \fn int tlSpiral::get(tlPoint *point)
///  Fills a tlPoint with the current position.
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
///  \remark
///  See get() to get a freshly allocated copy of the point.



///  \fn int tlSpiral::getX()
///  Return the current position in X.
///
///  \return
///  The current position in X on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn int tlSpiral::getY()
///  Return the current position in Y.
///
///  \return
///  The current position in Y on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn tlPoint *tlSpiral::getLast()
///  Return a tlPoint to the previous position.
///
///  \return
///  A pointer to a tlPoint on success, NULL otherwise.
///  See \ref errors "error management" for details.
///
///  \remark
///  The return value points to a copy of the tlPoint. Therefore,
///  you should explicitely delete the point after use to avoid memory leak.
///  See getLast(tlPoint *point) to avoid allocating new memory.



///  \fn int tlSpiral::getLast(tlPoint *point)
///  Fills a tlPoint with the previous position.
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
///  \remark
///  See getLast() to get a freshly allocated copy of the point.



///  \fn int tlSpiral::getLastX()
///  Return the previous position in X.
///
///  \return
///  The previous position in X on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn int tlSpiral::getLastY()
///  Return the previous position in Y.
///
///  \return
///  The previous position in Y on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn tlPoint *tlSpiral::getNext()
///  Return a tlPoint to the next position.
///
///  \return
///  A pointer to a tlPoint on success, NULL otherwise.
///  See \ref errors "error management" for details.
///
///  \remark
///  The return value points to a copy of the tlPoint. Therefore,
///  you should explicitely delete the point after use to avoid memory leak.
///  See getLast(tlPoint *point) to avoid allocating new memory.



///  \fn int tlSpiral::getNext(tlPoint *point)
///  Fills a tlPoint with the next position.
///
///  \return
///  0 on success, -1 otherwise.
///  See \ref errors "error management" for details.
///
///  \remark
///  See getNext() to get a freshly allocated copy of the point.



///  \fn int tlSpiral::getNextX()
///  Return the next position in X.
///
///  \return
///  The next position in X on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn int tlSpiral::getNextY()
///  Return the next position in Y.
///
///  \return
///  The next position in Y on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn int tlSpiral::getIndex()
///  Return the pixel array index of the current pixel.
///
///  \return
///  The pixel array index of the current pixel on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn int tlSpiral::getLastIndex()
///  Return the pixel array index of the previous pixel.
///
///  \return
///  The pixel array index of the previous pixel on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  \fn int tlSpiral::getNextIndex()
///  Return the pixel array index of the next pixel.
///
///  \return
///  The pixel array index of the next pixel on success, -1 otherwise.
///  See \ref errors "error management" for details.



///  Constructor, creates a spiral centered relatively to the default image size.
///
///  \note
///  See \ref errors "error management" for details.

tlSpiral::tlSpiral (tlRect * rect)
{
  tl_errno = TL_NO_ERROR;
  init (TL_DEFAULT_COLS / 2, TL_DEFAULT_ROWS / 2, TL_DEFAULT_COLS, TL_DEFAULT_ROWS, rect);
}



///  Constructor, creates a spiral centered at the given point.
///
///  \param x0			spiral starting point on X axis
///  \param y0			spiral starting point on Y axis
///  \param width	  image width
///  \param height	image height
///  \param rect		portion of the image to consider
///
///  \note
///  See \ref errors "error management" for details.

tlSpiral::tlSpiral (int    x0, 
                    int    y0, 
                    int    width, 
                    int    height, 
                    tlRect *rect)
{
  tl_errno = TL_NO_ERROR;
  init (x0, y0, width, height, rect);
}



///  Destructor.
///
///  \note
///  See \ref errors "error management" for details.

tlSpiral::~tlSpiral ()
{
  tl_errno = TL_NO_ERROR;
}



///  Constructor, creates a spiral centered at the given point.
///
///  \param p0			spiral starting point
///  \param width	  image width
///  \param height	image height
///  \param rect		portion of the image to consider
///
///  \note
///  See \ref errors "error management" for details.

tlSpiral::tlSpiral (tlPoint *p0, 
                    int     width, 
                    int     height, 
                    tlRect  *rect)
{
  tl_errno = TL_NO_ERROR;
  init (p0->x, p0->y, width, height, rect);
}



///  Initialize the spiral around a given point, for an image of a given
///  width, and within a certain portion of the image only.
///
///  \param p0			spiral starting point
///  \param width	  image width
///  \param height	image height
///  \param rect		portion of the image to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSpiral::init (tlPoint *p0, 
                int     width, 
                int     height, 
                tlRect  *rect)
{
  return init (p0->x, p0->y, width, height, rect);
}



///  Initialize the spiral around a given point, for an image of a given
///  width, and within a certain portion of the image only.
///
///  \param x0			spiral starting point on X axis
///  \param y0			spiral starting point on Y axis
///  \param width	image width
///  \param height	image height
///  \param rect		portion of the image to consider
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSpiral::init (int    x0, 
                int    y0, 
                int    width, 
                int    height, 
                tlRect *rect)
{
  this->x0     = x0;
  this->y0     = y0;
  this->x      = x0;
  this->y      = y0;
  this->nextX  = x0 + 1;
  this->nextY  = y0;
  this->lastX  = x0;
  this->lastY  = y0;
  this->width  = width;
  this->height = height;

  if (rect != NULL) {
    this->minX = TL_MAX ((rect->x), 0);
    this->maxX = TL_MIN ((rect->x) + (rect->width), width);
    this->minY = TL_MAX ((rect->y), 0);
    this->maxY = TL_MIN ((rect->y) + (rect->height), height);
  }
  else {
    this->minX = 0;
    this->maxX = width;
    this->minY = 0;
    this->maxY = height;
  }

  this->done = 0;
  this->condX = TL_MAX (abs (x0 - minX), abs (maxX - x0));
  this->condY = TL_MAX (abs (y0 - minY), abs (maxY - y0));

  cur_incr = 1;
  x_incr   = 0;
  y_incr   = 0;


  return 0;
}



///  Moves on to the next point on the spiral.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSpiral::toNext ()
{
  // vars
  int more = 1;

  // test
  if (done) return -1;

  // backup vars
  lastX = x;
  lastY = y;
  x = nextX;
  y = nextY;

  while (more) {

    // optimism...
    more = 0;

    // update vars (safe spiral)
    if (x_incr < abs (cur_incr)) {
      if (cur_incr > 0) nextX++;
      else nextX--;
      x_incr++;
    }
    else if (y_incr < abs (cur_incr)) {
      if (cur_incr > 0)	nextY++;
      else nextY--;
      y_incr++;
    }
    else {
      if (cur_incr < 0) cur_incr = (-1) * (cur_incr - 1);
      else cur_incr = (-1) * (cur_incr + 1);
      x_incr = y_incr = 0;
      more = 3;
    }
    if (nextX < minX) more++;
    else if (nextX >= maxX) more++;
    if (nextY < minY) more++;
    else if (nextY >= maxY) more++;

    // test for permanent exit
    if (more == 2) {
      if ((abs (nextY - y0) >= condY) && (abs (nextX - x0) >= condX)) {
        done = 1;
        more = 0;
      }
    }
  }

  return 0;
}



///  Moves on to the next point on the spiral and place the new
///  current location in a tlPoint.
///
///  \param point    point that receives the new current location
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tlSpiral::toNext (tlPoint * point)
{
  if (toNext () != -1) {
    return point->setLocation (x, y);
  }

  TL_ERRNO_MSG ("tlSpiral::toNext()", TL_ERROR_NOT_IN_IMAGE);
  return -1;
}
