/////////////////////////////////////////////////////////////////////////////////
//
//  This file is part of the TLIB computer vision library.
//  Copyright (C) 2003-2008 Sebastien Grange
//  Copyright (C) 2003-2007 VRAI Group, EPFL
//  All rights reserved.
// 
//  This library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License("GPL") version 2
//  as published by the Free Software Foundation.
// 
//  <http://www.tuyphon.com/tlib>
//
/////////////////////////////////////////////////////////////////////////////////



#include "tlTimer.h"

#ifdef WIN32
#include <stdio.h>
#include <conio.h>
#include <windows.h>
#endif

#ifdef LINUX
#include <sys/time.h>
#endif



///  \file tlTimer.h
///  \brief High resolution timers for TLIB.
///
///  LINUX: uses gettimeofday(), which provides resolution in microseconds.
///         The actual resolution is not nearly as precise, but is less
///         than 1 msec.
///
///  WIN32: uses GetTickCount (approx. 10 msec resolution) or the high
///         performance clock (less than 1 msec resolution)
///
///  The real resolution is machine dependant.

///  \def TL_MAX_TIMER
///  Maximum number of timers allowed.



#ifdef WIN32
// #define USE_GETTICK

#ifdef USE_GETTICK
static unsigned long _tl_timer_ref;
static unsigned long _tl_timer_array[TL_MAX_TIMER];
#else 
static LARGE_INTEGER _tl_timer_freq;
static LARGE_INTEGER _tl_timer_ref;
static LARGE_INTEGER _tl_timer_array[TL_MAX_TIMER];
#endif
#endif

#ifdef LINUX
struct timeval _tl_timer_ref;
struct timeval _tl_timer_array[TL_MAX_TIMER];
#endif



///  Initilize TL main timer.
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_timer_init (void)
{
#ifdef LINUX
  gettimeofday (&_tl_timer_ref, NULL);
  return TL_NO_ERROR;
#endif

#ifdef WIN32
#ifdef USE_GETTICK
  _tl_timer_ref = (unsigned long) GetTickCount ();
#else
  if (!QueryPerformanceFrequency (&_tl_timer_freq)) {
    TL_ERROR_MSG ("tl_timer_init(): warning: no high-resolution clock");
  }
  if (tlVerbose) {
    printf ("tl_timer_init(): high-resolution clock: %I64d ticks/sec\n",
      _tl_timer_freq.QuadPart);
  }
  if (!QueryPerformanceCounter (&_tl_timer_ref)) {
    TL_ERRNO_MSG ("tl_timer_init()", TL_ERROR);
    return -1;
  }
#endif
  return TL_NO_ERROR;
#endif
}



///  Poll TL main timer.
///
///  \return
///  Timer value [ms] on success, 0 otherwise.\n
///  See \ref errors "error management" for details.

unsigned long
tl_timer_get_time (void)
{

#ifdef LINUX
  struct timeval cTime;

  gettimeofday (&cTime, NULL);

  // compute difference in msec
  return ((long)
    ((cTime.tv_sec - _tl_timer_ref.tv_sec) * 1000.0 +
    ((float) (cTime.tv_usec - _tl_timer_ref.tv_usec)) / 1000.0));
#endif

#ifdef WIN32
#ifdef USE_GETTICK
  return (unsigned long) GetTickCount () - _tl_timer_ref;
#else
  {
    LARGE_INTEGER cTime;

    if (!QueryPerformanceCounter (&cTime)) {
      TL_ERRNO_MSG ("tl_timer_get_time()", TL_ERROR_TIMER);
      return 0;
    }
    return (unsigned long) ((cTime.QuadPart - _tl_timer_ref.QuadPart) * 1000 / _tl_timer_freq.QuadPart);	// msec
  }
#endif
#endif
}



///  Reset one of the TL timers.
///
///  \param index timer index
///
///  \return
///  0 on success, -1 otherwise.\n
///  See \ref errors "error management" for details.

int
tl_timer_reset (int index)
{
  if (index >= TL_MAX_TIMER) {
    return TL_ERROR;
  }

#ifdef LINUX
  gettimeofday (&_tl_timer_array[index], NULL);
  return TL_NO_ERROR;
#endif

#ifdef WIN32
#ifdef USE_GETTICK
  _tl_timer_array[index] = (unsigned long) GetTickCount ();
#else
  if (!QueryPerformanceCounter (&_tl_timer_array[index])) {
    TL_ERRNO_MSG ("tl_timer_reset(", TL_ERROR);
    return -1;
  }
#endif
  return TL_NO_ERROR;
#endif

}



///  Poll one of the TL timers.
///
///  \param index timer index
///
///  \return
///  Timer value [ms] 0 on success, 0 otherwise.\n
///  See \ref errors "error management" for details.

unsigned long
tl_timer_get (int index)
{
  if (index >= TL_MAX_TIMER) {
    return 0;
  }

#ifdef LINUX
  {
    struct timeval cTime;

    gettimeofday (&cTime, NULL);

    // compute difference in msec
    return ((long)
      ((cTime.tv_sec - _tl_timer_array[index].tv_sec) * 1000 +
      ((float) (cTime.tv_usec - _tl_timer_array[index].tv_usec))
      / 1000.0));
  }
#endif // LINUX

#ifdef WIN32
#ifdef USE_GETTICK
  return ((unsigned long) GetTickCount () - _tl_timer_array[index]);
#else
  {
    LARGE_INTEGER cTime;

    if (!QueryPerformanceCounter (&cTime)) {
      TL_ERROR_MSG ("tl_timer_get: error: unable to get tick");
      return TL_ERROR;
    }
    return (unsigned long) ((cTime.QuadPart - _tl_timer_array[index].QuadPart) * 1000 / _tl_timer_freq.QuadPart);
  }
#endif
#endif // WIN32
}
